
#include <windows.h>
#include <stdio.h>

int ddeInitialized=0, ddeCodePage=1004, deletebufferwhendone=0, 
	ddeIdInst = 0L, 
	ddeTimeout=10000;	// in milliseconds
size_t buffersize=0;
char *buffer=0;

void ddeInit()
{
	if ( !ddeInitialized )
	{	UINT rc = DdeInitialize(&ddeIdInst, 0, 
			//APPCLASS_STANDARD, 
			APPCMD_CLIENTONLY,
			0L);
		if ( rc != DMLERR_NO_ERROR )
		{	printf("Failed to initialize DDE %i", rc);
		}else
		{	ddeInitialized = 1;
		}
	}
}

void ddeCleanUp()
{
	if (ddeIdInst != 0)
	{	DdeUninitialize(ddeIdInst);
		ddeIdInst = 0;
	}
}

HSZ ddeAtomFromString(const char *s)
{
	HSZ hsz=0;
	if (!ddeIdInst)
		printf("DDE not initialized");
	hsz = DdeCreateStringHandle(ddeIdInst, s, ddeCodePage);
	if ( !hsz )
		printf("Failed to create DDE string");
	return hsz;
}

void ddeFreeString(HSZ hsz)
{
	// DS: Failure to free a string handle might indicate there's
	// some other severe error.
	int ok = (DdeFreeStringHandle(ddeIdInst, hsz) != 0);
	if (!ok)
		printf("Failed to free DDE string handle");
}

char *getBufferAtLeast(size_t bytes)
{
	if (buffersize < bytes)
	{	if (buffer )
			free(buffer);
		buffer = calloc(bytes, 1);
		buffersize = bytes;
	}
	return buffer;
}

HCONV makeConnection(const char *host,
                     const char *server,
                     const char *topic)
{
	HSZ hszServer, hszTopic;
	HCONV hConv;

	hszServer = ddeAtomFromString(server);
	if ( !hszServer )
	{	return 0;
	}
	hszTopic = ddeAtomFromString(topic);
	if ( !hszTopic )
	{	ddeFreeString(hszServer);
		return 0;
	}
	hConv = DdeConnect(ddeIdInst, hszServer, hszTopic,
	(PCONVCONTEXT) NULL);

	ddeFreeString(hszServer);
	ddeFreeString(hszTopic);

	if ( !hConv )
		printf("Failed to create connection to server '%s' on topic '%s'",
			server, topic);
	else
		return hConv;
	return 0;
}

int poke(HCONV con, const char *item, char *data)
{
	DWORD result;
	HSZ item_atom;
	int ok, size=-1, format=CF_TEXT;
	if (size < 0)
		size = strlen(data) + 1;

	item_atom = ddeAtomFromString(item);
	ok = DdeClientTransaction((LPBYTE)data, size, con,
		item_atom, format,
		XTYP_POKE,
		ddeTimeout,
		&result) != 0;

	ddeFreeString(item_atom);
	if (!ok)
		printf("DDE poke request failed");
	return ok;
}

char *request(HCONV conv, const char *item, int *size, int format)
{
	DWORD result, len;

	HSZ atom = ddeAtomFromString(item);
	HDDEDATA returned_data;
	char *data;

	returned_data = DdeClientTransaction(NULL, 0,
		conv, atom, format, XTYP_REQUEST, ddeTimeout, &result);
	if ( !returned_data )
	{	printf("DDE data request failed");
		return 0;
	}
	len = DdeGetData(returned_data, NULL, 0, 0);
	
	data = getBufferAtLeast(len);
	if (!data)
		printf("Buffer too small in request");
	DdeGetData(returned_data, (LPBYTE)data, len, 0);
	DdeFreeDataHandle(returned_data);
	ddeFreeString(atom);
	*size=len;
	return data;
}

int __cdecl main(int argc, char **argv)
{
	int i, n, size;
	char *host="localhost", 
		*service="C:/TEMP/AUCBLOGO.SERVICE", 
		*topic="MYTOPIC",
		item[256]="output ", 
		name[256], *start=0, *end=0, *result=0,
		*request_method=0, *query_string=0, *content_length=0,
		*val=0,
		*env[20]={
			"AUTH_TYPE",
			"CONTENT_TYPE",
			"GATEWAY_INTERFACE",
			"HTTP_*",
			"PATH_INFO",
			"PATH_TRANSLATED",
			"REMOTE_ADDR",
			"REMOTE_HOST",
			"REMOTE_IDENT",
			"REMOTE_USER",
			"SCRIPT_NAME",
			"SERVER_NAME",
			"SERVER_PORT",
			"SERVER_PROTOCOL",
			"SERVER_SOFTWARE",
			0};

	HCONV conv=0;

	ddeInit();
	while (!conv)
	{	conv=makeConnection(host, service, topic);
		Sleep(100);
	}
	if (conv)
	{	content_length=getenv("CONTENT_LENGTH");
		if (content_length)
			while (!poke(conv, "CONTENT_LENGTH", content_length))
				Sleep(100);

		request_method=getenv("REQUEST_METHOD");
		if (request_method)
		{	if (!strcmp(request_method, "POST"))
			{	size=atoi(content_length);
				buffer=getBufferAtLeast(size+1);
				fread(buffer, size, 1, stdin);
				buffer[size]='\0';
				while (!poke(conv, "QUERY_STRING", buffer))
					Sleep(100);
			}else
			{	query_string=getenv("QUERY_STRING");
				if (query_string)
					while (!poke(conv, "QUERY_STRING", query_string))
						Sleep(100);
			}
			while (!poke(conv, "REQUEST_METHOD", request_method))
				Sleep(100);
		}
		i=0;
		while (env[i])
		{	val=getenv(env[i]);
			if (val)
				while (!poke(conv, env[i], val))
					Sleep(100);
			i++;
		}

		strcpy(name, argv[1]);
		start=strrchr(name, '/');
		if (!start)
			start=name;
		else
			start++;
		end=strrchr(start, '.');
		if (!start || !end || start>=end)
			return 1;
		n=end-start;
		strncpy(item+strlen(item), start, n);
		result=0;
		while (!result)
		{	result=request(conv, item, &size, CF_TEXT);
			if (!result)
				Sleep(100);
		}
		if (result)
		{	fwrite(result, size, 1, stdout);
		}
	}
	ddeCleanUp();
	return 0;
}

/*
#include <wx/dde.h>
#include <wx/ipc.h>

class Connection: public wxDDEConnection
{
public:
};

class Client: public wxClient
{
public:
	wxConnectionBase *OnMakeConnection();
};

wxConnectionBase *Client::OnMakeConnection()
{
	Connection *con=new Connection;
	return con;
}

int __cdecl main(int argc, char **argv)
{
	Client *client=new Client;
	Connection *con=0;
	wxString host="localhost", 
		service="C:/temp/aucblogo.service", 
		topic="mytopic";
	while (!con)
		con=(Connection*)client->MakeConnection(host, service, topic);
	if (con)
	{	while (!con->Poke("request_method", getenv("request_method")))
			;
		while (!con->Poke("query_string", getenv("query_string")))
			;
		while (!con->Poke("content_length", getenv("content_length")))
			;
		wxString item="output ", name=argv[1];
		name=name.BeforeLast('.').AfterLast('/');
		char *result=0;
		while (!result)
			result=con->Request(item+name, 0, wxIPC_TEXT);
		if (result)
		{	int l=strlen(result);
			fwrite(result, l, 1, stdout);
		}
		delete con;
	}
	delete client;
	return 0;
}
*/