//___________________________________________________________________
//
//				AVIGen.cpp: implementation of the class AVIGen
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
	#define DOSWIN32
	#include <windows.h>
	#include <vfw.h>
	#include <comdef.h>
	#include <tchar.h>
#endif

#include <wx/string.h>
#include <wx/window.h>

#include <memory.h>
#include <string.h>
#include <stdio.h>

#ifndef WIN32
	#include <avifile-0.7/avifile.h>
	#include <avifile-0.7/avm_fourcc.h>
#endif

#include "Main.h"
#include "AVIGen.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#ifdef WIN32

AVIGen::AVIGen()
: filename(_("noname.avi")), rate(30),
pAVIFile(0), pStream(0), pStreamCompressed(0)
{
	memset(&bih,0,sizeof(BITMAPINFOHEADER));
}

AVIGen::AVIGen(charP filename_, LPBITMAPINFOHEADER lpbih, int rate_): 
	filename(filename_), rate(rate_),
	pAVIFile(0), pStream(0), pStreamCompressed(0)
{
		MakeExtAvi();
		SetBitmapHeader(lpbih);
}

AVIGen::AVIGen(charP filename_, wxWindow *window, int rate_): 
	filename(filename_), rate(rate_),
	pAVIFile(0), pStream(0), pStreamCompressed(0)
{
		MakeExtAvi();
		SetBitmapHeader(window);
}

AVIGen::~AVIGen()
{
	// Just checking that all allocated ressources have been released
	wxASSERT(pStream==0);
	wxASSERT(pStreamCompressed==0);
	wxASSERT(pAVIFile==0);
}

#endif

void AVIGen::MakeExtAvi()
{
	// finding avi
	if (filename.Find(_T("avi"))==wxNOT_FOUND)
	{	filename+=_T(".avi");
	}
}

void AVIGen::SetBitmapHeader(BITMAPINFOHEADER *lpbih)
{
	// checking that bitmap size are multiple of 4
	wxASSERT(lpbih->biWidth%4==0);
	wxASSERT(lpbih->biHeight%4==0);

	// copying bitmap info structure.
	// corrected thanks to Lori Gardi
	memcpy(&bih, lpbih, sizeof(BITMAPINFOHEADER));
}

void AVIGen::SetBitmapHeader(wxWindow *window)
{
	wxASSERT(window!=0);

	////////////////////////////////////////////////
	// Getting screen dimensions
	// Get client geometry 
	wxRect rect=window->GetClientRect(); 
	wxSize size(rect.width, rect.height); 

	/////////////////////////////////////////////////
	// changing size of image so dimension are multiple of 4
	size.x=(size.x/4)*4;
	size.y=(size.y/4)*4;

	// initialize m_bih
	memset(&bih,0, sizeof(BITMAPINFOHEADER));
	// filling bitmap info structure.
	bih.biSize=sizeof(BITMAPINFOHEADER);
	bih.biWidth=size.x;
	bih.biHeight=size.y;
	bih.biPlanes=1;
	bih.biBitCount=24;
	bih.biSizeImage=((bih.biWidth*bih.biBitCount+31)/32 * 4)*bih.biHeight; 
	bih.biCompression=BI_RGB;		//BI_RGB means BRG in reality
}

#ifdef WIN32

int AVIGen::InitEngine(bool newFile, int frame_)
{
	aviError="Ok";
	frame=frame_;

	int wVer = HIWORD(VideoForWindowsVersion());
	if (wVer < 0x010a)
	{	aviError=_("Version of Video for Windows too old (< 0x010A).");
		return 1;
	}

	// Step 1 : initialize AVI engine
	AVIFileInit();

	int createFlag;
//	if (newFile)
		createFlag=OF_WRITE | OF_CREATE;
//	else
//		createFlag=OF_REOPEN;// didn't work (crash!)
	// Step 2 : Open the movie file for writing....
	int err = AVIFileOpen(&pAVIFile,	// Address to contain the new file interface pointer
								 filename,	// Null-terminated string containing
												// the name of the file to open
								createFlag,	// Access mode to use when opening the file. 
								0);			// use handler determined from file extension.
												// Name your file .avi -> very important
	if (err != AVIERR_OK)
	{	aviError=_("AVI Engine failed to initialize. Check filename ")
				  +filename+' ';
		switch (err)
		{
		case AVIERR_BADFORMAT: 
			aviError+=_("The file couldn't be read, "
							 "indicating a corrupt file or an unrecognized format.");
			break;
		case AVIERR_MEMORY:		
			aviError+=_("The file could not be opened "
							 "because of insufficient memory."); 
			break;
		case AVIERR_FILEREAD:
			aviError+=_("A disk error occurred while reading the file."); 
			break;
		case AVIERR_FILEOPEN:		
			aviError+=_("A disk error occurred while opening the file.");
			break;
		case REGDB_E_CLASSNOTREG:		
			aviError+=_("According to the registry, the type of file specified in AVIFileOpen does not have a handler to process it");
			break;
		}
		return err;
	}

	// Fill in the header for the video stream....
	AVISTREAMINFO strHdr; // information for a single stream 

	memset(&strHdr, 0, sizeof(strHdr));
	strHdr.fccType                = streamtypeVIDEO;	// video stream type
	strHdr.fccHandler             = 0;
	strHdr.dwScale                = 1;					// should be one for video
	strHdr.dwRate                 = rate;		    // fps
	strHdr.dwSuggestedBufferSize  = bih.biSizeImage;// Recommended buffer size, 
																	// in bytes, for the stream.
	SetRect(&strHdr.rcFrame, 0,0, bih.biWidth,bih.biHeight);// rectangle for stream

	// Step 3 : Create the stream;
	err = AVIFileCreateStream(pAVIFile,	// file pointer
									&pStream,	// returned stream pointer
									&strHdr);	// stream header

	// Check it succeded.
	if (err != AVIERR_OK)
	{	aviError=_("AVI Stream creation failed. Check Bitmap info.");
		if (err==AVIERR_READONLY)
		{
			aviError+=_(" Read only file.");
		}
		return err;
	}

	// Step 4: Get codec and infos about codec
	AVICOMPRESSOPTIONS opts;
	memset(&opts, 0, sizeof(opts));
	opts.dwFlags = AVICOMPRESSF_VALID 
					 | AVICOMPRESSF_KEYFRAMES 
					 | AVICOMPRESSF_INTERLEAVE
					 | AVICOMPRESSF_DATARATE;
	opts.dwKeyFrameEvery = 10;
	opts.dwInterleaveEvery = 1;
	opts.dwBytesPerSecond = 128*1024;
	AVICOMPRESSOPTIONS FAR * aopts[1] = {&opts};
	// Poping codec dialog
	if (!AVISaveOptions(NULL, 0/*ICMF_CHOOSE_DATARATE*/, 1, 
		&pStream, (LPAVICOMPRESSOPTIONS FAR *) &aopts))
	{
		AVISaveOptionsFree(1,(LPAVICOMPRESSOPTIONS FAR *) &aopts);
		aviError=_("Video output stream canceled by user"); 
		return 2;
	}

	// Step 5:  Create a compressed stream using codec options.
	err = AVIMakeCompressedStream(&pStreamCompressed, pStream, &opts, 0);

	if (err != AVIERR_OK)
	{
		aviError=_("AVI Compressed Stream creation failed.");
		
		switch(err)
		{
		case AVIERR_NOCOMPRESSOR:
			aviError+=_(" A suitable compressor cannot be found.");
				break;
		case AVIERR_MEMORY:
			aviError+=_(" There is not enough memory to complete the operation.");
				break; 
		case AVIERR_UNSUPPORTED:
			aviError+=_("Compression is not supported for this type of data. This error might be returned if you try to compress data that is not audio or video.");
			break;
		}
		return err;
	}

	// releasing memory allocated by AVISaveOptionFree
	err=AVISaveOptionsFree(1,(LPAVICOMPRESSOPTIONS FAR *) &aopts);
	if (err!=AVIERR_OK)
	{
		aviError=_("Error releasing memory");
		return err;
	}

	// Step 6 : sets the format of a stream at the specified position
//	if (newFile)
		err = AVIStreamSetFormat(pStreamCompressed, 
					0,			// position
					&bih,	    // stream format
					bih.biSize +   // format size
					bih.biClrUsed * sizeof(RGBQUAD));

	if (err != AVIERR_OK)
	{
		aviError=_("AVI Compressed Stream format setting failed.");
		return err;
	}
	return err;
}

void AVIGen::ReleaseEngine()
{
	if (pStream)
	{	AVIStreamRelease(pStream);
		pStream=0;
	}

	if (pStreamCompressed)
	{	AVIStreamRelease(pStreamCompressed);
		pStreamCompressed=0;
	}

	if (pAVIFile)
	{	AVIFileRelease(pAVIFile);
		pAVIFile=0;
	}
	// Close engine
	AVIFileExit();
}

int AVIGen::AddFrame(charP pixels)
{
	int err;
	// compress bitmap
	err = AVIStreamWrite(pStreamCompressed,// stream pointer
								frame,				// time of this frame
								1,						// number to write
								pixels,				// image buffer
								bih.biSizeImage,	// size of this frame
								AVIIF_KEYFRAME,	// flags....
								0, 0);
	// updating frame counter
	frame++;
	return err;
}

#else //WIN32

AVIGen::AVIGen(): 
	filename(_("noname.avi")), rate(30), avifile(0)
{
}

AVIGen::AVIGen(charP filename_, wxWindow *window, int rate_): 
	filename(filename_), rate(rate_), avifile(0)
{
	avifile = avm::CreateWriteFile(filename);
	SetBitmapHeader(window);
	//FOURCC codec = fccDIV3;
	fourcc_t codec = fccMP42;
	//FOURCC codec = fccIV32;
	//FOURCC codec = fccCVID;

	vidstr = avifile->AddVideoStream(codec, &bih, rate);
	vidstr->Start();
}

AVIGen::~AVIGen()
{
	vidstr->Stop();
	delete avifile;
}

int AVIGen::InitEngine(bool newFile, int frame_)
{
	return 0;
}

int AVIGen::AddFrame(charP pixels)
{
	avm::CImage *frame=new CImage((uint8_t*)pixels, bih.biWidth, bih.biHeight);
	if (frame)
	{	vidstr->AddFrame(frame);
		delete frame;
	}
	return 0;
}

void AVIGen::ReleaseEngine()
{
}

#endif //WIN32
