#include "stdafx.h"

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>

#include "Main.h"
#include "Thread.h"
#include "Error.h"
#include "Init.h"
#include "File.h"
#include "IntToPtr_.h"
#include "PtrToInt_.h"
#include "String.h"
#include "Object.h"
#include "LogoObject.h"
#include "Pair.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "Int8Array.h"
#include "UInt8Array.h"
#include "FloatArray.h"
#include "Complex.h"
#include "ComplexArray.h"
#include "Bitmap.h"
#include "Color.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

int Array::maxStringLength=65536;

Array::Array(int length_, int origin_): Node(ARRAY), _circcounter(0)
{
	_origin=origin_;
	_length=0;
	_data = (NodeP*) GC_::calloc(length_+1, sizeof(NodeP));
	if (_data) 
		_length=length_;
	flag=ObjectType(0);
}

Array::Array(ArrayP array_): Node(ARRAY), _circcounter(0)
{
	_origin = array_->_origin;
	_length=0;
	_data = (NodeP*) GC_::calloc(array_->_length+1, sizeof(NodeP));
	if (!_data)
	{	_length=0;
		return;
	}
	_length= array_->_length;
	NodePP source = array_->_data, 
		  destination = _data;

	for (int i=_length;  i;  i--)
	{	*destination = *source++;
		check_oldyoung(*destination++);	
	}
	flag=ObjectType(0);
}

void Array::free() 
{
	if (_data)
	{	GC_::free(_data);	_data=0;
	}
}

void Array::inc_gc()
{
	NodePP d=_data;
	for (int i=_length;  i;  i--)
		fifo << *d++;
}

int Array::youngs() 
{
	int youngs=0;
	NodePP d=_data;
	for (int i=_length;  i;  i--)
	{	NodeP child = *d++;
		youngs+=gmark(child);
	}
	return youngs;
}

int Array::youngsPtr(PtrToInt_ &ptrToInt) 
{
	int youngs=0;
	NodePP d=_data;
	for (int i=_length;  i;  i--)
	{	NodeP child = *d++;
		youngs+=gmarkPtr(child, ptrToInt);
	}
	return youngs;
}

void Array::check_oldyoungs()
{
	NodePP d=_data;
	for (int i=_length;  i;  i--) 
		check_oldyoung(*d++);
}

void Array::setPtrs(IntToPtr_ &intToPtr)
{
	NodePP itemptr=_data;

	for (int i=_length;  i;  i--)
	{	*itemptr=(Node*)intToPtr.find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
DataInputStream &Array::read(DataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;

	_length=0;
	_data = (NodeP*) GC_::calloc(sizeof(NodeP), length+1);
	if (!_data)
	{	_length=0;
		return istr;
	}
	_length=length;
	NodePP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Array::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;

	NodePP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;
}

NodeP Array::copy(PtrToInt_ &ptrToInt)
{
	ArrayP arr=new Array(_length, _origin);
	if throwing return Unbound;
	NodePP itemptr=_data, itemptr2=arr->_data;
	for (int i=_length;  i;  i--)
		*itemptr2++ = ObjectP(ptrToInt.find(*itemptr++));
	return arr;
}

StringP Array::string()
{
	char textbuf[65536];
	Thread::This()->stringptr=textbuf;
	Thread::This()->stringlen=maxStringLength-1;
	int width=maxStringLength-5;
	this->print(stringbuffer.file(), width, 0, 0, Int::max);
	*(Thread::This()->stringptr)='\0';
	return new String(textbuf);
}

void Array::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	ThreadP th=Thread::This();
	if (circcounter()==th->circcounter)
	{	::print(strm, width, "...");
		return;
	}else
		setcirccounter(th->circcounter);

	int i=0, wid=0;
	if (width >= 2*_length-10 && _length)
		wid=width/_length;
	NodePP pp=data();

	if (noBrackets <= 0)
		::print(strm, width, '{');

	while (i < _length  &&  width > 0) 
	{	int w=wid;
		if (!*pp)
			::print(strm, w, "[]");
		else if ((*pp)->type==ARRAY)
		{	if (noSpacers <= 0)
			{	th->indent++;
				::print(strm, w, '\t');	
			}
			(*pp)->print(strm, w, depth-1, noBrackets-1, noSpacers-1);	
			if (noSpacers <= 0)
			{	::print(strm, w, endL);
				th->indent--;
				::printIndents(strm, w);
			}
		}
		else
		{	(*pp)->print(strm, w, depth-1, noBrackets-1, noSpacers-1);
		}
		if (++i < _length)
			if (*pp && ((*pp)->type!=ARRAY && (*pp)->isnolist())) 
				::print(strm, w, ' ');
		width-=wid-w;
		pp++;
	}
	if (i < _length) 
		::print(strm, width, "...");
	if (noBrackets <= 0)
		::print(strm, width, '}');

	if (/*backslashes && */(origin() != 1)) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", origin());
		::print(strm, width, org);
	}
	th->setcirccounter(th->circcounter+1);
}

void Array::printBin(FILEP strm)
{
	ThreadP th=Thread::This();
	int width=Int::max;
	if (circcounter()==th->circcounter)
	{	::print(strm, width, "...");
		return;
	}else
		setcirccounter(th->circcounter);
	int i=0, wid=length();
	NodePP pp=data();
	while (i++ < wid) 
	{	if (*pp)
			(*pp)->printBin(strm);
		pp++;
	}
	th->setcirccounter(th->circcounter+1);
}

NodeP Array::contentsTree(	wxArrayString &namearr, 
									wxArrayString &valarr, 
									wxArrayPtrVoid &objarr, 
									ObjectType want, 
									ObjectType dont,
									ObjectType must, 
									int &count, int level)
{
	objarr[count]=this;
	if (!expanded())
	{	StringP str=string();
		if (str  &&  str->is.string)
		{	valarr[count]=str->noFormat()->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
		}
		return this;
	}
	ThreadP th=Thread::This();
	if (circcounter()==th->circcounter)
	{	valarr[count]="...";
		return this;
	}else
		setcirccounter(th->circcounter);
	NodeP contentsList=0;
	NodePP dp=data();
	for (int i=0; i < _length;  i++)
	{	NodeP nd=(*dp++);
		if (nd)
		{	int lev=level;
			if (nd->type==OBJECT)
			{	nd=ObjectP(nd)->name();
				lev--;
			}else if (nd->type==LOGOOBJECT)
			{	nd=LogoObjectP(nd)->name();
				lev--;
			}else
				lev++;
			namearr[count]=wxString(' ',level-1);
			if (nd && nd->expandable())
				if (nd->expanded())
					namearr[count]+='-';
				else
					namearr[count]+='+';
			namearr[count] << i+_origin;
			objarr[count]=nd;
			if (nd->expanded())
			{	if (nd->type==LOGOOBJECT)
				{	valarr[count]=LogoObjectP(nd)->name()->ptr();
					incContentsArrays(namearr, valarr, objarr, count);
				}else
				{	valarr[count]=nd->_typeof()->ptr();
					incContentsArrays(namearr, valarr, objarr, count);
				}
			}
			contentsList=nd->contentsTree(
				namearr, valarr, objarr, want, dont, must, count, lev);
		}else
		{	valarr[count]="[]";
			incContentsArrays(namearr, valarr, objarr, count);
		}
	}
	th->setcirccounter(th->circcounter+1);
	if (!contentsList)
		return 0;
	return this;
}

NodeP Array::compare(NodeP nd)
{
	if (nd->type != ARRAY) 
		return Unbound;
	if (this==nd)
		return 0;
	ThreadP th=Thread::This();
	if (circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);

	ArrayP b=ArrayP(nd);
	NodePP dp=data(), dpb=b->data();

	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{	NodeP c=(*dp++)->compare(*dpb++);
		if (c)
		{	th->setcirccounter(th->circcounter+1);
			return c;
		}
		if throwing return _Minus_One;
	}
	th->setcirccounter(th->circcounter+1);
	if (!i & !j)
		return 0;
	if (i)
		return _One;
	return _Minus_One;
}

int  Array::count() {return length();}

int  Array::_sizeof() 
{	int size=0;
	ThreadP th=Thread::This();
	if (circcounter()==th->circcounter)
		return size;
	else
		setcirccounter(th->circcounter);
	NodePP d=_data;
	for (int i=_length;  i;  i--)
	{	size+=(*d++)->_sizeof();
	}
	th->setcirccounter(th->circcounter+1);
	return size;
}

bool Array::equalP(NodeP b)
{	if (b->type != ARRAY) 
		return false;

	if (this==b)
		return true;

	ThreadP th=Thread::This();
	if (circcounter()==th->circcounter)
		return circcounter()==b->circcounter();
	else
		setcirccounter(th->circcounter);

	ArrayP ba=ArrayP(b);
	NodePP dp=data(), dpb=ba->data();

	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
	{	if (!(*dp))
		{	if (*dpb)
			{	th->setcirccounter(th->circcounter+1);
				return false;
			}
			dp++;
			dpb++;
		}else if (!(*dp++)->equalP(*dpb++))
		{	th->setcirccounter(th->circcounter+1);
			return false;
		}
	}
	th->setcirccounter(th->circcounter+1);
	if (!i & !j)
		return true;

	return false;
}

NodeP Array::deepCopy()
{	ArrayP result=new Array(length(), origin());
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	for (int i=length();  i;  i--)
	{	NodeP s=*source++;
		if (s)
			*dest = s->deepCopy();
		if throwing return Unbound;
		dest++;
	}
	return result;
}

NodeP Array::first_() 
{	if (length() >= 1) 
		return *(ArrayP(this)->data());
	else
		return 0;
}

NodeP Array::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	return 0;
	}
	ArrayP result=new Array(l, o);
	if throwing return Unbound;
	NodePP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Array::firsts()
{	int l=length(), o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l, o);
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	while (l--)
	{	if (*source)
			*dest++=(*source++)->first_();
		else
			*dest++=0;
		if throwing return Unbound;
	}
	return result;
}

NodeP Array::butFirsts()
{	int l=length(), o=origin();
	if (l < 1)
	{	return new Array(0, o);
	}
	ArrayP result=new Array(l, o);
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	while (l--)
	{	NodeP nd=(*source++)->a_bfable(1);
		if throwing return Unbound;
		*dest++=nd->butFirst();
	}
	return result;
}

NodeP Array::last()
{	if (length())
		return *(data()+length()-origin());
	error(BAD_DATA, this); return Unbound;
}

NodeP Array::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l, o);
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Array::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l+1, o);
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	*dest++=nd;
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Array::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l+1, o);
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=nd;
	return result;
}

double Array::_min(double d)
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	NodePP dp=_data;
	for (int i=0;  i<_length;  i++, dp++)
	{	if (*dp)
		{	double dnew = (*dp)->_min(d);
			if throwing return Float::quietNaN;
			if (dnew != d)
			{	d=dnew;
				th->minimumPosition=i;
			}
		}
	}
	th->setcirccounter(th->circcounter+1);
	return d;
}

double Array::_max(double d)
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	NodePP dp=_data;
	for (int i=0;  i<_length;  i++, dp++)
	{	if (*dp)
		{	double dnew = (*dp)->_max(d);
			if throwing return Float::quietNaN;
			if (dnew != d)
			{	d=dnew;
				th->maximumPosition=i;
			}
		}
	}
	th->setcirccounter(th->circcounter+1);
	return d;
}

NodeP Array::Norm()
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	double d=0;  NodePP dp=_data;
	if (*dp && (*dp)->is.aggregate)
	{	ArrayP result=new Array(_length, _origin);
		if throwing return Unbound;
		NodePP rp=result->_data;
		for (int i=_length;  i;  i--)
		{	if (*dp)
				(*rp++) = (*dp++)->Norm();
			else
				(*rp++)=0;
			if throwing return Unbound;
		}
		th->setcirccounter(th->circcounter+1);
		return result;
	}
	else
	{	for (int i=_length;  i;  i--)
		{	double h = (*dp++)->Norm()->getnumber();
			if throwing return Unbound;
			d+= h*h;
		}
		th->setcirccounter(th->circcounter+1);
		return new Float(::sqrt(d));
	}
}

NodeP Array::maxNorm()
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	double d=0;  NodePP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++)->maxNorm()->getnumber();
		if (throwing) return Unbound;
		double ha=Float::abs(h);
		if (ha > d) d = ha;
	}
	th->setcirccounter(th->circcounter+1);
	return new Float(d);
}

NodeP Array::resize(int newsize)
{	ArrayP r=new Array(newsize);
	if throwing return Unbound;
	NodePP d=data();
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0 && d[i+1])
				{	NodeP diff = d[i+1]->sub(d[i]), slope=0;
					if throwing return Unbound;
					IntP dii=new Int(di);
					if throwing return Unbound;
					slope=diff->div(dii);

					for (int ii=di-1;  ii;  ii--)
					{	IntP iii=new Int(ii);
						if throwing return Unbound;
						NodeP nd=slope->mul(iii);
						if throwing return Unbound;
						r->_data[ni+ii] = d[i]->add(nd);
					}
				}
				oi=ni;
			}
		}
	}
	return r;
}

NodeP Array::saturateAbove(double limit) 
{
	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	ArrayP result=new Array(length());
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	if (*source)
		{	*dest++= (*source++)->saturateAbove(limit);
			if throwing return Unbound;
		}else
			*dest++=0;
	}
	th->setcirccounter(th->circcounter+1);
	return result;
}

NodeP Array::saturateBelow(double limit) 
{
	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	ArrayP result=new Array(length());
	if throwing return Unbound;
	NodePP source=data(), dest=result->data();
	if (*source && (*source)->is.aggregate)
	{	for (int i=length(); i>0; i--)
		{	if (*source)
			{	*dest++= (*source++)->saturateBelow(limit);
				if throwing return Unbound;
			}else
				*dest++=0;
		}
	}else
		for (int i=length(); i>0; i--)
		{	double s=(*source++)->getnumber();
			*dest++= new Float(s<limit? limit : s);
			if throwing return Unbound;
		}
	th->setcirccounter(th->circcounter+1);
	return result;
}

NodeP Array::item(int i)
{	if (i >= 0  &&  i < _length) 
		return _data[i];
	IntP ip=new Int(i);
	if throwing return Unbound;
	error(BAD_DATA, ip); return Unbound;
}

NodeP Array::item(NodeP index, NodeP result)
{
	if (index->type==INT_)
	{	int i=IntP(index)->int_-_origin;
		if (i >= 0  &&  i < _length) 
			return _data[i];
	}
	else if (index->type==ARRAY)
	{	ArrayP idx=ArrayP(index);
		int l=idx->length();
		ArrayP r=new Array(l);
		if throwing return Unbound;
		NodePP d=r->data(), i=idx->data();
		for (;  l;  l--)
		{	*d++=item(*i++, 0);
			if throwing return Unbound;
		}
		return r;
	}
	else if (index->type==INTARRAY)
	{	IntArrayP idx=IntArrayP(index);
		int l=idx->length();
		ArrayP r=new Array(l);
		if throwing return Unbound;
		NodePP d=r->data();
		intP i=idx->data();
		for (;  l;  l--)
		{	*d++=item(*i++);
			if throwing return Unbound;
		}
		return r;
	}
	else if (index->islist() && result->islist())
	{	PairP i=PairP(index), r=PairP(result);
		while (i && r)
		{	NodeP ret=item(i->first(), r->first());
			if throwing return Unbound;
			r->setfirst( ret );
			next(i);
			next(r);
		}
		if (!i && !r && !throwing) return result;
	}
//	ITEM_if_index_islist
	else if (index->is.string)
		return item(StringP(index)->number(), 0);
	error(BAD_DATA, index); return Unbound;
}

NodeP Array::items(NodeP start, NodeP end)
{	if (!start)
	{	error(BAD_DATA, start, 1); return Unbound;
	}
	if (!end)
	{	error(BAD_DATA, end, 1); return Unbound;
	}
	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	ArrayP newar=new Array(e-s+1);
					if throwing return Unbound;
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP Array::mdItem(NodeP index)
{
	NodeP obj=this;
	if (index->islist())
	{	PairP indexlist=PairP(index);
		while (indexlist)
		{	NodeP i=indexlist->first();
			if (!i)
			{	error(BAD_DATA, index); 
				return Unbound;
			}
			if (!obj)
			{	error(BAD_DATA, this);
				return Unbound;
			}
			obj=obj->item(i, 0);
			next(indexlist);
		}
		return obj;
	}
	error(BAD_DATA, this);
	return Unbound;
}

NodeP Array::setitem(int index, NodeP newitem)
{	index-=origin();
	if ((index < 0 || index >= length()) && !throwing) 
	{	IntP indexNode=new Int(index);
		if throwing return Unbound;
		error(BAD_DATA, indexNode);
	}
	if (!throwing) 
	{	_data[index] = newitem;
		check_oldyoung(newitem);
	}
	return Unbound;
}

NodeP Array::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (safe) 
	{	if (circular(newitem)) 
			error(BAD_DATA, newitem, 3);
	}
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, index);

		if (!throwing) 
		{	_data[i] = newitem;
			check_oldyoung(newitem);
		}
	}
	return Unbound;
}

NodeP Array::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->_length <= _length)
				{	NodePP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
					{	*d = *s++;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Int(*s++);
						if throwing return Unbound;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	int16P s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Int16(*s++);
						if throwing return Unbound;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT8ARRAY)
			{	Int8ArrayP n=Int8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	int8P s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Int8(*s++);
						if throwing return Unbound;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==UINT8ARRAY)
			{	UInt8ArrayP n=UInt8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	uint8P s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new UInt8(*s++);
						if throwing return Unbound;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)
			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Float(*s++);
						if throwing return Unbound;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==COMPLEXARRAY)
			{	ComplexArrayP n=ComplexArrayP(newitems);
				if (i+n->length() <= _length)
				{	complexdP s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Complexd(*s++);
						if throwing return Unbound;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}

NodeP Array::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	NodePP d=_data;   int l=_length;

	while (--l >= 0 && !throwing) 
		if (subobj->equalP(*d++))
			return _True;
	
	return _False;
}

NodeP Array::reverse()
{
	ArrayP result=new Array(_length);
	if throwing return Unbound;

	NodePP d=_data, rev=result->_data+_length-1;

	for (int i=_length; i>0; i--)
		*rev-- = *d++;
	return result;
}

NodeP Array::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	ArrayP val = new Array(length(), origin());
	if throwing return Unbound;
	NodePP
		a= data(),
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
	{	*b = *a++;
		check_oldyoung(*b++);	
	}
	b=val->data();
	for (i=places;  i;  i--)
	{	*b = *a++;
		check_oldyoung(*b++);	
	}
	return val;
}

NodeP Array::shuffle()
{	ArrayP result=new Array(ArrayP(this));
	if throwing return Unbound;
	NodePP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		NodeP tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}

NodeP Array::upper() 
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	ArrayP newar=new Array(ArrayP(this));
	if throwing return Unbound;
	NodePP d=newar->_data;
	for (int i=_length;  i;  i--)
	{	*d=(*d)->upper();
		d++;
		if throwing return Unbound;
	}
	th->setcirccounter(th->circcounter+1);
	return newar;
}

NodeP Array::lower() 
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return 0;
	else
		setcirccounter(th->circcounter);
	ArrayP newar=new Array(ArrayP(this));
	if throwing return Unbound;
	NodePP d=newar->_data;
	for (int i=_length;  i;  i--)
	{	*d=(*d)->lower();
		d++;
		if throwing return Unbound;
	}
	th->setcirccounter(th->circcounter+1);
	return newar;
}

PairP Array::list()
{
	NodePP d = &_data[_length];
	PairP p=0;
	if (!d) return PairP(Unbound);

	for (int i=_length;  i;  i--)
	{	p = new Pair(*--d, p);
		if throwing return PairP(Unbound);
	}
	return p;
}

ArrayP Array::array(int origin) 
{
	return new Array(ArrayP(this));
}

IntArrayP Array::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	if throwing return IntArrayP(Unbound);
	intP d = intarray->data();
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	NodeP h=*s;
		switch (type_of(h))
		{	case INT_:	*d++ =             IntP(h)->int_;	break;
			case FLOAT_:*d++ = Int::clip(FloatP(h)->float_);break;
			case NINT16:	*d++ =           Int16P(h)->int_;	break;
			case NINT8:	*d++ =            Int8P(h)->int_;	break;
			case STRING:if (h->numberP())
								*d++ = Int::clip(h->getnumber());
							else
							{	error(BAD_DATA, this); return IntArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return IntArrayP(Unbound);
		}
		s++;
	}
	return intarray;
}

Int16ArrayP Array::int16array(int origin) 
{
	Int16ArrayP int16array = new Int16Array(_length, _origin);
	if throwing return Int16ArrayP(Unbound);
	int16P d = int16array->data();
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	NodeP h=*s;
		switch (type_of(h))
		{	case INT_:	*d++ =   Int16::clip(IntP(h)->int_);	break;
			case FLOAT_:*d++ = Int16::clip(FloatP(h)->float_);	break;
			case NINT16:	*d++ =             Int16P(h)->int_;		break;
			case NINT8:	*d++ =              Int8P(h)->int_;		break;
			case STRING:if (h->numberP())
								*d++ = Int16::clip(h->getnumber());
							else
							{	error(BAD_DATA, this); return Int16ArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return Int16ArrayP(Unbound);
		}
		s++;
	}
	return int16array;
}

FloatArrayP Array::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	if throwing return FloatArrayP(Unbound);
	doubleP d = floatarray->data();
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	NodeP h=*s;
		switch (type_of(h))
		{	case INT_:	*d++ =   IntP(h)->int_;	break;
			case FLOAT_:*d++ = FloatP(h)->float_;break;
			case NINT16:	*d++ = Int16P(h)->int_;	break;
			case NINT8:	*d++ =  Int8P(h)->int_;	break;
			case STRING:if (h->numberP())
								*d++ = h->getnumber();
							else
							{	error(BAD_DATA, this); return FloatArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return FloatArrayP(Unbound);
		}
		s++;
	}
	return floatarray;
}

ComplexArrayP Array::complexarray(int origin) 
{
	ComplexArrayP carray = new ComplexArray(_length, _origin);
	if throwing return ComplexArrayP(Unbound);
	complexdP d = carray->data();
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	NodeP h=*s;
		switch (type_of(h))
		{	case INT_:	*d++ =   IntP(h)->int_;	break;
			case FLOAT_:*d++ = FloatP(h)->float_;break;
			case COMPLEX:*d++ = ComplexdP(h)->complex_;break;
			case NINT16:	*d++ = Int16P(h)->int_;	break;
			case NINT8:	*d++ =  Int8P(h)->int_;	break;
			case STRING:if (StringP(h)->complexNumberP())
								StringP(h)->getcomplex(d++);
							else
							{	error(BAD_DATA, this); return ComplexArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return ComplexArrayP(Unbound);
		}
		s++;
	}
	return carray;
}

BitmapP Array::bitmap(int width)
{
	width=length();
	ArrayP col=(*data())->a_array(1);
	if throwing return BitmapP(Unbound);
	int height=col->length();
	BitmapP bm=new Bitmap(width, height, width, height, 0);
	if throwing return BitmapP(Unbound);
	bm->pixels=ColorP(GC_::malloc(width*height*sizeof(Color)));
	if throwing return BitmapP(Unbound);
	ColorP d=bm->pixels;
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	ArrayP col=ArrayP(*s++);
		NodePP cold=col->_data;
		if (!cold) return BitmapP(Unbound);
		for (int j=col->_length;  j;  j--)
		{	if (!*cold)
			{	error(BAD_DATA, this);	return BitmapP(Unbound);
			}
			*d++ = int((*cold++)->getnumber());
		}
	}
	return bm;
}

void Array::setPixel(NodeP color)
{	NodePP d=data();
	int i=length();
	if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
	{	NodePP c=ArrayP(color)->data();
		while (i--)
		{	(*d++)->setPixel(*c++);
			if throwing return;
		}
	}else 
		while (i--)
		{	(*d++)->setPixel(color);
			if throwing return;
		}
}

void Array::setPixelXY(NodeP yin, NodeP color)
{	if (type_of(yin) == ARRAY)
	{	NodePP d=data();
		int i=length();
		ArrayP y=ArrayP(yin);
		if (y->length() != i)
		{	error(BAD_DATA, yin, 2);
			return;
		}
		NodePP dy=y->data();
		if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
		{	NodePP c=ArrayP(color)->data();
			while (i--)
			{	(*d++)->setPixelXY(*dy++, *c++);
				if throwing return;
			}
		}else 
			while (i--)
			{	(*d++)->setPixelXY(*dy++, color);
				if throwing return;
			}
	}
	else if (type_of(yin) == INT_ || type_of(yin) == FLOAT_)
	{	NodePP d=data();
		int i=length();
		if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
		{	NodePP c=ArrayP(color)->data();
			while (i--)
			{	(*d++)->setPixelXY(yin, *c++);
				if throwing return;
			}
		}else 
			while (i--)
			{	(*d++)->setPixelXY(yin, color);
				if throwing return;
			}
	}
	else
	{	error(BAD_DATA, yin, 2);
		return;
	}
}

void Array::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{	if (type_of(yin) == ARRAY)
	{	NodePP d=data();
		int i=length();
		ArrayP y=ArrayP(yin);
		if (y->length() != i)
		{	error(BAD_DATA, yin, 2); return;
		}
		if (type_of(zin) != ARRAY)
		{	error(BAD_DATA, zin, 3); return;
		}
		ArrayP z=ArrayP(zin);
		if (z->length() != i)
		{	error(BAD_DATA, zin, 3); return;
		}
		NodePP 
			dy=y->data(), 
			dz=z->data();
		if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
		{	NodePP c=ArrayP(color)->data();
			while (i--)
			{	(*d++)->setPixelXYZ(*dy++, *dz++, *c++);
				if throwing return;
			}
		}else 
			while (i--)
			{	(*d++)->setPixelXYZ(*dy++, *dz++, color);
				if throwing return;
			}
		return;
	}
	else if (type_of(yin) == INT_ || type_of(yin) == FLOAT_)
	{	if (type_of(zin) == INT_ || type_of(zin) == FLOAT_)
		{	NodePP d=data();
			int i=length();
			if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
			{	NodePP c=ArrayP(color)->data();
				while (i--)
				{	(*d++)->setPixelXYZ(yin, zin, *c++);
					if throwing return;
				}
			}else 
				while (i--)
				{	(*d++)->setPixelXYZ(yin, zin, color);
					if throwing return;
				}
			return;
		}
		else
		{	error(BAD_DATA, zin, 3); return;
		}
	}
	error(BAD_DATA, yin, 2); return;
}

void Array::setXYZ(NodeP ynode, NodeP znode)
{	int l=length();
	if (ynode->type==ARRAY && ArrayP(ynode)->length()==l)
	{	if (znode)
			if (znode->type==ARRAY && ArrayP(znode)->length()==l)
			{	NodePP x=data(), 
					y=ArrayP(ynode)->data(),
					z=ArrayP(znode)->data();
				while (l--)
				{	if (*x)
						(*x++)->setXYZ(*y++, *z++);
					if throwing return;
				}
			}
			else 
			{	error(BAD_DATA, znode, 3);	return;
			}
		else
		{	NodePP x=data(), 
				y=ArrayP(ynode)->data();
			while (l--)
			{	if (*x)
					(*x++)->setXYZ(*y++, 0);
				if throwing return;
			}
		}
	}
	else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}

bool Array::circular2()
{
	ThreadP th=Thread::This();
	if (circcounter()==th->circcounter)
		return true;
	else
		setcirccounter(th->circcounter);
	NodePP p=data();
	for (int i=_length;  i;  i--)
	{	if (*p && (*p)->circular2()) 
			return true;
		if throwing return false;
	}
	th->setcirccounter(th->circcounter+1);
	return false;
}

void Array::combSortInPlace()
{
	double gap=_length; //initialize gap size
	int swaps=0; //see bubblesort for an explanation
    
	while (!(gap<=1 && swaps==0))
	{	//update the gap value for a next comb
		if (gap > 1)
		{	gap/=1.3;
			if (gap==10 || gap==9)
				gap=11;
		}
		int i=0;
		//a single "comb" over the input list
		while (i+gap<_length)
		{	int j=int(i+gap);
			if (_data[i] && _data[i]->compare(_data[j]) == _One)
			{	NodeP tmp=_data[i];
				_data[i]=_data[j];
				_data[j]=tmp;
			}
			i++;
		}
	}
}

NodeP Array::combSort()
{
	ArrayP a=new Array(this);
	if throwing return Unbound;
	a->combSortInPlace();
	return a;
}
