
#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <m_apm.h>

#include "Main.h"
#include "Error.h"
#include "File.h"
#include "String.h"
#include "Int.h"
#include "Float.h"
#include "Complex.h"
#include "BigFloat.h"
#include "Object.h"
#include "Init.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

MAPM BigFloat::deg2rad = MAPM(MM_PI)/180;

BigFloat::BigFloat(StringP s):
	Node(BIGFLOAT)
{	if (s)
		float_=MAPM(s->ptr());
	else
		float_=0;
}

void BigFloat::free() {float_.~MAPM();}

DataInputStream &BigFloat::read(DataInputStream &istr)
{	wxString s;
	istr >> s;
	float_=s;
	return istr;
}
wxDataOutputStream &BigFloat::write(wxDataOutputStream &ostr)
{	ostr << type;
	wxString s('\0', MM_cpp_min_precision+20);
	float_.toString(charP(s.c_str()), MM_cpp_min_precision);
	return ostr << s;
}

NodeP BigFloat::copy(PtrToInt_ &ptrToInt)
{	return new BigFloat(float_);
}

StringP BigFloat::string()
{
	wxString s('\0', MM_cpp_min_precision+20);
	float_.toString(charP(s.c_str()), MM_cpp_min_precision);
	s.Trim();
	return new String(s);
}

void BigFloat::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, float_);
}

void BigFloat::printBin(FILEP strm) 
{
	::printBin(strm, float_);
}

double BigFloat::getnumber()
{
	double d=float_.to_double();
	if (m_apm_error)
	{	m_apm_reset_error();
		error(BAD_DATA, this);
		return Float::quietNaN;
	}
	return d;
}

int BigFloat::_sizeof() {return sizeof(M_APM)+1+(MM_cpp_min_precision>>1);}

double BigFloat::_min(double d) {return float_ < d ?  float_.to_double() : d;}
double BigFloat::_max(double d) {return float_ > d ?  float_.to_double() : d;}

NodeP BigFloat::Norm()
{	if (float_ >=0) return this;
	else return new BigFloat(-float_);
}

NodeP BigFloat::maxNorm()
{	if (float_ >=0) return this;
	else return new BigFloat(-float_);
}

NodeP BigFloat::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}

NodeP BigFloat::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}

NodeP BigFloat::reverse()
{	StringP s=new String(float_.IntegerStringSize());
	if throwing return Unbound;
	float_.toIntegerString(s->ptr());
	s=new String(s->ptr());
	if throwing return Unbound;
	s=StringP(s->reverse());
	if throwing return Unbound;
	return new BigFloat(s);
}

NodeP BigFloat::upper() {return string()->upper();}
NodeP BigFloat::lower() {return string();}
PairP BigFloat::list() {return string()->list();}

void BigFloat::init() 
{
	m_apm_cpp_precision(30);
//	deg2rad=MAPM(MM_PI)/180;
	deg2rad=MAPM(MM_DEG2RAD)+0;
	BigFloatP bfpi=new BigFloat(MAPM(MM_PI));
	if throwing return;
	_BigFloatPi->setvarPermanent(bfpi);
}

MAPM BigFloat::Remainder(MAPM x, MAPM d) 
{	if (d==0.0) 
		return Float::NaN;
	return x%d;
}

MAPM BigFloat::mod(MAPM x, MAPM d) 
{
	MAPM m=x % d;
	if ((m < 0) != (d < 0)) return m+d;
	return m;
}

void print(FILEP strm, int &width, MAPM d)
{
	wxString s;
	if (d.is_integer())
	{	s.Alloc(d.IntegerStringSize());
		d.toIntegerString(charP(s.c_str()));
	}
	else
	{	s.Alloc(d.StringSize(-1));
		d.toString(charP(s.c_str()), -1);
	}
	char ch;
	charP cp=(charP)s.c_str();
	while ((ch=*cp++) !='\0' && width > 0)
		print(strm, width, ch);
}

void printBin(FILEP strm, MAPM d)
{
	int width=Int::max;
	print(strm, width, d);
}

