//___________________________________________________________________
//
//                          Class Bitmap
//
//                          Definitions
//___________________________________________________________________

#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <stdio.h>

#include <wx/thread.h>
#include <wx/image.h>

#include "Main.h"
#include "Thread.h"
#include "Error.h"
#include "Color.h"
#include "File.h"
#include "Pair.h"
#include "String.h"
#include "Int.h"
#include "Object.h"
#include "Array.h"
#include "IntArray.h"
#include "Bitmap.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

Bitmap::Bitmap(int width_,int height_, 
					int winx_ ,int winy_  , Color *pixels_): 
	Node(BITMAP_) 
{
	width=width_;	height=height_; 
	winx=winx_; winy=winy_; 
	pixels=pixels_;
}

Bitmap::Bitmap(wxBitmap &bmp):
	Node(BITMAP_)
{
	pixels=0;
	winx=width=bmp.GetWidth();
	winy=height=bmp.GetHeight();
	wxImage image=bmp.ConvertToImage();
	if (!image.IsOk())
	{	error(BAD_DATA, 0); return;
	}
	int size=width*height; //sizeof(Color);
	unsigned char *pix=image.GetData(), *a=image.GetAlpha(), *p;
	pixels=ColorP(GC_::malloc(size*sizeof(Color)));
	if throwing return;
	ColorP c=pixels;
	for (int y=height-1; y>=0; y--)
	{	p=&pix[width*3*y];
		for (int i=width; i; i--)
		{	c->r=*p++;
			c->g=*p++;
			c->b=*p++;
			c->a=0xff;
			c++;
		}
	}
	if (a)
	{	c=pixels;
		for (int i=size; i; i--)
		{	c->a=*a++;
			c++;
		}
	}
}

void Bitmap::free()
{
	if (pixels)
	{	GC_::free(pixels); pixels=0;
	}
}

TextInputStream &Bitmap::read(TextInputStream &istr) 
{	
	istr >> width >> height >> winx >> winy;
	int size=winx*winy;
	pixels=(ColorP)GC_::malloc(size*sizeof(Color));
	if (!pixels)
	{	width=height=winx=winy=0;
		return istr;
	}
	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	istr >> *itemptr;
		itemptr++;
	}
	return istr;
}
DataInputStream &Bitmap::read(DataInputStream &istr) 
{	
	istr >> width >> height >> winx >> winy;
	int size=winx*winy;
	pixels=(ColorP)GC_::malloc(size*sizeof(Color));
	if (!pixels)
	{	width=height=winx=winy=0;
		return istr;
	}
	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	istr >> *itemptr;
		itemptr++;
	}
	return istr;
}

TextOutputStream &Bitmap::write(TextOutputStream &ostr) 
{
	ostr << int(type) << ' ' << width << ' ' << height << ' ' 
		<< winx << ' ' << winy << ' ';
	int size=winx*winy;
	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	ostr << *itemptr;
		itemptr++;
	}
	ostr << ' ';
	return ostr;
}
wxDataOutputStream &Bitmap::write(wxDataOutputStream &ostr) 
{
	ostr << type << width << height << winx << winy;
	int size=winx*winy;

	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	ostr << *itemptr;
		itemptr++;
	}
	return ostr;

}

StringP Bitmap::string()
{
	char s[32];
	sprintf(s, "Bitmap [%i %i]", width, height);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void Bitmap::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, "Bitmap [");
	::print(strm, width, int(this->width)); 
	::print(strm, width, ' ');
	::print(strm, width, int(height)); 
	::print(strm, width, ']');
}

NodeP Bitmap::contentsTree(wxArrayString &namearr, 
									wxArrayString &valarr, 
									wxArrayPtrVoid &objarr, 
									ObjectType want, 
									ObjectType dont,
									ObjectType must, 
									int &count, int level)
{
	objarr[count]=this;
	if (!expanded())
	{	StringP str=string();
		if (str  &&  str->is.string)
		{	valarr[count]=str->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
		}
		return this;
	}
	ColorP cp=pixels;
	for (int y=0; y < int(winy);  y++)
	{	namearr[count]=wxString(' ',level-1);
		namearr[count] << y;
		objarr[count]=this;
		wxString line=wxEmptyString;
		for (int x=0; x < int(winx); x++)
		{	Color c=(*cp++);
			wxString col;
			col.Printf("%02x%02x%02x%02x ", c.a, c.b, c.g, c.r);
			line+=col;
		}
		valarr[count]=line;
		incContentsArrays(namearr, valarr, objarr, count);
	}
	return this;
}

ArrayP Bitmap::array(int origin) 
{
	ArrayP array = new Array(winx, 0);
	if throwing return ArrayP(Unbound);
	NodePP d = array->data();
	ColorP s=pixels;

	for (int x=winx;  x;  x--)
	{	ArrayP col = new Array(winy, 0);
		if throwing return ArrayP(Unbound);
		*d++=col;
		NodePP cd=col->data();

		for (int y=winy;  y;  y--)
		{	*cd++=new Int(*s++);
			if throwing return ArrayP(Unbound);
		}
	}
	return array;
}

IntArrayP Bitmap::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(winx*winy, 0);
	if throwing return IntArrayP(Unbound);
	intP d = intarray->data();
	ColorP s=pixels;

	for (int i=winx*winy;  i;  i--)
		*d++ = *s++;

	return intarray;
}

void Bitmap::makeTransparent(Color c)
{
	for (unsigned int i=0; i < winx*winy; i++)
		if (pixels[i]==c)
			pixels[i].a=0;
}

void Bitmap::setPixel(int x, int y, Color c)
{
	if (x >= 0 && x < int(winx) && y >= 0 && y < int(winy))
		pixels[x+y*winx]=c;
}

Color Bitmap::Pixel(int x, int y)
{
	if (x >= 0 && x < int(winx) && y >= 0 && y < int(winy))
		return pixels[x+y*winx];
	return 123456789;
}

void Bitmap::fill(int x, int y, Color fillc, Color rim)
{
	if (x < 0 || x >= int(winx) || y < 0 || y >= int(winy))
		return;
	fillRecursiv(x,y, fillc, rim);
}

void Bitmap::fillRecursiv(int x, int y, Color fillc, Color rim)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*winx];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (c==rim || c==fillc) 
			break;
//		wxASSERT(p >= pixels);
//		wxASSERT(p < &pixels[winx*winy]);
		*p-- = fillc;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
	if (leftLimit > 0) 
		leftLimit--;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*winx];
	for (i=x+1; i < int(winx); i++) 
	{	c=*p;
		if (c==rim || c==fillc)
			break;
//		wxASSERT(p < &pixels[winx*winy]);
		*p++ = fillc;
		rightLimit=i;
	}
	if (rightLimit < int(winx-1))
		rightLimit++;
	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y-1)*winx];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fillc) 
				{	fillRecursiv(i, y-1, fillc, rim);
					lastBorder=false;
				}
			}
			else if (c==rim) 
				lastBorder=true;
			p++;
		}
	}
	// Below
	if (y < int(winy)-1) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y+1)*winx];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fillc) 
				{	fillRecursiv(i, y+1, fillc, rim);
					lastBorder=false;
				}
			} 
			else if (c==rim) 
				lastBorder=true;
			p++;
		}
	}
}

void Bitmap::fillSurface(int x, int y, Color fillc, Color rim)
{
	if (x < 0 || x >= int(winx) || y < 0 || y >= int(winy))
		return;
	fillSurfaceRecursiv(x,y, fillc, rim);
}

void Bitmap::fillSurfaceRecursiv(int x, int y, Color fillc, Color rim)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*winx];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (c==rim || c==fillc) 
			break;
//		wxASSERT(p >= pixels);
		*p-- = fillc;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
	if (leftLimit > 0) 
		leftLimit--;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*winx];
	for (i=x+1; i < int(winx); i++) 
	{	c=*p;
		if (c==rim || c==fillc)
			break;
//		wxASSERT(p < &pixels[pixelsSize]);
		*p++ = fillc;
		rightLimit=i;
	}
	if (rightLimit < int(winx))
		rightLimit++;
	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y-1)*winx];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fillc) 
				{	fillSurfaceRecursiv(i, y-1, fillc, rim);
					lastBorder=false;
				}
			}
			else if (c==rim) 
				lastBorder=true;
			p++;
		}
	}
	// Below
	if (y < int(winy)-1) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y+1)*winx];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fillc) 
				{	fillSurfaceRecursiv(i, y+1, fillc, rim);
					lastBorder=false;
				}
			} 
			else if (c==rim) 
				lastBorder=true;
			p++;
		}
	}
}

void Bitmap::bitPaste(BitmapP bm, int x, int y)
{
	int	winxi=bm->winx,
			winyi=bm->winy;
	Color c;
	for (int yi=0; yi < winyi; yi++)
	{	int	wyi=yi*winxi,
				wyyi=(y+yi)*winx;
		for (int xi=0; xi < winxi; xi++)
		{	c=bm->pixels[xi+wyi];
			if (c.a)
			{	if (x+xi >= 0 && x+xi < int(winx) && y+yi >= 0 && y+yi < int(winy))
				{	if (c.a==0xff)
						pixels[x+xi+wyyi]=c;
					else
					{	Color &d=pixels[x+xi+wyyi];
						int a=c.a, da=d.a, sa=a+da;
						d.r=(d.r*da+c.r*a)/sa;
						d.g=(d.g*da+c.g*a)/sa;
						d.b=(d.b*da+c.b*a)/sa;
						d.a=(d.a*da+c.a*a)/sa;
					}
				}
			}
		}
	}
}

void Bitmap::bitPaste(BitmapP bm, int x, int y, double sx, double sy)
{
	int winxi=bm->winx,
		 winyi=bm->winy;
	Color c;
	for (int yi=0; yi < winyi; yi++)
	{	int yyi=int(yi*sx);
		if (y+yyi >= 0 && y+yyi < int(winy))
		{	int wyi=yi*winxi,
				 wyyi=(y+yyi)*winx;
			for (int xi=0; xi < winxi; xi++)
			{	c=bm->pixels[xi+wyi];
				if (c.a)
				{	int xxi=int(xi*sx);
					if (x+xxi >= 0 && x+xxi < int(winx))
					{	if (c.a==0xff)
							pixels[x+xxi+wyyi]=c;
						else
						{	Color &d=pixels[x+xxi+wyyi];
							int ca=c.a, da=d.a, sa=ca+da;
							d.r=(d.r*da+c.r*ca)/sa;
							d.g=(d.g*da+c.g*ca)/sa;
							d.b=(d.b*da+c.b*ca)/sa;
							d.a=(d.a*da+c.a*ca)/sa;
						}
					}
				}
			}
		}
	}
}

//___________________________________________________________________

NodeP lloadImage(PairPP arg)
{
	StringP filename=arg[1]->a_string(1);
	if (throwing)
		return Unbound;

	wxImage img;
	if (!img.LoadFile(filename->ptr()))
	{	error(BAD_DATA, filename);
		return Unbound;
	}

	Color *rgba=(Color*)GC_::malloc(img.GetWidth()*img.GetHeight()*4),
		*rgbap=rgba;
	if (!rgba)
		return Unbound;
	if (img.HasAlpha())
	{	for (int y=img.GetHeight()-1;  y >= 0;  y--)
		{	for (int x=0;  x< img.GetWidth();  x++)
			{	rgbap->r = img.GetRed(x,y);
				rgbap->g = img.GetGreen(x,y);
				rgbap->b = img.GetBlue(x,y);
				rgbap->a = img.GetAlpha(x,y);
				rgbap++;
			}
		}
	}
	else
	{	for (int y=img.GetHeight()-1;  y >= 0;  y--)
		{	for (int x=0;  x< img.GetWidth();  x++)
			{	rgbap->r = img.GetRed(x,y);
				rgbap->g = img.GetGreen(x,y);
				rgbap->b = img.GetBlue(x,y);
				rgbap->a = 0xff;
				rgbap++;
			}
		}
	}
	return new Bitmap(
		img.GetWidth(),img.GetHeight(), 
		img.GetWidth(),img.GetHeight(), rgba);
}

NodeP lsaveImage(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	StringP filename=arg[2]->a_string(2);
	if (throwing)
		return Unbound;

	int16 mx=b->winx, my=b->winy;

	wxImage img(mx, my, false);
	img.InitAlpha();

	Color *rgba=b->pixels,
		*rgbap=rgba;

	for (int y=my-1; y>=0; y--)
	{	for (int x=0; x<mx; x++)
		{	img.SetRGB(x,y, rgbap->r, rgbap->g, rgbap->b);
			img.SetAlpha(x,y, rgbap->a);
			rgbap++;
		}
	}
	if (!img.SaveFile(filename->ptr()))
		error(BAD_DATA, filename, 2);	
	return Unbound;
}

NodeP lBitMakeTransparent(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	int c=arg[2]->a_color(2);
	if (!throwing)
		b->makeTransparent(c);
	return Unbound;
}

NodeP lBitSetPixel(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	if (arg[2]->type==INT_ ||  arg[2]->type==FLOAT_)
	{	int x=int(arg[2]->getnumber());
		if (arg[3]->type==INT_ ||  arg[3]->type==FLOAT_)
		{	int y=int(arg[3]->getnumber());
			Color c=arg[4]->a_color(4);
			b->setPixel(x, y, c);
		}
	}
	else if (arg[2]->type==INTARRAY)
	{	IntArrayP xa=IntArrayP(arg[2]);
		intP x=xa->data();
		int l=xa->length();
		if (arg[3]->type==INT_)
		{	int y=IntP(arg[3])->int_;
			if (arg[4]->type==INTARRAY)
			{	IntArrayP cia=IntArrayP(arg[4]);
				if (l != cia->length())
				{	error(BAD_DATA, cia, 4);
					return Unbound;
				}
				intP c=cia->data();
				while (l--)
					b->setPixel(*x++, y, *c++);
			}
			else if (arg[4]->type==INT_)
			{	int c=IntP(arg[4])->int_;
				while (l--)
					b->setPixel(*x++, y, c);
			}
			else
				error(BAD_DATA, arg[4], 4);
		}
		else
			error(BAD_DATA, arg[3], 3);
	}
	else
		error(BAD_DATA, arg[2], 2);
	return Unbound;
}

NodeP lBitPixel(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	int x=int(arg[2]->a_float(2)),
		y=int(arg[3]->a_float(3));
	if throwing return Unbound;
	return new Int(b->Pixel(x, y));
}

NodeP lBitMaxX(PairPP arg)
{

	BitmapP b=arg[1]->a_bitmap(1);
	if throwing return Unbound;
	return new Int(b->winx-1);
}

NodeP lBitMaxY(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	if throwing return Unbound;
	return new Int(b->winy-1);
}

NodeP lBitPasteTo(PairPP arg)
{
	enum Params {_bitmapTo=1, _bitmapFrom, _x, _y};
	BitmapP 
		bitmapTo=arg[_bitmapTo]->a_bitmap(_bitmapTo),
		bitmapFrom=arg[_bitmapFrom]->a_bitmap(_bitmapFrom);
	NodeP 
		xn=arg[_x]->a_number(_x),
		yn=arg[_y]->a_number(_y);
	if throwing return Unbound;
	int	x=int(arg[_x]->getnumber()),
			y=int(arg[_y]->getnumber());
	bitmapTo->bitPaste(bitmapFrom, x, y);
	return Unbound;
}

NodeP lBitPasteToScaled(PairPP arg)
{
	enum Params {_bitmapTo=1, _bitmapFrom, _x, _y, _scalex, _scaley};
	BitmapP 
		bitmapTo=arg[_bitmapTo]->a_bitmap(_bitmapTo),
		bitmapFrom=arg[_bitmapFrom]->a_bitmap(_bitmapFrom);
	NodeP 
		xn=arg[_x]->a_number(_x),
		yn=arg[_y]->a_number(_y),
		sxn=arg[_scalex]->a_number(_scalex),
		syn=arg[_scaley]->a_number(_scaley);
	if throwing return Unbound;
	int	x=int(arg[_x]->getnumber()),
			y=int(arg[_y]->getnumber());
	double scalex=arg[_scalex]->getnumber(),
			 scaley=arg[_scaley]->getnumber();
	if throwing return Unbound;
	bitmapTo->bitPaste(bitmapFrom, x, y, scalex, scaley);
	return Unbound;
}
