//___________________________________________________________________
//
//                           Class  Button
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/button.h>
#include <wx/frame.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "LogoObject.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "Button.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Point.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Button, wxButton)
	EVT_BUTTON	(-1, Button::OnClick	)
END_EVENT_TABLE()

Button::Button(ButtonNodeP buttonNode_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=buttonNode_;
	wxWindow *parent=(wxWindow*)console;
	if (buttonNode_->parent)
		parent=buttonNode_->parent->window;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			_("Button"));
}

Button::Button(ButtonNodeP buttonNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxButton(parent, id, label, 
			pos, size, style, 
			validator, name)
{
	ThreadP th=Thread::This();
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=buttonNode_;
	*d(onClick)=onClick_;
	*d(uonClick) = th->lexufun;
	*d(tonClick) = th->thisobj;
}

Button::~Button()
{
	if (*d(windowNode))
		ButtonNodeP(*d(windowNode))->window=0;
}

void Button::OnClick(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventButton(this);
	if throwing return;
	*th->eventFiFo << evt;
}

wxString Button::wxstring()
{
	wxString str;
	str << _("Button") << " [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << (*d(onClick))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] ";
	return str;
}

DataInputStream &Button::read(DataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style;
	
	istr >> label >> name >> (int&)*d(onClick)
		>> pos.x >> pos.y >> size.x >> size.y >> style;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	return istr;
}
wxDataOutputStream &Button::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(&**d(onClick))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle());
	return ostr;
}

PairP ButtonNode::onClick()
{
	if (button)
		return PairP(*button->d(Button::onClick));
	return 0;
}

void ButtonNode::setonClick(NodeP nd) 
{
	if (window)
	{	ThreadP th=Thread::This();
		  *button->d(Button::onClick) = nd;
		*button->d(Button::uonClick) = th->lexufun;
		*button->d(Button::tonClick) = th->thisobj;
		check_oldyoung(nd);
	}
}

ButtonNode::ButtonNode(): 
	WindowNode(BUTTON),
	button((ButtonP&)window)
{}

ButtonNode::ButtonNode(WindowNodeP parent_): 
	WindowNode(BUTTON, parent_),
	button((ButtonP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void ButtonNode::free() 
{	
	if (window) 
	{	ThreadP th=Thread::This();
		*ButtonP(window)->d(Button::windowNode)=0;
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		window->Destroy(); 
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::tButton:
				if (WindowP(((EventButton*)ev)->button)==window)
					((EventButton*)ev)->button=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		window=0;
	}
}
void ButtonNode::inc_gc() {fifo << onClick();}
int ButtonNode::youngs() {return gmark(onClick());}
int ButtonNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(onClick(), ptrToInt);
}
void ButtonNode::check_oldyoungs() {check_oldyoung(onClick());}

DataInputStream &ButtonNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode;
		window=WindowP(new Button(ButtonNodeP(this)));
		if (!window)
		{	error(OUT_OF_MEM); return istr;
		}
		window->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &ButtonNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (window)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent);
		window->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ButtonNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=WindowNodeP((WindowNode*)intToPtr.find(int(parent)));
	if (window)
	{	setonClick(PairP((Pair*)intToPtr.find(int(ButtonP(window)->onClick))));
	}
}

void ButtonNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	ButtonP b=ButtonP(window), b2=0;
	if (!b) return;
	wxPoint pos=b->GetPosition();
	b2=new Button(ButtonNodeP(this),
			p, -1, b->GetLabel(), 
			PairP(*b->d(Button::onClick)), pos, b->GetSize(), b->GetWindowStyle(), 
			wxDefaultValidator, b->GetName());
	if (!b2)
	{	error(OUT_OF_MEM); return;
	}
	window=WindowP(b2);
	setonClick(*b->d(Button::onClick));
	delete b;
	window=WindowP(b2);
}

StringP ButtonNode::string()
{
	if (button)
		return new String(button->wxstring());
	return _Null_Word->name();
}

wxString ButtonNode::wxstring()
{
	if (button)
		return button->wxstring();
	return wxEmptyString;
}

void ButtonNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lButton(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	PairP onClick=0;
	if (exist(3))
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxWindowID id=-1;
	if (exist(7))
	{	int i=arg[7]->a_int(7);
		if throwing return Unbound;
		id=i;	
	}
	wxString name=label;
	if (exist(8))
	{	StringP n=arg[8]->a_string(8);
		if throwing return Unbound;
		name=n->ptr();
	}
//	GUI_ENTER
	ButtonNodeP buttonNode=new ButtonNode(parentNode);
	if throwing return Unbound;
	ButtonP button=new Button(buttonNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	if (!button)
	{	error(OUT_OF_MEM); return Unbound;
	}
	buttonNode->window=WindowP(button);
//	GUI_LEAVE
	return buttonNode;
}

NodeP lButtonDestroy(PairPP arg)
{
	ButtonNodeP button=arg[1]->a_button(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=button->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(button));
			p->check_oldyoung(p->children);
		}
	}
	button->window->Destroy();
//	GUI_LEAVE
	return Unbound;
}

NodeP lButtonEnable(PairPP arg)
{
	ButtonNodeP button=arg[1]->a_button(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	button->window->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lButtonGetLabel(PairPP arg)
{
	ButtonNodeP button=arg[1]->a_button(1);
	if throwing return Unbound;
	StringP label=new String(button->window->GetLabel());
	return label;
}

NodeP lButtonOnClick(PairPP arg)
{
	ButtonNodeP button=arg[1]->a_button(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
//	GUI_ENTER
	button->setonClick(function);
//	GUI_LEAVE
	return Unbound;
}

NodeP lButtonSetLabel(PairPP arg)
{
	enum Params {_button=1, _label};
	ButtonNodeP button=arg[_button]->a_button(_button);
	if throwing return Unbound;
	StringP label=arg[_label]->a_string(_label);
	if throwing return Unbound;
	button->window->SetLabel(label->ptr());
	return Unbound;
}

