#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/object.h>
#include <wx/app.h>
#include <wx/listctrl.h>
#include <wx/aui/aui.h>
#include <wx/config.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>
#include <stdio.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "Line.h"
#include "Object.h"
#include "Table.h"
#include "String.h"

#include "Error.h"
#include "NodePStack.h"
#include "Stack.h"
#include "Workspace.h"

#include "Calls.h"
#include "Color.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Thread.h"
#include "Aui.h"
#include "EdFrame.h"
#include "Init.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Calls, wxListCtrl)
	EVT_LIST_ITEM_FOCUSED(-1, Calls::OnClick)
	EVT_LIST_ITEM_ACTIVATED(-1, Calls::OnClick)
END_EVENT_TABLE()

Calls::Calls(wxWindow* parent, wxWindowID id, 
				 const wxPoint& pos, 
				 const wxSize& size, 
				 long style, 
				 const wxValidator& validator, 
				 const wxString& name):
wxListCtrl(parent, id, pos, size, style, validator, name)
{
	inupdate=false;
	InsertColumn(0, _("Calls"), wxLIST_FORMAT_LEFT, 1200);
	theApp->calls=this;
	show();
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuCalls->Check(true);
}

Calls::~Calls()
{
	theApp->calls=0;
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuCalls->Check(false);
}

void Calls::OnClick(wxListEvent &ev)
{
	ThreadP th=Thread::This();
	LineP line=LineP(objarr[ev.GetIndex()]);
	if (line)
	{	charP fn=line->filename()->ptr();
		int col=line->col, row=line->row;
		if (th->IsMain())
		{	theApp->editorShow();
			if (theApp->edFrame)
			{	theApp->edFrame->mark(fn, col, row, 0);
				theApp->edFrame->unmark(0);
			}
		}
	}
}

void Calls::show(bool shown)
{
	update();
	wxAuiPaneInfo pane=theApp->aui->GetPane(this);
	if (pane.IsOk())
	{	pane.Show();
	}else
	{	wxString pi;
		if (theApp->config->Read("CallsPaneInfo", &pi))
			theApp->aui->LoadPaneInfo(pi, pane);
		else
			pane=wxAuiPaneInfo()
				.Name("Calls").Caption(_("Calls"))
				.Layer(1).Bottom()
				.MinimizeButton().PinButton().Show(true)
				.DestroyOnClose();
		theApp->aui->AddPane(this, pane);
	}
	theApp->aui->Update();
}

wxString Calls::OnGetItemText(long item, long column) const
{
	return callarr[item];
}

void Calls::update()
{
	if (inupdate)
		return;

	inupdate=true;
	ThreadP th=Thread::This();

	int y=0;
	PairP callslistStart=PairP(lCalls(PairPP(Thread::This()->argStack->index))),
		callslist=callslistStart;
	_CallsList->setvar(callslist);
	if throwing return;
	if (callslist)
		count=callslist->count();
	else
		count=0;

	th->GuiEnter();
	SetItemCount(0);
	callarr.Clear(); callarr.Alloc(count); callarr.Add(wxEmptyString, count);
	 objarr.Clear();  objarr.Alloc(count);  objarr.Add(0, count);
	while (callslist)
	{
		PairP cf=callslist->first();
		if (cf)
		{
if (y >= count)
{	c_out << _("Calls::update error ") << count << callslistStart << endL;
	break;
}
			StringP call=StringP(cf->string());
			if throwing 
			{	th->GuiLeave(); return;
			}
			call=new String(
				call->ptr(),0,call->length(), STRING, String::copy_noformat);
			if throwing 
			{	th->GuiLeave(); return;
			}
			callarr[y]=call->ptr();
			 objarr[y]=cf;
		}
		y++;
		next(callslist);
	}
	SetItemCount(count);
	if (count >= 1)
		RefreshItems(0, count-1);

	th->GuiLeave();
	inupdate=false;
}
/*
void Calls::update()
{
	DeleteAllItems();
	int y=0;
	PairP callslist=PairP(_Calls(argStack));

	while (callslist)
	{
		PairP cf=callslist->first();
		if (cf)
		{	StringP call=StringP(cf->string());
			call=new String(
				call->ptr(),0,call->length(), STRING, String::copy_noformat);
			InsertItem(y, call->ptr());
		}
		y++;
		next(callslist);
	}
}
*/


NodeP lupdateCalls(PairPP)
{
	if (theApp && theApp->calls) 
		theApp->calls->update();
	return Unbound;
}
