//___________________________________________________________________
//
//                           Class ChoiceBox
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/listbox.h>
#include <wx/choice.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Point.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "ChoiceBox.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Color.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyChoiceBox : public wxChoice
{
public:
	MyChoiceBox(wxWindow* parent, wxWindowID id, 
		const wxArrayString& choices, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = _("ChoiceBox"));

	// MSW sends extra kill focus event
	virtual ~MyChoiceBox()
	{
		if (parent) 
			parent->choiceBox = 0;
		parent=0;
	}

	ChoiceBox *parent;

	void OnChar(wxKeyEvent &event);         // pass chars to ChoiceBox
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp(wxKeyEvent &event);
	void OnSelect(wxCommandEvent &event);
	void Set(wxArrayString &choices);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyChoiceBox,wxChoice)
	EVT_CHAR		(MyChoiceBox::OnChar		)
	EVT_KEY_DOWN(MyChoiceBox::OnKeyDown	)
	EVT_KEY_UP	(MyChoiceBox::OnKeyUp		)
	EVT_CHOICE	(-1,	MyChoiceBox::OnSelect	)
END_EVENT_TABLE()

MyChoiceBox::MyChoiceBox(wxWindow *parent_, wxWindowID id,
							  const wxArrayString &choices_,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxChoice(parent_, id, pos, size, choices_, style, validator, name)
{
	parent = (ChoiceBox*)parent_;
}

void MyChoiceBox::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyChoiceBox::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyChoiceBox::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}

void MyChoiceBox::OnSelect(wxCommandEvent &event)
{
	if (parent) 
		parent->OnSelect(event);
}

void MyChoiceBox::Set(wxArrayString &choices)
{
	Clear();
	int nr=choices.GetCount();
	for (int i=0; i<nr; i++)
		Append(choices[i]);
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(ChoiceBox, wxControl)
	EVT_CHAR				(		ChoiceBox::OnChar		)
	EVT_KEY_DOWN		(		ChoiceBox::OnKeyDown	)
	EVT_KEY_UP			(		ChoiceBox::OnKeyUp		)
	EVT_CHOICE			(-1,	ChoiceBox::OnSelect		)
END_EVENT_TABLE()

ChoiceBox::ChoiceBox(wxWindow *parent)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	label=0;
	choiceBox=0;
}

ChoiceBox::ChoiceBox(ChoiceBoxNodeP choiceBoxNode_, int style):
label(0), choiceBox(0), bestSize(-999, -999)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=choiceBoxNode_;
	wxWindow *parent=(wxWindow*)console;
	if (choiceBoxNode_->parent)
		parent=choiceBoxNode_->parent->window;
	create(parent, -1, wxArrayString(), 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, _("ChoiceBox"));
}

ChoiceBox::ChoiceBox(ChoiceBoxNodeP choiceBoxNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			PairP onSelect_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), choiceBox(0), bestSize(-999, -999), choices(choices_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=choiceBoxNode_;
	*(PairP*)(d(onSelect))=onSelect_;
	bestSize = wxSize(-999, -999);
	choiceBoxNode_->window=WindowP(this);
	choiceBoxNode_->style=style;
	create(parent, id, choices, pos, size, style, validator, name);
}

bool ChoiceBox::create(wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
	{	label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, _("staticText"));
		if (!label)
		{	error(OUT_OF_MEM); return false;
		}
	}else
		label=0;

	if (label)
	{	choiceBox = new MyChoiceBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
		if (!choiceBox)
		{	error(OUT_OF_MEM); return false;
		}
	}else
	{	choiceBox = new MyChoiceBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);
		if (!choiceBox)
		{	error(OUT_OF_MEM); return false;
		}
	}
	choices.Clear();

	DoSetSize(pos.x, pos.y, width, height);
	SetInitialSize(wxSize(width, height));
	return true;
}

void ChoiceBox::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (choiceBox)   
		choiceBox->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize ChoiceBox::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxChoice *t=new wxChoice((wxWindow*)this, wxID_ANY, 
			wxPoint(0,0), wxDefaultSize, choices, GetWindowStyle());
		if (!t)
		{	error(OUT_OF_MEM); return wxDefaultSize;
		}
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

ChoiceBox::~ChoiceBox()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (choiceBox)
	{	delete choiceBox;
		choiceBox=0;
	}

	ThreadP th=Thread::This();
	EventPP out=th->eventFiFo->out;
	EventP ev;
	while (*th->eventFiFo)
	{	*th->eventFiFo >> ev;
		switch (ev->type)
		{
		case Event::ChoiceBoxSelect:
		case Event::ChoiceBoxChar:
		case Event::ChoiceBoxKeyDown:
		case Event::ChoiceBoxKeyUp:
			if (((EventChoiceBox*)ev)->choiceBox==this)
				((EventChoiceBox*)ev)->choiceBox=0;
			break;
		default:
			;
		}
	}
	th->eventFiFo->out=out;
}

void ChoiceBox::Set(wxArrayString &choices)
{
	if (choiceBox)
		choiceBox->Set(choices);
}

#define LB_ON(NAME)																\
void ChoiceBox::On##NAME(wxKeyEvent &event)								\
{																						\
	ThreadP th=Thread::This();													\
	EventP evt=new EventChoiceBox##NAME(this, event);	\
	if throwing return;															\
	*th->eventFiFo << evt;														\
	event.Skip();																	\
}

LB_ON(Char)
LB_ON(KeyDown)
LB_ON(KeyUp)

void ChoiceBox::OnSelect(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventChoiceBoxSelect(this);
	if throwing return;
	*th->eventFiFo << evt;
}

wxString ChoiceBox::wxstring()
{
	wxString str;
	str << _("ChoiceBox ")
		<< '[' << GetName() << "] ";
	if (onSelect)
		str << (*d(onSelect))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &ChoiceBox::read(DataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString name, fontName;
	int fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> name 
		>> (int&)*d(onSelect)
		>> (int&)*d(onChar)
		>> (int&)*d(onKeyDown)
		>> (int&)*d(onKeyUp)
		>> pos.x >> pos.y >> size.x >> size.y
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	int nr;
	istr >> nr;
	choices.Clear();
	for (int i=0; i<nr; i++)
	{	wxString item;
		istr >> item;
		choices.Add(item);
	}
	choiceBox->Set(choices);

	istr >> nr;
	choiceBox->SetSelection(nr);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &ChoiceBox::write(wxDataOutputStream &ostr) 
{
//	flush();
	ostr
		<< GetName() 
		<< GC_::ptrToInt->find(*d(onSelect))
		<< GC_::ptrToInt->find(*d(onChar))
		<< GC_::ptrToInt->find(*d(onKeyDown))
		<< GC_::ptrToInt->find(*d(onKeyUp))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;

	{	wxColour 
			f=choiceBox->GetForegroundColour(), 
			b=choiceBox->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=choiceBox->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();

	int nr=choiceBox->GetCount();
	ostr << nr;
	for (int i=0; i<nr; i++)
		ostr << choiceBox->GetString(i);

	wxArrayInt sel;
	nr=choiceBox->GetSelection();
	ostr << nr;
	return ostr;
}

bool ChoiceBox::SetBackgroundColour(const wxColour &c)
{
	if (choiceBox)
		return choiceBox->SetBackgroundColour(c);
	return false;
}

bool ChoiceBox::SetForegroundColour(const wxColour &c)
{
	if (choiceBox)
		return choiceBox->SetForegroundColour(c);
	return false;
}

bool ChoiceBox::SetFont(const wxFont &f)
{
	if (choiceBox)
		return choiceBox->SetFont(f);
	return false;
}

//___________________________________________________________________

#define NODE_ON(x)							\
PairP ChoiceBoxNode::on##x()				\
{													\
	if (window)									\
	return PairP(*ChoiceBoxP(window)->d(ChoiceBox::on##x));\
	return 0;									\
}													\
													\
void ChoiceBoxNode::seton##x(NodeP nd)	\
{													\
	if (window)								\
	{	*ChoiceBoxP(window)->d(ChoiceBox::on##x) = PairP(nd);		\
		check_oldyoung(nd);					\
	}												\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(Select)

ChoiceBoxNode::ChoiceBoxNode(): 
	WindowNode(CHOICEBOX),
	choiceBox((ChoiceBoxP&)window)
{}

ChoiceBoxNode::ChoiceBoxNode(WindowNodeP parent_): 
	WindowNode(CHOICEBOX, parent_),
	choiceBox((ChoiceBoxP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void ChoiceBoxNode::free() 
{	
	if (window) 
	{	ThreadP th=Thread::This();
	*ChoiceBoxP(window)->d(ChoiceBox::windowNode)=0;
		window->Destroy(); 
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::ChoiceBoxSelect:
			case Event::ChoiceBoxChar:
			case Event::ChoiceBoxKeyDown:
			case Event::ChoiceBoxKeyUp:
				if (WindowP(((EventChoiceBox*)ev)->choiceBox)==window)
					((EventChoiceBox*)ev)->choiceBox=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		window=0;
	}
}
void ChoiceBoxNode::inc_gc() 
{	fifo << onSelect()
		 << onChar() << onKeyDown() << onKeyUp();
}
int ChoiceBoxNode::youngs() 
{	return gmark(onSelect())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int ChoiceBoxNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(onSelect(), ptrToInt)
		+gmarkPtr(onChar(), ptrToInt)
		+gmarkPtr(onKeyDown(), ptrToInt)
		+gmarkPtr(onKeyUp(), ptrToInt);
}
void ChoiceBoxNode::check_oldyoungs() 
{	check_oldyoung(onSelect());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

DataInputStream &ChoiceBoxNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode >> style;
		window=WindowP(new ChoiceBox(ChoiceBoxNodeP(this), style));
		if throwing return istr;
		window->read(istr);
		parent=parentNode;
	}
	return istr;
}
wxDataOutputStream &ChoiceBoxNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (window)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent) 
			<< style;
		window->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ChoiceBoxNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (window)
	{	ChoiceBoxP t=ChoiceBoxP(window);
		setonSelect ((Pair*)intToPtr.find(int(t->onSelect )));
		setonChar   ((Pair*)intToPtr.find(int(t->onChar   )));
		setonKeyDown((Pair*)intToPtr.find(int(t->onKeyDown)));
		setonKeyUp  ((Pair*)intToPtr.find(int(t->onKeyUp  )));
	}
}

void ChoiceBoxNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	ChoiceBoxP t=ChoiceBoxP(window), t2=0;
	if (!t) return;
	wxPoint pos=t->GetPosition();
	t2=new ChoiceBox(ChoiceBoxNodeP(this),
			p, -1, t->choices, 
			PairP(*t->d(ChoiceBox::onSelect)), pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	if throwing return;
	window=WindowP(t2);
	wxArrayInt sel;
	int nr=t->choiceBox->GetSelection();
	t2->choiceBox->SetSelection(nr);
	setonSelect (*t->d(ChoiceBox::onSelect ));
	setonChar   (*t->d(ChoiceBox::onChar   ));
	setonKeyDown(*t->d(ChoiceBox::onKeyDown));
	setonKeyUp  (*t->d(ChoiceBox::onKeyUp  ));
	t2->choiceBox->SetForegroundColour(t->choiceBox->GetForegroundColour());
	t2->choiceBox->SetBackgroundColour(t->choiceBox->GetBackgroundColour());
	wxFont f=t->choiceBox->GetFont();
	t2->choiceBox->SetFont(f);
	delete t;
	choiceBox=t2;
}

StringP ChoiceBoxNode::string()
{
	if (choiceBox)
		return new String(choiceBox->wxstring());
	return _Null_Word->name();
}

wxString ChoiceBoxNode::wxstring()
{
	if (choiceBox)
		return choiceBox->wxstring();
	return wxEmptyString;
}

void ChoiceBoxNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lChoiceBox(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=Thread::This()->graph;
	WindowNodeP parentNode=0;
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString name=_("ChoiceBox");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	wxArrayString choices;
	{	PairP list=arg[3]->a_list(3);
		if throwing return Unbound;
		while (list)
		{	StringP s=list->first()->a_string(3); 
			if throwing return Unbound;
			choices.Add(s->ptr());
			next(list);
		}
	}
	PairP onSelect=0;
	if (exist(4))
	{	PairP os=arg[4]->a_runnable(4);
		if throwing return Unbound;
		onSelect=os;
	}
	long style = 0;
	if (exist(5))
	{	int s=arg[5]->a_int(5);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(6))
	{	PairP p=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(7))
	{	PairP s=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
//	GUI_ENTER
	ChoiceBoxNodeP choiceBoxNode=new ChoiceBoxNode(parentNode);
	if throwing return Unbound;
	ChoiceBoxP cb=new ChoiceBox(choiceBoxNode,
			parent, id, choices, 
			onSelect, 
			pos, size, style, wxDefaultValidator, name);
	if throwing return Unbound;
	cb=cb;
//	GUI_LEAVE
	return choiceBoxNode;
}

NodeP lChoiceBoxSelection(PairPP arg)
{
	ThreadP th=Thread::This();
	ChoiceBox *choiceBox=0;
	if (exist(1))
	{	ChoiceBoxNodeP choiceBoxNode=arg[1]->a_choiceBox(1);
		if throwing return Unbound;
		choiceBox=choiceBoxNode->choiceBox;
	}
	else
	{	choiceBox=th->choiceBoxValue;
	}
	if (!choiceBox) return 0;
	wxArrayInt sel;
//	GUI_ENTER
	int nr=choiceBox->choiceBox->GetSelection();
//	GUI_LEAVE
	return new Int(nr);
}

NodeP lChoiceBoxSetSelection(PairPP arg)
{
	ChoiceBoxNodeP choiceBoxNode=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	ChoiceBox *choiceBox=choiceBoxNode->choiceBox;
	if (!choiceBox) return 0;
	int sel=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	int nr=choiceBox->choiceBox->GetCount();
	if (sel < 0 || sel >= nr)
	{	error(BAD_DATA, arg[2], 2);
//		GUI_LEAVE
		return Unbound;
	}
	choiceBox->choiceBox->SetSelection(sel);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxSetChoices(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	wxArrayString choices;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
//	GUI_ENTER
	while (list)
	{	StringP s=list->first()->a_string(2);
		if throwing return Unbound;
		choices.Add(s->ptr());
		next(list);
	}
	choiceBox->choiceBox->choiceBox->Set(choices);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxAppend(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	choiceBox->choiceBox->choiceBox->Append(value->ptr());
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxSetItem(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	StringP str=arg[3]->a_string(3);
	if throwing return Unbound;
//	GUI_ENTER
	if (n < 0 || n >= int(choiceBox->choiceBox->choiceBox->GetCount()))
	{	error(BAD_DATA, arg[2]);
//		GUI_LEAVE
		return Unbound;
	}
	choiceBox->choiceBox->choiceBox->SetString(n, str->ptr());
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxRemoveItem(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	if (n < 0 || n >= int(choiceBox->choiceBox->choiceBox->GetCount()))
	{	error(BAD_DATA, arg[2]);
//		GUI_LEAVE
		return Unbound;
	}
	choiceBox->choiceBox->choiceBox->Delete(n);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxCount(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
//	GUI_ENTER
	int count=choiceBox->choiceBox->choiceBox->GetCount();
//	GUI_LEAVE
	return new Int(count);
}

#define _TC_ON(NAME)									\
NodeP lChoiceBoxOn##NAME(PairPP arg)			\
{															\
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	choiceBox->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(Select)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lChoiceBoxSetBackgroundColor(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	choiceBox->choiceBox->SetBackgroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxSetColor(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	choiceBox->choiceBox->SetForegroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxSetFontSize(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetPointSize(s);
	choiceBox->choiceBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxSetFontName(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetFaceName(s->ptr());
	choiceBox->choiceBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxSetFontStyle(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetStyle(s);
	choiceBox->choiceBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxSetFontWeight(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetWeight(s);
	choiceBox->choiceBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxEnable(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	choiceBox->choiceBox->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lChoiceBoxDestroy(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=choiceBox->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(choiceBox));
			p->check_oldyoung(p->children);
		}
	}
	choiceBox->choiceBox->Destroy();
//	GUI_LEAVE
	return Unbound;
}
