#include "stdafx.h"

#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/frame.h>
#include <wx/dialog.h>
#include <wx/listctrl.h>
#include <wx/thread.h>
#include <wx/treectrl.h>
#include <wx/listimpl.cpp>
#include <wx/timer.h>
#include <wx/notebook.h>

#include <stdio.h>

#include "Main.h"

#include "File.h"
#include "Pair.h"
#include "Procedure.h"
#include "IntNodeTable.h"
#include "Compiled.h"
#include "Primitive.h"
#include "Object.h"
#include "Table.h"
#include "PtrTable.h"
#include "LogoObject.h"
#include "Quote.h"
#include "Colon.h"
#include "Tree.h"
#include "Int.h"
#include "Float.h"
#include "String.h"

#include "Environment.h"
#include "Error.h"
#include "Eval.h"
#include "Compile.h"
#include "NodePStack.h"
#include "Stack.h"
#include "ToDoStack.h"
#include "Init.h"
#include "Workspace.h"

#include "Coms.h"
#include "Calls.h"
#include "Vars.h"
#include "UInt8.h"
#include "Color.h"
#include "Console.h"
#include "LogoApp.h"
#include "Graphics.h"

#include "EdFrame.h"
#include "Mem.h"
#include "Point.h"
#include "Graph.h"
#include "Pen.h"
#include "Font.h"
#include "Turtle.h"
#include "Event.h"
#include "Security.h"
#include "Thread.h"
#include "DebugState.h"


#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#ifdef WIN32
	#pragma inline_depth(12)
	#pragma inline_recursion(on)
#endif

#define create_name(x) #x,

const char Compiler::commandName[][end+1] =
{	Eval_Labels(create_name)
	"null"
};

#ifdef PROC_EVAL	// for debugging & profiling
	#ifdef _DEBUG		// for use with stack frames
		#define Goto(f)	{register Do fun=__##f.instruction; __asm{mov ebx,fun __asm mov ecx,this __asm mov esp,ebp __asm pop ebp __asm jmp ebx} return 0;}
	#else				// for use without stack frames
		#define Goto(f)	{return *intP(&__##f.instruction);}
	#endif
	#define TO(x)	}	Compiler::_##x()	{
#else	// for the optimized release version; the fastest Goto is a goto !-)
	#define Goto(f)  goto f;
	#define TO(x)	}	x:	{
#endif

NodeP lcompile(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP li=arg[1]->a_runnable(1);
	th->val_status=output_ok;
	Compiler c(li, __begin_line, th);
	return Unbound;
}

NodeP llink(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP li=arg[1]->a_list(1);
	while (li)
	{	NodeP nd=li->first();
		if (!nd)
		{	error(BAD_DATA, nd); return Unbound;
		}	
		ObjectP obj=nd->intern();
		if (obj->proc()->compiled())
		{	obj->proc()->compiled()->link();
		}else
		{	error(BAD_DATA, obj); return Unbound;
		}
		next(li);
	}
	return Unbound;
}

void checkBreak()
{
	ThreadP th=Thread::This();
	th->checkBreak();
}

void argStackGrow()
{
	ThreadP th=Thread::This();
	th->argStack->grow();
}

NodeP lcompileIfElse(PairPP arg);

NodeP lcompileIf(PairPP arg) 
{
	if (exist(3)) return lcompileIfElse(arg);
	PairP yes = arg[2]->a_runnable(2);
	yes->makeTree();
	if (throwing) return 0;
	return new Pair(yes, 0);
}

NodeP lcompileIfElse(PairPP arg) 
{
	PairP yes = arg[2]->a_runnable(2);
	PairP  no = arg[3]->a_runnable(3);
	yes->makeTree();
	no->makeTree();
	if (throwing) return 0;
	return new Pair(yes, no);
}

NodeP lcompileForever(PairPP arg) 
{
	PairP instructionlist = arg[1]->a_runnable(1);
	if throwing return 0;
	instructionlist->makeTree();
	if throwing return 0;
	return instructionlist;
}

NodeP lcompileRepeat(PairPP arg) 
{
	PairP instructionlist = arg[2]->a_runnable(2);
	if throwing return 0;
	instructionlist->makeTree();
	if throwing return 0;
	return instructionlist;
}

NodeP lcompileWhile(PairPP arg) 
{
	PairP cond = arg[1]->a_runnable(1);
	PairP instructionlist = arg[2]->a_runnable(2);
	if throwing return 0;
	cond->makeTree();
	if throwing return 0;
	instructionlist->makeTree();
	if throwing return 0;
	return new Pair(cond, instructionlist);
}

NodeP evaluate(PairPP arg)
{
	ThreadP th=Thread::This();
	Evaluator e;
	return e.eval(arg[1], __begin_line, th);
}

//___________________________________________________________________
//
//         A Rough, Unfinished Micro Assembler for Intel x86
//___________________________________________________________________

CodeMacro::CodeMacro(wxString name_, OpCode o): 
name(name_), length(1), adr(0), offs(0), obj(0)     {code[0]=char(o);}

CodeMacro::CodeMacro(wxString name_, OpCode o, char i): 
name(name_), length(2), adr(0), offs(0), obj(0)     {code[0]=char(o); code[1]=i;}

CodeMacro::CodeMacro(wxString name_, OpCode o, int16 i): 
name(name_), length(3), adr(0), offs(0), obj(0)     {code[0]=char(o); *int16P(code+1)=int16(i);}

CodeMacro::CodeMacro(wxString name_, OpCode o, int16 i, char i2): 
name(name_), length(4), adr(0), offs(0), obj(0)     {code[0]=char(o); *int16P(code+1)=int16(i); code[3]=i2;}

CodeMacro::CodeMacro(wxString name_, OpCode o, int i): 
name(name_), length(5), adr(0), offs(0), obj(0)     {code[0]=char(o); *intP(code+1)=int(i);}

CodeMacro::CodeMacro(wxString name_, OpCode o, int i, ObjectP obj_, bool isadr): 
name(name_), length(5), adr(i), offs(1), obj(obj_)  {code[0]=char(o); *intP(code+1)=int(i);}

CodeMacro::CodeMacro(wxString name_, OpCode o, Reg r): 
name(name_), length(1), adr(0), offs(0), obj(0)     {code[0]=char(o+r);}

CodeMacro::CodeMacro(wxString name_, OpCode o, Reg r, int i): 
name(name_), length(5), adr(0), offs(0), obj(0)     {code[0]=char(o+r); *intP(code+1)=int(i);}

CodeMacro::CodeMacro(wxString name_, OpCode o, ModRM m): 
name(name_), length(2), adr(0), offs(0), obj(0)     {code[0]=char(o); code[1]=char(m);}

CodeMacro::CodeMacro(wxString name_, OpCode o, ModRM m, void *mem): 
name(name_), length(6), adr(0), offs(0), obj(0)     {code[0]=char(o); code[1]=char(m); *intP(code+2)=int(mem);}

CodeMacro::CodeMacro(wxString name_, OpCode o, ModRM m, void *mem, int i): 
name(name_), length(10), adr(0), offs(0), obj(0)    {code[0]=char(o); code[1]=char(m); *intP(code+2)=int(mem); *intP(code+6)=i;}

CodeMacro::CodeMacro(wxString name_, OpCode o, ModRM m, char c): 
name(name_), length(3), adr(0), offs(0), obj(0)     {code[0]=char(o); code[1]=char(m); code[2]=c;}

CodeMacro::CodeMacro(wxString name_, OpCode o, ModRM m, int c): 
name(name_), length(6), adr(0), offs(0), obj(0)     {code[0]=char(o); code[1]=char(m); *intP(code+2)=c;}

CodeMacro::CodeMacro(wxString name_, OpCode o, OpCode o2, JmpCond jc, int i): 
name(name_), length(6), adr(0), offs(0), obj(0)     {code[0]=char(o); code[1]=char(o2+jc); *intP(code+2)=i;}
		

#define CAdr(f) (int(f)-(int(code->code)+code->length))
#define JAdr(f) (int(f)-code->length)

#define ADDrc(reg,c)			CodeMacro("ADD  "#reg","#c			, addc , ModRM(REG,AX,reg), c)
#define ADD(reg,sreg)		CodeMacro("ADD  "#reg","#sreg		, add  , ModRM(REG,sreg,reg))
#define ADDmr(mem,reg)		CodeMacro("ADD  "#mem","#reg		, add  , ModRM(INDIRECT,reg,DISP32), mem)
#define ADDrm(reg,mem)		CodeMacro("ADD  "#reg","#mem		, addr , ModRM(INDIRECT,reg,DISP32), mem)
#define CALL(func)			CodeMacro("CALL "#func				, call , CAdr(func)-5)
#define CALLf(func, obj)	CodeMacro(wxString("CALL ")+obj->name()->ptr()	, call , CAdr(func)-5)
#define CALLl(func, obj)	CodeMacro(wxString("CALL ")+obj->name()->ptr()	, call , CAdr(func)-5, obj, true)
#define CMPrc(reg,c)			CodeMacro("CMP  "#reg","#c			, cmpc , ModRM(REG,DI,reg), c)
#define CMP(reg,sreg)		CodeMacro("CMP  "#reg","#sreg		, cmp  , ModRM(REG,sreg,reg))
#define CMPmr(mem,reg)		CodeMacro("CMP  "#mem","#reg		, cmp  , ModRM(INDIRECT,reg,DISP32), mem)
#define CMPrm(reg,mem)		CodeMacro("CMP  "#reg","#mem		, cmpr , ModRM(INDIRECT,reg,DISP32), mem)
#define CMPri(reg,ireg)		CodeMacro("CMP  "#reg",["#ireg"]", cmpr , ModRM(INDIRECT,reg,ireg))
#define DEC(reg)				CodeMacro("DEC  "#reg				, decr , reg)
#define DECm(mem)				CodeMacro("DEC  "#mem				, decm , ModRM(INDIRECT,CX,DISP32), mem)
#define ENTER(c)				CodeMacro("ENTER "#c					, enter, int16(c), 0)
#define INC(reg)				CodeMacro("INC  "#reg				, incr , reg)
#define INCm(mem)				CodeMacro("INC  "#mem				, incm , ModRM(INDIRECT,AX,DISP32), mem)
#define JMP(mem)				CodeMacro("JMP  "#mem				, jmp  , mem-6)
#define JMPc(cond,mem)		CodeMacro("J"#cond"  "#mem			, jcc  , jcc2, cond, mem-6)
#define LEAVE					CodeMacro("LEAVE"						, leave)
#define MOV(reg,sreg)		CodeMacro("MOV  "#reg"," #sreg	, mov  , ModRM(REG,sreg,reg))
#define MOVrc(reg,c)			CodeMacro("MOV  "#reg"," #c		, movc , reg, int(c))
#define MOVmc(mem,c)			CodeMacro("MOV  "#mem"," #c		, movmc, ModRM(INDIRECT,AX,DISP32), mem, int(c))
#define MOVir(ireg,reg)		CodeMacro("MOV ["#ireg"],"#reg	, mov  , ModRM(INDIRECT,reg,ireg))
#define MOVri(reg,ireg)		CodeMacro("MOV  "#reg",["#ireg"]", movr , ModRM(INDIRECT,reg,ireg))
#define MOVicr(ireg,c,reg)	CodeMacro("MOV ["#ireg"+"#c"],"#reg		, mov  , ModRM(INDIRECT8,reg,ireg), char(c))
#define MOVric(reg,ireg,c)	CodeMacro("MOV  "#reg",["#ireg"+"#c"]"	, movr , ModRM(INDIRECT8,reg,ireg), char(c))
#define MOVmr(mem,reg)		CodeMacro("MOV  "#mem"," #reg		, mov  , ModRM(INDIRECT,reg,DISP32), mem)
#define MOVrm(reg,mem)		CodeMacro("MOV  "#reg"," #mem		, movr , ModRM(INDIRECT,reg,DISP32), mem)
#define POP(reg)				CodeMacro("POP  "#reg				, popr , reg)
#define POPm(mem)				CodeMacro("POP  "#mem				, pop  , ModRM(INDIRECT, AX,DISP32), mem)
#define PUSH(reg)				CodeMacro("PUSH "#reg				, pushr, reg)
#define PUSHm(mem)			CodeMacro("PUSH "#mem				, push_, ModRM(INDIRECT, SI,DISP32), mem)
#define PUSHc(c)				CodeMacro("PUSH "#c					, pushc, int(c))
#define RET						CodeMacro("RET  "						, ret)
#define SUB(reg,sreg)		CodeMacro("SUB  "#reg","#sreg		, sub  , ModRM(REG,sreg,reg))
#define SUBrc(reg,c)			CodeMacro("SUB  "#reg","#c			, subc , ModRM(REG,BP,reg), c)
#define SUBmr(mem,reg)		CodeMacro("SUB  "#mem","#reg		, sub  , ModRM(INDIRECT,reg,DISP32), mem)
#define SUBrm(reg,mem)		CodeMacro("SUB  "#reg","#mem		, subr , ModRM(INDIRECT,reg,DISP32), mem)

#define arg_PUSH_AX			MOVrm(CX,&th->argStack->index)	 \
								<<	ADDrc(CX,4)								\
								<<	CMPrm(CX,&th->argStack->max);		\
					{	int if_grow=code->length;						\
						*code	<<	JMPc(B,0)								\
								<<	PUSH(AX);								\
						*code	<<	CALL(argStackGrow)					\
								<< POP(AX)									\
								<< MOVrm(CX,&th->argStack->index)	\
								<<	ADDrc(CX,4);								\
						*(int*)(charP(code->code)+if_grow+2)=code->length-if_grow-6;	\
					}															\
						*code	<<	MOVir(CX,AX)					   \
								<<	MOVmr(&th->argStack->index, CX)

#define arg_PUSH(c)			MOVrc(DX,c)								\
								<<	MOVrm(CX,&th->argStack->index)	\
								<<	ADDrc(CX,4)								\
								<<	CMPrm(CX,&th->argStack->max);		\
					{	int if_grow=code->length;						\
						*code	<<	JMPc(B,0)								\
								<<	PUSH(DX);								\
						*code	<<	CALL(argStackGrow)					\
								<<	POP(DX)									\
								<<	MOVrm(CX,&th->argStack->index)	\
								<<	ADDrc(CX,4);							\
						*(int*)(charP(code->code)+if_grow+2)=code->length-if_grow-6;	\
					}															\
						*code	<<	MOVir(CX,DX)							\
								<<	MOVmr(&th->argStack->index, CX)

#define arg_POP_TILL_asp	GET_ASP(CX)							\
								<<	ADDrm(CX,&th->argStack->base)	 \
								<<	MOVmr(&th->argStack->index, CX)

#define SET_ASP_from_argStack							\
	if (code && code->code)								\
		*code << MOVrm(CX, &th->argStack->index)	\
				<< SUBrm(CX, &th->argStack->base)	\
				<< SET_ASP(CX);

#define GET_th(reg)			MOVric(reg, BP, -16)
#define SET_th(reg)			MOVicr(BP, -16, reg)

#define GET_ASP(reg)			MOVric(reg, BP, -12)
#define SET_ASP(reg)			MOVicr(BP, -12, reg)

#define GET_repcount(reg)	MOVric(reg, BP, -8)
#define SET_repcount(reg)	MOVicr(BP, -8, reg)

#define GET_user_repcount(reg)	MOVric(reg, BP, -4)
#define SET_user_repcount(reg)	MOVicr(BP, -4, reg)

#define GET_th_stoppingFlag(reg)								\
				GET_th(CX)											 \
			<< ADDrc(CX,int(&th->stoppingFlag)-int(th))  \
			<< MOVri(reg,CX)										   \

#define SET_th_stoppingFlag(reg)								\
				GET_th(CX)											 \
			<< ADDrc(CX,int(&th->stoppingFlag)-int(th))  \
			<< MOVir(CX,reg)										   \

#define SET_th_output_node(reg)								\
				GET_th(CX)											 \
			<< ADDrc(CX,int(&th->output_node)-int(th))  \
			<< MOVir(CX,reg)										   \

#define cdecl_asp_CALL(func, obj)				\
						*code	<< GET_ASP(CX)	 \
								<< ADDrm(CX,&th->argStack->base)	\
								<< PUSH(BP)		 \
								<< MOV(BP,SP)	  \
								<< PUSH(CX);   \
						*code	<< CALLf(func, obj)	    \
								<< ADDrc(SP,4)		  \
								<< POP(BP)

#define cdecl_asp_CALLl(func, obj)				\
						*code	<< GET_ASP(CX)	 \
								<< ADDrm(CX,&th->argStack->base)	\
								<< PUSH(BP)		 \
								<< MOV(BP,SP)	  \
								<< PUSH(CX);   \
						*code	<< CALLl(func, obj)	    \
								<< ADDrc(SP,4)		  \
								<< POP(BP)

#define fast_asp_CALL(func)					\
						*code	<< MOVrm(CX,&asp); \
						*code	<< CALL(func)

#define RET_Unbound				\
				MOVrc(AX,Unbound)	 \
			<< LEAVE					  \
			<< RET

//___________________________________________________________________
//
//		An explicit control evaluator, according to SICP, section 5.2.
//			<nodeList> is a flat list of expressions to evaluate.
//			<what> is the EvalCommand to begin with.
//			<Return value>  depends on what.
//___________________________________________________________________

Compiler::Compiler(PairP nodeList_, ToDo what, Thread *th_)
{
	th=th_;
	compile(nodeList_, what, th);
}

NodeP Compiler::compile(PairP nodeList_, ToDo what, Thread *th)

#ifndef PROC_EVAL
{
#endif
				{	oldtailcall=th->tailcall;
					tracing=1;
					old_ift_iff=th->ift_iff_flag;

					val=catch_tag=0;
					exp=unev= arg= parm=formals=0;
					proc=0;
					asp=*th->argStack;
					vsp=0;
					th->var=*th->varStack;
					
					Push(th->varPStack, th->var, eval);
					Push3(th->stack, th->fun, th->ufun, th->this_line, eval);
					Push(th->toDoStack, __end, eval);

					nodeList=nodeList_;

	code=new Compiled(ObjectP(th->fun));
	*code << ENTER(16);
	*code	<<	CALL(Thread::This)
			<<	SET_th(AX);

#ifdef PROC_EVAL
 					ToDo command = what;
					while(command.instruction)  
						command.i = (this->*command.instruction)();
					return val;
#else
					Push(th->toDoStack, what, eval);
					Goto( fetch_command )
#endif

TO( fetch_command )
					ToDo __command;
					Pop1(th->toDoStack, __command, fetch_command); 
#ifdef PROC_EVAL
					Goto( command )
#else
	#define CREATE_CASE(x)	case x:	goto x;

					switch (__command.mnemonic)
					{	Eval_Labels(CREATE_CASE)
					}
#endif

TO( begin_line )
					Push(th->toDoStack, __end_line, begin_line);	Goto( begin_sequence )
TO( end_line )
					if (val != Unbound 
					&& th->val_status!=output_ok
					&& th->val_status!=value_required
					&& th->val_status!=dont_care
					&& !throwing) 
						error(DK_WHAT, val);
					if (val != Unbound 
					&& th->val_status!=output_ok
					&& th->val_status!=value_required
					&& th->val_status!=dont_care) 
						val=0;
					Goto( fetch_command )

TO( tail_dispatch )
					th->tailcall = tailcall_tail;
					Goto( dispatch )
TO( dispatch )
					switch (exp->type)
					{
					case LINE:
					case PAIR:	// procedure application
					case TREE:
									{	ObjectP obj=ObjectP(exp->first());
//	if (obj->bit.proc_macro)
//		Push(th->intStack, code->length, dispatch);

										if (th->tailcall==tailcall_tail)
										{	if (!obj->bit.permanent 
											&&   obj->bit.proc_macro 
											||   obj==_Goto
											||   obj==_For)	// this is not very nice
																	// but I don't know how to avoid it
											{	// tail call to user-defined macro must be treated as non-tail
												// because the expression returned by the macro
												// remains to be evaluated in the caller's context
												unev=0; Goto( non_tail_eval )
											}
										}
										th->fun=obj;
									}
									if (exp->tail()) Goto( proc_with_args )
									asp=*th->argStack;	
	SET_ASP_from_argStack;
									// Evaluate an application of a procedure with no arguments. 
									Goto( apply_dispatch )	// apply the procedure 

					case COLON:	// variable
	{	int sp= th->varStack->whereIsVar(ObjectP(exp->first()));
	//									if (val==Unbound && !throwing)	
	//										error(NO_VALUE, ColonP(exp)->colon());
		if (sp < 0)
		{	c_out	<< "			;COLON global " << exp << endL;
			*code	<< MOVrc(CX,int(exp->first())+int(&(Object::aObject._var))-int(&Object::aObject))
					<< MOVri(AX,CX);
		}else
		{	c_out	<< "			;COLON on argStack " << exp << endL;
			*code	<<	MOVrm(CX,&th->argStack->base)
					<< GET_ASP(DX)
					<< ADD(CX,DX)
					<< SUBrc(CX,sp*4)
					<< MOVri(AX,CX);
		}
	}
									break;

					case QUOTE:
	c_out	<< "			;QUOTE " << exp << endL;
	*code	<< MOVrc(AX,QuoteP(exp)->quote());
									val = QuoteP(exp)->quote();	break; // quoted literal
//					case ARRAY:	//val = ArrayP(exp)->deepCopy();break; // array must be copied
//									val = NodeP(exp->array());	break; 
					default:		val=exp;		// self-evaluating
					}
	*code	<<	MOVrc(AX,val);
					Goto( fetch_command )

TO( proc_with_args )	// Evaluate an application of a procedure with arguments. 
					unev=exp->tail();
					asp=*th->argStack;
	SET_ASP_from_argStack;
					Push3(th->intStack, th->val_status, th->tailcall, th->ift_iff_flag, proc_with_args);
					Push(th->varPStack, th->var, proc_with_args);
					Push(th->stack, th->didnt_output_name, proc_with_args);
					Goto( arg_loop )
TO( arg_loop )
	 arg_loop_again:
					if (!unev) Goto( args_done )
					exp = unev->first();
					if (!exp) {exp=_Not_Enough_node; goto arg_is_function;}
//if (!is_valid_pointer(exp)) {lpause(0); return Unbound;}
					switch (exp->type)
					{
					case LINE:
					case PAIR:
					case TREE:
								 {	ObjectP obj=ObjectP(exp->first());
									// Fix a colonless var if no proc or prim exists. 
									if (obj && obj->bit.proc)
										goto arg_is_function;
									if (th->logoLanguage==_ucblogo)
									{	error(DK_HOW, obj); 
										break;
									}
									exp = new Colon(obj);
									unev->setfirst( exp );	// now we have a COLON
								 }
					case COLON:	// variable
	{	int sp= th->varStack->whereIsVar(ObjectP(exp->first()));
	//									if (val==Unbound && !throwing)	
	//										error(NO_VALUE, ColonP(exp)->colon());
		if (sp < 0)
		{	c_out	<< "			;COLON global " << exp << endL;
			*code	<< MOVrc(CX,int(exp->first())+int(&(Object::aObject._var))-int(&Object::aObject))
					<< MOVri(AX,CX)
					<< arg_PUSH_AX;	// no error-checking yet!
		}else
		{	c_out	<< "			;COLON on argStack " << exp << endL;
			*code	<<	MOVrm(CX,&th->argStack->base)
					<< GET_ASP(DX)
					<< ADD(CX,DX)
					<< SUBrc(CX,sp*4)
					<< MOVri(AX,CX)
					<< arg_PUSH_AX;
		}
	}
									break;

					case QUOTE:	
	c_out	<< "			;QUOTE " << exp << endL;
	*code	<< arg_PUSH(QuoteP(exp)->quote());
									val = QuoteP(exp)->quote();	break; // quoted literal
//					case ARRAY:	//val = ArrayP(exp)->deepCopy();break; // array must be copied
//									val = NodeP(exp->array());	break; 
					default:
	c_out	<< "			;self-evaluating " << exp << endL;
	*code	<< arg_PUSH(exp);
									val=exp;		// self-evaluating
					}
					Push(th->argStack, val, arg_loop);
					next(unev);
					goto arg_loop_again;
	 arg_is_function:
					if (exp==_Not_Enough_node && !throwing) 
					{	error(NOT_ENOUGH, 0); 
	if ((th->argStack->index-th->argStack->base)!=asp)
		*code << arg_POP_TILL_asp;
						th->argStack->pop_till(asp);
						Goto( args_done )
					}
	*code	<< GET_ASP(CX)
			<< PUSH(CX);
					Push(th->argPStack, asp, arg_is_function);
					Push7(th->stack, unev,proc, 
						th->fun,th->ufun,th->this_line, 
						th->last_ufun,th->last_line, arg_is_function);
					PushError(th->errorStack, ObjectP(exp->first()), th->ufun, th->this_line, arg_is_function);
					th->didnt_output_name=0;

					th->var=*th->varStack;
					th->tailcall = tailcall_arg;
					th->val_status=value_required;

					Push(th->toDoStack, __accumulate_arg, arg_is_function); 
					Goto( dispatch )   // evaluate the current argument 
TO( accumulate_arg_this )
					Pop1(th->intStack, th->proccallInc, accumulate_arg);
					asp-=th->proccallInc;
					th->argStack->pop_till(asp);
					Pop1(th->stack, th->thisobj, accumulate_arg);
					Goto( accumulate_arg )
TO( accumulate_arg )	// Put the evaluated argument on the argStack.
					if (val && !throwing)
					{	if (val->type==PROCCALL)
						{	Push(th->stack, th->thisobj, accumulate_arg);
							th->thisobj=LogoObjectP(PairP(val)->first());
							proc=PairP(val)->tail();
							th->fun=ProcedureP(proc)->object();
							asp+=th->proccallInc;
							Push(th->intStack, th->proccallInc, accumulate_arg);
							val=Unbound;
							Push(th->toDoStack, __accumulate_arg_this, accumulate_arg); 
							Goto( compound_apply )
						}
						else if (val==Unbound 
						&&	(  th->val_status==value_required
							|| th->val_status==output_ok)
						&& th->lexufun
						&& th->lexufun->type==OBJECT)
							val=new LogoObject(th->lexufun);
					}
					th->varStack->pop_till(th->var);
					th->last_call=th->fun;
					Pop(th->errorStack, arg_is_function);
					Pop7(th->stack, unev, proc, 
						th->fun, th->ufun, th->this_line, 
						th->last_ufun, th->last_line, arg_is_function);
	*code << POP(CX)
			<< SET_ASP(CX);
					Pop1(th->argPStack, asp, arg_is_function);

	*code << arg_PUSH_AX;	// no error checking yet!
					if (!unev)
						error(BAD_DATA, 0);
					else
						next(unev);
//					if (val==Unbound && !throwing) 
//					{	error(DIDNT_OUTPUT, th->fun);
//						Goto( arg_loop )
//					}
					if (!throwing)
						Push(th->argStack, val, accumulate_arg);
					Goto( arg_loop )
TO( args_done )
					Pop1(th->stack, th->didnt_output_name, proc_with_args);
					Pop1(th->varPStack, th->var, proc_with_args);
					Pop3(th->intStack, th->val_status, th->tailcall, th->ift_iff_flag, proc_with_args);

					th->last_call=th->fun;
					if (throwing) 
					{	val=Unbound;
						Goto( fetch_command )
					}
					Goto( apply_dispatch )

TO( apply_dispatch )	// Load in the procedure's definition and decide 
							// whether it's a compound procedure or a primitive procedure.
					if (!th->fun)
					{	error(BAD_DATA, 0); 
						// better than a direct crash, but no good error msg yet.
						lpause(PairPP(th->argStack->index));
						Goto( fetch_command )
					}
					else
					{	proc=ObjectP(th->fun)->proc_or_prim();
						if (!ObjectP(th->fun)->bit.proc)
						{	if (th->varStack->isvar(ObjectP(th->fun)))
							{	NodeP v=th->varStack->find(ObjectP(th->fun));	
								// fixing colonless var before output
								val=v;	
								Goto( fetch_command );
							}
						}
					}
	//				if ( ObjectP(th->fun)->bit.proc_macro )
	//				{	Push2(th->intStack, th->val_status, th->tailcall, apply_dispatch);
	//					th->val_status=value_required;
	//					Push(th->toDoStack, __macro_return, apply_dispatch);
	//				}
					if (proc->is.primitive)		// primitive_apply
					{	if (!ObjectP(th->fun)->bit.proc_traced) 
						{
						FunctionP f=PrimitiveP(proc)->function();

	if (!ObjectP(th->fun)->bit.proc_traced) 
	{	if (ObjectP(th->fun)->bit.proc_macro 
//		|| PrimitiveP(proc)->function()==_bye
//		|| PrimitiveP(proc)->function()==_callCode
//		|| PrimitiveP(proc)->function()==_local
//		|| PrimitiveP(proc)->function()==_make
		)
		{	c_out << "			;compiler macro " << th->fun << endL;
			if (PrimitiveP(proc)->function()==lrepeat)
			{	val = lcompileRepeat((PairPP)&th->argStack->base[asp]);
				Push(th->toDoStack, __fetch_command, apply_dispatch); 
				Goto( repeat_continuation )
			}else if (PrimitiveP(proc)->function()==lforever)
			{	val = lcompileForever((PairPP)&th->argStack->base[asp]);
				Push(th->toDoStack, __fetch_command, apply_dispatch); 
				Goto( forever_continuation )
			}else if (PrimitiveP(proc)->function()==lif)
			{	val = lcompileIf((PairPP)&th->argStack->base[asp]);
				Push(th->toDoStack, __fetch_command, apply_dispatch); 
				if (PairP(val)->tail())
					Goto( ifElse_continuation )
				else
					Goto( if_continuation )
			}else if (PrimitiveP(proc)->function()==lifElse)
			{	val = lcompileIfElse((PairPP)&th->argStack->base[asp]);
				Push(th->toDoStack, __fetch_command, apply_dispatch); 
				Goto( ifElse_continuation )
			}else if (PrimitiveP(proc)->function()==lwhile)
			{	val = lcompileWhile((PairPP)&th->argStack->base[asp]);
				Push(th->toDoStack, __fetch_command, apply_dispatch); 
				Goto( while_continuation )
			}else if (PrimitiveP(proc)->function()==lrun)
			{	val = Unbound;
				cdecl_asp_CALL(evaluate, ObjectP(th->fun));
			}else if (PrimitiveP(proc)->function()==lrunResult)
			{	val = Unbound;
				cdecl_asp_CALL(evaluate, ObjectP(th->fun));
			}else 
				val = PrimitiveP(proc)->call(
					(PairPP)&th->argStack->base[asp]);	// TODO
		}else if (PrimitiveP(proc)->function()==loutput)
		{	*code << MOVrm(AX,&th->argStack->index)
					<< LEAVE
					<< RET;
		}else if (PrimitiveP(proc)->function()==lstop)
		{	*code << RET_Unbound;
		}else if (PrimitiveP(proc)->function()==llocal)
		{	val = compileLocal((PairPP)&th->argStack->base[asp]);
		}else if (PrimitiveP(proc)->function()==lmake)
		{	val = compileMake((PairPP)&th->argStack->base[asp]);
		}else
		{	c_out << "			;call " << th->fun << endL;
			cdecl_asp_CALL(f, ObjectP(th->fun));
			//fast_asp_CALL(f);
			if (th->argStack->index-th->argStack->base != asp)
				*code << arg_POP_TILL_asp;
			val=Unbound;
		}
	}
//							val = PrimitiveP(proc)->call(
//								(PairPP)&th->argStack->base[asp]);	// call the primitive
							if (!val || val->type!=PROCCALL)
								th->argStack->pop_till(asp);
						}else
						{	tracing = ObjectP(th->fun)->bit.proc_traced;
							if (th->lineInfoInTrace)
							{	c_out << '[' << th->this_line->col << th->this_line->row << ']';
							}
							for (int i=0; i < th->trace_level; i++) 
								c_out << ' ';
							c_out << "(" << th->fun << ' ';
							if (asp!=*th->argStack)
								for (NodePP arg=th->argStack->base+asp+1;  
								arg<=NodePP(th->argStack->index);  arg++)
								{	c_out << (*arg)->maybe_quote();
									if (arg+1<=NodePP(th->argStack->index))
										c_out << ' ';
								}
							c_out << ')' << endL;
						
							val = PrimitiveP(proc)->call(
								(PairPP)&th->argStack->base[asp]);	// call the primitive
							if (!val || val->type!=PROCCALL)
								th->argStack->pop_till(asp);

							if (!throwing) 
							{	for (int i=0;  i < th->trace_level;  i++) 
									c_out << ' ';
								c_out << th->fun;
								if (val==Unbound)
									c_out << _(" stops") << endL;
								else 
									c_out << _(" outputs ") << val->maybe_quote() << endL;
							}
						}
					}
					else if (proc->type==PROC_ || proc->type==INFIXPROC)
					{	if (ObjectP(th->fun)->proc()->compiled())
						{	c_out << "			;call " << th->fun << endL;
							FunctionP f=ProcedureP(proc)->compiled()->code;
							cdecl_asp_CALLl(f, ObjectP(th->fun));
							if (th->argStack->index-th->argStack->base != asp)
								*code << arg_POP_TILL_asp;
							val=Unbound;
							Goto( fetch_command )
						}else
						{	c_out << "			;call " << th->fun << endL;
							cdecl_asp_CALLl(0, ObjectP(th->fun));
							if (th->argStack->index-th->argStack->base != asp)
								*code << arg_POP_TILL_asp;
							Goto( compound_apply )
						}
					}
					else 
					{	//if (ufun) 
						//	untreeify_proc(ufun);
						if (!throwing) val=error(DK_HOW, th->fun);
					}
					Goto( fetch_command )

TO( compound_apply )
					if (!ObjectP(th->fun)->bit.permanent && th->thisobj)
					{	TableP t=th->thisobj->table();
						ObjectP o=t->find(ObjectP(th->fun)->name());
						if (o && o->bit.proc)
						{	proc=o->proc();
							th->fun=ProcedureP(proc)->object();							
						}
					}
					th->checkBreak();
					tracing = ObjectP(th->fun)->bit.proc_traced;
					if (tracing) 
					{	if (th->lineInfoInTrace)
						{	if (unev)
								c_out << '[' << LineP(unev)->col << ' ' << LineP(unev)->row << ']';
							else if (th->this_line && th->this_line->first())
								c_out << '[' << LineP(th->this_line->first())->col << ' '
									<< LineP(th->this_line->first())->row << ']';
						}
						for (int i=0; i < th->trace_level; i++) 
							writer << ' ';
						th->trace_level++;
						writer << "( " << th->fun << ' ';
					}
					Goto( lambda_apply )
TO( lambda_apply )					// Bind the actuals to the formals 
					vsp=*th->varStack;	// remember where we came in 

//					if (th->tailcall!=tailcall_tail || th->fun!=th->ufun || !th->ufun)
					{	Push(th->intStack, th->ift_iff_flag, lambda_apply);
						Push(th->stack, th->lexufun, lambda_apply);
	Push(th->stack, code, lambda_apply);
						Push(th->toDoStack, __compound_apply_pop, lambda_apply);
						if (th->fun->type==OBJECT && !ObjectP(th->fun)->bit.proc_dyn)
						{	Push3(th->frameStack, th->lastFrame, th->frame, vsp, lambda_apply);
							Push2(th->stack, th->ufun, th->parents, lambda_apply);
							if (!th->ufun || ObjectP(th->fun)->parent()==th->ufun)
							{	th->lastFrame=*th->frameStack-2;
								th->lexufun=ObjectP(th->fun);
							}
							th->frame=vsp;
							th->ift_iff_flag=-1;
							Push(th->toDoStack, __compound_apply_end, lambda_apply);
						}
					}
					Push(th->argPStack, asp, lambda_apply);
					formals=ProcedureP(proc)->formals();
	code=new Compiled(ObjectP(th->fun));
	ObjectP(th->fun)->proc()->setcompiled(code);
	*code << ENTER(16);
	*code	<<	CALL(Thread::This)
			<<	SET_th(AX);
					Goto( begin_formal_loop )
TO( begin_formal_loop )
					formalNr=0;
					if (formals)
					{	parm=formals->first();
						if (type_of(parm)==INT_)	// default # args 
							Goto( after_formals )

						if (asp != *th->argStack)
						{	arg = PairP(th->argStack->base[++asp]);
							if (tracing) 
								writer << arg->maybe_quote() << ' ';
						}else
							arg = PairP(Unbound);

						switch (type_of(parm))
						{
						case OBJECT: 
							{	if (arg)
									if (arg->type==VARPARAM)
									{	error(BAD_DATA, arg); break;
									}
								ObjectP parmobj = ObjectP(parm);
/*	*code 
		<< PUSH(BP)
		<< MOV(BP,SP)
		<< MOVrc(AX,arg)
		<< PUSH(AX)
		<<	MOVrc(CX,&(th->varStack->index))
		<< PUSH(CX)
		<< MOVrc(DX,parmobj)
		<<	MOVrc(CX,vsp);
	*code
//		<< CALL(Object_not_local)
		<< POP(BP);
	c_out	
		<< " vsp=" << vsp
		<< " parmobj=" << parmobj 
		<< " arg=" << arg 
		<< " varStack=" << th->varStack->index-th->varStack->base << endL;
*/
								if (parmobj->not_local(vsp)) 
								{	PushVar(th->varStack, parmobj, arg, begin_formal_loop);
/*	formalNr++;
	*code <<	GET_ASP(CX)
			<<	ADDrc(CX,4)
			<< SET_ASP(CX)
			<<	ADDrm(CX,&th->argStack->base)
			<<	MOVri(AX,CX)
			<<	arg_PUSH_AX;
*/								}else
									th->varStack->setVar(parmobj, arg);
								parmobj->tell_shadow();
								break;
							}
						case VARPARAM: 
							{	ObjectP parmobj = ColonP(parm)->colon();
								if (!arg)
								{	error(BAD_DATA, arg); break;
								}
								if (arg->type!=VARPARAM)
								{	error(BAD_DATA, arg); break;
								}
								ObjectP alias=ColonP(arg)->colon();
								th->varStack->push_alias(parmobj, alias);
								break;
							}
						case LINE:
						case PAIR:		// parm is optional or rest 
							{	ObjectP parm1 = ObjectP(parm->first());

								if (!parm->tail())	// parm is rest 
								{
									NodeP value=th->argStack->to_list(asp-1);	// SLOW 8-(
									if (parm1->not_local(vsp)) 
									{	PushVar(th->varStack, parm1, value, begin_formal_loop);
/*	formalNr++;
	*code <<	GET_ASP(CX)
			<<	ADDrc(CX,4)
			<< SET_ASP(CX)
			<<	ADDrm(CX,&th->argStack->base)
			<<	MOVri(AX,CX)
			<<	arg_PUSH_AX;
*/									}else
										th->varStack->setVar(parm1, value);

									parm1->tell_shadow();
									if (tracing && asp!=*th->argStack) 
									{	NodePP arg=(NodePP)&th->argStack->base[asp];
										while (arg!=NodePP(th->argStack)) 
											writer << (*++arg)->maybe_quote() << ' ';
									}
									Goto( after_formals )
								}

								if (arg==Unbound)		// use default 
								{
									Push2(th->varPStack, th->var, vsp, begin_formal_loop);
									Push8(th->stack, 
										th->didnt_output_name, proc,  
										th->fun, th->ufun, th->this_line,  
										th->last_ufun, th->last_line,
										formals, begin_formal_loop);

									Push(th->argPStack, asp, begin_formal_loop);
									Push2(th->intStack, th->ift_iff_flag, th->val_status, begin_formal_loop);

									th->var=*th->varStack;
									th->tailcall = tailcall_arg;
									th->val_status=value_required;

									nodeList=parm->tail();
									if (!throwing)
									{	ObjectP olexufun=th->lexufun;
										if (ObjectP(th->fun)->bit.proc_dyn)
											th->lexufun=0;
										else
											th->lexufun=ObjectP(th->fun);
										nodeList->makeTree();
										th->lexufun=olexufun;
									}else
										nodeList=0;

									if (nodeList && nodeList->type==TREE) 
									{	unev = TreeP(nodeList)->tree();
										val=Unbound;
										Push(th->toDoStack, __set_args_continue, begin_formal_loop);
										Goto( sequence )
									}else
										val=Unbound;
									Goto( set_args_continue )
								}
								Goto( set_args_continue2 )
							}
						default: ;
						}
						Goto( end_formal_loop )
					}
					if (asp != *th->argStack)
						error(TOO_MUCH, formals);
					Goto( after_formals )
TO( set_args_continue )
					Pop1(th->varPStack, vsp, begin_formal_loop);
					Pop1(th->argPStack, asp, begin_formal_loop);
					Pop1(th->stack, formals, begin_formal_loop);
					parm=formals->first();
					th->varStack->pop_till(th->var);
					Pop2(th->intStack, th->ift_iff_flag, th->val_status, begin_formal_loop);
					Pop7(th->stack, 
						th->didnt_output_name, proc,  
						th->fun, th->ufun, th->this_line,  
						th->last_ufun, th->last_line, begin_formal_loop);
					Pop1(th->varPStack, th->var, begin_formal_loop);

					arg=PairP(val);
					Goto( set_args_continue2 )
TO( set_args_continue2 )
					ObjectP parmobj=parm->first()->intern();
					if (ObjectP(parmobj)->not_local(vsp)) 
					{	PushVar(th->varStack, parmobj, arg, set_args_continue2);
	formalNr++;
	*code <<	arg_PUSH(arg);
					}else
						th->varStack->setVar(parmobj, arg);
					ObjectP(parm->first())->tell_shadow();
//					ObjectP(parm->first())->setvar( arg );
					Goto( end_formal_loop )
TO( end_formal_loop )
					next(formals);
					Goto( begin_formal_loop )
TO( after_formals )
					Pop1(th->argPStack, asp, lambda_apply);
//	if (th->argStack->index-th->argStack->base != asp)
//		*code << arg_POP_TILL_asp;
/*	*code	<< GET_ASP(CX)
			<<	ADDrc(CX, formalNr*4)
			<<	SET_ASP(CX);
*/					th->argStack->pop_till(asp);

					if (checkThrowing()) 
					{	val=Unbound;
						Goto( fetch_command )
					}
					if ((th->tailcall==in_sequence) 
					|| th->ufun!=th->last_ufun || !th->ufun)
					{	Push(th->funStack, th->fun, after_formals);
						Push2(th->argNrStack, vsp,
							th->varStack->index-th->varStack->base-th->var, after_formals);
						Push(th->toDoStack, __compound_apply_end_pop_fun, after_formals);
					}
					vsp=0;
					tracing= !th->fun->is.list;
					if (tracing && ObjectP(th->fun)->bit.proc_traced) 
					{	if (!throwing) 
							writer << ')';
						*th->Writer << endL;
						Push(th->stack, th->fun, after_formals);
						Push(th->toDoStack, __compound_apply_continue, after_formals);
					}
					val=Unbound;
					th->last_ufun=th->ufun;

			//		if (!fun->is.list)	
					// commented out to make apply in procedure text form work
					if (th->fun->type==OBJECT 
					&& !ObjectP(th->fun)->bit.proc_dyn)
						th->lexufun=th->fun->intern();
					else
						th->lexufun=0;
					th->ufun=th->fun->intern();
					
					th->last_line=th->this_line;
					th->this_line=0;
					proc = th->fun->is.list ?
						LineP(th->fun)->anonymous_function()
						: ObjectP(th->fun)->proc();

					nodeList=ProcedureP(proc)->bodylist();		// get the body ... 
					
					if (th->fun->type!=OBJECT 
					|| (th->fun->type==OBJECT && ObjectP(th->fun)->bit.proc_dyn))
					{	ObjectP olexufun=th->lexufun;
						th->lexufun=0;
						nodeList->makeTreeFromBody();
						th->lexufun=olexufun;
					}else
						nodeList->makeTreeFromBody();

					if (!nodeList || nodeList->type!=TREE 
					|| !TreeP(nodeList)->tree()) //Kludge to make empty to def work
						Goto( fetch_command )

					unev = TreeP(nodeList)->tree();
					if (!throwing) 
						th->stoppingFlag=RUN;

					th->output_node=PairP(Unbound);

					switch (th->val_status)
					{	case value_required:				th->val_status=output_ok; break;
						case value_maybe_ok_in_macro:	th->val_status=dont_care; break;
						default:								th->val_status=no_value_allowed;
					}
					Goto( sequence )
TO( begin_sequence )
					nodeList->makeTree();
					if (!nodeList  ||  nodeList->type != TREE)
					{	val=Unbound;
						Goto( fetch_command )
					}
					unev = TreeP(nodeList)->tree();
					val=Unbound;
					Goto( sequence )
TO( sequence )	// Evaluate each expression in the sequence.  
					if (!running  ||  val != Unbound) 
						Goto( fetch_command )		// Stop as soon as  val!=Unbound.

					if (unev && unev->is.line) 
					{	if (Line::the_generation != LineP(unev)->generation && th->ufun)
						{	// something redefined while we're running 
							LineP bl = LineP(th->ufun->proc()->bodylist());
							if (bl && bl->is.line)
							{	int col=LineP(unev)->col, row=LineP(unev)->row;
								LineP l = TreeP(bl)->tree(), t=0;
//c_out << l << endL;
//console->dispatchMessages();
								t=findLine(col, row, l);
//c_out << t << endL;
//console->dispatchMessages();
//c_out << '[' << col << ' ' << row << ']' << endL;
								if (t)
								{	bl->untreeify();
//c_out << 'B' << bl << 'X';
									th->setfilename(bl->filename());
									th->row=bl->row;
									th->col=bl->col;
									bl->makeTreeFromBody();
									l = TreeP(bl)->tree();
									t=findLine(col, row, l);
									if (t)
									{	unev=t;
//c_out << "A" << t << "Z" << endL;
									//	Goto( begin_sequence )
									}
									else
									{	c_out << "Bug: line not found: " << unev << endL; //lpause(0);
//										t=findLine(col, row, l);
									}
								}
								// else line[col row] not found, so we cant fix it (we are in trouble!!!)
								else
								{	c_out << "Bug: line not found: " << unev << endL; //lpause(0); 
//									t=findLine(col, row, l);
								}
							}
						}
					}
//c_out << "A" << unev << "Z" << endL;
					if (unev && unev->type==TREE) 
					{	th->this_line = TreeP(unev)->tree();

						if (th->ufun  
						&&  ObjectP(th->ufun)->bit.proc_stepped
						&&  !th->inpause) 
						{	if (tracing) 
							{	c_out << '(' << debugStates.last 
									//<< '|' << th->this_line->linenr
									<< ')';
								for (int i=1;  i < th->trace_level;  i++) 
									c_out << ' ';
							}
//							c_out << this_line;
//c_out << th->this_line->row << th->this_line << endL;
							if (!th->this_line->filename()->equal(_Unnamed))
							{	theApp->editorShow();
								if (theApp->edFrame)
								{	theApp->edFrame->unmark(1);
									int col, row;
									if (th->this_line->first() 
									&& th->this_line->first()->is.list)
									{	col=LineP(th->this_line->first())->col;
										row=LineP(th->this_line->first())->row;
									}else
									{	col=th->this_line->col;
										row=th->this_line->row;
									}
									theApp->edFrame->mark(
										th->this_line->filename()->ptr(), 
										col, row, 1);
								}
							}
							bool oldinstep=th->instep;
							th->instep=true;
							lupdateCalls(0);
							if (th->updateVarsOnStep)
							{	lupdateVars(0);
							}
							if (th->stepNr < 10)
								wxMilliSleep(200-20*th->stepNr);
							
							if (console)
							{	console->flush();
								console->updateGraph();
							}
							if (th->stepNr < 20  || (th->stepNr & 0x3)==0)
								th->dispatchMessages();
							th->instep=oldinstep;
#ifndef _DEBUG
							if (th->stepNr==0 || !th->stepArmed)
#endif
							{	oldinstep=th->instep;
								th->instep=true;
								th->stepArmed=false;
								c_in.read("");
								th->instep=oldinstep;
							}
							th->stepNr++;
							if (theApp->edFrame)
								theApp->edFrame->unmark(1);
						}
					}

					exp=unev->first();
					wxASSERT(unev->is.list || unev->type==QUOTE || unev->type==COLON);
					th->this_line=LineP(unev);

					if (th->ufun)
						if	(ObjectP(th->ufun)->bit.proc_stepped
						&&  !th->inpause) 
						{
//static NodeP t0=0;
//static NodeP t1=0;
//t0=lTimeFine(0);
							if (tracing) 
							{	if (th->this_line)
									c_out << '(' << debugStates.last 
										//<< '|' << th->this_line->linenr 
										<< ')';
								for (int i=1;  i < th->trace_level;  i++) 
									c_out << ' ';
							}
//							c_out << exp;//this_line->first();
//c_out << th->this_line->row << th->this_line << endL;
							if (th->this_line  &&  th->this_line->is.line)
								if (!th->this_line->filename()->equal(_Unnamed))
								{	if (!theApp->edFrame)
										theApp->editorShow();
									if (theApp->edFrame)
									{	theApp->edFrame->unmark(1);
										int col, row;
										if (th->this_line->first() 
										&& th->this_line->first()->is.list)
										{	col=LineP(th->this_line->first())->col;
											row=LineP(th->this_line->first())->row;
										}else
										{	col=th->this_line->col;
											row=th->this_line->row;
										}
										theApp->edFrame->mark(
											th->this_line->filename()->ptr(), 
											col, row, 1);
									}
								}
//t1=lTimeFine(0);
							bool oldinstep=th->instep;
							th->instep=true;
								
							lupdateCalls(0);
							if (th->updateVarsOnStep)
							{	lupdateVars(0);
							}
							if (th->stepNr < 10)
								wxMilliSleep(200-20*th->stepNr);

							if (console)
							{	console->flush();
								console->updateGraph();
							}
							if (th->stepNr < 20  || (th->stepNr & 0x3)==0)
								th->dispatchMessages();
							th->instep=oldinstep;
#ifndef _DEBUG
							if (th->stepNr==0 || !th->stepArmed)
#endif
							{	oldinstep=th->instep;
								th->instep=true;
								th->stepArmed=false;
								c_in.read("");
								th->instep=oldinstep;
							}
							th->stepNr++;
							if (theApp->edFrame)
								theApp->edFrame->unmark(1);
//t8=lTimeFine(0);
//c_out << t1->sub(t0) << "seconds1" << endL;
						}
					next(unev);

					if (exp->is.list
					&& ObjectP(exp->first())->proc_or_prim()->type==TAILFORM)
					{
						PrimitiveP expProc = ObjectP(exp->first())->prim();
						Priority pri = expProc->_priority;

						switch (pri)
						{
						case priority_OUTPUT: 
								PushError(th->errorStack, 
									ObjectP(exp->first()), th->ufun, th->this_line, sequence);
								th->didnt_get_output = th->errorStack->top();
								Pop(th->errorStack, sequence);	// pretty bad, this. Should be in some other places.
								th->didnt_output_name=0;

								if (!th->ufun && !throwing && th->val_status!=output_ok) 
								{	error(AT_TOPLEVEL, exp->first());
									val=Unbound; Goto( fetch_command )
								}else 
									switch (th->val_status)
									{
									case output_ok:
									case dont_care:
											th->val_status=value_required;
											if (exp->tail())
												exp = exp->second();
											else
												error(BAD_DATA, exp);
											Goto( tail_dispatch )

									case no_value_allowed:
									case value_required:
											th->val_status=value_required;
											exp = exp->second();
											unev=0;
											Goto( non_tail_eval )    // compute value then give error
									default: ;
									}
								break;

						case priority_STOP:
								if (!th->ufun && !throwing && th->val_status!=output_ok) 
								{	error(AT_TOPLEVEL, exp->first());
									val=Unbound; Goto( fetch_command )
								}else 
									switch (th->val_status)
									{
									case value_required:
									case output_ok:			th->didnt_output_name=th->fun;
									case no_value_allowed:
									case dont_care:			val=Unbound; 
	*code	<< RET_Unbound;
																	Goto( fetch_command )
									default: ;
									}
								break;
						

						default:	// maybeoutput
								if (exp->tail())
									exp = exp->second();
								else
									error(BAD_DATA, exp);
								if (exp && !exp->first() && !exp->tail()) // new in 4.684
								{	val=Unbound;
									Goto( fetch_command )
								}
								th->val_status=value_maybe_ok_in_macro;
								Goto( tail_dispatch )
						}
					}
					if (!unev) 
					{	switch (th->val_status)
						{
						case output_ok:
						case no_value_in_macro:
								th->didnt_output_name=th->fun;
								unev = (PairP)Unbound;
								Goto( non_tail_eval )

						default: ;
						}
						Goto( tail_dispatch )
					}
					if (unev->first())
					{
						Pair &lst=*PairP(unev->first());
						if (        lst.islist()
						&&  ObjectP(lst.first())->proc_or_prim()->type==TAILFORM 
						&&  ObjectP(lst.first())->prim()->_priority==priority_STOP
						&&  th->ufun) 
						{	switch (th->val_status)
							{
							case value_required:
							case output_ok:			th->didnt_output_name=th->fun;
							case no_value_allowed:
							case dont_care:			Goto( tail_dispatch )
							default: ;
							}
						}
					}
					Goto( non_tail_eval )
TO( non_tail_eval )
					Push(th->varPStack, th->var, proc_with_args);
					Push7(th->stack, unev, proc,  
						th->fun, th->ufun, th->this_line,  
						th->last_ufun, th->last_line, proc_with_args);
					Push2(th->intStack, th->ift_iff_flag, th->val_status, proc_with_args);
					th->var = *th->varStack;
					th->tailcall = in_sequence;
					Push(th->toDoStack, __sequence_continue, non_tail_eval);	Goto( dispatch )
TO( sequence_continue_this )
					Pop1(th->intStack, th->proccallInc, sequence_continue);
					asp-=th->proccallInc;
					th->argStack->pop_till(asp);
					Pop1(th->stack, th->thisobj, sequence_continue);
					Goto( sequence_continue )
TO( sequence_continue )
					if (val)
						if (val->type==PROCCALL)
						{	Push(th->stack, th->thisobj, sequence_continue);
							th->thisobj=LogoObjectP(PairP(val)->first());
							proc=PairP(val)->tail();
							th->fun=ProcedureP(proc)->object();
							asp+=th->proccallInc;
							Push(th->intStack, th->proccallInc, sequence_continue);
							val=Unbound;
							Push(th->toDoStack, __sequence_continue_this, sequence_continue); 
							Goto( compound_apply )
						}
					th->varStack->pop_till(th->var);
					Pop1(th->varPStack, th->var, proc_with_args);
					Pop7(th->stack, unev, proc,  
						th->fun, th->ufun, th->this_line,  
						th->last_ufun, th->last_line, proc_with_args);

 					if (th->dont_fix_ift) 
					{	Pop2(th->intStack, th->dont_fix_ift, th->val_status, proc_with_args);
						th->dont_fix_ift=false;
					}else
						Pop2(th->intStack, th->ift_iff_flag, th->val_status, proc_with_args);

					if (th->stoppingFlag==MACRO_RETURN) 
					{
						if (unev==Unbound) unev=0;

						bool fromGoto=false;
						if (val)
							if (val->is.list  &&  PairP(val)->first()==NodeP(_Tag))
								fromGoto=true;
						if (fromGoto)
							unev = PairP(val)->tail();	// from goto 
						else
						{	StringP fn=th->filename();
							if (val)
								th->setfilename(LineP(val)->filename());
							if (unev)
								unev = PairP(val)->append(unev);	// slow
							else
								unev=PairP(val);
							th->setfilename(fn);
						}
						val=Unbound;
						th->stoppingFlag=RUN;

						if (!unev) Goto( fetch_command )
					}
					else 
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound);
												break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (th->val_status==value_required  &&  !throwing)
												{	if (val==Unbound) 
														th->didnt_output_name=_Stop;
													else
														th->didnt_output_name=_Output;

													error(DIDNT_OUTPUT, _Output);
												}
												Goto( fetch_command )
								default: ;
								}
						default: ;
						}
	*code	<<	GET_th_stoppingFlag(DX)
			<<	CMPrc(DX,STOP);
	int if_stop=code->length;
	*code	<< JMPc(NZ,0);
	*code	<< MOVrc(CX,Unbound)
			<< SET_th_output_node(CX)
			<< RET_Unbound;
	*(int*)(charP(code->code)+if_stop+2)=code->length-if_stop-6;	

					if (val!=Unbound && !throwing 
					&& th->val_status!=dont_care 
					&& th->val_status!=output_ok
					&& th->val_status!=value_required) 
					{	error(!unev ?	DK_WHAT_UP
										:  DK_WHAT,  
									val);
						val=Unbound;
					}
					if (!throwing && (!unev || unev==Unbound)) 
					{	if (val==Unbound 
						&&	(  th->val_status==value_required
							|| th->val_status==output_ok)
						&& th->lexufun)
							val=new LogoObject(th->lexufun);
						else if (val==Unbound 
						&& th->val_status == value_required) 
							error(DIDNT_OUTPUT, th->fun);
						Goto( fetch_command )
					}
					Goto( sequence ) 

TO( compound_apply_continue )	// Only get here if tracing 
					Pop1(th->stack, th->fun, after_formals);
					--th->trace_level;
					if (!throwing) 
					{	if (th->lineInfoInTrace)
							if (th->this_line)
								c_out << '[' << th->this_line->col << ' ' << th->this_line->row << ']';
						for (int i=0; i < th->trace_level; i++) 
							writer << ' ';
						writer << th->fun;
						if (val==Unbound)
							writer << _(" stops") << endL;
						else 
							writer << _(" outputs ") << val->maybe_quote() << endL;
					}
					Goto( fetch_command )

TO( compound_apply_end_pop_fun )
					Pop(th->funStack, after_formals);
					Pop(th->argNrStack, after_formals);
					Pop(th->argNrStack, after_formals);
					Goto( fetch_command )

TO( compound_apply_end_this )
					Pop1(th->intStack, th->proccallInc, compound_apply_end_this);
					Pop1(th->stack, th->thisobj, compound_apply_end_this);
					if (th->tailcall != tailcall_tail)
					{	asp-=th->proccallInc;
						th->argStack->pop_till(asp);
					}
					Goto( compound_apply_end )

TO( compound_apply_end )
					if (!throwing)
					{	if (val && val->type==PROCCALL)
						{	//th->tailcall=in_sequence;
							//if (th->tailcall != tailcall_tail)
							{	Push(th->stack, th->thisobj, compound_apply_end_this);
								Push(th->intStack, th->proccallInc, compound_apply_end_this);
							}
							th->thisobj=LogoObjectP(PairP(val)->first());
							proc=PairP(val)->tail();
							th->fun=ProcedureP(proc)->object();
							asp+=th->proccallInc;
							val=Unbound;
							//if (th->tailcall != tailcall_tail)
								Push(th->toDoStack, __compound_apply_end_this, compound_apply_end); 
							//else
							//	Push(th->toDoStack, __compound_apply_end); 
							Goto( compound_apply )
						}
	/*					if (val==Unbound 
						&&	(  th->val_status==value_required
							|| th->val_status==output_ok)
						&& th->lexufun
						&& th->lexufun->type==OBJECT)
							val=new LogoObject(th->lexufun);
	*/				}
					th->varStack->pop_till(th->frame);
					Pop3(th->frameStack, th->lastFrame, th->frame, vsp, lambda_apply);
					Pop2(th->stack, th->ufun, th->parents, lambda_apply);
					Goto( fetch_command )
TO( compound_apply_pop )
	*code << LEAVE 
			<< RET;
	code->setassembly(PairP(code->assembly()->reverse()));
	if (th->fun)
		ObjectP(th->fun)->proc()->setcompiled(code);
	Pop1(th->stack, (NodeP&)code, lambda_apply);
					Pop1(th->intStack, th->ift_iff_flag, lambda_apply);
					Pop1(th->stack, th->lexufun, lambda_apply);
					Goto( fetch_command )	

TO( begin_apply )							// This is for APPLY.
					th->fun=PairP(val)->first();
					
					if (th->fun->isnolist() 
					&& !(th->fun && (th->fun->type==OBJECT || th->fun->is.string))  
					&& !throwing)
						error(DK_HOW_UNREC, th->fun, 1);

//					if (!PairP(val)->tail() && !throwing)
//						error(NOT_ENOUGH, val, 2);
					if ((PairP(val)->tail()->isnolist() 
					&& PairP(val)->tail()->type!=ARRAY) && !throwing)
						error(APPLY_BAD_DATA, PairP(val)->tail(), 2);

					int applyArgCount=0;
					if (!throwing)
					{	asp=*th->argStack;
	SET_ASP_from_argStack;
						if (PairP(val)->tail())
							if (PairP(val)->tail()->is.list)
								PushList(th->argStack, PairP(val)->tail(), begin_apply);
							else if (PairP(val)->tail()->type==ARRAY)
								Push(th->argStack, ArrayP(PairP(val)->tail()), begin_apply);
							else
								error(APPLY_BAD_DATA, val, 2);
						else if (!PairP(val)->first())
							error(NOT_ENOUGH, val, 2);
						applyArgCount=th->argStack->index-th->argStack->base-asp;
						val=Unbound;
					}
					if (!throwing && th->fun) 
					{	if (th->fun->islist())		    // template 
						{	PairP lst=PairP(th->fun);
							if (lst->first()->islist() && lst->tail()) 
							{	if (lst->second()->islist())		// procedure text form 
								{	proc = LineP(lst)->anonymous_function();
									tracing=0;
									Goto( lambda_apply )
								}
								formals=lst->first();		// lambda form 

								Push(th->varPStack, th->var, begin_apply);
								Push(th->intStack, th->tailcall, begin_apply);
								th->tailcall = in_sequence;
//								Push(th->argPStack, asp);
//								asp=th->argStack;
//								Push(th->argStack, formals);
//								llocal((PairPP)&th->argStack.base[asp]);	
								// bind the formals locally 
//								*th->argPStack >> asp;
								Push(th->argPStack, asp, begin_apply);
								Pop1(th->intStack, th->tailcall, begin_apply);
								
								int i=applyArgCount;
								while (formals &&  asp!=*th->argStack  && !throwing)
								{	
									if (!(i > 0))
									{	error(NOT_ENOUGH, formals, 2);
										break;
									}
//									ObjectP(formals->first())->
//										setvar( th->argStack.base[++asp] );
									PushVar(th->varStack, ObjectP(formals->first()->intern()), 
										th->argStack->base[++asp], begin_apply);
									next(formals); 
									i--;
								}
								if (i > 0)
									error(TOO_MUCH, val, 2);

								Pop1(th->argPStack, asp, begin_apply);
//						th->argStack->pop_till(asp);
								nodeList = lst->tail();
								Push(th->argPStack, th->qmList, begin_apply);

								Push(th->toDoStack, __after_lambda, begin_apply); 
								goto lambda_questionmark;
							}else		// question-mark form 
							{	Push(th->argPStack, th->qmList, begin_apply);
								th->qmList=asp;
								nodeList=lst;
	lambda_questionmark:
								Push(th->argPStack, asp, begin_apply);

//					ObjectP olexufun=th->lexufun;
//					if ((th->ufun->type==OBJECT 
//					&& ObjectP(th->ufun)->bit.proc_dyn))
//						th->lexufun=0;
								nodeList->makeTree();
//					if ((th->ufun->type==OBJECT 
//					&& ObjectP(th->ufun)->bit.proc_dyn))
//						th->lexufun=olexufun;

								if (!nodeList || nodeList->type!=TREE) 
								{	error(DK_HOW_UNREC, nodeList, 2);
									Goto( questionmark_failed )
								}							
								unev = TreeP(nodeList)->tree();

								Push(th->stack, th->didnt_output_name, lambda_questionmark);
								Push2(th->intStack, th->val_status, th->tailcall, lambda_questionmark);

								th->val_status=value_maybe_ok_in_macro;

								Push(th->toDoStack, __questionmark_continue, lambda_questionmark);	
								Goto( sequence )
							}
						}else		// name of procedure to apply 
						{	int min, max, n;

							th->fun = th->fun->intern();	// 1st) try finding th->fun in internal routines ;-)
//							if (ObjectP(th->fun)->proc_or_prim()==Unbound  &&  th->fun!=_Null_Word)
//								th->fun = th->fun->internRoot();

							proc=Unbound;
							if (ObjectP(th->fun)->proc_or_prim()==Unbound  &&  th->fun!=_Null_Word)
							{	th->fun->silent_load(environment.logolib.c_str());	// 2nd) try <logolib>/<th->fun> 
								ObjectP procobj=th->namespce->find(ObjectP(th->fun)->name());
								proc=Unbound;
								if (procobj && procobj->type==OBJECT && procobj->bit.proc)
								{	proc=procobj->proc_or_prim();
									if (proc != Unbound)
										th->fun=ProcedureP(proc)->object();
								}
							}
							if (ObjectP(th->fun)->proc_or_prim()==Unbound  &&  th->fun!=_Null_Word)
							{	th->fun->silent_load(0);	// 3rd) try ./<th->fun>.lg (the user directory)
								ObjectP procobj=th->namespce->find(ObjectP(th->fun)->name());
								proc=Unbound;
								if (procobj && procobj->type==OBJECT && procobj->bit.proc)
								{	proc=procobj->proc_or_prim();
									if (proc != Unbound)
										th->fun=ProcedureP(proc)->object();
								}
							}
							if (proc==Unbound)
								proc=ObjectP(th->fun)->proc_or_prim();

							if (proc==Unbound && !throwing)
								val=error(DK_HOW_UNREC, th->fun);
							
							if (!throwing) 
							{	if (type_of(proc) == PROC_) 
								{	min= ProcedureP(proc)->min;
									max= ProcedureP(proc)->max;
								}else 
								{	Primitive &prim= *PrimitiveP(proc);

									if (prim.std < 0)		// special form
									{	error(DK_HOW_UNREC, th->fun);	// can't apply 
										Goto( fetch_command )
									}else 
									{	min = prim.min;
										max= prim.max;
									}
								}
								n = int(th->argStack->index-th->argStack->base)-asp;

								if (n < min) 
									error(NOT_ENOUGH, 0);
								else if (n > max && max >= 0) 
									error(TOO_MUCH, 0);
								else 
									Goto( apply_dispatch )
							}
						}
					}
					Goto( fetch_command )
TO( after_lambda )
					th->varStack->pop_till(th->var);
					Pop1(th->varPStack, th->var, begin_apply);
					Goto( fetch_command )


TO( questionmark_continue )
					Pop2(th->intStack, th->val_status, th->tailcall, lambda_questionmark);
					Pop1(th->stack, th->didnt_output_name, lambda_questionmark);
					
					if (th->tailcall) 
					{	switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val != Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
								default: ;
								}
						default: ;
						}
					}
					Goto( questionmark_failed )
TO( questionmark_failed )
					Pop2(th->argPStack, th->qmList, asp, begin_apply);
					if (asp>=0)
						th->argStack->pop_till(asp);
//					if (th->qmList>=0)
//						th->argStack->pop_till(th->qmList);// new in 4.65
//					asp=th->qmList;
					Goto( fetch_command )

TO( macro_return )	//___________________ MACROS ______________________
					Pop2(th->intStack, th->val_status, th->tailcall, apply_dispatch);
	{
//		int macro_start;
//		Pop1(th->intStack, macro_start, dispatch);
		// Rest TODO
	}
					while (!val->islist() && !throwing)
						val=error(ERR_MACRO, val);
					
					if (!throwing) 
					{	if (type_of(val)==TODO)
						{	ToDoNode &tdn = *ToDoNodeP(val);
							Push(th->toDoStack, tdn.toDo, macro_return);
							val = tdn.tail();
							Goto( fetch_command )
						}
						if (ObjectP(th->fun)->bit.permanent)
						{
							if (th->tailcall==in_sequence) 
							{	PairP(val)->makeTree();
								th->stoppingFlag=MACRO_RETURN;
								if (val->istree()) 
									val= TreeP(LineP(val))->tree();
								else
									val=0;
								Goto( fetch_command )
							}
						}else
						{	if (th->tailcall==in_sequence) 
							{//	ObjectP olexufun=th->lexufun;
							//	th->lexufun=th->last_ufun;
								PairP(val)->makeTree();
							//	th->lexufun=olexufun;
								th->stoppingFlag=MACRO_RETURN;
								if (val->istree()) 
									val= TreeP(LineP(val))->tree();
								else
									val=0;
								Goto( fetch_command )
							}
						}
						nodeList= PairP(val);
						Goto( begin_sequence )
					}
					val=Unbound;
					Goto( fetch_command )


TO( runresult_continuation )
					nodeList=PairP(val);
					th->val_status=value_maybe_ok_in_macro;
					Push(th->toDoStack, __runresult_followup, runresult_continuation);
					Goto( begin_sequence )
TO( runresult_followup )
					if (val==Unbound) val=0;
					else 
						val = new Pair(val, 0);
					Goto( fetch_command )


TO( if_continuation )
					nodeList = PairP(val)->first();
					val=Unbound;
	*code	<< MOVrm(CX,&th->argStack->index)
			<< MOVrc(AX,Unbound)
			<< SUBrc(CX,4)
			<< MOVri(DX,CX)
			<< arg_POP_TILL_asp
			<< CMPrc(DX,_True);
	int if_start=code->length;
					Push(th->intStack, if_start, if_continuation);
	*code <<	JMPc(NZ,0);
					Push2(th->intStack, th->val_status, th->tailcall, if_continuation);
					th->val_status=output_ok;
					Push(th->stack, nodeList, if_continuation);

					Push(th->toDoStack, __if_followup, if_continuation);
					Goto( begin_sequence )
TO( if_followup )
					Pop1(th->stack, nodeList, if_continuation);
					Pop2(th->intStack, th->val_status, th->tailcall, if_continuation);

					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
								default: ;
								}
						default:	;
						}
					}
					th->checkBreak();
	*code	<<	GET_th_stoppingFlag(DX)
			<< CMPrc(DX,THROWING);
	int if_throwing=code->length;
	*code	<< JMPc(NZ, 0)
			<< RET_Unbound;
	*(int*)(charP(code->code)+if_throwing+2)=code->length-if_throwing-6;	
					val=Unbound;
	
	int if_start;
	Pop1(th->intStack, if_start, if_continuation);
	*(int*)(charP(code->code)+if_start+2)=code->length-if_start-6;	
	*code	<< arg_POP_TILL_asp;
	
					th->argStack->pop_till(asp);
					Goto( fetch_command )


TO( ifElse_continuation )
					nodeList = PairP(val);
					val=Unbound;
	*code	<< MOVrm(CX,&th->argStack->index)
			<< MOVrc(AX,Unbound)
			<< SUBrc(CX,8)
			<< MOVri(DX,CX)
			<< arg_POP_TILL_asp
			<< CMPrc(DX,_True);
	int if_else=code->length;
	*code <<	JMPc(NZ,0);
					Push2(th->intStack, th->val_status, th->tailcall, ifElse_continuation);
					th->val_status=output_ok;
					Push(th->stack, nodeList, ifElse_continuation);
					nodeList=nodeList->first();
					Push(th->intStack, if_else, ifElse_continuation);
					Push(th->toDoStack, __ifElse_else, ifElse_continuation);
					Goto( begin_sequence )
TO( ifElse_else )
	int if_then_end=code->length;
	*code <<	JMP(0);
	int if_else;
	Pop1(th->intStack, if_else, ifElse_continuation);
	*(int*)(charP(code->code)+if_else+2)=code->length-if_else-6;	
	Push(th->intStack, if_then_end, ifElse_else);
					Pop1(th->stack, nodeList, ifElse_continuation);
					nodeList=nodeList->tail();
					Push(th->toDoStack, __ifElse_followup, ifElse_else);
					Goto( begin_sequence )
TO( ifElse_followup )
	int if_then_end;
	Pop1(th->intStack, if_then_end, ifElse_else);
					Pop2(th->intStack, th->val_status, th->tailcall, ifElse_continuation);
					val=Unbound;
	*(int*)(charP(code->code)+if_then_end+1)=code->length-if_then_end-5;	
	*code	<< arg_POP_TILL_asp;
	
					th->argStack->pop_till(asp);
					Goto( fetch_command )


TO( forever_continuation )
					Push2(th->intStack, th->repcount, th->user_repcount, forever_continuation);
					th->repcount = -1;
					nodeList = PairP(val);
					val=Unbound;
					th->user_repcount=0;
	*code	<< GET_repcount(CX)
			<<	PUSH(CX)
			<< GET_user_repcount(CX)
			<< PUSH(CX)
			<< arg_POP_TILL_asp
			<< MOVrc(DX,0)
			<< SET_user_repcount(DX);
	Push(th->intStack, code->length, forever_continuation);
	*code	<< GET_user_repcount(DX)
			<< INC(DX)
			<< SET_user_repcount(DX)
			<< MOVmr(&th->user_repcount,DX);
					Goto( forever_again )
TO( forever_again )
					th->user_repcount++;
					Push2(th->intStack, th->val_status, th->tailcall, forever_again);
					th->val_status=no_value_in_macro;
					Push(th->stack, nodeList, forever_again);

					Push(th->toDoStack, __forever_followup, forever_again);
					Goto( begin_sequence )
TO( forever_followup )
					Pop1(th->stack, nodeList, forever_again);
					Pop2(th->intStack, th->val_status, th->tailcall, forever_again);


					if (val != Unbound && !throwing) error(DK_WHAT, val);
	*code << CMPrc(AX, Unbound);
	int if_val_Unbound=code->length;
	*code	<< JMPc(Z, 0)
			<< PUSHc(1)
			<< PUSH(AX)
			<<	PUSHc(DK_WHAT);
	*code	<< CALL(error)
			<< POP(CX)
			<< POP(CX)
			<< POP(CX)
			<< RET_Unbound;
	*(int*)(charP(code->code)+if_val_Unbound+2)=code->length-if_val_Unbound-6;	

	/*				if (th->stoppingFlag)
					{	switch (th->stoppingFlag)
						{
						case BREAK:	th->stoppingFlag=RUN;
										th->output_node=PairP(Unbound); 
										switch (th->val_status)
										{
										case value_required:
												error(DK_WHAT_UP, val);
										default: ;
										}												
										goto forever_end;
						case CONTINUE:	th->stoppingFlag=RUN; 
										th->output_node=PairP(Unbound); 
										break;
						default: ;
						}
					}
	*/
	*code	<<	GET_th_stoppingFlag(DX)
			<< CMPrc(DX,BREAK);
	int if_break=code->length;
	int forever_start;
	Pop1(th->intStack, forever_start, forever_continuation);
	Push(th->intStack, forever_start, forever_continuation);
	*code	<< JMPc(Z, 0)
			<< CMPrc(DX,CONTINUE)
			<<	JMPc(Z, forever_start-code->length);
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												goto forever_end;
								default: ;
								}
						default:	;
						}
					}
					if (th->repcount >0) --th->repcount;	// negative means forever 
					if (th->repcount==0) goto forever_end;

					th->checkBreak();
	*code	<<	CALL(checkBreak)
	//				if (running) Goto( forever_again )
			<<	GET_th_stoppingFlag(DX)
			<<	CMPrc(DX,RUN);
	Pop1(th->intStack, forever_start, forever_continuation);
	*code	<< JMPc(Z,forever_start-code->length);

					val=Unbound;
forever_end:	
	*(int*)(charP(code->code)+if_break+2)=code->length-if_break-6;	
	*code	<< POP(CX)
			<< SET_user_repcount(CX)
			<< POP(CX)
			<< SET_repcount(CX)
			<< arg_POP_TILL_asp;
	//		<< MOVrc(AX,Unbound);
					Pop2(th->intStack, th->repcount, th->user_repcount, forever_continuation);
					th->argStack->pop_till(asp);
					Goto( fetch_command )


TO( repeat_continuation )
					Push2(th->intStack, th->repcount, th->user_repcount, repeat_continuation);
					th->repcount = 1;
					nodeList = PairP(val);
					val=Unbound;
					th->user_repcount=0;
	*code	<< GET_repcount(CX)
			<<	PUSH(CX)
			<< GET_user_repcount(CX)
			<< PUSH(CX)
	//		<< MOVrc(AX,th->repcount)	// case count==0 is not yet handled!
			<< MOVrm(CX,&th->argStack->index)
			<< SUBrc(CX,4)
			<< MOVri(DX,CX)
			<< ADDrc(DX,int(&(Int::aInt.int_))-int(&Int::aInt))
			<< MOVri(CX,DX)
			<< SET_repcount(CX)
			<< MOVrc(DX,0)
			<< SET_user_repcount(DX)
			<< arg_POP_TILL_asp;
	int repeat_start=code->length;
	Push(th->intStack, repeat_start, repeat_continuation);
	*code	<< GET_user_repcount(DX)
			<< INC(DX)
			<< SET_user_repcount(DX)
			<< MOVmr(&th->user_repcount,DX);
					Goto( repeat_again )
TO( repeat_again )
					th->user_repcount++;
					Push2(th->intStack, th->val_status, th->tailcall, repeat_again);
					th->val_status=no_value_in_macro;
					Push(th->stack, nodeList, repeat_again);

					Push(th->toDoStack, __repeat_followup, repeat_again);
					Goto( begin_sequence )
TO( repeat_followup )
					Pop1(th->stack, nodeList, repeat_again);
					Pop2(th->intStack, th->val_status, th->tailcall, repeat_again);


					if (val != Unbound && !throwing) error(DK_WHAT, val);
	*code << CMPrc(AX, Unbound);
	int if_val_Unbound=code->length;
	*code	<< JMPc(Z, 0)
			<< PUSHc(1)
			<< PUSH(AX)
			<<	PUSHc(DK_WHAT);
	*code	<< CALL(error)
			<< POP(CX)
			<< POP(CX)
			<< POP(CX)
			<< RET_Unbound;
	*(int*)(charP(code->code)+if_val_Unbound+2)=code->length-if_val_Unbound-6;	

	/*				if (th->stoppingFlag)
					{	switch (th->stoppingFlag)
						{
						case BREAK:	th->stoppingFlag=RUN;
										th->output_node=PairP(Unbound); 
										switch (th->val_status)
										{
										case value_required:
												error(DK_WHAT_UP, val);
										default: ;
										}												
										goto repeat_end;
						case CONTINUE:	th->stoppingFlag=RUN; 
										th->output_node=PairP(Unbound); 
										break;
						default: ;
						}
					}
	*/
	*code	<<	GET_th_stoppingFlag(DX)
			<< CMPrc(DX,BREAK);
	int if_break=code->length;
	int repeat_start;
					Pop1(th->intStack, repeat_start, repeat_continuation);
					Push(th->intStack, repeat_start, repeat_continuation);
	*code	<< JMPc(Z, 0)
			<< CMPrc(DX,CONTINUE);
	int if_continue=code->length;
	*code	<< JMPc(NZ, 0)
			<< MOVrc(DX, 0)
			<< SET_th_stoppingFlag(DX)
			<<	JMPc(Z, repeat_start-code->length);
	*(int*)(charP(code->code)+if_continue+2)=code->length-if_continue-6;	
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												goto repeat_end;
								default: ;
								}
						default:	;
						}
					}
					if (th->repcount >0) --th->repcount;	// negative means forever 
	//				if (th->repcount==0) goto repeat_end;

					th->checkBreak();
	*code	<<	CALL(checkBreak)
	//				if (running) Goto( forever_again )
			<<	GET_th_stoppingFlag(DX)
			<<	CMPrc(DX,RUN);
	int if_not_running=code->length;
	*code	<< JMPc(NZ,0);
	*code	<< GET_repcount(CX)
			<< DEC(CX)
			<< SET_repcount(CX)
			<< CMPrc(CX,0);
	Pop1(th->intStack, repeat_start, repeat_continuation);
	*code	<< JMPc(NBE,repeat_start-code->length);
	*(int*)(charP(code->code)+if_break+2)=code->length-if_break-6;	
	*(int*)(charP(code->code)+if_not_running+2)=code->length-if_not_running-6;	

					val=Unbound;
	repeat_end:	
	*code	<< POP(CX)
			<< SET_user_repcount(CX)
			<< POP(CX)
			<< SET_repcount(CX)
			<< arg_POP_TILL_asp
			<< MOVrc(AX,Unbound);
					Pop2(th->intStack, th->repcount, th->user_repcount, repeat_continuation);
					th->argStack->pop_till(asp);
					Goto( fetch_command )


TO( while_continuation )
					Push(th->intStack, th->user_repcount, while_continuation);
					th->user_repcount = 0;
					Push(th->stack, th->loopcond, while_continuation);
					th->loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
	*code	<< GET_repcount(CX)
			<<	PUSH(CX)
			<< GET_user_repcount(CX)
			<< PUSH(CX)
			<< MOVrc(CX,0)
			<< SET_repcount(CX)
			<< SET_user_repcount(CX)
			<< arg_POP_TILL_asp;
					Goto( while_cond )
TO( while_cond )
					th->user_repcount++;
	int while_cond=code->length;
	Push(th->intStack, while_cond, while_cond);
	*code	<< GET_user_repcount(DX)
			<< INC(DX)
			<< SET_user_repcount(DX)
			<< MOVmr(&th->user_repcount,DX);
					Push2(th->intStack, th->val_status, th->tailcall, while_cond);
					th->val_status=value_required;
					Push(th->stack, nodeList, while_cond);
					nodeList=th->loopcond;
					Push(th->toDoStack, __while_again, while_cond);	Goto( begin_sequence )
TO( while_again )
	*code	<< arg_POP_TILL_asp;
	*code	<< CMPrc(AX,_True)
			<< MOVrc(AX,Unbound);
	int while_start=code->length;
	Push(th->intStack, while_start, while_again);
	*code <<	JMPc(NZ,0);
					if (val==_False)
					{	Pop1(th->stack, nodeList, while_cond);
						Pop2(th->intStack, th->val_status, th->tailcall, while_cond);
						val=Unbound;
						Goto( while_end )
					}
					th->val_status=no_value_in_macro;
					Pop1(th->stack, nodeList, while_cond);	// get it again
					Push(th->stack, nodeList, while_cond);

					Push(th->toDoStack, __while_followup, while_again);
					Goto( begin_sequence )
TO( while_followup )
					Pop1(th->stack, nodeList, while_cond);
	int while_cond, while_start;
					Pop1(th->intStack, while_start, while_again);
					Pop2(th->intStack, th->val_status, th->tailcall, while_cond);
					Pop1(th->intStack, while_cond, while_cond);

					if (val != Unbound && !throwing) error(DK_WHAT, val);
	*code << CMPrc(AX, Unbound);
	int if_val_Unbound=code->length;
	*code	<< JMPc(Z, 0)
			<< PUSHc(1)
			<< PUSH(AX)
			<<	PUSHc(DK_WHAT);
	*code	<< CALL(error)
			<< POP(CX)
			<< POP(CX)
			<< POP(CX)
			<< RET_Unbound;
	*(int*)(charP(code->code)+if_val_Unbound+2)=code->length-if_val_Unbound-6;	

					switch (th->stoppingFlag)
					{
					case BREAK:	th->stoppingFlag=RUN;
									th->output_node=PairP(Unbound); 
									th->val_status=no_value_in_macro;
									Goto( while_end );
					case CONTINUE:	th->stoppingFlag=RUN; 
									th->output_node=PairP(Unbound); 
									break;
					default: ;
					}
	*code	<<	GET_th_stoppingFlag(DX)
			<< CMPrc(DX,BREAK);
	int if_break=code->length;
	*code	<< JMPc(Z, 0)
			<< CMPrc(DX,CONTINUE);
	int if_continue=code->length;
	*code	<< JMPc(NZ, 0)
			<< MOVrc(DX, 0)
			<< SET_th_stoppingFlag(DX)
			<<	JMPc(Z, while_cond-code->length);
	*(int*)(charP(code->code)+if_continue+2)=code->length-if_continue-6;	
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( while_end );
								default: ;
								}
						default: ;
						}
					}
					th->checkBreak();
	//				if (running) Goto( while_cond )
	*code	<<	GET_th_stoppingFlag(DX)
			<< CMPrc(DX,THROWING);
	int if_throwing=code->length;
	*code	<< JMPc(NZ, 0)
			<< RET_Unbound;
	*(int*)(charP(code->code)+if_throwing+2)=code->length-if_throwing-6;	
	*code <<	JMP(while_cond-code->length+1);
	Push(th->intStack, while_start, while_followup);
					val=Unbound;
					Goto( while_end )
TO( while_end )
	int while_start;
	Pop1(th->intStack, while_start, while_followup);
	*(int*)(charP(code->code)+while_start+2)=code->length-while_start-6;	
	*code	<< arg_POP_TILL_asp;
					Pop1(th->stack, th->loopcond, while_continuation);
					Pop1(th->intStack, th->user_repcount, while_continuation);
					Goto( fetch_command )

TO( until_continuation )
					Push(th->intStack, th->user_repcount, until_continuation);
					th->user_repcount = 0;
					Push(th->stack, th->loopcond, until_continuation);
					th->loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
					Goto( until_cond )
TO( until_cond )
					th->user_repcount++;
					Push2(th->intStack, th->val_status, th->tailcall, until_cond);
					th->val_status=value_required;
					Push(th->stack, nodeList, until_cond);
					nodeList=th->loopcond;
					Push(th->toDoStack, __until_again, until_cond);	Goto( begin_sequence )
TO( until_again )
					if (val==_True)
					{	Pop1(th->stack, nodeList, until_cond);
						Pop2(th->intStack, th->val_status, th->tailcall, until_cond);
						val=Unbound;
						Goto( until_end )
					}
					th->val_status=no_value_in_macro;
					Pop1(th->stack, nodeList, until_cond);	// get it again
					Push(th->stack, nodeList, until_cond);

					Push(th->toDoStack, __until_followup, until_again);	Goto( begin_sequence )
TO( until_followup )
					Pop1(th->stack, nodeList, until_cond);
					Pop2(th->intStack, th->val_status, th->tailcall, until_cond);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (th->stoppingFlag)
					{

					case BREAK:	th->stoppingFlag=RUN;
									th->output_node=PairP(Unbound); 
									th->val_status=no_value_in_macro;
									Goto( until_end );
					case CONTINUE:	th->stoppingFlag=RUN; 
									th->output_node=PairP(Unbound); 
									break;
					default: ;
					}
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( until_end );
								default: ;
								}
						default: ;
						}
					}
					th->checkBreak();
					if (running) Goto( until_cond )

					val=Unbound;
					Goto( until_end )
TO( until_end )
					Pop1(th->stack, th->loopcond, until_continuation);
					Pop1(th->intStack, th->user_repcount, until_continuation);

					Goto( fetch_command )

TO( do_while_continuation )
					Push(th->intStack, th->user_repcount, do_while_continuation);
					th->user_repcount = 0;
					Push(th->stack, th->loopcond, do_while_continuation);
					th->loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
					Goto( do_while_again )
TO( do_while_again )
					th->user_repcount++;
					Push2(th->intStack, th->val_status, th->tailcall, do_while_again);
					th->val_status=no_value_in_macro;
					Push(th->stack, nodeList, do_while_again);


					Push(th->toDoStack, __do_while_followup, do_while_again);
					Goto( begin_sequence )
TO( do_while_followup )
					Pop1(th->stack, nodeList, do_while_again);
					Pop2(th->intStack, th->val_status, th->tailcall, do_while_again);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (th->stoppingFlag)
					{
					case BREAK:	th->stoppingFlag=RUN;
									th->output_node=PairP(Unbound); 
									th->val_status=no_value_in_macro;
									Goto( do_while_end );
					case CONTINUE:	th->stoppingFlag=RUN; 
									th->output_node=PairP(Unbound); 
									break;
					default: ;
					}
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)


													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( do_while_end );
								default: ;
								}
						default: ;
						}
					}
					th->val_status=value_required;
					Push(th->stack, nodeList, do_while_followup);
					nodeList=th->loopcond;
					Push(th->toDoStack, __do_while_cond, do_while_followup);
					Goto( begin_sequence )
TO( do_while_cond )
					Pop1(th->stack, nodeList, do_while_followup);
					if (val==_False)
					{	val=Unbound;
						Goto( do_while_end )
					}

					th->checkBreak();
					if (running) Goto( do_while_again )

					val=Unbound;
					Goto( do_while_end )
TO( do_while_end )
					Pop1(th->stack, th->loopcond, do_while_continuation);
					Pop1(th->intStack, th->user_repcount, do_while_continuation);
					Goto( fetch_command )

TO( do_until_continuation )
					Push(th->intStack, th->user_repcount, do_until_continuation);
					th->user_repcount = 0;
					Push(th->stack, th->loopcond, do_until_continuation);
					th->loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
					Goto( do_until_again )
TO( do_until_again )
					th->user_repcount++;
					Push2(th->intStack, th->val_status, th->tailcall, do_until_again);
					th->val_status=no_value_in_macro;
					Push(th->stack, nodeList, do_until_again);

					Push(th->toDoStack, __do_until_followup, do_until_again);	Goto( begin_sequence )
TO( do_until_followup )
					Pop1(th->stack, nodeList, do_until_again);
					Pop2(th->intStack, th->val_status, th->tailcall, do_until_again);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (th->stoppingFlag)
					{
					case BREAK:	th->stoppingFlag=RUN;
									th->output_node=PairP(Unbound); 
									th->val_status=no_value_in_macro;
									Goto( do_until_end );
					case CONTINUE:	th->stoppingFlag=RUN; 
									th->output_node=PairP(Unbound); 
									break;
					default: ;
					}
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( do_until_end );
								default: ;

								}
						default: ;
						}
					}
					th->val_status=value_required;
					Push(th->stack, nodeList, do_until_followup);
					nodeList=th->loopcond;
					Push(th->toDoStack, __do_until_cond, do_until_followup);
					Goto( begin_sequence )
TO( do_until_cond )
					Pop1(th->stack, nodeList, do_until_followup);
					if (val==_True)
					{	val=Unbound;
						Goto( do_until_end )
					}
					th->checkBreak();
					if (running) Goto( do_until_again )

					val=Unbound;
					Goto( do_until_end )
TO( do_until_end )
					Pop1(th->stack, th->loopcond, do_until_continuation);
					Pop1(th->intStack, th->user_repcount, do_until_continuation);
					Goto( fetch_command )

TO( for_continuation )
					Push3(th->floatStack, th->fori, th->forend, th->forstep, for_continuation);
					Push(th->stack, th->forname, for_continuation);
					PairP pval=PairP(val);
					th->forname= pval->first()->intern();
					th->fori   = FloatP(pval->second())->float_;
					th->forend = FloatP(pval->third ())->float_;
					th->forstep= FloatP(pval->fourth())->float_;
					nodeList =      pval->fifth();
					PushVar(th->varStack, th->forname, new Float(th->fori), for_continuation);
					th->forname->tell_shadow();
								
					val=Unbound;
					Goto( for_again )
TO( for_again )
					Push2(th->intStack, th->val_status, th->tailcall, for_again);
					th->val_status=no_value_in_macro;
					Push(th->stack, nodeList, for_again);
					Push(th->varPStack, *th->varStack, for_again);

					Push(th->toDoStack, __for_followup, for_again);	Goto( begin_sequence )
TO( for_followup )
					Pop1(th->varPStack, vsp, for_again);
					th->varStack->pop_till(vsp);
					Pop1(th->stack, nodeList, for_again);
					Pop2(th->intStack, th->val_status, th->tailcall, for_again);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (th->stoppingFlag)
					{
					case BREAK:	th->stoppingFlag=RUN;
									th->output_node=PairP(Unbound); 
									switch (th->val_status)
									{
									case value_required:
											error(DK_WHAT_UP, val);
									default: ;
									}												
									goto for_end;
					case CONTINUE:	th->stoppingFlag=RUN; 
									th->output_node=PairP(Unbound); 
									break;
					default:	;
					}
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												goto for_end;
								default: ;
								}
						default: ;
						}
					}
					{	NodeP forval=th->varStack->find(th->forname);
						if (forval->type==FLOAT_)
							th->fori = FloatP(forval)->float_;
						else if (forval->type==INT_)
							th->fori = IntP(forval)->int_;
						else
							error(BAD_DATA, forval);
					}
					th->fori+=th->forstep;
					if (th->forstep >= 0)
					{	if (th->fori > th->forend) 
							goto for_end;
					}else
						if (th->fori < th->forend) 
							goto for_end;
					th->varStack->setVar(th->forname, new Float(th->fori) );

					th->checkBreak();
					if (running) Goto( for_again )

					val=Unbound;
	for_end:
					Pop(th->varStack, for_continuation);
					Pop1(th->stack, th->forname, for_continuation);
					Pop3(th->floatStack, th->fori, th->forend, th->forstep, for_continuation);
					Goto( fetch_command )

TO( ifor_continuation )
					Push3(th->intStack, th->ifori, th->iforend, th->iforstep, ifor_continuation);
					Push(th->stack, th->forname, ifor_continuation);
					PairP pval=PairP(val);
					th->forname= pval->first()->intern();
					th->ifori   = IntP(pval->second())->int_;
					th->iforend = IntP(pval->third ())->int_;
					th->iforstep= IntP(pval->fourth())->int_;
					nodeList =      pval->fifth();
					PushVar(th->varStack, th->forname, new Int(th->ifori), ifor_continuation);
					th->forname->tell_shadow();
								
					val=Unbound;
					Goto( ifor_again )
TO( ifor_again )
					Push2(th->intStack, th->val_status, th->tailcall, ifor_again);
					th->val_status=no_value_in_macro;
					Push(th->stack, nodeList, ifor_again);
					Push(th->varPStack, *th->varStack, ifor_again);

					Push(th->toDoStack, __ifor_followup, ifor_again);
					Goto( begin_sequence )
TO( ifor_followup )
					Pop1(th->varPStack, vsp, ifor_again);
					th->varStack->pop_till(vsp);
					Pop1(th->stack, nodeList, ifor_again);
					Pop2(th->intStack, th->val_status, th->tailcall, ifor_again);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (th->stoppingFlag)
					{
					case BREAK:	th->stoppingFlag=RUN;
									th->output_node=PairP(Unbound); 
									switch (th->val_status)
									{
									case value_required:
											error(DK_WHAT_UP, val);
									default: ;
									}												
									goto ifor_end;
					case CONTINUE:	th->stoppingFlag=RUN; 
									th->output_node=PairP(Unbound); 
									break;
					default:	;
					}
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;

													}
												goto ifor_end;
								default: ;
								}
						default: ;
						}
					}
					{	NodeP forval=th->varStack->find(th->forname);
						if (forval->type==INT_)
							th->ifori = IntP(forval)->int_;
						else if (forval->type==FLOAT_)
							th->ifori = int(FloatP(forval)->float_);
						else
							error(BAD_DATA, forval);
					}
					th->ifori+=th->iforstep;
					if (th->iforstep >= 0)
					{	if (th->ifori > th->iforend) 
							goto ifor_end;
					}else
						if (th->ifori < th->iforend) 
							goto ifor_end;
					th->varStack->setVar(th->forname, new Int(th->ifori) );

					th->checkBreak();
					if (running) Goto( ifor_again )

					val=Unbound;
	ifor_end:	Pop(th->varStack, ifor_continuation);
					Pop1(th->stack, th->forname, ifor_continuation);
					Pop3(th->intStack, th->ifori, th->iforend, th->iforstep, ifor_continuation);
					Goto( fetch_command )

TO( foreach_continuation )
					Push(th->intStack, th->user_repcount, foreach_continuation);
					Push(th->stack, th->loopcond, foreach_continuation);
					Push(th->varPStack, th->var, foreach_continuation);
					th->var=*th->varStack;
					th->loopcond = PairP(val)->first();
					nodeList = PairP(val)->tail ();
					th->user_repcount=0;
					Goto( foreach_again )
TO( foreach_again )
					if (!th->loopcond->first() || th->loopcond->first()==NodeP(_Null_Word)) 
						Goto( foreach_end )

					th->user_repcount++;
					PushVar(th->varStack, ___Template_Lists, th->loopcond, foreach_again);
					val=new Pair(nodeList, th->loopcond->firsts());
					th->loopcond = PairP(th->loopcond->butFirsts());

					Push2(th->intStack, th->val_status, th->tailcall, foreach_again);
					Push(th->stack, nodeList, foreach_again);
					th->val_status=no_value_in_macro;

					th->fun=PairP(val)->first();
					
					while (type_of(th->fun)==ARRAY  &&  !throwing)
						th->fun=ObjectP(error(APPLY_BAD_DATA, th->fun));

					while (PairP(val)->tail()->isnolist() && !throwing)
						PairP(val)->settail( 
							error(APPLY_BAD_DATA, PairP(val)->tail()) );

					if (!throwing)
					{	asp=*th->argStack;
	SET_ASP_from_argStack;
						PushList(th->argStack, PairP(val)->tail(), foreach_again);
						val=Unbound;
					}
					Push(th->argPStack, th->qmList, foreach_again);
					th->qmList=asp;
					
					Push(th->toDoStack, __foreach_followup, foreach_again);
					Goto( begin_sequence )
TO( foreach_followup )
					if (th->qmList>=0)
						th->argStack->pop_till(th->qmList);// new in 4.65
					Pop1(th->argPStack, th->qmList, foreach_again);
					Pop1(th->stack, nodeList, foreach_again);

					Pop2(th->intStack, th->val_status, th->tailcall, foreach_again);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (th->stoppingFlag)
					{
					case BREAK:	th->stoppingFlag=RUN;
									th->output_node=PairP(Unbound); 
									switch (th->val_status)
									{
									case value_required:
											error(DK_WHAT_UP, val);
									default: ;
									}												
									Goto( foreach_end );
					case CONTINUE:	th->stoppingFlag=RUN; 
									th->output_node=PairP(Unbound); 
									break;
					default:	;

					}
					if (th->tailcall) 
					{
						switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( foreach_end );
								default: ;
								}
						default: ;
						}
					}
					th->checkBreak();
					if (running) Goto( foreach_again )

					val=Unbound;
					Goto( foreach_end )
TO( foreach_end )
					th->varStack->pop_till(th->var);
					Pop1(th->varPStack, th->var, foreach_continuation);
					Pop1(th->stack, th->loopcond, foreach_continuation);
					Pop1(th->intStack, th->user_repcount, foreach_continuation);
					Goto( fetch_command )

TO( catch_continuation )
					nodeList = PairP(val)->tail ();
					catch_tag= PairP(val)->first();

					th->val_status=value_maybe_ok_in_macro;

					Push3(th->stack, catch_tag, th->didnt_output_name, th->erract, catch_continuation);
					Push2(th->intStack, th->val_status, th->tailcall, catch_continuation);
					Push(th->argPStack, asp, catch_continuation);

					if (catch_tag->equal(_error)) 
					{	th->erract=Unbound;
					}
					Push(th->toDoStack, __catch_followup, catch_continuation);
					Goto( begin_sequence )
TO( catch_followup )
					Pop2(th->intStack, th->val_status, th->tailcall, catch_continuation);
					Pop3(th->stack, catch_tag, th->didnt_output_name, th->erract, catch_continuation);
					Pop1(th->argPStack, asp, catch_continuation);
					th->argStack->pop_till(asp);

					if (th->tailcall) 
					{	switch (th->val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (th->stoppingFlag)
								{
								case RUN:	th->output_node=PairP(Unbound); break;
								case STOP:	th->output_node=PairP(Unbound);
								case OUTPUT:th->stoppingFlag=RUN;
												val=th->output_node;
												if (val!=Unbound)
													switch (th->val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;

													}
								default: ;
								}
						default: ;
						}
					}
					if ( throwing 
					&& th->throw_node->equal(catch_tag)) 
					{	th->throw_node=Unbound;
						th->stoppingFlag=RUN;
						val=th->output_node;
					}
					Goto( fetch_command )

TO( goto_continuation )
					if (!throwing) 
					{	if (!th->ufun) 
						{	error(AT_TOPLEVEL, _Goto);
							val=Unbound;
							Goto( fetch_command )
						}
						    proc = ObjectP(th->ufun)->proc_or_prim();

						nodeList = ProcedureP(proc)->bodylist();
						    unev = TreeP(nodeList)->tree();

						while (unev  &&  !checkThrowing()) 
						{
							if (unev->type==TREE)
								th->this_line = TreeP(unev)->tree();
							exp=unev->first();
							next(unev);

							if (exp->islist()
							&& ObjectP(exp->first())==_Tag
							&& exp->tail()
							&& exp->second()->type==QUOTE 
							&& val->equal(QuoteP(exp->second())->quote()) ) 
							{
								val = new Pair(_Tag, unev);
								th->stoppingFlag=MACRO_RETURN;
								Goto( fetch_command )
							}
						}
						if (!throwing)
							error(BAD_DATA_UNREC, val);
					}
					val=Unbound;
					Goto( fetch_command )

TO( polygon_continuation )
					nodeList=PairP(val);
					th->val_status=no_value_allowed;
					Push(th->toDoStack, __polygon_followup, polygon_continuation);
					Goto( begin_sequence )
TO( polygon_followup )
					lPolyEnd(0);
					Goto( fetch_command )

TO( tess_continuation )
					nodeList=PairP(val);
					th->val_status=no_value_allowed;
					Push(th->toDoStack, __tess_followup, tess_continuation);
					Goto( begin_sequence )
TO( tess_followup )
					lTessEnd(0);
					Goto( fetch_command )

TO( surface_continuation )
					nodeList=PairP(val);
					th->val_status=no_value_allowed;
					Push(th->toDoStack, __surface_followup, surface_continuation);
					Goto( begin_sequence )
TO( surface_followup )
					lSurfaceEnd(0);
					Goto( fetch_command )

TO( graphic_continuation )
					nodeList=PairP(val);
					th->val_status=no_value_allowed;
					Push(th->toDoStack, __graphic_followup, graphic_continuation);
					Goto( begin_sequence )
TO( graphic_followup )
					val=lGraphicEnd(0);
					Goto( fetch_command )

TO( end )
					th->tailcall=oldtailcall;
					th->ift_iff_flag=old_ift_iff;

					Pop3(th->stack, th->fun, th->ufun, th->this_line, eval);
					th->varStack->pop_till(th->var);
					Pop1(th->varPStack, th->var, eval);
#ifdef PROC_EVAL
					return 0;
				}
#else
				}
				return code;
}
#endif

NodeP Compiler::compileMake(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);

	if (!throwing) 
	{	ObjectP what=0;
		if (name->type==OBJECT)	// slow if many locals exist!
		{	what=ObjectP(name);
		}else
		{	what = name->intern();
		}
//		th->varStack->setVar(what, arg[2]);
	int sp= th->varStack->whereIsVar(what);
	if (sp < 0)
	{	c_out	<< "			;MAKE global " << what << "=" << arg[2] << endL;
		*code	<< MOVrm(CX,&th->argStack->index)
				<< MOVri(DX,CX)
				<< MOVrc(CX,int(what)+int(&(Object::aObject._var))-int(&Object::aObject))
				<< MOVir(CX,DX);
	}else
	{	c_out	<< "			;MAKE on argStack " << what << "=" << arg[2] << endL;
		*code	<<	MOVrm(CX,&th->argStack->index)
				<< MOVri(DX,CX)
				<< SUBrc(CX,sp*4+8)
				<< MOVir(CX,DX);
	}
	*code << MOVrc(AX,Unbound);
	*code	<<	arg_POP_TILL_asp;
		if (what->bit.val_traced) 
		{
			NodeP tvar=0;	tvar = arg[2]->maybe_quote();
			writer << _("Make ") << QuoteP(new Quote(what)) << ' ' << tvar;

			ThreadP th=Thread::This();
			if (th->ufun) 
				writer << _(" in ") << th->ufun << endL << th->this_line;
			writer << endL;
		}
	}
	return Unbound;
}

NodeP Compiler::compileLocal(PairPP arg)
{
	if (th->tailcall==tailcall_tail) 
		return Unbound;

	PairP li = arg[1];
	while (li->islist()  &&  exist(2)  &&  !throwing)
		li = error(BAD_DATA, li);

	int vsp = *th->varStack;

	*code << arg_POP_TILL_asp;
	int n=0;
	if (li->islist())		// local [a b]
		while (li  &&  !throwing) 
		{
			while (! li->first()->is.word  &&  !throwing)
				li->setfirst( error(BAD_DATA, li->first()) ); // prevent crash in _apply
			
			if (!throwing) 
			{	ObjectP a = li->first()->intern();
				li->setfirst(a);	// local [a b] faster next time
				if (a->not_local(vsp)) 
				{	PushVar(th->varStack, a, Unbound, local);
	*code <<	arg_PUSH(Unbound);
					n++;
				}else
					th->varStack->setVar(a, Unbound);
				a->tell_shadow();
			}
			if (checkThrowing())
				break;
			next(li);
		}
	else		// (local "a "b)
	{	int nr=1;
		while (arg!=PairPP(th->argStack->index)  &&  !throwing) 
		{
			if (arg!=PairPP(th->argStack->index)  
			&&  *++arg  &&  !(*arg)->is.word)
				error(BAD_DATA, *arg, nr);

			ObjectP a = (*arg)->intern();
			if (!throwing) 
			{
				if (a->not_local(vsp)) 
				{	PushVar(th->varStack, a, Unbound, local);
	*code <<	arg_PUSH(Unbound);
					n++;
				}else
					th->varStack->setVar(a, Unbound);
				a->tell_shadow();
			}
			if (checkThrowing()) break;
			nr++;
		}
	}
	*code	<< GET_ASP(CX)
			<<	ADDrc(CX, n*4)
			<<	SET_ASP(CX);
	th->var = *th->varStack;    // so eval won't undo our work
	return Unbound;
}

