
#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

#include <limits>

#include "Main.h"
#include "Thread.h"
#include "File.h"
#include "String.h"
#include "Float.h"
#include "Complex.h"
#include "Error.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

const int Complexd::digits = std::numeric_limits<double>::digits10+10;

DataInputStream &Complexd::read(DataInputStream &istr)
{	double r,i;
	istr >> r >> i;
	complex_=complexd(r,i);
	return istr;
}
wxDataOutputStream &Complexd::write(wxDataOutputStream &ostr)
{	return ostr << type << complex_.real() << complex_.imag();
}

NodeP Complexd::copy(PtrToInt_ &ptrToInt)
{	return new Complexd(complex_);
}

int Complexd::_sizeof() {return sizeof(complexd);}

StringP Complexd::string()
{
	char format[20], s[digits*2+2];
	int prec=Thread::This()->precision;
	if (complex_.imag() >= 0)
	{	sprintf(format, "%%-.%ig+%%-.%igi", prec, prec);
		sprintf(s, format, complex_.real(),complex_.imag());
	}else
	{	sprintf(format, "%%-.%ig%%-.%igi", prec, prec);
		sprintf(s, format, complex_.real(),complex_.imag());
	}
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void Complexd::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, complex_);
}

void Complexd::printBin(FILEP strm) 
{
	::printBin(strm, complex_);
}

double Complexd::getnumber() 
{
	return ::sqrt(Float::sqr(complex_.real())+Float::sqr(complex_.imag()));
}

double Complexd::_min(double d) 
{	double f=Norm()->getnumber();
	return f < d ?  f : d;
}

double Complexd::_max(double d) 
{	double f=Norm()->getnumber();
	return f > d ?  f : d;
}

NodeP Complexd::Norm()
{	return new Float(::sqrt(Float::sqr(complex_.real())+Float::sqr(complex_.imag())));
}

NodeP Complexd::maxNorm()
{	return Norm();
}

NodeP Complexd::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}

NodeP Complexd::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(complexd))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	int *bits=(intP)&complex_;
	return boolnode((bits[i>>5]>>(i & 31)) & 1);
}

NodeP Complexd::memberP(NodeP subobj, bool notp, bool substr)
{
	return string()->memberP(subobj, notp, substr);
}
NodeP Complexd::readbin(FileP file)
{
	ThreadP th=Thread::This();
	complexd num;
	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read complex binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(complexd), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(complexd);
			if (th->rstringlen >= 0)
			{	num = *complexdP(th->rstringptr);	// reading from string
				th->rstringptr+=sizeof(complexd);
			}else
				error(FILE_ERROR, new String
					(_("memory block too small to read complex")));
		}
	complex_=num;
	return this;
}

PairP Complexd::list() 
{
	return string()->list();
}

double Complexd::abs(complexd x) 
{	return ::sqrt(Float::sqr(x.real())+Float::sqr(x.imag()));
}

complexd Complexd::rnd()
{
	return complexd(Float::rnd(),Float::rnd());
}

complexd Complexd::sqr(complexd x) {return x*x;}

complexd Complexd::Remainder(complexd x, double d) 
{	if (d==0.0) 
	{	error(BAD_DATA, new Complexd(0));
		return Float::NaN;
	}
	return complexd(fmod(x.real(),d), fmod(x.imag(),d));
}
complexd Complexd::Remainder(double x, complexd d) 
{	return Float::NaN;
}
complexd Complexd::Remainder(complexd x, complexd d) 
{	return Float::NaN;
}

complexd Complexd::mod(complexd x, double d) 
{
	if (d==0.0) 
	{	error(BAD_DATA, new Float(0));
		return Float::NaN;
	}
	double r=fmod(x.real(),d), i=fmod(x.imag(),d);

	if ((r < 0) != (d < 0)) 
		if ((i < 0) != (d < 0)) 
			return complexd(r+d,i+d);
		else
			return complexd(r+d,i);

	if ((i < 0) != (d < 0)) 
		return complexd(r,i+d);
	return complexd(r,i);
}
complexd Complexd::mod(double x, complexd d) 
{	return Float::NaN;
}
complexd Complexd::mod(complexd x, complexd d) 
{	return Float::NaN;
}

NodeP lreadComplexBin(PairPP) 
{
	NodeP n=new Complexd();
	return n->readbin(Thread::This()->Reader);
}

