#include "stdafx.h"

#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

#include "Main.h"
#include "Thread.h"
#include "File.h"
#include "String.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "Float.h"
#include "Object.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "Int8Array.h"
#include "UInt8Array.h"
#include "FloatArray.h"
#include "Complex.h"
#include "ComplexArray.h"
#include "Pair.h"
#include "Error.h"
#include "Init.h"
#include "NodePStack.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

ComplexArray::ComplexArray(int length_, int origin_): Node(COMPLEXARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (complexdP) GC_::calloc(length_+1, sizeof(complexd));
	if (!_data) 
		return;
	_length=length_;
	flag=ObjectType(0);
}

ComplexArray::ComplexArray(ComplexArrayP array_): Node(COMPLEXARRAY)
{
	_origin = array_->_origin;
	_length=0;
	_data = (complexdP) GC_::calloc(array_->_length+1, sizeof(complexd));
	if (!_data) 
		return;
	_length= array_->_length;
	complexdP 
		source = array_->_data, 
		destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
	flag=ObjectType(0);
}

void ComplexArray::free() {GC_::free(_data); _data=0;}

DataInputStream &ComplexArray::read(DataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (complexdP) GC_::calloc(sizeof(complexd), length+1);
	if (!_data)
		return istr;
	_length=length;
	complexdP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	double re,im;
		istr >> re >> im;
		*itemptr=complexd(re,im);
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &ComplexArray::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	complexdP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	ostr << itemptr->real() << itemptr->imag();
		itemptr++;
	}
	return ostr;
}

NodeP ComplexArray::copy(PtrToInt_ &ptrToInt)
{
	ComplexArrayP arr=new ComplexArray(this);
	return arr;
}

StringP ComplexArray::string()
{
 	const int maxlength=65536;
	char textbuf[maxlength];
	Thread::This()->stringptr=textbuf;
	Thread::This()->stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*(Thread::This()->stringptr)='\0';
	return new String(textbuf);
}

void ComplexArray::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	int i=0, wid=0;
	complexdP pp=data();
						
	if (width >= 2*_length-10)
		wid=width/_length;

	::print(strm, width, '{');
	while (i < _length) 
	{	int w=wid;
		::print(strm, w, *pp++);	
		if (++i < length())
			::print(strm, w, ' ');
		width-=wid-w;
	}
	if (i < length()) 
		::print(strm, width, "...");
	::print(strm, width, '}');

	if (origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", origin());
		::print(strm, width, org);
	}
}

void ComplexArray::printBin(FILEP strm)
{
	ThreadP th=Thread::This();
	int wid=length();
	complexdP pp=data();
	if (!strm)
	{	if (th->stringlen > 0)     // printing to string
			th->stringlen-=sizeof(complexd)*wid;
		if (th->stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*complexdP(th->stringptr) = *pp++;
				th->stringptr+=sizeof(complexd);
			}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print ComplexArray to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(data());
	int width=Int::max;
		for (int l=sizeof(complexd)*wid; l; l--)
			::print(strm, width, *cp++);
		return;
	}
	fwrite(pp, sizeof(complexd), wid, strm);
}

NodeP ComplexArray::contentsTree(	wxArrayString &namearr, 
											wxArrayString &valarr, 
											wxArrayPtrVoid &objarr, 
											ObjectType want, 
											ObjectType dont,
											ObjectType must, 
											int &count, int level)
{
	objarr[count]=this;
	if (!expanded())
	{	StringP str=string();
		if (str  &&  str->is.string)
		{	valarr[count]=str->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
		}
		return this;
	}
	complexdP dp=_data;
	for (int i=0; i < _length;  i++)
	{	namearr[count]=wxString(' ',level-1);
		namearr[count] << i+_origin;
		objarr[count]=this;
		complexd d=(*dp++);
		if (d.imag() > 0)
			valarr[count].Printf("%10g +%10g ", d.real(), d.imag());
		else if (d.imag() < 0)
			valarr[count].Printf("%10g -%10g ", d.real(), abs(d.imag()));
		else
			valarr[count].Printf("%10g", d.real());
		incContentsArrays(namearr, valarr, objarr, count);
	}
	return this;
}

int ComplexArray::count() {return length();}

int ComplexArray::_sizeof() {return length()*sizeof(double);}

NodeP ComplexArray::compare(NodeP nd)
{	if (nd->type != COMPLEXARRAY) 
		return Unbound;
	if (this==nd)
		return 0;

	ComplexArrayP b=ComplexArrayP(nd);
	complexdP dp=_data, dpb=b->_data;
	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{	complexd c=((*dp++) - (*dpb++));
		double d=sqrt(c.real()*c.real()+c.imag()*c.imag());
		if (d)
			return (d > 0 ?  _One : _Minus_One);
	}
	if (!i & !j)
		return 0;
	if (i)
		return _One;
	return _Minus_One;
}

bool ComplexArray::equalP(NodeP b)
{
	if (b->type != COMPLEXARRAY) 
		return false;
	if (this==b)
		return true;

	ComplexArrayP ba=ComplexArrayP(b);
	complexdP dp=_data, dpb=ba->_data;
	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
	{	complexd c=(*dp++) - (*dpb++);
		if (c.real()!=0 || c.imag()!=0)
			return false;
	}
	if (!i & !j)
		return true;
	return false;
}

NodeP ComplexArray::first_() 
{	if (length() >= 1) 
		return new Complexd(*(ComplexArrayP(this)->data()));
	else
		return 0;
}

NodeP ComplexArray::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ComplexArrayP result=new ComplexArray(l, o);
	if throwing return Unbound;
	complexdP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP ComplexArray::last()
{	return new Complexd(*(data()+length()-origin()));
}

NodeP ComplexArray::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ComplexArrayP result=new ComplexArray(l, o);
	if throwing return Unbound;
	complexdP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP ComplexArray::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	ComplexArrayP result=new ComplexArray(l+1, o);
	if throwing return Unbound;
	complexdP source=data(), dest=result->data();
	*dest++=nd->getnumber();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP ComplexArray::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	ComplexArrayP result=new ComplexArray(l+1, o);
	if throwing return Unbound;
	complexdP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=nd->getnumber();
	return result;
}

double ComplexArray::_min(double d)
{
	ThreadP th=Thread::This();
	complexdP dp=_data;
	for (int i=0;  i<_length;  i++)
	{	complexd c=*dp++;
		double n=sqrt(c.real()*c.real()+c.imag()*c.imag());
		if (n < d)
		{	d=n;
			th->minimumPosition=i+_origin;
		}
	}
	return d;
}

double ComplexArray::_max(double d)
{
	ThreadP th=Thread::This();
	complexdP dp=_data;
	for (int i=0;  i<_length;  i++)
	{	complexd c=*dp++;
		double n=sqrt(c.real()*c.real()+c.imag()*c.imag());
		if (n > d)
		{	d=n;
			th->maximumPosition=i+_origin;
		}
	}
	return d;
}

NodeP ComplexArray::Norm()
{	FloatArrayP result=new FloatArray(_length);
	if throwing return Unbound;
	doubleP d=result->data();
	complexdP s=_data;
	for (int i=_length;  i>0;  i--)
	{	complexd c=(*s++);
		*d++=sqrt(c.real()*c.real()+c.imag()*c.imag());
	}
	return result;
}

NodeP ComplexArray::maxNorm()
{	FloatArrayP result=new FloatArray(_length);
	if throwing return Unbound;
	doubleP d=result->data();
	complexdP s=_data;
	for (int i=_length;  i>0;  i--)
	{	complexd c=(*s++);
		double x=Float::abs(c.real());
		double y=Float::abs(c.imag());
		if (x > y)
			*d++=x;
		else
			*d++=y;
	}
	return result;
}

NodeP ComplexArray::resize(int newsize)
{	ComplexArrayP r=new ComplexArray(newsize);
	if throwing return Unbound;
	complexdP d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (Complexd::abs(d[i]) && di-1 > 0)
				{	complexd diff = d[i+1]-d[i];
					complexd slope = diff/complexd(di);
					for (int ii=di-1;  ii;  ii--)
						r->_data[ni+ii] = d[i]+slope*complexd(ii);
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f;
				complexd diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=d[int(i)]+frac*diff;
			}
		}
	}
	return r;
}

NodeP ComplexArray::lowPassFilter(double weight)
{
	ComplexArrayP result=new ComplexArray(length());
	if throwing return Unbound;
	complexdP source=data(), dest=result->data();
	complexd median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=median;
	}
	Float::reset_FPU();
	return result;
}

NodeP ComplexArray::lowPassFilter2(double weight)
{
	ComplexArrayP result=new ComplexArray(length());
	if throwing return Unbound;
	complexdP source=data(), dest=result->data();
	complexd median=*source;
	for (int j=4; j>0; j--)
	{	for (int i=length(); i>0; i--)
		{	median=(median*weight+(*source++))/(weight+1);
			(*dest++)+=median;
		}
		for (int i=length(); i>0; i--)
		{	median=(median*weight+(*--source))/(weight+1);
			(*--dest)+=median;
		}
	}
	Float::reset_FPU();
	return result;
}

NodeP ComplexArray::saturateAbove(double limit) 
{
	ComplexArrayP result=new ComplexArray(length());
	if throwing return Unbound;
	complexdP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	double s=Complexd::abs(*source++);
		*dest++= s>limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP ComplexArray::saturateBelow(double limit) 
{
	ComplexArrayP result=new ComplexArray(length());
	if throwing return Unbound;
	complexdP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	double s=Complexd::abs(*source++);
		*dest++= s<limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP ComplexArray::item(int i)
{	if (i >= 0  &&  i < _length) 
		return new Complexd(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}

NodeP ComplexArray::items(NodeP start, NodeP end)
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	ComplexArrayP newar=new ComplexArray(e-s+1);
					if throwing return Unbound;
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP ComplexArray::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	ComplexdP ni=newitem->a_complexOrReal(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni->complex_;
	}
	return Unbound;
}

NodeP ComplexArray::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	complexdP d=_data;   int l=_length;
	ComplexdP f=new Complexd(0);
	if throwing return Unbound;

	while (--l >= 0 && !throwing) 
	{	f->complex_=(*d++);
		if (subobj->equalP(f))
			return _True;
	}
	return _False;
}

NodeP ComplexArray::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					complexdP d = _data+i;
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =   IntP(h)->int_;	break;
							case FLOAT_:*d++ = FloatP(h)->float_;break;
							case NINT16:	*d++ = Int16P(h)->int_;	break;
							case NINT8:	*d++ =  Int8P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = h->getnumber();
											else
											{	error(BAD_DATA, n, 3); 
												return ComplexArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);
								return ComplexArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					complexdP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	int16P s=n->data();
					complexdP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT8ARRAY)
			{	Int8ArrayP n=Int8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	int8P s=n->data();
					complexdP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==UINT8ARRAY)
			{	UInt8ArrayP n=UInt8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	uint8P s=n->data();
					complexdP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==COMPLEXARRAY)
			{	ComplexArrayP n=ComplexArrayP(newitems);
				if (i+n->length() <= _length)
				{	complexdP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}

NodeP ComplexArray::reverse()
{
	ComplexArrayP result=new ComplexArray(_length);
	if throwing return Unbound;

	complexdP d=_data, rev=result->_data+_length-1;
	for (int i=_length; i>0; i--)
		*rev-- = *d++;
	return result;
}

NodeP ComplexArray::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	ComplexArrayP val = new ComplexArray(length(), origin());
	if throwing return Unbound;
	complexdP
		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}

NodeP ComplexArray::shuffle()
{	ComplexArrayP result=new ComplexArray(ComplexArrayP(this));
	if throwing return Unbound;
	complexdP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		complexd tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}

PairP ComplexArray::list()
{
	complexdP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
	{	p = new Pair(new Complexd(*--d), p);
		if throwing return PairP(Unbound);
	}
	return p;
}

ArrayP ComplexArray::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	if throwing return ArrayP(Unbound);
	NodePP d = array->data();
	complexdP s=_data;

	for (int i=_length;  i;  i--)
	{	*d++ = new Complexd(*s++);
		if throwing return ArrayP(Unbound);
	}
	return array;
}

IntArrayP ComplexArray::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	if throwing return IntArrayP(Unbound);
	intP d = intarray->data();
	complexdP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int::clip(Complexd::abs(*s++));

	return intarray;
}

Int16ArrayP ComplexArray::int16array(int origin) 
{
	Int16ArrayP int16array = new Int16Array(_length, _origin);
	if throwing return Int16ArrayP(Unbound);
	int16P d = int16array->data();
	complexdP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int16::clip(Complexd::abs(*s++));

	return int16array;
}

FloatArrayP ComplexArray::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	if throwing return FloatArrayP(Unbound);
	doubleP d = floatarray->data();
	complexdP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Complexd::abs(*s++);

	return floatarray;
}

ComplexArrayP ComplexArray::complexarray(int origin) 
{
	return new ComplexArray(ComplexArrayP(this));
}

void ComplexArray::combSortInPlace()
{
	double gap=_length; //initialize gap size
	int swaps=0; //see bubblesort for an explanation
    
	while (!(gap<=1 && swaps==0))
	{	//update the gap value for a next comb
		if (gap > 1)
		{	gap/=1.3;
			if (gap==10 || gap==9)
				gap=11;
		}
		int i=0;
		//a single "comb" over the input list
		while (i+gap<_length)
		{	int j=int(i+gap);
		if (Complexd::abs(_data[i]) > Complexd::abs(_data[j]))
			{	complexd tmp=_data[i];
				_data[i]=_data[j];
				_data[j]=tmp;
			}
			i++;
		}
	}
}

NodeP ComplexArray::combSort()
{
	ComplexArrayP a=new ComplexArray(this);
	if throwing return Unbound;
	a->combSortInPlace();
	return a;
}

NodeP ComplexArray::readbin(FileP file)
{
	ThreadP th=Thread::This();
	complexdP ip=data();

	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read ComplexArray binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(double), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(double)*_length;
			if (th->rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *doubleP(th->rstringptr);	// reading from string
					th->rstringptr+=sizeof(double);
				}
			else
				error(FILE_ERROR, new String
					(_("memory block too small to read ComplexArray")));
		}
	return this;
}

NodeP lreadComplexArrayBin(PairPP arg) 
{
	enum Params {_size=1, _origin};
	int size=arg[_size]->a_int(_size);
	if throwing return Unbound;
	int origin=1;
	if (exist(_origin))
	{	origin=arg[_origin]->a_int(_origin);
		if throwing return Unbound;
	}
	ComplexArrayP a=new ComplexArray(size, origin);
	if throwing return Unbound;
	return a->readbin(Thread::This()->Reader);
}

double ComplexArray::abs(complexd c)
{
	return Complexd::abs(c);
}