#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/utils.h>
#include <wx/app.h>
#include <wx/textctrl.h>
#include <wx/timer.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
	#include <process.h>
	#include <direct.h>
#else
	#include <unistd.h>
#endif
	
#include "Main.h"

#include "String.h"
#include "File.h"
#include "Line.h"
#include "List.h"
#include "Tree.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Table.h"
#include "Procedure.h"
#include "Color.h"

#include "Error.h"
#include "Eval.h"
#include "NodePStack.h"
#include "Stack.h"
#include "Init.h"
#include "LogoApp.h"
#include "Environment.h"
#include "Security.h"
#include "Thread.h"
#include "Workspace.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

NodeP lignore(PairPP)
{
	return Unbound;
}

NodeP lisCaseIgnored(PairPP)
{
	ThreadP th=Thread::This();
	return th->caseIgnored ? _True : _False;
}

NodeP lsetCaseIgnored(PairPP arg)
{
	ThreadP th=Thread::This();
	int ci=arg[1]->a_bool(1);
	if (!throwing)
	{	th->caseIgnored=(ci==1);
	}
	return Unbound;
}

NodeP lgetEnv(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	wxString val;
	if (wxGetEnv(name->ptr(), &val))
		return new String(val);
	else
		return _Null_Word;
}

NodeP lsetEnv(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	if (!arg[2])
	{	error(BAD_DATA, arg[2], 2); return Unbound;
	}
	StringP val=arg[2]->string();
	if throwing return Unbound;
	if (!wxSetEnv(name->ptr(), val->ptr()))
		error(BAD_DATA, val, 2);
	return Unbound;
}

NodeP lgetEnvTable(PairPP arg)
{
#ifdef WIN32
	charP env=GetEnvironmentStrings();
	charP c=env;
	TableP table=new Table(121);
	if throwing return Unbound;
	int i=1;
	for (c=env; *c; c++)
	{	charP c2=strchr(c, '=');
		StringP name=String::make(c, 0, c2-c, STRING, String::copy);
		if throwing return Unbound;
		c2++;
		int len=strlen(c2);
		StringP val=String::make(c2, 0, len, STRING, String::copy);
		if throwing return Unbound;
		table->setitem(name, val, false);
		if throwing return Unbound;
		while (*c) 
			*c++; 
		i++;
	}
	FreeEnvironmentStrings(env);
#else
//	TODO: test it!
	charP *env=environment.env;
	if (!env) return 0;
	TableP table=new Table(121);
	if throwing return Unbound;
	for (int i=0; env[i]; i++)
	{	for (charP c=env[i]; *c; c++)
		{	charP c2=strchr(c, '=');
			StringP name=String::make(c, 0, c2-c, STRING, String::copy);
			if throwing return Unbound;
			c2++;
			int len=strlen(c2);
			StringP val=String::make(c2, 0, len, STRING, String::copy);
			if throwing return Unbound;
			table->setitem(name, val, false);
			if throwing return Unbound;
			while (*c) 
				*c++; 
		}
	}
#endif
	return table;
}

NodeP lstop(PairPP) 
{
	if throwing return Unbound;
	Thread::This()->stoppingFlag=STOP;
	return Unbound;
}

NodeP loutput(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!throwing) 
	{	th->stoppingFlag=OUTPUT; 
		th->output_node=arg[1];
	}
	return Unbound;
}

NodeP lrun(PairPP arg) // macroized
{    
	NodeP instructions = arg[1]->a_runnable(1);
	return instructions;
}

class RunSecureTimer: public wxTimer
{
public:
	void Notify()
	{	ThreadP th=Thread::This();
		if (!throwing)
		{	th->throw_node=_runsecuretimer;
			th->stoppingFlag=THROWING;
			th->output_node = (PairP)Unbound;
		}
	}
};

NodeP lrunSecure(PairPP arg)
{  
	enum Params 
	{	_level=1, _runlist, 
		_stacksize, _recordsize, _memsize, _nodes,
		_timeout, _namespce, _vars,
	};
	ThreadP th=Thread::This();
	Security *oldSecurity=th->security;
	Security security=*oldSecurity;
	int level=arg[_level]->a_int(_level);
	if throwing return Unbound;
	if (exist(_stacksize))
	{	security.stacksize=arg[_stacksize]->a_int(_stacksize);
		if throwing return Unbound;
	}
	if (exist(_recordsize))
	{	security.recordsize=arg[_recordsize]->a_int(_recordsize);
		if throwing return Unbound;
	}
	if (exist(_memsize))
	{	security.memsize=arg[_memsize]->a_int(_memsize);
		if throwing return Unbound;
	}
	if (exist(_nodes))
	{	security.nodes=arg[_nodes]->a_int(_nodes);
		if throwing return Unbound;
	}
	if (exist(_timeout))
	{	security.timeout=arg[_timeout]->a_int(_timeout);
		if throwing return Unbound;
	}
	TableP namespce=th->namespce;
	ObjectP namespceobj=th->namespceobj;
	if (exist(_namespce))
	{	StringP str=arg[_namespce]->a_string(_namespce);
		if throwing return Unbound;
		ObjectP obj=str->internRoot();
		if (obj != _Null_Word)
		{	if (obj->table()==Unbound)
			{	TableP table=new Table(31);
				if throwing return Unbound;
				obj->settable(table);
			}
			if (obj->var()==Unbound)
				obj->setvar(0);
			namespce=obj->table();
		}else
			namespce=root;
		namespceobj=obj;
	}
	PairP vars=0;
	if (exist(_vars))
	{	vars=arg[_vars]->a_list(_vars);
		if throwing return Unbound;
	}
	security.level &= ~level;
	if (security < *oldSecurity)
	{	error(SECURITY_ERROR, new String(_("to use less security"))); return Unbound;
	}
	th->security=&security;

	TableP ons=th->namespce;
	ObjectP onso=th->namespceobj;
	th->namespce=namespce;
	th->namespceobj=namespceobj;

	PairP instructions = arg[_runlist]->a_runnable(_runlist);
	if throwing 
	{	th->security=oldSecurity;
		th->namespce=ons;
		th->namespceobj=onso;
		return Unbound;
	}
	Evaluator e;
	th->val_status=value_maybe_ok_in_macro;
	
	int vsp=*th->varStack;
	PairP v=vars;
	while (v)
	{	NodeP var=v->first();
		ObjectP obj=var->internRoot();
		NodeP val=th->varStack->find(obj);
		obj=var->intern();
		PushVar(th->varStack, obj, val, dummy);
		next(v);
	}
//c_out << int(security.memsize) << endL;
	NodeP result=0;
	if (security.timeout !=-1)
	{	RunSecureTimer t;
		t.Start(security.timeout, wxTIMER_ONE_SHOT);
		result=e.eval(instructions, __begin_line, th);
		t.Stop();
	}else
		result=e.eval(instructions, __begin_line, th);

	th->varStack->pop_till(vsp);
	th->namespce=ons;
	th->namespceobj=onso;
	th->security=oldSecurity;
	return result;
}

NodeP lrunResult(PairPP arg) 
{	
	NodeP result=lrun(arg);
	if throwing return 0;
	return new ToDoNode(__runresult_continuation, result);
}
NodeP lapply(PairPP arg)	// The logo word APPLY. 
{
	PairP inst=new Pair(arg[1], arg[2]);
	if throwing return 0;
	return new ToDoNode(__begin_apply, inst);
}
NodeP lgoTo(PairPP arg) 
{
	return new ToDoNode(__goto_continuation, arg[1]);
}
NodeP ltag(PairPP) {return Unbound;}

NodeP lcatch(PairPP arg) 
{	
	ThreadP th=Thread::This();
	if (_runsecuretimer->equalP(arg[1]) && th->security->timeout != -1)
	{	error(SECURITY_ERROR, new String(_("to use the runSecureTimer twice"))); 
		return Unbound;
	}
	NodeP inst=lrun(arg+1);
	if throwing return 0;
	PairP list=new Pair(arg[1], inst);
	if throwing return 0;
	return new ToDoNode(__catch_continuation, list);
}

NodeP lthrow(PairPP arg) 
{
	if throwing return Unbound;
	if (arg[1]->equal(_error)) 
	{	if (exist(2))
			error(USER_ERR, arg[2]);
		else
			error(USER_ERR, Unbound);
	}else 
	{	Thread::This()->stoppingFlag=THROWING;
		Thread::This()->throw_node=arg[1];
	
		if (exist(2))
			Thread::This()->output_node=arg[2];
		else
			Thread::This()->output_node=(PairP)Unbound;
	}
	return Unbound;
}

NodeP lquestionmark(PairPP arg)		// The logo word ?
{
	if (noarg) 
		if (Thread::This()->qmList>=0) 
			return Thread::This()->argStack->base[Thread::This()->qmList+1];
		else return error(BAD_DATA_UNREC, Thread::This()->fun);
	else
	{	int argnum = arg[1]->a_pos_int(1)->int_;

		if (throwing) return Unbound;

		if (Thread::This()->qmList < 0)
			return error(BAD_DATA, new Int(argnum));

		PairPP np = PairPP(&(Thread::This()->argStack->base[Thread::This()->qmList+argnum]));
//		PairPP np = &qmList[1];
//		int i=argnum;
//		while (--i > 0  &&  np!=argStack  &&  *++np)
//			;
		if (!argnum || unsigned(np)>=unsigned(Thread::This()->argStack->index) || !*np)
			return error(BAD_DATA_UNREC, new Int(argnum));

		return *np;
	}
}

NodeP ltest(PairPP arg) 
{
	int b = arg[1]->a_bool(1);

	if (throwing || Thread::This()->tailcall != in_sequence) 
		return Unbound;

	Thread::This()->ift_iff_flag=b;
	Thread::This()->dont_fix_ift=true;
	return Unbound;
}

NodeP lifTrue(PairPP arg) 
{
	if (Thread::This()->ift_iff_flag < 0)
		return error(NO_TEST, 0);
	else
		if (Thread::This()->ift_iff_flag > 0)
			return lrun(arg);
		else
			return 0;
}

NodeP lifFalse(PairPP arg) 
{
	if (Thread::This()->ift_iff_flag < 0)
		return error(NO_TEST, 0);
	else
		if (Thread::This()->ift_iff_flag==0)
			return lrun(arg);
		else
			return 0;
}

NodeP lifElse(PairPP arg)  // macroized
{    
	int cond = arg[1]->a_bool(1);
	NodeP
		yes= arg[2]->a_runnable(2),
		no = arg[3]->a_runnable(3);

	if (throwing) return 0;
	return cond? yes : no;
}

NodeP lif(PairPP arg) 	// macroized
{
	if (exist(3)) return lifElse(arg);
	
	 int pred = arg[1]->a_bool(1);
	NodeP yes = arg[2]->a_runnable(2);

	if throwing return 0;
	return pred? yes : 0;
}

NodeP lcase(PairPP arg)
{
	NodeP value=arg[1];
	PairP conds=arg[2]->a_list(2);
	if (!conds)
		return 0;
	if throwing return 0;

	Evaluator e;
	Thread::This()->val_status=value_maybe_ok_in_macro;

	for (PairP l=conds;  l;  next(l))
	{	if (l->first())
		{	if (l->first()->isnolist())
			{	error(BAD_DATA, l, 2);
				return 0;
			}
			PairP cond=l->first()->first();
			if (!cond)
			{	error(BAD_DATA, l, 2);
				return 0;
			}
			if (cond->equal(_Else))
				return l->first()->tail();

			cond=cond->a_runnable(2);
			if throwing return 0;
			NodeP caseval;
			caseval=e.eval(cond, __begin_sequence, Thread::This());


			if (caseval && caseval->equal(value))
				return l->first()->tail();
		}
		else if (value==NodeP(0))
			return l->tail();
	}
	return 0;
}

NodeP lforever(PairPP arg) 
{
	NodeP instructionlist = arg[1]->a_runnable(1);
	if throwing return 0;
	PairP list=new Pair( new Int(-1), instructionlist);
	return new ToDoNode(	__repeat_continuation, list);
}

NodeP lrepeat(PairPP arg) 
{
	IntP count = arg[1]->a_pos_int(1);
	PairP instructionlist = arg[2]->a_runnable(2);

	if (throwing || !count->int_) return 0;
	PairP list=new Pair(count, instructionlist);
	if throwing return 0;
	return new ToDoNode(__repeat_continuation, list);
}

NodeP lrepeatCount(PairPP)
{
	return new Int(Thread::This()->user_repcount);
}

NodeP lwhile(PairPP arg) 
{
	PairP         cond = arg[1]->a_runnable(1),
		instructionlist = arg[2]->a_runnable(2);
	if throwing return 0;
	PairP list=new Pair(cond, instructionlist);
	if throwing return 0;
	return new ToDoNode(__while_continuation, list);
}

NodeP luntil(PairPP arg) 
{
	PairP         cond = arg[1]->a_runnable(1),
		instructionlist = arg[2]->a_runnable(2);
	if throwing return 0;
	PairP list=new Pair(cond, instructionlist);
	if throwing return 0;
	return new ToDoNode(__until_continuation, list);
}

NodeP ldo_while(PairPP arg) 
{
	PairP
		instructionlist = arg[1]->a_runnable(1),
					  cond = arg[2]->a_runnable(2);
	if throwing return 0;
	PairP list=new Pair(cond, instructionlist);
	if throwing return 0;
	return new ToDoNode(__do_while_continuation, list);
}

NodeP ldo_until(PairPP arg) 
{
	PairP
		instructionlist = arg[1]->a_runnable(1),
					  cond = arg[2]->a_runnable(2);
	if throwing return 0;
	PairP list=new Pair(cond, instructionlist);
	if throwing return 0;
	return new ToDoNode(__do_until_continuation, list);
}

NodeP lfor(PairPP arg)	// What a messy primitive! But faster than for_old
{
	PairP l=arg[1]->a_runnable(1),
		instructionlist = arg[2]->a_runnable(2);

	if throwing return 0;
	if (!l 
	|| !l->first() || !l->tail() 
	|| !l->second() || !l->tail()->tail()
	|| !l->third())
	{	error(BAD_DATA, l);
		return 0;
	}
	NodeP name=l->first()->a_name(1);
	if throwing return 0;

	PairP start=l->second(),
		end=l->third(),
		step=0;

	if (start->isnolist())
	{	start=new Line(start, 0);
		if throwing return 0;
	}
	if (end->isnolist())
	{	end=new Line(end, 0);
		if throwing return 0;
	}
	double dstart, dend, dstep;

	Evaluator e;
	NodeP startval, endval, stepval;
	Thread::This()->val_status=value_maybe_ok_in_macro;

	startval=e.eval(start, __begin_sequence, Thread::This());
	if throwing return 0;
	dstart=startval->a_float(1);
	if throwing return 0;

	  endval=e.eval( end , __begin_sequence, Thread::This());
	if throwing return 0;
	dend=endval->a_float(1);
	if throwing return 0;

	if (l->tail()->tail()->tail() && l->third())
	{	step=l->fourth();
		if (step->isnolist())
		{	step=new Line(step, 0);
			if throwing return 0;
		}
		stepval=e.eval(step, __begin_sequence, Thread::This());
		if throwing return 0;
		dstep=stepval->a_float(1);
	}
	else
		dstep= dend-dstart >= 0 ? 1 : -1;

	if ((dend-dstart > 0  &&  dstep < 0)
	||  (dend-dstart < 0  &&  dstep > 0))
		return 0;

	if (dstart==int(dstart) && dend==int(dend) && dstep==int(dstep))
		return new ToDoNode
		(	__ifor_continuation,
			new Pair(name,
			new Pair(new Int(int(dstart)), 
			new Pair(new Int(int(dend)), 
			new Pair(new Int(int(dstep)), 
			new Pair(instructionlist, 0)))))
		);

	return new ToDoNode
	(	__for_continuation, 
		new Pair(name,
		new Pair(new Float(dstart), 
		new Pair(new Float(dend), 
		new Pair(new Float(dstep), 
		new Pair(instructionlist, 0)))))
	);
}

NodeP lforeach(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP list=0;
	NodeP instructions=*(th->argStack->index);
	instructions=instructions->a_runnable(2);
	if (!instructions)
		return 0;

	NodePP i=th->argStack->index-1;
	if (!*i)
		return 0;

	while (i != NodePP(arg))
	{	if (*i && (*i)->type==PROC_)
			push(list, ProcedureP(*i--)->bodywords());
		else
			push(list, *i--);
	}
	return new ToDoNode
	(	__foreach_continuation,
		new Pair(list, instructions)
	);
}

NodeP lbreak(PairPP)
{
	Thread::This()->stoppingFlag=BREAK;
	return Unbound;
}

NodeP lcontinueLoop(PairPP)
{
	Thread::This()->stoppingFlag=CONTINUE;
	return Unbound;
}

NodeP lbye(PairPP) 
{
	ThreadP th=Thread::This();
	if (!th->security->shell)
	{	error(SECURITY_ERROR, new String(_("to use bye"))); return Unbound;
	}
	if (!Thread::This()->ufun  
	&&  loader.file()==File::conin 
	&& theApp->mainShown) 
		c_out << _("Thank you for using Logo.") << endL
			  << _("Have a nice day.") << endL;

	theApp->action=EXIT_LOGO;
	return Unbound;
}

NodeP lLogoVersion(PairPP)
{
//	StringP os=0;
/*	switch (theApp->osVersion)
	{
case wxCURSES:			os=new String("Curses"); break;				//  Text-only CURSES 
case wxXVIEW_X:		os=new String("XView X"); break;				//  Sun's XView OpenLOOK toolkit 
case wxMOTIF_X:		os=new String("Motif X"); break;				//  OSF Motif 1.x.x 
case wxCOSE_X:			os=new String("COSE"); break;					//  OSF Common Desktop Environment 
case wxNEXTSTEP:		os=new String("NeXTStep"); break;			//  NeXTStep 
case wxMAC:				os=new String("Mac OS 8/9/X with Mac paths"); break;	//  Apple Mac OS 8/9/X with Mac paths 
case wxMAC_DARWIN:	os=new String("Mac OS X with Unix paths"); break;		//  Apple Mac OS X with Unix paths 
case wxBEOS:			os=new String("BeOS"); break;					//  BeOS 
case wxGTK:				os=new String("GTK on X"); break;			//  GTK on X 
case wxGTK_WIN32:		os=new String("GTK on Win32"); break;		//  GTK on Win32 
case wxGTK_OS2:		os=new String("GTK on OS/2"); break;		//  GTK on OS/2 
case wxGTK_BEOS:		os=new String("GTK on BeOS"); break;		//  GTK on BeOS 
case wxGEOS:			os=new String("GEOS"); break;					//  GEOS 
case wxOS2_PM:			os=new String("OS/2 Workplace"); break;	//  OS/2 Workplace 
case wxWINDOWS:		os=new String("Windows or WfW"); break;	//  Windows or WfW 
case wxMICROWINDOWS:	os=new String("MicroWindows"); break;		//  MicroWindows 
case wxPENWINDOWS:	os=new String("Pen Windows"); break;		//  Windows for Pen Computing 
case wxWINDOWS_NT:	os=new String("Windows NT"); break;			//  Windows NT 
case wxWIN32S:			os=new String("Windows 32S API"); break;	//  Windows 32S API 
case wxWIN95:			os=new String("Windows 95"); break;			//  Windows 95 
case wxWIN386:			os=new String("Watcom 32-bit supervisor modus"); break;	//  Watcom 32-bit supervisor modus 
case wxWINDOWS_CE:	os=new String("Windows CE (generic)"); break;				//  Windows CE (generic) 
case wxWINDOWS_POCKETPC: os=new String("Windows CE PocketPC"); break;			//  Windows CE PocketPC 
case wxWINDOWS_SMARTPHONE: os=new String("Windows CE Smartphone"); break;		//  Windows CE Smartphone 
case wxMGL_UNIX:		os=new String("MGL with direct hardware access"); break;	//  MGL with direct hardware access 
case wxMGL_X:			os=new String("MGL on X"); break;				//  MGL on X 
case wxMGL_WIN32:		os=new String("MGL on Win32"); break;			//  MGL on Win32 
case wxMGL_OS2:		os=new String("MGL on OS/2"); break;			//  MGL on OS/2 
case wxMGL_DOS:		os=new String("MGL on MS-DOS"); break;			//  MGL on MS-DOS 
case wxWINDOWS_OS2:	os=new String("Native OS/2 PM"); break;		//  Native OS/2 PM 
case wxUNIX:			os=new String("wxBase under Unix"); break;	//  wxBase under Unix 
case wxX11:				os=new String("Plain X11 and Universal widgets"); break;	//  Plain X11 and Universal widgets 
case wxPALMOS:			os=new String("PalmOS"); break;					//  PalmOS 
case wxDOS:				os=new String("wxBase under MS-DOS"); break;	//  wxBase under MS-DOS 
case wxUNKNOWN_PLATFORM: 
default:					os=new String("Unknown Platform"); break;
	}
*/
	wxString str;
	str << theApp->osMajor << "." << theApp->osMinor;
	double osVersionNr;
	str.ToDouble(&osVersionNr);

	return new List(
		new String(aUCBLogo), 
		new Float(versionNumber),
		new String(theApp->osDescription), 
		new Float(osVersionNr),
		_ListEnd);
}

NodeP lgetArchitectureName(PairPP arg)
{
	wxPlatformInfo plati;
	return new String(plati.GetArchName());
}

NodeP lgetEndiannessName(PairPP arg)
{
	wxPlatformInfo plati;
	return new String(plati.GetEndiannessName());
}

NodeP lgetOperatingSystemFamilyName(PairPP arg)
{
	wxPlatformInfo plati;
	return new String(plati.GetOperatingSystemFamilyName());
}

NodeP lgetOperatingSystemIdName(PairPP arg)
{
	wxPlatformInfo plati;
	return new String(plati.GetOperatingSystemIdName());
}

NodeP lgetPortIdName(PairPP arg)
{
	wxPlatformInfo plati;
	return new String(plati.GetPortIdName());
}

NodeP lOSScreenSize(PairPP)
{
	return new List(
		new Int(theApp->max.x),
		new Int(theApp->max.y), _ListEnd);
}

NodeP lsetProcessIdleInterleaving(PairPP arg)
{
	ThreadP th=Thread::This();
	int i=arg[1]->a_int(1);
	if throwing return Unbound;
	if (i > 0)
		th->processIdleInterleaving=i;
	else
		error(BAD_DATA, arg[1]);
	return Unbound;
}

NodeP lThread(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->thread)
	{	error(SECURITY_ERROR, new String(_("to use Thread"))); return Unbound;
	}
	PairP commands=arg[1]->a_runnable(1);
	if throwing return Unbound;
	ThreadP thread=new Thread(wxTHREAD_DETACHED);
	if (!thread)
	{	error(OUT_OF_MEM); return Unbound;
	}
	thread->Create(0x10000, commands, __begin_line, Thread::This()->graph, true);
	thread->Run();
	return Unbound;
}

NodeP lThreadJoinable(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->thread)
	{	error(SECURITY_ERROR, new String(_("to use Thread"))); return Unbound;
	}
	PairP commands=arg[1]->a_runnable(1);
	if throwing return Unbound;
	ThreadP thread=new Thread(wxTHREAD_JOINABLE);
	if (!thread)
	{	error(OUT_OF_MEM); return Unbound;
	}
	thread->Create(0x10000, commands, __begin_line, Thread::This()->graph, true);
	thread->Run();
	thread->Wait();
//	while (thread->IsAlive() && thread->stoppingFlag!=THROWING)
//		theApp->dispatchMessages();
	NodeP output=thread->output_node;
	delete thread;
	return output;
}

void __stdcall fiberEntry(void *p)
{
	ThreadP th=ThreadP(p);
	th->Entry();
}

NodeP lFiber(PairPP arg)
{
#ifdef WIN32
	ThreadP th=Thread::This();
	if (!th->security->thread)
	{	error(SECURITY_ERROR, new String(_("to use Fiber"))); return Unbound;
	}
	PairP commands=arg[1]->a_runnable(1);
	if throwing return Unbound;
	ThreadP thread=new Thread(wxTHREAD_DETACHED);
	if (!thread)
	{	error(OUT_OF_MEM); return Unbound;
	}
	thread->Create(0, commands, __begin_line, Thread::This()->graph, false);
//	thread->Run();

	if (!th->fiber)
		th->fiber=(Fiber*)ConvertThreadToFiber(th);
	if (!th->fiber)
	{	c_out << _("; ConvertThreadToFiber failed with code ") 
		<< int(GetLastError()) << endL;
		return 0;
	}
	thread->fiber=(Fiber*)CreateFiber(0x10000, fiberEntry, thread);
	if (!thread->fiber)
	{	error(FILE_ERROR, new String("CreateFiber error"));
		return Unbound;
	}
	thread->threadFiber=th->threadFiber;
	thread->startFiber=th->startFiber;
	ThreadP nf=th->nextFiber;
	th->nextFiber=thread;
	thread->prevFiber=th;
	thread->nextFiber=nf;
	nf->prevFiber=thread;
#else
// TODO: use PCL or something like this to emulate our own Fibers
	error(NOT_YET_SUPPORTED, 
		new String(_("Fibers are not yet supported under Linux")));
#endif
	return Unbound;
}

NodeP lgetProcessId(PairPP arg)
{
	return new Int(wxGetProcessId());
}

NodeP lgetThreadId(PairPP arg)
{
	ThreadP th=Thread::This();
	return new Int(th->GetId());
}

NodeP lsetNamespace(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->namespce)
	{	error(SECURITY_ERROR, new String(_("to set the namespace"))); 
		return Unbound;
	}
	StringP namespce=arg[1]->a_string(1);
	ObjectP obj=namespce->internRoot();
	if (obj != _Null_Word)
	{	if (obj->table()==Unbound)
		{	TableP table=new Table(31);
			if throwing return Unbound;
			obj->settable(table);
		}
		if (obj->var()==Unbound)
			obj->setvar(0);
		th->namespce=obj->table();
	}else
		th->namespce=root;
	th->namespceobj=obj;
	return Unbound;
}

NodeP lNamespace(PairPP)
{
	ThreadP th=Thread::This();
	if (!th->security->namespce)
	{	error(SECURITY_ERROR, new String(_("to read out the namespace"))); 
		return Unbound;
	}
	return th->namespceobj->fullName();
}

NodeP lsetLogoLanguage(PairPP arg)
{
	enum Paras {_logoLanguage=1};
	ThreadP th=Thread::This();
	if (!th->security->language)
	{	error(SECURITY_ERROR, new String(_("to change the LogoLanguage"))); 
		return Unbound;
	}
	StringP logoLanguage=arg[_logoLanguage]->a_string(_logoLanguage);
	if throwing return Unbound;
	if (logoLanguage->equalP(_aucblogo))
	{	th->logoLanguage=_aucblogo;
		_Assign_Sign->setprim(_lmake->prim());
		_Equality_Op->setprim(_lisequal->prim());
	}else if (logoLanguage->equalP(_ucblogo))
	{	th->logoLanguage=_ucblogo;
		_Assign_Sign->setprim(_lisequal->prim());
		_Equality_Op->setvar( Unbound );
	}else
		error(BAD_DATA, logoLanguage);
	return Unbound;
}

NodeP lLogoLanguage(PairPP)
{
	ThreadP th=Thread::This();
	return th->logoLanguage;
}

NodeP lsetLineInfoInTrace(PairPP arg)
{
	enum Paras {_lineInfoInTrace=1};
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to change GUI properties"))); 
		return Unbound;
	}
	int lineInfoInTrace=arg[_lineInfoInTrace]->a_bool(_lineInfoInTrace);
	if throwing return Unbound;
	th->lineInfoInTrace=(lineInfoInTrace!=0);
	return Unbound;
}