//___________________________________________________________________
//
//                           Class Console
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"

#include <wx/textctrl.h>
//#include <wx/html/helpctrl.h>
#include <wx/help.h>
#include <wx/timer.h>
#include <wx/clipbrd.h>
#include <wx/filename.h>
#include <wx/notebook.h>
#include <wx/listbox.h>
#include <wx/menu.h>
#include <wx/dialog.h>
#include <wx/process.h>
#include <wx/caret.h>
#include <wx/aui/aui.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>

#include <fstream>

#include <math.h>

#ifdef WIN32
	#include <process.h>
	#include <conio.h>
#else
	#include <unistd.h>
#endif

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Line.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "TextControl.h"

#include "Error.h"
#include "Init.h"
#include "NodePStack.h"
#include "Files.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Mouse.h"
#include "Font.h"
#include "Turtle.h"
#include "Recorder.h"

#include "Event.h"
#include "Console.h"
#include "Consolew.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"

#include "EdFrame.h"
#include "Thread.h"
#include "Security.h"
#include "DebugState.h"
#include "History.h"
#include "Environment.h"
#include "Aui.h"
#include "Bitmap.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

static wxCriticalSection writeBufferCS;

ConsoleFrame::ConsoleFrame(wxWindow *parent)
:wxPanel(parent, ID_CONSOLEFRAME, 
			wxPoint(0,0), wxSize(600,600), 
			wxNO_BORDER)
{
	ThreadP th=Thread::This();
	fileMenu = new wxFlatMenu;
	if (!fileMenu)
		return;
	fileMenu->Append(wxID_NEW   ,_("&New\tCtrl-N")    ,_("New File"));
	fileMenu->Append(wxID_OPEN  ,_("&Load to Workspace...\tCtrl-L"),_("Load File into Workspace"));
	fileMenu->Append(wxID_SAVE  ,_("&Save Workspace...\tCtrl-S"),_("Save Workspace into File"));
//	fileMenu->Append(wxID_SAVEAS,_("Save &As...")     ,_("Save as"));
	fileMenu->AppendSeparator();
	fileMenu->Append(ID_CONSOLE_OPEN  ,_("Op&en Console..."),_("Open Console"));
	fileMenu->Append(ID_CONSOLE_SAVE  ,_("Sa&ve Console..."),_("Save Console"));
	fileMenuConsoleAutoSave=fileMenu->AppendCheckItem(
		ID_CONSOLE_AUTOSAVE, _("Autosave Console"));
	fileMenuConsoleAutoSave->Check(theApp->consoleAutosave);
	fileMenu->AppendSeparator();
	fileMenu->Append(ID_PICTURE_OPEN  ,_("Open Picture..."),_("Open Picture"));
	fileMenu->Append(ID_PICTURE_SAVE  ,_("Save.Picture.."),_("Save Picture"));
	fileMenu->Append(ID_PICTURE_TEXT_OPEN  ,_("Open Picture in Textmode..."),_("Open Picture in Textmode"));
	fileMenu->Append(ID_PICTURE_TEXT_SAVE  ,_("Save.Picture in Textmode..."),_("Save Picture in Textmode"));
	fileMenuPictureAutoSave=fileMenu->AppendCheckItem(
		ID_PICTURE_AUTOSAVE, _("Autosave Picture"));
	fileMenuPictureAutoSave->Check(theApp->pictureAutosave);
	fileMenu->AppendSeparator();
//	nodesAutosave=nodesAutosave_;
#ifdef PERSISTENCE
	fileMenu->Append(ID_NODES_OPEN  ,_("Open Nodes..."),_("Open Nodes"));
	fileMenu->Append(ID_NODES_SAVE  ,_("Save Nodes..."),_("Save Nodes"));
	fileMenuNodesAutoSave=fileMenu->AppendCheckItem(
		ID_NODES_AUTOSAVE, _("Autosave Nodes"));
	fileMenuNodesAutoSave->Check(theApp->nodesAutosave);
	fileMenu->AppendSeparator();
#endif
	fileMenu->Append(wxID_PRINT ,_("&Print...\tCtrl-P")  ,_("Print File"));
	fileMenu->Append(wxID_PRINT_SETUP ,_("P&rint Setup..."),_("Print Setup"));
	fileMenu->Append(ID_PAGE_SETUP, _("Pa&ge Setup..."), _("Page setup"));
	fileMenu->Append(wxID_PREVIEW ,_("Pr&int Preview")  ,_("Print Preview"));
	fileMenu->AppendSeparator();
	fileMenu->Append(wxID_EXIT  ,_("E&xit\tAlt-X")    ,_("Exit"));

	editMenu = new wxFlatMenu;
	if (!editMenu)
		return;
	editMenu->Append(wxID_UNDO  ,_("&Undo\tCtrl-Z") , _("Undo Edit"));
	editMenu->Append(wxID_CUT   ,_("Cu&t\tCtrl-X")  , _("Cut Text"));
	editMenu->Append(wxID_COPY  ,_("&Copy\tCtrl-C") , _("Copy Text"));
	editMenu->Append(wxID_PASTE ,_("&Paste\tCtrl-V"), _("Paste Text"));
	editMenu->AppendSeparator();
	editMenu->Append(ID_EDIT_REPEAT ,_("Repeat &last command\tF2"), _("Repeat last command"));
	editMenu->Append(ID_EDIT_RESET  ,_("&Reset\tCtrl-F2"), _("Call reset.lg"));
	editMenu->AppendSeparator();
	editMenu->Append(ID_EDIT_GOTO_CURSOR_LINE,_("&Goto Line under Cursor\tCtrl-G"), _("Goto Line under Cursor"));
	editMenu->AppendSeparator();
	editMenuWrapLines=editMenu->AppendCheckItem(
		ID_EDIT_WRAP_LINES,_("&Wrap Lines"), _("Wrap Lines"));
	editMenuWrapLines->Check(!theApp->consoleHSCROLL);
	editMenu->AppendSeparator();
	editMenu->Append(ID_EDIT_LANGUAGE, _("Choose &Language"), _("Choose Language"));

	helpMenu = new wxFlatMenu;
	if (!helpMenu)
		return;
	helpMenu->Append(wxID_HELP_CONTENTS,_("&Help\tF1"),_("Help"));
	helpMenu->AppendSeparator();
	helpMenu->Append(wxID_ABOUT,_("&About"),_("About"));

	menuBar=new wxFlatMenuBar(this);
	if (!menuBar)
		return;
	menuBar->Append(fileMenu  ,_("&File"));
	menuBar->Append(editMenu  ,_("&Edit"));
	menuBar->Append(helpMenu  ,_("&Help"));

	int style=wxTE_MULTILINE | wxTE_PROCESS_TAB | wxTE_RICH2 | wxTE_NOHIDESEL;
	if (theApp->consoleHSCROLL)
		style|=wxHSCROLL;
	console=new Console(this, ID_CONSOLE, "", 
		wxDefaultPosition, wxDefaultSize, style,
		wxDefaultValidator, _("Console"), theApp->consoleAutosave, theApp->caretpos);

	wxBoxSizer *sizer=new wxBoxSizer(wxVERTICAL);
	sizer->Add(menuBar, 0, wxEXPAND);
	sizer->Add(console, 1, wxEXPAND);
	SetSizer(sizer);

	show();
	if (theApp->mainFrame && theApp->mainFrame->viewMenuConsole)
		theApp->mainFrame->viewMenuConsole->Check(true);
}

ConsoleFrame::~ConsoleFrame()
{
	theApp->memoryFrame=0;
	if (theApp->mainFrame && theApp->mainFrame->viewMenuConsole)
		theApp->mainFrame->viewMenuConsole->Check(false);
}

void ConsoleFrame::show(bool shown)
{
Show();
	wxAuiPaneInfo &pane=theApp->aui->GetPane(this);
	if (pane.IsOk())
	{	pane.Show();
	}else
		theApp->aui->AddPane(this, 
			wxAuiPaneInfo()
			.Name("Console").Caption(_("Console"))
			.Center()
			.BestSize(600,600)
			.MaximizeButton().MinimizeButton().PinButton()
		);
	theApp->aui->Update();
}

BEGIN_EVENT_TABLE(Console, wxTextCtrl)
	EVT_SET_FOCUS		(Console::OnSetFocus				)
	EVT_ENTER_WINDOW	(Console::OnEnterWindow			)
	EVT_LEAVE_WINDOW	(Console::OnLeaveWindow			)
	EVT_CHAR				(Console::OnChar					)
	EVT_KEY_DOWN		(Console::OnKeyDown				)
	EVT_KEY_UP			(Console::OnKeyUp					)
	EVT_LEFT_DOWN		(Console::OnMouseLeftDown		)
	EVT_RIGHT_DOWN		(Console::OnMouseRightDown		)
	EVT_MIDDLE_DOWN	(Console::OnMouseMiddleDown	)
	EVT_LEFT_UP			(Console::OnMouseLeftUp			)
	EVT_RIGHT_UP		(Console::OnMouseRightUp		)
	EVT_MIDDLE_UP		(Console::OnMouseMiddleUp		)
	EVT_LEFT_DCLICK	(Console::OnMouseLeftDClick	)	
	EVT_RIGHT_DCLICK	(Console::OnMouseRightDClick	)
	EVT_MIDDLE_DCLICK	(Console::OnMouseMiddleDClick	)
	EVT_MOTION			(Console::OnMouseMotion			)
	EVT_MOUSE_EVENTS	(Console::OnMouse					)
	EVT_CONTEXT_MENU	(Console::OnContextMenu			)
	EVT_HELP	(wxEVT_HELP, Console::OnHelp				)
	EVT_MENU(ID_GOTO_LINE, Console::OnGotoLine)
	EVT_MENU(ID_OPEN_FILE_AT_CURSOR, Console::OnOpenFileAtCursor)
	EVT_MENU_RANGE(ID_CONSOLE_INSERT, ID_CONSOLE_INSERT+1000, Console::OnInsert)
	EVT_IDLE(Console::OnIdle)
	EVT_END_PROCESS(wxID_ANY, Console::OnProcessTerm)
END_EVENT_TABLE()

Console::Console(wxWindow* parent, wxWindowID id, 
			const wxString& value, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name,
			bool autosave_,
			int caretpos):
wxTextCtrl(parent, id, value, pos, size, style, validator, name),
autosave(autosave_)
{
	writeBufferSize=0x1000;
	readIndex=0;
	charAvailable=lineAvailable=false;
	charMode=false;
	
	standout='\001';
	plainMode='\002';
	insertMode=true;
	
	coord=wxPoint(0,0);
	Max.x=0x1000; // hard coded; should be improved;
	Max.y=0x1000; // but seems to be faily big enough at the moment...
	Thread::This()->interactive=true;

	textColor=0;
	backgroundColor=(unsigned int)(-1);

#ifdef WIN32
	fontSize=10;
#else
	fontSize=10;
#endif
	fontFamily=wxMODERN;
	fontStyle=wxNORMAL;
	fontWeight=wxNORMAL;
	fontName="Courier New";
	underline=false;
	tabSize=80; // in 0.1 millimeter
	setDefaultStyle();
//	GraphP gr=Thread::This()->graph;
//	mouse=gr->mouse;
	mouse=new Mouse;
	mouseEvents=true;
	for (int i=0; i < contextMenuNr; i++)
		contextMenu[i]=0;

	consoleFileName=theApp->consoleFileName;
	if (autosave)
	{	if (wxFileExists(consoleFileName))
		{	LoadFile(consoleFileName);
			SetStyle(0, GetLastPosition(), GetDefaultStyle());
			SetInsertionPoint(caretpos);
		}
	}
	theApp->console=this;
	console=this;
}

Console::~Console() 
{
	if (contextMenu[0])
		delete contextMenu[0];

	if (autosave)
	{	SaveFile(consoleFileName);
	}
	if (mouse)
		delete mouse;
}

void Console::OnSetFocus(wxFocusEvent& event)
{
	static bool inOnSetFocus=false;
	if (inOnSetFocus)
	{	event.Skip();
		return;
	}
	inOnSetFocus=true;
	if (theApp->consoleFrame)
		theApp->consoleFrame->SetFocus();
	SetFocus();
	event.Skip();
	inOnSetFocus=false;
}

void Console::OnEnterWindow(wxMouseEvent &event) 
{
//	GetParent()->SetFocus();
//	SetFocus();
	event.Skip();
}

void Console::OnLeaveWindow(wxMouseEvent &event) 
{
/*	if (this==theApp->graph && theApp->consoleFrame)
	{	theApp->consoleFrame->SetFocus();
		if (console)
			console->SetFocus();
	}else
*/
	event.Skip();
}

void Console::OnProcessTerm(wxProcessEvent& WXUNUSED(event))
{
	delete process;
	process=0;
	c_out << _("Shell ended. Now you're back to Logo.") << endL;
}

void Console::OnChar(wxKeyEvent &event) 
{
	ThreadP th=Thread::This();
	int c=event.GetKeyCode();
	EventP evt=new EventChar(event);
	if throwing return;
	*th->eventFiFo << evt;
	if (charMode) 
	{	if (theApp->action!=NOP)
			return;
		switch (c)
		{
		case WXK_CANCEL:
			WriteText('\n');
			theApp->action=CONTROL_PAUSE;
			return;
		
		case WXK_PAUSE:
			WriteText('\n');
			if (event.ControlDown())
				theApp->action=CONTROL_PAUSE;
			else
				theApp->action=PAUSE;
			return;
		
		case WXK_F9:		
			theApp->action=ON_RUN;	
			return;
		}
		if (c > 255)
		{	ch=char(255);
			chExt=c;
		}else
			ch=char(c);
		charAvailable=true;
		// we must handle Events ourselfes, 
		// because no foreign code is allowed on the stack for the persistence
		return;
	}
	switch(c)
	{
	case '\t':			WriteText('\t');	break;	// tab
	case WXK_CANCEL:	WriteText('\n');	
							theApp->action=CONTROL_PAUSE;	break;
	case WXK_PAUSE:	WriteText('\n');
							if (event.ControlDown())
								theApp->action=CONTROL_PAUSE;
							else
								theApp->action=PAUSE;
							break;
//	case WXK_F7:		if (event.ShiftDown()) 
//								theApp->action=ON_REVERSE_STEP;
//							else
//								theApp->action=ON_STEP;	
//							break;
//	case WXK_F8:		if (event.ShiftDown()) 
//								theApp->action=ON_REVERSE_STEP_OVER;
//							else
//								theApp->action=ON_STEP_OVER;	
//							break;
	case WXK_F9:		theApp->action=ON_RUN;	break;
	default:				
		if (event.ControlDown())
			switch (c)
			{
			case WXK_UP:
			case WXK_DOWN: 
				{	
					long selstart=GetInsertionPoint(), x,y;
					PositionToXY(selstart, &x, &y);
					wxString line=GetLineText(y);
#ifdef WIN32
					POINT pos;
					::GetCaretPos(&pos);
					theApp->history->show(pos.y+GetPosition().y, line);
#else
// TODO: test it!
//					wxPoint pos=GetCaret()->GetPosition();
/*					int cx=0, cy=0;
					gtk_text_view_buffer_to_window_coords (
						(GtkTextView*)m_text, GTK_TEXT_WINDOW_WIDGET,
						x, y, &cx, &cy);
//					ClientToScreen(&cx,&cy);
					wxPoint pos(cx,cy);
					theApp->history->show(pos.y, line);
*/
					wxTextCoord col=0, row=0;
					HitTest(wxPoint(0,0), &col, &row);
					y-=row;
					int yy=GetFont().GetPointSize()*y;
					theApp->history->show(yy+GetPosition().y, line);
#endif
					break;
				}
			default:	
				event.Skip();
			}
		else
			event.Skip();
	}
}

void Console::setKeyModifiers(wxKeyEvent &event)
{
	if (event.ShiftDown())
		mouse->buttons|=8;
	else
		mouse->buttons&=~8;

	if (event.ControlDown())
		mouse->buttons|=16;
	else
		mouse->buttons&=~16;

	if (event.AltDown())
		mouse->buttons|=32;
	else
		mouse->buttons&=~32;

	if (event.MetaDown())
		mouse->buttons|=64;
	else
		mouse->buttons&=~64;
}

void Console::OnKeyDown(wxKeyEvent &event) 
{
	ThreadP th=Thread::This();
	int c=event.GetKeyCode();
	EventP evt=new EventKeyDown(event);
	if throwing return;
	*th->eventFiFo << evt;
	if (charMode) 
	{	setKeyModifiers(event);
		if (c==WXK_PAUSE || c==WXK_CANCEL 
		|| c==WXK_CONTROL || c==WXK_SHIFT || c==WXK_ALT)
		{	event.Skip();
			return;
		}		
		if ((event.ControlDown() 
		|| event.AltDown() 
		|| event.MetaDown()
		||	c > 255)
		&& !event.ShiftDown())
		{	ch=char(255);
			chExt=c;
			charAvailable=true;
			event.Skip();
			return;
		}
		if (c==WXK_RETURN 
		||  c==WXK_BACK)
		{	ch=c;
			charAvailable=true;
			return;
		}
		event.Skip();
		return;
	}
	switch (c)
	{
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			if (event.ControlDown() && !event.AltDown())
			{	wxCommandEvent cev;
				cev.SetId(ID_LAYOUT_LOAD1+c-'1');
				theApp->mainFrame->OnLayoutLoad(cev);
				return;
			}
		default:
			;
	}
	if (c==WXK_TAB && event.ControlDown())
	{	theApp->editorShow();
		return;
	}
	if (c==WXK_F7)
	{	if (event.ShiftDown()) 
			theApp->action=ON_REVERSE_STEP;
		else
			theApp->action=ON_STEP;	
		return;
	}else if (c==WXK_F8)
	{	if (event.ShiftDown()) 
			theApp->action=ON_REVERSE_STEP_OVER;
		else
			theApp->action=ON_STEP_OVER;	
		return;
	}
	else if ((c!=WXK_RETURN 
	  && c!=WXK_NUMPAD_ENTER)
	|| event.ControlDown() || event.AltDown()
#ifdef WIN32
 || event.MetaDown()
#endif
	)
	{	if (event.ControlDown() && c=='V')
			Paste();
		else
			event.Skip();
	}else
	{	int index=GetInsertionPoint(); // end of Line
		long x,y;
		PositionToXY(index, &x, &y);
		wxString line=GetLineText(y);
	#ifdef WIN32
		if (line.Length() > 0)
			line[line.Length()-1]='\n';
	#else
		line+='\n';
	#endif
		index=XYToPosition(0,y+1);
	#ifdef WIN32
		if (index > GetLastPosition() || index < 0)
	#else
		if (index > GetLastPosition() ||  index < 0
		|| (index==GetLastPosition() && GetRange(index-1, index)!='\n'))
	#endif
		{	SetInsertionPoint(GetLastPosition());
			WriteText('\n');
			SetInsertionPoint(GetLastPosition());
		}else
			SetInsertionPoint(index);
		finishLine(line);
	}
}

void Console::OnKeyUp(wxKeyEvent &event) 
{
	ThreadP th=Thread::This();
	int c=event.GetKeyCode();
	if (c==WXK_F7 || c==WXK_F8)
	{	th->stepNr=0;
		th->stepArmed=false;
	}
	setKeyModifiers(event);
	switch (theApp->action)
	{
	case EXIT_LOGO:
	case EXIT_DELAYED:
	case RECREATE_GUI:
	case HIDE_MAIN:
		break;
	default:
		{	EventP evt=new EventKeyUp(event);
			if throwing return;
			*th->eventFiFo << evt;
		}
	}
}

void Console::setMouse(wxMouseEvent &event, MouseP mouse) 
{
	int x=event.GetX(), y=event.GetY();

	if (event.LeftUp())
		mouse->buttons&=~1;
	if (event.RightUp())
		mouse->buttons&=~2;
	if (event.MiddleUp())
		mouse->buttons&=~4;

	if (event.LeftDown())
		mouse->buttons|=1;
	if (event.RightDown())
		mouse->buttons|=2;
	if (event.MiddleDown())
		mouse->buttons|=4;

	if (event.ShiftDown())
		mouse->buttons|=8;
	else
		mouse->buttons&=~8;

	if (event.ControlDown())
		mouse->buttons|=16;
	else
		mouse->buttons&=~16;

	if (event.AltDown())
		mouse->buttons|=32;
	else
		mouse->buttons&=~32;

	if (event.MetaDown())
		mouse->buttons|=64;
	else
		mouse->buttons&=~64;

	HitTest(wxPoint(x,y), (long*)&mouse->textx, (long*)&mouse->texty);
	if (mouseEvents)
		event.Skip();
}

void Console::OnMouse(wxMouseEvent &event) 
{
	setMouse(event, mouse);
}

#define EVENT_ON_MOUSE(NAME)									\
void Console::OnMouse##NAME(wxMouseEvent &event)		\
{																		\
	ThreadP th=Thread::This();									\
	setMouse(event, mouse);										\
	EventP evt=new EventTextMouse##NAME(mouse);			\
	if throwing return;											\
	*th->eventFiFo << evt;										\
}

EVENT_ON_MOUSE(LeftDown)
EVENT_ON_MOUSE(RightDown)
EVENT_ON_MOUSE(MiddleDown)

EVENT_ON_MOUSE(LeftUp)
EVENT_ON_MOUSE(RightUp)
EVENT_ON_MOUSE(MiddleUp)

EVENT_ON_MOUSE(LeftDClick)
EVENT_ON_MOUSE(RightDClick)
EVENT_ON_MOUSE(MiddleDClick)

EVENT_ON_MOUSE(Motion)

void Console::updateGraph()
{
	ThreadP th=Thread::This();
	if (th->graph)
		th->graph->updateGraph();
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
bool Console::kbhit()
{
	ThreadP th=Thread::This();
	bool oldCharMode=charMode;
	charMode=true;
	if (th->IsMain())
	{	while (theApp->Pending())
		{	th->dispatch();
			ACTION;
			
			if (charAvailable)
				break;
			switch (theApp->action)
			{
			case EXIT_LOGO:
			case EXIT_DELAYED:
			case RECREATE_GUI:
			case HIDE_MAIN:
				return false;
			default:
				;
			}
		}
		th->processIdle();
	}
	charMode=oldCharMode;
	return charAvailable;
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

char Console::getch()
{
	ThreadP th=Thread::This();
	updateGraph();

	if (!charAvailable)
	{	idle=true;
		while (true)
		{	if (theApp->Pending())
				th->dispatch();
			th->processIdle();
			
			if (charAvailable)
				break;
			switch (theApp->action)
			{
			case EXIT_LOGO:
			case EXIT_DELAYED:
			case RECREATE_GUI:
			case HIDE_MAIN:
				break;
			default:
				;
			}
		}
		idle=false;
	}
	charAvailable=false;
	return ch;
}

int Console::getchExt()
{
	charAvailable=false;
	return chExt;
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
char Console::getc_()
{
	ThreadP th=Thread::This();
	if (!lineAvailable) 
	{	updateGraph();
		idle=true;
		while (true)
		{
				th->dispatch();
				th->processIdle();
				ACTION;

				switch (theApp->action)
				{
				case EXIT_LOGO:
				case EXIT_DELAYED:
				case RECREATE_GUI:
				case HIDE_MAIN:
										idle=false;
										return ' ';
				case ON_RUN:	{	theApp->action=NOP;
										if (theApp->edFrame)
										{	wxCommandEvent cev;
											theApp->edFrame->OnRun(cev); 
										}
										break;
									}
				case ON_STEP:	{	theApp->action=NOP;
										if (theApp->edFrame)
										{	wxCommandEvent cev;
											theApp->edFrame->OnStep(cev); 
										}
										break;
									}
				case ON_REVERSE_STEP:	
									{	theApp->action=NOP;
										if (theApp->edFrame)
										{	wxCommandEvent cev;
											theApp->edFrame->OnReverseStep(cev); 
										}
										break;
									}
				case ON_STEP_OVER:
									{	theApp->action=NOP;
										if (theApp->edFrame)
										{	wxCommandEvent cev;
											theApp->edFrame->OnStepOver(cev); 
										}
										break;
									}
				case ON_REVERSE_STEP_OVER:
									{	theApp->action=NOP;
										if (theApp->edFrame)
										{	wxCommandEvent cev;
											theApp->edFrame->OnReverseStepOver(cev); 
										}
										break;
									}
				case ON_LOAD:	{	theApp->action=NOP;
										if (theApp->edFrame)
										{	wxCommandEvent cev;
											theApp->edFrame->OnLoad(cev); 
										}
										break;
									}
				case PAUSE:		{	char c='p'-'a'+1; // control-p
										theApp->action=NOP;
										idle=false;
										return c;
									}
				case CONTROL_PAUSE:
									{	char c='q'-'a'+1; // control-q
										theApp->action=NOP;
										if (th->graph && th->graph->recorder != th->graph->recorderSave)
										{	delete th->graph->recorder;
											th->graph->recorder=th->graph->recorderSave;
										}
										if (th->graph)
										{	if (th->graph->turtle->polyMode)
												 th->graph->turtle->polyEnd();
											if (th->graph->turtle->tessMode)
												 th->graph->turtle->tessEnd();
											if (th->graph->turtle->surfaceMode)
												 th->graph->turtle->surfaceEnd();
										}
										EventChar   ::function=0;
										EventKeyDown::function=0;
										EventKeyUp  ::function=0;

										EventTextMouseLeftDown  ::function=0;
										EventTextMouseLeftUp    ::function=0;
										EventTextMouseLeftDClick::function=0;

										EventTextMouseRightDown  ::function=0;
										EventTextMouseRightUp    ::function=0;
										EventTextMouseRightDClick::function=0;

										EventTextMouseMiddleDown  ::function=0;
										EventTextMouseMiddleUp    ::function=0;
										EventTextMouseMiddleDClick::function=0;

										EventTextMouseMotion::function=0;

										EventMouseLeftDown  ::function=0;
										EventMouseLeftUp    ::function=0;
										EventMouseLeftDClick::function=0;

										EventMouseRightDown  ::function=0;
										EventMouseRightUp    ::function=0;
										EventMouseRightDClick::function=0;

										EventMouseMiddleDown  ::function=0;
										EventMouseMiddleUp    ::function=0;
										EventMouseMiddleDClick::function=0;

										EventMouseMotion::function=0;

//										charMode=false;
										idle=false;
										return c;
									}
				default: 
					;
				}
//			}
			if (lineAvailable) break;
			if throwing 
			{	idle=false;
				return 0;
			}
			if (!th->IsMain() && th->TestDestroy())
			{	th->throw_node=_Stop;
				th->stoppingFlag=THROWING;
			}
			if (GC_::is_running)
			{	th->nrPaused++;
				while (theApp->Pending())
					th->dispatch();
				while (GC_::is_running)
				{	th->yield();
				}
				th->nrPaused--;
			}
		}
	}
	idle=false;
	char c;
	if (readIndex <= readLine.length())
	{	c=readLine[readIndex];
		readIndex++;
	}
	if (c=='q'-'a'+1 || c=='p'-'a'+1 || c=='\n') // control-q, control-p
	{	lineAvailable=false;
		readIndex=0;
//		readLine.resize(0);
		charAvailable=false;
		if (readLine=="reset\n")
			readLine=oldReadLine;
	}
	return c;
}

#ifdef WIN32
//	#pragma optimize("", on)
#endif

char Console::putc_(char c, FILEP strm)
{
	if (strm==File::conout) 
	{	switch(c)
		{
		case '\007':theApp->Tone(0, 100); break;

		case '\n':
#ifdef WIN32
						putc_('\r', strm);	// and do the rest...
#endif
		default:		writeBufferCS.Enter();
						if (!c)
							writeBuffer+='0';
						else
							writeBuffer+=char(c);
						writeBufferCS.Leave();
						if (//char(c)=='\n' || 
						writeBuffer.length() > writeBufferSize)
							flush();
		}
		return 0;
	}
	if (strm==File::textcontrol)
	{	if (Thread::This()->textControl 
		&& (Thread::This()->textControl->textControl!=0))
		{	TextControl *tc=Thread::This()->textControl->textControl;
			switch(c)
			{
			case '\007':theApp->Tone(0, 100); break;
			case '\n':
#ifdef WIN32
							putc_('\r', strm);	// and do the rest...
#endif
			default:		tc->writeBuffer+=char(c);
							if (char(c)=='\n' 
								|| tc->writeBuffer.length() > tc->writeBufferSize)
								tc->flush();
			}
		}
		else
		{	Thread::This()->Writer=File::ConOut;
			putc_(ch, File::conout);
		}
		return 0;
	}
	if (c==10)
		return putc(c, strm);
	return putc(c, strm);
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
void Console::flush()
{
	static volatile bool inFlush=false;
	if (inFlush)
		return;
	inFlush=true;

	ThreadP th=Thread::This();
/*static NodeP t0=0;
t0=lTimeFine(0);
*/
	if (!writeBuffer.length())
	{	inFlush=false;
		return;
	}
//	th->GuiEnter();
	long from, to;
	GetSelection(&from, &to);
//	th->GuiLeave();
//t1=lTimeFine(0);

	writeBufferCS.Enter();
#ifdef WIN32
	if (!insertMode)
		to=from+writeBuffer.length();
//	Replace(from, to, writeBuffer);
	wxString buf=writeBuffer;
	writeBuffer="";
	writeBufferCS.Leave();

//	th->GuiEnter();
	::SendMessage(HWND(GetHandle()), EM_REPLACESEL, 0, (LPARAM)buf.c_str());
//	th->GuiLeave();
#else
	wxString buf=writeBuffer;
	writeBuffer="";
	writeBufferCS.Leave();
	long ip=GetInsertionPoint();
	if (from < 0)
		from=ip;
	if (to < 0)
		to=ip;
	if (!insertMode)
	{	to=from+buf.length();
		if (to > GetLastPosition())
			to=GetLastPosition();
//		Remove(from, to);
	}
	Replace(from, to, buf);
//	WriteText(writeBuffer);
//	long ipnew=ip+writeBuffer.length();
//	SetInsertionPoint(ipnew);
#endif

//t2=lTimeFine(0);

	if (th->GuiTryEnter())
	{
//t3=lTimeFine(0);
//		if (th->graph)
//			th->graph->tryEnter();
//t4=lTimeFine(0);
//		while (theApp->Pending())
//			theApp->dispatch();
		th->checkBreak();
//t5=lTimeFine(0);
//		theApp->processIdle();
//t6=lTimeFine(0);
		ACTION;
		th->GuiLeave();
	}
	inFlush=false;
/*t7=lTimeFine(0);
c_out << t1->sub(t0) << "seconds1" << endL;
*/
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

void Console::setTextSelection(int x1, int y1, int x2, int y2)
{
	int pos1=XYToPosition(x1, y1),

		pos2=XYToPosition(x2, y2);
	SetSelection(pos1, pos2);
}

void Console::update_coords(char ch) 
{
	int width=Int::max;
	switch(ch)
	{
	case '\n':
		{	coord.x=0;		// newline
			coord.y++;
			for (int i=0; i < Thread::This()->marginX; i++) 
				print(File::conout, width, ' ');
			break;
		}		
	case '\b': if (coord.x > 0) --coord.x;		// backspace
			break;

	case '\t':	coord.x &=~07; coord.x+=8; break;	// tab
	case '\001':                    // bold mode

	case '\002': break;      // plain mode
	case '\007':               // ...???
		coord.x++;

		if (coord.x > Max.x) 
		{

			if (Max.x != 0) // well, it could be...
			{
				coord.y += coord.x/Max.x;
				coord.x %= Max.x;
			}
		}
		if (coord.y > Max.y) 
			coord.y = Max.y;
	}
}

void Console::finishLine(wxString line)
{
	readLine=line;
	theApp->history->append(line);
	lineAvailable=true;
	readIndex=0;
}

void Console::OnEditRepeat(wxCommandEvent &event)
{
	lineAvailable=true;
	readIndex=0;
}

void Console::OnEditReset(wxCommandEvent &event)
{
	oldReadLine=readLine;
	readLine="reset\n";
	lineAvailable=true;
	readIndex=0;
}

NodeP Console::_Cursor() 
{
	long pos=GetInsertionPoint();
	long x,y;
	PositionToXY(pos, &x, &y);
	return new Pair
	(	new Int(x),
		new Pair(new Int(y),0) );
}

NodeP Console::_setCursor(PairP args) 
{
	PairP vec = args->a_pos_int_vector(1);
	if (!throwing) 
	{	flush();
		wxPoint oldCoord=coord;
		coord.x=IntP(vec-> first())->int_;
		coord.y=IntP(vec->second())->int_;

		if ((coord.x < 0 || coord.y < 0) && !throwing) 
		{	error(BAD_DATA, vec);
			coord=oldCoord;
			return Unbound;
		}
		if (coord.y >= GetNumberOfLines())
		{	int end=GetLastPosition();
			SetInsertionPoint(end);
			long x,y;
			PositionToXY(end, &x, &y);
			int missingLines=coord.y-y;
			wxString CRLFs;
			while (missingLines)
			{
#ifdef WIN32
				CRLFs+="\r\n";
#else
				CRLFs+="\n";
#endif
				missingLines--;
			}
			*this << CRLFs;
		}
		if (coord.x > GetLineLength(coord.y))
		{	SetInsertionPoint(
				XYToPosition( GetLineLength(coord.y), coord.y ));
			WriteText(wxString(
				coord.x+1-GetLineLength(coord.y), ' ').c_str());
		}
		SetInsertionPoint(XYToPosition(coord.x, coord.y));
	}
	return Unbound;
}

NodeP Console::_setTextColor(PairPP arg) 
{
	textColor = arg[1]->a_color(1);

	if (!throwing) 
	{
		backgroundColor = arg[2]->a_color(2);

		if (!throwing) 
		{	flush();
			SetDefaultStyle(wxTextAttr(
				int(textColor), 
				int(backgroundColor)));
		}
	}
	return Unbound;
}

void Console::charModeOn()
{
	if (charMode) 
		return; 
	flush();
	charMode=true; 
}
void Console::charModeOff() 
{
	if (!EventChar::function)
		charMode=false;
}

void Console::setDefaultStyle()
{
#ifdef WIN32
	wxTextAttr textattr(
		int(textColor), int(backgroundColor),
		wxFont(fontSize, fontFamily, fontStyle, fontWeight, underline,
			fontName));
	wxArrayInt tabs;
	tabs.Add(0,10);
	for (int i=0; i<10; i++)
		tabs[i]=tabSize*i;
	textattr.SetTabs(tabs);

	SetDefaultStyle(textattr);
#else
	SetFont(	wxFont(fontSize, fontFamily, fontStyle,
					fontWeight, underline, fontName));
#endif
}

void Console::modeBold     (){fontWeight=wxBOLD;  setDefaultStyle();}
void Console::modeItalic   (){fontStyle=wxITALIC; setDefaultStyle();}
void Console::modeUnderline(){underline=true;     setDefaultStyle();}
void Console::modePlain()
{	fontStyle=wxNORMAL; fontWeight=wxNORMAL; underline=false;
	setDefaultStyle();
}

NodeP Console::_setTextSize(PairP arg)
{
	double s=arg->a_float(1);
	if throwing return Unbound;
	fontSize=int(s);
	setDefaultStyle();
	return Unbound;
}

NodeP Console::_setTextFont(PairP arg)
{
	StringP name=arg->a_string(1);
	if (!throwing)
	{	if (name->length() > 32)
		{	error(BAD_DATA,arg);
			return Unbound;
		}

		flush();
		name=String::make(name->ptr(), 0, 
			name->length(), STRING, String::copy);
		fontName=name->ptr();
		
		setDefaultStyle();
	}
	return Unbound;
}

void Console::toggleInsertMode()
{
	ThreadP th=Thread::This();
	flush();
	insertMode=!insertMode;
#ifdef _WIN32
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY , 0 );
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
#else
	EmulateKeyPress(wxKeyEvent(WXK_INSERT));
//	error(NOT_YET_SUPPORTED, new String("Console::toggleInsertMode"));
#endif
	th->dispatchMessages();
}

NodeP Console::_insertMode()
{
	if (!insertMode)
		toggleInsertMode();
	return Unbound;
}

NodeP Console::_overwriteMode()
{
	if (insertMode)
		toggleInsertMode();
	return Unbound;
}

NodeP Console::_clearText(PairP)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI")));
		return Unbound;
	}
	Clear();
	return Unbound;
}

NodeP Console::_setMargins(PairP args) 
{
	PairP arg;
	arg = args->a_pos_int_vector(1);
	if (!throwing) 
	{	Thread::This()->marginX = IntP(arg->   first())->int_;
		Thread::This()->marginY = IntP(arg->second())->int_;
		_clearText(NIL);
	}
	return Unbound;
}

NodeP Console::_standout(NodeP arg) 
{
	char textbuf[65536];
	Thread::This()->stringptr=textbuf;
	Thread::This()->stringlen=65535;

	stringbuffer << '\001' << arg << '\002';
	*(Thread::This()->stringptr)='\0';
	return String::make(textbuf, 0, strlen(textbuf), STRING, String::copy);
}

void Console::OnSaveDocument(wxString pathname)
{
	ThreadP th=Thread::This();
	NodeP pn=new String(pathname);
	if throwing return;
	NodePP asp=th->argStack->index;
	th->argStack->push(pn);
	lsave(PairPP(asp));
	th->argStack->pop_till(PairPP(asp));
	return;
}

char Console::charUnderCursor()
{
	int ip=GetInsertionPoint();
	if (ip==GetLastPosition())
		return '\0';
	wxString tc=GetRange(ip, ip+1);
	return tc[0];
}

wxString Console::wordUnderCursor(bool *paren) 
{
	long selstart,selend;
	GetSelection(&selstart, &selend);
	wxString line;

	if (selstart==selend)
	{	long x,y;
		if (selstart < 0)
			selstart=GetInsertionPoint();
		PositionToXY(selstart, &x, &y);
		line=GetLineText(y);
		long lineindex=XYToPosition(0,y);
		if (lineindex<0)
			return wxEmptyString;
		int pos=selstart-lineindex,
			start=pos-1,
			len=line.length();

		if (start<0)
			start=0;
		else
		{	while (start>=0 && start<len && isalnum(line[(unsigned int)start])) 
			{	start--;
			}
			start++;
		}
		int end=pos;
		while (isalnum(line[(unsigned int)end]) && end<len) 
		{	end++;
		}
		line=line.Mid(start, end-start);
		if (paren && line[end]==')')
			*paren=true;
	}else
	{	line=GetStringSelection();
	}
	return line;
}

void Console::OnHelp(wxHelpEvent &event) 
{
	wxString line=wordUnderCursor();
	if (!theApp->help)
	{	theApp->help=new wxHelpController();
		if (!theApp->help)
		{	error(OUT_OF_MEM); return;
		}
		theApp->help->Initialize(theApp->helpFileName);
	}
#ifdef WIN32
	theApp->help->KeywordSearch(line);
#else
// TODO: test it!
//	theApp->help->KeywordSearch(line);
	theApp->help->DisplaySection(line);
#endif
}

void Console::OnGotoLine(wxCommandEvent &event)
{
	bool paren=false;
	wxString word=wordUnderCursor(&paren);
	unsigned long nr;
	if (word.ToULong(&nr))
		if (paren)
			debugStates.switchTo(nr);
		else
			theApp->edFrame->gotoLine(nr);
}

void Console::OnOpenFileAtCursor(wxCommandEvent &event)
{
	wxString word=wordUnderCursor()+".lg";
	if (wxFileExists(word))
	{	theApp->editorShow();
		if (theApp->edFrame)
			theApp->edFrame->FileOpen(word);
	}else
	{	wxFileName fn(environment.logolib, word);
		if (fn.FileExists())
		{	theApp->editorShow();
			if (theApp->edFrame)
				theApp->edFrame->FileOpen(word);
		}
	}
}

NodeP lCursor(PairPP)     
{	if (console)
		return console->_Cursor();
	else if (consolew)
		return consolew->_Cursor();
	return 0;
}

NodeP lsetCursor(PairPP arg) 
{	if (console)
		return console->_setCursor(arg[1]);
	else if (consolew)
		return consolew->_setCursor(arg[1]);
	return Unbound;
}
NodeP lsetTextColor(PairPP arg) 
{	if (console)
		return console->_setTextColor(arg);
	else if (consolew)
		return consolew->_setTextColor(arg);
	return Unbound;
}
NodeP lsetTextSize(PairPP arg) 
{	if (console)
		return console->_setTextSize(arg[1]);
	return Unbound;
}

NodeP lsetTextFont(PairPP arg) 
{	if (console)
		return console->_setTextFont(arg[1]);
	return Unbound;
}
NodeP linsertMode(PairPP)     
{	if (console)
		return console->_insertMode();
	return Unbound;
}
NodeP loverwriteMode(PairPP)     
{	if (console)
		return console->_overwriteMode();
	return Unbound;
}
NodeP lboldTextMode(PairPP)     
{	if (console)
		console->modeBold(); 
	return Unbound;
}
NodeP litalicTextMode(PairPP)     
{	if (console)
		console->modeItalic(); 
	return Unbound;
}
NodeP lunderlineTextMode(PairPP)    
{	if (console)
		console->modeUnderline(); 
	return Unbound;
}
NodeP lplainTextMode(PairPP)     
{	if (console)
		console->modePlain(); 
	return Unbound;
}
NodeP lclearText(PairPP)     
{	if (console)
		return console->_clearText(NIL);
	return Unbound;
}
NodeP lsetMargins(PairPP arg) 
{	if (console)
		return console->_setMargins(arg[1]);
	return Unbound;
}
NodeP lstandout(PairPP arg) 
{	if (console)
		return console->_standout(arg[1]);
	return Unbound;
}

NodeP ldispatchMessages(PairPP arg) 
{
	ThreadP th=Thread::This();
//	console->updateGraph();	
	th->dispatchMessages(); 
	return Unbound;
}

NodeP lCharUnderCursor(PairPP) 
{
	char c[2];
	if (console)
	{	c[0]=console->charUnderCursor();
		c[1]=0;
		return new String(c);
	}
	return new String(_T("?"));
}

NodeP lWordUnderCursor(PairPP) 
{
	if (console)
	{	wxString word=console->wordUnderCursor();
		return new String(word);
	}
	return new String(_T("?"));
}

NodeP lsetTextSelection(PairPP arg)
{
	PairP start=arg[1]->a_pos_int_vector(1),
		end=arg[2]->a_pos_int_vector(2);
	if throwing return Unbound;

	int startx=int(start->first()->getnumber()),
		starty=int(start->second()->getnumber()),
		endx=int(end->first()->getnumber()),
		endy=int(end->second()->getnumber());

	if (console)
		console->setTextSelection(startx, starty, endx, endy);
	return Unbound;
}

NodeP lenableTextMouseEvents(PairPP)
{
	if (console)
		console->mouseEvents=true;
	return Unbound;
}

NodeP ldisableTextMouseEvents(PairPP)
{
	if (console)
		console->mouseEvents=false;
	return Unbound;
}

Console::MenuTree Console::command[]=
{
#include "toc.h"
	{0,0}
};

void Console::createMenu(int &used, int &i, int level)
{
	int parent=used;
	contextMenu[parent] = new wxMenu;
	if (!contextMenu[parent])
	{	error(OUT_OF_MEM); return;
	}
	while (true)
	{	int newlevel=command[i].level;

		if (newlevel >= level)
		{	
			int nextlevel=command[i+1].level;
			if (nextlevel == newlevel)
			{	contextMenu[parent]->Append(ID_CONSOLE_INSERT+i, command[i].str);
				i++;
			}
			else if (nextlevel > newlevel)
			{	used++;
				int item=i, child=used;
				i++;
				createMenu(used, i, newlevel);
				contextMenu[parent]->Append(ID_CONSOLE_INSERT+item, 
					command[item].str, contextMenu[child]);
				nextlevel=command[i].level;
				if (nextlevel < newlevel)
					break;
			}
			else
			{	contextMenu[parent]->Append(ID_CONSOLE_INSERT+i, command[i].str);
				i++;
				break;
			}
		}
		else
		{	i++;
			break;
		}
	};
}

void Console::OnContextMenu(wxContextMenuEvent &event)
{
	if (!contextMenu[0])
	{
		int used=0, i=0, level=1; 
		createMenu(used, i, level);

		wxMenu &m = *contextMenu[0];
		m.AppendSeparator();
		m.Append(wxID_UNDO, _("&Undo"));
		m.Append(wxID_REDO, _("&Redo"));
		m.AppendSeparator();
		m.Append(wxID_CUT, _("Cu&t"));
		m.Append(wxID_COPY, _("&Copy"));
		m.Append(wxID_PASTE, _("&Paste"));
		m.Append(wxID_CLEAR, _("&Delete"));
		m.AppendSeparator();
		m.Append(wxID_SELECTALL, _("Select &All"));
		m.AppendSeparator();
		m.Append(ID_GOTO_LINE, _("&Goto Line"));
		m.Append(ID_OPEN_FILE_AT_CURSOR, _("&Open File at Cursor"));
	}
	PopupMenu(contextMenu[0]);

	return;
}

void Console::OnInsert(wxCommandEvent &event)
{
	WriteText(command[event.GetId()-ID_CONSOLE_INSERT].str);
	WriteText(' ');
}

void Console::OnIdle(wxIdleEvent &event)
{
	static volatile bool inOnIdle=false;
	if (inOnIdle)
		return;
	inOnIdle=true;
	if (writeBuffer.length() > 0)
		flush();
	inOnIdle=false;
}

void Console::Paste()
{
	wxClipboard clipboard;
	if (clipboard.Open())
	{
		if (clipboard.IsSupported(wxDF_TEXT))
		{
			wxTextDataObject data;
			clipboard.GetData(data);
			WriteText(data.GetText());
		}  
		clipboard.Close();
	}
}

BitmapP Console::bitmap()
{
	int w,h;
	GetClientSize(&w, &h);
	wxClientDC dc(this);
#ifdef WIN32
	wxBitmap bmp(w, h, dc);
	wxMemoryDC mdc(bmp);
#else
	wxMemoryDC mdc;
	wxBitmap bmp(w, h);
	mdc.SelectObject(bmp);
#endif
	mdc.Blit(0,0, w,h, &dc, 0,0);
	BitmapP bitmap=new Bitmap(bmp);
	return bitmap;
}

NodeP lConsoleSetFocus(PairPP)
{
	if (console)
		console->SetFocus();
	return Unbound;
}

NodeP lisKey(PairPP) 
{
	ThreadP th=Thread::This();
	if (!th->security->keyboard)
	{	error(SECURITY_ERROR, new String(_("to use keyboard"))); return Unbound;
	}
	if (reader.file()==File::conin && Thread::This()->interactive) 
	{	th->dispatchMessages();
		if (console)
			console->flush();
		NodeP result= (File::kbhit() != 0) ? _True : _False;
		return result;
	}else
	{	int c=reader.getc_();
		if (reader.eof())
			return _False;
		else 
		{	reader.ungetc(c);
			return _True;
		}
	}
}

char tmp_filename[500]="";

NodeP ledit(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	if (environment.editor != wxEmptyString)
	{	if (tmp_filename[0]=='\0'  ||  !noarg) 
			sprintf(tmp_filename, "%s/logo%d.lg",
				environment.tempdir.c_str(), 
#ifdef WIN32
				_getpid());
#else
				getpid());
#endif
		if (!noarg) 
		{	FILEP old_file = writer.file();
			writer.setfile( fopen(tmp_filename, "w") );
			if (writer) 
			{	arg[1]->printOut(-1);
				writer.close();
				writer.setfile(old_file);
			}else 
			{	error(FILE_ERROR, new String(_("Could not create editor file")));
				writer.setfile(old_file);
				return Unbound;
			}
		}
		if throwing return Unbound;

	#ifdef WIN32
		if ( _spawnlp(P_WAIT, 
				environment.editor, 
				environment.editorname, tmp_filename, 0) )
	#else
		wxString command=wxString(environment.editor)+' '+wxString(tmp_filename);
	//	wxEnableTopLevelWindows(false);
	//	if (wxExecute(command, wxEXEC_SYNC))
		system(command.c_str());
		if (false)
	#endif
		{	error(FILE_ERROR, new String(_("Could not launch the editor")));
			return Unbound;
		}
	#ifndef WIN32
	//	wxEnableTopLevelWindows(true);
	#endif
		StringP old_line = th->currentLine;
		FILEP   old_file = loader.file();

		loader.setfile( fopen(tmp_filename, "r") );
		if (loader) 
		{	ValStatus save_val_status = th->val_status;

			while (!loader.eof() && !throwing )
			{	th->currentLine = loader.read("");
				PairP exec_list = th->currentLine->parser(true);
				th->val_status = no_value_allowed;
				if (exec_list) 
					exec_list->eval_driver();
			}
			loader.close();
			th->val_status = save_val_status;
		}else
			error(FILE_ERROR, new String(_("Could not read editor file")));

		loader.setfile(old_file);
		th->currentLine=old_line;
	}
	else
		if (tmp_filename[0]=='\0' || !noarg)
		{
//			c_out << _(";Hint: better save your procedures in a file and edit that in the editor!") << endL;

			sprintf(tmp_filename, "%s/logo%d.lg",
				environment.tempdir.c_str(), 
#ifdef WIN32
				_getpid());
#else
				getpid());
#endif
			if (!noarg) 
			{	FILEP old_file = writer.file();
				writer.setfile( fopen(tmp_filename, "w") );
				if (writer) 
				{	arg[1]->printOut(-1);
					writer.close();
					writer.setfile(old_file);
				}else 
				{	error(FILE_ERROR, new String(_("Could not create editor file")));
					writer.setfile(old_file);
					return Unbound;
				}
			}
			if throwing return Unbound;
			theApp->editorShow();
			if (theApp->edFrame)
				theApp->edFrame->FileOpen(tmp_filename);
/*			c_out << _(";editing... press any key in the console to load changes! ");
//			console->charModeOn();
			console->getch();
//			console->charModeOff();
			c_out << _("OK") << endL;
			theApp->edFrame->FileSave(tmp_filename);
			theApp->edFrame->FileClose(tmp_filename);
			console->SetFocus();

			StringP old_line = th->currentLine;
			FILEP   old_file = loader.file();

			loader.setfile( fopen(tmp_filename, "r") );
			if (loader) 
			{	ValStatus save_val_status = th->val_status;

				while (!loader.eof() && !throwing )
				{	th->currentLine = loader.read("");
					PairP exec_list = th->currentLine->parser(true);
					th->val_status = no_value_allowed;
					if (exec_list) 
						exec_list->eval_driver();
				}
				loader.close();

				th->val_status = save_val_status;
			}else
				error(FILE_ERROR, new String(_("Could not read editor file")));

			loader.setfile(old_file);
			th->currentLine=old_line;
*/		}
		else
		{	theApp->editorShow();
			if (theApp->edFrame)
				theApp->edFrame->FileOpen(tmp_filename);
		}
	return Unbound;
}

NodeP leditFile(PairPP arg) 
{
	ThreadP th=Thread::This();
	StringP str = arg[1]->string();

	if (!throwing) 
	{//	String::copy_decode(tmp_filename, str->ptr(), str->length());
		String::copy(tmp_filename, str->ptr(), str->length());
		return ledit(PairPP(th->argStack->index));
	}else
		return Unbound;
}

NodeP lConsoleBitmap(PairPP arg)
{
	return theApp->console->bitmap();
}

