//___________________________________________________________________
//
//                        Class Console
//___________________________________________________________________

extern Function 
	lsetTextColor, lsetTextSize, lsetTextFont, 
	linsertMode, loverwriteMode, 
	lboldTextMode, litalicTextMode, lunderlineTextMode, lplainTextMode,
	lclearText, lCursor, lsetMargins, lstandout,
	ldispatchMessages, lCharUnderCursor, lWordUnderCursor,
	lsetTextSelection,
	lenableTextMouseEvents, ldisableTextMouseEvents,
	lConsoleSetFocus,
	ledit, leditFile, lConsoleBitmap;

class Mouse;
class History;
class wxProcessEvent;
class wxFlatMenuBar;
class wxFlatMenu;
class wxFlatMenuItem;

class ConsoleFrame: public wxPanel
{
public:
	wxFlatMenuItem
		*fileMenuConsoleAutoSave,
		*fileMenuPictureAutoSave,
		*fileMenuNodesAutoSave,
		*editMenuWrapLines;
	wxFlatMenu *fileMenu, *editMenu, *viewMenu, *helpMenu;
	wxFlatMenuBar *menuBar;

	ConsoleFrame(wxWindow *parent);
	~ConsoleFrame();
	void show(bool shown=true);
};

class Console: public wxTextCtrl
{
	friend Function lStandout;
	public:
		bool charAvailable, lineAvailable, charMode, insertMode, 
			mouseEvents, autosave, idle;
		char ch;
		int chExt;
		wxPoint coord, Max;
		wxString readLine, oldReadLine, consoleFileName;
		Mouse *mouse;
		wxProcess *process;

		Console(wxWindow* parent, wxWindowID id, 
			const wxString& value = "", 
			const wxPoint& pos = wxDefaultPosition, 
			const wxSize& size = wxDefaultSize, 
			long style = 0, 
			const wxValidator& validator = wxDefaultValidator, 
			const wxString& name = wxTextCtrlNameStr,
			bool autosave_ = true,
			int caretpos = 0);
		~Console();

		void OnProcessTerm(wxProcessEvent &event);
		void OnSetFocus(wxFocusEvent &event);
		void OnEnterWindow(wxMouseEvent &event);
		void OnLeaveWindow(wxMouseEvent &event);
		void OnChar		(wxKeyEvent &event);
		void OnKeyDown	(wxKeyEvent &event);
		void OnKeyUp		(wxKeyEvent &event);
		void OnEnter		(wxKeyEvent &event);
		void setMouse		(wxMouseEvent &event, Mouse *mouse);
		void OnMouse		(wxMouseEvent &event);
		void OnMouseLeftDown		(wxMouseEvent &event);
		void OnMouseRightDown		(wxMouseEvent &event);
		void OnMouseMiddleDown		(wxMouseEvent &event);
		void OnMouseLeftUp  		(wxMouseEvent &event);
		void OnMouseRightUp  		(wxMouseEvent &event);
		void OnMouseMiddleUp  		(wxMouseEvent &event);
		void OnMouseLeftDClick		(wxMouseEvent &event);
		void OnMouseRightDClick	(wxMouseEvent &event);
		void OnMouseMiddleDClick	(wxMouseEvent &event);
		void OnMouseMotion			(wxMouseEvent &event);
		void OnHelp		(wxHelpEvent &event);
		void OnEditRepeat(wxCommandEvent &event);
		void OnEditReset	(wxCommandEvent &event);
		void OnGotoLine(wxCommandEvent &event);
		void OnOpenFileAtCursor(wxCommandEvent &event);
		void OnIdle(wxIdleEvent &event);
		void setKeyModifiers(wxKeyEvent &event);
		void updateGraph();
		void update_coords(char ch);
		void finishLine(wxString line); 
		void flush();
		void setTextSelection(int x1, int y1, int x2, int y2);
		void charModeOn(); 
		void charModeOff(); 
		void setDefaultStyle();
		void modeBold(); 
		void modeItalic(); 
		void modeUnderline(); 
		void modeStrikeout(); 
		void modePlain();
		void toggleInsertMode();
	
		bool kbhit();
		char getch(); 
		char getc_(); 
		char putc_(char c, FILEP strm); 
		char charUnderCursor();
		int getchExt();
		wxString linebreak(); 
		wxString wordUnderCursor(bool *paren=0);

		NodeP _Cursor(); 
		NodeP _setCursor(PairP args); 
		NodeP _setMargins(PairP args);
		NodeP _setTextColor(PairPP arg);
		NodeP _setTextSize(PairP args);
		NodeP _setTextFont(PairP args); 
		NodeP _insertMode(); 
		NodeP _overwriteMode();
		NodeP _standout(NodeP arg);
		NodeP _clearText(PairP);
		BitmapP bitmap();

	protected:
		bool underline;
		char standout, plainMode;
		int fontSize, fontFamily, fontStyle, fontWeight;
		unsigned int readIndex, writeBufferSize, tabSize;
		wxString writeBuffer, fontName;
		Color textColor, backgroundColor;

#define contextMenuNr 1000
		wxMenu *contextMenu[contextMenuNr];

		struct MenuTree
		{	int level;
			charP str;
		};

		static MenuTree command[];

		void OnDestroy();
		void OnSaveDocument(wxString pathname);
		void createMenu(int &used, int &i, int level);
		void OnContextMenu(wxContextMenuEvent &event);
		void OnInsert(wxCommandEvent &event);
		void Paste();

		DECLARE_EVENT_TABLE()
};

