
#include "stdafx.h"
#include <math.h>
#include <stdio.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/string.h>
#include <wx/thread.h>
#include <wx/app.h>
#include <wx/frame.h>
#include <wx/notebook.h>
#include <wx/timer.h>
#include <wx/dialog.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>


#include "Main.h"
#include "Node.h"
#include "File.h"
#include "PtrToInt_.h"
#include "IntToPtr_.h"
#include "Object.h"
#include "Table.h"
#include "LogoObject.h"
#include "Line.h"
#include "String.h"
#include "TurtleNode.h"
#include "Stack.h"
#include "NodePStack.h"
#include "EdFrame.h"
#include "Point.h"
#include "Graph.h"
#include "UInt8.h"
#include "Color.h"
#include "Pen.h"
#include "Font.h"
#include "Turtle.h"
#include "Recorder.h"
#include "LogoApp.h"
#include "Init.h"
#include "Thread.h"
#include "DebugState.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

extern TableP root;

//___________________________________________________________________
//
//									class DebugState
//___________________________________________________________________

class DebugState
{
public:
	TableP root2;
	charP filename;
	int col, row;
	LogoObjectP thisobj;
	TurtleNodeP turtleNode;

	VarStack   varStack;	// stack for storing local variables
	NodePStack funStack;	// for the calls window
	IntStack argNrStack; // range into varStack to actual arguments given to fun

	int graphi;

	DebugState();

	void mark();	// Iterative mark procedure
	void markPtr(PtrToInt_ &ptrToInt);// mark and insert ptr into ptrToInt table

	void check_oldyoung(NodeP new_); 
	int gmark(NodeP child), 
		gmarkPtr(NodeP child, PtrToInt_ &ptrToInt);

	void copyRoot();
	void copyToRoot();
};

DebugState::DebugState(): graphi(0)
{
	root2=TableP(Unbound);
	thisobj=0;
	turtleNode=0;
}

void DebugState::mark()
{
	if (root2)
		root2->mark();
	if (thisobj)
		thisobj->mark();
	if (turtleNode)
		turtleNode->mark();
	varStack.mark();
	funStack.mark();
}

void DebugState::markPtr(PtrToInt_ &ptrToInt)
{
	if (root2)
		root2->markPtr(ptrToInt);
	if (thisobj)
		thisobj->markPtr(ptrToInt);
	if (turtleNode)
		turtleNode->markPtr(ptrToInt);
	varStack.markPtr(ptrToInt);
	funStack.markPtr(ptrToInt);
}

void DebugState::copyRoot()
{
	ThreadP th=Thread::This();
//	GC_::run(true);
	PtrToInt_ ptrToInt(GC_::mem_used);
	GC_::mark_phasePtr(ptrToInt);

	IntToPtr_ intToPtr(GC_::mem_used);

	PtrToInt_::PtrIntPairP obj=ptrToInt.start;
	GC_::disabled=true;
	while (obj)
	{	NodeP n=NodeP(obj->p)->copy(ptrToInt);
		intToPtr.insert(n);
		obj=obj->nxt;
	}
	IntToPtr_::PtrIntPairP unbound_pair=
		intToPtr.findPair(ptrToInt.find(Unbound));
	unbound_pair->p=Unbound;
	int i;
	for (i=0; i < intToPtr.size; i++)
	{	IntToPtr_::PtrIntPairP obj;
		for (obj=intToPtr.data[i];  obj;  obj=obj->tail)
			if (obj && obj->p) 
				((Node*)(obj->p))->setPtrs(intToPtr);
	}
	root2=TableP(intToPtr.find(ptrToInt.find(root)));
	thisobj=LogoObjectP(intToPtr.find(ptrToInt.find(th->thisobj)));
	turtleNode=TurtleNodeP(intToPtr.find(ptrToInt.find(th->turtleNode)));
	th->varStack->copy(varStack, intToPtr, ptrToInt);
	th->funStack->copy(funStack, intToPtr, ptrToInt);
	th->argNrStack->copyTo(argNrStack);
	graphi=RECORDER.items;
	GC_::disabled=false;
	return;
}

void DebugState::copyToRoot()
{
	ThreadP th=Thread::This();
	root=root2;
	th->thisobj=thisobj;
	th->turtleNode=turtleNode;
	varStack.copyTo(*th->varStack);
	funStack.copyTo(*th->funStack);
	argNrStack.copyTo(*th->argNrStack);
	if (DebugStates::updateGraph)
	{	RECORDER.items=graphi;
		RECORDER.playAgain();
	}
	if (filename)
	{	theApp->editorShow();
		if (theApp->edFrame)
		{	theApp->edFrame->unmark(1);
			theApp->edFrame->mark(filename, col, row, 1);
		}
	}
}
//___________________________________________________________________

int DebugStates::defaultSize=0;
bool DebugStates::updateGraph=true;

DebugStates debugStates;

DebugStates::DebugStates(): size(0), active(0), first(0), last(0), base(0)
{}

void DebugStates::mark()
{
	if (!base)
		return;
	int i=first;
	while (i!=last)
	{	base[i].mark();
		i++;
		if (i >= size)
			i=0;
	}
	base[i].mark();
}

void DebugStates::markPtr(PtrToInt_ &ptrToInt)
{
	if (!base)
		return;
	int i=first;
	while (i!=last)
	{	base[i].markPtr(ptrToInt);
		i++;
		if (i >= size)
			i=0;
	}
	base[i].markPtr(ptrToInt);
}

bool DebugStates::init(int size_)
{
	base=DebugStateP(GC_::calloc(size_+1, sizeof(DebugState)));
	if (!base)
		return false;
	size=size_;
	return true;
}

void DebugStates::free()
{
	if (base)
	{	switchToLast();
		for (int i=0; i<size; i++)
		{	if (base[i].varStack.base)
				base[i].varStack.free();
			if (base[i].funStack.base)
				base[i].funStack.free();
			if (base[i].argNrStack.base)
				base[i].argNrStack.free();
		}
		GC_::free(base); base=0;
		first=active=last=0;
	}
	theApp->ringBufferSize=size;
}

void DebugStates::copyRoot()
{
	if (!size)
		return;
	ThreadP th=Thread::This();
	base[active].copyRoot();
	if (th->this_line 
	&& !th->this_line->filename()->equal(_Unnamed))
	{	base[active].filename=th->this_line->filename()->ptr();
		base[active].col=th->this_line->col;
		base[active].row=th->this_line->row;
	}else
	{	base[active].filename=0;
		base[active].row=0;
	}
	active++;
	if (active >= size)
		active=0;
	last=active;
	base[last].root2=0;
	base[last].thisobj=0;
//	base[last].turtleNode=0;
	base[last].varStack.free();
	base[last].funStack.free();
	base[last].argNrStack.free();
	base[last].graphi=0;
	if (last==first)
	{	first++;
		if (first >= size)
			first=0;
	}
}

bool DebugStates::switchToNext()
{
	if (active==last)
	{	return false;
	}
	active++;
	if (active >= size)
		active=0;
	base[active].copyToRoot();
	return true;
}

void DebugStates::switchToPrev()
{
	ThreadP th=Thread::This();
	if (active==last)
	{	base[last].root2=root;
		base[last].thisobj=th->thisobj;
		base[last].turtleNode=th->turtleNode;
		th->varStack->copyTo(base[last].varStack);
		th->funStack->copyTo(base[last].funStack);
		th->argNrStack->copyTo(base[last].argNrStack);
		base[last].graphi=RECORDER.items;
	}
	int old=active;
	active--;
	if (active < 0)
		active=size-1;
	if (active==first)
	{	active=old;
		return;
	}
	base[active].copyToRoot();
}

void DebugStates::switchTo(int state)
{
	ThreadP th=Thread::This();
	if (state >= 0 && state < size)
	{	if (active==last)
		{	base[last].root2=root;
			base[last].thisobj=th->thisobj;
			base[last].turtleNode=th->turtleNode;
			th->varStack->copyTo(base[last].varStack);
			th->funStack->copyTo(base[last].funStack);
			th->argNrStack->copyTo(base[last].argNrStack);
			base[last].graphi=RECORDER.items;
		}
		int old=active;
		active=state;
		if (active < 0)
			active=size-1;
		if (active==first 
		|| (active < first && first < last)
		|| (active > last && first < last))
		{	active=old;
			return;
		}
		base[active].copyToRoot();
	}
}

void DebugStates::switchToLast()
{
	if (active==last)
		return;
	active=last;
	base[active].copyToRoot();
}
