#include "stdafx.h"
#include "Main.h"
#include "Node.h"
#include "Pair.h"
#include "Float.h"
#include "FloatArray.h"
#include "Point.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

NodeP lESimGetAllFields(PairPP arg)
{
	int sx=arg[1]->a_int(1),
		sy=arg[2]->a_int(2);
	if (sx < 1)
	{	error(BAD_DATA, arg[1], 1); return Unbound;
	}
	if (sy < 1)
	{	error(BAD_DATA, arg[2], 2); return Unbound;
	}
	FloatArrayP 
		lrho=arg[3]->a_floatarray(3),
		lphi=arg[4]->a_floatarray(4),
		lE=arg[5]->a_floatarray(5),
		lB=arg[6]->a_floatarray(6),
		lj=arg[7]->a_floatarray(7);
	if throwing return Unbound;
	doubleP 
		rho=lrho->data(), 
		phi=lphi->data(), 
		E=lE->data(),
		B=lB->data(),
		j=lj->data();
	DPoint null(0,0,0);

	for (int iy=0; iy<sy; iy++)
	{	for (int ix=0; ix<sx; ix++)
		{	DPoint &Ei=*(DPoint*)&E[(ix+sx*iy)*3];
			Ei=0;
			double &phii=phi[ix+sx*iy];
			phii=0;
			for (int jy=0; jy<sy; jy++)
				for (int jx=0; jx<sx; jx++)
				{	DPoint rij=DPoint(double(ix-jx)/sx, double(iy-jy)/sy, 0);
					if (rij != null)
					{	double nrij=rij.Norm();
						Ei+=rij/(nrij*nrij*nrij)*rho[jx+sx*jy];
						phii+=rho[jx+sx*jy]/nrij;
					}
				}
			DPoint &Bi=*(DPoint*)&B[(ix+sx*iy)*3];
			Bi=0;
			for (int jy=0; jy<sy; jy++)
				for (int jx=0; jx<sx; jx++)
				{	DPoint rij=DPoint(double(ix-jx)/sx, double(iy-jy)/sy, 0);
					if (rij != null)
					{	double nrij=rij.Norm();
						DPoint &jj=*(DPoint*)&j[(jx+sx*jy)*3];
						Bi+=jj.X(rij)/(nrij*nrij*nrij);
					}
				}
			checkThrowing();
			if throwing return Unbound;
		}
	}
	return Unbound;
}

NodeP lESimOneStep(PairPP arg)
{
	enum params
	{	_sx=1,_sy,_sz,_st,
		_t,_dt,_eps,_v,
		_rho,_sigma,_phi,
		_E, _B, _j,
	};
	int 
		sx=arg[_sx]->a_int(_sx),
		sy=arg[_sy]->a_int(_sy),
		sz=arg[_sz]->a_int(_sz),
		st=arg[_st]->a_int(_st);
	if (sx < 1)
	{	error(BAD_DATA, arg[_sx], _sx); return Unbound;
	}
	if (sy < 1)
	{	error(BAD_DATA, arg[_sy], _sy); return Unbound;
	}
	if (sz < 1)
	{	error(BAD_DATA, arg[_sz], _sz); return Unbound;
	}
	if (st < 1)
	{	error(BAD_DATA, arg[_st], _st); return Unbound;
	}
//	double c=1;	//Lichtgeschwindigkeit in Pixel pro Frame
	int t=int(arg[_t]->a_float(_t));
	double dt=arg[_dt]->a_float(_dt);
	double eps=arg[_eps]->a_float(_eps);
	double v=arg[_v]->a_float(_v);
	FloatArrayP 
		lrho=arg[_rho]->a_floatarray(_rho),
		lsigma=arg[_sigma]->a_floatarray(_sigma),
		lphi=arg[_phi]->a_floatarray(_phi),
		lE=arg[_E]->a_floatarray(_E),
		lB=arg[_B]->a_floatarray(_B),
		lj=arg[_j]->a_floatarray(_j);
	if throwing return Unbound;
	doubleP 
		rho=lrho->data(), 
		sigma=lsigma->data(), 
		phi=lphi->data(), 
		E=lE->data(),
		B=lB->data(),
		j=lj->data();
	DPoint null(0,0,0);

	int sxy=sx*sy;
	int sit=sxy*(t % st);
	for (int iy=0; iy<sy; iy++)
	{	for (int ix=0; ix<sx; ix++)
		{	DPoint &Bi=*(DPoint*)&B[(ix+sx*iy+sit)*3];
			Bi=0;
			for (int jy=0; jy<sy; jy++)
				for (int jx=0; jx<sx; jx++)
				{	DPoint rij=DPoint(double(ix-jx)/sx, double(iy-jy)/sy, 0);
					if (rij != null)
					{	double nrij=rij.Norm();
						int jt=int(t-nrij/v);
						if (jt < 0) jt=0;
						int sjt=sxy*(jt % st);
						int sjt2=sxy*((jt+1) % st);
						DPoint &jj=*(DPoint*)&j[(jx+sx*jy+sjt)*3];
						DPoint &Ej = *(DPoint*)&E[(jx+sx*jy+sjt )*3];
						DPoint &Ej2= *(DPoint*)&E[(jx+sx*jy+sjt2)*3];
						DPoint dEj_dt=(Ej-Ej2);
						double nrij3=nrij*nrij*nrij;
						Bi+=(jj+dEj_dt).X(rij)/nrij3;
					}
				}
			Bi*=eps;
			checkThrowing();
			if throwing return Unbound;
		}
	}
	for (int iy=0; iy<sy; iy++)
	{	for (int ix=0; ix<sx; ix++)
		{	DPoint &Ei=*(DPoint*)&E[(ix+sx*iy+sit)*3];
			Ei=0;
			for (int jy=0; jy<sy; jy++)
				for (int jx=0; jx<sx; jx++)
				{	DPoint rij=DPoint(double(ix-jx)/sx, double(iy-jy)/sy, 0);
					if (rij != null)
					{	double nrij=rij.Norm();
						int jt=int(t-nrij/v);
						if (jt < 0) jt=0;
						int sjt=sxy*(jt % st);
						int sjt2=sxy*((jt+1) % st);
						double nrij3=nrij*nrij*nrij;
						DPoint &Bj = *(DPoint*)&B[(jx+sx*jy+sjt )*3];
						DPoint &Bj2= *(DPoint*)&B[(jx+sx*jy+sjt2)*3];
						DPoint dBj_dt=(Bj2-Bj);
						Ei+=dBj_dt.X(rij)/nrij3;
					}
				}
			checkThrowing();
			if throwing return Unbound;
		}
	}
	for (int iy=0; iy<sy; iy++)
	{	for (int ix=0; ix<sx; ix++)
		{	if (sigma[ix+sx*iy] > 0)
			{	DPoint &Ei=*(DPoint*)&E[(ix+sx*iy+sit)*3];
//				Ei=0;
				double &phii=phi[ix+sx*iy];
				phii=0;
				for (int jy=0; jy<sy; jy++)
					for (int jx=0; jx<sx; jx++)
					{	DPoint rij=DPoint(double(ix-jx)/sx, double(iy-jy)/sy, 0);
						if (rij != null)
						{	double nrij=rij.Norm();
							Ei+=rij/(nrij*nrij*nrij)*rho[jx+sx*jy];
							phii+=rho[jx+sx*jy]/nrij;
						}
					}
				checkThrowing();
				if throwing return Unbound;
			}
		}
	}
	for (int iy=1; iy<sy-1; iy++)
	{	for (int ix=1; ix<sx-1; ix++)
		{	//if (sigma[ix+sx*iy] > 0)
			{	DPoint &Ei=*(DPoint*)&E[(ix+sx*iy+sit)*3];
				DPoint &E1=*(DPoint*)&E[(ix-1+sx*iy+sit)*3];
				DPoint &E2=*(DPoint*)&E[(ix+1+sx*iy+sit)*3];
				DPoint &E3=*(DPoint*)&E[(ix+sx*(iy-1)+sit)*3];
				DPoint &E4=*(DPoint*)&E[(ix+sx*(iy+1)+sit)*3];
				Ei=(Ei*4+E1+E2+E3+E4)/8;
			}
		}
	}
	for (int iy=1; iy<sy-1; iy++)
	{	for (int ix=1; ix<sx-1; ix++)
		{	DPoint &Bi=*(DPoint*)&B[(ix+sx*iy+sit)*3];
			DPoint &B1=*(DPoint*)&B[(ix-1+sx*iy+sit)*3];
			DPoint &B2=*(DPoint*)&B[(ix+1+sx*iy+sit)*3];
			DPoint &B3=*(DPoint*)&B[(ix+sx*(iy-1)+sit)*3];
			DPoint &B4=*(DPoint*)&B[(ix+sx*(iy+1)+sit)*3];
			Bi=(Bi*4+B1+B2+B3+B4)/8;
		}
	}
	doubleP drho=doubleP(GC_::calloc(sx*sy, sizeof(double)));
	for (int iy=1; iy<sy-1; iy++)
	{	for (int ix=1; ix<sx-1; ix++)
		{	DPoint &Ei=*(DPoint*)&E[(ix+sx*iy+sit)*3];
//			DPoint &Bi=*(DPoint*)&B[(ix+sx*iy+sit)*3];
			double d, h=sigma[ix+sx*iy]*Float::abs(rho[ix+sx*iy])*dt;
			DPoint &ji=*(DPoint*)&j[(ix+sx*iy+sit)*3];
			ji=0;
			d=Ei.x*h;
			drho[ix-1+sx*iy]-=d;
			drho[ix+1+sx*iy]+=d;
			ji.x=d;

			d=Ei.y*h;
			drho[ix+sx*(iy-1)]-=d;
			drho[ix+sx*(iy+1)]+=d;
			ji.y=d;
		}
	}
	for (int iy=0; iy<sy; iy++)
		for (int ix=0; ix<sx; ix++)
			rho[ix+sx*iy]+=drho[ix+sx*iy];
	GC_::free(drho); drho=0;
	return Unbound;
}
