//___________________________________________________________________
//
//                  aUCBLogo Editor module header
//                         based on stctest
//             written by Otto Wyss and Andreas Micheler
//___________________________________________________________________

class Edit;
class FileDropTarget;

class EdFrame: public wxPanel
{
	friend class App;
	friend class AppBook;
	friend class AppAbout;
	friend class EditPrint;

public:
	wxNotebook *notebook;
	int fontSize;
	bool inReverseStep;

	//! constructor
	EdFrame(wxWindow *parent, 
		const wxString &title, wxPoint pos, wxSize size,
		int argc, wxChar** argv);

	//! destructor
	~EdFrame ();

	void show(bool shown=true);

	//! event handlers
	//! common
	void OnClose (wxCloseEvent &event);
	void OnAbout (wxCommandEvent &event);
	void OnExit (wxCommandEvent &event);
	void OnTimerEvent (wxTimerEvent &event);
	//! file
	void OnFileNew			(wxCommandEvent &event);
	void OnFileNewFrame	(wxCommandEvent &event);
	void OnFileOpen		(wxCommandEvent &event);
	void OnFileOpenFrame	(wxCommandEvent &event);
	void OnFileSave		(wxCommandEvent &event);
	void OnFileSaveAs		(wxCommandEvent &event);
	void OnFileClose		(wxCommandEvent &event);
	void OnSetFont			(wxCommandEvent &event);
	void callOnRun			(wxCommandEvent &event);
	void OnRun				(wxCommandEvent &event); 
	void callOnStep		(wxCommandEvent &event);
	void OnStep				(wxCommandEvent &event); 
	void callOnReverseStep(wxCommandEvent &event);
	void OnReverseStep    (wxCommandEvent &event);
	void callOnStepOver  (wxCommandEvent &event);
	void OnStepOver      (wxCommandEvent &event);
	void callOnReverseStepOver(wxCommandEvent &event);
	void OnReverseStepOver    (wxCommandEvent &event);
	void OnBreakpoint		(wxCommandEvent &event);
	void OnLoad          (wxCommandEvent &event);
	void OnAutoLoad      (wxCommandEvent &event);
	void OnSetRunCommand	(wxCommandEvent &event); 
	void OnSetRingBuffer (wxCommandEvent &event); 
	void OnSetRunDir		(wxCommandEvent &event); 
	void OnSetHelpFileName(wxCommandEvent &event);
	//! properties
	void OnProperties (wxCommandEvent &event);
	//! print
	void OnPrintSetup (wxCommandEvent &event);
	void OnPrintPreview (wxCommandEvent &event);
	void OnPrint (wxCommandEvent &event);
	//! edit events
	void OnEdit (wxCommandEvent &event);

	void OnSelChange(wxNotebookEvent& event);
	void OnPageNext(wxCommandEvent &event);
	void OnPagePrev(wxCommandEvent &event);
	void OnConsoleSetFocus(wxCommandEvent &event);
	void OnMakeBackupFiles(wxCommandEvent &event);
	void OnSetFocus(wxFocusEvent &event);

	void FileOpen (wxString fname);
	void FileSave(charP filename);
	void FileClose(charP filename);
	void mark(charP filename, int col, int row, int markernr);
	void unmark(int markernr);
	void gotoLine(int line);
	void run();
	void togglePause(Edit * edit, int col, int row, StringP fn, PairP parent);
	void togglePauseTree(Edit * edit, int col, int row, StringP fn, PairP parent, bool clear);
	void togglePauseTree2(Edit * edit, int col, int row, StringP fn, PairP parent, int &n, bool clear);

private:
#ifdef WIN32
	FileDropTarget *fileDropTarget;
#endif

	//! creates the application menu bar
	wxFlatMenuBar *m_menuBar;
	void CreateMenu ();

	// print preview position and size
	wxRect DeterminePrintSize ();

	DECLARE_EVENT_TABLE()
};

class EdAbout: public wxDialog 
{
public:
	//! constructor
	EdAbout (wxWindow *parent,
			  int milliseconds = 0,
			  long style = 0);

	//! destructor
	~EdAbout ();

	// event handlers
	void OnTimerEvent (wxTimerEvent &event);

private:
	// timer
	wxTimer *m_timer;

	DECLARE_EVENT_TABLE()
};

