//___________________________________________________________________
//
// File:        edit.cpp
// Purpose:     STC Edit class
// Maintainer:  Andreas Micheler
// Created:     2003-09-01
// RCS-ID:      $Id: edit.cpp,v 1.8 2005/03/03 19:50:33 ABX Exp $
// Licence:     wxWindows licence
//___________________________________________________________________

//----------------------------------------------------------------------------
// headers
//----------------------------------------------------------------------------

#include "stdafx.h"
//! wxWidgets headers
#include <wx/app.h>
#include <wx/file.h>     // raw file io support
#include <wx/filename.h> // filename support
#include <wx/tokenzr.h>  // string tokenizer
#include <wx/arrstr.h>   // wxArrayString
#include <wx/help.h> // html help
#include <wx/dynarray.h>
#include <wx/arrimpl.cpp>
#include <wx/notebook.h>
#include <wx/numdlg.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>

//! application headers
#include "Main.h"
#include "File.h"
#include "LogoApp.h"
#include "Environment.h"
#include "EdDefsExt.h"     // additional definitions
#include "EdPrefs.h"
#include "Edit.h"        // edit module
#include "EdFrame.h"
#include "String.h"
#include "Pair.h"
#include "Thread.h"
#include "Error.h"
#include "Color.h"
#include "Console.h"
#include "MainFrame.h"

#ifdef _DEBUG
	#include <crtdbg.h>
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

#define NONAME _("noname.lg")

WX_DEFINE_OBJARRAY(intArray);

BEGIN_EVENT_TABLE (Edit, wxStyledTextCtrl)
	// common
	EVT_SET_FOCUS(							  Edit::OnSetFocus)
	EVT_ENTER_WINDOW(						  Edit::OnEnterWindow)
	EVT_LEAVE_WINDOW(						  Edit::OnLeaveWindow)
	EVT_SIZE (                         Edit::OnSize)
	// edit
	EVT_MENU (wxID_CLEAR,              Edit::OnEditClear)
	EVT_MENU (wxID_CUT,                Edit::OnEditCut)
	EVT_MENU (wxID_COPY,               Edit::OnEditCopy)
	EVT_MENU (wxID_PASTE,              Edit::OnEditPaste)
	EVT_MENU (myID_INDENTINC,          Edit::OnEditIndentInc)
	EVT_MENU (myID_INDENTRED,          Edit::OnEditIndentRed)
	EVT_MENU (wxID_SELECTALL,          Edit::OnEditSelectAll)
	EVT_MENU (myID_SELECTLINE,         Edit::OnEditSelectLine)
	EVT_MENU (wxID_REDO,               Edit::OnEditRedo)
	EVT_MENU (wxID_UNDO,               Edit::OnEditUndo)
	// find
	EVT_MENU (wxID_FIND,               Edit::OnFind)
	EVT_MENU (myID_FINDNEXT,           Edit::OnFindNext)
	EVT_MENU (myID_FINDPREVIOUS,       Edit::OnFindPrevious)
	EVT_MENU (myID_REPLACE,            Edit::OnReplace)
	EVT_MENU (myID_REPLACENEXT,        Edit::OnReplaceNext)
	EVT_MENU (myID_REPLACEALL,         Edit::OnReplaceAll)
	EVT_MENU (myID_FINDWHOLEWORD,      Edit::OnFindWholeWord)
	EVT_MENU (myID_SKIP_COMMENTS,      Edit::OnSkipComments)
	EVT_MENU (myID_BRACEMATCH,         Edit::OnBraceMatch)
	EVT_MENU (myID_GOTO,               Edit::OnGoto)
	EVT_MENU (myID_GOTODEF,            Edit::OnGotoDef)
	EVT_MENU (myID_GOTOPREV,           Edit::OnGotoPrev)
	// view
	EVT_MENU_RANGE (myID_HILIGHTFIRST, myID_HILIGHTLAST,
												Edit::OnHilightLang)
	EVT_MENU (myID_DISPLAYEOL,         Edit::OnDisplayEOL)
	EVT_MENU (myID_INDENTGUIDE,        Edit::OnIndentGuide)
	EVT_MENU (myID_LINENUMBER,         Edit::OnLineNumber)
	EVT_MENU (myID_LONGLINEON,         Edit::OnLongLineOn)
	EVT_MENU (myID_WHITESPACE,         Edit::OnWhiteSpace)
	EVT_MENU (myID_FOLDTOGGLE,         Edit::OnFoldToggle)
	EVT_MENU (myID_FOLDTOGGLEALL,      Edit::OnFoldToggleAll)
	EVT_MENU (myID_OVERTYPE,           Edit::OnSetOverType)
	EVT_MENU (myID_READONLY,           Edit::OnSetReadOnly)
	EVT_MENU (myID_WRAPMODEON,         Edit::OnWrapmodeOn)
	EVT_MENU (myID_CHARSETANSI,        Edit::OnUseCharset)
	EVT_MENU (myID_CHARSETMAC,         Edit::OnUseCharset)
	// extra
	EVT_MENU (myID_CHANGEUPLOW,        Edit::OnCorrectCase)
	EVT_MENU (myID_CHANGELOWER,        Edit::OnChangeCase)
	EVT_MENU (myID_CHANGEUPPER,        Edit::OnChangeCase)
	EVT_MENU (myID_CONVERTCR,          Edit::OnConvertEOL)
	EVT_MENU (myID_CONVERTCRLF,        Edit::OnConvertEOL)
	EVT_MENU (myID_CONVERTLF,          Edit::OnConvertEOL)
	EVT_MENU (myID_MAKETOASSIGN,       Edit::OnMakeToAssign)
	EVT_MENU (wxID_HELP,               Edit::OnHelp)

	EVT_MENU (myID_EXPORT_HTML,        Edit::OnExportHTML)
	// stc
    EVT_STC_MARGINCLICK (wxID_ANY,    Edit::OnMarginClick)
    EVT_STC_CHARADDED   (wxID_ANY,    Edit::OnCharAdded)
    EVT_STC_MODIFIED    (wxID_ANY,    Edit::OnModified)

	EVT_CONTEXT_MENU(Edit::OnContextMenu)
	EVT_MENU_RANGE(ID_EDITOR_INSERT, ID_EDITOR_INSERT+1000, Edit::OnInsert)
	EVT_CHAR(Edit::OnChar)
	EVT_KEY_UP(Edit::OnKeyUp)
	EVT_KEY_DOWN(Edit::OnKeyDown)
END_EVENT_TABLE()

Edit::Edit (wxString filename, wxWindow *parent, wxWindowID id,
			const wxPoint &pos,
			const wxSize &size,
			long style, int fontSize_)
	: wxStyledTextCtrl (parent, id, pos, size, style) 
{
    m_filename = filename;

	m_LineNrID = 0;
	m_DividerID = 1;
	m_FoldingID = 2;

	// initialize language
	m_language = NULL;

	// default font for all styles
	SetViewEOL (g_CommonPrefs.displayEOLEnable);
	SetIndentationGuides (g_CommonPrefs.indentGuideEnable);
	SetEdgeMode (g_CommonPrefs.longLineOnEnable?
				 wxSTC_EDGE_LINE: wxSTC_EDGE_NONE);
	SetViewWhiteSpace (g_CommonPrefs.whiteSpaceEnable?
					   wxSTC_WS_VISIBLEALWAYS: wxSTC_WS_INVISIBLE);
	SetOvertype (g_CommonPrefs.overTypeInitial);
	SetReadOnly (g_CommonPrefs.readOnlyInitial);
	SetWrapMode (g_CommonPrefs.wrapModeInitial?
				 wxSTC_WRAP_WORD: wxSTC_WRAP_NONE);
	fontSize=fontSize_;
	setFonts();

	// set visibility
	SetVisiblePolicy (wxSTC_VISIBLE_SLOP, 100);
//	SetXCaretPolicy (wxSTC_CARET_EVEN|wxSTC_VISIBLE_STRICT|wxSTC_CARET_SLOP, 1);
	SetYCaretPolicy (wxSTC_CARET_EVEN|wxSTC_VISIBLE_STRICT|wxSTC_CARET_SLOP, 10);

	// markers
	MarkerDefine (wxSTC_MARKNUM_FOLDER,        wxSTC_MARK_DOTDOTDOT, _T("BLACK"), _T("BLACK"));
	MarkerDefine (wxSTC_MARKNUM_FOLDEROPEN,    wxSTC_MARK_ARROWDOWN, _T("BLACK"), _T("BLACK"));
	MarkerDefine (wxSTC_MARKNUM_FOLDERSUB,     wxSTC_MARK_EMPTY,     _T("BLACK"), _T("BLACK"));
	MarkerDefine (wxSTC_MARKNUM_FOLDEREND,     wxSTC_MARK_DOTDOTDOT, _T("BLACK"), _T("WHITE"));
	MarkerDefine (wxSTC_MARKNUM_FOLDEROPENMID, wxSTC_MARK_ARROWDOWN, _T("BLACK"), _T("WHITE"));
	MarkerDefine (wxSTC_MARKNUM_FOLDERMIDTAIL, wxSTC_MARK_EMPTY,     _T("BLACK"), _T("BLACK"));
	MarkerDefine (wxSTC_MARKNUM_FOLDERTAIL,    wxSTC_MARK_EMPTY,     _T("BLACK"), _T("BLACK"));

	// miscelaneous
	m_LineNrMargin = TextWidth (wxSTC_STYLE_LINENUMBER, _T("_999999"));
	m_FoldingMargin = 16;
	CmdKeyClear (wxSTC_KEY_TAB, 0); // this is done by the menu accelerator key
	SetLayoutCache (wxSTC_CACHE_PAGE);
	m_wholeword=2;
	skipComments=false;
	allFolded=false;

	AutoCompSetIgnoreCase(true);
	AutoCompSetAutoHide(true);
	AutoCompSetDropRestOfWord(true);

#ifdef WIN32
	fileDropTarget=new FileDropTarget();
	if (fileDropTarget)
		SetDropTarget(fileDropTarget);
#endif
	posArray.Add(0);
	marker.Add(-1, 25);

	for (int i=0; i < contextMenuNr; i++)
		contextMenu[i]=0;

	SetUseHorizontalScrollBar(false);
}

Edit::~Edit () 
{
	if (contextMenu[0])
		delete contextMenu[0];
}

//----------------------------------------------------------------------------
// common event handlers

void Edit::OnSetFocus(wxFocusEvent& event)
{
//	theApp->edFrame->SetFocus();
//	SetFocus();
	event.Skip();
}

void Edit::OnEnterWindow(wxMouseEvent &event) 
{
//	theApp->edFrame->SetFocus();
//	SetFocus();
	event.Skip();
}

void Edit::OnLeaveWindow(wxMouseEvent &event) 
{
/*	if (this==theApp->graph && theApp->consoleFrame)
	{	theApp->consoleFrame->SetFocus();
		if (console)
			console->SetFocus();
	}else
*/
	event.Skip();
}

static bool F7KeyPressed=false;
static bool F8KeyPressed=false;

void Edit::OnChar(wxKeyEvent &event)
{
	event.Skip();
}

void Edit::OnKeyUp(wxKeyEvent &event)
{
	ThreadP th=Thread::This();
	int c=event.GetKeyCode();
	if (c==WXK_F7)
	{	th->stepNr=0;
		th->stepArmed=false;
		F7KeyPressed=false;
	}else if (c==WXK_F8)
	{	th->stepNr=0;
		th->stepArmed=false;
		F8KeyPressed=false;
	}else
		event.Skip();
	if (theApp->autoloadEdit 
	&& c!=WXK_UP && c!=WXK_DOWN && c!=WXK_LEFT && c!=WXK_RIGHT)
	{	if (theApp->edFrame)
		{	wxCommandEvent cev(wxEVT_COMMAND_MENU_SELECTED);
			cev.SetEventObject(theApp->edFrame);
			cev.SetId(myID_LOAD);
			GetEventHandler()->AddPendingEvent(cev);
		}
	}
}

void Edit::OnKeyDown(wxKeyEvent &event)
{
	int c=event.GetKeyCode();
	switch (c)
	{
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			if (event.ControlDown() && !event.AltDown())
			{	wxCommandEvent cev;
				cev.SetId(ID_LAYOUT_LOAD1+c-'1');
				theApp->mainFrame->OnLayoutLoad(cev);
				return;
			}
		default:
			;
	}
	if (c==WXK_F7 && !F7KeyPressed)
	{	if (theApp->edFrame)
		{	wxCommandEvent cev;
			theApp->edFrame->OnStep(cev);
		}
		F7KeyPressed=true;
	}else if (c==WXK_F8 && !F8KeyPressed)
	{	if (theApp->edFrame)
		{	wxCommandEvent cev;
			theApp->edFrame->OnStepOver(cev);
		}
		F8KeyPressed=true;
	}else
		event.Skip();
}

void Edit::OnSize( wxSizeEvent& event ) 
{
	int x = GetClientSize().x +
			(g_CommonPrefs.lineNumberEnable? m_LineNrMargin: 0) +
			(g_CommonPrefs.foldEnable? m_FoldingMargin: 0);
	if (x > 0) SetScrollWidth (x);
	event.Skip();
}

// edit event handlers
void Edit::OnEditRedo (wxCommandEvent &WXUNUSED(event)) 
{
	if (!CanRedo()) return;
	Redo ();
}

void Edit::OnEditUndo (wxCommandEvent &WXUNUSED(event)) 
{
	if (!CanUndo()) return;
	Undo ();
}

void Edit::OnEditClear (wxCommandEvent &WXUNUSED(event)) 
{
	if (GetReadOnly()) return;
	Clear ();
}

void Edit::OnEditCut (wxCommandEvent &WXUNUSED(event)) 
{
	if (GetReadOnly() || (GetSelectionEnd()-GetSelectionStart() <= 0)) return;
	Cut ();
}

void Edit::OnEditCopy (wxCommandEvent &WXUNUSED(event)) 
{
	if (GetSelectionEnd()-GetSelectionStart() <= 0) return;
	Copy ();
}

void Edit::OnEditPaste (wxCommandEvent &WXUNUSED(event)) 
{
	if (!CanPaste()) return;
	Paste ();
}

void Edit::GotoPosVisible(int pos)
{
	GotoPos(pos);
	EnsureVisibleEnforcePolicy(LineFromPosition(pos));
}

void Edit::selectWord()
{
	int pos = GetCurrentPos(),
		start = WordStartPosition(pos, true),
		end = WordEndPosition(pos, true);
	SetSelection(end, start);
}

void Edit::OnFind(wxCommandEvent &WXUNUSED(event)) 
{
	int pos=GetCurrentPos();
	posArray.Add(pos);
	pos = GetCurrentPos();
	int start = WordStartPosition(pos, true),
		end = WordEndPosition(pos, true);
	lastFindExpr=GetTextRange(start, end);
	lastFindExpr=wxGetTextFromUser(_("Find Text:"), _("Find"), lastFindExpr);
	if (lastFindExpr=="")
		return;
	int style;
	do
	{	pos=FindText(pos+1, GetLength(), lastFindExpr, m_wholeword);
		if (pos < 0)
			return;
		style=GetStyleAt(pos);
	} while (skipComments && (style==1 || style==2));
	GotoPosVisible(pos);
	selectWord();
}

void Edit::OnFindNext(wxCommandEvent &WXUNUSED(event)) 
{
	int pos=GetCurrentPos();
	posArray.Add(pos);
	GotoPos(WordEndPosition(pos, true));
	pos=GetCurrentPos();
	
	int style;
	do
	{	pos=FindText(pos+1, GetLength(), lastFindExpr, m_wholeword);
		if (pos < 0)
			return;
		style=GetStyleAt(pos);
	} while (skipComments && (style==1 || style==2));
	GotoPosVisible(pos);
	selectWord();
}

void Edit::OnFindPrevious(wxCommandEvent &WXUNUSED(event)) 
{
	int pos=GetCurrentPos();
	posArray.Add(pos);
	GotoPos(WordStartPosition(pos, true));
	pos=GetCurrentPos();

	int style;
	do
	{	pos=FindText(pos-1, 0, lastFindExpr, m_wholeword);
		if (pos < 0)
			return;
		style=GetStyleAt(pos);
	} while (skipComments && (style==1 || style==2));
	GotoPosVisible(pos);
	selectWord();
}

void Edit::OnReplace(wxCommandEvent &WXUNUSED(event)) 
{
	int pos = GetCurrentPos(),
		start = WordStartPosition(pos, true),
		end = WordEndPosition(pos, true);
	posArray.Add(pos);
	lastFindExpr=GetTextRange(start, end);
	lastFindExpr=wxGetTextFromUser(_("Find Text:"), _("Find & Replace"), lastFindExpr);
	if (lastFindExpr=="")
		return;
	int style;
	do
	{	pos=FindText(pos+1, GetLength(), lastFindExpr, m_wholeword);
		if (pos < 0)
			return;
		style=GetStyleAt(pos);
	} while (skipComments && (style==1 || style==2));
	GotoPosVisible(pos);
	SetSelection(pos, pos+lastFindExpr.Length());

	lastReplaceExpr=wxGetTextFromUser(_("Replace Text:"), _("Find & Replace"), lastFindExpr);
	if (lastReplaceExpr=="")
		return;
	ReplaceSelection(lastReplaceExpr);
}

void Edit::OnReplaceNext(wxCommandEvent &WXUNUSED(event)) 
{
	int pos=GetCurrentPos();
	posArray.Add(pos);

	if (lastFindExpr=="")
		return;
	int style;
	do
	{	pos=FindText(pos+1, GetLength(), lastFindExpr, m_wholeword);
		if (pos < 0)
			return;
		style=GetStyleAt(pos);
	} while (skipComments && (style==1 || style==2));
	GotoPos(pos);
	SetSelection(pos, pos+lastFindExpr.Length());

	ReplaceSelection(lastReplaceExpr);
}

void Edit::OnReplaceAll(wxCommandEvent &WXUNUSED(event)) 
{
	int pos=GetCurrentPos();
	posArray.Add(pos);

	wxCommandEvent cev;
	OnReplace(cev);

	if (lastFindExpr=="")
		return;
	while (true)
	{	int style;
		do
		{	pos=FindText(pos+1, GetLength(), lastFindExpr, m_wholeword);
			if (pos < 0)
				return;
			style=GetStyleAt(pos);
		} while (skipComments && (style==1 || style==2));
		GotoPos(pos);
		SetSelection(pos, pos+lastFindExpr.Length());
	
		ReplaceSelection(lastReplaceExpr);
	}
}

void Edit::OnFindWholeWord(wxCommandEvent &WXUNUSED(event)) 
{
	m_wholeword=2-m_wholeword;
}

void Edit::OnSkipComments(wxCommandEvent &WXUNUSED(event)) 
{
	skipComments=!skipComments;
}

void Edit::OnBraceMatch (wxCommandEvent &WXUNUSED(event)) 
{
	int min = GetCurrentPos ();
	int max = BraceMatch (min);
	if (max > (min+1)) 
	{	BraceHighlight (min+1, max);
		SetSelection (min+1, max);
	}else
	{	BraceBadLight (min);
	}
}

void Edit::OnGoto(wxCommandEvent &WXUNUSED(event)) 
{
	int pos=GetCurrentPos();
	posArray.Add(pos);
	int line=LineFromPosition(pos);
	line=wxGetNumberFromUser("", _("Line number:"), _("Goto Line"),
		line, 1, GetLineCount()+1);
	if (line==-1)
	{	wxMessageBox(
			wxString(_("Invalid Line Number ")) << line 
				<< _(" Range is: ") << 1 << ".." << GetLineCount()+1, 
			_("Goto Line Error!"));
		return;
	}
	gotoLine(line);
}

void Edit::gotoLine(int line)
{
	int pos=PositionFromLine(line-1);
	GotoPosVisible(pos);
	SetFocus();
}

bool Edit::gotoDef(wxString preposition, wxString word, int pos, int start, int end)
{
	wxString toword=preposition+word;
	bool wholeWord=true;
	pos=0;
	do
	{	pos=FindText(pos+1, GetLength(), toword, 0);
		if (pos < 0)
		{	word=word+"=";
			pos=GetCurrentPos();
			do
			{	pos=FindText(pos-1, 0, word, 0);
			} while (pos > 0 && isalnum(GetCharAt(pos-1)));

			if (pos < 0)
			{	pos=end+1;
				do 
				{	pos=FindText(pos+1, GetLength(), word, 0);
				} while (pos > 0 && isalnum(GetCharAt(pos-1)));
				if (pos < 0)
					return false;
			}
			GotoPosVisible(pos);

			int l=LineFromPosition(pos);
			if (GetFoldExpanded(l)==false)
				ToggleFold(l);
			return false;
		}
		int afterword=pos+toword.Length();
		if (afterword <= GetLength())
		{	wxChar c=GetCharAt(afterword);
			if (!isspace(c))
				wholeWord=false;
		}
	} while (!wholeWord);
	GotoPosVisible(pos);

	int l=LineFromPosition(pos);
	if (GetFoldExpanded(l)==false)
		ToggleFold(l);
	posArray.Add(pos);
	return true;
}

void Edit::OnGotoDef(wxCommandEvent &WXUNUSED(event)) 
{
	int pos = GetCurrentPos(),
		start = WordStartPosition(pos, true),
		end = WordEndPosition(pos, true);
	wxString word=GetTextRange(start, end);

	if (!gotoDef("to ", word, pos, start, end) 
	&&  !gotoDef("be ", word, pos, start, end))
	{	if (theApp->edFrame)
		{	wxNotebook *nb=theApp->edFrame->notebook;
			int count=nb->GetPageCount();
			for (int i=0;  i<count;  i++)
			{	Edit *ed=(Edit*)nb->GetPage(i);
				if (ed->gotoDef("to ", word, 0, 0, 0) 
				 || ed->gotoDef("be ", word, 0, 0, 0))
				{	nb->SetSelection(i);
					ed->SetFocus();
					return;
				}
			}
			// not yet found, so try /lib/<word>.lg
			wxFileName fn(environment.logolib, word+".lg");
			if (fn.FileExists())
			{	theApp->edFrame->FileOpen(fn.GetLongPath());
				return;
			}
			// not yet found, so try <currentDir>/<word>.lg
			if (wxFileExists(word+".lg"))
			{	theApp->edFrame->FileOpen(word+".lg");
				return;
			}
		}
	}
}

void Edit::OnGotoPrev(wxCommandEvent &WXUNUSED(event)) 
{
	int pos=posArray.Last();
	int c=posArray.GetCount();
	if (c > 1)
		posArray.RemoveAt(c-1);
	GotoPosVisible(pos);
}

void Edit::OnEditIndentInc (wxCommandEvent &WXUNUSED(event)) 
{
	CmdKeyExecute (wxSTC_CMD_TAB);
}

void Edit::OnEditIndentRed (wxCommandEvent &WXUNUSED(event)) 
{
	CmdKeyExecute (wxSTC_CMD_BACKTAB);
}

void Edit::OnEditSelectAll (wxCommandEvent &WXUNUSED(event)) 
{
	SetSelection (0, GetTextLength ());
}

void Edit::OnEditSelectLine (wxCommandEvent &WXUNUSED(event)) 
{
	int lineStart = PositionFromLine (GetCurrentLine());
	int lineEnd = PositionFromLine (GetCurrentLine() + 1);
	SetSelection (lineStart, lineEnd);
}

void Edit::OnHilightLang (wxCommandEvent &event) 
{
	InitializePrefs (g_LanguagePrefs [event.GetId() - myID_HILIGHTFIRST].name);
}

void Edit::OnDisplayEOL (wxCommandEvent &WXUNUSED(event)) 
{
	SetViewEOL (!GetViewEOL());
}

void Edit::OnIndentGuide(wxCommandEvent &WXUNUSED(event)) 
{
	SetIndentationGuides(!GetIndentationGuides());
}

void Edit::OnLineNumber(wxCommandEvent &WXUNUSED(event)) 
{
	SetMarginWidth(m_LineNrID,
					GetMarginWidth (m_LineNrID) == 0? m_LineNrMargin: 0);
}

void Edit::OnLongLineOn(wxCommandEvent &WXUNUSED(event)) 
{
	SetEdgeMode(GetEdgeMode() == 0? wxSTC_EDGE_LINE: wxSTC_EDGE_NONE);
}

void Edit::OnWhiteSpace(wxCommandEvent &WXUNUSED(event)) 
{
	SetViewWhiteSpace(GetViewWhiteSpace() == 0?
	                   wxSTC_WS_VISIBLEALWAYS: wxSTC_WS_INVISIBLE);
}

void Edit::OnFoldToggle(wxCommandEvent &WXUNUSED(event)) 
{
	ToggleFold(GetCurrentLine());
}

void Edit::OnFoldToggleAll(wxCommandEvent &WXUNUSED(event)) 
{
	int lo=1;
	for (int line=1; line < GetLineCount(); line++)
	{	int l=GetFoldParent(line);
		if (l != lo)
			if (allFolded)
			{	if (GetFoldExpanded(l)==false)
					ToggleFold(l);
			}else
			{	if (GetFoldExpanded(l)==true)
					ToggleFold(l);
			}
		lo=l;
	}
	allFolded = !allFolded;
}

void Edit::OnSetOverType(wxCommandEvent &WXUNUSED(event)) 
{
	SetOvertype (!GetOvertype());

}

void Edit::OnSetReadOnly (wxCommandEvent &WXUNUSED(event)) 
{
	SetReadOnly (!GetReadOnly());
}

void Edit::OnWrapmodeOn (wxCommandEvent &WXUNUSED(event)) 
{
	SetWrapMode (GetWrapMode() == 0? wxSTC_WRAP_WORD: wxSTC_WRAP_NONE);
}

void Edit::OnUseCharset (wxCommandEvent &event) 
{
	int Nr;
	int charset = GetCodePage();
	switch (event.GetId()) 
	{	case myID_CHARSETANSI: {charset = wxSTC_CHARSET_ANSI; break;}
		case myID_CHARSETMAC : {charset = wxSTC_CHARSET_ANSI; break;}
	}
	for (Nr = 0; Nr < wxSTC_STYLE_LASTPREDEFINED; Nr++) 
	{	StyleSetCharacterSet (Nr, charset);
	}
	SetCodePage (charset);
}

void Edit::OnChangeCase (wxCommandEvent &event) 
{
	switch (event.GetId()) 
	{	case myID_CHANGELOWER: 
		{	CmdKeyExecute (wxSTC_CMD_LOWERCASE); break;
		}
		case myID_CHANGEUPPER: 
		{	CmdKeyExecute (wxSTC_CMD_UPPERCASE); break;
		}
	}
}

void Edit::OnConvertEOL (wxCommandEvent &event) 
{
	int eolMode = GetEOLMode();
	switch (event.GetId()) 
	{	case myID_CONVERTCR  : { eolMode = wxSTC_EOL_CR  ; break;}
		case myID_CONVERTCRLF: { eolMode = wxSTC_EOL_CRLF; break;}
		case myID_CONVERTLF  : { eolMode = wxSTC_EOL_LF  ; break;}
	}
	ConvertEOLs (eolMode);
	SetEOLMode (eolMode);
}

void Edit::OnMakeToAssign(wxCommandEvent &WXUNUSED(event)) 
{
	wxString text=GetSelectedText();

	for (size_t i=0; i < text.Length(); i++)
	{	if (text[i]==';')	// skip comments
			while (text[i] != '\n')
				i++;
		if (text.Mid(i, 6).Lower() == _T("make \"") 
		&& (!i || !isalnum(text[i-1])))
		{	text.Remove(i, 6);
			while (text[i]!=' ' && i < text.Length())
			{	if (text[i]=='(' || text[i]==')' 
				||  text[i]=='[' || text[i]==']'
				||  text[i]=='{' || text[i]=='}' || text[i]==';')
				{	text=text.Left(i)+'='+text.Right(text.Length()-i);
					break;
				}
				i++;
			}
			if (i < text.Length())
				text[i]='=';
		}
	}
	ReplaceSelection(text);
}

//! misc
void Edit::OnMarginClick (wxStyledTextEvent &event) 
{
	if (event.GetMargin() == 2) 
	{	int lineClick = LineFromPosition (event.GetPosition());

		int levelClick = GetFoldLevel (lineClick);
		if ((levelClick & wxSTC_FOLDLEVELHEADERFLAG) > 0) 
			ToggleFold (lineClick);
	}
}

//! returns true if there is a word in "wordlist" that starts with "word"
bool Edit::HasWord(wxString word, wxString &wordlist)
{
	word.MakeLower();

	wxStringTokenizer tkz(wordlist.Lower(), _T(" "));
	while (tkz.HasMoreTokens())
	{	wxString token = tkz.GetNextToken();

		if (token.Cmp(word)==0)
			continue;
		if (token.StartsWith(word))
			return true;
	}
	return false;
}

extern wxString aUCBLogo_dynkeystring;

int __cdecl Edit::CompareCaseInsensitive(const wxString& first, 
													  const wxString& second)
{
    return first.CmpNoCase(second);
}

wxArrayString Edit::SortedKeywordArray()
{
	wxArrayString wordarray;
	wxString all=aUCBLogoWordlist+aUCBLogo_dynkeystring;

	wxStringTokenizer tkz(all, _T(" "));
	while ( tkz.HasMoreTokens() )
		wordarray.Add(tkz.GetNextToken());

//	wordarray.Sort(CompareCaseInsensitive);
	return wordarray;
}


wxString Edit::SortedKeywords()
{
	wxArrayString wordarray=SortedKeywordArray();
	wxString words;
	for (size_t i=1; i < wordarray.GetCount(); i++)
		words+=wxString(" ")+wordarray.Item(i);
	return words;
}

void Edit::mark(int col, int row, int markernr)
{
	marker[markernr]=row;
	MarkerAdd(row, markernr);
	GotoPosVisible(PositionFromLine(row)+col);
}

void Edit::unmark(int markernr)
{
	if (MarkerGet(marker[markernr]) | (1<<markernr))
		MarkerDelete(marker[markernr], markernr);
}

bool Edit::togglemark(int col, int row)
{
	bool deleted=false;
	for (int i=2; i<25; i++)
		if (marker[i]==row)
		{	MarkerDelete(marker[i], i);
			marker[i]=-1;
			deleted=true;
		}

	if (!deleted)
		for (int i=2; i<25; i++)
			if (marker[i]==-1)
			{	mark(col, row, i);
				break;
			}

	return !deleted;
}

void Edit::OnModified(wxStyledTextEvent &event) 
{
	if (event.GetModificationType() != wxSTC_MOD_CHANGEMARKER)
		unmark(0);
}

void Edit::OnCharAdded (wxStyledTextEvent &event) 
{
/*	int pos = GetCurrentPos();
	int start = WordStartPosition(pos, true);

	if (pos - start >= 1)
	{	wxString words=SortedKeywords();
		if (HasWord(GetTextRange(start, pos), words))
		{	if (!AutoCompActive())
				AutoCompShow(pos-start, words);
		}else
			AutoCompCancel();
	}
*/
	char chr = (char)event.GetKey();
	int currentLine = GetCurrentLine();
	// Change this if support for mac files with \r is needed
	if (chr == '\n') 
	{	int lineInd = 0;
		if (currentLine > 0) 
			lineInd = GetLineIndentation(currentLine-1);
		
		if (lineInd == 0) return;
		SetLineIndentation (currentLine, lineInd);
		GotoPos(PositionFromLine (currentLine) + lineInd/GetIndent());
	}
}

inline bool Edit::IsAWordChar(const int ch) 
{	return isalnum(ch) || ch == '_';
}

void Edit::OnCorrectCase (wxCommandEvent &WXUNUSED(event)) 
{
	wxArrayString wordarray=SortedKeywordArray();
	wxString text=GetSelectedText();
	size_t len=text.Length(), end=0;

	for (size_t start=0; start < len; start++)
	{	if (text[start]==';')	// skip comments
			while (text[start]!='\n')
				start++;

		end=start;
		while (IsAWordChar(text[end]) && end < text.Length())
			end++;

		if (end != start)
		{	wxString word=text.Mid(start, end-start);
			int found=wordarray.Index(word, false);

			if (found != wxNOT_FOUND)
			{	wxString newword=wordarray.Item(found);
				for (size_t i=start; i < end; i++)
					text[i]=newword[i-start];
			}
			start=end;
		}
	}
	ReplaceSelection(text);
}

void Edit::OnHelp(wxCommandEvent &WXUNUSED(event))
{
	int pos = GetCurrentPos(),
		start = WordStartPosition(pos, true),
		end = WordEndPosition(pos, true);
	wxString word=GetTextRange(start, end);

	if (!theApp->help)
	{	theApp->help=new wxHelpController();
		if (!theApp->help)
		{	error(OUT_OF_MEM); return;
		}
		theApp->help->Initialize(theApp->helpFileName);
	}
	theApp->help->KeywordSearch(word);
//	theApp->help->DisplayLow(word);
}

//----------------------------------------------------------------------------
// private functions

void Edit::setFonts()
{
	wxFont font (fontSize, wxMODERN, wxNORMAL, wxNORMAL);
	StyleSetFont (wxSTC_STYLE_DEFAULT, font);
	StyleSetForeground (wxSTC_STYLE_DEFAULT, *wxBLACK);
	StyleSetBackground (wxSTC_STYLE_DEFAULT, *wxWHITE);
	StyleSetForeground (wxSTC_STYLE_LINENUMBER, wxColour (_T("DARK GREY")));
	StyleSetBackground (wxSTC_STYLE_LINENUMBER, *wxWHITE);
	StyleSetForeground(wxSTC_STYLE_INDENTGUIDE, wxColour (_T("DARK GREY")));
	InitializePrefs(DeterminePrefs(GetFilename()));
}

wxString Edit::DeterminePrefs (const wxString &filename) 
{
	LanguageInfo const* curInfo;

	// determine language from filepatterns
	int languageNr;
	for (languageNr = 0; languageNr < g_LanguagePrefsSize; languageNr++) 
	{
		curInfo = &g_LanguagePrefs [languageNr];
		wxString filepattern = curInfo->filepattern;
		filepattern.Lower();

		while (!filepattern.empty())
		{	wxString cur = filepattern.BeforeFirst (';');

			if ((cur == filename)
			||  (cur == (filename.BeforeLast ('.') + _T(".*")))
			||  (cur == (_T("*.") + filename.AfterLast ('.')))) 
				return curInfo->name;
		
			filepattern = filepattern.AfterFirst (';');
		}
	}
	return wxEmptyString;
}

bool Edit::InitializePrefs (const wxString &name) 
{
	// initialize styles
	StyleClearAll();
	LanguageInfo *curInfo = 0;

	// determine language
	bool found = false;
	int languageNr;
	for (languageNr = 0; languageNr < g_LanguagePrefsSize; languageNr++) 
	{
		curInfo = &g_LanguagePrefs [languageNr];
		if (curInfo->name == name) 
		{	found = true; break;
		}
	}
	if (!found) return false;

	// set lexer and language
	SetLexer (curInfo->lexer);
	m_language = curInfo;

	// set margin for line numbers
	SetMarginType (m_LineNrID, wxSTC_MARGIN_NUMBER);
	StyleSetForeground (wxSTC_STYLE_LINENUMBER, wxColour (_T("DARK GREY")));
	StyleSetBackground (wxSTC_STYLE_LINENUMBER, *wxWHITE);
	SetMarginWidth (m_LineNrID, 0); // start out not visible

	// default fonts for all styles!
	int Nr;
	for (Nr = 0; Nr < wxSTC_STYLE_LASTPREDEFINED; Nr++) 

	{
		wxFont font (fontSize, wxMODERN, wxNORMAL, wxNORMAL);
		StyleSetFont (Nr, font);
	}

	// set common styles
	StyleSetForeground (wxSTC_STYLE_DEFAULT, wxColour (_T("DARK GREY")));
	StyleSetForeground (wxSTC_STYLE_INDENTGUIDE, wxColour (_T("DARK GREY")));

	// initialize settings
	if (g_CommonPrefs.syntaxEnable) 
	{
		int keywordnr = 0;
		for (Nr = 0; Nr < STYLE_TYPES_COUNT; Nr++) 
		{
			if (curInfo->styles[Nr].type == -1) continue;
			const StyleInfo &curType = g_StylePrefs [curInfo->styles[Nr].type];
			wxFont font (//curType.fontsize
				fontSize, wxMODERN, wxNORMAL, wxNORMAL, false,
			curType.fontname);
			StyleSetFont (Nr, font);
			if (curType.foreground) 
				StyleSetForeground (Nr, wxColour (curType.foreground));
			
			if (curType.background) 
				StyleSetBackground (Nr, wxColour (curType.background));
			
			StyleSetBold (Nr, (curType.fontstyle & mySTC_STYLE_BOLD) > 0);
			StyleSetItalic (Nr, (curType.fontstyle & mySTC_STYLE_ITALIC) > 0);
			StyleSetUnderline (Nr, (curType.fontstyle & mySTC_STYLE_UNDERL) > 0);
			StyleSetVisible (Nr, (curType.fontstyle & mySTC_STYLE_HIDDEN) == 0);
			StyleSetCase (Nr, curType.lettercase);

			const wxString pwords = curInfo->styles[Nr].words;
			if (pwords!=_T("")) 
			{	SetKeyWords (keywordnr, pwords);
				keywordnr += 1;
			}
		}
	}

	// set margin as unused
	SetMarginType (m_DividerID, wxSTC_MARGIN_SYMBOL);
	SetMarginWidth (m_DividerID, 0);
	SetMarginSensitive (m_DividerID, false);

	// folding
	SetMarginType (m_FoldingID, wxSTC_MARGIN_SYMBOL);
	SetMarginMask (m_FoldingID, wxSTC_MASK_FOLDERS);
    StyleSetBackground (m_FoldingID, *wxWHITE);
	SetMarginWidth (m_FoldingID, 0);
	SetMarginSensitive (m_FoldingID, false);

	if (g_CommonPrefs.foldEnable) 
	{	SetMarginWidth (m_FoldingID, curInfo->folds != 0? m_FoldingMargin: 0);
		SetMarginSensitive (m_FoldingID, curInfo->folds != 0);
		SetProperty (_T("fold"), curInfo->folds != 0? _T("1"): _T("0"));
		SetProperty (_T("fold.comment"),
		(curInfo->folds & mySTC_FOLD_COMMENT) > 0? _T("1"): _T("0"));
		SetProperty (_T("fold.compact"),
		(curInfo->folds & mySTC_FOLD_COMPACT) > 0? _T("1"): _T("0"));
		SetProperty (_T("fold.preprocessor"),

		(curInfo->folds & mySTC_FOLD_PREPROC) > 0? _T("1"): _T("0"));
		SetProperty (_T("fold.html"),
		(curInfo->folds & mySTC_FOLD_HTML) > 0? _T("1"): _T("0"));
		SetProperty (_T("fold.html.preprocessor"),
		(curInfo->folds & mySTC_FOLD_HTMLPREP) > 0? _T("1"): _T("0"));
		SetProperty (_T("fold.comment.python"),
		(curInfo->folds & mySTC_FOLD_COMMENTPY) > 0? _T("1"): _T("0"));
		SetProperty (_T("fold.quotes.python"),
		(curInfo->folds & mySTC_FOLD_QUOTESPY) > 0? _T("1"): _T("0"));
	}
	SetFoldFlags (wxSTC_FOLDFLAG_LINEBEFORE_CONTRACTED |
	wxSTC_FOLDFLAG_LINEAFTER_CONTRACTED);

	// set spaces and indention
	SetTabWidth (3);
	SetUseTabs (true);
	SetTabIndents (true);
	SetBackSpaceUnIndents (true);
	SetIndent (g_CommonPrefs.indentEnable? 3: 0);

	// others
	SetViewEOL (g_CommonPrefs.displayEOLEnable);
	SetIndentationGuides (g_CommonPrefs.indentGuideEnable);
	SetEdgeColumn (80);
	SetEdgeMode (g_CommonPrefs.longLineOnEnable? wxSTC_EDGE_LINE: wxSTC_EDGE_NONE);
	SetViewWhiteSpace (g_CommonPrefs.whiteSpaceEnable?
	wxSTC_WS_VISIBLEALWAYS: wxSTC_WS_INVISIBLE);
	SetOvertype (g_CommonPrefs.overTypeInitial);
	SetReadOnly (g_CommonPrefs.readOnlyInitial);
	SetWrapMode (g_CommonPrefs.wrapModeInitial?
	wxSTC_WRAP_WORD: wxSTC_WRAP_NONE);

	MarkerDefine(0, 0, *wxWHITE, wxColour(255,128,128));
	MarkerDefine(1, 0, *wxWHITE, *wxGREEN);
	for (int i=2; i<25; i++)
		MarkerDefine(i, 0, *wxWHITE, wxColour("YELLOW"));

	aUCBLogo_dynkeystring=wxEmptyString;
	Colourise(0, -1);
	Colourise(0, -1);
	return true;

}

bool Edit::LoadFile () 
{
	// get filname
	if (!m_filename) 
	{	wxFileDialog dlg (this, _("Open file"), wxEmptyString, wxEmptyString,
			_("Any file (*)|*"), wxOPEN | wxFILE_MUST_EXIST | wxCHANGE_DIR);

		if (dlg.ShowModal() != wxID_OK) return false;
		m_filename = dlg.GetPath();
	}
	// load file
	return LoadFile (m_filename);
}

bool Edit::LoadFile (const wxString &filename) 
{
	// load file in edit and clear undo
    if (!filename.empty()) m_filename = filename;
	//wxFile file (m_filename);
	//if (!file.IsOpened()) return false;
//	ClearAll ();
	//long lng = file.Length ();
	//if (lng > 0)
	//{	wxString buf;
	//	wxChar *buff = buf.GetWriteBuf (lng);
	//	file.Read (buff, lng);
	//	buf.UngetWriteBuf ();
	//	InsertText (0, buf);
	//}
	//file.Close();

	wxStyledTextCtrl::LoadFile(m_filename);

	EmptyUndoBuffer();

	// determine lexer language
	wxFileName fname (m_filename);
	InitializePrefs (DeterminePrefs (fname.GetFullName()));
	return true;
}

bool Edit::SaveFile () 

{
	// return if no change
	if (!Modified()) return true;

	// get filname
	if (!m_filename) 
	{
        wxFileDialog dlg (this, _("Save file"), 
			wxEmptyString, wxEmptyString, _("Any file (*)|*"),
			wxSAVE | wxOVERWRITE_PROMPT);

		if (dlg.ShowModal() != wxID_OK) return false;
		m_filename = dlg.GetPath();
	}
	return SaveFile (m_filename);
}

bool Edit::SaveFile (const wxString &filename) 
{
	// return if no change
//	if (!Modified()) return true;

//	// save edit in file and clear undo
//	if (!filename.empty()) m_filename = filename;
//	wxFile file (m_filename, wxFile::write);
//	if (!file.IsOpened()) return false;
//	wxString buf = GetText();
//	bool okay = file.Write (buf);
//	file.Close();
//	if (!okay) return false;
//	EmptyUndoBuffer();
//	SetSavePoint();

//	return true;

	if (theApp->backup)
		if (wxFile::Exists(filename))
			wxCopyFile(filename, filename+".bak");

	return wxStyledTextCtrl::SaveFile(filename);
}

bool Edit::Modified () 
{
	// return modified state
	return (GetModify() && !GetReadOnly());
}

//----------------------------------------------------------------------------
// EditProperties
//----------------------------------------------------------------------------

EditProperties::EditProperties (Edit *edit, long style)
: wxDialog (edit, wxID_ANY, wxEmptyString,
            wxDefaultPosition, wxDefaultSize,
            style | wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER) 
{
	// sets the application title
	SetTitle (_("Properties"));
	wxString text;

	// fullname
	wxBoxSizer *fullname = new wxBoxSizer (wxHORIZONTAL);
	if (!fullname)
	{	error(OUT_OF_MEM); return;
	}
	fullname->Add (10, 0);
   fullname->Add (new wxStaticText (this, wxID_ANY, _("Full filename"),
                                     wxDefaultPosition, 
									 wxSize(80, wxDefaultCoord)),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL);
   fullname->Add (new wxStaticText (this, wxID_ANY, edit->GetFilename()),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL);

	// text info
	wxGridSizer *textinfo = new wxGridSizer (4, 0, 2);
	if (!textinfo)
	{	error(OUT_OF_MEM); return;
	}
    textinfo->Add (new wxStaticText (this, wxID_ANY, _("Language"),
                                     wxDefaultPosition, 
									 wxSize(80, wxDefaultCoord)),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
	if (edit->m_language)
	{    textinfo->Add (new wxStaticText (this, wxID_ANY, edit->m_language->name),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);
	}else
	{    textinfo->Add (new wxStaticText (this, wxID_ANY, _("none")),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);
	}
    textinfo->Add (new wxStaticText (this, wxID_ANY, _("Lexer-ID: "),
                                     wxDefaultPosition, 
									 wxSize(80, wxDefaultCoord)),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
    text = wxString::Format (_T("%d"), edit->GetLexer());
    textinfo->Add (new wxStaticText (this, wxID_ANY, text),
                   0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);
    wxString EOLtype = wxEmptyString;
    switch (edit->GetEOLMode()) 
	{	case wxSTC_EOL_CR  : {EOLtype = _T(" CR   (Unix)"); break; }
		case wxSTC_EOL_CRLF: {EOLtype = _T("CRLF (Windows)"); break; }
		case wxSTC_EOL_LF  : {EOLtype = _T(" LF (Macintosh)"); break; }
    }
    textinfo->Add (new wxStaticText (this, wxID_ANY, _("Line endings"),
                                     wxDefaultPosition, 
									 wxSize(80, wxDefaultCoord)),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
    textinfo->Add (new wxStaticText (this, wxID_ANY, EOLtype),
                   0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);

	// text info box
	wxStaticBoxSizer *textinfos = new wxStaticBoxSizer (
					 new wxStaticBox (this, wxID_ANY, _("Informations")),
					 wxVERTICAL);
	if (!textinfos)
	{	error(OUT_OF_MEM); return;
	}
	textinfos->Add (textinfo, 0, wxEXPAND);
	textinfos->Add (0, 6);

	// statistic
	wxGridSizer *statistic = new wxGridSizer (4, 0, 2);
	if (!statistic)
	{	error(OUT_OF_MEM); return;
	}
	statistic->Add (new wxStaticText (this, wxID_ANY, _("Total lines"),
											 wxDefaultPosition, 
								 wxSize(80, wxDefaultCoord)),
					  0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
	text = wxString::Format (_T("%d"), edit->GetLineCount());
	statistic->Add (new wxStaticText (this, wxID_ANY, text),
					  0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);
	statistic->Add (new wxStaticText (this, wxID_ANY, _("Total chars"),
											 wxDefaultPosition, 
								 wxSize(80, wxDefaultCoord)),
					  0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
	text = wxString::Format (_T("%d"), edit->GetTextLength());
	statistic->Add (new wxStaticText (this, wxID_ANY, text),
					  0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);

	statistic->Add (new wxStaticText (this, wxID_ANY, _("Current line"),
											 wxDefaultPosition, 
								 wxSize(80, wxDefaultCoord)),
					  0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
	text = wxString::Format (_T("%d"), edit->GetCurrentLine()+1);
	statistic->Add (new wxStaticText (this, wxID_ANY, text),
					  0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);

	statistic->Add (new wxStaticText (this, wxID_ANY, _("Current column"),
											 wxDefaultPosition, 
								 wxSize(80, wxDefaultCoord)),
					  0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
	int pos=edit->GetCurrentPos(),
		line=edit->GetCurrentLine(),
		col=pos-edit->PositionFromLine(line);
	text = wxString::Format (_T("%d"), col+1);
	statistic->Add (new wxStaticText (this, wxID_ANY, text),
					  0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);

	statistic->Add (new wxStaticText (this, wxID_ANY, _("Current pos"),
											 wxDefaultPosition, 
								 wxSize(80, wxDefaultCoord)),
					  0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
	text = wxString::Format (_T("%d"), edit->GetCurrentPos());
	statistic->Add (new wxStaticText (this, wxID_ANY, text),
					  0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);

	// char/line statistics
	wxStaticBoxSizer *statistics = new wxStaticBoxSizer (
						new wxStaticBox (this, wxID_ANY, _("Statistics")),
						wxVERTICAL);
	if (!statistics)
	{	error(OUT_OF_MEM); return;
	}
	statistics->Add (statistic, 0, wxEXPAND);
	statistics->Add (0, 6);

	// total pane
	wxBoxSizer *totalpane = new wxBoxSizer (wxVERTICAL);
	if (!totalpane)
	{	error(OUT_OF_MEM); return;
	}
	totalpane->Add (fullname, 0, wxEXPAND | wxLEFT | wxRIGHT | wxTOP, 10);
	totalpane->Add (0, 6);
	totalpane->Add (textinfos, 0, wxEXPAND | wxLEFT | wxRIGHT, 10);
	totalpane->Add (0, 10);
	totalpane->Add (statistics, 0, wxEXPAND | wxLEFT | wxRIGHT, 10);
	totalpane->Add (0, 6);
	wxButton *okButton = new wxButton (this, wxID_OK, _("OK"));
	if (!okButton)
	{	error(OUT_OF_MEM); return;
	}
	okButton->SetDefault();
	totalpane->Add (okButton, 0, wxALIGN_CENTER | wxALL, 10);

	SetSizerAndFit (totalpane);
	ShowModal();
}

//----------------------------------------------------------------------------
// EditPrint
//----------------------------------------------------------------------------

EditPrint::EditPrint (Edit *edit, wxChar *title)
: wxPrintout(title) 
{
	m_edit = edit;
	m_printed = 0;
}

bool EditPrint::OnPrintPage (int page) 
{
	wxDC *dc = GetDC();
	if (!dc) return false;

	// scale DC
	PrintScaling (dc);

	// print page
	if (page == 1) m_printed = 0;
	m_printed = m_edit->FormatRange (1, m_printed, m_edit->GetLength(),
									 dc, dc, m_printRect, m_pageRect);
	return true;
}

bool EditPrint::OnBeginDocument (int startPage, int endPage) 
{
	if (!wxPrintout::OnBeginDocument (startPage, endPage)) 
		return false;
	return true;
}

void EditPrint::GetPageInfo (int *minPage, int *maxPage, 
							 int *selPageFrom, int *selPageTo) 
{
	// initialize values
	*minPage = 0;
	*maxPage = 0;

	*selPageFrom = 0;
	*selPageTo = 0;


	// scale DC if possible
	wxDC *dc = GetDC();
	if (!dc) return;
	PrintScaling (dc);

	// get print page informations and convert to printer pixels
	wxSize ppiScr;
	GetPPIScreen (&ppiScr.x, &ppiScr.y);
	wxSize page = g_pageSetupData->GetPaperSize();
	page.x = static_cast<int> (page.x * ppiScr.x / 25.4);
	page.y = static_cast<int> (page.y * ppiScr.y / 25.4);
	m_pageRect = wxRect (0, 0,  page.x, page.y);

	// get margins informations and convert to printer pixels
	wxPoint pt = g_pageSetupData->GetMarginTopLeft();
	int left = pt.x;
	int top = pt.y;
	pt = g_pageSetupData->GetMarginBottomRight();
	int right = pt.x;
	int bottom = pt.y;
	top = static_cast<int> (top * ppiScr.y / 25.4);
	bottom = static_cast<int> (bottom * ppiScr.y / 25.4);
	left = static_cast<int> (left * ppiScr.x / 25.4);
	right = static_cast<int> (right * ppiScr.x / 25.4);
	m_printRect = wxRect (left,
								 top,
								 page.x - (left + right),
								 page.y - (top + bottom));

	// count pages
	while (HasPage (*maxPage)) 
	{	m_printed = m_edit->FormatRange (0, m_printed, m_edit->GetLength(),
									   dc, dc, m_printRect, m_pageRect);
		*maxPage += 1;
	}
	m_printed=0;
	if (*maxPage > 0) *minPage = 1;
	*selPageFrom = *minPage;
	*selPageTo = *maxPage;
}

bool EditPrint::HasPage (int WXUNUSED(page)) 
{
	return (m_printed < m_edit->GetLength());
}

bool EditPrint::PrintScaling (wxDC *dc)
{
	// check for dc, return if none
	if (!dc) return false;

	// get printer and screen sizing values
	wxSize ppiScr;
	GetPPIScreen (&ppiScr.x, &ppiScr.y);
	if (ppiScr.x == 0)  // most possible guess 96 dpi
	{	ppiScr.x = 96;
		ppiScr.y = 96;
	}
	wxSize ppiPrt;
	GetPPIPrinter (&ppiPrt.x, &ppiPrt.y);
	if (ppiPrt.x == 0)  // scaling factor to 1
	{	ppiPrt.x = ppiScr.x;
		ppiPrt.y = ppiScr.y;
	}
	wxSize dcSize = dc->GetSize();
	wxSize pageSize;
	GetPageSizePixels (&pageSize.x, &pageSize.y);

	// set user scale
	float scale_x = (float)(ppiPrt.x * dcSize.x) /
					(float)(ppiScr.x * pageSize.x);
	float scale_y = (float)(ppiPrt.y * dcSize.y) /
					(float)(ppiScr.y * pageSize.y);
	dc->SetUserScale (scale_x, scale_y);

	return true;
}

Edit::MenuTree Edit::command[]=
{
#include "toc.h"
	{0,0}
};

void Edit::createMenu(int &used, int &i, int level)
{
	int parent=used;
	contextMenu[parent] = new wxMenu;
	if (!contextMenu[parent])
	{	error(OUT_OF_MEM); return;
	}
	while (true)
	{	int newlevel=command[i].level;

		if (newlevel >= level)
		{	
			int nextlevel=command[i+1].level;
			if (nextlevel == newlevel)
			{	contextMenu[parent]->Append(ID_EDITOR_INSERT+i, command[i].str);
				i++;
			}
			else if (nextlevel > newlevel)
			{	used++;
				int item=i, child=used;
				i++;
				createMenu(used, i, newlevel);
				contextMenu[parent]->Append(ID_EDITOR_INSERT+item, 
					command[item].str, contextMenu[child]);
				nextlevel=command[i].level;
				if (nextlevel < newlevel)
					break;
			}
			else
			{	contextMenu[parent]->Append(ID_EDITOR_INSERT+i, command[i].str);
				i++;
				break;
			}
		}
		else
		{	i++;
			break;
		}
	};
}

void Edit::OnContextMenu(wxContextMenuEvent &event)
{
	if (!contextMenu[0])
	{
		int used=0, i=0, level=1; 
		createMenu(used, i, level);

		wxMenu &m = *contextMenu[0];
		m.AppendSeparator();
		m.Append(wxID_UNDO, _("&Undo"));
		m.Append(wxID_REDO, _("&Redo"));
		m.AppendSeparator();
		m.Append(wxID_CUT, _("Cu&t"));
		m.Append(wxID_COPY, _("&Copy"));
		m.Append(wxID_PASTE, _("&Paste"));
		m.Append(wxID_CLEAR, _("&Delete"));
		m.AppendSeparator();
		m.Append(wxID_SELECTALL, _("Select &All"));
	}
	PopupMenu(contextMenu[0]);
	return;
}

void Edit::OnInsert(wxCommandEvent &event)
{
	AddText(command[event.GetId()-ID_EDITOR_INSERT].str);
	AddText(' ');
}

void Edit::OnExportHTML(wxCommandEvent &event)
{
	// get filname
	if (!m_filename) 
	{
        wxFileDialog dlg (this, _("Save file"), 
			wxEmptyString, wxEmptyString, _("Any file (*)|*"),
			wxSAVE | wxOVERWRITE_PROMPT);

		if (dlg.ShowModal() != wxID_OK) 
			return;
		m_filename = dlg.GetPath();
	}
	wxString htmlName=m_filename+".html";
	wxFFileOutputStream output(htmlName);
	TextOutputStream out(output);
	wxString html=RenderAsHTML();
	out << html;
}

const wxString Edit::RenderAsHTML()
{
	wxString htmlString;

	int iPos, iStyle, iOldStyle=0, iMax;
	wxChar c;

	iOldStyle=-1;

	//force styling, then read document letter by letter
	iMax=GetLength();
	Colourise(0,iMax);

	wxColor *colors=new wxColor[g_StylePrefsSize];
	intP fontstyles=(intP)GC_::malloc(g_StylePrefsSize*sizeof(intP));

	for (int i=0; i<g_StylePrefsSize; i++)
	{	const StyleInfo &style=g_StylePrefs[m_language->styles[i].type];
		colors[i]=wxColor(style.foreground);
		fontstyles[i]=style.fontstyle;
	}
	for (iPos =0; iPos < iMax; iPos++)
	{
		iStyle = GetStyleAt (iPos);
		if (iStyle != iOldStyle)
		{	int p=iPos, s=iStyle;
			bool justSpace=true;
			do
			{	c=GetCharAt(p);
				s=GetStyleAt(p);
				if (s!=iStyle)
					break;
				if (c!=wxT(' ') && c!=wxT('\t') && c!=wxT('\n'))
					justSpace=false;
				p++;
			}while (justSpace && p < iMax);
			if (!justSpace)
			{	if  (iPos > 0)
				{	if (fontstyles[iOldStyle]==mySTC_STYLE_BOLD)
						htmlString << wxT("</b>");
					htmlString << wxT("</font>");
				}
				wxString sR, sG, sB;
				sR.Printf ( wxT("%02X"), colors[iStyle].Red());
				sG.Printf ( wxT("%02X"), colors[iStyle].Green());
				sB.Printf ( wxT("%02X"), colors[iStyle].Blue());
				htmlString << wxT("<font color=\"#") << sR << sG << sB << wxT("\">"); ;
				if	(fontstyles[iStyle]==mySTC_STYLE_BOLD)
					htmlString << wxT("<b>");
			}
		}
		c = GetCharAt(iPos);
		//translate <  & \n etc
		switch (c)
		{
			case wxT('\n') :
				htmlString << wxT("\n<br />");
				break ;
			case wxT('\r') :
				htmlString << wxT("\n<br />");
				break ;
			case wxT('<') :
				htmlString << wxT("&lt;");
				break ;
			case wxT('>') :
				htmlString << wxT("&gt;");
				break ;
			case wxT('&') :
				htmlString << wxT("&amp;");
				break ;
			case  wxT(' ') :
				htmlString << wxT("&nbsp;");
				break ;
			case  wxT('\t') :
				for (int i=0; i<String::tabSize; i++)
					htmlString << wxT("&nbsp;");
				break ;

			default:
				htmlString << c;
				break ;
		}
		iOldStyle = iStyle;
	}
	if (fontstyles[iOldStyle]==mySTC_STYLE_BOLD)
		htmlString << wxT("</b>");
	htmlString << wxT("</font>");
	delete[] colors;
	GC_::free(fontstyles); fontstyles=0;
	return htmlString;
}

NodeP lrenderLogoFileAsHTML(PairPP arg)
{
	StringP fname=arg[1]->a_string(1);
	if throwing return Unbound;
	if (!wxFileExists(fname->ptr()))
	{	error(FILE_ERROR, new String(fname, new String(_("doesn't exist!"))));
	}
	theApp->editorShow();
	if (!theApp->edFrame)
		return Unbound;
	Edit *edit=new Edit(fname->ptr(), theApp->edFrame);
	if (!edit)
	{	error(OUT_OF_MEM); return Unbound;
	}
	edit->LoadFile(fname->ptr());
	StringP result=0;
	wxString html=edit->RenderAsHTML();
	if throwing return Unbound;
	result=new String(html);
	delete edit;
	return result;
}

NodeP lrenderLogoAsHTML(PairPP arg)
{
	StringP str=arg[1]->a_string(1);
	if throwing return Unbound;
	theApp->editorShow();
	if (!theApp->edFrame)
		return Unbound;
	Edit *edit=new Edit(NONAME, theApp->edFrame);
	if (!edit)
	{	error(OUT_OF_MEM); return Unbound;
	}
	str=new String(str);
	if throwing return Unbound;
	edit->AddText(str->ptr());
	StringP result=0;
	wxString html=edit->RenderAsHTML();
	if throwing return Unbound;
	result=new String(html);
	delete edit;
	return result;
}
