//___________________________________________________________________
//
//								class Event and fellows
//___________________________________________________________________

#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <stdio.h>
#include <malloc.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/slider.h>
#include <wx/button.h>
#include <wx/tglbtn.h>
#include <wx/checkbox.h>
#include <wx/radiobut.h>
#include <wx/listctrl.h>
#include <wx/frame.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"
#include "GC.h"
#include "File.h"
#include "Pair.h"
#include "String.h"
#include "Int.h"
#include "UInt8.h"
#include "Object.h"
#include "LogoObject.h"
#include "Line.h"
#include "Window.h"
#include "Frame.h"
#include "Slider.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "Color.h"
#include "Point.h"
#include "Mouse.h"
#include "Console.h"
#include "Graph.h"
#include "LogoApp.h"
#include "Event.h"
#include "Thread.h"
#include "Stack.h"
#include "NodePStack.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

//___________________________________________________________________

EventFiFo::~EventFiFo() {if (allocated) GC_::free(base); base=0;}

EventFiFo &EventFiFo::operator<<(const EventP val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

EventFiFo &EventFiFo::operator>>(EventP &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}

void EventFiFo::expand()
{
	EventPP newbase = (EventP*) GC_::calloc(size+expand_size, sizeof(EventP));

	if (newbase==EventPP(0))	// no room to increase EventFiFo
	{	c_out << _("; Warning: Not enough memory to aquire more events, some will be lost.") << endL;
		_overflow=true;
	}
	else
	{	EventPP i=newbase;

		if (out)			// transfer old EventFiFo to new EventFiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;
		}				
		if (allocated) 
		{	GC_::free(base); base=0;
		}
		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}

//___________________________________________________________________

#define EVENT_KEY(NAME)									\
PairP Event##NAME::function=0;						\
																\
void Event##NAME::process()							\
{																\
	if (function)											\
	{	ThreadP th=Thread::This();						\
		th->keyboardValue=e->m_keyCode;				\
		th->keyEvent=e;									\
		ValStatus vs=Thread::This()->val_status;			\
		Thread::This()->val_status = no_value_allowed;	\
		function->eval_driver();						\
		Thread::This()->val_status=vs;						\
	}															\
}

EVENT_KEY(Char)
EVENT_KEY(KeyDown)
EVENT_KEY(KeyUp)
//___________________________________________________________________

void EventWindowClose::process()
{
	if ((window!=0) && *window->d(Window_::onClose))
	{	ThreadP th=Thread::This();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;					
		th->lexufun=ObjectP(*window->d(Window_::uonClose));				
		LogoObjectP thisobj=th->thisobj;		
		th->thisobj=LogoObjectP(*window->d(Window_::tonClose));			
		PairP(*window->d(Window_::onClose))->eval_driver();		
		th->thisobj=thisobj;						
		th->lexufun=lexufun;								
		th->val_status=vs;
		*PairP(window->d(Window_::onClose))=0;
		wxCloseEvent cev;
		window->OnClose(cev);
	}
}

#define EVENT_WINDOW0(NAME)											\
void EventWindow##NAME::process()									\
{																				\
	if ((window!=0) && window->on##NAME)							\
	{	ThreadP th=Thread::This();										\
		ValStatus vs=th->val_status;									\
		th->val_status = no_value_allowed;							\
		ObjectP lexufun=th->lexufun;									\
		th->lexufun=ObjectP(*window->d(Window_::uon##NAME));	\
		LogoObjectP thisobj=th->thisobj;								\
		th->thisobj=LogoObjectP(*window->d(Window_::ton##NAME));\
		PairP(*window->d(Window_::on##NAME))->eval_driver();	\
		th->thisobj=thisobj;												\
		th->lexufun=lexufun;												\
		th->val_status=vs;												\
	}																			\
}

EVENT_WINDOW0(Paint)
EVENT_WINDOW0(Erase)
EVENT_WINDOW0(SetFocus)
EVENT_WINDOW0(KillFocus)

#define EVENT_WINDOW_KEY(NAME)				\
void EventWindow##NAME::process()			\
{														\
	if ((window!=0) && *window->d(Window_::on##NAME))\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*window->d(Window_::uon##NAME));	\
		LogoObjectP thisobj=th->thisobj;		\
		th->thisobj=LogoObjectP(*window->d(Window_::ton##NAME));	\
		PairP(*window->d(Window_::on##NAME))->eval_driver();	\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_WINDOW_KEY(Char)
EVENT_WINDOW_KEY(KeyDown)
EVENT_WINDOW_KEY(KeyUp)

void EventWindowMove::process()									
{																				
	if ((window!=0) && window->onMove)
	{	ThreadP th=Thread::This();									
		th->moveEvent=e;
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*window->d(Window_::uonMove));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*window->d(Window_::tonMove));
		PairP(*window->d(Window_::onMove))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}

void EventWindowSize::process()									
{																				
	if ((window!=0) && window->onSize)
	{	ThreadP th=Thread::This();									
		th->sizeEvent=e;
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*window->d(Window_::uonSize));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*window->d(Window_::tonSize));
		PairP(*window->d(Window_::onSize))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}

#define EVENT_WINDOW_MOUSE(NAME)								\
void EventWindow##NAME::process()							\
{																		\
	if ((window!=0) && *window->d(Window_::on##NAME))	\
	{	ThreadP th=Thread::This();								\
		if (th->graph)												\
		{	th->graph->mouse->x = e->GetX();					\
			th->graph->mouse->y = e->GetY();					\
			th->graph->mouse->z = e->GetWheelRotation();	\
			th->graph->mouse->buttons =						\
				int(e->LeftDown()) |								\
				int(e->RightDown())<<1 |						\
				int(e->MiddleDown())<<2 |						\
				int(e->ShiftDown())<<3 |						\
				int(e->ControlDown())<<4 |						\
				int(e->AltDown())<<5 |							\
				int(e->MetaDown())<<6;							\
		}																\
		ValStatus vs=th->val_status;							\
		th->val_status = no_value_allowed;					\
		ObjectP lexufun=th->lexufun;							\
		th->lexufun=ObjectP(*window->d(Window_::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;									\
		th->thisobj=LogoObjectP(*window->d(Window_::ton##NAME));	\
		PairP(*window->d(Window_::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;										\
		th->lexufun=lexufun;										\
		th->val_status=vs;										\
	}																	\
}

EVENT_WINDOW_MOUSE(LeftDown)
EVENT_WINDOW_MOUSE(RightDown)
EVENT_WINDOW_MOUSE(MiddleDown)
EVENT_WINDOW_MOUSE(LeftUp)
EVENT_WINDOW_MOUSE(RightUp)
EVENT_WINDOW_MOUSE(MiddleUp)
EVENT_WINDOW_MOUSE(LeftDClick)
EVENT_WINDOW_MOUSE(RightDClick)
EVENT_WINDOW_MOUSE(MiddleDClick)
EVENT_WINDOW_MOUSE(Motion)
EVENT_WINDOW_MOUSE(EnterWindow)
EVENT_WINDOW_MOUSE(LeaveWindow)
EVENT_WINDOW_MOUSE(MouseWheel)
EVENT_WINDOW_MOUSE(Mouse)
//___________________________________________________________________
#define EVENT_FRAME_KEY(NAME)					\
														\
void EventFrame##NAME::process()				\
{														\
	if ((frame!=0) && *frame->d(Frame::on##NAME))		\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*frame->d(Frame::uon##NAME));			\
		LogoObjectP thisobj=th->thisobj;		\
		th->thisobj=LogoObjectP(*frame->d(Frame::ton##NAME));			\
		PairP(*frame->d(Frame::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_FRAME_KEY(Char)
EVENT_FRAME_KEY(KeyDown)
EVENT_FRAME_KEY(KeyUp)

void EventFrameClose::process()
{
	if ((frame!=0) && frame->onClose)
	{	ThreadP th=Thread::This();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;					
		th->lexufun=ObjectP(*frame->d(Frame::uonClose));				
		LogoObjectP thisobj=th->thisobj;		
		th->thisobj=LogoObjectP(*frame->d(Frame::tonClose));
		PairP(*frame->d(Frame::onClose))->eval_driver();		
		th->thisobj=thisobj;						
		th->lexufun=lexufun;								
		th->val_status=vs;
		*frame->d(Frame::onClose)=0;
		wxCloseEvent cev;
		frame->OnClose(cev);
	}
}
//___________________________________________________________________

#define EVENT_GRAPH_KEY(NAME)					\
														\
void EventGraph##NAME::process()				\
{														\
	if ((graph!=0) && *graph->d(Graph::on##NAME))		\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*graph->d(Graph::uon##NAME));	\
		LogoObjectP thisobj=th->thisobj;		\
		th->thisobj=LogoObjectP(*graph->d(Graph::ton##NAME));			\
		PairP(*graph->d(Graph::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;							\
		th->val_status=vs;						\
	}													\
}

EVENT_GRAPH_KEY(Char)
EVENT_GRAPH_KEY(KeyDown)
EVENT_GRAPH_KEY(KeyUp)
//___________________________________________________________________

#define EVENT_TEXT_MOUSE(NAME)								\
PairP EventTextMouse##NAME::function=0;					\
																		\
EventTextMouse##NAME::EventTextMouse##NAME(Mouse *m):	\
	Event(TextMouse##NAME),										\
	x(m->textx), y(m->texty), buttons(m->buttons) {}	\
																		\
void EventTextMouse##NAME::process()						\
{																		\
	if (function)													\
	{	if (console)												\
		{	console->mouse->textx = int(x);					\
			console->mouse->texty = int(y);					\
			console->mouse->buttons = buttons;				\
		}																\
		ValStatus vs=Thread::This()->val_status;					\
		Thread::This()->val_status = no_value_allowed;			\
		function->eval_driver();								\
		Thread::This()->val_status=vs;								\
	}																	\
}

EVENT_TEXT_MOUSE(LeftDown)
EVENT_TEXT_MOUSE(RightDown)
EVENT_TEXT_MOUSE(MiddleDown)

EVENT_TEXT_MOUSE(LeftUp)
EVENT_TEXT_MOUSE(RightUp)
EVENT_TEXT_MOUSE(MiddleUp)

EVENT_TEXT_MOUSE(LeftDClick)
EVENT_TEXT_MOUSE(RightDClick)
EVENT_TEXT_MOUSE(MiddleDClick)

EVENT_TEXT_MOUSE(Motion)
//___________________________________________________________________

#define EVENT_MOUSE(NAME)										\
PairP EventMouse##NAME::function=0;							\
																		\
EventMouse##NAME::EventMouse##NAME(Mouse *m):			\
	Event(Mouse##NAME),											\
	x(m->x), y(m->y), z(m->z), buttons(m->buttons) {}	\
																		\
void EventMouse##NAME::process()								\
{																		\
	if (function)													\
	{	if (console)												\
		{	console->mouse->x = x;								\
			console->mouse->y = y;								\
			console->mouse->z = z;								\
			console->mouse->buttons = buttons;				\
		}																\
		ValStatus vs=Thread::This()->val_status;			\
		Thread::This()->val_status = no_value_allowed;	\
		function->eval_driver();								\
		Thread::This()->val_status=vs;						\
	}																	\
}

PairP EventMouseLeftDown::function=0;

EventMouseLeftDown::EventMouseLeftDown(Mouse *m):
	Event(MouseLeftDown),
	x(m->x), y(m->y), z(m->z), buttons(m->buttons) {}

void EventMouseLeftDown::process()
{
	if (function)
	{	if (console)
		{	console->mouse->x = x;
			console->mouse->y = y;
			console->mouse->z = z;
			console->mouse->buttons = buttons;
		}
		ValStatus vs=Thread::This()->val_status;
		Thread::This()->val_status = no_value_allowed;
		function->eval_driver();
		Thread::This()->val_status=vs;
	}
}

//EVENT_MOUSE(LeftDown)
EVENT_MOUSE(RightDown)
EVENT_MOUSE(MiddleDown)

EVENT_MOUSE(LeftUp)
EVENT_MOUSE(RightUp)
EVENT_MOUSE(MiddleUp)

EVENT_MOUSE(LeftDClick)
EVENT_MOUSE(RightDClick)
EVENT_MOUSE(MiddleDClick)

EVENT_MOUSE(Motion)
//___________________________________________________________________

#define EVENT_GRAPH_MOUSE(NAME)								\
EventGraph##NAME::EventGraph##NAME(Graph *g, Mouse *m):\
	EventGraph(g, Graph##NAME),								\
	x(m->x), y(m->y), buttons(m->buttons) {}				\
																		\
void EventGraph##NAME::process()								\
{																		\
	if ((graph!=0) && *graph->d(Graph::on##NAME))		\
	{	ThreadP th=Thread::This();								\
		if (console)												\
		{	console->mouse->x = x;								\
			console->mouse->y = y;								\
			console->mouse->z = z;								\
			console->mouse->buttons = buttons;				\
		}																\
		ValStatus vs=th->val_status;							\
		th->val_status = no_value_allowed;					\
		ObjectP lexufun=th->lexufun;							\
		th->lexufun=ObjectP(*graph->d(Graph::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;								\
		th->thisobj=LogoObjectP(*graph->d(Graph::ton##NAME));	\
/*		if (th->graph && th->graph!=graph)							\
		{	th->graph->leave();											\
			graph->enter();												\
		}																		\
*/		PairP(*graph->d(Graph::on##NAME))->eval_driver();		\
/*		if (th->graph && th->graph!=graph)							\
		{	graph->leave();												\
			th->graph->enter();											\
		}																		\
*/		th->thisobj=thisobj;										\
		th->lexufun=lexufun;										\
		th->val_status=vs;										\
	}																	\
}

EventGraphMotion::EventGraphMotion(Graph *g, Mouse *m):\
	EventGraph(g, GraphMotion),								\
	x(m->x), y(m->y), buttons(m->buttons) {}				\
																		\
void EventGraphMotion::process()								\
{																		\
	if ((graph!=0) && *graph->d(Graph::onMotion))		\
	{	ThreadP th=Thread::This();								\
		if (console)												\
		{	console->mouse->x = x;								\
			console->mouse->y = y;								\
			console->mouse->z = z;								\
			console->mouse->buttons = buttons;				\
		}																\
		ValStatus vs=th->val_status;							\
		th->val_status = no_value_allowed;					\
		ObjectP lexufun=th->lexufun;							\
		th->lexufun=ObjectP(*graph->d(Graph::uonMotion));		\
		LogoObjectP thisobj=th->thisobj;								\
		th->thisobj=LogoObjectP(*graph->d(Graph::tonMotion));	\
/*		GraphP og=th->graph;												\
		int oo=og->graphOwned;											\
		if (th->graph && th->graph!=graph)							\
		{	og->graphOwned=1;												\
			og->leave();													\
			graph->enter();												\
		}																		\
*/
/*c_out << th->lexufun << th->thisobj; console->flush();*/	\
		PairP(*graph->d(Graph::onMotion))->eval_driver();		\
/*		if (th->graph && th->graph!=graph)							\
		{	graph->leave();												\
			og->enter();													\
			og->graphOwned=oo;											\
		}																		\
*/		th->thisobj=thisobj;										\
		th->lexufun=lexufun;										\
		th->val_status=vs;										\
	}																	\
}

EVENT_GRAPH_MOUSE(LeftDown)
EVENT_GRAPH_MOUSE(RightDown)
EVENT_GRAPH_MOUSE(MiddleDown)

EVENT_GRAPH_MOUSE(LeftUp)
EVENT_GRAPH_MOUSE(RightUp)
EVENT_GRAPH_MOUSE(MiddleUp)

EVENT_GRAPH_MOUSE(LeftDClick)
EVENT_GRAPH_MOUSE(RightDClick)
EVENT_GRAPH_MOUSE(MiddleDClick)

//EVENT_GRAPH_MOUSE(Motion)
//___________________________________________________________________

EventSlider::EventSlider(SliderP slider_):
	Event(tSlider), slider(slider_) 
{}

void EventSlider::process()
{
	if ((slider!=0) && slider->onScroll)
	{	ThreadP th=Thread::This();
		th->sliderValue=slider->GetValue();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*slider->d(Slider::uonScroll));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*slider->d(Slider::tonScroll));
		PairP(*slider->d(Slider::onScroll))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}
//___________________________________________________________________

EventIntControl::EventIntControl(IntControlP intControl_):
	Event(tIntControl), intControl(intControl_) 
{}

void EventIntControl::process()
{
	if ((intControl!=0) && *intControl->d(IntControl::onChange))
	{	ThreadP th=Thread::This();
		th->intControlValue=intControl->GetValue();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*intControl->d(IntControl::uonChange));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*intControl->d(IntControl::tonChange));
		PairP(*intControl->d(IntControl::onChange))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}
//___________________________________________________________________

EventFloatControl::EventFloatControl(FloatControlP floatControl_):
	Event(tFloatControl), floatControl(floatControl_) 
{}

void EventFloatControl::process()
{
	if ((floatControl!=0) && *floatControl->d(FloatControl::onChange))
	{	ThreadP th=Thread::This();
		th->floatControlValue=floatControl->GetValue();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*floatControl->d(FloatControl::uonChange));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*floatControl->d(FloatControl::tonChange));
		PairP(*floatControl->d(FloatControl::onChange))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}
//___________________________________________________________________

EventButton::EventButton(ButtonP button_):
	Event(tButton), button(button_) 
{}

void EventButton::process()
{
	if ((button!=0) && *button->d(Button::onClick))
	{	ThreadP th=Thread::This();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*button->d(Button::uonClick));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*button->d(Button::tonClick));
		PairP(*button->d(Button::onClick))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}
//___________________________________________________________________

EventToggleButton::EventToggleButton(ToggleButtonP toggleButton_):
	Event(tToggleButton), toggleButton(toggleButton_) 
{}

void EventToggleButton::process()
{
	if ((toggleButton!=0) && *toggleButton->d(ToggleButton::onClick))
	{	ThreadP th=Thread::This();
		th->toggleButtonValue=toggleButton->GetValue();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*toggleButton->d(ToggleButton::uonClick));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*toggleButton->d(ToggleButton::tonClick));
		PairP(*toggleButton->d(ToggleButton::onClick))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}
//___________________________________________________________________

EventCheckBox::EventCheckBox(CheckBoxP checkBox_):
	Event(tCheckBox), checkBox(checkBox_) 
{}

void EventCheckBox::process()
{
	if ((checkBox!=0) && *checkBox->d(CheckBox::onClick))
	{	ThreadP th=Thread::This();
		th->checkBoxValue=checkBox->GetValue();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*checkBox->d(CheckBox::uonClick));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*checkBox->d(CheckBox::tonClick));
		PairP(*checkBox->d(CheckBox::onClick))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}
//___________________________________________________________________

EventRadioButton::EventRadioButton(RadioButtonP radioButton_):
	Event(tRadioButton), radioButton(radioButton_) 
{}

void EventRadioButton::process()
{
	if ((radioButton!=0) && *radioButton->d(RadioButton::onClick))
	{	ThreadP th=Thread::This();
		th->radioButtonValue=radioButton->GetValue();
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*radioButton->d(RadioButton::uonClick));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*radioButton->d(RadioButton::tonClick));
		PairP(*radioButton->d(RadioButton::onClick))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}
//___________________________________________________________________

#define EVENT_TEXT_KEY(NAME)					\
void EventText##NAME::process()				\
{														\
	if ((textControl!=0) && *textControl->d(TextControl::on##NAME))		\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*textControl->d(TextControl::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;		\
		th->thisobj=LogoObjectP(*textControl->d(TextControl::ton##NAME));	\
		PairP(*textControl->d(TextControl::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_TEXT_KEY(Char)
EVENT_TEXT_KEY(KeyDown)
EVENT_TEXT_KEY(KeyUp)

#define EVENT_TEXT_CE(NAME)									\
void EventText##NAME::process()								\
{																		\
	if ((textControl!=0) && *textControl->d(TextControl::on##NAME))\
	{	ThreadP th=Thread::This();								\
		th->textControlValue=textControl->GetValue();	\
		ValStatus vs=th->val_status;							\
		th->val_status = no_value_allowed;					\
		ObjectP lexufun=th->lexufun;							\
		th->lexufun=ObjectP(*textControl->d(TextControl::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;												\
		th->thisobj=LogoObjectP(*textControl->d(TextControl::ton##NAME));	\
		PairP(*textControl->d(TextControl::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_TEXT_CE(Change)
EVENT_TEXT_CE(Enter)

//___________________________________________________________________

EventListBox::EventListBox(ListBox_ *listBox_, Type type_):
	Event(type_)
{	listBox=listBox_;
}

#define EVENT_LB_KEY(NAME)						\
void EventListBox##NAME::process()			\
{														\
	if ((listBox!=0) && *listBox->d(ListBox_::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		th->listBoxValue=listBox;				\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*listBox->d(ListBox_::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;		\
		th->thisobj=LogoObjectP(*listBox->d(ListBox_::ton##NAME));\
		PairP(*listBox->d(ListBox_::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_LB_KEY(Char)
EVENT_LB_KEY(KeyDown)
EVENT_LB_KEY(KeyUp)

#define EVENT_LB_SD(NAME)						\
void EventListBox##NAME::process()			\
{														\
	if ((listBox!=0) && *listBox->d(ListBox_::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->listBoxValue=listBox;				\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*listBox->d(ListBox_::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;		\
		th->thisobj=LogoObjectP(*listBox->d(ListBox_::ton##NAME));\
		PairP(*listBox->d(ListBox_::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

void EventListBoxSelect::process()
{
	if ((listBox!=0) && *listBox->d(ListBox_::onSelect))
	{	ThreadP th=Thread::This();
		th->listBoxValue=listBox;
		ValStatus vs=th->val_status;
		th->val_status = no_value_allowed;
		ObjectP lexufun=th->lexufun;
		th->lexufun=ObjectP(*listBox->d(ListBox_::uonSelect));
		LogoObjectP thisobj=th->thisobj;
		th->thisobj=LogoObjectP(*listBox->d(ListBox_::tonSelect));
		PairP(*listBox->d(ListBox_::onSelect))->eval_driver();
		th->thisobj=thisobj;
		th->lexufun=lexufun;
		th->val_status=vs;
	}
}

//EVENT_LB_SD(Select)
EVENT_LB_SD(DClick)
//___________________________________________________________________

#define EVENT_CB_KEY(NAME)						\
void EventChoiceBox##NAME::process()		\
{														\
	if ((choiceBox!=0) && *choiceBox->d(ChoiceBox::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		th->choiceBoxValue=choiceBox;			\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*choiceBox->d(ChoiceBox::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;											\
		th->thisobj=LogoObjectP(*choiceBox->d(ChoiceBox::ton##NAME));	\
		PairP(*choiceBox->d(ChoiceBox::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_CB_KEY(Char)
EVENT_CB_KEY(KeyDown)
EVENT_CB_KEY(KeyUp)

#define EVENT_CB_S(NAME)						\
void EventChoiceBox##NAME::process()		\
{														\
	if ((choiceBox!=0) && *choiceBox->d(ChoiceBox::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->choiceBoxValue=choiceBox;			\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*choiceBox->d(ChoiceBox::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;											\
		th->thisobj=LogoObjectP(*choiceBox->d(ChoiceBox::ton##NAME));	\
		PairP(*choiceBox->d(ChoiceBox::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_CB_S(Select)
//___________________________________________________________________

#define EVENT_COMBOBOX_KEY(NAME)						\
void EventComboBox##NAME::process()					\
{																\
	if ((comboBox!=0) && *comboBox->d(ComboBox::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		th->comboBoxValue=comboBox->GetValue();			\
		th->staticComboBox=comboBox;			\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*comboBox->d(ComboBox::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;											\
		th->thisobj=LogoObjectP(*comboBox->d(ComboBox::ton##NAME));	\
		PairP(*comboBox->d(ComboBox::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_COMBOBOX_KEY(Char)
EVENT_COMBOBOX_KEY(KeyDown)
EVENT_COMBOBOX_KEY(KeyUp)

#define EVENT_CB(NAME)							\
void EventComboBox##NAME::process()			\
{														\
	if ((comboBox!=0) && *comboBox->d(ComboBox::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->comboBoxValue=comboBox->GetValue();			\
		th->staticComboBox=comboBox;			\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*comboBox->d(ComboBox::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;										\
		th->thisobj=LogoObjectP(*comboBox->d(ComboBox::ton##NAME));	\
		PairP(*comboBox->d(ComboBox::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_CB(Select)
EVENT_CB(Change)
EVENT_CB(Enter)
//___________________________________________________________________

EventListControlItemSelected::EventListControlItemSelected(
	ListControl *listControl_, wxListEvent event):
EventListControl(listControl_, ListControlItemSelected)
{
	row=event.GetIndex();
	col=event.GetColumn();
	text=event.GetText();
}

EventListControlItemActivated::EventListControlItemActivated(
	ListControl *listControl_, wxListEvent event):
EventListControl(listControl_, ListControlItemActivated)
{
	row=event.GetIndex();
	col=event.GetColumn();
	text=event.GetText();
}

EventListControlColClick::EventListControlColClick(
	ListControl *listControl_, wxListEvent event):
EventListControl(listControl_, ListControlColClick)
{
	row=event.GetIndex();
	col=event.GetColumn();
	text=event.GetText();
}

#define EVENT_LC_KEY(NAME)								\
void EventListControl##NAME::process()				\
{																\
	if ((listControl!=0) && *listControl->d(ListControl::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->keyboardValue=e->m_keyCode;		\
		th->keyEvent=e;							\
		th->listControlValue=listControl;	\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*listControl->d(ListControl::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;											\
		th->thisobj=LogoObjectP(*listControl->d(ListControl::ton##NAME));	\
		PairP(*listControl->d(ListControl::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_LC_KEY(Char)
EVENT_LC_KEY(KeyDown)
EVENT_LC_KEY(KeyUp)

#define EVENT_LC_ASC(NAME)								\
void EventListControl##NAME::process()				\
{																\
	if ((listControl!=0) && *listControl->d(ListControl::on##NAME))			\
	{	ThreadP th=Thread::This();				\
		th->listControlValue=listControl;	\
		th->listControlRow=row;					\
		th->listControlCol=col;					\
		th->listControlText=text;				\
		ValStatus vs=th->val_status;			\
		th->val_status = no_value_allowed;	\
		ObjectP lexufun=th->lexufun;			\
		th->lexufun=ObjectP(*listControl->d(ListControl::uon##NAME));		\
		LogoObjectP thisobj=th->thisobj;											\
		th->thisobj=LogoObjectP(*listControl->d(ListControl::ton##NAME));	\
		PairP(*listControl->d(ListControl::on##NAME))->eval_driver();		\
		th->thisobj=thisobj;						\
		th->lexufun=lexufun;						\
		th->val_status=vs;						\
	}													\
}

EVENT_LC_ASC(ItemSelected)
EVENT_LC_ASC(ItemActivated)
EVENT_LC_ASC(ColClick)
//___________________________________________________________________

NodeP lKeyboardValue(PairPP)
{
	ThreadP th=Thread::This();
	return new Int(th->keyboardValue);
}

NodeP lKeyboardRawCode(PairPP)
{
	ThreadP th=Thread::This();
	if (th->keyEvent)
		return new Int(th->keyEvent->m_rawCode);
	else
		return new Int(0);
}

NodeP lKeyboardRawFlags(PairPP)
{
	ThreadP th=Thread::This();
	if (th->keyEvent)
		return new Int(th->keyEvent->m_rawFlags);
	else
		return new Int(0);
}

NodeP lSizeEventWidth(PairPP)
{
	ThreadP th=Thread::This();
	if (th->sizeEvent)
		return new Int(th->sizeEvent->GetSize().GetWidth());
	else
		return new Int(0);
}

NodeP lSizeEventHeight(PairPP)
{
	ThreadP th=Thread::This();
	if (th->sizeEvent)
		return new Int(th->sizeEvent->GetSize().GetHeight());
	else
		return new Int(0);
}

NodeP lMoveEventX(PairPP)
{
	ThreadP th=Thread::This();
	if (th->moveEvent)
		return new Int(th->moveEvent->GetPosition().x);
	else
		return new Int(0);
}

NodeP lMoveEventY(PairPP)
{
	ThreadP th=Thread::This();
	if (th->moveEvent)
		return new Int(th->moveEvent->GetPosition().y);
	else
		return new Int(0);
}

NodeP lOnChar(PairPP arg)
{
	PairP commandlist=arg[1]->a_runnable(1);
	if (!throwing)
	{	EventChar::function=commandlist;
		if (commandlist)
			File::charModeOn();
	}
	return Unbound;
}

#define LOGO_ON_HANDLER(NAME)						\
NodeP lOn##NAME(PairPP arg)						\
{															\
	PairP commandlist=arg[1]->a_runnable(1);	\
	if (!throwing)										\
		Event##NAME::function=commandlist;		\
	return Unbound;									\
}

LOGO_ON_HANDLER(KeyDown)
LOGO_ON_HANDLER(KeyUp)

#define LOGO_ON_TM(NAME) LOGO_ON_HANDLER(TextMouse##NAME)

LOGO_ON_TM(LeftDown)
LOGO_ON_TM(RightDown)
LOGO_ON_TM(MiddleDown)

LOGO_ON_TM(LeftUp)
LOGO_ON_TM(RightUp)
LOGO_ON_TM(MiddleUp)

LOGO_ON_TM(LeftDClick)
LOGO_ON_TM(RightDClick)
LOGO_ON_TM(MiddleDClick)

LOGO_ON_TM(Motion)

#define LOGO_ON_M(NAME) LOGO_ON_HANDLER(Mouse##NAME)

LOGO_ON_M(LeftDown)
LOGO_ON_M(RightDown)
LOGO_ON_M(MiddleDown)

LOGO_ON_M(LeftUp)
LOGO_ON_M(RightUp)
LOGO_ON_M(MiddleUp)

LOGO_ON_M(LeftDClick)
LOGO_ON_M(RightDClick)
LOGO_ON_M(MiddleDClick)

LOGO_ON_M(Motion)

NodeP lsendCharEvent(PairPP arg)
{
	wxWindow *win=arg[1]->a_wxwindow(1);
	int key=arg[2]->a_int(2);
	int raw=arg[3]->a_int(3);
	int flags=arg[4]->a_int(4);
	if throwing return Unbound;
	wxKeyEvent ev(wxEVT_CHAR);
	ev.m_x=0;
	ev.m_y=0;
	ev.m_keyCode=key;
	ev.m_rawCode=raw;
	ev.m_rawFlags=flags;
	ev.SetEventObject(win);
	win->GetEventHandler()->AddPendingEvent(ev);
	return Unbound;
}

NodeP lsendKeyDownEvent(PairPP arg)
{
	wxWindow *win=arg[1]->a_wxwindow(1);
	int key=arg[2]->a_int(2);
	int raw=arg[3]->a_int(3);
	int flags=arg[4]->a_int(4);
	if throwing return Unbound;
	wxKeyEvent ev(wxEVT_KEY_DOWN);
	ev.m_x=0;
	ev.m_y=0;
	ev.m_keyCode=key;
	ev.m_rawCode=raw;
	ev.m_rawFlags=flags;
	ev.SetEventObject(win);
	win->GetEventHandler()->AddPendingEvent(ev);
	return Unbound;
}

NodeP lsendKeyUpEvent(PairPP arg)
{
	wxWindow *win=arg[1]->a_wxwindow(1);
	int key=arg[2]->a_int(2);
	int raw=arg[3]->a_int(3);
	int flags=arg[4]->a_int(4);
	if throwing return Unbound;
	wxKeyEvent ev(wxEVT_KEY_UP);
	ev.m_x=0;
	ev.m_y=0;
	ev.m_keyCode=key;
	ev.m_rawCode=raw;
	ev.m_rawFlags=flags;
	ev.SetEventObject(win);
	win->GetEventHandler()->AddPendingEvent(ev);
	return Unbound;
}

NodeP lresendKeyEvent(PairPP arg)
{
	ThreadP th=Thread::This();
	wxWindow *win=arg[1]->a_wxwindow(1);
	if throwing return Unbound;
/*
	wxKeyEvent *ev=(wxKeyEvent*)th->keyEvent->Clone();
	ev->SetEventObject(win);
	ev->SetId(win->GetId());
	win->GetEventHandler()->AddPendingEvent(*ev);
*/
#ifdef WIN32
	DWORD msg;
	if (th->keyEvent->GetEventType()==wxEVT_KEY_DOWN)
		msg=WM_KEYDOWN;
	else if (th->keyEvent->GetEventType()==wxEVT_KEY_UP)
		msg=WM_KEYUP;
	else
		msg=WM_CHAR;
	SendMessage(HWND(win->GetHWND()), msg, th->keyEvent->m_rawCode, 0);
#else
// TODO
#endif
	return Unbound;
}

