//___________________________________________________________________
//
//                        Class Event and fellows
//___________________________________________________________________

// Why are those classes necessary?
// Because for the persistence there is no foreign code allowed on the stack!!!
// (look at GC_::set_C_stack, GC_::save and GC_::load !)
// Conclusion: The wx event handling mechanism can't be used 
// for events with assigned logo procedures.

extern Function 
	lKeyboardValue, lKeyboardRawCode, lKeyboardRawFlags, lresendKeyEvent,
	lMoveEventX, lMoveEventY, 
	lSizeEventWidth, lSizeEventHeight, 
	lOnChar, lOnKeyDown, lOnKeyUp, 
	lOnTextMouseLeftDown  , lOnTextMouseRightDown  , lOnTextMouseMiddleDown  , 
	lOnTextMouseLeftUp    , lOnTextMouseRightUp    , lOnTextMouseMiddleUp    , 
	lOnTextMouseLeftDClick, lOnTextMouseRightDClick, lOnTextMouseMiddleDClick,
	lOnTextMouseMotion,
	lOnMouseLeftDown  , lOnMouseRightDown  , lOnMouseMiddleDown  , 
	lOnMouseLeftUp    , lOnMouseRightUp    , lOnMouseMiddleUp    , 
	lOnMouseLeftDClick, lOnMouseRightDClick, lOnMouseMiddleDClick,
	lOnMouseMotion,
	
	lsendCharEvent, lsendKeyDownEvent, lsendKeyUpEvent;

class Mouse;
class Slider;
class IntControl;
class FloatControl;
class Button;
class ToggleButton;
class CheckBox;
class RadioButton;
class TextControl;
class ListBox_;
class ChoiceBox;
class ComboBox;
class ListControl;
class wxListEvent;
class Window_;
class Frame;
class Graph;
class Event;	typedef Event *EventP;	typedef EventP *EventPP;

class EventFiFo		// is a First-in-First-out buffer class.
{
	public:
		bool allocated, _overflow;
		int size, expand_size;
		EventPP base, max, in, out;

		EventFiFo(int expand_size_) 
		{	expand_size=expand_size_;
			allocated=_overflow=false; 
			size=0; 
			base=max=in=out=0; 
			expand();
		}
		~EventFiFo();

		EventFiFo &operator<<(const EventP val);
		EventFiFo &operator>>(EventP &val);

		operator bool() {return in!=out;}
		bool overflow() {return _overflow;}
		void expand();
};

class Event
{
	public:
		enum Type
		{	NOP,
			Char, KeyDown, KeyUp,
			TextMouseLeftDown  , TextMouseRightDown  , TextMouseMiddleDown  ,
			TextMouseLeftUp    , TextMouseRightUp    , TextMouseMiddleUp    ,
			TextMouseLeftDClick, TextMouseRightDClick, TextMouseMiddleDClick,
			TextMouseMotion,
			MouseLeftDown  , MouseRightDown  , MouseMiddleDown  ,
			MouseLeftUp    , MouseRightUp    , MouseMiddleUp    ,
			MouseLeftDClick, MouseRightDClick, MouseMiddleDClick,
			MouseMotion,
			tSlider,
			tIntControl, 
			tFloatControl, 
			tButton,
			tToggleButton,
			tCheckBox,
			tRadioButton,
			TextControlChange, TextControlEnter, 
			TextControlChar, TextControlKeyDown, TextControlKeyUp,
			ListBoxSelect, ListBoxDClick,
			ListBoxChar,	ListBoxKeyDown, ListBoxKeyUp,
			ChoiceBoxSelect, ChoiceBoxDClick,
			ChoiceBoxChar,	ChoiceBoxKeyDown, ChoiceBoxKeyUp,
			ComboBoxSelect, ComboBoxChange, ComboBoxEnter, 
			ComboBoxChar,	ComboBoxKeyDown, ComboBoxKeyUp,
			ListControlItemSelected, ListControlItemActivated,
			ListControlColClick,
			ListControlChar,	ListControlKeyDown, ListControlKeyUp,

			WindowClose, WindowPaint, WindowErase, WindowSetFocus, WindowKillFocus, 
			WindowMove, WindowSize, 
			WindowChar, WindowKeyDown, WindowKeyUp, 
			WindowLeftDown  , WindowRightDown  , WindowMiddleDown  ,
			WindowLeftUp    , WindowRightUp    , WindowMiddleUp    ,
			WindowLeftDClick, WindowRightDClick, WindowMiddleDClick,
			WindowMotion, WindowMouseWheel, WindowMouse,
			WindowEnterWindow, WindowLeaveWindow,

			FrameChar, FrameKeyDown, FrameKeyUp, FrameClose,
			GraphChar, GraphKeyDown, GraphKeyUp,
			GraphLeftDown  , GraphRightDown  , GraphMiddleDown  ,
			GraphLeftUp    , GraphRightUp    , GraphMiddleUp    ,
			GraphLeftDClick, GraphRightDClick, GraphMiddleDClick,
			GraphMotion,
		};

		Type type;

		Event(): type(NOP) {}
		Event(Type type_): type(type_) {}
		virtual ~Event() {}
		
		virtual void process()=0;
};

#define EVENT_CLASS_KEY(NAME)							\
class Event##NAME: public Event						\
{																\
	public:													\
		wxKeyEvent *e;										\
		Event##NAME(wxKeyEvent e_):					\
			Event(NAME)										\
		{	e=(wxKeyEvent*)e_.Clone();					\
		}														\
		~Event##NAME(){delete e;}				\
		void process();									\
																\
		static PairP function;							\
};

EVENT_CLASS_KEY(Char)
EVENT_CLASS_KEY(KeyDown)
EVENT_CLASS_KEY(KeyUp)


class EventWindow: public Event
{
	public:
		Window_ *window;
		EventWindow(Window_ *window_, Type type_):
			Event(type_)
		{	window=window_;
		}
};

#define EVENT_CLASS_WINDOW(NAME, EVENT)						\
class EventWindow##NAME: public EventWindow					\
{																			\
	public:																\
		wx##EVENT *e;													\
		EventWindow##NAME(Window_ *window_, wx##EVENT e_):	\
			EventWindow(window_, Window##NAME)					\
		{	e=(wx##EVENT*)e_.Clone();								\
		}																	\
		~EventWindow##NAME(){delete e;}							\
		void process();												\
};	

EVENT_CLASS_WINDOW(Close, CloseEvent)
EVENT_CLASS_WINDOW(Paint, PaintEvent)
EVENT_CLASS_WINDOW(Erase, EraseEvent)
EVENT_CLASS_WINDOW(SetFocus, FocusEvent)
EVENT_CLASS_WINDOW(KillFocus, FocusEvent)
EVENT_CLASS_WINDOW(Move, MoveEvent)
EVENT_CLASS_WINDOW(Size, SizeEvent)
EVENT_CLASS_WINDOW(Char, KeyEvent)
EVENT_CLASS_WINDOW(KeyDown, KeyEvent)
EVENT_CLASS_WINDOW(KeyUp, KeyEvent)
EVENT_CLASS_WINDOW(LeftDown, MouseEvent)
EVENT_CLASS_WINDOW(RightDown, MouseEvent)
EVENT_CLASS_WINDOW(MiddleDown, MouseEvent)
EVENT_CLASS_WINDOW(LeftUp, MouseEvent)
EVENT_CLASS_WINDOW(RightUp, MouseEvent)
EVENT_CLASS_WINDOW(MiddleUp, MouseEvent)
EVENT_CLASS_WINDOW(LeftDClick, MouseEvent)
EVENT_CLASS_WINDOW(RightDClick, MouseEvent)
EVENT_CLASS_WINDOW(MiddleDClick, MouseEvent)
EVENT_CLASS_WINDOW(Motion, MouseEvent)
EVENT_CLASS_WINDOW(Mouse, MouseEvent)
EVENT_CLASS_WINDOW(EnterWindow, MouseEvent)
EVENT_CLASS_WINDOW(LeaveWindow, MouseEvent)
EVENT_CLASS_WINDOW(MouseWheel, MouseEvent)


class EventFrame: public Event
{
	public:
		Frame *frame;
		EventFrame(Frame *frame_, Type type_):
			Event(type_)
		{	frame=frame_;
		}
};

#define EVENT_CLASS_FRAME_KEY(NAME)					\
class EventFrame##NAME: public EventFrame			\
{																\
	public:													\
		wxKeyEvent *e;										\
		EventFrame##NAME(Frame *frame_, wxKeyEvent e_):	\
			EventFrame(frame_, Frame##NAME)			\
		{	e=(wxKeyEvent*)e_.Clone();					\
		}														\
		~EventFrame##NAME(){delete e;}				\
		void process();									\
};

EVENT_CLASS_FRAME_KEY(Char)
EVENT_CLASS_FRAME_KEY(KeyDown)
EVENT_CLASS_FRAME_KEY(KeyUp)

class EventFrameClose: public EventFrame
{
	public:
		EventFrameClose(Frame *frame_):
			EventFrame(frame_, FrameClose)
		{}
		void process();
};

class EventGraph: public Event
{
	public:
		Graph *graph;
		EventGraph(Graph *graph_, Type type_):
			Event(type_)
		{	graph=graph_;
		}
};

#define EVENT_CLASS_GRAPH_KEY(NAME)					\
class EventGraph##NAME: public EventGraph			\
{																\
	public:													\
		wxKeyEvent *e;										\
		EventGraph##NAME(Graph *graph_, wxKeyEvent &e_):	\
			EventGraph(graph_, Graph##NAME)			\
		{	e=(wxKeyEvent*)e_.Clone();					\
		}														\
		~EventGraph##NAME(){delete e;}				\
		void process();									\
};

EVENT_CLASS_GRAPH_KEY(Char)
EVENT_CLASS_GRAPH_KEY(KeyDown)
EVENT_CLASS_GRAPH_KEY(KeyUp)

#define EVENT_CLASS_MOUSE(NAME)							\
class Event##NAME: public Event							\
{																	\
	public:														\
		double x,y,z;											\
		int buttons;											\
		Event##NAME(Mouse *m);								\
		void process();										\
																	\
		static PairP function;								\
};

EVENT_CLASS_MOUSE(TextMouseLeftDown)
EVENT_CLASS_MOUSE(TextMouseRightDown)
EVENT_CLASS_MOUSE(TextMouseMiddleDown)

EVENT_CLASS_MOUSE(TextMouseLeftUp)
EVENT_CLASS_MOUSE(TextMouseRightUp)
EVENT_CLASS_MOUSE(TextMouseMiddleUp)

EVENT_CLASS_MOUSE(TextMouseLeftDClick)
EVENT_CLASS_MOUSE(TextMouseRightDClick)
EVENT_CLASS_MOUSE(TextMouseMiddleDClick)

EVENT_CLASS_MOUSE(TextMouseMotion)


EVENT_CLASS_MOUSE(MouseLeftDown)
EVENT_CLASS_MOUSE(MouseRightDown)
EVENT_CLASS_MOUSE(MouseMiddleDown)

EVENT_CLASS_MOUSE(MouseLeftUp)
EVENT_CLASS_MOUSE(MouseRightUp)
EVENT_CLASS_MOUSE(MouseMiddleUp)

EVENT_CLASS_MOUSE(MouseLeftDClick)
EVENT_CLASS_MOUSE(MouseRightDClick)
EVENT_CLASS_MOUSE(MouseMiddleDClick)

EVENT_CLASS_MOUSE(MouseMotion)

#define EVENT_CLASS_GRAPH_MOUSE(NAME)					\
class EventGraph##NAME: public EventGraph				\
{																	\
	public:														\
		double x,y,z;											\
		int buttons;											\
		EventGraph##NAME(Graph *graph_, Mouse *m);		\
		void process();										\
};

EVENT_CLASS_GRAPH_MOUSE(LeftDown)
EVENT_CLASS_GRAPH_MOUSE(RightDown)
EVENT_CLASS_GRAPH_MOUSE(MiddleDown)
EVENT_CLASS_GRAPH_MOUSE(LeftUp)
EVENT_CLASS_GRAPH_MOUSE(RightUp)
EVENT_CLASS_GRAPH_MOUSE(MiddleUp)
EVENT_CLASS_GRAPH_MOUSE(LeftDClick)
EVENT_CLASS_GRAPH_MOUSE(RightDClick)
EVENT_CLASS_GRAPH_MOUSE(MiddleDClick)
EVENT_CLASS_GRAPH_MOUSE(Motion)

class EventSlider: public Event
{
	public:
		Slider *slider;
		EventSlider(Slider *slider_);
		void process();
};

class EventIntControl: public Event
{
	public:
		IntControl *intControl;
		EventIntControl(IntControl *intControl_);
		void process();
};

class EventFloatControl: public Event
{
	public:
		FloatControl *floatControl;
		EventFloatControl(FloatControl *floatControl_);
		void process();
};

class EventButton: public Event
{
	public:
		Button *button;
		EventButton(Button *button_);
		void process();
};

class EventToggleButton: public Event
{
	public:
		ToggleButton *toggleButton;
		EventToggleButton(ToggleButton *toggleButton_);
		void process();
};

class EventCheckBox: public Event
{
	public:
		CheckBox *checkBox;
		EventCheckBox(CheckBox *checkBox_);
		void process();
};

class EventRadioButton: public Event
{
	public:
		RadioButton *radioButton;
		EventRadioButton(RadioButton *radioButton_);
		void process();
};

class EventText: public Event
{
	public:
		TextControl *textControl;
		EventText(TextControl *textControl_, Type type_):
			Event(type_)
		{	textControl=textControl_;
		}
};

#define EVENT_CLASS_TEXT_KEY(NAME)								\
class EventText##NAME: public EventText						\
{																			\
	public:																\
		wxKeyEvent *e;													\
		EventText##NAME(TextControl *textControl_, wxKeyEvent &e_):\
			EventText(textControl_, TextControl##NAME)		\
		{	e=(wxKeyEvent*)e_.Clone();								\
		}																	\
		~EventText##NAME(){delete e;}							\
		void process();												\
};

EVENT_CLASS_TEXT_KEY(Char)
EVENT_CLASS_TEXT_KEY(KeyDown)
EVENT_CLASS_TEXT_KEY(KeyUp)

class EventTextChange: public EventText
{
	public:
		EventTextChange(TextControl *textControl_):
			EventText(textControl_, TextControlChange)
		{}
		void process();
};

class EventTextEnter: public EventText
{
	public:
		EventTextEnter(TextControl *textControl_):
			EventText(textControl_, TextControlChange)
		{}
		void process();
};

class EventListBox: public Event
{
	public:
		ListBox_ *listBox;
		EventListBox(ListBox_ *listBox_, Type type_);
};

#define EVENT_CLASS_LISTBOX_KEY(NAME)							\
class EventListBox##NAME: public EventListBox				\
{																			\
	public:																\
		wxKeyEvent *e;													\
		EventListBox##NAME(ListBox_ *listBox_, wxKeyEvent e_):		\
			EventListBox(listBox_, ListBox##NAME)				\
		{	e=(wxKeyEvent*)e_.Clone();					\
		}														\
		~EventListBox##NAME(){delete e;}				\
		void process();												\
};

EVENT_CLASS_LISTBOX_KEY(Char)
EVENT_CLASS_LISTBOX_KEY(KeyDown)
EVENT_CLASS_LISTBOX_KEY(KeyUp)

class EventListBoxSelect: public EventListBox
{
	public:
		EventListBoxSelect(ListBox_ *listBox_):
			EventListBox(listBox_, ListBoxSelect)
		{}
		void process();
};

class EventListBoxDClick: public EventListBox
{
	public:
		EventListBoxDClick(ListBox_ *listBox_):
			EventListBox(listBox_, ListBoxDClick)
		{}
		void process();
};

class EventChoiceBox: public Event
{
	public:
		ChoiceBox *choiceBox;
		EventChoiceBox(ChoiceBox *choiceBox_, Type type_):
			Event(type_)
		{	choiceBox=choiceBox_;
		}
};

#define EVENT_CLASS_CHOICEBOX_KEY(NAME)							\
class EventChoiceBox##NAME: public EventChoiceBox				\
{																				\
	public:																	\
		wxKeyEvent *e;														\
		EventChoiceBox##NAME(ChoiceBox *choiceBox_, wxKeyEvent e_):	\
			EventChoiceBox(choiceBox_, ChoiceBox##NAME)			\
		{	e=(wxKeyEvent*)e_.Clone();					\
		}														\
		~EventChoiceBox##NAME(){delete e;}				\
		void process();													\
};

EVENT_CLASS_CHOICEBOX_KEY(Char)
EVENT_CLASS_CHOICEBOX_KEY(KeyDown)
EVENT_CLASS_CHOICEBOX_KEY(KeyUp)

class EventChoiceBoxSelect: public EventChoiceBox
{
	public:
		EventChoiceBoxSelect(ChoiceBox *choiceBox_):
			EventChoiceBox(choiceBox_, ChoiceBoxSelect)
		{}
		void process();
};

class EventComboBox: public Event
{
	public:
		ComboBox *comboBox;
		EventComboBox(ComboBox *comboBox_, Type type_):
			Event(type_)
		{	comboBox=comboBox_;
		}
};

#define EVENT_CLASS_COMBOBOX_KEY(NAME)							\
class EventComboBox##NAME: public EventComboBox				\
{																			\
	public:																\
		wxKeyEvent *e;													\
		EventComboBox##NAME(ComboBox *comboBox_, wxKeyEvent e_):	\
			EventComboBox(comboBox_, ComboBox##NAME)			\
		{	e=(wxKeyEvent*)e_.Clone();					\
		}														\
		~EventComboBox##NAME(){delete e;}				\
		void process();												\
};

EVENT_CLASS_COMBOBOX_KEY(Char)
EVENT_CLASS_COMBOBOX_KEY(KeyDown)
EVENT_CLASS_COMBOBOX_KEY(KeyUp)

#define EVENT_CLASS_COMBOBOX(NAME)							\
class EventComboBox##NAME: public EventComboBox			\
{																		\
	public:															\
		EventComboBox##NAME(ComboBox *comboBox_):			\
			EventComboBox(comboBox_, ComboBox##NAME)		\
		{}																\
		void process();											\
};

EVENT_CLASS_COMBOBOX(Select)
EVENT_CLASS_COMBOBOX(Change)
EVENT_CLASS_COMBOBOX(Enter)

class EventListControl: public Event
{
	public:
		ListControl *listControl;
		EventListControl(ListControl *listControl_, Type type_):
			Event(type_)
		{	listControl=listControl_;
		}
};

#define EVENT_CLASS_LISTCONTROL_KEY(NAME)						\
class EventListControl##NAME: public EventListControl		\
{																			\
	public:																\
		wxKeyEvent *e;														\
		EventListControl##NAME(ListControl *listControl_, wxKeyEvent e_):		\
			EventListControl(listControl_, ListControl##NAME)				\
		{	e=(wxKeyEvent*)e_.Clone();					\
		}														\
		~EventListControl##NAME(){delete e;}				\
		void process();												\
};

EVENT_CLASS_LISTCONTROL_KEY(Char)
EVENT_CLASS_LISTCONTROL_KEY(KeyDown)
EVENT_CLASS_LISTCONTROL_KEY(KeyUp)

class EventListControlItemSelected: public EventListControl
{
	int row, col;
	wxString text;
	public:
		EventListControlItemSelected(ListControl *listControl_,
											  wxListEvent event);
		void process();
};

class EventListControlItemActivated: public EventListControl
{
	int row, col;
	wxString text;
	public:
		EventListControlItemActivated(ListControl *listControl_,
												wxListEvent event);
		void process();
};

class EventListControlColClick: public EventListControl
{
	int row, col;
	wxString text;
	public:
		EventListControlColClick(ListControl *listControl_,
										 wxListEvent event);
		void process();
};

