#include "stdafx.h"

#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

#include <limits>

#include "Main.h"
#include "Thread.h"
#include "File.h"
#include "String.h"
#include "Int.h"
#include "Float.h"
#include "Object.h"
#include "Pair.h"
#include "Error.h"
#include "NodePStack.h"
#include "Init.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

const int Float::digits = std::numeric_limits<double>::digits10+10;
const int Float::mantDigits = std::numeric_limits<double>::digits10;

const double 
	Float::max_ = std::numeric_limits<double>::max(),
	Float::pi = 3.141592653589793227020265931059839203954,
	Float::deg2rad = pi/180,
	Float::     epsilon = std::numeric_limits<double>::epsilon(),
	Float::         NaN = std::numeric_limits<double>::signaling_NaN(),
	Float::quietNaN = std::numeric_limits<double>::quiet_NaN();

DataInputStream &Float::read(DataInputStream &istr)
{	return istr >> float_;
}
wxDataOutputStream &Float::write(wxDataOutputStream &ostr)
{	return ostr << type << float_;
}

NodeP Float::copy(PtrToInt_ &ptrToInt)
{	return new Float(float_);
}

StringP Float::string()
{
	char format[10], s[digits];
	sprintf(format, "%%-.%ig", Thread::This()->precision);
	sprintf(s, format, float_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void Float::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, float_);
}

void Float::printBin(FILEP strm) 
{
	::printBin(strm, float_);
}

int  Float::_sizeof() {return sizeof(double);}

double Float::_min(double d) {return float_ < d ?  float_ : d;}
double Float::_max(double d) {return float_ > d ?  float_ : d;}

NodeP Float::Norm()
{	if (float_ >=0) return this;
	else return new Float(-float_);
}

NodeP Float::maxNorm()
{	if (float_ >=0) return this;
	else return new Float(-float_);
}

NodeP Float::saturateAbove(double limit) 
{
	return new Float(float_>limit? limit : float_);
}

NodeP Float::saturateBelow(double limit) 
{
	return new Float(float_<limit? limit : float_);
}

NodeP    Float::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}

NodeP    Float::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(double))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	int *bits=(intP)&float_;
	return boolnode((bits[i>>5]>>(i & 31)) & 1);
}

NodeP Float::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}

NodeP Float::reverse()
{	return string()->reverse();
}

NodeP Float::readbin(FileP file)
{
	ThreadP th=Thread::This();
	double num;
	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read double binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(double), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(double);
			if (th->rstringlen >= 0)
			{	num = *doubleP(th->rstringptr);	// reading from string
				th->rstringptr+=sizeof(double);
			}else
				error(FILE_ERROR, new String
					(_("memory block too small to read float")));
		}
	float_=num;
	return this;
}

NodeP Float::upper() {return string()->upper();}
NodeP Float::lower() {return string();}
PairP Float::list() {return string()->list();}

void Float::init() 
{
	reset_FPU();
	_FloatMax   ->setvarPermanent(new Float(max_));
	_NaN        ->setvarPermanent(new Float(NaN));
	_QuietNaN   ->setvarPermanent(new Float(quietNaN));
	_Pi         ->setvarPermanent(new Float(pi ));
	_Determinant->setvarPermanent(new Float( 0 ));
}

Float::x87_ControlWord 
	Float::x87_overflow_mask = {1,0,1,1, 0,0,0,0, 0,0,0};	// 0xD was: 0x1F

Float::x87_ControlWord 
	Float::x87_status = {1,1,1,1, 1,1,0,1, 3, 0, 1};	// should be 0x1320

void Float::reset_FPU()
{
	// for FPU=80x87:
	x87_status.precision_control = _64bit;
	x87_status.rounding_control = nearest;
#ifdef WIN32
	_asm finit
	_asm fldcw x87_status;
#else
	asm("finit;  fldcw %0": :"m" (x87_status));
#endif
}

double Float::rnd()
{
	union DoubleStruct
	{	double f;  
		struct {unsigned int lo, hi;};
	} d;
	d.lo =   Int::randomMT();
	d.hi = (Int::randomMT() & 0xfffff) | 0x3ff00000;
	return d.f-1;
}

double Float::abs(double x) {return x<0 ? -x : x;}
double Float::sqr(double x) {return x*x;}

double Float::cutError(double x) 
{
	if ((x>0 && x<epsilon)||(x<0 && x>-epsilon)) 
		return 0; 
	return x;
}

double Float::Remainder(double x, double d) 
{	if (d==0.0) 
	{	error(BAD_DATA, new Float(0));
		return Float::NaN;
	}
	return fmod(x,d);
}

double Float::mod(double x, double d) 
{
	if (d==0.0) 
	{	error(BAD_DATA, new Float(0));
		return Float::NaN;
	}
	double m=fmod(x,d);
	if ((m < 0) != (d < 0)) return m+d;
	return m;
}

NodeP lsetPrintPrecision(PairPP arg)
{
	ThreadP th=Thread::This();
	int p=arg[1]->a_int(1);
	if (!throwing)
	{	if (p > Float::digits-10)
		{	error(BAD_DATA, arg[1]);
			return Unbound;
		}
		th->precision=p;
	}
	return Unbound;
}

NodeP lreadFloatBin(PairPP) 
{
	NodeP n=new Float();
	if throwing return Unbound;
	return n->readbin(Thread::This()->Reader);
}

NodeP lsetFPUStatusFlags(PairPP arg)
{
	enum Params 
	{	_invalid_operation=1, _denormalized_operand, _zerodivide,
		_overflow, _underflow, _precision, _mask_interrupts
	};
	bool 
		invalid_operation=true,
		denormalized_operand=false,
		zerodivide=true,
		overflow=true,
		underflow=false,
		precision=false,
		mask_interrupts=false;

	if (exist(_invalid_operation))
		invalid_operation=arg[_invalid_operation]->a_bool(_invalid_operation)!=0;
	if (exist(_denormalized_operand))
		denormalized_operand=arg[_denormalized_operand]->a_bool(_denormalized_operand)!=0;
	if (exist(_zerodivide))
		zerodivide=arg[_zerodivide]->a_bool(_zerodivide)!=0;
	if (exist(_overflow))
		overflow=arg[_overflow]->a_bool(_overflow)!=0;
	if (exist(_underflow))
		underflow=arg[_underflow]->a_bool(_underflow)!=0;
	if (exist(_precision))
		precision=arg[_precision]->a_bool(_precision)!=0;
	if (exist(_mask_interrupts))
		mask_interrupts=arg[_mask_interrupts]->a_bool(_mask_interrupts)!=0;

	if throwing return Unbound;

	// for FPU=80x87:
	Float::x87_overflow_mask.invalid_operation=invalid_operation;
	Float::x87_overflow_mask.denormalized_operand=denormalized_operand;
	Float::x87_overflow_mask.zerodivide=zerodivide;
	Float::x87_overflow_mask.overflow=overflow;
	Float::x87_overflow_mask.underflow=underflow;
	Float::x87_overflow_mask.precision=precision;
	Float::x87_overflow_mask.dummy=false;
	Float::x87_overflow_mask.mask_interrupts=mask_interrupts;
	Float::x87_overflow_mask.precision_control=0;
	Float::x87_overflow_mask.rounding_control=0;
	Float::x87_overflow_mask.infinity=0;
	// should be 0x1320 normally

	return Unbound;
}
