#include "stdafx.h"

#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

//#include <limits>

#include "Main.h"
#include "Thread.h"
#include "File.h"
#include "String.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "Float.h"
#include "Object.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "Int8Array.h"
#include "UInt8Array.h"
#include "FloatArray.h"
#include "Complex.h"
#include "ComplexArray.h"
#include "Pair.h"
#include "Error.h"
#include "Init.h"
#include "NodePStack.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

FloatArray::FloatArray(int length_, int origin_): Node(FLOATARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (doubleP) GC_::calloc(length_+1, sizeof(double));
	if (!_data) 
		return;
	_length=length_;
	flag=ObjectType(0);
}

FloatArray::FloatArray(FloatArrayP array_): Node(FLOATARRAY)
{
	_origin = array_->_origin;
	_length=0;
	_data = (doubleP) GC_::calloc(array_->_length+1, sizeof(double));
	if (!_data) 
		return;
	_length= array_->_length;
	doubleP source = array_->_data, 
		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
	flag=ObjectType(0);
}

void FloatArray::free() {GC_::free(_data); _data=0;}

DataInputStream &FloatArray::read(DataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (doubleP) GC_::calloc(sizeof(double), length+1);
	if (!_data)
		return istr;
	_length=length;
	doubleP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (double&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &FloatArray::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	doubleP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << *itemptr++;
	return ostr;
}

NodeP FloatArray::copy(PtrToInt_ &ptrToInt)
{
	FloatArrayP arr=new FloatArray(this);
	return arr;
}

StringP FloatArray::string()
{
 	const int maxlength=65536;
	char textbuf[maxlength];
	Thread::This()->stringptr=textbuf;
	Thread::This()->stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*(Thread::This()->stringptr)='\0';
	return new String(textbuf);
}

void FloatArray::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	int i=0, wid=0;
	doubleP pp=data();
						
	if (width >= 2*_length-10)
		wid=width/_length;

	::print(strm, width, '{');
	while (i < _length) 
	{	int w=wid;
		::print(strm, w, *pp++);	
		if (++i < length())
			::print(strm, w, ' ');
		width-=wid-w;
	}
	if (i < length()) 
		::print(strm, width, "...");
	::print(strm, width, '}');

	if (origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", origin());
		::print(strm, width, org);
	}
}

void FloatArray::printBin(FILEP strm)
{
	ThreadP th=Thread::This();
	int wid=length();
	doubleP pp=data();
	if (!strm)
	{	if (th->stringlen > 0)     // printing to string
			th->stringlen-=sizeof(double)*wid;
		if (th->stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*doubleP(th->stringptr) = *pp++;
				th->stringptr+=sizeof(double);
			}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print floatarray to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(data());
	int width=Int::max;
		for (int l=sizeof(double)*wid; l; l--)
			::print(strm, width, *cp++);
		return;
	}
	fwrite(pp, sizeof(double), wid, strm);
}

NodeP FloatArray::contentsTree(	wxArrayString &namearr, 
											wxArrayString &valarr, 
											wxArrayPtrVoid &objarr, 
											ObjectType want, 
											ObjectType dont,
											ObjectType must, 
											int &count, int level)
{
	objarr[count]=this;
	if (!expanded())
	{	StringP str=string();
		if (str  &&  str->is.string)
		{	valarr[count]=str->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
		}
		return this;
	}
	doubleP dp=_data;
	for (int i=0; i < _length;  i++)
	{	namearr[count]=wxString(' ',level-1);
		namearr[count] << i+_origin;
		objarr[count]=this;
		double d=(*dp++);
		valarr[count].Printf("%g", d);
		incContentsArrays(namearr, valarr, objarr, count);
	}
	return this;
}

int FloatArray::count() {return length();}

int FloatArray::_sizeof() {return length()*sizeof(double);}
/*
int FloatArray::compare(NodeP nd)
{	if (nd->type != FLOATARRAY) 
		return Node::boolerror;
	if (this==nd)
		return 0;

	FloatArrayP b=FloatArrayP(nd);
	doubleP dp=_data, dpb=b->_data;
	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{	double c=(*dp++) - (*dpb++);
		if (c)
			return (c > 0 ?  1 : -1);
	}
	if (!i & !j)
		return 0;
	if (i)
		return 1;
	return -1;
}
*/
bool FloatArray::equalP(NodeP b)
{
	if (b->type != FLOATARRAY) 
		return false;
	if (this==b)
		return true;

	FloatArrayP ba=FloatArrayP(b);
	doubleP dp=_data, dpb=ba->_data;
	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
		if ((*dp++) - (*dpb++))
			return false;
	if (!i & !j)
		return true;
	return false;
}

NodeP FloatArray::first_() 
{	if (length() >= 1) 
		return new Float(*(FloatArrayP(this)->data()));
	else
		return 0;
}

NodeP FloatArray::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l, o);
	if throwing return Unbound;
	doubleP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP FloatArray::last()
{	return new Float(*(data()+length()-origin()));
}

NodeP FloatArray::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l, o);
	if throwing return Unbound;
	doubleP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP FloatArray::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l+1, o);
	if throwing return Unbound;
	doubleP source=data(), dest=result->data();
	*dest++=nd->getnumber();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP FloatArray::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l+1, o);
	if throwing return Unbound;
	doubleP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=nd->getnumber();
	return result;
}

double FloatArray::_min(double d)
{
	ThreadP th=Thread::This();
	doubleP dp=_data;
	for (int i=0;  i<_length;  i++)
	{	double n=*dp++;
		if (n < d)
		{	d=n;
			th->minimumPosition=i+_origin;
		}
	}
	return d;
}

double FloatArray::_max(double d)
{
	ThreadP th=Thread::This();
	doubleP dp=_data;
	for (int i=0;  i<_length;  i++)
	{	double n=*dp++;
		if (n > d)
		{	d=n;
			th->maximumPosition=i+_origin;
		}
	}
	return d;
}

NodeP FloatArray::Norm()
{	double d=0;  doubleP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++);
		d+= h*h;
	}
	return new Float(::sqrt(d));
}

NodeP FloatArray::maxNorm()
{	double d=0;  doubleP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++);
		double ha=Float::abs(h);
		if (ha > d) d = ha;
	}
	return new Float(d);
}

NodeP FloatArray::Norm3()
{	int len=_length/3;
	FloatArrayP result=new FloatArray(len);
	if throwing return Unbound;
	doubleP d=result->data();
	doubleP s=_data;
	for (int i=len;  i>0;  i--)
	{	double x=(*s++);
		double y=(*s++);
		double z=(*s++);
		*d++=sqrt(x*x+y*y+z*z);
	}
	return result;
}

NodeP FloatArray::resize(int newsize)
{	FloatArrayP r=new FloatArray(newsize);
	if throwing return Unbound;
	doubleP d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	double diff = d[i+1]-d[i];
					double slope = diff/di;


					for (int ii=di-1;  ii;  ii--)
						r->_data[ni+ii] = d[i]+slope*ii;
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f,
					diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=d[int(i)]+frac*diff;
			}
		}
	}
	return r;
}

NodeP FloatArray::lowPassFilter(double weight)
{
	FloatArrayP result=new FloatArray(length());
	if throwing return Unbound;
	doubleP source=data(), dest=result->data();
	double median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=median;
	}
	Float::reset_FPU();
	return result;
}

NodeP FloatArray::lowPassFilter2(double weight)
{
	FloatArrayP result=new FloatArray(length());
	if throwing return Unbound;
	doubleP source=data(), dest=result->data();
	double median=*source;
	for (int j=4; j>0; j--)
	{	for (int i=length(); i>0; i--)
		{	median=(median*weight+(*source++))/(weight+1);
			(*dest++)+=median;
		}
		for (int i=length(); i>0; i--)
		{	median=(median*weight+(*--source))/(weight+1);
			(*--dest)+=median;
		}
	}
	Float::reset_FPU();
	return result;
}

NodeP FloatArray::saturateAbove(double limit) 
{
	FloatArrayP result=new FloatArray(length());
	if throwing return Unbound;
	doubleP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	double s=*source++;
		*dest++= s>limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP FloatArray::saturateBelow(double limit) 
{
	FloatArrayP result=new FloatArray(length());
	if throwing return Unbound;
	doubleP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	double s=*source++;
		*dest++= s<limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP FloatArray::item(int i)
{	if (i >= 0  &&  i < _length) 
		return new Float(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}

NodeP FloatArray::items(NodeP start, NodeP end)
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	FloatArrayP newar=new FloatArray(e-s+1);
					if throwing return Unbound;
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP FloatArray::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	double ni=newitem->a_float(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni;
	}
	return Unbound;
}

NodeP FloatArray::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	doubleP d=_data;   int l=_length;
	FloatP f=new Float(0);
	if throwing return Unbound;

	while (--l >= 0 && !throwing) 
	{	f->float_=(*d++);
		if (subobj->equalP(f))
			return _True;
	}
	return _False;
}

NodeP FloatArray::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					doubleP d = _data+i;
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =   IntP(h)->int_;	break;
							case FLOAT_:*d++ = FloatP(h)->float_;break;
							case NINT16:	*d++ = Int16P(h)->int_;	break;
							case NINT8:	*d++ =  Int8P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = h->getnumber();
											else
											{	error(BAD_DATA, n, 3); 
												return FloatArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);
								return FloatArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					doubleP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	int16P s=n->data();
					doubleP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT8ARRAY)
			{	Int8ArrayP n=Int8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	int8P s=n->data();
					doubleP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==UINT8ARRAY)
			{	UInt8ArrayP n=UInt8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	uint8P s=n->data();
					doubleP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)
			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==COMPLEXARRAY)
			{	ComplexArrayP n=ComplexArrayP(newitems);
				if (i+n->length() <= _length)
				{	complexdP s=n->data();
					doubleP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Complexd::abs(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}

NodeP FloatArray::reverse()
{
	FloatArrayP result=new FloatArray(_length);
	if throwing return Unbound;

	doubleP d=_data, rev=result->_data+_length-1;
	for (int i=_length; i>0; i--)
		*rev-- = *d++;
	return result;
}

NodeP FloatArray::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	FloatArrayP val = new FloatArray(length(), origin());
	if throwing return Unbound;
	doubleP
		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}

NodeP FloatArray::shuffle()
{	FloatArrayP result=new FloatArray(FloatArrayP(this));
	if throwing return Unbound;
	doubleP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		double tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}

PairP FloatArray::list()
{
	doubleP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
	{	p = new Pair(new Float(*--d), p);
		if throwing return PairP(Unbound);
	}
	return p;
}

ArrayP FloatArray::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	if throwing return ArrayP(Unbound);
	NodePP d = array->data();
	doubleP s=_data;

	for (int i=_length;  i;  i--)
	{	*d++ = new Float(*s++);
		if throwing return ArrayP(Unbound);
	}
	return array;
}

IntArrayP FloatArray::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	if throwing return IntArrayP(Unbound);
	intP d = intarray->data();
	doubleP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int::clip(*s++);

	return intarray;
}

Int16ArrayP FloatArray::int16array(int origin) 
{
	Int16ArrayP int16array = new Int16Array(_length, _origin);
	if throwing return Int16ArrayP(Unbound);
	int16P d = int16array->data();
	doubleP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int16::clip(*s++);

	return int16array;
}

FloatArrayP FloatArray::floatarray(int origin) 
{
	return new FloatArray(FloatArrayP(this));
}

ComplexArrayP FloatArray::complexarray(int origin) 
{
	ComplexArrayP complexarray = new ComplexArray(_length, _origin);
	if throwing return ComplexArrayP(Unbound);
	complexdP d = complexarray->data();
	doubleP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return complexarray;
}

void FloatArray::combSortInPlace()
{
	double gap=_length; //initialize gap size
	int swaps=0; //see bubblesort for an explanation
    
	while (!(gap<=1 && swaps==0))
	{	//update the gap value for a next comb
		if (gap > 1)
		{	gap/=1.3;
			if (gap==10 || gap==9)
				gap=11;
		}
		int i=0;
		//a single "comb" over the input list
		while (i+gap<_length)
		{	int j=int(i+gap);
			if (_data[i] > _data[j])
			{	double tmp=_data[i];
				_data[i]=_data[j];
				_data[j]=tmp;
			}
			i++;
		}
	}
}

NodeP FloatArray::combSort()
{
	FloatArrayP a=new FloatArray(this);
	if throwing return Unbound;
	a->combSortInPlace();
	return a;
}

NodeP FloatArray::readbin(FileP file)
{
	ThreadP th=Thread::This();
	doubleP ip=data();

	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read floatarray binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(double), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(double)*_length;
			if (th->rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *doubleP(th->rstringptr);	// reading from string
					th->rstringptr+=sizeof(double);
				}
			else
				error(FILE_ERROR, new String
					(_("memory block too small to read floatarray")));
		}
	return this;
}

NodeP lreadFloatArrayBin(PairPP arg) 
{
	enum Params {_size=1, _origin};
	int size=arg[_size]->a_int(_size);
	if throwing return Unbound;
	int origin=1;
	if (exist(_origin))
	{	origin=arg[_origin]->a_int(_origin);
		if throwing return Unbound;
	}
	FloatArrayP a=new FloatArray(size, origin);
	if throwing return Unbound;
	return a->readbin(Thread::This()->Reader);
}
