#include "stdafx.h"
//___________________________________________________________________
//
//                           Class  Frame
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/textctrl.h>
#include <wx/sizer.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Table.h"
#include "LogoObject.h"
#include "Int.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "BoxSizer.h"
#include "Point.h"
#include "Graph.h"
#include "Frame.h"
#include "Window.h"
#include "UInt8.h"
#include "Color.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"

#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Frame, wxFrame)
	EVT_CHAR		(Frame::OnChar		)
	EVT_KEY_DOWN(Frame::OnKeyDown	)
	EVT_KEY_UP	(Frame::OnKeyUp	)
	EVT_CLOSE	(Frame::OnClose	)
END_EVENT_TABLE()

Frame::Frame(FrameNodeP frameNode_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=frameNode_;

	thread=Thread::This();
	wxWindow *parent=0;
	if (WindowNodeP(*d(windowNode))->parent)
		parent=WindowNodeP(*d(windowNode))->parent->window;
	Create(parent, -1, _("Frame"),
			wxDefaultPosition, wxDefaultSize, 
			wxDEFAULT_FRAME_STYLE, _("Frame"));
}

Frame::Frame(FrameNodeP frameNode_,
			wxWindow* parent, wxWindowID id, 
			const wxString& title_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxString& name):
wxFrame(parent, id, title_, pos, size, style, name),
title(title_)
{
	ThreadP th=Thread::This();
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=frameNode_;
	thread=Thread::This();
}

Frame::~Frame()
{
	if (*d(windowNode))
		(*(WindowNodeP*)d(windowNode))->window=0;
	if (!Thread::list.Find(thread))
		return;
	EventFiFo *fifo=thread->eventFiFo;
	EventPP out=fifo->out;
	EventP ev;
	while (*fifo)
	{	*fifo >> ev;
		switch (ev->type)
		{
		case Event::FrameChar:
		case Event::FrameKeyDown:
		case Event::FrameKeyUp:
		case Event::FrameClose:
			if (((EventFrame*)ev)->frame==this)
				((EventFrame*)ev)->frame=0;
			break;
		default:
			;
		}
	}
	fifo->out=out;
}

void Frame::OnClose(wxCloseEvent &event)
{
	if (onClose)
	{	EventP evt=new EventFrameClose(this);
		if throwing return;
		*thread->eventFiFo << evt;
		event.Skip(false);// wait for the onClose handler
	}else
	{	if (*d(windowNode))
			(*(WindowNodeP*)d(windowNode))->window=0;
		delete this;
//		event.Skip();
	}
}

#define FRAME_ON(NAME)														\
void Frame::On##NAME(wxKeyEvent &event)								\
{																					\
	EventP evt=new EventFrame##NAME(this, event);	\
	if throwing return;														\
	*thread->eventFiFo << evt;												\
	event.Skip();																\
}

FRAME_ON(Char)
FRAME_ON(KeyDown)
FRAME_ON(KeyUp)

bool Frame::setShape(PairP list)
{
	if (!list)
		return true;
	int n=list->count();
	wxPoint *points=new wxPoint[n];
	if (!points)
	{	error(OUT_OF_MEM); return false;
	}
	int i=0;
	while (list)
	{	PairP p=list->first()->a_vector(2);
		if throwing 
		{	delete points;
			return false;
		}
		int x=int(p->first ()->getnumber()),
			 y=int(p->second()->getnumber());
		points[i]=wxPoint(x,y);
		i++;
		next(list);
	}
	SetShape(wxRegion(n, points));
	delete points;
	return true;
}

wxString Frame::wxstring()
{
	wxString str;
	str << _("Frame") << " [" 
		<< GetTitle() << "]["
		<< GetPosition().x << ' ' << GetPosition().y << "][" 
		<< GetSize().x << ' ' << GetSize().y << "] "
		<< GetWindowStyle() << " [" << GetName() << ']';
	return str;
}

DataInputStream &Frame::read(DataInputStream &istr) 
{	
	wxString name, fontName;

	wxPoint pos;
	wxSize size;
	int style, fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> title >> pos.x >> pos.y >> size.x >> size.y >> style >> name
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight
#define READ_ON(NAME)	>> (int&)*d(on##NAME) >> (int&)*d(uon##NAME) >> (int&)*d(ton##NAME)
		READ_ON(Char)
		READ_ON(KeyDown)
		READ_ON(KeyUp)
		READ_ON(Close)
		>> (int&)*d(shape)
		>> (int&)*d(boxSizer);

	SetTitle(title); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetName(name); 
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont font=GetFont();
	font.SetPointSize(fontSize);
	font.SetFaceName(fontName);
	font.SetStyle(fontStyle);
	font.SetWeight(fontWeight);
	SetFont(font);
	return istr;
}
wxDataOutputStream &Frame::write(wxDataOutputStream &ostr) 
{
	wxColour f=GetForegroundColour(), b=GetBackgroundColour();
	wxFont font=GetFont();
	ostr
		<< GetTitle() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle()) << GetName()
		<< f.Red() << f.Green() << f.Blue() 
		<< b.Red() << b.Green() << b.Blue() 
		<< font.GetPointSize() << font.GetFaceName() 
		<< font.GetStyle() << font.GetWeight();
#define WRITE_ON(NAME)	\
	ostr << GC_::ptrToInt->find(*d(on##NAME));		\
	ostr << GC_::ptrToInt->find(*d(uon##NAME));	\
	ostr << GC_::ptrToInt->find(*d(ton##NAME));

	WRITE_ON(Char)
	WRITE_ON(KeyDown)
	WRITE_ON(KeyUp)
	WRITE_ON(Close)

	ostr << GC_::ptrToInt->find(*d(shape))
		<< GC_::ptrToInt->find(*d(boxSizer));
	return ostr;
}

#define NODE_ON(x)						\
PairP FrameNode::on##x()				\
{												\
	if (frame)								\
	return PairP(*frame->d(Frame::on##x));\
	return 0;								\
}												\
												\
ObjectP FrameNode::uon##x()			\
{												\
	if (frame)								\
		return ObjectP(*frame->d(Frame::uon##x));		\
	return 0;								\
}												\
												\
LogoObjectP FrameNode::ton##x()		\
{												\
	if (frame)								\
		return LogoObjectP(*frame->d(Frame::ton##x));	\
	return 0;								\
}												\
												\
void FrameNode::seton##x(NodeP nd)	\
{												\
	if (frame)								\
	{	ThreadP th=Thread::This();		\
		  *frame->d(Frame::on##x) = nd;		\
		*frame->d(Frame::uon##x) = th->lexufun;	\
		*frame->d(Frame::ton##x) = th->thisobj;	\
		check_oldyoung(nd);				\
	}											\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(Close)

PairP FrameNode::shape()
{
	if (frame)
		return PairP(*frame->d(Frame::shape));
	return 0;
}

void FrameNode::realsetshape(PairP shape_)
{
	if (frame && frame->setShape(shape_))
		setshape(shape_);
}

void FrameNode::setshape(PairP shape_)
{
	if (frame)
	{	*frame->d(Frame::shape)=shape_;
		check_oldyoung(shape_);
	}
}

BoxSizerNodeP FrameNode::boxSizer()
{
	if (frame)
		return BoxSizerNodeP(*frame->d(Frame::boxSizer));
	return 0;
}

void FrameNode::setSizer(BoxSizerNodeP boxSizer_)
{
	if ((frame!=0) && boxSizer_)
	{	*frame->d(Frame::boxSizer)=boxSizer_;
		check_oldyoung(boxSizer_);
		frame->SetSizer(boxSizer_->boxSizer);
		frame->Layout();
	}
}

FrameNode::FrameNode(): 
	WindowNode(FRAME), frame((FrameP&)window)
{}

FrameNode::FrameNode(WindowNodeP parent_): 
	WindowNode(FRAME), frame((FrameP&)window)
{
	WindowNodeP p=parent_;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void FrameNode::free() 
{	
	if (frame) 
	{	
//		GUI_ENTER
		*frame->d(Frame::windowNode)=0;
		WindowPairP c=children;
		while (c)
		{	if (c->first())
				c->first()->setParent(WindowNodeP(0));
			c=WindowPairP(c->tail());
		}				
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		frame->Destroy();
		if (frame->thread)
		{	EventFiFo *fifo=frame->thread->eventFiFo;
			EventPP out=fifo->out;
			EventP ev;
			while (*fifo)
			{	*fifo >> ev;
				switch (ev->type)
				{
				case Event::FrameChar:
				case Event::FrameKeyDown:
				case Event::FrameKeyUp:
				case Event::FrameClose:
					if (((EventWindow*)ev)->window==window)
						((EventWindow*)ev)->window=0;
					break;
				default:
					;
				}
			}
			fifo->out=out;
		}
		frame=0;
//		GUI_LEAVE
	}
}
void FrameNode::inc_gc() 
{	fifo << children 
		<< onChar() << onKeyDown() << onKeyUp() << onClose() 
		<< uonChar() << uonKeyDown() << uonKeyUp() << uonClose() 
		<< tonChar() << tonKeyDown() << tonKeyUp() << tonClose() 
		<< shape();
}
int FrameNode::youngs() 
{	return gmark(children)
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp())
		+gmark(uonChar())+gmark(uonKeyDown())+gmark(uonKeyUp())
		+gmark(tonChar())+gmark(tonKeyDown())+gmark(tonKeyUp())
		+gmark(onClose())
		+gmark(shape());
}
int FrameNode::youngsPtr(PtrToInt_ &ptrToInt)
{	return 
		 gmarkPtr(children, ptrToInt)
		+gmarkPtr(onChar(), ptrToInt)
		+gmarkPtr(onKeyDown(), ptrToInt)
		+gmarkPtr(onKeyUp(), ptrToInt)
		+gmarkPtr(onClose(), ptrToInt)
		+gmarkPtr(uonChar(), ptrToInt)
		+gmarkPtr(uonKeyDown(), ptrToInt)
		+gmarkPtr(uonKeyUp(), ptrToInt)
		+gmarkPtr(uonClose(), ptrToInt)
		+gmarkPtr(tonChar(), ptrToInt)
		+gmarkPtr(tonKeyDown(), ptrToInt)
		+gmarkPtr(tonKeyUp(), ptrToInt)
		+gmarkPtr(tonClose(), ptrToInt)
		+gmarkPtr(shape(), ptrToInt);
}
void FrameNode::check_oldyoungs() 
{	check_oldyoung(children);
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
	check_oldyoung(onClose());
	check_oldyoung(uonChar());
	check_oldyoung(uonKeyDown());
	check_oldyoung(uonKeyUp());
	check_oldyoung(uonClose());
	check_oldyoung(tonChar());
	check_oldyoung(tonKeyDown());
	check_oldyoung(tonKeyUp());
	check_oldyoung(tonClose());
	check_oldyoung(shape());
}

DataInputStream &FrameNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr 
			>> (int&)parentNode 
			>> (int&)children;
		frame=new Frame(FrameNodeP(this));
		if (!frame)
		{	error(OUT_OF_MEM); return istr;
		}
		frame->read(istr);
		parent=parentNode;
	}
	return istr;
}
wxDataOutputStream &FrameNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (frame)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent)
			<< GC_::ptrToInt->find(children);
		frame->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void FrameNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	children = (WindowPair*)intToPtr.find(int(children));
	check_oldyoung(children);
	if (frame)
	{
#define SETON(NAME)	\
	*frame->d(Frame:: on##NAME)=((      Pair*)intToPtr.find(int(*frame->d(Frame:: on##NAME))));	\
	*frame->d(Frame::uon##NAME)=((    Object*)intToPtr.find(int(*frame->d(Frame::uon##NAME))));	\
	*frame->d(Frame::ton##NAME)=((LogoObject*)intToPtr.find(int(*frame->d(Frame::ton##NAME))));

		SETON(Char)
		SETON(KeyDown)
		SETON(KeyUp)
		SETON(Close)

		setshape		(PairP((Pair*)intToPtr.find(int(frame->shape	   ))));
		*frame->d(Frame::boxSizer) = (BoxSizerNode*)intToPtr.find(int(frame->boxSizer));
		frame->Show();
	}
}

void FrameNode::aftersetPtrs()
{
	if (parent)
		return;

	recreate();
}

void FrameNode::recreate()
{
	FrameP f=FrameP(frame), f2=0;
	if (!frame)
		return;
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	if (!f) return;
	f2=new Frame(FrameNodeP(this),
			p, -1, f->GetTitle(), f->GetPosition(), 
			f->GetSize(), f->GetWindowStyle(), f->GetName());
	if (!f2)
	{	error(OUT_OF_MEM); return;
	}
	f2->SetForegroundColour(f->GetForegroundColour());
	f2->SetBackgroundColour(f->GetBackgroundColour());
	wxFont font=f->GetFont();
	f2->SetFont(font);
	frame=f2;
#define SET_ON(NAME) \
	*frame->d(Frame:: on##NAME)=(*f->d(Frame:: on##NAME));	\
	*frame->d(Frame::uon##NAME)=(*f->d(Frame::uon##NAME));	\
	*frame->d(Frame::ton##NAME)=(*f->d(Frame::ton##NAME));	

		SET_ON(Char)
		SET_ON(KeyDown)
		SET_ON(KeyUp)
		SET_ON(Close)

	realsetshape(PairP(*f->d(Frame::shape)));
	*frame->d(Frame::boxSizer)=*f->d(Frame::boxSizer);
	delete f;

	frame=f2;
	frame->Show();

	PairP c=children;
	while (c)
	{	if (c->first())
			c->first()->recreate();
		next(c);
	}
}

void FrameNode::aftersetPtrs3()
{
	if (frame)
		setSizer(BoxSizerNodeP(*frame->d(Frame::boxSizer)));
}

StringP FrameNode::string()
{
	if (frame)
		return new String(frame->wxstring());
	return _Null_Word->name();
}

wxString FrameNode::wxstring()
{
	if (frame)
		return frame->wxstring();
	return wxEmptyString;
}

void FrameNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lFrame(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString title=_("Frame");
	if (exist(2))
	{	StringP t=arg[2]->a_string(2);
		if throwing return Unbound;
		title=t->ptr();
		title.Truncate(t->length());
	}
	long style=wxDEFAULT_FRAME_STYLE;
	if (exist(3) && arg[3])
	{	int s=arg[3]->a_int(3);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(4) && arg[4])
	{	PairP p=arg[4]->a_int_vector(4);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(5) && arg[5])
	{	PairP s=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=title;
	if (exist(6))
	{	StringP n=arg[6]->a_string(6);
		if throwing return Unbound;
		name=n->ptr();
		name.Truncate(n->length());
	}
//	GUI_ENTER
	FrameNodeP frameNode=new FrameNode(parentNode);
	if throwing return Unbound;
	FrameP frame=new Frame(frameNode,
			parent, id, title, pos, size, style, name);
	if (!frame)
	{	error(OUT_OF_MEM); return Unbound;
	}
	frame->Show();
	frameNode->frame=frame;
#ifndef WIN32
	th->dispatchMessages();
#endif
//	GUI_LEAVE
	return frameNode;
}

NodeP lFrameDestroy(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;

//	GUI_ENTER
	WindowPairP c=frame->children;
	while (c)
	{	if (c->first())
			c->first()->setParent(WindowNodeP(0));
		c=WindowPairP(c->tail());
	}				
	{	WindowNodeP p=frame->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(frame));
			p->check_oldyoung(p->children);
		}
	}
	frame->frame->Destroy();
	frame->frame=0;
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetFocus(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
//	GUI_ENTER
	frame->frame->SetFocus();
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameEnable(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	frame->frame->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameMaximize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	FrameP(frame->frame)->Maximize(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameIconize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	FrameP(frame->frame)->Iconize(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameFullScreen(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	int style=arg[3]->a_int(3);
	if throwing return Unbound;
//	GUI_ENTER
	FrameP(frame->frame)->ShowFullScreen(state!=0, style);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetClientSize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int w=arg[2]->a_int(2);
	if throwing return Unbound;
	int h=arg[3]->a_int(3);
	if throwing return Unbound;
//	GUI_ENTER
	frame->frame->SetClientSize(w,h);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetBackgroundColor(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	FrameP(frame->frame)->SetBackgroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetColor(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	frame->frame->SetForegroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetFontSize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=frame->frame->GetFont();
	f.SetPointSize(s);
	frame->frame->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetFontName(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=frame->frame->GetFont();
	f.SetFaceName(s->ptr());
	frame->frame->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetFontStyle(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=frame->frame->GetFont();
	f.SetStyle(s);
	frame->frame->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetFontWeight(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=frame->frame->GetFont();
	f.SetWeight(s);
	frame->frame->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetShape(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
//	GUI_ENTER
	frame->realsetshape(list);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFrameSetSizer(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	BoxSizerNodeP boxSizer=arg[2]->a_boxSizer(2);
	if throwing return Unbound;
//	GUI_ENTER
	frame->setSizer(boxSizer);
//	GUI_LEAVE
	return Unbound;
}

#define _FRAME_ON(NAME)								\
NodeP lFrameOn##NAME(PairPP arg)					\
{															\
	FrameNodeP frame=arg[1]->a_frame(1);		\
	if throwing return Unbound;					\
	PairP commandlist=arg[2]->a_runnable(2);	\
	if throwing return Unbound;					\
	frame->seton##NAME(commandlist);				\
	return Unbound;									\
}

_FRAME_ON(Char)
_FRAME_ON(KeyDown)
_FRAME_ON(KeyUp)
_FRAME_ON(Close)
