//___________________________________________________________________
//
//                        Class Graph
//___________________________________________________________________

#ifndef WINDOWNODE
	#include "Window.h"
#endif

extern Function 
	lGraph, lGraphDestroy, 
	lGraphCurrent, lGraphSetCurrent, lGraphSetFocus,
	lGraphOnChar, lGraphOnKeyDown, lGraphOnKeyUp,
	lGraphOnLeftDown  , lGraphOnRightDown  , lGraphOnMiddleDown  ,
	lGraphOnLeftUp    , lGraphOnRightUp    , lGraphOnMiddleUp    ,
	lGraphOnLeftDClick, lGraphOnRightDClick, lGraphOnMiddleDClick,
	lGraphOnMotion,
	lsetPlayOnSwap, lMousePos, lMouseX, lMouseY, lMouseZ, lMouseButtons;

class Mouse;
class Screen_;
class Turtle;
class Recorder;

class Console;
class AVIGen;
class ShadowObject;
class FTGLFontManager;
class DispatchMutex;

#ifdef OPENGL
	#include <wx/glcanvas.h>
#endif

class Graph 
#ifdef OPENGL
	: public wxGLCanvas
#else
	: public wxWindow
#endif
{
	public:
		bool direct, printing, saving, savingToObj,
			doubleBuffer, playOnSwap, 
			was1, autosave, inread, inSwapBuffers, inOnPaint, inOnSize;
		int active;
		float fRadius, colorTolerance;
		wxSize saveSize, size;
		Point winorigin;
		wxString pictureFileName;
		Screen_ *screen;
		Turtle *turtle;
		Recorder *recorder, *recorderSave;
		Mouse *mouse;
		AVIGen *aviGen;
		ShadowObject *shadowObject;
		FTGLFontManager *ftglFontManager;

		ThreadP thread;
		int graphOwned;
		int waitingForGraph;
		DispatchMutex *csGraph, *csWaitingForGraph;

		float nearPlane, farPlane;
		wxRect oldRect;
#ifdef WIN32
		wxSize m_szPage, saveWinsize;
		BITMAPINFO m_bmi;
		HBITMAP m_hDib, m_hOldDib;
		void *m_pBitmapBits;
		HDC m_hMemDC, m_hOldDC;
		HGLRC m_hMemRC, m_hOldRC;
#endif
		wxPalette palette, *oldPalette;

#ifndef OPENGL
		wxPen pen;
		wxBrush brush;
		wxBitmap *bitmap;
		wxDC *dc;
		wxMemoryDC *mdc;
#endif

		enum Data
		{	WINDOW_DATA,
			turtleNode, 
			dataCount
		};
		NodeP data[dataCount];
/*		PairP 
			onChar, onKeyUp, onKeyDown, 
			onMouseLeftDown  , onMouseRightDown  , onMouseMiddleDown  ,
			onMouseLeftUp    , onMouseRightUp    , onMouseMiddleUp    ,
			onMouseLeftDClick, onMouseRightDClick, onMouseMiddleDClick,
			onMouseMotion;
		ObjectP 
			uonChar, uonKeyUp, uonKeyDown, 
			uonMouseLeftDown  , uonMouseRightDown  , uonMouseMiddleDown  ,
			uonMouseLeftUp    , uonMouseRightUp    , uonMouseMiddleUp    ,
			uonMouseLeftDClick, uonMouseRightDClick, uonMouseMiddleDClick,
			uonMouseMotion;
		LogoObjectP 
			tonChar, tonKeyUp, tonKeyDown, 
			tonMouseLeftDown  , tonMouseRightDown  , tonMouseMiddleDown  ,
			tonMouseLeftUp    , tonMouseRightUp    , tonMouseMiddleUp    ,
			tonMouseLeftDClick, tonMouseRightDClick, tonMouseMiddleDClick,
			tonMouseMotion;
		GraphNodeP graphNode;
		TurtleNodeP turtleNode;
*/
		static int attribs[];

		static int vmt;

		virtual NodePP d(int i) {return &data[i];}

		Graph(wxWindow *parent);
		Graph(GraphNodeP graphNode_);
		Graph(GraphNodeP graphNode_,
			wxWindow* parent, wxWindowID id, 
			const wxPoint& pos, const wxSize& size, 
			long style=0, const wxString& name="GLCanvas", 
			int* attribList = 0, const wxPalette& palette = wxNullPalette,
			bool pictureAutosave = true);

		~Graph();

		void create(wxWindow* parent, wxWindowID id, 
			const wxPoint& pos, const wxSize& size, 
			long style=0, const wxString& name="GLCanvas", 
			int* attribList = 0, const wxPalette& palette = wxNullPalette,
			bool pictureAutosave = true);
		void create2();
		void reset();

		void enter();
		void leave();
		bool tryEnter();

		wxString wxstring();
		void setPtrs(IntToPtr_ &intToPtr);
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		void 
			show(bool shown=true),
			OnSize(wxSizeEvent &event),
			OnChar(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp(wxKeyEvent &event),
			OnPaint(wxPaintEvent &event),
			OnMove (wxMoveEvent  &event),
			OnErase(wxEraseEvent &event),
			drawTurtles(bool incNr=true),
			swapBuffers(),
			setFrustum(double w, double h),
			setOpenGLState(),
			clear(),
			clearScreen(),
			screen_reset(int centerP),
			screen_clear(int centerp),
			setColor(Color c),
			updateGraph(),
			autoload(), save(), 
			saveVector(wxString filename,
						  int sort, int options, int colornr),
			update(), 
			print(wxDC* dc, wxPoint displacement),
			videoStart(charP filename, int rate),
			videoFrame(),
			videoEnd();

		NodeP saveScreen(charP filename);
	
#ifdef OPENGL
		wxGLContext *glContext() {return m_glContext;}
#endif
		void 
			OnDestroy(),
			OnSetFocus				(wxFocusEvent &event),
			OnEnterWindow			(wxMouseEvent &event),
			OnLeaveWindow			(wxMouseEvent &event),
			setMouse					(wxMouseEvent &event, Mouse *mouse),
			OnMouse					(wxMouseEvent& event),
			OnMouseLeftDown		(wxMouseEvent &event),
			OnMouseRightDown		(wxMouseEvent &event),
			OnMouseMiddleDown		(wxMouseEvent &event),
			OnMouseLeftUp  		(wxMouseEvent &event),
			OnMouseRightUp  		(wxMouseEvent &event),
			OnMouseMiddleUp  		(wxMouseEvent &event),
			OnMouseLeftDClick		(wxMouseEvent &event),
			OnMouseRightDClick	(wxMouseEvent &event),
			OnMouseMiddleDClick	(wxMouseEvent &event),
			OnMouseMotion			(wxMouseEvent &event),

			beginPrinting(wxDC* pDC),
			endPrinting(wxDC* pDC),
			beginSaving(),
			endSaving();

		bool 
			OnMouseWheel(int nFlags, int16 zDelta, wxPoint pt),
			OnPreparePrinting();

#ifdef WIN32
		bool setDCPixelFormat(HDC hDC, DWORD dwFlags);
#endif

//		virtual void OnPrint(CDC* pDC, CPrintInfo* pInfo);
//		virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);

		DECLARE_EVENT_TABLE()
};
typedef Graph *GraphP;

#define TURTLE		(th->graph->turtle)
#define SCREEN		(th->graph->screen)
#define MEMSCREEN	(th->graph->memscreen)
#define RECORDER	(*(th->graph->recorder))

class GraphNode: public WindowNode
{
	public:
		GraphP &graph;
		PairP textures;

		static const GraphNode aGraphNode;

		operator_new;

		TurtleNodeP turtleNode();
		void setturtleNode(TurtleNodeP tn);

		GraphNode();
		GraphNode(WindowNodeP parent_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr(PtrToInt_ &ptrToInt);
		void check_oldyoungs();
		
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);
		void aftersetPtrs();
		void recreate();

		StringP string();
		wxString wxstring();
		void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);

		int count();

		StringP _typeof();
};

class GraphLocker
{
public:
	GraphLocker();
	~GraphLocker();
};
