#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "String.h"
#include "File.h"
#include "Object.h"
#include "Init.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

const int 
	Int::max = unsigned(-1) >> 1,
	Int::min = -Int::max-1,
	Int::umax = unsigned(-1),
	Int::digits = sizeof(int)*256/100 + 3;

const __int64 Int::i64max = ((unsigned __int64)(-1)) >> 1;


int volatile Int::genSymNr = 0;

DataInputStream &Int::read(DataInputStream &istr)
{	return istr >> int_;
}
wxDataOutputStream &Int::write(wxDataOutputStream &ostr)
{	return ostr << type << int_;
}

NodeP Int::copy(PtrToInt_ &ptrToInt)
{	return new Int(int_);
}

StringP Int::string()
{
	char s[digits];
	sprintf(s, "%i", int_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void Int::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, int_);
}

void Int::printBin(FILEP strm) 
{
	::printBin(strm, int_);
}

int  Int::_sizeof() {return sizeof(int);}

double Int::_min(double d) {return int_ < d ?  int_ : d;}
double Int::_max(double d) {return int_ > d ?  int_ : d;}

NodeP Int::Norm()
{	if (int_ >=0) return this;

	else return new Int(-int_);
}

NodeP Int::maxNorm()
{	if (int_ >=0) return this;
	else return new Int(-int_);
}

NodeP Int::saturateAbove(double limit) 
{
	return new Int(int_>limit? int(limit) : int_);
}

NodeP Int::saturateBelow(double limit) 
{
	return new Int(int_<limit? int(limit) : int_);
}

NodeP Int::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}

NodeP Int::bititem(NodeP index)
{
	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(int))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return boolnode((int_>>i) & 1);
}

NodeP Int::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (newitem && newitem->type==OBJECT)
	{	int bit=newitem->a_bool(3);
		if (!throwing)
		{	if (i<0 || i >= int(sizeof(int))<<3)
			{	error(BAD_DATA, index);
				return Unbound;
			}
			if (bit)
				int_|=(1<<i);
			else
				int_&=~(1<<i);
		}
	}
	else
	{	StringP num=string();
		num->setitem(index, newitem, safe);
		int_=int(num->getnumber());
	}
	return Unbound;
}

NodeP Int::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}

NodeP Int::reverse()
{	return string()->reverse();
}

NodeP Int::readbin(FileP file)
{
	ThreadP th=Thread::This();
	int num;
	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read int binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(int), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(int);
			if (th->rstringlen >= 0)
			{	num = *intP(th->rstringptr);	// reading from string
				th->rstringptr+=sizeof(int);
			}else
				error(FILE_ERROR, new String
					(_("memory block too small to read int")));
		}
	int_=num;
	return this;
}

NodeP Int::upper() {return this;}
NodeP Int::lower() {return this;}
PairP Int::list() {return string()->list();}

void Int::init() 
{
	seedMT(time(0));
		 _Zeroo->setvarPermanent(new Int( 0)); 
	      _One->setvarPermanent(new Int( 1));
	_Minus_One->setvarPermanent(new Int(-1));

	_IntMax   ->setvarPermanent(new Int(max));
	_Int16Max ->setvarPermanent(new Int16(Int16::max));
	_Int8Max  ->setvarPermanent(new Int8 (Int8 ::max));
	_UInt8Max ->setvarPermanent(new UInt8(UInt8::max));

	_IntMin   ->setvarPermanent(new Int(min));
	_Int16Min ->setvarPermanent(new Int16(Int16::min));
	_Int8Min  ->setvarPermanent(new Int8 (Int8 ::min));
	_UInt8Min ->setvarPermanent(new UInt8(UInt8::min));
}
#ifdef WIN32
	#pragma warning(disable: 4307)
#endif
int Int::Remainder(   int x,    int d) 
{	if (!d) 
	{	error(BAD_DATA, new Int(0));
		return Int::max+1;	// set overflow flag
	}
	return x%d;
}

int Int::mod(int x, int d) 
{
	if (!d) 
	{	error(BAD_DATA, new Int(0));
		return Int::max+1;	// set overflow flag
	}
	int m=x%d;
	if ((m < 0) != (d < 0)) return m+d;
	return m;
}

#pragma warning(default: 4307)

int Int::gcd(int a, int b)
{
	if (!b) return a;
	return gcd(b, a % b);
}

int Int::clip(double d)
{
	if (d > Int::max) return  Int::max;
	if (d <-Int::max) return -Int::max-1;
	return int(d);
}

StringP Int::tochar()
{
	char c = char(int_);
	return new String(&c, 0, 1, STRING, String::copy);
}

