#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "Int16.h"
#include "String.h"
#include "File.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

const int16 
	Int16::max = (uint16)(-1) >> 1,
	Int16::min = -Int16::max-1,
	Int16::digits = sizeof(int16)*256/100 + 3;

DataInputStream &Int16::read(DataInputStream &istr)
{	return istr >> int_;
}
wxDataOutputStream &Int16::write(wxDataOutputStream &ostr)
{	return ostr << type << int_;
}

NodeP Int16::copy(PtrToInt_ &ptrToInt)
{	return new Int16(int_);
}

StringP Int16::string()
{
	char s[digits];
	sprintf(s, "%i", int_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void Int16::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, int_);
}

void Int16::printBin(FILEP strm) 
{
	::printBin(strm, int_);
}

int Int16::_sizeof() {return sizeof(int16);}

double Int16::_min(double d) {return int_ < d ?  int_ : d;}
double Int16::_max(double d) {return int_ > d ?  int_ : d;}

NodeP Int16::Norm()
{	if (int_ >=0) return this;
	else return new Int16(-int_);
}

NodeP Int16::maxNorm()
{	if (int_ >=0) return this;
	else return new Int16(-int_);
}

NodeP Int16::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}

NodeP Int16::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(int16))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return boolnode((int_>>i) & 1);
}

NodeP Int16::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (newitem && newitem->type==OBJECT)
	{	int bit=newitem->a_bool(3);
		if (!throwing)
		{	if (i<0 || i >= int(sizeof(int16))<<3)
			{	error(BAD_DATA, index);
				return Unbound;
			}
			if (bit)
				int_|=(1<<i);
			else
				int_&=~(1<<i);
		}
	}else
	{	StringP num=string();
		num->setitem(index, newitem, safe);
		int_=int(num->getnumber());
	}
	return Unbound;
}

NodeP Int16::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}

NodeP Int16::reverse()
{	return string()->reverse();
}

NodeP Int16::readbin(FileP file)
{
	ThreadP th=Thread::This();
	int16 num;
	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read int16 binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(int16), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(int16);
			if (th->rstringlen >= 0)
			{	num = *int16P(th->rstringptr);	// reading from string
				th->rstringptr+=sizeof(int16);
			}else
				error(FILE_ERROR, new String
					(_("memory block too small to read int16")));
		}
	int_=num;
	return this;
}

NodeP Int16::upper() {return this;}
NodeP Int16::lower() {return this;}
PairP Int16::list() {return string()->list();}

int16 Int16::clip(double d)
{
	if (d > Int16::max) return  Int16::max;
	if (d <-Int16::max) return -Int16::max-1;
	return int16(d);
}

int16 Int16::clip(int d)
{
	if (d > Int16::max) return  Int16::max;

	if (d <-Int16::max) return -Int16::max-1;
	return int16(d);
}
