#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "Int8.h"
#include "String.h"
#include "File.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

const char 
	Int8::max = (unsigned char)(-1) >> 1,
	Int8::min = -Int8::max-1,

	Int8::digits = sizeof(char)*256/100 + 3;

DataInputStream &Int8::read(DataInputStream &istr)
{	return istr >> (unsigned char&)int_;
}
wxDataOutputStream &Int8::write(wxDataOutputStream &ostr)
{	return ostr << type << (unsigned char)int_;
}

NodeP Int8::copy(PtrToInt_ &ptrToInt)
{	return new Int8(int_);
}

StringP Int8::string()
{
	char s[digits];
	sprintf(s, "%i", int_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void Int8::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, int_);
}

void Int8::printBin(FILEP strm) 
{
	::printBin(strm, int_);
}

int Int8::_sizeof() {return sizeof(char);}

double Int8::_min(double d) {return int_ < d ?  int_ : d;}
double Int8::_max(double d) {return int_ > d ?  int_ : d;}

NodeP Int8::Norm()
{	if (int_ >=0) return this;
	else return new Int8(-int_);
}

NodeP Int8::maxNorm()
{	if (int_ >=0) return this;
	else return new Int8(-int_);
}

NodeP Int8::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}

NodeP Int8::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(char))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return boolnode((int_>>i) & 1);
}

NodeP Int8::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (newitem && newitem->type==OBJECT)
	{	int bit=newitem->a_bool(3);
		if (!throwing)
		{	if (i<0 || i >= int(sizeof(char))<<3)
			{	error(BAD_DATA, index);
				return Unbound;
			}
			if (bit)
				int_|=(1<<i);
			else
				int_&=~(1<<i);
		}
	}
	else
	{	StringP num=string();
		num->setitem(index, newitem, safe);
		int_=int(num->getnumber());
	}
	return Unbound;
}

NodeP Int8::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}

NodeP Int8::reverse()
{	return string()->reverse();
}

NodeP Int8::readbin(FileP file)
{
	ThreadP th=Thread::This();
	char num;
	File::charModeOn();
	th->inputBlocking++;

	if (!setjmp(th->inputBlockingBuffer))
	{	
		if (th->interactive && file->file()==File::conin)
			num=File::getch();
		else if (file->file())
		{	fread(&num, sizeof(char), 1, file->file());
			if (file->eof()) return 0;
		}
		else if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(char);
			if (th->rstringlen >= 0)
			{	num = *charP(th->rstringptr);	// reading from string
				th->rstringptr+=sizeof(char);
			}else
				error(FILE_ERROR, new String
					(_("memory block too small to read char")));
		}
	}
	th->inputBlocking=0;
	if (file->eof()) return 0;
	int_=num;
	return this;
}

NodeP Int8::upper() {return this;}
NodeP Int8::lower() {return this;}
PairP Int8::list() {return string()->list();}

char Int8::clip(double d)
{
	if (d > Int8::max) return  Int8::max;
	if (d <-Int8::max) return -Int8::max-1;
	return char(d);
}

char Int8::clip(int d)
{
	if (d > Int8::max) return  Int8::max;
	if (d <-Int8::max) return -Int8::max-1;
	return char(d);
}
