#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "String.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "Object.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "Int8Array.h"
#include "UInt8Array.h"
#include "FloatArray.h"
#include "Pair.h"
#include "File.h"
#include "Init.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

Int8Array::Int8Array(int length_, int origin_): Node(INT8ARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (int8P) GC_::calloc(length_+1, sizeof(int8));
	if (!_data) 
		return;
	_length=length_;
}

Int8Array::Int8Array(Int8ArrayP array_): Node(INT8ARRAY)
{
	_length=0;
	_origin = array_->_origin;
	_data = (int8P) GC_::calloc(array_->_length+1, sizeof(int8));
	if (!_data) 
		return;
	_length = array_->_length;
	int8P source = array_->_data, 
		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
}

void Int8Array::free() {GC_::free(_data); _data=0;}

DataInputStream &Int8Array::read(DataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (int8P) GC_::calloc(sizeof(int8), length+1);
	if (!_data) 
		return istr;
	_length=length;
	int8P itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (int8&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Int8Array::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	int8P itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << *itemptr++;
	return ostr;
}

NodeP Int8Array::copy(PtrToInt_ &ptrToInt)
{
	Int8ArrayP arr=new Int8Array(this);
	return arr;
}

StringP Int8Array::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	Thread::This()->stringptr=textbuf;
	Thread::This()->stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*(Thread::This()->stringptr)='\0';
	return new String(textbuf);
}

void Int8Array::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	int i=0, wid=0;
	int8P pp=data();

	if (width >= 2*_length-10)
		wid=width/_length;

	if (noBrackets <= 0)
		::print(strm, width, "{");

	while (i < _length) 
	{	int w=wid;
		::print(strm, w, int(*pp++));	
		if (++i < length())
			::print(strm, w, ' ');
		width-=wid-w;
	}
	if (i < length()) 
		::print(strm, width, "...");
	if (noBrackets <= 0)
		::print(strm, width, '}');

	if (origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", origin());
		::print(strm, width, org);
	}
}

void Int8Array::printBin(FILEP strm)
{
	ThreadP th=Thread::This();
	int wid=length();
	int8P pp=data();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen-=sizeof(int8)*wid;
		if (th->stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*int8P(th->stringptr) = *pp++;
				th->stringptr+=sizeof(int8);
			}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print int8array to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(data());
		int width=Int::max;	
		for (int l=sizeof(int)*wid; l; l--)
			::print(strm, width, *cp++);
		return;
	}
	fwrite(pp, sizeof(int8), wid, strm);
}

NodeP Int8Array::contentsTree(	wxArrayString &namearr, 
										wxArrayString &valarr, 
										wxArrayPtrVoid &objarr, 
										ObjectType want, 
										ObjectType dont,
										ObjectType must, 
										int &count, int level)
{
	objarr[count]=this;
	if (!expanded())
	{	StringP str=string();
		if (str  &&  str->is.string)
		{	valarr[count]=str->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
		}
		return this;
	}
	int8P dp=_data;
	for (int i=0; i < _length;  i++)
	{	namearr[count]=wxString(' ',level-1);
		namearr[count] << i+_origin;
		objarr[count]=this;
		int d=(*dp++);
		valarr[count].Printf("%6i %4x %c", d, d, d);
		incContentsArrays(namearr, valarr, objarr, count);
	}
	return this;
}

int Int8Array::count() {return length();}

int Int8Array::_sizeof() {return length()*sizeof(int8);}

NodeP Int8Array::compare(NodeP nd)
{	if (nd->type != INT8ARRAY) 
		return Unbound;
	if (this==nd)
		return 0;

	Int8ArrayP b=Int8ArrayP(nd);
	int8P dp=_data, dpb=b->_data;
	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{	int c=(*dp++) - (*dpb++);
		if (c)
			return (c > 0 ?  _One : _Minus_One);
	}
	if (!i & !j)
		return 0;
	if (i)
		return _One;
	return _Minus_One;
}

bool Int8Array::equalP(NodeP b)
{
	if (b->type != INT8ARRAY) 
		return false;
	if (this==b)
		return true;

	Int8ArrayP ba=Int8ArrayP(b);
	int8P dp=_data, dpb=ba->_data;
	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
		if ((*dp++) - (*dpb++))
			return false;
	if (!i & !j)
		return true;
	return false;
}

NodeP Int8Array::first_() 
{
	if (length() >= 1) 
		return new Int8(*(Int8ArrayP(this)->data()));
	else
		return 0;
}

NodeP Int8Array::butFirst()
{
	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	Int8ArrayP result=new Int8Array(l, o);
	if throwing return Unbound;
	int8P source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Int8Array::last()
{
	return new Int8(*(data()+length()-origin()));
}

NodeP Int8Array::butLast()
{
	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	Int8ArrayP result=new Int8Array(l, o);
	if throwing return Unbound;
	int8P source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Int8Array::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	Int8ArrayP result=new Int8Array(l+1, o);
	if throwing return Unbound;
	int8P source=data(), dest=result->data();
	*dest++=Int8::clip(nd->getnumber());
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Int8Array::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	Int8ArrayP result=new Int8Array(l+1, o);
	int8P source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=Int8::clip(nd->getnumber());
	return result;
}

double Int8Array::_min(double d)
{
	ThreadP th=Thread::This();
	int8P dp=_data;
	for (int i=0;  i<_length;  i++)
	{	int8 n=*dp++;
		if (n < d)
		{	d=n;
			th->minimumPosition=i+_origin;
		}
	}
	return d;
}

double Int8Array::_max(double d)
{
	ThreadP th=Thread::This();
	int8P dp=_data;
	for (int i=0;  i<_length;  i++)
	{	int8 n=*dp++;
		if (n > d)
		{	d=n;
			th->maximumPosition=i+_origin;
		}
	}
	return d;
}

NodeP Int8Array::Norm()
{	double d=0;  int8P dp=_data;
	for (int i=_length;  i;  i--)
	{	int h = (*dp++);
		if (throwing) return Unbound;
		d+= h*h;
	}
	return new Float(::sqrt(d));
}

NodeP Int8Array::maxNorm()
{	int d=0;  int8P dp=_data;
	for (int i=_length;  i;  i--)
	{	int8 h = (*dp++);
		if (throwing) return Unbound;
		int8 ha=::abs(h);
		if (ha > d) d = ha;
	}
	return new Int(d);
}

NodeP Int8Array::resize(int newsize)
{	Int8ArrayP r=new Int8Array(newsize);
	if throwing return Unbound;
	int8P d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	int diff = d[i+1]-d[i];
					double slope = double(diff)/di;

					for (int ii=di-1;  ii;  ii--)
						r->_data[ni+ii] = int8(d[i]+slope*ii);
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f,
					diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=int8(d[int(i)]+frac*diff+0.5);
			}
		}
	}
	return r;
}

NodeP Int8Array::lowPassFilter(double weight)
{
	Int8ArrayP result=new Int8Array(length());
	if throwing return Unbound;
	int8P source=data(), dest=result->data();
	double median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=int8(median);
	}
	Float::reset_FPU();
	return result;
}

NodeP Int8Array::saturateAbove(double limit) 
{
	Int8ArrayP result=new Int8Array(length());
	if throwing return Unbound;
	int8P source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	int8 s=*source++;
		*dest++= s>limit? int8(limit) : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP Int8Array::saturateBelow(double limit) 
{
	Int8ArrayP result=new Int8Array(length());
	if throwing return Unbound;
	int8P source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	int8 s=*source++;
		*dest++= s<limit? int8(limit) : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP Int8Array::item(int i)
{	if (i >= 0  &&  i < _length) 
		return new Int8(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}

NodeP Int8Array::items(NodeP start, NodeP end)
{
	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	Int8ArrayP newar=new Int8Array(e-s+1);
					if throwing return Unbound;
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP Int8Array::setitem(NodeP index, NodeP newitem, bool safe)
{
	int i=index->a_int(1);
	int8 ni=newitem->a_int8(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni;
	}
	return Unbound;
}

NodeP Int8Array::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	int8P d=_data;   int l=_length;
	Int8P i=new Int8(0);
	if throwing return Unbound;

	while (--l >= 0 && !throwing) 
	{	i->int_=(*d++);
		if (subobj->equalP(i))
			return _True;
	}
	return _False;
}

NodeP Int8Array::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 

	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					int8P d=&_data[i];
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =  Int8::clip(IntP(h)->int_);	break;
							case FLOAT_:*d++ =Int8::clip(FloatP(h)->float_);break;
							case NINT8:	*d++ =            Int8P(h)->int_;	break;
							case NINT16:*d++ =           Int16P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = Int8::clip(h->getnumber());
											else
											{	error(BAD_DATA, n, 3); 
												return IntArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);	
								return IntArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					int8P d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int8::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT8ARRAY)
			{	Int8ArrayP n=Int8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	int8P s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==UINT8ARRAY)
			{	UInt8ArrayP n=UInt8ArrayP(newitems);
				if (i+n->length() <= _length)
				{	uint8P s=n->data();
					int8P d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)
			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->data();
					int8P d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int8::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}

NodeP Int8Array::reverse()
{
	Int8ArrayP result=new Int8Array(_length);
	if throwing return Unbound;

	int8P d=_data, rev=result->_data+_length-1;
	for (int i=_length; i>0; i--)
		*rev-- = *d++;
	return result;
}

NodeP Int8Array::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	Int8ArrayP val = new Int8Array(length(), origin());
	if throwing return Unbound;
	int8P
		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}

NodeP Int8Array::shuffle()
{	Int8ArrayP result=new Int8Array(Int8ArrayP(this));
	if throwing return Unbound;
	int8P d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		int8 tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}

PairP Int8Array::list()
{
	int8P d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
	{	p = new Pair(new Int8(*--d), p);
		if throwing return PairP(Unbound);
	}
	return p;
}

ArrayP Int8Array::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	if throwing return ArrayP(Unbound);
	NodePP d = array->data();
	int8P s=_data;

	for (int i=_length;  i;  i--)
	{	*d++ = new Int8(*s++);
		if throwing return ArrayP(Unbound);
	}
	return array;
}

IntArrayP Int8Array::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	if throwing return IntArrayP(Unbound);
	intP d = intarray->data();
	int8P s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return intarray;
}

Int8ArrayP Int8Array::int8array(int origin) 
{
	return new Int8Array(Int8ArrayP(this));
}

FloatArrayP Int8Array::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	if throwing return FloatArrayP(Unbound);
	doubleP d = floatarray->data();
	int8P s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return floatarray;
}

StringP Int8Array::tochar()
{
	StringP result=new String(length());
	if throwing return StringP(Unbound);
	charP dest=result->ptr();
	int8P src=data();
	for (int i=0; i<_length; i++)
		*dest++=char(*src++);
	return result;
}

void Int8Array::combSortInPlace()
{
	double gap=_length; //initialize gap size
	int swaps=0; //see bubblesort for an explanation
    
	while (!(gap<=1 && swaps==0))
	{	//update the gap value for a next comb
		if (gap > 1)
		{	gap/=1.3;
			if (gap==10 || gap==9)
				gap=11;
		}
		int i=0;
		//a single "comb" over the input list
		while (i+gap<_length)
		{	int j=int(i+gap);
			if (_data[i] > _data[j])
			{	int8 tmp=_data[i];
				_data[i]=_data[j];
				_data[j]=tmp;
			}
			i++;
		}
	}
}

NodeP Int8Array::combSort()
{
	Int8ArrayP a=new Int8Array(this);
	if throwing return Unbound;
	a->combSortInPlace();
	return a;
}

NodeP Int8Array::readbin(FileP file)
{
	ThreadP th=Thread::This();
	int8P ip=data();

	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read int8array binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(int8), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(int8)*_length;
			if (th->rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *int8P(th->rstringptr);	// reading from string
					th->rstringptr+=sizeof(int8);
				}
			else
				error(FILE_ERROR, new String
					(_("memory block too small to read int8array")));
		}
	return this;
}
