#include "stdafx.h"
//___________________________________________________________________
//
//                          Class IntControl
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "spinctli.h"

#include "Main.h"
#include "File.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "Pair.h"
#include "WindowPair.h"
#include "Point.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "IntControl.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Color.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(IntControl, wxSpinCtrlInt)
	EVT_SPINCTRL	(-1, IntControl::OnChange)
END_EVENT_TABLE()

IntControl::IntControl(IntControlNodeP intControlNode_, int style)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=intControlNode_;
	wxWindow *parent=(wxWindow*)console;
	if (intControlNode_->parent)
		parent=intControlNode_->parent->window;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			1,1,1,1,
			_("IntControl"));
}

IntControl::IntControl(IntControlNodeP intControlNode_, 
			wxWindow* parent, wxWindowID id, 
			int value, int minValue, int maxValue, 
			PairP onChange_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxSpinCtrlInt(parent, id, 
			wxEmptyString,
			pos, size, style, 
			minValue, maxValue, value,
			name)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=intControlNode_;
	*(PairP*)(d(onChange))=onChange_;
	SetValue(value);
	IntControlNodeP(*d(windowNode))->window=WindowP(this);
	IntControlNodeP(*d(windowNode))->style=style;
}

IntControl::~IntControl()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
}

void IntControl::OnChange(wxSpinEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventIntControl(this);
	if throwing return;
	*th->eventFiFo << evt;
	event.Skip();
}

wxString IntControl::wxstring()
{
	wxString str;
	str << _("IntControl") << " " << GetMin() << ' ' << GetValue() << ' ' << GetMax() << ' ' 
		<< '[' << GetName() << "] ";
	if (onChange)
		str << (*d(onChange))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &IntControl::read(DataInputStream &istr) 
{	
	int value, min, max;
	wxPoint pos;
	wxSize size;
	wxString name;
	
	istr >> min >> value >> max >> name >> (int&)*d(onChange)
		>> pos.x >> pos.y >> size.x >> size.y;

	SetRange(min, max);
	SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	return istr;
}
wxDataOutputStream &IntControl::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetMin() << GetValue() << GetMax() << GetName() 
		<< GC_::ptrToInt->find(*d(onChange))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;
	return ostr;
}

PairP IntControlNode::onChange()
{
	if (window)
		return PairP(*IntControlP(window)->d(IntControl::onChange));
	return 0;
}

void IntControlNode::setonChange(NodeP nd) 
{
	if (window)
	{	*IntControlP(window)->d(IntControl::onChange) = PairP(nd);  
		check_oldyoung(nd);
	}
}

IntControlNode::IntControlNode(): 
	WindowNode(INTCONTROL), 
	intControl((IntControlP&)window)
{}

IntControlNode::IntControlNode(WindowNodeP parent_): 
	WindowNode(INTCONTROL, parent_), 
	intControl((IntControlP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void IntControlNode::free() 
{	
	if (window) 
	{	ThreadP th=Thread::This();
//		GUI_ENTER
		*IntControlP(window)->d(IntControl::windowNode)=0;
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		window->Destroy(); 
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::tIntControl:
				if (WindowP(((EventIntControl*)ev)->intControl)==window)
					((EventIntControl*)ev)->intControl=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		window=0;
//		GUI_LEAVE
	}
}
void IntControlNode::inc_gc() {fifo << onChange();}
int IntControlNode::youngs() {return gmark(onChange());}
int IntControlNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(onChange(), ptrToInt);
}
void IntControlNode::check_oldyoungs() {check_oldyoung(onChange());}

DataInputStream &IntControlNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode >> style;
		window=WindowP(new IntControl(IntControlNodeP(this), style));
		if (!window)
		{	error(OUT_OF_MEM); return istr;
		}
		window->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &IntControlNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (window)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent)
			<< style;
		window->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void IntControlNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (window)
		setonChange((Pair*)intToPtr.find(int(IntControlP(window)->onChange)));
}

void IntControlNode::recreate()
{	wxWindow *p=0;
	if (parent)
		p=parent->window;
	IntControlP i=IntControlP(window), i2=0;
	i2=new IntControl(IntControlNodeP(this),
			p, -1, i->GetValue(), i->GetMin(), i->GetMax(),
			PairP(*i->d(IntControl::onChange)), 
			i->GetPosition(), i->GetSize(), style, 
			wxDefaultValidator, i->GetName());
	if (!i2)
	{	error(OUT_OF_MEM); return;
	}
	window=WindowP(i2);
	setonChange(*i->d(IntControl::onChange));
	delete i;
	window=WindowP(i2);
}

StringP IntControlNode::string()
{
	if (intControl)
		return new String(intControl->wxstring());
	return _Null_Word->name();
}

wxString IntControlNode::wxstring()
{
	if (intControl)
		return intControl->wxstring();
	return wxEmptyString;
}

void IntControlNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lIntControl(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=Thread::This()->graph;
	WindowNodeP parentNode=0;
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString name=_("IntControl");
	if (arg[2])
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	int minValue;
	{	int v=arg[3]->a_int(3);
		if throwing return Unbound;
		minValue=v;	
	}
	int value;
	{	int v=arg[4]->a_int(4);
		if throwing return Unbound;
		value=v;	
	}
	int maxValue;
	{	int v=arg[5]->a_int(5);
		if throwing return Unbound;
		maxValue=v;	
	}
	PairP onChange=0;
	if (exist(6))
	{	PairP os=arg[6]->a_runnable(6);
		if throwing return Unbound;
		onChange=os;
	}
	long style = wxSP_ARROW_KEYS;
	if (exist(7))
	{	int s=arg[7]->a_int(7);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(8))
	{	PairP p=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(9))
	{	PairP s=arg[9]->a_int_vector(9);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
//	GUI_ENTER
	IntControlNodeP intControlNode=new IntControlNode(parentNode);
	if throwing return Unbound;
	IntControlP ic=new IntControl(intControlNode,
			parent, id, value, minValue, maxValue, 
			onChange, pos, size, style, wxDefaultValidator, name);
	if (!ic)
	{	error(OUT_OF_MEM); return Unbound;
	}
//	GUI_LEAVE
	return intControlNode;
}

NodeP lIntControlValue(PairPP arg)
{
	ThreadP th=Thread::This();
	if (exist(1))
	{	IntControlNodeP intControl=arg[1]->a_intControl(1);
		if throwing return Unbound;
//		GUI_ENTER
		IntP val=new Int(IntControlP(intControl->window)->GetValue());
//		GUI_LEAVE
		return val;
	}
	else
		return new Int(th->intControlValue);
}

NodeP lIntControlSetValue(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	IntControlP(intControl->window)->SetValue(value);
//	GUI_LEAVE
	return Unbound;
}

NodeP lIntControlSetRange(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	int min=arg[2]->a_int(2);
	if throwing return Unbound;
	int max=arg[3]->a_int(3);
	if throwing return Unbound;
//	GUI_ENTER
	IntControlP(intControl->window)->SetRange(min, max);
//	GUI_LEAVE
	return Unbound;
}

NodeP lIntControlOnChange(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
//	GUI_ENTER
	intControl->setonChange(function);
//	GUI_LEAVE
	return Unbound;
}

NodeP lIntControlEnable(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	intControl->window->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lIntControlDestroy(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=intControl->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(intControl));
			p->check_oldyoung(p->children);
		}
	}
	intControl->window->Destroy();
//	GUI_LEAVE
	return Unbound;
}
