#include "stdafx.h"
#include <malloc.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

#include "Main.h"
#include "File.h"
#include "Error.h"
#include "IntNodeTable.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

IntNodeTable::IntNodeTable(int size_):
	size(0)
{
	data = (IntNodePairP*)(GC_::calloc(size_, sizeof(IntNodePairP)));
	if (data)
		size=size_;
}

IntNodeTable::~IntNodeTable()
{
	for (int i=0; i < size; i++)
	{	if (data[i])
			delete data[i];
	}
	if (data)
	{	GC_::free(data); data=0;
	}
}

int IntNodeTable::hash(int i)
{
	unsigned int index = (unsigned int)i;
	if (size)
		index %= size;
	else
		index=0;
	return index;
}

IntNodeTable::IntNodePairP IntNodeTable::find(int i)
{
	IntNodePairP obj;
	for (obj=data[hash(i)];  obj;  obj=obj->tail)
		if (obj->i==i)
			return obj;
	return 0;
}

IntNodeTable::IntNodePairP IntNodeTable::insert(int i, NodeP nd)
{
	IntNodePairP &obj=data[hash(i)];
	obj = new IntNodePair(i, nd, obj);
	if (!obj)
		error(OUT_OF_MEM);
	return obj;
}

int IntNodeTable::item(int index)
{	IntNodePairP obj=find(index);
	if (!obj)
		return 0;
	return obj->i;
}

void IntNodeTable::setitem(int index, NodeP newitem)
{	IntNodePairP obj=find(index);
	if (!obj)
		obj=insert(index, newitem);
	obj->nd=newitem;
}

void IntNodeTable::removeitem(int index)
{	int i=hash(index);
	IntNodePairP obj=data[i], last=0;
	while (obj)
	{	if (index==obj->i)
			break;
		last=obj;
		obj=obj->tail;
	}
	if (!obj)
	{//	error(BAD_DATA, index);
		return;
	}
	if (!last)
	{	data[i]=obj->tail;
	}
	else
		last->tail=obj->tail;
}

