#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>

#include "Main.h"
#include "File.h"
#include "IntToPtr_.h"
#include "PtrToInt_.h"
#include "String.h"
#include "Line.h"
#include "Object.h"
#include "Procedure.h"
#include "Mem.h"
#include "Init.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

int Line::the_generation=1;

StringP Line::filename() {return _filename;}

void Line::setfilename(NodeP nd) 
{
	wxASSERT(is.line);
	wxASSERT(nd && nd->is.string);
	_filename=StringP(nd);  
	check_oldyoung(nd);
}

Line::Line() 
{	type=LINE; _tail=0; _first=0; 
	generation=the_generation;
	ThreadP th=Thread::This();
	if (th)
	{	col=th->col;
		tcol=th->tcol;
		row=th->row;
		setfilename(th->filename());
	}
}

Line::Line(NodeP first_, NodeP tail_)
{	ThreadP th=Thread::This();
	type=LINE; setfirst(first_); settail(tail_);
	generation=the_generation;
	col=th->col; 
	tcol=th->tcol; 
	row=th->row; 
	wxASSERT(is_nil_or_ok(th->filename()));
	setfilename(th->filename());
}
Line::Line(NodeP first_, NodeP tail_, NodeType type_)
{	ThreadP th=Thread::This();
	type=type_; setfirst(first_); settail(tail_);
	generation=the_generation;
	col=th->col; 
	tcol=th->tcol; 
	row=th->row; 
	wxASSERT(is_nil_or_ok(th->filename()));
	setfilename(th->filename());
}

Line::Line(const charP str)
{	type=LINE; _tail=0;
	generation=the_generation;
	ThreadP th=Thread::This();
	if (th)
	{	col=th->col;
		tcol=th->tcol;
		row=th->row;
		setfilename(th->filename());
	}
	setfirst(new String(str));
	flag=ObjectType(0);
}

Line::Line(const wxChar *str)
{	type=LINE; _tail=0;
	generation=the_generation;
	ThreadP th=Thread::This();
	if (th)
	{	col=th->col;
		tcol=th->tcol;
		row=th->row;
		setfilename(th->filename());
	}
	setfirst(new String(str));
	flag=ObjectType(0);
}

void Line::inc_gc() {fifo << first() << tail() << filename();}

int Line::youngs() {return gmark(first())+gmark(tail())+gmark(filename());}

int Line::youngsPtr(PtrToInt_ &ptrToInt) 
{	return 
		gmarkPtr(first(), ptrToInt)
		+gmarkPtr(tail(), ptrToInt)
		+gmarkPtr(filename(), ptrToInt);
}

void Line::check_oldyoungs() 
{	check_oldyoung(first()); 
	check_oldyoung(tail());
	check_oldyoung(filename());
}

void Line::setPtrs(IntToPtr_ &intToPtr)
{	setfirst(  (Node*)intToPtr.find(int(_first  )));
	settail (  (Node*)intToPtr.find(int(_tail   )));
	setfilename((String*)intToPtr.find(int(_filename)));
}
DataInputStream &Line::read(DataInputStream &istr) 
{	return istr >> (int&)_first >> (int&)_tail 
		>> col >> tcol >> row >> generation >> (int&)_filename;
}
wxDataOutputStream &Line::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(first()) 
		<< GC_::ptrToInt->find(tail())
		<< col << tcol << row << generation
		<< GC_::ptrToInt->find(filename());
}

NodeP Line::copy(PtrToInt_ &ptrToInt)
{
	LineP line=new Line();
	if throwing return Unbound;
	line->_tail=PairP(ptrToInt.find(tail()));
	line->_first=PairP(ptrToInt.find(first()));
	line->_filename=StringP(ptrToInt.find(filename()));
	line->col=col;
	line->tcol=tcol;
	line->row=row;
	line->generation=generation;
	return line;
}

NodeP Line::deepCopy()
{	if (!this) 
		return 0;
	NodeP f=0, t=0;
	if (first())
		f=first()->deepCopy();
	if (tail())
		t=tail()->deepCopy();
	LineP l=new Line(f, t, type);
	if throwing return Unbound;
	l->col=col;
	l->tcol=tcol;
	l->row=row;
	l->generation=generation;
	l->setfilename(filename());
	return l;
}

NodeP Line::deepCopyLine()
{	if (!this) 
		return 0;
	NodeP f=0, t=0;
	if (first())
		f=first()->deepCopyLine();
	if (tail())
		t=tail()->deepCopyLine();
	LineP l=new Line(f, t, type);
	if throwing return Unbound;
	l->col=col;
	l->tcol=tcol;
	l->row=row;
	l->generation=generation;
	l->setfilename(filename());
	return l;
}

ProcedureP Line::anonymous_function() 
{
	return define_helper(false, true);
}

void Line::init()
{
	ThreadP th=Thread::This();
	th->setfilename(_Unnamed->name());
	th->filenames=new Pair(th->filename(), 0);
}

StringP Line::findFile(StringP filename_)
{
	ThreadP th=Thread::This();
	StringP f=0;
	for (PairP fnlist=th->filenames;  fnlist;  next(fnlist))
		if (fnlist->first() && fnlist->first()->equal(filename_))
			f=StringP(fnlist->first());
	if (f)
		return f;
	else
	{	PairP fns=new Pair(filename_, th->filenames);
		if throwing return StringP(Unbound);
		th->filenames=fns;
	}
	return filename_;
}

