//___________________________________________________________________
//
//                        Class LogoApp
//
//                         Definitions
//___________________________________________________________________

#include "stdafx.h"

#ifdef WIN32
//	#include <vld.h>
	#include <mmsystem.h>
	#include <mmreg.h>
#else
	#include <sys/time.h>
#endif

#ifdef OPENGL
	#include <GL/glu.h>
#endif

#include <wx/evtloop.h>
#include <wx/image.h>
#include <wx/listctrl.h>
//#include <wx/html/helpctrl.h>
#include <wx/help.h>
#include <wx/splitter.h>
#include <wx/config.h>
#include <wx/datetime.h>
#include <wx/sound.h>
#include <wx/notebook.h>
#include <wx/treectrl.h>
#include <wx/ipc.h>
#include <wx/print.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/stc/stc.h>
#include <wx/sysopt.h>
#include <wx/platform.h>
#include <wx/aui/aui.h>
#include <wxMidi.h>

extern "C" void pm_term(void); 

#include "Main.h"
#include "Pair.h"
#include "Line.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "File.h"
#include "Object.h"
#include "String.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "Int8Array.h"
#include "UInt8Array.h"
#include "Init.h"
#include "List.h"
#include "Error.h"
#include "NodePStack.h"
#include "TurtleNode.h"
#include "Table.h"
#include "PtrTable.h"

#include "Color.h"
#include "Point.h"
#include "Pen.h"
#include "Font.h"
#include "Turtle.h"
#include "FTGLFontManager.h"

#include "Event.h"
#include "History.h"
#include "Console.h"
#include "Consolew.h"
#include "Graph.h"
#include "Calls.h"
#include "Vars.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "Edit.h"
#include "EdFrame.h"
#include "EdDefsExt.h"
#include "EdPrefs.h"
#include "MemoryWindow.h"
#include "Midi.h"
#include "LogoApp.h"
#include "Environment.h"
#include "Security.h"
#include "Thread.h"
#include "Server.h"
#include "DebugState.h"
#include "Aui.h"

#ifdef WIN32
	#include "io.h"
#endif

#ifdef _DEBUG
	#undef THIS_FILE
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

static char THIS_FILE[]= __FILE__;

//#define EAT_MEM

const int eat_mem_size=550*1000*1000;
charP eat_mem_mem;

#ifdef WIN32
DECLARE_APP(LogoApp)
IMPLEMENT_APP(LogoApp)
#else
wxAppConsole *wxCreateApp()
{
	wxAppConsole::CheckBuildOptions(WX_BUILD_OPTIONS_SIGNATURE,
												"your program");
	return new LogoApp;
}

wxAppInitializer wxTheAppInitializer((wxAppInitializerFunction) wxCreateApp);

LogoApp &wxGetApp()
{
	return *wx_static_cast(LogoApp*, wxApp::GetInstance());
}
#endif

LogoApp *theApp;
Console *console;
MemoryWindow *memoryWindow;

void (__cdecl *LogoApp::callback)()=0;

#define NEED_MEM_LogoApp_OnInit(x)						\
	if (!x)														\
	{	delete heapReserve;									\
		heapReserve=0;											\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x": Insufficient Memory";						\
		wxMessageBox(msg,	aUCBLogo+" "+_("Error"));			\
		return false;											\
	}

bool LogoApp::OnInit()
{
	theApp=this;
	heapReserve=new char[0x10000];
	if (!heapReserve)
		return false;

	Thread::main=new Thread(wxTHREAD_JOINABLE);
	if (!Thread::main)
		return false;

#ifdef WIN32
	Thread::main->fiber=(Fiber*)ConvertThreadToFiber(Thread::main);

	DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), 
		GetCurrentProcess(), &Thread::mainHandle, 
		0, TRUE, DUPLICATE_SAME_ACCESS); 
#else
//	TODO
#endif

	action=NOP;
	saved=false;
	aui=0;
	mainShown=true;
	mainFrame=0;
	consoleFrame=0;
	console=0;
	graph=0;
	edFrame=0;
	calls=0;
	vars=0;
	stackvars=0;
	max.x=wxSystemSettings::GetMetric(wxSYS_SCREEN_X);
	max.y=wxSystemSettings::GetMetric(wxSYS_SCREEN_Y);
	platformInfo=new wxPlatformInfo();
	osVersion=wxGetOsVersion(&osMajor, &osMinor);
	osDescription=platformInfo->GetOperatingSystemFamilyName();

	aUCBLogo=::aUCBLogo;
	SetAppName(aUCBLogoVersion);
	config = new wxConfig(aUCBLogoVersion);
	NEED_MEM_LogoApp_OnInit(config);

	autoloadEdit=false;
	config->Read("MakeBackupFiles", &backup, true);

	environment.comspec   = config->Read("comspec"  , environment.comspec  );
	environment.editor    = config->Read("editor"   , environment.editor   );
	environment.fontsdir  = config->Read("fontsdir" , environment.fontsdir );
	environment.helpfiles = config->Read("helpfiles", environment.helpfiles);
	environment.logolib   = config->Read("logolib"  , environment.logolib  );
	environment.tempdir   = config->Read("tempdir"  , environment.tempdir  );
	environment.servicedir= config->Read("servicedir", environment.servicedir  );

	runDirectory = config->Read("RunDirectory", wxGetCwd());
	wxString exeDir = wxPathOnly(argv[0]);
	if (exeDir != runDirectory)
	{
#ifdef WIN32
		helpFileName = exeDir+wxFileName::GetPathSeparator()+aUCBLogo+_T(".chm");
#else
		helpFileName = exeDir+wxFileName::GetPathSeparator()+_T("wxhelp")+
			wxFileName::GetPathSeparator()+aUCBLogo+_T(".hhp");
#endif
		runDirectory=exeDir;
	}else
	{
#ifdef WIN32
		helpFileName = config->Read("HelpFileName", 
			wxGetCwd()+wxFileName::GetPathSeparator()+aUCBLogo+_T(".chm"));
#else
		helpFileName = config->Read("HelpFileName",
			wxGetCwd()+_T("/wxhelp/aUCBLogo.hhp"));
#endif
	}
	showGuiOnError = (config->Read("ShowGuiOnError", true) != 0);
	help=0;

	wxInitAllImageHandlers();

	consolew=new Consolew();
	wxString userName=wxGetCwd()+wxFileName::GetPathSeparator()+wxGetUserName();
	historyFileName=userName+"-"+aUCBLogo+_T(".history");
	history=new History(0, historyFileName);

	server=new Server;
#ifdef WIN32
	wxString service=environment.servicedir
		+"/aucblogo.service";	
//	wxString service="C:/temp/aucblogo.service";
#else
	wxString service="aucblogo.service";
#endif
//	service << wxGetProcessId();
	bool err=0;
	if (server)
		err=server->Create(service);
	return true;
}

#define NEED_MEM_LogoApp_createGUI(x)					\
	if (!x)														\
	{	delete heapReserve;									\
		heapReserve=0;											\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x": Insufficient Memory";						\
		wxMessageBox(msg,	"aUCBLogo Error");			\
		return;													\
	}

void LogoApp::createGUI()
{
	locale=0;
	if(!config->Read(wxT("Language"), 
		&language, wxLANGUAGE_UNKNOWN))
		language = wxLANGUAGE_UNKNOWN;
	loadLocale(language);

	graphShown=false;
	config->Read("GraphShown", &graphShown);
	stackvarsShown=false;
	config->Read("StackvarsShown", &stackvarsShown);
	varsShown=false;
	config->Read("VarsShown", &varsShown);
	callsShown=false;
	config->Read("CallsShown", &callsShown);
	editorShown=false;
	config->Read("EditorShown", &editorShown);

	wxString userName=wxGetCwd()+wxFileName::GetPathSeparator()+wxGetUserName();
	consoleFileName=userName+"-"+aUCBLogo+_T(".console");
	pictureFileName=userName+"-"+aUCBLogo+_T(".picture");
	  nodesFileName=userName+"-"+aUCBLogo+_T(".nodes");

	caretpos=config->Read("ConsoleCaretPos", long(0));
	config->Read("ConsoleAutoSave", &consoleAutosave, true);
	config->Read("ConsoleHSCROLL", &consoleHSCROLL, false);
	config->Read("PictureAutoSave", &pictureAutosave, false);
	config->Read("NodesAutoSave", &nodesAutosave, false);
	config->Read("DoubleBuffer", &doubleBuffer, true);
	config->Read("PlayOnSwap", &playOnSwap, true);
	config->Read("ErallOnRun", &erallOnRun, true);
	config->Read("SetFocusOnRun", &setFocusOnRun, true);

	config->Read("RingBufferSize", &ringBufferSize, 
		long(DebugStates::defaultSize));

	if (mainShown)
		mainShow();

	memblockx=config->Read("MemoryBlockX", 5);
	memblocky=config->Read("MemoryBlockY", 5);
	memborder=config->Read("MemoryBorder", 1);
	memposx  =config->Read("MemoryPosX", max.x/2);
	memposy  =config->Read("MemoryPosY", (long int)(0));
	memsizex =config->Read("MemorySizeX", max.x/2);
	memsizey =config->Read("MemorySizeY", int(max.y*0.96));

	memoryFrame=0;
	memoryShown=false;
	config->Read("MemoryShown", &memoryShown);
	if (memoryShown)
		memoryShow();

	ThreadP th=Thread::This();
	th->throw_node=0;
	th->stoppingFlag=RUN;
	action=NOP;
	StringP startup=new String(theApp->langdir+_("startup"));
	startup->Node::silent_load(0);
	if throwing
	{	if (th->throw_node==_error   ) error_print();
		else if (th->throw_node!=_toplevel 
				&& th->throw_node!=_system)
		{	error(NO_CATCH_TAG, th->throw_node);	error_print();
		}
		th->stoppingFlag=RUN;
		th->inpause=false;
	}
	if (stopping || outputting)
	{	c_out << _("; You must be in a procedure to use OUTPUT or STOP.") << endL;
		th->stoppingFlag=RUN;
	}
	th->argStack->clean();
	if (console)
		th->dispatchMessages();

	init_aUCBLogoWordLists();
	loadEditorFiles();

	if (graph)
	{
#ifdef WIN32
#ifdef OPENGL
//		graph->SetCurrent();
//		graph->setOpenGLState();
#endif
#else
		//TODO: test it!
#ifdef OPENGL
		graph->SetCurrent();
		graph->setOpenGLState();
#endif
#endif
	}
	auiPerspective=config->Read("AuiPerspective", auiPerspective);
	if (mainFrame && !aui->LoadPerspective(auiPerspective))
		aui->Update();

	while (Pending())
		th->dispatch();

	if (mainFrame)
		mainFrame->SetFocus();
	while (Pending())
		th->dispatch();
	if (consoleFrame)
		consoleFrame->SetFocus();
	if (console)
	{	console->SetFocus();
		console->SetSelection(caretpos,caretpos);
	}
}

void LogoApp::destroyGUI()
{
//	if (!console)
//		return;

	ThreadP th=Thread::This();

	long language = wxLANGUAGE_UNKNOWN;
	if (locale)
		language = locale->GetLanguage();
	config->Write("Language", language);

	if (mainFrame)
	{	auiPerspective=aui->SavePerspective();
		config->Write("AuiPerspective", auiPerspective);
	}
	config->Write("MemoryShown", ::memoryWindow != 0);
	if (::memoryWindow)
	{
		memoryFrame->GetPosition(&memposx , &memposy );
		memoryFrame->GetSize    (&memsizex, &memsizey);
		config->Write("MemoryBlockX", memoryFrame->memoryWindow->blockx);
		config->Write("MemoryBlockY", memoryFrame->memoryWindow->blocky);
		config->Write("MemoryBorder", memoryFrame->memoryWindow->border);
		delete memoryFrame;

		config->Write("MemoryPosX", memposx);
		config->Write("MemoryPosY", memposy);
		config->Write("MemorySizeX", memsizex);
		config->Write("MemorySizeY", memsizey);
	}
	config->Write("DoubleBuffer", doubleBuffer);
	config->Write("PlayOnSwap", playOnSwap);
	config->Write("ErallOnRun", erallOnRun);
	config->Write("SetFocusOnRun", setFocusOnRun);

	config->Write("RingBufferSize", ringBufferSize);

	config->Write("StackvarsShown", stackvars != 0);
	config->Write("VarsShown"     ,      vars != 0);
	config->Write("CallsShown"    ,     calls != 0);
	config->Write("EditorShown"   ,   edFrame != 0);

	config->Write("comspec"  , environment.comspec  );
	config->Write("editor"   , environment.editor   );
	config->Write("fontsdir" , environment.fontsdir );
	config->Write("helpfiles", environment.helpfiles);
	config->Write("logolib"  , environment.logolib  );
	config->Write("tempdir"  , environment.tempdir  );
	config->Write("servicedir", environment.servicedir  );

	editorHide();
	mainHide();

	while (Pending())
		th->dispatch();

	delete locale; locale=0;
}

NodeP lrecreateGUI(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	theApp->action=RECREATE_GUI;
	return Unbound;
}

void LogoApp::mainShow()
{
	if (mainFrame)
		return;
	ThreadP th=Thread::This();
	aui=new Aui;
	th->gtTable=new PtrTable(7);
	int posx=config->Read("PosX", (long int)(0)),
		posy=config->Read("PosY", (long int)(0)),
		sizex=config->Read("SizeX", max.x),
		sizey=config->Read("SizeY", int(max.y*0.95));
	mainFrame=new MainFrame(NULL, ID_MAINFRAME, aUCBLogoVersion, 
		wxPoint(posx, posy), wxSize(sizex, sizey),
		wxDEFAULT_FRAME_STYLE, 
		aUCBLogo, 
		consoleAutosave, caretpos, pictureAutosave, nodesAutosave);
	NEED_MEM_LogoApp_createGUI(mainFrame);
	if (!mainFrame)
		return;
	mainFrame->Show();
	mainFrame->Raise();
//	graph->screen_clear(false);

	graphNode=new GraphNode();
	graphNode->window=WindowP(Thread::This()->graph);
	if (graphNode->graph)
	{	*graphNode->graph->d(Graph::windowNode)=graphNode;
		if (!graph->turtleNode)
			*graph->d(Graph::turtleNode)=new TurtleNode(graph->turtle, graphNode);
		th->gtTable->setitem(
			*graph->d(Graph::windowNode), 
			*graph->d(Graph::turtleNode));
	}
	while (Pending())
		th->dispatch();

#ifdef WIN32
	if (consolew)
		consolew->show(SW_HIDE);
#else
//	TODO
#endif
}

void LogoApp::saveMainConfig()
{
	if (!mainFrame)
		return;
	int posx, posy, sizex, sizey;
	if (mainFrame->IsIconized())
		mainFrame->Iconize(false);
	mainFrame->GetPosition(&posx, &posy);
	mainFrame->GetSize(&sizex, &sizey);
	config->Write("PosX", posx);
	config->Write("PosY", posy);
	config->Write("SizeX", sizex);
	config->Write("SizeY", sizey);

//	config->Write("SplitterRatio" , mainFrame->splitter ->ratio);
//	config->Write("SplitterRatio2", mainFrame->splitter2->ratio);
//	config->Write("SplitterRatio3", mainFrame->splitter3->ratio);
//	config->Write("SplitterRatio4", mainFrame->splitter4->ratio);

	int caretpos=console->GetInsertionPoint();
	config->Write("ConsoleCaretPos", caretpos);
	config->Write("ConsoleAutoSave", console->autosave);
	config->Write("ConsoleHSCROLL", consoleHSCROLL);
	config->Write("PictureAutoSave", mainFrame->graphAutosave);
	config->Write("NodesAutoSave", mainFrame->nodesAutosave);
}

void LogoApp::mainHide()
{
#ifdef WIN32
	if (consolew)
		consolew->show(SW_SHOW);
#else
//	TODO
#endif
	if (mainFrame)
	{	delete mainFrame; mainFrame=0;
		delete aui; aui=0;
	}
}

NodeP lhideLogoMainFrame(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	theApp->action=HIDE_MAIN;
	return Unbound;
}

NodeP lshowLogoMainFrame(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	theApp->mainShow();
	return Unbound;
}

NodeP lhideLogoConsolew(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
#ifdef WIN32
	if (consolew)
		consolew->show(SW_HIDE);
#else
//	TODO
#endif
	return Unbound;
}

NodeP lshowLogoConsolew(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
#ifdef WIN32
	if (consolew)
		consolew->show(SW_SHOW);
#else
//	TODO
#endif
	return Unbound;
}

void LogoApp::loadEditorFiles()
{
	editorShown=true;
	config->Read("EditorShown", &editorShown);
	if (mainShown && editorShown)
		editorShow();
}

void LogoApp::editorShow()
{
	if (!mainShown || edFrame)
	{	wxFocusEvent event;
		edFrame->OnSetFocus(event);
		return;
	}
	int edposx=config->Read("EdPosX", max.x/2),
		edposy=config->Read("EdPosY", (long int)(0)),
		edsizex=config->Read("EdSizeX", max.x/2),
		edsizey=config->Read("EdSizeY", int(max.y*0.96));

	editFontSize=config->Read("EditFontSize", 10);

	edFrame = new EdFrame(mainFrame, aUCBLogo, wxPoint(edposx, edposy), 
		wxSize(edsizex, edsizey), argc, argv);
	if (!edFrame)
	{	error(OUT_OF_MEM); return;
	}
	int fileCount=config->Read("FilesOpen", (long int)(0));
	int i;
	for (i=0;  i < fileCount;  i++)
	{	wxString name;
		name.Printf("File%i", i+1);
		edFrame->FileOpen( config->Read(name, _("noname.lg")) );
	}
	unsigned int selectedPage=config->Read("SelectedPage", (long int)(-1));
	if (selectedPage >= 0 && selectedPage < edFrame->notebook->GetPageCount())
		edFrame->notebook->SetSelection(selectedPage);

	edFrame->Show();

	for (i=0;  i < fileCount;  i++)
	{	wxString name;
		name.Printf("Pos%i", i+1);
		((Edit*)edFrame->notebook->GetPage(i))
			->GotoPosVisible( config->Read(name, (long int)(0)) );
	}
}

void LogoApp::saveEditorConfig()
{
	if (!edFrame)
		return;
//	if (edFrame->IsIconized())
//		edFrame->Iconize(false);
	int posx, posy, sizex, sizey;
	edFrame->GetPosition(&posx, &posy);
	edFrame->GetSize(&sizex, &sizey);
	config->Write("EdPosX", posx);
	config->Write("EdPosY", posy);
	config->Write("EdSizeX", sizex);
	config->Write("EdSizeY", sizey);

	int count=edFrame->notebook->GetPageCount();
	config->Write("FilesOpen", count);

	for (int i=0; i < count;  i++)
	{	Edit *edit=(Edit*)edFrame->notebook->GetPage(i);
		wxString filename=edit->GetFilename();
		wxString name;
		name.Printf("File%i", i+1);
		config->Write(name, filename);

		int pos=edit->GetCurrentPos();
		name.Printf("Pos%i", i+1);
		config->Write(name, pos);
	}
	int selectedPage=edFrame->notebook->GetSelection();
	config->Write("SelectedPage", selectedPage);
	config->Write("EditFontSize", edFrame->fontSize);
}

void LogoApp::editorHide()
{
	if (edFrame)
	{	wxCloseEvent closeEv;
		edFrame->OnClose(closeEv);
	}
}

void LogoApp::varsShow()
{
	if (!mainShown || vars)
		return;
	vars=new Vars(mainFrame, ID_VARS);
	if (!vars)
	{	error(OUT_OF_MEM); return;
	}
}

void LogoApp::varsHide()
{
	if (vars)
	{	wxAuiPaneInfo api=aui->GetPane(vars);
		wxString pi=aui->SavePaneInfo(api);
		theApp->config->Write("VarsPaneInfo", pi);
		aui->DetachPane(vars);
		delete vars;
		aui->Update();
	}
}

void LogoApp::stackvarsShow()
{
	if (!mainShown || stackvars)
		return;
	stackvars=new StackVars(mainFrame, ID_STACKVARS);
	if (!stackvars)
	{	error(OUT_OF_MEM); return;
	}
}

void LogoApp::stackvarsHide()
{
	if (stackvars)
	{	wxAuiPaneInfo api=aui->GetPane(stackvars);
		wxString pi=aui->SavePaneInfo(api);
		theApp->config->Write("StackvarsPaneInfo", pi);
		aui->DetachPane(stackvars);
		delete stackvars;
		aui->Update();
	}
}

void LogoApp::callsShow()
{
	if (!mainShown || calls)
		return;
	calls=new Calls(mainFrame, ID_CALLS);
	if (!calls)
	{	error(OUT_OF_MEM); return;
	}
}

void LogoApp::callsHide()
{
	if (calls)
	{	wxAuiPaneInfo api=aui->GetPane(calls);
		wxString pi=aui->SavePaneInfo(api);
		theApp->config->Write("CallsPaneInfo", pi);
		aui->DetachPane(calls);
		delete calls;
		aui->Update();
	}
}

int LogoApp::OnExit () 
{
	config->Write("MakeBackupFiles", backup);
	config->Write("RunDirectory", runDirectory);
	config->Write("HelpFileName", helpFileName);
	config->Write("ShowGuiOnError", showGuiOnError);
	delete config;
	if (help)
	{	help->Quit();
		delete help;
	}
	if (server)
		delete server;	
	if (heapReserve)
		delete heapReserve;
	if (history)
		delete history;
	if (consolew)
		delete consolew;
	return 0;
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
void LogoApp::doAction()
{
	ThreadP th=Thread::This();
	GraphP graph=th->graph;
	switch (action)
	{
	case EXIT_LOGO:
		if (graph && graph->recorder != graph->recorderSave)
		{	action=EXIT_DELAYED;
			c_out << _(";Warning: EXIT is delayed because ") << endL
			<< _(";the program is in the middle of GraphicStart..GraphicEnd") 
			<< _(";(press control-pause to break)") << endL;
			break;
		}
		if (graph && graph->turtle->polyMode)
		{	action=EXIT_DELAYED;
			c_out << _(";Warning: EXIT is delayed because ") << endL
			<< _(";the program is in the middle of PolyStart..PolyEnd") 
			<< _(";(press control-pause to break)") << endL;
			break;
		}
		if (graph && graph->turtle->tessMode)
		{	action=EXIT_DELAYED;
			c_out << _(";Warning: EXIT is delayed because ") << endL
			<< _(";the program is in the middle of TessStart..TessEnd") 
			<< _(";(press control-pause to break)") << endL;
			break;
		}
		if (graph && graph->turtle->surfaceMode)
		{	action=EXIT_DELAYED;
			c_out << _(";Warning: EXIT is delayed because ") << endL
			<< _(";the program is in the middle of SurfaceStart..SurfaceEnd") 
			<< _(";(press control-pause to break)") << endL;
			break;
		}
	case EXIT_DELAYED:
		if (graph 
		&&	(graph->recorder != graph->recorderSave
		|| graph->turtle->polyMode
		|| graph->turtle->tessMode
		|| graph->turtle->surfaceMode))
			break;
	case RECREATE_GUI:
		if (graph)
			graph->save();
		if (mainFrame && mainFrame->nodesAutosave && !saved)
		{	GC_::save(charP(theApp->nodesFileName.c_str()));
			saved=true;
		}
		Thread::This()->stoppingFlag=THROWING;
		Thread::This()->throw_node = NodeP(_system);
		break;

	case HIDE_MAIN:
		Thread::This()->stoppingFlag=THROWING;
		Thread::This()->throw_node = NodeP(_system);
		break;

	case LOAD_NODES:	
		{	action=NOP;
			StringP name=0;
			{	wxFileDialog fd(mainFrame, _("Choose a Nodes file"),"","",
					_("Nodes Files (*.zip)|*.zip|All Files (*.*)|*.*"),
					wxOPEN|wxCHANGE_DIR);
				if (fd.ShowModal()==wxID_OK)
				name=new String(fd.GetPath());
				if throwing break;
			}
			if (name)
				GC_::load(name->ptr());
			break;
		}
	case SAVE_NODES:	
		{	action=NOP;
			StringP name=0;
			{	wxFileDialog fd(mainFrame, _("Choose a Nodes file"),"","",
					_("Nodes Files (*.zip)|*.zip|All Files (*.*)|*.*"),
					wxSAVE|wxCHANGE_DIR);
				if (fd.ShowModal()==wxID_OK)
				name=new String(fd.GetPath());
				if throwing break;
			}
			if (name)
				GC_::save(name->ptr());
			break;
		}
	default: 
		;
	}
	th->eventDispatch();
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

void LogoApp::prepareToDraw(void)
{
	ThreadP th=Thread::This();
	if (th->graph)
		return;
	if (!mainFrame)
		mainShow();
	if (mainFrame)
	{	if (mainFrame->inGraphicsMode) 
			return;
		else //if (console)
		{//	console->dispatchMessages();
			mainFrame->splitScreen();
		}
	}
}

void LogoApp::memoryShow()
{
	if (memoryFrame || !mainFrame)
		return;
	memoryFrame=new MemoryFrame(mainFrame, -1, _("Memory"), 
		wxPoint(memposx, memposy), wxSize(memsizex,memsizey),
		wxNO_BORDER,
//		wxDEFAULT_FRAME_STYLE,
		_("Memory"), memblockx, memblocky, memborder);
	if (!memoryFrame)
	{	error(OUT_OF_MEM); return;
	}
}	

void LogoApp::memoryHide()
{
	if (memoryFrame)
	{	wxAuiPaneInfo api=aui->GetPane(memoryFrame);
		wxString pi=aui->SavePaneInfo(api);
		theApp->config->Write("MemoryPaneInfo", pi);
		aui->DetachPane(memoryFrame);
		delete memoryFrame;
		aui->Update();
	}
}	


NodeP lLogoInstance(PairPP)
{
	return new Int(int(theApp->GetInstance()));
}

NodeP lyield(PairPP)
{
	ThreadP th=Thread::This();
	th->yield();
	return Unbound;
}

NodeP lcountThreads(PairPP)
{
	IntP nr=new Int(Thread::list.GetCount());
	return nr;
}

NodeP lThreads(PairPP)
{
	PairP list=0;
	ThreadList::Node *node;
	for (node = Thread::list.GetFirst();  node;  node = node->GetNext())
	{	ThreadP th2=node->GetData()->startFiber;
		ThreadP th0=th2;
		while (th2)
		{	if throwing return Unbound;
			list=new Pair(new Int(th2->GetId()), list);
			if throwing return Unbound;
			if (th2->nextFiber==th0)
				break;
			else
				th2=th2->nextFiber;
		}
	}
	return list;
}

bool LogoApp::killThread(int id)
{
	ThreadList::Node *node;
	for (node = Thread::list.GetFirst();  node;  node = node->GetNext())
	{	ThreadP th2=node->GetData()->startFiber;
		ThreadP th0=th2;
		while (th2)
		{	if (th2->GetId()==(unsigned int)id)
			{	th2->throw_node=_toplevel;
				th2->stoppingFlag=THROWING;
				return true;
			}
			if (th2->nextFiber==th0)
				break;
			else
				th2=th2->nextFiber;
		}
	}
	return false;
}

NodeP lkillThread(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->thread)
	{	error(SECURITY_ERROR, new String(_("to use Thread"))); return Unbound;
	}
	if (arg[1] && arg[1]->numberP())
	{	int id=arg[1]->a_int(1);
		if throwing return Unbound;
		theApp->killThread(id);
	}else if (arg[1] && arg[1]->is.list)
	{	PairP list=arg[1];
		while (list)
		{	if (list->circcounter()==th->circcounter)
				break;
			else
				list->setcirccounter(th->circcounter);
			if (list->first())
			{	int id=list->first()->a_int(1);
				if throwing return Unbound;
				theApp->killThread(id);
			}
			next(list);
		}
		th->setcirccounter(th->circcounter+1);
	}else if (arg[1])
	{	error(BAD_DATA, arg[1]);
	}
	return Unbound;
}

NodeP lsetErallOnRun(PairPP arg)
{
	int erallOnRun=arg[1]->a_bool(1);
	if throwing return Unbound;
	theApp->erallOnRun = (erallOnRun==1);
	return Unbound;
}

NodeP lsetFocusOnRun(PairPP arg)
{
	int setFocusOnRun=arg[1]->a_bool(1);
	if throwing return Unbound;
	theApp->setFocusOnRun = (setFocusOnRun==1);
	return Unbound;
}

NodeP lsetShowGuiOnError(PairPP arg)
{
	int showGuiOnError=arg[1]->a_bool(1);
	if throwing return Unbound;
	theApp->showGuiOnError = (showGuiOnError==1);
	return Unbound;
}

bool LogoApp::loadLocale(int language)
{
	if(language == wxLANGUAGE_UNKNOWN) 
		return false;
	wxArrayString dirs, names;
	wxArrayLong identifiers;
	getInstalledLanguages(dirs, names, identifiers);
	for(size_t i = 0; i < identifiers.Count(); i++)
	{	if(identifiers[i] == language)
		{	if(locale) 
				delete locale;
			locale = new wxLocale;
			locale->Init(identifiers[i]);
//c_out << int(identifiers[i]) << ' ' << wxPathOnly(argv[0]) << endL;
			locale->AddCatalogLookupPathPrefix(wxPathOnly(argv[0]));
//c_out << aUCBLogo+wxT(".mo") << endL;
			bool result=locale->AddCatalog(aUCBLogo+wxT(".mo"));
			result=result;
//c_out << "loaded=" << int(result) << endL;
			langdir=dirs[i];
			return true;
		}
	}
	return false;
}

void LogoApp::getInstalledLanguages(wxArrayString &dirs, 
												wxArrayString &names, 
											   wxArrayLong &identifiers)
{
	names.Clear();
	identifiers.Clear();
	wxDir dir(wxPathOnly(argv[0]));
	wxString filename;	
	const wxLanguageInfo *langinfo;	
	wxString name = wxLocale::GetLanguageName(wxLANGUAGE_DEFAULT);
	if(!name.IsEmpty())
	{	dirs.Add("");
		names.Add(_("Default"));
		identifiers.Add(wxLANGUAGE_DEFAULT);		
	}
	for(bool cont = dir.GetFirst(&filename, wxEmptyString, wxDIR_DIRS); 
		cont; cont = dir.GetNext(&filename))
	{
//c_out << _("wxTranslationHelper: Directory found = \"")
//	<< filename.GetData() << endL;
		langinfo = wxLocale::FindLanguageInfo(filename);
		if(langinfo != NULL)
		{
			if(wxFileExists(dir.GetName()+wxFileName::GetPathSeparator()+
				filename+wxFileName::GetPathSeparator()+
				aUCBLogo+wxT(".mo")))
			{
				dirs.Add(dir.GetName()+wxFileName::GetPathSeparator()+
				filename+wxFileName::GetPathSeparator());

				names.Add(langinfo->Description);
				identifiers.Add(langinfo->Language);
			}
		}
	}
}

bool LogoApp::askUserForLanguage(wxArrayString &dirs, 
											wxArrayString &names, 
											wxArrayLong & identifiers)
{
	ThreadP th=Thread::This();
	wxCHECK_MSG(names.Count() == identifiers.Count(), false, 
		_("Array of language names and identifiers should have the same size."));
	long index = wxGetSingleChoiceIndex(_("Select the language"), 
			_("Language"), names);
	if(index != -1)
	{	
		StringP unload=new String(langdir+_("unload"));
		unload->Node::silent_load(0);
		if throwing
		{	if (th->throw_node==_error) error_print();
			th->stoppingFlag=RUN;
			th->inpause=false;
		}
		if (stopping || outputting)
		{	c_out << _("; You must be in a procedure to use OUTPUT or STOP.") << endL;
			th->stoppingFlag=RUN;
		}
		th->argStack->clean();
		
//		for (int i=0; i<100; i++)
			while (Pending())
				th->dispatch();

		if(locale)
		{	delete locale;
		}
		locale = new wxLocale;
		locale->Init(identifiers[index]);
		locale->AddCatalogLookupPathPrefix(wxPathOnly(argv[0]));
//c_out << _("wxTranslationHelper: Path Prefix = \"")
//	<< wxPathOnly(argv[0]) << endL;
		locale->AddCatalog(aUCBLogo+wxT(".mo"));
//c_out << _("wxTranslationHelper: Catalog Name = \"")
//	<< aUCBLogo << endL;
		langdir=dirs[index];
		action=RECREATE_GUI;
		return true;
	}
	return false;
}

NodeP lwait(PairPP arg) 
{
	ThreadP th=Thread::This();
	IntP i = arg[1]->a_pos_int(1);

	if throwing return Unbound;

	if (console)
		console->flush();

	if (i->int_ > 0) 
	{	int n=i->int_;
		while (n >= 60) 
		{
#ifdef WIN32
			Sleep(1000);
#else
			sleep(1);
#endif
			n-=60;
			th->dispatchMessages();
			if (checkThrowing()) n=0;
		}
		if (n > 0)
#ifdef WIN32
			Sleep(n*1000/60);
#else
			usleep(n*1000000/60);
#endif
	}
	return Unbound;
}

NodeP lwaitMS(PairPP arg) 
{
	IntP i = arg[1]->a_pos_int(1);

	if throwing return Unbound;

	ThreadP th=Thread::This();

	if (console)
		console->flush();

	if (i->int_ > 0) 
	{	int n=i->int_;
		while (n >= 1000) 
		{
#ifdef WIN32

			if (th->IsMain())
				Sleep(1000);
			else
				th->Sleep(1000);
#else
			sleep(1);
#endif
			n-=1000;
			th->dispatchMessages();
			if (checkThrowing()) n=0;
		}
		if (n > 0)
#ifdef WIN32
			if (th->IsMain())
				Sleep(n);
			else
				th->Sleep(n);
#else
			usleep(n*1000);
#endif
	}
	return Unbound;
}

void LogoApp::waituS(int n)
{
	ThreadP th=Thread::This();
	while (n >= 1000000) 
	{
#ifdef WIN32
		Sleep(1000);
#else
		sleep(1);
#endif
		n-=1000000;
		th->dispatchMessages();
		if (checkThrowing()) n=0;
	}
#ifdef WIN32
	while (n >= 100000) 
	{
		Sleep(100);
		n-=100000;
		if (checkThrowing()) n=0;
	}
#endif
	if (n > 0)
#ifdef WIN32
	{	TimeUnion f,c;
		QueryPerformanceCounter(&c.i);
		QueryPerformanceFrequency(&f.i);
		if (!f.u)
			return;

		double t,t0=double(c.u)/f.u;
		do
		{	QueryPerformanceCounter(&c.i);
			t=double(c.u)/f.u;
		} while (1000000*(t-t0) < n);

	}
#else
		usleep(n);
#endif
}

NodeP lwaituS(PairPP arg) 
{
	IntP i = arg[1]->a_pos_int(1);

	if throwing return Unbound;

	if (i->int_ > 0) 
	{	int n=i->int_;
		theApp->waituS(n);
	}
	return Unbound;
}
/*
NodeP t0=lTimeFine(0);
NodeP t=lTimeFine(0);
c_out << t->sub(t0) << "seconds" << endL;
*/
NodeP lTimeFine(PairPP)
{
#ifdef WIN32
	TimeUnion f,c;
	QueryPerformanceFrequency(&f.i);
	QueryPerformanceCounter(&c.i);
	if (f.u)
		return new Float(double(c.u)/f.u);
	else
		return Unbound;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	return new Float(tv.tv_sec+tv.tv_usec/1000000.0);
#endif
}

NodeP lTimeMilli(PairPP)
{

#ifdef WIN32
	TimeUnion f,c;
	QueryPerformanceFrequency(&f.i);
	QueryPerformanceCounter(&c.i);
	if (f.u)
		return new Float(1000*double(c.u)/f.u);
	else
		return Unbound;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	return new Float(1000.0*tv.tv_sec+tv.tv_usec/1000.0);
#endif
}

NodeP lTimeu(PairPP)
{
#ifdef WIN32
	TimeUnion f,c;
	QueryPerformanceFrequency(&f.i);
	QueryPerformanceCounter(&c.i);
	if (f.u)
		return new Float(1000000*double(c.u)/f.u);
	else
		return Unbound;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	return new Float(1000000.0*tv.tv_sec+tv.tv_usec);
#endif
}

NodeP lTimerFreq(PairPP)
{
#ifdef WIN32
	TimeUnion f;
	QueryPerformanceFrequency(&f.i);
	if (f.u)
		return new Float(double(f.u)/1000000);
	else
		return Unbound;
#else
	return new Int(1);
#endif
	return Unbound;
}

NodeP lTimeuRes(PairPP)
{
#ifdef WIN32
	TimeUnion f;
	QueryPerformanceFrequency(&f.i);
	if (f.u)
		return new Float(1000000./f.u);
	else
		return Unbound;
#else
	return new Int(1);
#endif
}

NodeP lMIPS(PairPP)
{
#ifdef WIN32
	TimeUnion f,c0, c;
	QueryPerformanceFrequency(&f.i);
	if (f.u)
	{	QueryPerformanceCounter(&c0.i);
		_asm
		{	mov eax, 100000000
		again:
			dec eax
			jnz again
		}
		QueryPerformanceCounter(&c.i);
		return new Float(200*double(f.u)/(c.u-c0.u));
	}else
		return Unbound;
#else
	timeval tv, tv2;
	gettimeofday(&tv, 0);
	asm("movl $100000000,%eax;  again:  decl %eax;  jnz again");
	gettimeofday(&tv2, 0);
	return new Float(200/(tv2.tv_sec-tv.tv_sec+(tv2.tv_usec-tv.tv_usec)/1000000.0));
#endif
	return Unbound;
}

NodeP lTime(PairPP)
{
	wxDateTime t=wxDateTime::UNow();
	return new List(
		new Int(t.GetHour()), new Int(t.GetMinute()), new Int(t.GetSecond()), 
		new Int(t.GetMillisecond()), 
		new Int(t.GetWeekDay()), 
		new Int(t.GetDay()), new Int(t.GetMonth()+1), new Int(t.GetYear()), _ListEnd);
	// There's a bug in GetMonth(), so we need the +1!
}

void LogoApp::Sound(double frequency, double duration)
{
	if (frequency < 19) frequency=19;

	switch (osVersion)
	{
#ifdef WIN32
	case wxOS_WINDOWS_NT:
		Beep((int)frequency, duration); break;
/*		{	int count = (int)(1193180/frequency);
			PortOut(0x43, char(0xB6));
			PortOut(0x42,  count     && 0xff);
			PortOut(0x42, (count>>8) && 0xff);
			int bits=PortIn(0x61);
			PortOut(0x61, bits|3);
			if (duration!=INFINITE)
			{	waituS(duration*1000);
				int bits=PortIn(0x61);
				PortOut(0x61, bits & 0xF8);
			}
			break;
		}
*/
	case wxOS_WINDOWS:
		int count = (int)(1193180/frequency);
		_asm 
		{	mov al, 0xB6 // set timer 
			out 0x43, al
			mov eax, count
			out 0x42, al
			mov al, ah
			out 0x42, al
			in al, 0x61
			or al, 3 // set speaker bits
			out 0x61, al // sound on.
		}
		if (duration!=INFINITE)
		{	Sleep(duration);
			_asm 
			{	in al, 0x61
				and al, 0xf8 // reset speaker bits
				out 0x61, al // sound off.
			}
		}
#endif
	}
}

void LogoApp::Tone(double pitch, double duration)
{
	double frequency = 440*pow(2, pitch/12);
	Sound(frequency, duration);
}

NodeP lSound(PairPP arg) 
{
	double freq=arg[1]->a_float(1),
		duration=arg[2]->a_float(2);
	if throwing return Unbound;
	theApp->Sound(freq, duration);
	return Unbound;
}

NodeP lTone(PairPP arg) 
{
	double pitch=arg[1]->a_float(1),
		 duration=arg[2]->a_float(2);
	if throwing return Unbound;
	theApp->Tone(pitch, duration);
	return Unbound;
}

class SoundsThread: public wxThread
{
	int length;
	double *freq, *duration;
public:
	static volatile int finished;
	static volatile bool stop;

	SoundsThread(int length_, double *freq_, double *duration_): 
		length(length_), freq(freq_), duration(duration_) 
	{}

	ExitCode Entry()
	{	double *f=freq, *d=duration;
		for (int i=length;  i>0;  i--)
		{	theApp->Sound(*f++, *d++);
			if (stop)
				break;
		}
		finished++;
		stop=false;
		delete freq;
		delete duration;
		return 0;
	}
};

volatile int SoundsThread::finished=0;
volatile bool SoundsThread::stop=false;

NodeP lSounds(PairPP arg) 
{
	int length=0;
	double *freq=0, *duration=0;

	if (exist(2))
	{	PairP flist=arg[1]->a_list(1);
		if throwing return Unbound;
		NodeP	dnode=arg[2];
		if (!dnode)
		{	error(BAD_DATA, arg[2], 2);
			return Unbound;
		}
		length=flist->count();
		freq=new double[length]; double *f=freq;
		duration=new double[length]; double *d=duration;

		if (!f || !d)
		{	error(OUT_OF_MEM); return Unbound;
		}
		if (dnode->islist())
		{	PairP dlist=PairP(dnode);
			while (flist && flist->first() 
				 && dlist && dlist->first())
			{	*f++=flist->first()->a_float(1);
				*d++=dlist->first()->a_int(1);
				if throwing return Unbound;
				next(flist);
				next(dlist);
			}
		}
		else if (dnode->numberP())
		{	double di=dnode->getnumber();
			while (flist && flist->first())
			{	*f++=flist->first()->a_float(1);
				*d++=di;
				if throwing return Unbound;
				next(flist);
			}
		}
	}else
	{	PairP list=arg[1]->a_list(1);
		if throwing return Unbound;
		length=list->count()/2;
		freq=new double[length]; double *f=freq;
		duration=new double[length]; double *d=duration;
		if (!f || !d)
		{	error(OUT_OF_MEM); return Unbound;
		}
		while (list && list->first() && list->tail() && list->second())
		{	*f++=list->first()->a_float(1);
			*d++=list->second()->a_int(1);
			if throwing return Unbound;
			next(list);
			next(list);
		}
	}
	if throwing return Unbound;
	SoundsThread *t=new SoundsThread(length, freq, duration);
	if (!t)
	{	error(OUT_OF_MEM); return Unbound;
	}
	t->Create(0x100);
	t->Run();
	return Unbound;
}

NodeP lSoundsStop(PairPP)
{
	SoundsThread::stop=true;
	return Unbound;
}

NodeP lSoundsFinished(PairPP)
{
	return new Int(SoundsThread::finished);
}

class TonesThread: public wxThread
{
	int length;
	double *pitch, *duration;
public:
	static volatile int finished;
	static volatile bool stop;

	TonesThread(int length_, double *pitch_, double *duration_): 
		length(length_), pitch(pitch_), duration(duration_) 
	{}

	ExitCode Entry()
	{	double *p=pitch, *d=duration;
		for (int i=length;  i>0;  i--)
		{	theApp->Tone(*p++, *d++);
			if (stop)
				break;
		}
		finished++;
		stop=false;
		delete pitch;
		delete duration;
		return 0;
	}
};

volatile int TonesThread::finished=0;
volatile bool TonesThread::stop=false;

NodeP lTones(PairPP arg) 
{
	int length=0;
	double *pitch=0, *duration=0;

	if (exist(2))
	{	PairP plist=arg[1]->a_list(1);
		if throwing return Unbound;
		NodeP	dnode=arg[2];
		if (!dnode)
		{	error(BAD_DATA, arg[2], 2);
			return Unbound;
		}
		length=plist->count();
		pitch=new double[length]; double *p=pitch;
		duration=new double[length]; double *d=duration;
		if (!p || !d)
		{	error(OUT_OF_MEM); return Unbound;
		}
		if (dnode->islist())
		{	PairP dlist=PairP(dnode);
			while (plist && plist->first() 
				 && dlist && dlist->first())
			{	*p++=plist->first()->a_float(1);
				*d++=dlist->first()->a_int(1);
				if throwing return Unbound;
				next(plist);
				next(dlist);
			}
		}
		else if (dnode->numberP())
		{	double di=dnode->getnumber();
			while (plist && plist->first())
			{	*p++=plist->first()->a_float(1);
				*d++=di;
				if throwing return Unbound;
				next(plist);
			}
		}
	}else
	{	PairP list=arg[1]->a_list(1);
		if throwing return Unbound;
		length=list->count()/2;
		pitch=new double[length]; double *p=pitch;
		duration=new double[length]; double *d=duration;
		if (!p || !d)
		{	error(OUT_OF_MEM); return Unbound;
		}
		while (list && list->first() && list->tail() && list->second())
		{	*p++=list->first()->a_float(1);
			*d++=list->second()->a_int(1);
			if throwing return Unbound;
			next(list);
			next(list);
		}
	}
	if throwing return Unbound;
	TonesThread *t=new TonesThread(length, pitch, duration);
	if (!t)
	{	error(OUT_OF_MEM); return Unbound;
	}
	t->Create(0x100);
	t->Run();
	return Unbound;
}

NodeP lTonesStop(PairPP)
{
	TonesThread::stop=true;
	return Unbound;
}

NodeP lTonesFinished(PairPP)
{
	return new Int(TonesThread::finished);
}

NodeP lplayWave(PairPP arg)
{
#ifdef WIN32
	int flags=arg[2]->a_int(2);
	if (arg[1] && type_of(arg[1])==INT16ARRAY)
	{	PlaySound(charP(Int16ArrayP(arg[1])->data()), 
			0, (flags|SND_MEMORY) & ~SND_FILENAME);
		return Unbound;
	}
	else if (arg[1] && type_of(arg[1])==INT8ARRAY)
	{	PlaySound(charP(Int8ArrayP(arg[1])->data()), 
			0, (flags|SND_MEMORY) & ~SND_FILENAME);
		return Unbound;
	}
	else if (arg[1] && type_of(arg[1])==UINT8ARRAY)
	{	PlaySound(charP(UInt8ArrayP(arg[1])->data()), 
			0, (flags|SND_MEMORY) & ~SND_FILENAME);
		return Unbound;
	}
	else if (arg[1] && type_of(arg[1])==INTARRAY)
	{	PlaySound(charP(IntArrayP(arg[1])->data()), 
			0, (flags|SND_MEMORY) & ~SND_FILENAME);
		return Unbound;
	}
	StringP filename=0;
	if (arg[1])
		filename=arg[1]->a_string(1);
	if (!throwing)
	{	if (arg[1])
			PlaySound(filename->ptr(), 0, (flags|SND_FILENAME) & ~SND_MEMORY);
		else
			PlaySound(0, 0, flags);
	}
#else
	StringP filename=0;
	if (arg[1])
		filename=arg[1]->a_string(1);
	if (!throwing)
		if (arg[1])
		{	wxSound s(filename->ptr());
			if (s.IsOk())
			{	wxSound::Stop();
				s.Play(wxSOUND_SYNC);//unsigned(flags));
			}
		}
		else
			wxSound::Stop();
#endif
	return Unbound;
}


#ifdef WIN32

bool inWaveOutProc=false;

struct WaveInfo
{
	HWAVEOUT hwo;
	WAVEHDR *wh;
	bool done;
	charP data;
	TimeUnion time;
};

const int WaveInfoNr=128;
WaveInfo waveInfo[WaveInfoNr]={0};

void WaveInfoDestroy()
{
	for (int i=0; i<WaveInfoNr; i++)
	{	WaveInfo &wi=waveInfo[i];
		if (wi.wh)
			delete wi.wh;
	}
}

void CALLBACK WaveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD index, 
								  DWORD wh, DWORD dwParam2) 
{ 
	switch(uMsg) 
	{	case WOM_OPEN: break; 
		case WOM_DONE: 
		{	inWaveOutProc=true;
			WaveInfo &wi=waveInfo[index];
			wi.done=true;
			inWaveOutProc=false;
			break;
		}
		case WOM_CLOSE:
			break; 
	}
}

bool PlayWaveFast(charP data, int length) 
{
	WAVEHDR *wh=0;
	HWAVEOUT hwo; 
	bool samedata=false;

	while (inWaveOutProc)
		Sleep(10);

	for (int i=0; i<WaveInfoNr; i++)
	{	WaveInfo &wi=waveInfo[i];
		if (data==wi.data && wi.done)
		{	wh=wi.wh;
			hwo=wi.hwo;
			wi.done=false;
			QueryPerformanceCounter(&wi.time.i);
			break;
		}
	}
	int OFFSET_WAVEDATA=44, 
		HEADER_SIZE=OFFSET_WAVEDATA;

	WAVEFORMATEX *wf=(WAVEFORMATEX*)&data[20];
	int cbSize=128;
//	if (wf->wFormatTag != WAVE_FORMAT_PCM)
//		cbSize=wf->cbSize;

	if (!wh)
	{	
		int index=-1;
		for (int i=0; i<WaveInfoNr; i++)// try to find an empty waveInfo first
		{	WaveInfo &wi=waveInfo[i];
			if (!wi.hwo) 
			{	index=i;
				break;
			}
		}
		__int64 t=-1;
		while (index < 0)	// no empty found, so search the oldest waveInfo
		{	for (int i=0; i<WaveInfoNr; i++)
			{	WaveInfo &wi=waveInfo[i];
				if (wi.time.u < t) 
				{	index=i;
					t=wi.time.u;
				}
			}
			if (index < 0)
				Sleep(10);
		}
		if (index >= 0)
		{	WaveInfo &wi=waveInfo[index];
			if (wi.hwo && !wi.done) 
			{	if (waveOutUnprepareHeader(wi.hwo, wi.wh, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) 
					return true; 
				wh=wi.wh;
				hwo=wi.hwo;
				goto weida;
			}	
		}
		wh = new WAVEHDR; 
		if (!wh)
		{	error(OUT_OF_MEM); return true;
		}
		ZeroMemory((void*)wh, sizeof(WAVEHDR)); 
	
		wh->lpData = data+OFFSET_WAVEDATA; 
		wh->dwBufferLength = length-HEADER_SIZE-cbSize; 
		
		// Find a waveOut device and open it 
		int devs=waveOutGetNumDevs();
		for(int devid = 0; devid < devs; devid++) 
		{
			if(devid == devs) // Error, no free devices found 
				return true; 
			if(waveOutOpen(&hwo, WAVE_MAPPER, wf, 
					unsigned long(WaveOutProc), index, 
					CALLBACK_FUNCTION) == MMSYSERR_NOERROR) 
			{
				WaveInfo &wi=waveInfo[index];
				wi.hwo=hwo;
				wi.wh=wh;
				wi.data=data;
				wi.done=false;
				QueryPerformanceCounter(&wi.time.i);
				break; 	// Usable device found, stop searching 
			}
		} 
	}
weida:
	wh->lpData = data+OFFSET_WAVEDATA; 
	wh->dwBufferLength = length-HEADER_SIZE-cbSize; 
	wh->dwFlags=0;

	if (waveOutPrepareHeader(hwo, wh, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) 
		return true; 

	if (waveOutWrite(hwo, wh, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) 
		return true; 
	return false; 
}
#endif

NodeP lplayWaveFast(PairPP arg)
{
#ifdef WIN32
	if (arg[1] && type_of(arg[1])==INT16ARRAY)
	{	PlayWaveFast(charP(Int16ArrayP(arg[1])->data()), 
			Int16ArrayP(arg[1])->length()*2);
	}
	else if (arg[1] && type_of(arg[1])==INT8ARRAY)
	{	PlayWaveFast(charP(Int8ArrayP(arg[1])->data()), 
			Int8ArrayP(arg[1])->length());
	}
	else if (arg[1] && type_of(arg[1])==UINT8ARRAY)
	{	PlayWaveFast(charP(UInt8ArrayP(arg[1])->data()), 
			UInt8ArrayP(arg[1])->length());
	}
	else if (arg[1] && type_of(arg[1])==INTARRAY)
	{	PlayWaveFast(charP(IntArrayP(arg[1])->data()), 
			IntArrayP(arg[1])->length()*4);
	}
#endif
	return Unbound;
}

//___________________________________________________________________
//
//                        Logo Main Run loop
//___________________________________________________________________

#define NEED_MEM_LogoApp_MainLoop(x)					\
	if (!x)														\
	{	delete heapReserve;									\
		heapReserve=0;											\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x << _(": Insufficient Memory");			\
		wxMessageBox(msg,	_("aUCBLogo Error"));		\
		return false;											\
	}

bool LogoApp::OnExceptionInMainLoop()
{
	return false;
}

int LogoApp::MainLoop()
{
	ThreadP th=Thread::This();
	wxEventLoop *eventLoop = new wxEventLoop,
		*oldEventLoop=wxEventLoop::GetActive();
	NEED_MEM_LogoApp_MainLoop(eventLoop);
	wxEventLoop::SetActive(eventLoop);

	while (Pending())
		th->dispatch();
	th->processIdle();
	
#ifdef EAT_MEM
	try {
		eat_mem_mem=charP(calloc(eat_mem_size, 1));
		_main(argc, argv);  //////////  main  ///////////
	}
	catch (...)
	{
		free(eat_mem_mem);
		throw;
	}
	free(eat_mem_mem);
#else
	_main(argc, argv);  //////////  main  ///////////
#endif
#ifdef WIN32
	UnloadIODLL();
	pm_term();
#endif
	destroyGUI();

	wxEventLoop::SetActive(oldEventLoop);
	delete eventLoop;
	theApp=0;
#ifdef MIDI
	delete midiSystem;
#endif
	delete platformInfo;
#ifdef WIN32
	CloseHandle(Thread::mainHandle);
#endif
	delete Thread::main;	Thread::main=0;
	return 0;
}

#ifdef LOGODLL

Export int __cdecl logomain(int argc, char **argv, void (__cdecl *callback)())
{
	wxString args;
	for(int i=1;i<argc;++i)
	{	if(!args.empty())
		{	args+=" ";
		}
		args+="\"";
		args+=argv[i];
		args+="\"";
	}
	char *cmdline=strdup(args.c_str());
	LogoApp::callback=callback;
	int r=WinMain(GetModuleHandle(0) ,0 , cmdline, SW_SHOW);
	GC_::free(cmdline); cmdline=0;
	return 0;//r;
}

#ifdef WIN32
BOOL WINAPI DllMain(HANDLE hInst, 
                    ULONG ul_reason_for_call,
                    LPVOID lpReserved)
{
	return TRUE;
}
#else
//	TODO
#endif

#else

#ifdef WIN32
int __cdecl main(int argc, charP *argv)
#else
int __cdecl main(int argc, charP *argv, charP *env)
#endif
{
	wxString args;
	for(int i=1;i<argc;++i)
	{	if(!args.empty())
		{	args+=" ";
		}
		args+="\"";
		args+=argv[i];
		args+="\"";
	}
#ifdef WIN32
	char *cmdline=_strdup(args.c_str());
#else
	char *cmdline=strdup(args.c_str());
#endif
#ifndef WIN32
	environment.env=env;
#endif
#ifdef WIN32
	int r=WinMain(GetModuleHandle(0) ,0 , cmdline, SW_SHOW);
#else
	int r=wxEntry(argc, argv);
#endif
	r=r;
	GC_::free(cmdline); cmdline=0;
	return 0;//r;
}

#endif
