#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/wfstream.h>
#include <wx/zipstrm.h>
#include <wx/ptr_scpd.h>
#include <wx/glcanvas.h>
#include <wx/textctrl.h>
#include <wx/frame.h>
#include <wx/msw/threadinternal.h>

#include <GL/glu.h>

#include <m_apm.h>

#include "Main.h"

#include "Segment.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "List.h"
#include "Procedur.h"
#include "Object.h"
#include "Primitiv.h"
#include "Table.h"
#include "Struct.h"
#include "Tree.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "Complex.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Mem.h"

#include "PtrToString.h"
#include "StringToPtr.h"

#include "Error.h"
#include "NodePStack.h"
#include "Stack.h"
#include "Eval.h"
#include "ToDoStack.h"
#include "Init.h"
#include "Print.h"

#include "Event.h"
#include "Color.h"
#include "Point.h"
#include "Pen.h"
#include "Font.h"
#include "Screen.h"
#include "Turtle.h"
#include "Recorder.h"
//#include "AVIGen.h"
#include "Graph.h"
#include "Console.h"
#include "MemoryWindow.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"
#include "DebugState.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)

	#define ROOT_SIZE 1021
#else
	#define ROOT_SIZE 1021	// root_size should be a prime number to achieve a fine random distribution
#endif

TableP root;	// All logo objects are stored in the root table

static wxCriticalSection gcCS;

NodeP Unbound=0;
IntP Tree::the_generation=0;

//#define GC_DEBUG 1
//#define SEG_SIZE_ONE
//#define GC_ON_NEWNODE

const unsigned int
	MemBlockSize=0x10,	// might be compiler-specific!!!
	MemBlockMask=~(MemBlockSize-1);

const int
#ifdef SEG_SIZE_ONE
	GC_::segment_size=1,	GC_::fifo_size=100,	// Xtreme GC debugging ;-)))
#else
	#ifdef _DEBUG
		GC_::segment_size=0x4000,	GC_::fifo_size=0x4000,
	#else
//		GC_::segment_size=0x100000,	GC_::fifo_size=0x100000,	// 1 MN
		GC_::segment_size=0x10000,	GC_::fifo_size=0x10000,	// 64 k Nodes
	#endif
#endif

	// GC heuristic parameters. These parameters can be modified to fine tune
	// the performance of the GC program. The values below are a good set of
	// default parameters that should work well for most data.

	GC_::min_free = 1+GC_::segment_size*4/10,	// if fewer than min_free nodes are free after a GC run,
												 // then a new segment of nodes is allocated
	GC_::generations = 4,
												// If a node of generation gen gets older than age_threshold[gen],
	gc_age_threshold[GC_::generations] =		// then it will be promoted to an older generation.
		//{0,0,0,0};	// to deactivate generational GC
		//{1,1,1,1};
		{2,2,2,2};
		//{4,4,4,4};	// old values
		//{3, 11, 37, 113};	// optimization experiment with prime numbers

NodeP generation[GC_::generations]={0};	// list of nodes in the same generation

//___________________________________________________________________

SegmentP Segment::list=0, Segment::start=0;

Segment::Segment()
{
	base = (Object*)::calloc(GC_::segment_size, sizeof(Object));
}

Segment::~Segment()
{
	free((Node*)base);
}

bool Segment::add()
{
	SegmentP newseg = new Segment;
	if ((newseg!=0) && newseg->base) 
	{	newseg->baseu=unsigned(newseg->base);
		newseg->maxu=unsigned(&newseg->base[GC_::segment_size-1]);

		newseg->_tail = list;
		if (list)
			list->prev = newseg;
		newseg->prev = 0;
		list=newseg;
		if (!start)
			start=newseg;

		for (int p=GC_::segment_size-1;  p >=0;  p--)
		{	newseg->base[p]._next = GC_::free_node;
			GC_::free_node = &newseg->base[p];
			GC_::free_node->type = NTFREE;
			*intP((Node*)GC_::free_node) = Node::vmtNode;
		}
		GC_::mem_free += GC_::segment_size;
		if (memoryWindow)
		{	wxPaintEvent pev;
			memoryWindow->OnPaint(pev);
		}
		return true;
	}else
		return false;
}

#ifdef WIN32
extern void WaveInfoDestroy();
#endif

void Segment::freeAll()
{
	theApp->mainFrame->initialized=false;
#ifdef WIN32
	WaveInfoDestroy();
#endif
	SegmentP seg=list;
	while (seg)
	{	for (int p=0;  p < GC_::segment_size;  p++)
		{	NodeP n = &seg->base[p];
			if (n)
				n->free();
		}
		next(seg);
	}
	seg=list;
	while (seg)
	{	SegmentP s=seg;
		next(seg);
		delete s;
	}
	m_apm_free_all_mem();
	debugStates.free();
}

bool is_nil_or_ok(volatile Node *nd)
{
	if (!nd) return true;
	unsigned ptr = unsigned(nd);

	for (SegmentP seg=Segment::list;  seg;  next(seg)) 
	{
#ifdef SEG_SIZE_ONE
		if (ptr == seg->baseu)
			return true;
#else
		if (   ptr >= seg->baseu
		&&   ptr <= seg->maxu
		&&  (ptr - seg->baseu) % sizeof(Object) == 0)
			return true;
#endif
	}
	return false;
}

//#pragma optimize("", off)

bool is_valid_pointer(volatile Node *nd)
{
	if (!nd) return false;
	unsigned ptr = unsigned(nd);

	for (SegmentP seg=Segment::list;  seg;  next(seg)) 
	{
#ifdef SEG_SIZE_ONE
		if (ptr == seg->baseu)
			return nd->type != NTFREE;
#else
		if (   ptr >= seg->baseu
		&&   ptr <= seg->maxu
		&&  (ptr - seg->baseu) % sizeof(Object) == 0 )
			return nd->type != NTFREE;
#endif
	}
	return false;
}

//#pragma optimize("", on)

#ifdef GC_DEBUG

	#define print_gc(text, var)			  \
		File::print(cout.file(), text, var);\
		if (dribbler)								 \
			fprintf(dribbler.file(), text, var); \
		examined=0;										\

	#define print_gc_object							 \
		File::print(conout,							  \
			"%s p=%s v=%s mark=%2i gcNo.=%2i \n",	\
			obj, obj->proc, obj->val,					 \
			obj->mark_gc, gc_run_numero);				  \

	#define inc_examined	examined++
	#define freed_sofar_init	int freed_sofar=0
	#define freed_sofar_is(n)	freed_sofar=(n)
#else
	#define print_gc(text, var)
	#define print_gc_object
	#define inc_examined
	#define freed_sofar_init
	#define freed_sofar_is(n)
#endif

//___________________________________________________________________

FiFo::~FiFo() {if (allocated) free(base);}

FiFo &FiFo::operator<<(const NodeP val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

FiFo &FiFo::operator>>(NodeP &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}

void FiFo::expand()	// transfer is untested!!!!
{
	NodePP newbase = (NodeP*)calloc(1, sizeof(NodeP)*(size+expand_size));

	if (newbase==NodePP(0))	// no room to increase FiFo
	{	c_out << "; Warning: Not enough memory to run garbage collector." << endL;
		c_out << "; GC disabled - Save important data and exit!" << endL;
		_overflow=true;
	}
	else
	{	NodePP i=newbase;

		if (out)			// transfer old FiFo to new FiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;

		}				
		if (allocated) 
			free(base);

		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}

//___________________________________________________________________

int Table::maxStringLength=65536;

Table::Table(int size_): 
	Node(TABLE), allocated(false), size(0), base(0)
{
	self_resize(size_);
}

Table::~Table() 
{
	if (allocated) 
	{	GC_::free((NodeP*)base); 
		allocated=false;
	}

}

void Table::free() 
{
	if (allocated) 
	{	GC_::free((NodeP*)base); 
		allocated=false;
	}
}

void Table::inc_gc()
{
	if (!allocated) return;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			fifo << obj;
}

int Table::youngs() 
{
	if (!allocated) return 0;

	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmark(obj);
	return youngs;
}

int Table::youngsPtr(PtrToInt_ &ptrToInt) 
{
	if (!allocated) return 0;

	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmarkPtr(obj, ptrToInt);
	return youngs;
}

void Table::check_oldyoungs()
{
	if (!allocated) return;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			check_oldyoung(obj);
}

void Table::setPtrs(IntToPtr_ &intToPtr)
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
	{	*itemptr=(Object*)intToPtr.find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
wxDataInputStream &Table::read(wxDataInputStream &istr) 
{	
	int siz;
	istr >> allocated >> siz;
	size=0;
	base = (ObjectP*) GC_::calloc(sizeof(ObjectP), siz+1);
	if (!base)
		return istr;
	size=siz;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Table::write(wxDataOutputStream &ostr) 
{
	ostr << type << allocated << size;

	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;
}

NodeP Table::copy(PtrToInt_ &ptrToInt)
{
	TableP tab=new Table(size);
	ObjectPP itemptr=base, itemptr2=tab->base;
	for (int i=size;  i;  i--)
		*itemptr2++ = ObjectP(ptrToInt.find(*itemptr++));
	return tab;
}

StringP Table::string()
{
	char textbuf[65536];
	ThreadP th=Thread::This();
	th->stringptr=textbuf;
	th->stringlen=maxStringLength-1;
	stringbuffer << NodeP(this);
	*th->stringptr='\0';
	return new String(textbuf);
}

void Table::print(FILEP strm, int width, int depth)
{
	PairP list=table()->contentsNameVar(VALS);
	if (list)
		list->print(strm, width, depth);
}

void Table::printBin(FILEP strm)
{
	PairP list=contentsNameVar(VALS);
	if (list)
		list->printBin(strm);
}

int Table::hash(StringP str)
{
	if (!str)
		str=_Null_Word->name();
	unsigned int *i= (unsigned int*) str->ptr(),   index=0;

	for (char len=char(str->length());  len>0;  len-=sizeof(int))
	{	switch (len)
		{	case 0:	break;
			case 1:	index^= *i & 0x000000ff;	break;
			case 2:	index^= *i & 0x0000ffff;	break;

			case 3:	index^= *i & 0x00ffffff;	break;
			case 4:	index^= *i;	break;
			default:	index^= *i++ << 5;
		}
	};
	index %= size;
	return index;
}

ObjectP Table::find(StringP str) 
{
	ObjectP obj;
	for (obj=base[hash(str)];  obj;  next(obj))
		if (str->equal(obj->name()))
			break;

	return obj;
}

ObjectP Table::operator[](StringP str) {return find(str);}


ProcedureP Table::findProc(StringP filename, int linenr)
{
	int maxlinenr=-1;
	ObjectP obj=0;
	ObjectPP itemptr=base;
	LineP line=0;
	ProcedureP result=0;

	for (int i=size;  i;  i--)
		for (obj = *itemptr++;  obj;  next(obj))
			if (obj->proc() && obj->proc()!=Unbound && !obj->bit.permanent)
			{	line=LineP(obj->proc()->bodylist());
				if (line->filename()==filename)
					if (line->linenr <= linenr && line->linenr > maxlinenr)
					{	maxlinenr=line->linenr;
						result=obj->proc();
					}
			}

	return result;
}

ObjectP Table::insert(StringP str)
{
	ObjectPP obj=&base[hash(str)];
	*obj = new Object(str, *obj);
	check_oldyoung(*obj);
	return *obj;
}

PairP Table::objectlist()
{
	PairP li=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			push(li, obj);

	li = li->mergeSort();
	return li;
}

void Table::self_resize(int newsize)
{
	ObjectPP newbase = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(newsize+1));

	if (newbase==ObjectPP(0))
		c_out << "; Warning: Not enough memory to resize the table." << endL;
	else
	{	if (allocated) 
		{	PairP li=0;	li=objectlist();

			GC_::free((NodeP*)base);	base=newbase;	size=newsize;
			while (li)
			{	int i=hash(ObjectP(li->first())->name());
				base[i] = ObjectP(li->first());
				next(li);
			}
		}else
		{	allocated=true;	base=newbase;	size=newsize;
		}
	}
}

void Table::mark()
{
	_marked = run_numero();
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))

			obj->mark();

	print_gc("%2i ", examined);
}

void Table::mark_all_but_objects()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->mark_all_but_objects();

	print_gc("%2i ", examined);
}

void Table::clean()
{
	_marked = run_numero();
	mark_all_but_objects();

	print_gc("Table::clean: ", 0);
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
	{	ObjectP lastobj=0;

		for (ObjectP obj = *itemptr;  obj;  next(obj)) 
		{//	print_gc_object;

			if ( obj->marked() != GC_::run_numero()
			&&  obj->proc_or_prim()==Unbound
			&&  obj->var()==Unbound
			&&  obj->table()==Unbound
			&& !(obj->bit.permanent | obj->bit.permanentobj) )
			{
				inc_examined;
				if (!lastobj)
					next(*itemptr);
				else
					lastobj->settail( obj->tail() );
			}else		// has a value, don't gctwa
				lastobj=obj;
		}
		itemptr++;
	}
	print_gc("%2i collected\n", examined);
}

void Table::setflag(ObjectType want, ObjectType newflag)
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->setflag(want, newflag);
}

void Table::clearflag(ObjectType want, ObjectType newflag)
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->clearflag(want, newflag);
}

PairP Table::contents(ObjectType contentsType, ObjectType obj_type) 
{
	PairP contentsList=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			NodeP sub = obj->contents(contentsType, obj_type);
			if (sub)
				if (sub->is.list)
					if (contentsList)
						contentsList=PairP(contentsList->sentence(sub));
					else
						contentsList=PairP(sub);
				else
					push(contentsList, sub);
		}
	contentsList = contentsList->mergeSort();
	return contentsList;
}

PairP Table::contentsNameVar(ObjectType want) 
{
	PairP contentsList=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			StringP name = obj->name();
			if (name)
				if (obj->bit.proc && want & PROCS)
					push(contentsList, 
						new Pair(name, 0));
				else if (obj->bit.val && want & VALS)
					push(contentsList, 
						new Pair(name, 
						new Pair(obj->var(), 0)));
		}
	contentsList = contentsList->mergeSort();
	return contentsList;
}

PairP Table::contentsTree(ObjectType want, 
								  ObjectType dont,
								  ObjectType must, intP count)
{
	PairP contentsList=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{	PairP ct=obj->contentsTree(want, dont, must, count);
			if (ct)
				push(contentsList, ct);
		}

	if (!contentsList)
		return 0;

	contentsList = contentsList->mergeSort();
	return new Pair(type_Table, contentsList);
}

PairP Table::contents(ObjectType obj_type)
{
	PairP ret=0;
	push(ret, contents(TABLES, obj_type));
	push(ret, contents(  VALS, obj_type));
	push(ret, contents( PROCS, obj_type));
	return ret;
}

NodeP Table::erase_all(bool procs, bool vals, bool tables) 
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String("to erase definitions")); return Unbound;
	}
	int redef = _RedefP->var()->equal(_True);
	
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)

		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			if (procs 
			&&  obj->bit.proc
			&& !obj->bit.proc_buried 
			&& !obj->proc_or_prim()->is.primitive)
				obj->setproc( ProcedureP(Unbound) );

			if (vals  
			&&  obj->bit.val
			&& !obj->bit.val_buried)
				obj->setvar(Unbound);

			if (tables && !obj->bit.proc_buried)
				obj->settable(TableP(Unbound));
		}
	return Unbound;
}

//___________________________________________________________________

Struct::Struct(PairP data_): 
	Node(STRUCT), size(0), base(0), _first(0)
{
	base = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(data_->count()+1));
	if (!base)
		return;
	size=data_->count();
	ObjectP last=0;
	PairP np;
	for (np=data_;  np;  next(np))
	{	StringP f=np->first()->first()->string();
		if (String::caseIgnored)
			f=StringP(f->lower());
		ObjectP obj=find(f);
		if (obj)
		{	error(BAD_DATA, f);
			return;
		}
		obj=insert(f);

		NodeP t=np->first()->second()->intern(), n=t;
		if (n==_Int)
			n=new Int();
		else if (n==_Int16)
			n=new Int16();

		else if (n==_Int8)
			n=new Int8();
		else if (n==_UInt8)
			n=new UInt8();
		else if (n==_Float)
			n=new Float;
		else if (n==_Complex_)
			n=new Complexd();
		else if (n==_Word_)
			n=new String();
		else if (n==_IntArray_)
			n=new IntArray();
		else if (n==_Int16Array_)
			n=new Int16Array();
		else if (n==_FloatArray_)
			n=new FloatArray();
		else 
			error(BAD_DATA, t);
		
		if throwing return;
		
		if (np->first()->tail()->tail())
		{	NodeP third=np->first()->third(), n3=third;
		
			if (t==_Int)
				n3=n3->_int();

			else if (t==_Int16)
				n3=n3->toint16();
			else if (t==_Int8)
				n3=n3->toint8();
			else if (t==_UInt8)
				n3=n3->touint8();
			else if (t==_Float)
				n3=n3->_float();
			else if (t==_Word_)
			{	if (n3->numberP())
					n3=new String(int(third->getnumber()));
				else
					n3=n3->string();
			}
			else if (t==_IntArray_)
			{	if (n3->numberP())
					n3=new IntArray(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else if (t==_Int16Array_)
			{	if (n3->numberP())
					n3=new Int16Array(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else if (t==_FloatArray_)
			{	if (n3->numberP())
					n3=new FloatArray(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else
			{	error(BAD_DATA, third);
				return;
			}
			if (n->type!=n3->type && !(n->is.string && n3->is.string))
			{	error(BAD_DATA, third);
				return;
			}
			n=n3;
		}
		obj->setvar(n);
		if (last)
			last->settable(TableP(obj));
		else
			setfirst(obj);
		last=obj;
	}
	last->settable(TableP(0));
}

void Struct::free() 
{
	GC_::free((NodeP*)base); 
}

void Struct::inc_gc()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			fifo << obj;
}

int Struct::youngs() 
{
	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmark(obj);
	return youngs;
}

int Struct::youngsPtr(PtrToInt_ &ptrToInt) 
{
	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmarkPtr(obj, ptrToInt);
	return youngs;
}

void Struct::check_oldyoungs()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			check_oldyoung(obj);


}

void Struct::setPtrs(IntToPtr_ &intToPtr)
{
	setfirst(ObjectP((Object*)intToPtr.find(int(_first))));
	ObjectPP itemptr=base;


	for (int i=size;  i;  i--)
	{	*itemptr=(Object*)intToPtr.find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
wxDataInputStream &Struct::read(wxDataInputStream &istr) 
{	
	int siz;
	istr >> siz >> (int&)_first;

	base = (ObjectP*) GC_::calloc(sizeof(ObjectP), siz+1);
	if (!base)
		return istr;
	size=siz;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Struct::write(wxDataOutputStream &ostr) 
{
	ostr << type << size 
		<< GC_::ptrToInt->find(first());

	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;
}

NodeP Struct::copy(PtrToInt_ &ptrToInt)
{
	StructP str=new Struct();
	str->base = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(size+1));
	if (!str->base)
		return 0;
	str->size=size;
	str->_first=ObjectP(ptrToInt.find(first()));
	ObjectPP itemptr=base, itemptr2=str->base;
	for (int i=size;  i;  i--)
		*itemptr2++ = ObjectP(ptrToInt.find(*itemptr++));
	return str;
}

StringP Struct::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	ThreadP th=Thread::This();
	th->stringptr=textbuf;
	th->stringlen=maxlength;
	stringbuffer << NodeP(this);
	*th->stringptr='\0';
	return new String(textbuf);
}

void Struct::print(FILEP strm, int width, int depth)
{
	PairP lst=list();
	if (lst)
		lst->print(strm, width, depth);
}

void Struct::printBin(FILEP strm, StructP struct_)
{
	for (ObjectP obj=struct_->first();  obj;  obj=ObjectP(obj->proc()))
		if (obj->var())
			obj->var()->printBin(strm);
}

int Struct::hash(StringP str)
{
	unsigned int *i= (unsigned int*) str->ptr(),   index=0;

	for (char len=char(str->length());  len>0;  len-=sizeof(int))
	{	switch (len)
		{	case 0:	break;
			case 1:	index^= *i & 0x000000ff;	break;
			case 2:	index^= *i & 0x0000ffff;	break;
			case 3:	index^= *i & 0x00ffffff;	break;
			case 4:	index^= *i;	break;
			default:	index^= *i++ << 5;
		}
	};
	index %= size;
	return index;

}

ObjectP Struct::find(StringP str) 
{
	ObjectP obj;
	for (obj=base[hash(str)];  obj;  next(obj))
		if (str->equal(obj->name()))
			break;

	return obj;
}

ObjectP Struct::operator[](StringP str) {return find(str);}

ObjectP Struct::insert(StringP str)
{
	ObjectP &obj=base[hash(str)];
	obj = new Object(str, obj);
	check_oldyoung(obj);
	return obj;
}

void Struct::mark()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->mark();

	print_gc("%2i ", examined);
}

//___________________________________________________________________

bool GC_::initialized=false;
int volatile 
	GC_::is_running=false, 
	GC_::disabled=false;

int 
	GC_::mem_free=0,	// how many free nodes are available in the allocated segments
	GC_::mem_peak=0,	// peak memory (when most nodes were newed while least freed)
	GC_::mem_used=0,	// newed-freed nodes

	GC_::mark_gen=0,  GC_::max_gen=0,
	GC_::_run_numero=0,
	GC_::examined=0;

GC_::Interrupt GC_::interrupt = GC_::no_interrupt;

NodeP
	GC_::free_node=0,	// list of free nodes
	GC_::oldyoungs=0,	// list of nodes that point to younger nodes
	GC_::SpareNode=0;	// here goes newnode if no free_node is available

PairP GC_::reserve_tank=0;	// list of nodes not in any generation

FiFo GC_::fifo(GC_::fifo_size);

PtrToInt_ *GC_::ptrToInt;

PtrToString *GC_::primitivesPtrToString;
StringToPtr *GC_::primitivesStringToPtr;

void GC_::init()
{
	memset(generation, 0, sizeof(generation));
	Segment::add();
	reserve_tank_fill();
	oldyoungs = Unbound = new Node();
	SpareNode=new Node(SPARE_NODE);
	root=new Table(ROOT_SIZE);
}

void GC_::reserve_tank_fill() 
{
	PairP p=0;
	for (int i=50;  i;  i--)		// make pairs not in any generation
	{	if (!free_node) break;

		PairP newnd = PairP(free_node);   
		free_node = newnd->next_();	// remove newnd from the free_node list

		inc_mem_nodes();
		newnd->setnext(0);
		newnd->setoldyoung_next(0);

		newnd->settype_PAIR();
		newnd->_tail=p;	p=newnd;	// insert newnd into reserve tank list
		newnd->setfirst(0);
	}
	reserve_tank=p;
}

void GC_::reserve_tank_check() 
{
	if (!reserve_tank) 
		reserve_tank_fill();
}

void GC_::reserve_tank_use() 
{
	PairP nd=reserve_tank;	reserve_tank=0;
	while(nd) 
	{	nd->setnext(free_node);	free_node=nd;	// insert nd into the free_node list
		nd->settype_NTFREE();
		mem_free++;  mem_used--;
		next(nd);
	}
}


void GC_::mark()	// Iterative mark procedure
{
	if (!this) return;
	if (gen > mark_gen) return;   // too old
	if (marked()==run_numero()) return;  // already marked
	
	fifo << NodeP((Node*)this);

	while (fifo)
	{	NodeP nd;	fifo >> nd;
		if (nd)
			if (nd->gen <= mark_gen)
				if (nd->marked() != run_numero())
				{	nd->_marked = run_numero();		inc_examined;
					nd->inc_gc();
				}
	}
}

void GC_::markPtr(PtrToInt_ &ptrToInt)	// Iterative mark procedure
{
	if (!this) return;
	if (gen > mark_gen) return;   // too old
	if (marked()==run_numero()) return;  // already marked
	
	fifo << NodeP((Node*)this);

	while (fifo)
	{	NodeP nd;	fifo >> nd;
		if (nd)
			if (nd->marked() != run_numero())
			{	nd->_marked = run_numero();		inc_examined;
				ptrToInt.insertNode(nd);
				nd->inc_gc();
			}
	}
}

int GC_::gmark(NodeP child)
{
	if (child)
		if (child->gen < gen)
		{	child->mark();
			return 1;
		}
	return 0;
}

int GC_::gmarkPtr(NodeP child, PtrToInt_ &ptrToInt)
{
	if (child)
		if (child->gen < gen)
		{	child->markPtr(ptrToInt);
			return 1;
		}
	return 0;
}

void _cdecl GC_::mark_many(NodeP arg1, ...)
{
	va_list args;
	va_start(args, arg1);

	NodeP arg=arg1;
	while (arg!=_ListEnd)
	{	arg->mark();
		arg = va_arg(args, NodeP);
	}
	va_end(args);
	_ListEnd->mark();
	print_gc(";Globals %2i + ", examined);
}

void _cdecl GC_::mark_manyPtr(PtrToInt_ &ptrToInt, NodeP arg1, ...)
{
	va_list args;
	va_start(args, arg1);

	NodeP arg=arg1;
	while (arg!=_ListEnd)
	{	arg->markPtr(ptrToInt);
		arg = va_arg(args, NodeP);
	}
	va_end(args);
	_ListEnd->markPtr(ptrToInt);
	print_gc(";Globals %2i + ", examined);

}

void GC_::mark_C_stack()	// Check C Stack for Node pointers and mark valid ones
{
	NodeP  top_here=0,			// top is just now allocated on the Stack, 
			*top_stack = &top_here;  //top_stack points to it

	ThreadP th=Thread::main;

	NodeP 
		*top = 0,
		*bottom = th->bottomStack;

	if (th==Thread::This())
		top=top_stack;
	else
	{	CONTEXT context;
		context.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
		HANDLE h=Thread::mainHandle;
		if (!GetThreadContext(h, &context))
			return;

		Node *nd=0;
		nd=(Node*)context.Eax;
		if (is_valid_pointer(nd)) 
			nd->mark();
		nd=(Node*)context.Ebx;
		if (is_valid_pointer(nd)) 
			nd->mark();
		nd=(Node*)context.Ecx;
		if (is_valid_pointer(nd)) 
			nd->mark();
		nd=(Node*)context.Edx;
		if (is_valid_pointer(nd)) 
			nd->mark();
		nd=(Node*)context.Esi;
		if (is_valid_pointer(nd)) 
			nd->mark();
		nd=(Node*)context.Edi;
		if (is_valid_pointer(nd)) 
			nd->mark();

		top = (NodeP*)context.Esp;
	}
	if (bottom)
	{	if ((unsigned int)top < (unsigned int)bottom)	// check direction stack grows
		{	for (NodeP *sp=top;  (unsigned int)sp <= (unsigned int)bottom;  sp++)
				if (is_valid_pointer(*sp)) 
					(*sp)->mark();
		}else
		{	for (NodeP *sp=top;  (unsigned int)sp >= (unsigned int)bottom;  sp--)
				if (is_valid_pointer(*sp)) 
					(*sp)->mark();
		}
	}

	for (ThreadList::Node *node = Thread::list.GetFirst();  node;  node = node->GetNext())
   {
		th=node->GetData();

		NodeP 
			*top = 0,
			*bottom = th->bottomStack;

		if (th==Thread::This())
			top=top_stack;
		else
		{	CONTEXT context;
			context.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
			HANDLE h=th->GetInternal()->GetHandle();
			if (!GetThreadContext(h, &context))
				return;

			Node *nd=0;
			nd=(Node*)context.Eax;
			if (is_valid_pointer(nd)) 
				nd->mark();
			nd=(Node*)context.Ebx;
			if (is_valid_pointer(nd)) 
				nd->mark();
			nd=(Node*)context.Ecx;
			if (is_valid_pointer(nd)) 
				nd->mark();
			nd=(Node*)context.Edx;
			if (is_valid_pointer(nd)) 
				nd->mark();
			nd=(Node*)context.Esi;
			if (is_valid_pointer(nd)) 
				nd->mark();
			nd=(Node*)context.Edi;
			if (is_valid_pointer(nd)) 
				nd->mark();

			top = (NodeP*)context.Esp;
		}
		if (bottom)
		{	if ((unsigned int)top < (unsigned int)bottom)	// check direction stack grows
			{	for (NodeP *sp=top;  (unsigned int)sp <= (unsigned int)bottom;  sp++)
					if (is_valid_pointer(*sp)) 
						(*sp)->mark();
			}else
			{	for (NodeP *sp=top;  (unsigned int)sp >= (unsigned int)bottom;  sp--)
					if (is_valid_pointer(*sp)) 
						(*sp)->mark();
			}
		}
	}
	print_gc("C stack %2i + ", examined);
}

void GC_::mark_C_stackPtr(PtrToInt_ &ptrToInt)	
// Check C Stack for Node pointers and mark valid ones
{
	NodeP  top_here=0,			// top is just now allocated on the Stack, 
			*top_stack = &top_here;  //top_stack points to it

	ThreadP th=Thread::main;

	NodeP 
		*top = 0,
		*bottom = th->bottomStack;

	if (th==Thread::This())
		top=top_stack;
	else
	{	CONTEXT context;
		context.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
		HANDLE h=Thread::mainHandle;
		if (!GetThreadContext(h, &context))
			return;

		Node *nd=0;
		nd=(Node*)context.Eax;
		if (is_valid_pointer(nd)) 
			nd->markPtr(ptrToInt);
		nd=(Node*)context.Ebx;
		if (is_valid_pointer(nd)) 
			nd->markPtr(ptrToInt);
		nd=(Node*)context.Ecx;
		if (is_valid_pointer(nd)) 
			nd->markPtr(ptrToInt);
		nd=(Node*)context.Edx;
		if (is_valid_pointer(nd)) 
			nd->markPtr(ptrToInt);
		nd=(Node*)context.Esi;
		if (is_valid_pointer(nd)) 
			nd->markPtr(ptrToInt);
		nd=(Node*)context.Edi;
		if (is_valid_pointer(nd)) 
			nd->markPtr(ptrToInt);

		top = (NodeP*)context.Esp;
	}
	if (bottom)
	{	if ((unsigned int)top < (unsigned int)bottom)	// check direction stack grows
		{	for (NodeP *sp=top;  (unsigned int)sp <= (unsigned int)bottom;  sp++)
				if (is_valid_pointer(*sp)) 
					(*sp)->markPtr(ptrToInt);
		}else
		{	for (NodeP *sp=top;  (unsigned int)sp >= (unsigned int)bottom;  sp--)
				if (is_valid_pointer(*sp)) 
					(*sp)->markPtr(ptrToInt);
		}
	}

	for (ThreadList::Node *node = Thread::list.GetFirst();  node;  node = node->GetNext())
   {
		th=node->GetData();

		NodeP 
			*top = 0,
			*bottom = th->bottomStack;

		if (th==Thread::This())
			top=top_stack;
		else
		{	CONTEXT context;
			context.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
			HANDLE h=th->GetInternal()->GetHandle();
			if (!GetThreadContext(h, &context))
				return;

			Node *nd=0;
			nd=(Node*)context.Eax;
			if (is_valid_pointer(nd)) 
				nd->markPtr(ptrToInt);
			nd=(Node*)context.Ebx;
			if (is_valid_pointer(nd)) 
				nd->markPtr(ptrToInt);
			nd=(Node*)context.Ecx;
			if (is_valid_pointer(nd)) 
				nd->markPtr(ptrToInt);
			nd=(Node*)context.Edx;
			if (is_valid_pointer(nd)) 
				nd->markPtr(ptrToInt);
			nd=(Node*)context.Esi;
			if (is_valid_pointer(nd)) 
				nd->markPtr(ptrToInt);
			nd=(Node*)context.Edi;
			if (is_valid_pointer(nd)) 
				nd->markPtr(ptrToInt);

			top = (NodeP*)context.Esp;
		}
		if (bottom)
		{	if ((unsigned int)top < (unsigned int)bottom)	// check direction stack grows
			{	for (NodeP *sp=top;  (unsigned int)sp <= (unsigned int)bottom;  sp++)
					if (is_valid_pointer(*sp)) 
						(*sp)->markPtr(ptrToInt);
			}else
			{	for (NodeP *sp=top;  (unsigned int)sp >= (unsigned int)bottom;  sp--)
					if (is_valid_pointer(*sp)) 
						(*sp)->markPtr(ptrToInt);
			}
		}
	}
	print_gc("C stack %2i + ", examined);
}

NodePP GC_::mark_oldyoung(NodePP oldyoung)	// Mark/traverse pointers to younger generations only
{
	Node &nd = **oldyoung;

	if (nd.gen <= mark_gen)		// too young
		return &(nd.oldyoung_next);

	if (!nd.youngs())		// nd no longer points to a node of a younger generation
	{
		*oldyoung = nd.oldyoung_next;  nd.oldyoung_next=0;	// remove nd from the list of oldyoungs
		return oldyoung;
	}
	return &(nd.oldyoung_next);
}

NodePP GC_::mark_oldyoungPtr(NodePP oldyoung, PtrToInt_ &ptrToInt)
// Mark/traverse pointers to younger generations only
{
	Node &nd = **oldyoung;

	if (!nd.youngsPtr(ptrToInt))	// nd no longer points to a node of a younger generation
	{
		*oldyoung = nd.oldyoung_next;  nd.oldyoung_next=0;	// remove nd from the list of oldyoungs
		return oldyoung;
	}
	return &(nd.oldyoung_next);
}


void GC_::mark_oldyoungs()	// mark pointers from old generations to young
{
	NodePP oldyoung = &oldyoungs;  

	while (*oldyoung != Unbound)
		oldyoung = mark_oldyoung(oldyoung);

	print_gc("oldyoungs %2i marked ", examined);
}

void GC_::mark_oldyoungsPtr(PtrToInt_ &ptrToInt)	// mark pointers from old generations to young
{
	NodePP oldyoung = &oldyoungs;  

	while (*oldyoung != Unbound)
		oldyoung = mark_oldyoungPtr(oldyoung, ptrToInt);

	print_gc("oldyoungs %2i marked ", examined);
}

void _cdecl GC_::find_many(IntToPtr_ &intToPtr, intP manyPtrNrs, NodePP arg1, ...)
{
	va_list args;
	va_start(args, arg1);

	NodePP arg=arg1;
	int i=0;
	while (*arg!=_ListEnd)
	{	*arg = (Node*)intToPtr.find(manyPtrNrs[i]);
		 arg = va_arg(args, NodePP);
		i++;
	}
	va_end(args);
	_ListEnd = (Object*)intToPtr.find(manyPtrNrs[i]);
	print_gc(";Globals %2i + ", examined);
}

void GC_::mark_phase(bool clean_root)	// mark all accessable node pointers of generations<=mark_gen
{
mark_phase:
	_run_numero++;

	mark_many(Unbound, R_Texture::textures, 
		MANY_NODES(LIST_MANY) 0);

	ThreadP th=Thread::main;
	mark_many(MANY_TH_NODES(LIST_MANY) 0);
	th->     stack->mark();
	th->  varStack->mark();
	th->errorStack->mark();
	th->  argStack->mark();
	th->  funStack->mark();

	for (ThreadList::Node *node = Thread::list.GetFirst();  node;  node = node->GetNext())
   {
		ThreadP th=node->GetData();
		mark_many(MANY_TH_NODES(LIST_MANY) 0);
		th->     stack->mark();
		th->  varStack->mark();
		th->errorStack->mark();
		th->  argStack->mark();
		th->  funStack->mark();
	}
	print_gc("NodePStacks %2i + ", examined);

	mark_C_stack();

	mark_oldyoungs();		// mark pointers from old generations to young

	if (fifo.overflow()) return;

	if (root)
		if (clean_root) 
		{	root->clean();	clean_root=false;
			goto mark_phase;
		}
		else
			root->mark();
	debugStates.mark();
}

void GC_::mark_phasePtr(PtrToInt_ &ptrToInt)
// mark all accessable node pointers of generations<=mark_gen
{
	_run_numero++;
	mark_gen=max_gen;

	mark_manyPtr(ptrToInt, Unbound, MANY_NODES(LIST_MANY) 0);
	ThreadP th=Thread::main;
	mark_manyPtr(ptrToInt, MANY_TH_NODES(LIST_MANY) 0);
	th->     stack->markPtr(ptrToInt);
	th->  varStack->markPtr(ptrToInt);
	th->errorStack->markPtr(ptrToInt);
	th->  argStack->markPtr(ptrToInt);
	th->  funStack->markPtr(ptrToInt);

	for (ThreadList::Node *node = Thread::list.GetFirst();  node;  node = node->GetNext())
   {
		ThreadP th=node->GetData();
		mark_manyPtr(ptrToInt, MANY_TH_NODES(LIST_MANY) 0);
		th->     stack->markPtr(ptrToInt);
		th->  varStack->markPtr(ptrToInt);
		th->errorStack->markPtr(ptrToInt);
		th->  argStack->markPtr(ptrToInt);
		th->  funStack->markPtr(ptrToInt);
	}
	print_gc("NodePStacks %2i + ", examined);

	mark_C_stackPtr(ptrToInt);
	mark_oldyoungsPtr(ptrToInt);	// mark pointers from old generations to young
	if (fifo.overflow()) 
		return;

	if (root)
		root->markPtr(ptrToInt);
}

void GC_::check_oldyoung(NodeP new_child) 
{
	if (new_child)
		if (new_child->gen < gen  &&  !oldyoung_next)
		{
//			gcCS.Enter();
			oldyoung_next = oldyoungs;	
			oldyoungs=(Node*)this;	// insert this into the list of oldyoung pointers
//			gcCS.Leave();
		}
}


int GC_::sweep_phase(int gen)	// promote marked to next gen, free the rest
{
	freed_sofar_init;
	int num_freed=0;

	for (int i=gen;  i >=0;  i--)
	{
		NodePP lastnd = &generation[i];

		for (NodeP nd= *lastnd;  nd;  nd = *lastnd) 
		{
			if (nd->marked() != run_numero()) 			// free the nd
			{	num_freed++,  mem_used--;   mem_free++;

				*lastnd = nd->_next;					   // remove nd from its generation
				nd->_next=free_node;   free_node=nd;	// insert nd into the free_node list

				if (nd->oldyoung_next) 
				{	NodePP prev = &oldyoungs;
					while (*prev != nd)					// search nd in the list of oldyoungs (expensive!)
						prev = &((*prev)->oldyoung_next);

					*prev = nd->oldyoung_next;   nd->oldyoung_next=0;	// remove nd from the list of oldyoungs
				}
				nd->free();
				nd->settype_NTFREE();
				if (memoryWindow)
					memoryWindow->updateNode(nd);
			}else
			{	nd->age--;
				if (nd->age==0  &&  i < generations-1)		// promote to next gen
				{
					*lastnd = nd->_next;									// remove nd form its gen


					nd->_next= generation[i+1];    generation[i+1] = nd;	// insert nd into next gen

					nd->gen = i+1;
					nd->age = gc_age_threshold[i+1];

					nd->check_oldyoungs();

					if (max_gen==i) 
						max_gen++;
				}else								// keep in this gen
					lastnd = &(nd->_next);
			}
		}
		print_gc("%i+", num_freed - freed_sofar);
		freed_sofar_is(num_freed);
	}
	print_gc("=%2i freed\n", num_freed);
	return num_freed;
}

	
void GC_::gc(bool full)	// Garbage Collector
{
	if (!initialized)
		return;

	static int next_gen=0;

	if (disabled)
	{	if (!free_node)
			if (!Segment::add())
				error(OUT_OF_MEM_UNREC, 0);
		return;
	}
	if (fifo.overflow()) 
	{	if (!Segment::add() && !free_node)

			error(OUT_OF_MEM_UNREC, 0);
		return;
	}
	checkThrowing();

	int gen =		// deepest generation to garbage collect 
		mark_gen = full ?  max_gen
							 :  next_gen;

	bool clean_all =		// garbage collect root objects
		(gen==max_gen  &&  max_gen > 1) || full;

	mark_phase(clean_all);

#ifdef _DEBUG
	int num_freed =
#endif
		sweep_phase(gen);

	if (mem_free > min_free)
		next_gen=0;			// check only nursery gen next time
	else if (gen < max_gen)
		next_gen=gen+1;	// need to check more gens next time
	else	// all gens have been collected; allocate more memory
	{	if (!Segment::add()  &&  mem_free < 50  &&  !free_node)
			error(OUT_OF_MEM_UNREC, 0);
		next_gen=0;	
	}
	if (theApp && !theApp->heapReserve)
	{	theApp->heapReserve=new char[0x11000];
		if (theApp->heapReserve)	// well, at least *try* to fix it
		{	delete theApp->heapReserve;
			theApp->heapReserve=new char[0x10000];
		}
	}
}

void GC_::inc_mem_nodes()
{
	mem_used++;
	mem_free--;
	if (mem_used > mem_peak) 
		mem_peak=mem_used;
	ThreadP th=Thread::This();
	if (th)
	{	if (th->security->nodes < unsigned(mem_used) && !throwing)
			error(SECURITY_NODES);
	}
}

void GC_::run(bool full) 
{
//	if (is_running)
//		return;

	static wxCriticalSection runCS;

	runCS.Enter();

	is_running++;
	wxASSERT(is_running==1);

	ThreadP th=Thread::This();
	HANDLE h=Thread::mainHandle;
	int result=0;
	if (!th->IsMain())
	{	result=SuspendThread(h);
		wxASSERT(result!=-1);
	}
	Thread::listCS.Enter();
	ThreadList::Node *node;
	for (node = Thread::list.GetFirst();  node;  node = node->GetNext())
	{	ThreadP th2=node->GetData();
		if (th2 != th)
			if (th2->IsRunning())
			{	wxThreadError err=th2->Pause();
				wxASSERT(err==wxTHREAD_NO_ERROR);
			}
	}
	Thread::listCS.Leave();
	
	interrupt=no_interrupt;
	gc(full);
	is_running=0;

	if (interrupt) 
	{	if (interrupt==pause) 
			logo_pause(0);
		else 
			logo_stop(0);
	}

	if (!th->IsMain())
		result=ResumeThread(h);

	Thread::listCS.Enter();
	for (node = Thread::list.GetFirst();  node;  node = node->GetNext())
   {	ThreadP th2=node->GetData();
		if (th2 != th)
			if (th2->IsPaused())
				th2->Resume();
	}
	Thread::listCS.Leave();
	runCS.Leave();
}

void *GC_::malloc_GC
{
	ThreadP th=Thread::This();
	size=(size? ((size-1) & MemBlockMask)+MemBlockSize : MemBlockSize);
//if (size > 1000000 || size < MemBlockSize)
//{	c_out << "bogus!" << int(size) << endL;
//	return 0;
//}
	if (th)
	{	if (th->security->memsize < th->memsize+size && !throwing)
		{	error(SECURITY_MEMSIZE); return 0;
		}
	}
#ifdef _DEBUG
	void *mem=::_malloc_dbg(size, blockType, filename, linenumber);
#else
	void *mem=::malloc(size);
#endif
	if (!mem)
	{	run(true);				 // collect garbage 
		mem=::malloc(size);	// and try again
		if (!mem) 
			error(OUT_OF_MEM, 0);
	}
	if (mem && th)
		th->memsize+=size;
	return mem;
}

void *GC_::calloc_GC
{
	ThreadP th=Thread::This();
	int realsize=(num*size ? ((num*size-1) & MemBlockMask)+MemBlockSize : MemBlockSize);
	if (num)
		size=realsize/num;
//if (size > 1000000 || realsize < MemBlockSize)
//{	c_out << "bogus!" << int(size) << endL;
//	return 0;
//}
	if (th)
	{	if (th->security->memsize < th->memsize+num*size && !throwing)
		{	error(SECURITY_MEMSIZE); return 0;
		}
	}
	void *mem=::calloc(num, size);
	if (!mem)
	{	run(true);						 // collect garbage 
		mem=::calloc(num, size);	// and try again
		if (!mem) 
			error(OUT_OF_MEM, 0);
	}
	if (mem && th)
		th->memsize+=realsize;
	return mem;
}

void *GC_::realloc_GC
{
	ThreadP th=Thread::This();
	size=(size? ((size-1) & MemBlockMask)+MemBlockSize : MemBlockSize);
//if (size > 1000000 || size < MemBlockSize)
//{	c_out << "bogus!" << int(size) << endL;
//	return memblock;
//}
	int old=0;
	if (th)
	{	if (memblock)
		{	old=_msize(memblock);
			old= old? ((old-1) & MemBlockMask)+MemBlockSize : MemBlockSize;
			if (th->security->memsize < th->memsize+size-old && !throwing)
			{	error(SECURITY_MEMSIZE); return 0;
			}
		}else if (th->security->memsize < th->memsize+size && !throwing)
		{	error(SECURITY_MEMSIZE); return 0;
		}
	}
	void *mem=::realloc(memblock, size);
	if (!mem)
	{	run(true);								 // collect garbage 
		mem=::realloc(memblock, size);	// and try again
		if (!mem) 
			error(OUT_OF_MEM, 0);
	}
	if (mem && th)
		th->memsize+=size-old;
	return mem;
}

void GC_::free(void *mem)
{
	ThreadP th=Thread::This();
	if (th)
	{	int size=_msize(mem);
		th->memsize-= size? ((size-1) & MemBlockMask)+MemBlockSize : MemBlockSize;
	}
	::free(mem);	
}

NodeP GC_::newnode() 
{
	gcCS.Enter();
	
#ifndef GC_ON_NEWNODE
	while (!free_node && !throwing) 
#endif
		run(false);

	if (!free_node)
	{	if (SpareNode->oldyoung_next) 
		{	NodePP prev = &oldyoungs;
			while (*prev != SpareNode)	// search nd in the list of oldyoungs
				prev = &((*prev)->oldyoung_next);

			*prev = SpareNode->oldyoung_next;   
			SpareNode->oldyoung_next=0;	// remove nd from the list of oldyoungs
		}
		SpareNode->free();
		gcCS.Leave();
		return SpareNode;
	}
	NodeP newnd=free_node;	free_node=newnd->_next;	// remove newnd from the free_node list

	newnd->_next= generation[0];	generation[0]=newnd;	// insert newnd into the nursery generation

	newnd->oldyoung_next=0;
	newnd->_marked=0; 
	newnd->gen=0;
	newnd->age = gc_age_threshold[0];

	inc_mem_nodes();

	gcCS.Leave();
	return newnd;
}
//___________________________________________________________________

wxDataInputStream &operator>>(wxDataInputStream &istr, Node &i)
{	return i.read(istr);
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Node &i)
{	return i.write(ostr);
}
//___________________________________________________________________

#ifdef WIN32
	#pragma optimize("", off)
#endif
void GC_::set_C_Stack(int stacksize, int *stack, int bi)
{
	int top=0,					 // top is just now allocated on the Stack, 
		*top_stack = &top;  //top_stack points to it
	int b;

	NodePP bottomStack=Thread::This()->bottomStack;	// TODO: this is totally bogus!
	int *sp=(int*)int(bottomStack);
	if ((unsigned int)bottomStack < (unsigned int)top_stack)	// check direction stack grows
	{	if (top_stack-(int*)int(bottomStack) <= stacksize+64)
			set_C_Stack(stacksize, stack, bi);
		for (int i=0; i < stacksize; i++)
			*sp++=stack[i];
		delete stack;
	}
	else
	{	int real_stacksize=(int*)int(Thread::This()->bottomStack)-top_stack;
		if (real_stacksize <= stacksize+64)
			set_C_Stack(stacksize, stack, bi);
		b=int(bottomStack-bi);
		sp=(int*)int(bottomStack);
		for (int i=0; i < stacksize; i++)
			*sp--=stack[i];
		delete stack;
	}
#ifdef WIN32
	_asm mov esp,b
	_asm pop ebp
	_asm pop eax
	_asm mov esp,ebp
	_asm pop ebp
	_asm ret;
#else
	asm("mov %0,%%esp": :"g" (b));
	asm("pop %ebp");
	asm("pop %eax");
	asm("mov %ebp,%esp");
	asm("pop %ebp");
	asm("ret");
#endif
}
#ifdef WIN32
	#pragma optimize("", on)
#endif


wxDECLARE_SCOPED_PTR(wxZipEntry, wxZipEntryPtr);

void GC_::load(charP name)
{
	root->erase_all(true, true, true);
	run(true);

	wxFFileInputStream *input=new wxFFileInputStream(name);
	if (!input->Ok())
		return;
	
	wxZipInputStream *zip=new wxZipInputStream(*input);
	wxZipEntryPtr *entry=new wxZipEntryPtr;

	if (entry->reset(zip->GetNextEntry()), entry->get() != 0)
	{
		wxDataInputStream *istr=new wxDataInputStream(*zip);

		int i, tableSize, rootPtrNr, manyPtrNrs[HOW_MANY_NODES];
		*istr >> tableSize >> rootPtrNr;
		for (i=0; i < HOW_MANY_NODES; i++)
			*istr >> manyPtrNrs[i];

		ThreadP th=Thread::This();
		th->  argStack->read(*istr);
		th->     stack->read(*istr);
		th->  funStack->read(*istr);
		th->  intStack->read(*istr);
		th->floatStack->read(*istr);
		th-> toDoStack->read(*istr);
		th->  varStack->read(*istr);
		th->errorStack->read(*istr);
		th-> argPStack->read(*istr);
		th-> varPStack->read(*istr);
		th->argNrStack->read(*istr);
	
		IntToPtr_ *intToPtr= new IntToPtr_(tableSize);
		RECORDER.intToPtr=new IntToPtr_(Recorder::tableSize);
		primitivesStringToPtr=new StringToPtr(
			ROOT_SIZE, Primitive::primitives);

		int stacksize;
		*istr >> stacksize;
		int *intstack=new int[stacksize], 
			*argpstack=new int[stacksize], 
//			*varpstack=new int[stacksize], 
				 *stack=new int[stacksize],
		*consolestack=new int[stacksize],
		  *stackstack=new int[stacksize];
		for (i=0; i<stacksize; i++)
			*istr >> intstack[i];
		for (i=0; i<stacksize; i++)
			*istr >> argpstack[i];
//		for (i=0; i<stacksize; i++)
//			*istr >> varpstack[i];
		for (i=0; i<stacksize; i++)
			*istr >> consolestack[i];
		int bi;
		*istr >> bi;
		for (i=0; i<stacksize; i++)
			*istr >> stack[i];
		for (i=0; i<stacksize; i++)
			*istr >> stackstack[i];
		int v;
		*istr >> v;
		th->var = v;
		*istr
			>> th->dont_fix_ift
			>> th->ift_iff_flag
			>> th->repcount
			>> th->user_repcount	// counters for repeat & while
			>> th->ifori
			>> th->iforend
			>> th->iforstep	// ...and for for
			>> th->trace_level	// indentation level when tracing
			>> th->fori
			>> th->forend
			>> th->forstep	// counters for for
			>> (int&)th->tailcall
			>> (int&)th->val_status;
		bool insertMode;
		*istr	>> insertMode >> console->charMode;
		if (!insertMode)
			console->toggleInsertMode();
		*istr >> th->graph->doubleBuffer
			>> int(Int::genSymNr);

/*		bool avi;	// shall append to Video Stream, but crashes!
		*istr >> avi;
		if (avi)		// if there was a AVI Video Stream open it again
		{	wxString *filename=new wxString;
			int rate, frame;
			*istr >> *filename >> rate >> frame;
			if (graph->aviGen)
			{	graph->aviGen->ReleaseEngine();
				delete graph->aviGen;
				graph->aviGen=0;
			}
			graph->aviGen=new AVIGen(charP(filename->c_str()), 
											 graph, rate);
			int err=graph->aviGen->InitEngine();//true, frame);
			//delete filename;
		}

*/
		intToPtr->readNodes(*istr);

		root=(Table*)intToPtr->find(rootPtrNr);
		find_many(*intToPtr, manyPtrNrs, MANY_NODES(LIST_MANY_ADDRESS) 0);
		th->  argStack->setPtrs(*intToPtr);
		th->     stack->setPtrs(*intToPtr);
		th->  varStack->setPtrs(*intToPtr);
		th->errorStack->setPtrs(*intToPtr);
		th->  funStack->setPtrs(*intToPtr);

#define FIND_AND_SET_PERMANENT(x) \
		if (_##x) _##x=root->find(_##x->name());
#define FIND_AND_SET_PERMANENT_TYPE(x) \
		if (type_##x) type_##x=root->find(type_##x->name());

		Permanent_Objects(FIND_AND_SET_PERMANENT)
		Permanent_Objects2(FIND_AND_SET_PERMANENT)
		Node_Type_Strings(FIND_AND_SET_PERMANENT_TYPE, UNUSED)

		for (i=0; i<stacksize; i++)
		{	if (intstack[i])
				stack[i]=int(intToPtr->find(intstack[i]));
			if (argpstack[i])
				stack[i]=int(argpstack[i]+th->argStack->base);
//			if (varpstack[i])
//				stack[i]=int(varpstack[i]+Evaluator::varStack.base);
			if (consolestack[i])
				stack[i]=int(console);
			if (stackstack[i])
				stack[i]=int(Thread::This()->bottomStack-stackstack[i]);
		}
		delete intstack;
		delete argpstack;
//		delete varpstack;
		delete consolestack;
		delete stackstack;

		delete primitivesStringToPtr;
		delete RECORDER.intToPtr;
		delete intToPtr;
		delete istr;
		delete entry;
		delete zip;
		delete input;
		set_C_Stack(stacksize, stack, bi);
	}
}

void GC_::save(charP name)
{
	NodeP  top=0,				// top is just now allocated on the Stack, 
			*top_stack = &top;  //top_stack points to it

	run(true);
	ptrToInt=new PtrToInt_(mem_used);
	primitivesPtrToString=new PtrToString(
		ROOT_SIZE, Primitive::primitives);

	mark_phasePtr(*ptrToInt);

	wxFFileOutputStream output(name);
	wxZipOutputStream zip(output);
	zip.PutNextEntry(_T("nodes.dat"));
	wxDataOutputStream ostr(zip);

	Recorder::ptrToInt=new PtrToInt_(Recorder::tableSize);

#define FIND_AND_WRITE_PTR(x) << ptrToInt->find(x)

	ostr << mem_used 
		<< ptrToInt->find(root)
		MANY_NODES(FIND_AND_WRITE_PTR);

	ThreadP th=Thread::This();
	th->  argStack->write(ostr);
	th->     stack->write(ostr);
	th->  funStack->write(ostr);
	th->  intStack->write(ostr);
	th->floatStack->write(ostr);
	th-> toDoStack->write(ostr);
	th->  varStack->write(ostr);
	th->errorStack->write(ostr);
	th-> argPStack->write(ostr);
	th-> varPStack->write(ostr);
	th->argNrStack->write(ostr);

	int stacksize;
	NodePP bottomStack=Thread::This()->bottomStack;	// TODO: totally bogus!

	if ((unsigned int)bottomStack < (unsigned int)top_stack)	// check direction stack grows
	{	stacksize=int(top_stack)-int(bottomStack)+1;
		ostr << stacksize;
		int *intstack=new int[stacksize];
		int i=0;
		NodeP *sp;
		for (sp=bottomStack;  (unsigned int)sp <= (unsigned int)top_stack;  sp++, i++)
			if (is_valid_pointer(*sp)) 
				intstack[i]=ptrToInt->find(*sp);
			else
				intstack[i]=0;
		for (i=0; i<stacksize; i++)
			ostr << intstack[i];
		delete intstack;
		for (sp=bottomStack;  (unsigned int)sp <= (unsigned int)top_stack;  sp++)
			ostr << int(*sp);
	}else
	{	stacksize=bottomStack-top_stack+1;
		ostr << stacksize;
		int *intstack=new int[stacksize],
			*argpstack=new int[stacksize],
//			*varpstack=new int[stacksize],
			*consolestack=new int[stacksize],
			*stackstack=new int[stacksize];
		int i=0;
		NodeP *sp;
		for (sp=bottomStack;  (unsigned int)sp >= (unsigned int)top_stack;  sp--, i++)
		{	if (is_valid_pointer(*sp)) 
				intstack[i]=ptrToInt->find(*sp);
			else
				intstack[i]=0;

			unsigned ptr = unsigned(*sp);
			if ( ptr >= unsigned(th->argStack->base )
			&&   ptr <= unsigned(th->argStack->index)
			&&  (ptr -  unsigned(th->argStack->base )) % sizeof(PairPP) == 0 )
				argpstack[i]=NodePP((NodeP*)&*(*sp))-th->argStack->base;
			else
				argpstack[i]=0;

/*			if ( ptr >= unsigned(Evaluator::varStack->base )
			&&   ptr <= unsigned(Evaluator::varStack->index)
			&&  (ptr -  unsigned(Evaluator::varStack->base )) % sizeof(PairPP) == 0 )
				varpstack[i]=VarP((Var*)&*(*sp))-Evaluator::varStack->base;
			else
				varpstack[i]=0;
*/
			if (int(*sp)==int(console))
				consolestack[i]=1;
			else
				consolestack[i]=0;
		}

		for (i=0; i<stacksize; i++)
			ostr << intstack[i];
		for (i=0; i<stacksize; i++)
			ostr << argpstack[i];
/*		for (i=0; i<stacksize; i++)
			ostr << varpstack[i];
*/		for (i=0; i<stacksize; i++)
			ostr << consolestack[i];
		delete argpstack;
//		delete varpstack;
		delete consolestack;
		i=0;
		for (sp=bottomStack;  (unsigned int)sp >= (unsigned int)top_stack;  sp--, i++)
			intstack[i]=int(*sp);

		int b;
#ifdef WIN32
		_asm mov b,ebp
#else
		asm("mov %%ebp,%0":"=g" (b));
#endif
		int bi=intP(int(bottomStack))-intP(b);
		ostr << bi;
		int *p=intP(intstack[bi]);
		for (i=0; i<stacksize; i++)
			ostr << intstack[i];

		for (i=0; i<stacksize; i++)
		{	p=intP(intstack[i]);
			unsigned n=intP(int(bottomStack))-p;
			if (n<unsigned(stacksize) || n>unsigned(-stacksize))
				stackstack[i]=n;
			else
				stackstack[i]=0;
		}
		for (i=0; i<stacksize; i++)
			ostr << stackstack[i];

		delete stackstack;
		delete intstack;
	}
	int v=th->var;
	ostr << v 
		<< th->dont_fix_ift
		<< th->ift_iff_flag
		<< th->repcount
		<< th->user_repcount	// counters for repeat & while
		<< th->ifori
		<< th->iforend
		<< th->iforstep	// ...and for for
		<< th->trace_level	// indentation level when tracing
		<< th->fori
		<< th->forend
		<< th->forstep	// counters for for
		<< th->tailcall
		<< th->val_status
		<< console->insertMode
		<< console->charMode
		<< th->graph->doubleBuffer
		<< Int::genSymNr;

/*	if (graph->aviGen)
	{	ostr.Write8(true); 
		ostr
			<< graph->aviGen->filename 
			<< graph->aviGen->rate 
			<< graph->aviGen->frame;
		graph->aviGen->ReleaseEngine();
		delete graph->aviGen;
		graph->aviGen=0;
	}
	else
		ostr << false;
*/	
	ptrToInt->writeNodes(ostr);

	delete RECORDER.ptrToInt;
	delete primitivesPtrToString;
	delete ptrToInt;
}
//___________________________________________________________________

NodeP lGC(PairPP arg) 
{
	GC_::run(exist(1));
//char *c;
//for (int i=0; i<35; i++)
//{	c=(charP)malloc(i); 
//		c_out << int(_msize(c)) << (i? ((i-1) & MemBlockMask)+MemBlockSize : MemBlockSize) << endL; 
//		free(c);
//}
	return Unbound;
}

NodeP lNodes(PairPP) 
{
	GC_::run(true);	// get real in-use figures

	int peak=GC_::mem_peak,
		used=GC_::mem_used;
	GC_::mem_peak=GC_::mem_used;

	ThreadP th=Thread::This();
	return new Pair
	(	new Int(used),
		new Pair
		(	new Int(peak), 
			new Pair( new Int(th->memsize), 0)
		)
	);
}

NodeP lloadNodes(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GC_::load(name->ptr());
	return Unbound;
}

NodeP lsaveNodes(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GC_::save(name->ptr());
	return Unbound;
}

NodeP lshrinkStacks(PairPP)
{
	ThreadP th=Thread::This();
	th->intStack->shrink();
	th->varPStack->shrink();
	th->argPStack->shrink();
	th->floatStack->shrink();
	th->stack->shrink();
	th->argStack->shrink();
	th->varStack->shrink();
	th->toDoStack->shrink();
	th->errorStack->shrink();
	return Unbound;
}

NodeP lsetStackNoisy(PairPP arg)
{
	bool noisy=arg[1]->a_bool(1)!=0;
	if throwing return Unbound;
	IntStack::noisy=noisy;
	return Unbound;
}
