#include "stdafx.h"
#include "Main.h"
#include "Node.h"
#include "File.h"
#include "Pair.h"
#include "Int.h"
#include "Float.h"
#include "Array.h"
#include "FloatArray.h"
#include "Complex.h"
#include "ComplexArray.h"
#include "Point.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

double sqr(double x) {return x*x;}

double NMRCarterNu(double I, double m, 
						  double nu0, double nuQ, double nuref, 
						  double eta, double K1, double K2,
						  double theta, double phi)
{
	theta*=Float::deg2rad;
	phi*=Float::deg2rad;
	double mu=cos(theta), cos2phi=cos(2*phi);
	return nu0 + (0.5*nuQ*(m-0.5)+K1*nuref/2)*(3*sqr(mu)-1) 
		- 0.5*(K2*nuref+(m-0.5)*nuQ*eta)*sqr(sin(theta))*cos2phi
		+ (sqr(nuQ)/(32*nu0))*(1-sqr(mu))
			*((102*m*(m-1)-18*I*(I+1)+39)*sqr(mu)*(1+2.0/3*eta*cos2phi)
				-(6*m*(m-1)-2*I*(I+1)+3)*(1-2.0/3*eta*cos2phi))
		+ sqr(eta)/72*sqr(nuQ)/nu0*(24*m*(m-1)-4*I*(I+1)+9-(30*m*(m-1)-2*I*(I+1)+12)*sqr(mu)
			-(51.0/2*m*(m-1)-9.0/4*I*(I+1)+39.0/4)*sqr(cos2phi)*sqr(sqr(mu)-1));
}

NodeP lNMRCarterNu(PairPP arg)
{
	enum Params
	{	_I=1, _m, 
		_nu0, _nuQ, _nuref, 
		_eta, _K1, _K2,
		_theta, _phi,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		m  =arg[_m  ]->a_float(_m  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		theta=arg[_theta]->a_float(_theta),
		phi=arg[_phi]->a_float(_phi);
	if throwing return Unbound;

	return new Float(
		NMRCarterNu(I, m, nu0, nuQ, nuref, 
						 eta, K1, K2, theta, phi));
}

NodeP lNMRCarterIntensity_oneTheta(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_theta, _nphi, _Inu0, _nnu,
		_intensity, _expo,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		theta=arg[_theta]->a_float(_theta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		expo=arg[_expo]->a_float(_expo);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dphi=360/nphi;

	double ifactors[5][5]=
	{	{1,0,0,0,0},
		{4,3,0,0,0},
		{9,8,5,0,0},
		{16,15,12,7},
		{25,24,21,16,9},
	};

	for (double m=-I+1; m<=I; m++)
	{	double ifactor=ifactors[int(I-0.5)][int(Float::abs(m-0.5))];
		for (double phi=0; phi<360; phi+=dphi)
		{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
										eta, K1, K2, theta, phi);
			int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
			if (inu>=0 && inu<ilen)
				intensity[inu]+=ifactor;
		}
		checkThrowing();
		if throwing return Unbound;
	}
	return lintensity;
}

NodeP lNMRCarterIntensity(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_ntheta, _nphi, _Inu0, _nnu,
		_intensity, _expo,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		ntheta=arg[_ntheta]->a_int(_ntheta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		expo=arg[_expo]->a_float(_expo);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dtheta=180/ntheta,
		dphi=360/nphi;

	double ifactors[5][5]=
	{	{1,0,0,0,0},
		{4,3,0,0,0},
		{9,8,5,0,0},
		{16,15,12,7},
		{25,24,21,16,9},
	};
	for (double m=-I+1; m<=I; m++)
	{	double ifactor=ifactors[int(I-0.5)][abs(int(m-0.5))];
		for (double theta=0; theta<90; theta+=dtheta)
		{
//			double factor=sin(Float::deg2rad*theta)*expo
//				+pow(Float::abs(cos(Float::deg2rad*theta)), 1000*(1-expo))*20*(1-expo);
			double 
				abssin=Float::abs(sin(Float::deg2rad*theta)),
				factor=pow(abssin, expo);
//			y=(sin x)*f+(abs cos x)^(1000*(1-f))*10*(1-f)
			double amount=ifactor*abssin;
			if (factor!=0)
				amount/=factor;
			for (double phi=0; phi<360; phi+=dphi)
			{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
											eta, K1, K2, theta, phi);
				int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
				if (inu>=0 && inu<ilen)
					intensity[inu]+=amount;
			}
			checkThrowing();
			if throwing return Unbound;
		}
	}
	return lintensity;
}

NodeP lNMRCarterIntensity_MAS(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_ntheta, _nphi, _Inu0, _nnu,
		_intensity, _expo,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		ntheta=arg[_ntheta]->a_int(_ntheta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		expo=arg[_expo]->a_float(_expo);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dtheta=180/ntheta,
		dphi=360/nphi,
		thetaMAS=54.7,
		r=Float::deg2rad;

	double ifactors[5][5]=
	{	{1,0,0,0,0},
		{4,3,0,0,0},
		{9,8,5,0,0},
		{16,15,12,7},
		{25,24,21,16,9},
	};
	for (double m=-I+1; m<=I; m++)
	{	double ifactor=ifactors[int(I-0.5)][abs(int(m-0.5))];
		for (double theta=Float::rnd()*dtheta; theta<180; theta+=dtheta)
		{	double factor=sin(r*theta)*expo
				+pow(Float::abs(cos(r*theta)), 1000*(1-expo))*20*(1-expo);
//			y=(sin x)*f+(abs cos x)^(1000*(1-f))*10*(1-f)
			DPoint 
				a(	cos(r*thetaMAS), 0, sin(r*thetaMAS));

			if (factor!=0)
				for (double phi=Float::rnd()*dphi/factor; phi<360; phi+=dphi/factor)
				{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
												eta, K1, K2, theta, phi);
					int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
					if (inu>=0 && inu<ilen)
					{	DPoint 
							b(	cos(r*phi)*cos(r*theta), 
								sin(r*phi)*cos(r*theta), 
								sin(r*theta)),
							c(a*b);
						double d=c.x+c.y+c.z;
						intensity[int(inu)]+=Float::abs(d);
					}
				}
			checkThrowing();
			if throwing return Unbound;
		}
	}
	return lintensity;
}

NodeP lNMRCarterIntensity_m(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_ntheta, _nphi, _Inu0, _nnu,
		_intensity,
		_m,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		ntheta=arg[_ntheta]->a_int(_ntheta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		m=arg[_m]->a_float(_m);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dtheta=180/ntheta,
		dphi=360/nphi;

	for (double theta=Float::rnd()*dtheta; theta<180; theta+=dtheta)
	{	for (double phi=Float::rnd()*dphi; phi<360; phi+=dphi)
		{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
										eta, K1, K2, theta, phi);
			int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
			if (inu>=0 && inu<ilen)
				intensity[inu]++;
		}
		checkThrowing();
		if throwing return Unbound;
	}
	return lintensity;
}

NodeP lGaussianSmooth(PairPP arg)
{
	enum Params
	{	_width=1, _smoothWidth, _compWidth, _data
	};
	double 
		width=arg[_width]->a_float(_width),
		smoothWidth=arg[_smoothWidth]->a_float(_smoothWidth),
		compWidth=arg[_compWidth]->a_float(_compWidth);
	FloatArrayP ldata=arg[_data]->a_floatarray(_data);
	if throwing return Unbound;
	doubleP s=ldata->data();
	int len=ldata->length();
	if (smoothWidth==0)
	{	error(BAD_DATA, arg[2], 2); return Unbound;
	}
	FloatArrayP result=new FloatArray(len);
	if throwing return Unbound;
	doubleP r=result->data();
	double factor=1.0/len*width/smoothWidth/sqrt(2.0);
	int clen=int(len/width*compWidth);
	if (clen>len)
		clen=len;
	doubleP e=(doubleP)GC_::malloc(2*clen*sizeof(double));
	for (int j=-clen; j<clen; j++)
		e[j+clen]=exp(-sqr(j*factor));
/*	for (int i=0; i<len; i++)
	{	double y=s[i];
		for (int j=-clen; j<clen; j++)
		{	int k=i+j;
			if (k >= 0 && k < len)
			{	r[k]+=y*e[j+clen];
			}
		}
	}
*/	for (int i=0; i<len; i++)
	{	double y=0;
		for (int j=-clen; j<clen; j++)
		{	int k=i-j;
			if (k >= 0 && k < len)
			{	y+=s[k]*e[j+clen];
			}
		}
		r[i]=y;
	}
	GC_::free(e); e=0;
	return result;
}

NodeP lNMRBlochSim(PairPP arg)
{
	enum Params 
	{	_Mx=1, _My, _P,
		_signalOn, _frequency, _MSamples, _B1c, _dBmax,
		_tau, _tauab, _taubc,
		_spins,
		_T1, _T2, _T2e,
		_ww, _expo
	};
	FloatArrayP nMx=arg[_Mx]->a_floatarray(_Mx);
	FloatArrayP nMy=arg[_My]->a_floatarray(_My);
	FloatArrayP nP =arg[_P ]->a_floatarray(_P );
	bool signalOn=(arg[_signalOn]->a_bool(_signalOn)!=0);
	double frequency=arg[_frequency]->a_float(_frequency);
	double MSamples=arg[_MSamples]->a_float(_MSamples);
	double B1c=arg[_B1c]->a_float(_B1c);
	double dBmax=arg[_dBmax]->a_float(_dBmax);
	double tau=arg[_tau]->a_float(_tau);
	double tauab=arg[_tauab]->a_float(_tauab);
	double taubc=arg[_taubc]->a_float(_taubc);
	int spins=arg[_spins]->a_int(_spins);
	double T1=arg[_T1]->a_float(_T1);
	double T2=arg[_T2]->a_float(_T2);
	double T2e=arg[_T2e]->a_float(_T2e);
	double ww=arg[_ww]->a_float(_ww);
	double expo=arg[_expo]->a_float(_expo);
	if throwing return Unbound;
	int IMAX=nMx->count();
	double *Mx=nMx->data();
	double *My=nMy->data();
	double *P =nP ->data();
	double 
		IM=IMAX/MSamples,
		f=frequency*IM,
		inull=Float::rnd()*100, 
		tau_=tau/IM, tauab_=tauab/IM, taubc_=taubc/IM,
		t0=0.35-tauab_,
		t1=t0+2*tau_, 
		t2=t1+tauab_, 
		t3=t2+tau_,
		t4=t3+taubc_,
		t5=t4+2*tau_,
		t6=t5+tauab_;
	const int n=1000;
	B1c=B1c/frequency;
	double 
		PI2=2*Float::pi,
		gamma=f*PI2,
		B[3], B0=1, B1=0, 
		t=t0, dt=1.0/IMAX,
		M0=1, M[n][3], *Mj=0, Mnorm=0,
		dM_dt[n][3], d2M_dt2[3], *dM_dtj,
		Bj[3], dB[n][3];
	int i0=int(t*IMAX);
	
//	srand(0);
	for (int j=0; j<n; j++)
	{	M[j][0]=0;
		M[j][1]=M0;
		M[j][2]=0;
		dM_dt[j][0]=0;
		dM_dt[j][1]=0;
		dM_dt[j][2]=0;
		dB[j][0]=0;
		dB[j][1]=pow(abs((2.0*j)/spins-1), expo)*dBmax;
//		dB[j][1]=sqr(sqr(Float::rnd()))*dBmax;
//		if (Float::rnd() > 0.5)
//			dB[j][1]=-dB[j][1];
		if (j < spins/2)
			dB[j][1]=-dB[j][1];
		dB[j][2]=0;
	}
	double Mges[3]={0,0,0};
	for (int j=0; j<spins; j++)
	{	Mges[0]+=M[j][0];
		Mges[1]+=M[j][1];
		Mges[2]+=M[j][2];
	}
	Mges[0]=Mges[0]/spins;
	Mges[1]=Mges[1]/spins;
	Mges[2]=Mges[2]/spins;
	int spins3rt=int(pow(spins, 0.333)),
		spins3rt2=spins3rt*spins3rt;

	for (int i=i0; i<IMAX; i++)
	{	if (signalOn)
			B1=B1c*((t<t1)|((t>t2)&(t<t3))|((t>t4)&(t<t5)));
		else
			B1=B1c*(       ((t>t2)&(t<t3))|((t>t4)&(t<t5)));
		B[0]=B1*sin(PI2*f*t);
		B[1]=B0;
		B[2]=0; //B1*cos -omega*t
		for (int j=0; j<spins; j++)
		{	Mj=M[j];
//			Bj[0]=B[0]+dB[j][0];
//			Bj[1]=B[1]+dB[j][1];
//			Bj[2]=B[2]+dB[j][2];
			int k,l,p;
			k=j-1;
			if (k<0)
				k=spins-1;
			l=j-spins3rt;
			if (l<0)
				l=spins-spins3rt;
			p=j-spins3rt2;
			if (p<0)
				p=spins-spins3rt;
			Bj[0]=B[0]+dB[j][0]+(M[k][0]+M[l][0]+M[p][0])*ww;
			Bj[1]=B[1]+dB[j][1]+(M[k][1]+M[l][1]+M[p][1])*ww;
			Bj[2]=B[2]+dB[j][2]+(M[k][2]+M[l][2]+M[p][2])*ww;
			dM_dtj=dM_dt[j];
			dM_dtj[0]=(gamma*(Mj[1]*Bj[2]-Mj[2]*Bj[1])-    Mj[0] /T2e);
			dM_dtj[1]=(gamma*(Mj[2]*Bj[0]-Mj[0]*Bj[2])+(M0-Mj[1])/T1);
			dM_dtj[2]=(gamma*(Mj[0]*Bj[1]-Mj[1]*Bj[0])-    Mj[2] /T2);
			d2M_dt2[0]=dM_dtj[0]-dM_dt[j][0];
			d2M_dt2[1]=dM_dtj[1]-dM_dt[j][1];
			d2M_dt2[2]=dM_dtj[2]-dM_dt[j][2];
			M[j][0]=Mj[0]+dM_dtj[0]*dt+d2M_dt2[0]*dt/2;
			M[j][1]=Mj[1]+dM_dtj[1]*dt+d2M_dt2[1]*dt/2;
			M[j][2]=Mj[2]+dM_dtj[2]*dt+d2M_dt2[2]*dt/2;
			Mnorm=sqrt(sqr(M[j][0])+sqr(M[j][1])+sqr(M[j][2]));
			if (Mnorm > M0) 
			{	M[j][0]=M[j][0]/Mnorm;
				M[j][1]=M[j][1]/Mnorm;
				M[j][2]=M[j][2]/Mnorm;
			}
		}
		t+=dt;

		for (int j=0; j<spins; j++)
		{	Mges[0]+=M[j][0];
			Mges[1]+=M[j][1];
			Mges[2]+=M[j][2];
		}
		Mges[0]=Mges[0]/spins;
		Mges[1]=Mges[1]/spins;
		Mges[2]=Mges[2]/spins;
		if (i>=0)
		{	Mx[i]=Mges[0];// M[0] in coil direction, M[1] parallel B0
			My[i]=Mges[1];
			P [i]=B1/B1c;
//			Ur[i]=char(10*sin(PI2*f*t));
		}
	}
	return Unbound;
}

NodeP lNMRBlochSim2(PairPP arg)
{
	enum Params 
	{	_Mx=1, _My, _P,
		_signalOn, _frequency, _MSamples, _B1c, _dBmax,
		_tau, _tau2, _tau3, _tauab, _taubc,
		_spins,
		_T1, _T2, _T2e,
		_ww, _expo
	};
	FloatArrayP nMx=arg[_Mx]->a_floatarray(_Mx);
	FloatArrayP nMy=arg[_My]->a_floatarray(_My);
	FloatArrayP nP =arg[_P ]->a_floatarray(_P );
	bool signalOn=(arg[_signalOn]->a_bool(_signalOn)!=0);
	double frequency=arg[_frequency]->a_float(_frequency);
	double MSamples=arg[_MSamples]->a_float(_MSamples);
	double B1c=arg[_B1c]->a_float(_B1c);
	double dBmax=arg[_dBmax]->a_float(_dBmax);
	double tau=arg[_tau]->a_float(_tau);
	double tau2=arg[_tau2]->a_float(_tau2);
	double tau3=arg[_tau3]->a_float(_tau3);
	double tauab=arg[_tauab]->a_float(_tauab);
	double taubc=arg[_taubc]->a_float(_taubc);
	int spins=arg[_spins]->a_int(_spins);
	double T1=arg[_T1]->a_float(_T1);
	double T2=arg[_T2]->a_float(_T2);
	double T2e=arg[_T2e]->a_float(_T2e);
	double ww=arg[_ww]->a_float(_ww);
	double expo=arg[_expo]->a_float(_expo);
	if throwing return Unbound;
	int IMAX=nMx->count();
	double *Mx=nMx->data();
	double *My=nMy->data();
	double *P =nP ->data();
	double 
		IM=IMAX/MSamples,
		f=frequency*IM,
		inull=Float::rnd()*100, 
		tau_=tau/IM, tau2_=tau2/IM, tau3_=tau3/IM, tauab_=tauab/IM, taubc_=taubc/IM,
		t0=0.35-tauab_,
		t1=t0+tau_, 
		t2=t1+tauab_, 
		t3=t2+tau2_,
		t4=t3+taubc_,
		t5=t4+tau3_,
		t6=t5+tauab_;
	const int n=1000;
	B1c=B1c/frequency;
	double 
		PI2=2*Float::pi,
		gamma=f*PI2,
		B[3], B0=1, B1=0, 
		t=t0, dt=1.0/IMAX,
		M0=1, M[n][3], *Mj=0, Mnorm=0,
		dM_dt[n][3], d2M_dt2[3], *dM_dtj,
		Bj[3], dB[n][3];
	int i0=int(t*IMAX);
	
//	srand(0);
	for (int j=0; j<n; j++)
	{	M[j][0]=0;
		M[j][1]=M0;
		M[j][2]=0;
		dM_dt[j][0]=0;
		dM_dt[j][1]=0;
		dM_dt[j][2]=0;
		dB[j][0]=0;
		dB[j][1]=pow(abs((2.0*j)/spins-1), expo)*dBmax;
//		dB[j][1]=sqr(sqr(Float::rnd()))*dBmax;
//		if (Float::rnd() > 0.5)
//			dB[j][1]=-dB[j][1];
		if (j < spins/2)
			dB[j][1]=-dB[j][1];
		dB[j][2]=0;
	}
	double Mges[3]={0,0,0};
	for (int j=0; j<spins; j++)
	{	Mges[0]+=M[j][0];
		Mges[1]+=M[j][1];
		Mges[2]+=M[j][2];
	}
	Mges[0]=Mges[0]/spins;
	Mges[1]=Mges[1]/spins;
	Mges[2]=Mges[2]/spins;
	int spins3rt=int(pow(spins, 0.333)),
		spins3rt2=spins3rt*spins3rt;

	for (int i=i0; i<IMAX; i++)
	{	if (signalOn)
			B1=B1c*((t<t1)|((t>t2)&(t<t3))|((t>t4)&(t<t5)));
		else
			B1=B1c*(       ((t>t2)&(t<t3))|((t>t4)&(t<t5)));
		B[0]=B1*sin(PI2*f*t);
		B[1]=B0;
		B[2]=0; //B1*cos -omega*t
		for (int j=0; j<spins; j++)
		{	Mj=M[j];
//			Bj[0]=B[0]+dB[j][0];
//			Bj[1]=B[1]+dB[j][1];
//			Bj[2]=B[2]+dB[j][2];
			int k,l,p;
			k=j-1;
			if (k<0)
				k=spins-1;
			l=j-spins3rt;
			if (l<0)
				l=spins-spins3rt;
			p=j-spins3rt2;
			if (p<0)
				p=spins-spins3rt;
			Bj[0]=B[0]+dB[j][0]+(M[k][0]+M[l][0]+M[p][0])*ww;
			Bj[1]=B[1]+dB[j][1]+(M[k][1]+M[l][1]+M[p][1])*ww;
			Bj[2]=B[2]+dB[j][2]+(M[k][2]+M[l][2]+M[p][2])*ww;
			dM_dtj=dM_dt[j];
			dM_dtj[0]=(gamma*(Mj[1]*Bj[2]-Mj[2]*Bj[1])-    Mj[0] /T2e);
			dM_dtj[1]=(gamma*(Mj[2]*Bj[0]-Mj[0]*Bj[2])+(M0-Mj[1])/T1);
			dM_dtj[2]=(gamma*(Mj[0]*Bj[1]-Mj[1]*Bj[0])-    Mj[2] /T2);
			d2M_dt2[0]=dM_dtj[0]-dM_dt[j][0];
			d2M_dt2[1]=dM_dtj[1]-dM_dt[j][1];
			d2M_dt2[2]=dM_dtj[2]-dM_dt[j][2];
			M[j][0]=Mj[0]+dM_dtj[0]*dt+d2M_dt2[0]*dt/2;
			M[j][1]=Mj[1]+dM_dtj[1]*dt+d2M_dt2[1]*dt/2;
			M[j][2]=Mj[2]+dM_dtj[2]*dt+d2M_dt2[2]*dt/2;
			Mnorm=sqrt(sqr(M[j][0])+sqr(M[j][1])+sqr(M[j][2]));
			if (Mnorm > M0) 
			{	M[j][0]=M[j][0]/Mnorm;
				M[j][1]=M[j][1]/Mnorm;
				M[j][2]=M[j][2]/Mnorm;
			}
		}
		t+=dt;

		for (int j=0; j<spins; j++)
		{	Mges[0]+=M[j][0];
			Mges[1]+=M[j][1];
			Mges[2]+=M[j][2];
		}
		Mges[0]=Mges[0]/spins;
		Mges[1]=Mges[1]/spins;
		Mges[2]=Mges[2]/spins;
		if (i>=0)
		{	Mx[i]=Mges[0];// M[0] in coil direction, M[1] parallel B0
			My[i]=Mges[1];
			P [i]=B1/B1c;
//			Ur[i]=char(10*sin(PI2*f*t));
		}
	}
	return Unbound;
}

NodeP lNMRBlochSimSpectrum(PairPP arg)
{
	enum Params 
	{	_Mx=1, _My, _P,
		_signalOn, _frequency, _MSamples, _B1c, _dBmax,
		_tau, _tauab, 
		_spins,
		_T1, _T2, _T2e,
		_ww, _expo
	};
	FloatArrayP nMx=arg[_Mx]->a_floatarray(_Mx);
	FloatArrayP nMy=arg[_My]->a_floatarray(_My);
	FloatArrayP nP =arg[_P ]->a_floatarray(_P );
	bool signalOn=(arg[_signalOn]->a_bool(_signalOn)!=0);
	double frequency=arg[_frequency]->a_float(_frequency);
	double MSamples=arg[_MSamples]->a_float(_MSamples);
	double B1c=arg[_B1c]->a_float(_B1c);
	double dBmax=arg[_dBmax]->a_float(_dBmax);
	double tau=arg[_tau]->a_float(_tau);
	double tauab=arg[_tauab]->a_float(_tauab);
	int spins=arg[_spins]->a_int(_spins);
	double T1=arg[_T1]->a_float(_T1);
	double T2=arg[_T2]->a_float(_T2);
	double T2e=arg[_T2e]->a_float(_T2e);
	double ww=arg[_ww]->a_float(_ww);
	double expo=arg[_expo]->a_float(_expo);
	if throwing return Unbound;
	int IMAX=nMx->count();
	double *Mx=nMx->data();
	double *My=nMy->data();
	double *P =nP ->data();
	double 
		IM=IMAX/MSamples,
		f=frequency*IM,
		inull=Float::rnd()*100, 
		tau_=tau/IM, tauab_=tauab/IM,
		t0=0.35-tauab_,
		t1=t0+tau_,
		t2=t1+tauab_,
		t3=t2+2*tau_,
		t4=t3+tauab_;
	const int n=1000;
	B1c=B1c/frequency;
	double 
		PI2=2*Float::pi,
		gamma=f*PI2,
//		T2=0.1, T2e=T2, T1=10,
		B[3], B0=1, B1=0, 
		t=t0, dt=1.0/IMAX,
		M0=1, M[n][3], *Mj=0, Mnorm=0,
		dM_dt[n][3], d2M_dt2[3], *dM_dtj,
		Bj[3], dB[n][3];
	int i0=int(t*IMAX);
	
//	srand(0);
	for (int j=0; j<spins; j++)
	{	M[j][0]=0;
		M[j][1]=M0;
		M[j][2]=0;
		dM_dt[j][0]=0;
		dM_dt[j][1]=0;
		dM_dt[j][2]=0;
		dB[j][0]=0;
		dB[j][1]=pow(abs((2.0*j)/spins-1), expo)*dBmax;
//		dB[j][1]=sqr(sqr(Float::rnd()))*dBmax;
//		if (Float::rnd() > 0.5)
//			dB[j][1]=-dB[j][1];
		if (j < spins/2)
			dB[j][1]=-dB[j][1];
		dB[j][2]=0;
	}
	double Mges[3]={0,0,0};
	for (int j=0; j<spins; j++)
	{	Mges[0]+=M[j][0];
		Mges[1]+=M[j][1];
		Mges[2]+=M[j][2];
	}
	Mges[0]=Mges[0]/spins;
	Mges[1]=Mges[1]/spins;
	Mges[2]=Mges[2]/spins;
	int spins3rt=int(pow(spins, 0.333)),
		spins3rt2=spins3rt*spins3rt;
	for (int i=i0; i<IMAX; i++)
	{	if (signalOn)
			B1=B1c*(((t>t0)&(t<t1))|((t>t2)&(t<t3)));
		else
			B1=B1c*(                ((t>t2)&(t<t3)));
		B[0]=B1*sin(PI2*f*t);
		B[1]=B0;
		B[2]=0; //B1*cos -omega*t
		for (int j=0; j<spins; j++)
		{	Mj=M[j];
//			Bj[0]=B[0]+dB[j][0];
//			Bj[1]=B[1]+dB[j][1];
//			Bj[2]=B[2]+dB[j][2];
			int k,l,p;
			k=j-1;
			if (k<0)
				k=spins-1;
			l=j-spins3rt;
			if (l<0)
				l=spins-spins3rt;
			p=j-spins3rt2;
			if (p<0)
				p=spins-spins3rt;
			Bj[0]=B[0]+dB[j][0]+(M[k][0]+M[l][0]+M[p][0])*ww;
			Bj[1]=B[1]+dB[j][1]+(M[k][1]+M[l][1]+M[p][1])*ww;
			Bj[2]=B[2]+dB[j][2]+(M[k][2]+M[l][2]+M[p][2])*ww;
			dM_dtj=dM_dt[j];
			dM_dtj[0]=(gamma*(Mj[1]*Bj[2]-Mj[2]*Bj[1])-    Mj[0] /T2e);
			dM_dtj[1]=(gamma*(Mj[2]*Bj[0]-Mj[0]*Bj[2])+(M0-Mj[1])/T1);
			dM_dtj[2]=(gamma*(Mj[0]*Bj[1]-Mj[1]*Bj[0])-    Mj[2] /T2);
			d2M_dt2[0]=dM_dtj[0]-dM_dt[j][0];
			d2M_dt2[1]=dM_dtj[1]-dM_dt[j][1];
			d2M_dt2[2]=dM_dtj[2]-dM_dt[j][2];
			M[j][0]=Mj[0]+dM_dtj[0]*dt+d2M_dt2[0]*dt/2;
			M[j][1]=Mj[1]+dM_dtj[1]*dt+d2M_dt2[1]*dt/2;
			M[j][2]=Mj[2]+dM_dtj[2]*dt+d2M_dt2[2]*dt/2;
			Mnorm=sqrt(sqr(M[j][0])+sqr(M[j][1])+sqr(M[j][2]));
			if (Mnorm > M0) 
			{	M[j][0]=M[j][0]/Mnorm;
				M[j][1]=M[j][1]/Mnorm;
				M[j][2]=M[j][2]/Mnorm;
			}
		}
		t+=dt;

		Mges[0]=0;
		Mges[1]=0;
		Mges[2]=0;
		for (int j=0; j<spins; j++)
		{	Mges[0]+=M[j][0];
			Mges[1]+=M[j][1];
			Mges[2]+=M[j][2];
		}
		Mges[0]=Mges[0]/spins;
		Mges[1]=Mges[1]/spins;
		Mges[2]=Mges[2]/spins;
		if (i>=0)
		{	Mx[i]=Mges[0];// M[0] in coil direction, M[1] parallel B0
			My[i]=Mges[1];
			P [i]=B1/B1c;
//			Ur[i]=char(10*sin(PI2*f*t));
		}
	}
	return Unbound;
}

NodeP lNMR_QMsim(PairPP arg)
{
	enum Params
	{	_Hz=1, _Hq, _chi0, _B,
		_S, _Mx, _My, _Mz,
	};
	ArrayP Hz=arg[_Hz]->a_array(_Hz);
	if (Hz->count()!=3)
	{	error(BAD_DATA, arg[_Hz], _Hz); return Unbound;
	}
	ArrayP Hq=arg[_Hq]->a_array(_Hq);
	ArrayP 
		Hzx=ArrayP(Hz->data()[0]), 
		Hzy=ArrayP(Hz->data()[1]), 
		Hzz=ArrayP(Hz->data()[2]);
	int chii=std::min(Hzx->count(), Hq->count());
	ComplexArrayP chi0=arg[_chi0]->a_complexarray(_chi0);
	ArrayP B=arg[_B]->a_array(_B);
	doubleP 
		Bx=FloatArrayP(B->data()[0])->data(), 
		By=FloatArrayP(B->data()[1])->data(), 
		Bz=FloatArrayP(B->data()[2])->data();
	int imax=
		std::min(B->data()[0]->count(), 
		std::min(B->data()[1]->count(), 
					B->data()[2]->count()));
	FloatArrayP
		Mx=arg[_Mx]->a_floatarray(_Mx),
		My=arg[_My]->a_floatarray(_My),
		Mz=arg[_Mz]->a_floatarray(_Mz);
	doubleP
		mx=Mx->data(),
		my=My->data(),
		mz=Mz->data();
	ArrayP S=arg[_S]->a_array(_S);
	if (S->count()!=3)
	{	error(BAD_DATA, arg[_S], _S); return Unbound;
	}
	ArrayP 
		Sx=ArrayP(S->data()[0]), 
		Sy=ArrayP(S->data()[1]), 
		Sz=ArrayP(S->data()[2]);

	double h=6.62620e-34,
		h_bar=h/(2*Float::pi);
	ArrayP chi_nd=new Array(chii);
	ComplexArrayP *chi=(ComplexArrayP*)(chi_nd->data());
	for (int i=0; i<chii; i++)
	{	chi[i]=new ComplexArray(imax);
	}
	for (int j=0; j<chii; j++)
		chi[j]->data()[0]=chi0->data()[j];

	complexd xx=0, yy=0, zz=0;
	for (int i=1; i<imax; i++)
	{	
/*		for (int j=0; j<chii; j++)
		{	complexdP 
				Hzxj=ComplexArrayP(Hzx->data()[j])->data(),
				Hzyj=ComplexArrayP(Hzy->data()[j])->data(),
				Hzzj=ComplexArrayP(Hzz->data()[j])->data(),
				Hqj=ComplexArrayP(Hq->data()[j])->data(),
				chij=chi[j]->data();
			for (int k=0; k<chii; k++)
			{	complexd xyz=0;
				xyz+=Hzxj[k]*Bx[i];
				xyz+=Hzyj[k]*By[i];
				xyz+=Hzzj[k]*Bz[i];
				xyz+=Hqj[k]+1.0*(k==j);
				chij[i]+= xyz*chi[k]->data()[i-1];
			}
		}
/*/
		xx=0; yy=0; zz=0;
		for (int j=0; j<chii; j++)
		{	complexd x=0, y=0, z=0, q=0;
			for (int k=0; k<chii; k++)
			{	complexd chiki1=chi[k]->data()[i-1];
				x+=ComplexArrayP(Hzx->data()[k])->data()[j]*chiki1;
				y+=ComplexArrayP(Hzy->data()[k])->data()[j]*chiki1;
				z+=ComplexArrayP(Hzz->data()[k])->data()[j]*chiki1;
				q+=ComplexArrayP(Hq ->data()[k])->data()[j]*chiki1;
			}
			x*=Bx[i];
			y*=By[i];
			z*=Bz[i];
			chi[j]->data()[i]+= x+y+z+q+ chi[j]->data()[i-1];
		}
//*/
		complexd tmp=0;
		for (int j=0; j<chii; j++)
		{	complexd c=chi[j]->data()[i];
			tmp+=	c*complexd(c.real(),-c.imag());
		}
		tmp=sqrt(tmp);
		for (int j=0; j<chii; j++)
			chi[j]->data()[i]/=tmp;

		xx=0;	yy=0;	zz=0;
		for (int j=0; j<chii; j++)
		{	complexd x=0, y=0, z=0;
			for (int k=0; k<chii; k++)
			{	complexd chiki=chi[k]->data()[i];
				x+=ComplexArrayP(Sx->data()[k])->data()[j]*chiki;
				y+=ComplexArrayP(Sy->data()[k])->data()[j]*chiki;
				z+=ComplexArrayP(Sz->data()[k])->data()[j]*chiki;
			}
			complexd chiji=chi[j]->data()[i];
			complexd chijiconj=complexd(chiji.real(), -chiji.imag());
			xx+=x*chijiconj;
			yy+=y*chijiconj;
			zz+=z*chijiconj;
		}
		mx[i]=xx.real()/h;
		my[i]=yy.real()/h;
		mz[i]=zz.real()/h;
	}
	return chi_nd;
}

