#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/html/helpctrl.h>

#include <stdio.h>
#include <ctype.h>

#include "Main.h"

#include "File.h"
#include "String.h"
#include "Pair.h"
#include "Line.h"
#include "Procedure.h"
#include "Primitive.h"
#include "Object.h"
#include "Table.h"
#include "LogoObject.h"
#include "List.h"
#include "Quote.h"
#include "Colon.h"
#include "Runparsed.h"
#include "Tree.h"
#include "Int.h"
#include "UInt8.h"
#include "Table.h"

#include "Environment.h"
#include "Color.h"
#include "Error.h"
#include "Stack.h"
#include "Init.h"
#include "LogoApp.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#define SET_th_LINE(l)									\
	if (l && l->is.line)									\
	{	th->col=LineP(l)->col;							\
		th->tcol=LineP(l)->tcol;						\
		th->row=LineP(l)->row;							\
		th->setfilename(LineP(l)->filename());		\
	}

//static wxMutex parenMutex(wxMUTEX_RECURSIVE);

bool iswhitespace(char ch) {return ch==' ' || ch=='\t'|| ch=='\n'|| ch=='\0';}
bool isparen(char ch) {return ch=='(' || ch==')' || ch==';';}
//bool isunary(char ch) {return ch=='!';}
bool isdot  (char ch) {return ch=='.';}
bool issinglequote(char ch) {return ch=='\'';}
bool isassignmentop(char ch)
{	return
		ch=='=' || ch=='!' || ch=='<' || ch=='>'
	|| ch=='+' || ch=='-' || ch=='*' || ch=='/';
}
bool isinfix(char ch)
{	return       /* ch=='!' ||*/ ch=='^' 
	|| ch=='*' || ch=='/' || /*ch=='%' ||*/ ch=='+' || ch=='-'
	|| ch=='=' || ch=='<' || ch=='>';
}

bool Tree::dont_know_how_to;

LineP String::runparse_node(PairPP ndsptr, 
									 bool &was_assignment,
									 bool &was_dot) 
{
	ThreadP th=Thread::This();
	NodeType wtype = type;
	   charP wptr  = ptr();
	     int wlen  = length(),	wcount=0;
	   HeadP whead = head();

	bool monadic_minus=false;

	LineP outline=0, tnode=0, lastnode=0;	// to construct the output
	
	while (wcount < wlen) 
	{	if (*wptr=='\\')
		{	wptr++, wcount++;
			if (wcount < wlen)
			{	wptr++, wcount++;
			}
		}else if (*wptr==';')		// semicolon comment
		{	*ndsptr=0;
			break;
		}
		else if (*wptr=='"')		// quoted
		{	int tcount=0;
			charP tptr = ++wptr;
			wcount++;
			bool vbar=false;
			while (wcount < wlen  &&  (!isparen(*wptr) || vbar)) 
			{	//if (wtype==BACKSLASHED  &&  backslash.is(*wptr))
				//	wtype=UNBOUND;    // flag for "\( case
				if (*wptr=='\\')
				{	wptr++, wcount++, tcount++;
				}else if (*wptr=='|')
					vbar=!vbar;
				if (wcount < wlen)
				{	wptr++, wcount++, tcount++;
				}
			}
//			if (wtype==UNBOUND) 
//			{	wtype=BACKSLASHED;
//				tnode = new Line
//				(	new Quote( 
//						make(tptr, 0, tcount, wtype, copy_decode)->intern() ),
//					0
//				);
//			}else
	//		th->caseIgnored=false;
			StringP s=make(tptr, whead, tcount, wtype, copy);
			if throwing return LineP(Unbound);
			s=s->noPepper();
			if throwing return LineP(Unbound);
			tnode = new Line
			(	new Quote( s->intern() ),
				0
			);
	//		th->caseIgnored=true;
		}
		else if (*wptr==':' && wptr[1]!=':')	// colon (asks for contents of a variable)
		{	int tcount=0;
			charP tptr = ++wptr;
			wcount++;
			while (wcount < wlen  &&  !isparen(*wptr)  &&  !isinfix(*wptr)  
			&& (th->logoLanguage==_ucblogo 
				|| (!isdot(*wptr) && !issinglequote(*wptr))))
				wptr++, wcount++, tcount++;
			ObjectP obj=make(tptr, whead, tcount, wtype, copy)->intern();
			if (th->logoLanguage==_ucblogo && obj->bit.proc)
			{	error(IS_ALREADY_A_PROC, obj);
				return 0;
			}
			tnode = new Line(new Colon(obj), 0);
		}
		else if (th->logoLanguage==_aucblogo && *wptr=='&')	// ampersand (used as var parameter)
		{	int tcount=0;
			charP tptr = ++wptr;
			wcount++;
			while (wcount < wlen  &&  !isparen(*wptr)  &&  !isinfix(*wptr)  
			&& !isdot(*wptr) && !issinglequote(*wptr))
				wptr++, wcount++, tcount++;
			ObjectP obj=make(tptr, whead, tcount, wtype, copy)->intern();
//			if (obj->bit.proc)
//				c_out << endL << "; Warning: the variable " << obj 
//					<< " shadows a proc or primitive!" << endL;
			tnode = new Line(new Colon(obj, VARPARAM), 0);
		}
		else if (*wptr=='-'	// minus sign with space before and no space after is unary
		&&  (wcount==0 || wptr[-1]=='=' || wptr[-1]=='(' || isinfix(wptr[-1]))
		&&  monadic_minus==false 
		&&  wcount+1 < wlen  &&  !iswhitespace(wptr[1])  &&  wptr[1] != '=') 
		{
//			tnode = new Line( new Int(0), 0);
			monadic_minus=true;
			continue;
		}
		else if (isparen(*wptr) || isinfix(*wptr) || *wptr=='!' 
		|| (th->logoLanguage==_aucblogo && isdot(*wptr)))	// paren or infix
		{	if (monadic_minus)
			{	tnode = new Line(	_Negate, 0 );
			}
			else if (isassignmentop(*wptr)  &&  wcount+1 < wlen  &&  wptr[1]=='=')	
			//  +=  -=  *=  /=  %=  ==  !=
			{
				if (th->logoLanguage==_aucblogo 
				&& *wptr != '=' &&  *wptr != '!' 
				&&  *wptr != '<' &&  *wptr != '>')	
					// not the comparsion ops, only assignments
					if (tnode)
					{	if (!tnode->tail() 
						&&  tnode->first() 
						&&  tnode->first()->type != COLON
						&& !(tnode->first()->type==OBJECT 
							&& ObjectP(tnode->first())->bit.proc))
							lastnode->setfirst( new Quote(lastnode->first()));
							// automatically quote the left arg
					}else
						was_assignment=true;	// 'x=' is the first thing in this expression; 
													// runparse has to notice that somehow
				tnode = new Line
					(	(new String(wptr, whead, 2, STRING, copy))->intern(),
						0
					);
				wptr++, wcount++;
			}
/*			else if (isunary(*wptr))	// Kludge for factorial operator '!'
			{	tnode = new Line
					(	(new String(wptr, whead, 1, STRING, copy))->intern(),
						new Int(0)
					);
			}
*/			else		// normal infix-ops + - * / = ect.
			{	if (th->logoLanguage==_aucblogo  &&  *wptr=='=')	// assignment sign
				{	bool setItem=false, 
						setItem2D=false,
						setItem3D=false;
					if (outline)
						if (outline->tail()) 
							if (ObjectP(outline->second())==_Dot
							&& outline->tail()->tail())
							{	PairP list=outline->tail()->tail();
								while (list)
								{	if (ObjectP(list->first())==_Dot
									&& list->tail())
									{	setItem2D=true; 
										next(list);
										break;
									}
									next(list);
								}
								if (setItem2D)
								{	while (list)
									{	if (ObjectP(list->first())==_Dot
										&& list->tail())
										{	setItem2D=false; 
											setItem3D=true; 
											break;
										}
										next(list);
									}
								}
								if (!setItem2D && !setItem3D)
									setItem=true;
							}
					if (setItem)
					// Kludge for name.item=value
					{	PairP 
							name=outline->first(),
							item=outline->tail()->tail();
						
						outline=new Line
							(	__SetItem, 
								item
							);
						lastnode=LineP(item);
						while (lastnode->tail() && !throwing)
							lastnode=LineP(lastnode->tail());
						tnode=new Quote(name);
					}
					else if (setItem2D)
					// even bigger KLUDGE for name.i.j=value
					{	PairP 
							name=outline->first(),
							item1=outline->tail()->tail(),
							item2=0,
							list=item1, lastlist=list;
						while (list && ObjectP(list->first())!=_Dot)
						{	lastlist=list;
							next(list);							
						}
						item2=list;
						if (item2)
							next(item2);

						lastlist->settail(item2);
						outline=
							new Line(_mdSetItem, 
							new Line(_List,
							item1));
						lastnode=LineP(item2);
						while (lastnode->tail() && !throwing)
							lastnode=LineP(lastnode->tail());
						tnode=new Quote(name);
					}
					else if (setItem3D)
					// even bigger KLUDGE for name.i.j.k=value
					{	PairP 
							name=outline->first(),
							item1=outline->tail()->tail(),
							item2=0,
							item3=0,
							list=item1, lastlist=list;
//c_out << "L)" << list << endL;
						while (list && ObjectP(list->first())!=_Dot)
						{	lastlist=list;
							next(list);							
						}
						item2=list;
						if (item2)
							next(item2);
						lastlist->settail(item2);

						list=item2;
						while (list && ObjectP(list->first())!=_Dot)
						{	lastlist=list;
							next(list);							
						}
						item3=list;
						if (item3)
							next(item3);
						lastlist->settail(item3);

						outline=
							new Line(_mdSetItem, 
							new Line(_Left_Paren,
							new Line(_List, item1)));
						lastnode=LineP(item3);
						while (lastnode->tail() && !throwing)
							lastnode=LineP(lastnode->tail());
						lastnode->settail(new Line(_Right_Paren, 0));
						lastnode=LineP(lastnode->tail());
//c_out << "1)"<< item1 << endL;
//c_out << "2)"<< item2 << endL;
//c_out << "3)"<< item3 << endL;
//c_out << "O)"<< outline << endL;
//c_out << lastnode << endL;
						tnode=new Quote(name);
					}
					else
					{	if (tnode)
						{	if (!tnode->tail() 
							&&   tnode->first() 
							&&   tnode->first()->type != COLON)
//							&& !(tnode->first()->type == OBJECT
//								&& ObjectP(tnode->first())->proc_or_prim() != Unbound))
								lastnode->setfirst( new Quote(lastnode->first()));	
								// automatically quote the left arg
						}else
							was_assignment=true;	// '=' is the first thing in this expression; 
														// runparse has to notice that somehow
						tnode = new Line
						(	(new String(wptr, whead, 1, STRING, copy))->intern(), 
							0
						);
					}
				}else if (th->logoLanguage==_aucblogo  &&  *wptr=='.')
				{	tnode=new Line(_Dot, 0);
					was_dot=true;
				}else 
					tnode = new Line
					(	(new String(wptr, whead, 1, STRING, copy))->intern(), 
						0
					);
			}
			monadic_minus=false;
			wptr++, wcount++;
		}
		else if (th->logoLanguage==_aucblogo  &&  issinglequote(*wptr))	// singlequote
		{	int tcount=0;	// code copied from "quoted" part
			charP tptr = ++wptr;
			wcount++;
			while (wcount < wlen  && !isparen(*wptr) && !isinfix(*wptr)) 
			{//	if (wtype==BACKSLASHED  &&  backslash.is(*wptr))
			//		wtype=UNBOUND;    // flag for "\( case
				wptr++, wcount++, tcount++;
			}
			//if (wtype==UNBOUND) 
			//{	wtype=BACKSLASHED;
			//	tnode = new Line
			//		(	_Dot,
			//			new Line
			//			(	new Quote
			//				(	make(tptr, 0, tcount, wtype, 
			//						copy_decode)->intern() 
			//				),
			//				0
			//			)
			//		);
			//}else
				tnode = new Line
					(	_Dot,
						new Line
						(	new Quote
							(	make(tptr, whead, tcount, wtype, copy)->intern() 
							),
							0
						)
					);
		}
		else				// number or ?5
		{	int tcount=0, dotcount=0;
			charP tptr = wptr, dotptr=0;
			bool gotdot=false, goti=false;

			enum NumType
			{	digits_so_far, 
				just_saw_e,		// so minus can be next
				no_longer_eligible,	// even if an 'e' comes along
				question_mark,
				nothing_yet
			};
			NumType numtype = nothing_yet;

			if (*wptr=='?')	// turn ?5 to (? 5) 
			{	numtype = question_mark;
				wptr++, wcount++, tcount++;
			}
			while (wcount < wlen  
			&&  !isparen(*wptr) 
			&& (!(isinfix(*wptr) || issinglequote(*wptr))
				|| (numtype==just_saw_e  &&  (*wptr=='-' || *wptr=='+')))) 
			{
				if (numtype==just_saw_e  &&  (*wptr=='-' || *wptr=='+'))
				{	wptr++, wcount++, tcount++;
					continue;
				}
				if (numtype==nothing_yet  &&  is_digit(*wptr)) 
					numtype=digits_so_far;
				
				if (numtype==digits_so_far  &&  tcount > 0  
				&& (*wptr=='e' || *wptr=='E'))
					numtype = just_saw_e;
				else 
					if (!(	is_digit(*wptr) || (!gotdot && isdot(*wptr)) 
							|| (!goti && (*wptr=='i' || *wptr=='I'))) 
						/*|| numtype==just_saw_e*/)
						numtype = no_longer_eligible;

				if (isdot(*wptr)) 
					if (numtype==no_longer_eligible || numtype==question_mark)
					{	if (th->logoLanguage==_aucblogo)
							break;
					}else
					{	gotdot=true;
						dotcount=tcount;
						dotptr=wptr;
					}
				if (*wptr=='i' || *wptr=='I') 
					if (numtype==digits_so_far || numtype==just_saw_e)
					{	wptr++, wcount++, tcount++;
						goti=true;
						break;
					}
					else
						numtype = no_longer_eligible;//goti=true;
				wptr++, wcount++, tcount++;
			}
			if (numtype==question_mark  &&  tcount > 1)	// ?5 syntax
			{	LineP qmtnode = 
					new Line
					(	_Left_Paren, 
						new Line
						(	_Query,
							new Line
							(	make(tptr+1, whead, tcount-1, wtype, copy)->number(),
								0
							)
						)
					);
				if (!outline) 
					outline=qmtnode;
				else 
					lastnode->settail( qmtnode );
				lastnode = LineP(qmtnode->tail()->tail());

				tnode = new Line( _Right_Paren, 0 );
			}
			else if (numtype==digits_so_far  ||  numtype==just_saw_e  &&  tcount > 0) 
			{	tnode = new Line
					(	make(tptr, whead, tcount, wtype, copy)->number(),
						0
					);
			}else if (gotdot)
			{	tnode = 
				new Line(make(tptr, whead, dotcount, wtype, copy)->number(),
				new Line(_Dot,
				new Line(
				make(dotptr+1, whead, wptr-(dotptr+1), wtype, copy)->intern(),
					0)));
			}else
			{	StringP str=make(tptr, whead, tcount, wtype, copy);
				if throwing return LineP(Unbound);
				ObjectP obj=str->intern();
				tnode = new Line(obj, 0);
			}
		}
		if throwing return LineP(Unbound);
		if (!outline) 
			outline=tnode;
		else
			lastnode->settail( tnode );
		if (tnode)
			while (tnode->tail())
				tnode=LineP(tnode->tail());
		lastnode=tnode;
	}
	return outline;
}

LineP Pair::runparse(bool deep) 
{
	if (type_of(this)==RUNPARSED)
		return ((Runparsed*)this)->parsed();
	
	if (isnolist()) 
	{	error(BAD_DATA_UNREC, this);
		return 0;
	}

//	wxMutexError err=parenMutex.Lock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);

	LineP outline=0, tnode=0, lastnode=0, olastnode=0;
	PairP ndlist=this;
	bool was_dot=false, owas_dot=false;

	ThreadP th=Thread::This();
	while (ndlist) 
	{	NodeP nd = ndlist->first();

		if (ndlist->is.line)
		{	if (LineP(ndlist)->filename()->type==STRING)
			{	th->col=LineP(ndlist)->col;
				th->tcol=LineP(ndlist)->tcol;
				th->row=LineP(ndlist)->row;
				th->setfilename(LineP(ndlist)->filename());
//c_out << "RP" << th->filename() << th->nr << nd << endL;
			}
		}
		next(ndlist);
		
		if (!nd)
			tnode= new Line(0,0);
		else if (!nd->is.word)
		{	if (deep)
				tnode=PairP(nd)->runparse(deep);
			else
			{	tnode = new Line(nd, 0);
			}
		}else 
		{	if (nd==_Negate) 
			{	tnode=new Line(nd, 0);
			}
			else if (!nd->numberP())
			{	bool was_assignment=false;
				tnode = nd->string()->
					runparse_node(&ndlist, was_assignment, was_dot);

//				if ( owas_dot && olastnode)
//					if (!olastnode->tail() && olastnode->first())
//c_out << olastnode;

				if ( was_assignment && lastnode)
					if (!lastnode->tail() && lastnode->first())
						if (lastnode->first()->type != COLON
						&& lastnode->first()->type != QUOTE)
							lastnode->setfirst( new Quote(lastnode->first()));
			}else
				tnode = new Line(nd->number(), 0);
		}
		if (tnode) 
		{	if (!outline) 
				outline=tnode;
			else
				lastnode->settail( tnode );
			
			owas_dot=was_dot;
			was_dot=false;
			olastnode=lastnode;
			lastnode=tnode;
			while (lastnode->tail()) 
			{	lastnode=LineP(lastnode->tail());
				if (checkThrowing()) break;
			}
		}
		if (checkThrowing()) break;
	}
//	c_out << "<RP>" << outline << "</rp>" << endL; lpause(0);
//	err=parenMutex.Unlock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);

	return outline;
}

NodeP lrunParse(PairPP arg) 
{
	NodeP s=arg[1];
	if (type_of(s)==ARRAY  &&  !throwing)	// arrays not yet allowed in runparse! array-to-list, then runparse!
	{	error(BAD_DATA, s);
		return Unbound;
	}
	PairP val=PairP(s);
	if (!throwing && s && !s->is.aggregate)
		val = StringP(s)->parser(true);

	if (!throwing) return val->runparse();
	return Unbound;
}

// Make a list of the next n expressions, where n is between min and max.
// Set args to immediately after the last expression.
 
LineP Procedure::gather_some_args(int min, int max, 
											 LinePP args, bool inparen, 
											 int oldPriority, ObjectPP ifnode) 
{
	ThreadP th=Thread::This();
	if (!*args
	|| (*args)->first() == NodeP(_Right_Paren)
	|| (type_of((*args)->first())==OBJECT
		&& (type_of(ObjectP((*args)->first())->proc_or_prim())==INFIX
			|| type_of(ObjectP((*args)->first())->proc_or_prim())==INFIXPROC)) ) 
	{
		if (min > 0) 
		{	_Not_Enough_node=new Line(0,0);
			return _Not_Enough_node;
		}
	}
	else if (max==0) 
	{	if (ifnode)
			if ((*args)->first()->islist())	// if -> ifelse kludge
			{	error(IF_WARNING, 0);
				*ifnode=_IfElse;

				NodeP retval=0;
				retval = Line::paren_expr(args, false, oldPriority);
				retval = retval->paren_infix(args, oldPriority, inparen, false);

				PairP someArgs=gather_some_args(
					min, max, args, inparen, oldPriority, 0);

				if (retval && retval->is.line)
				{	th->col=LineP(retval)->col;
					th->tcol=LineP(retval)->tcol;
					th->row=LineP(retval)->row;
					th->setfilename(LineP(retval)->filename());
				}
				return new Line(retval, someArgs);
			}
	}
	else 
	{	if (max < 0) max=0;   // negative max means unlimited

		if (!(*args)->first()
		|| ( (*args)->first() != NodeP(_Right_Paren)
			&& (type_of((*args)->first()) != OBJECT
				|| (type_of( ObjectP((*args)->first())->proc_or_prim() ) != INFIX
					&& type_of( ObjectP((*args)->first())->proc_or_prim() ) != INFIXPROC)) ) ) 
		{
			NodeP retval=0;
			retval = Line::paren_expr(args, false, oldPriority);
			retval = retval->paren_infix(args, oldPriority, inparen, false);
			bool savecr=false;
			int col=0, tcol=0, row=0;
			StringP fn=0;
			if (retval && !retval->is.line)
			{	savecr=true;
				col=th->col;
				tcol=th->tcol;
				row=th->row;
				fn=th->filename();
			}

			PairP someArgs=gather_some_args(
				min-1, max-1, args, inparen, oldPriority, ifnode);
			if (savecr)
			{	th->col=col;
				th->tcol=tcol;
				th->row=row;
				th->setfilename(fn);
			}else if (retval && retval->is.line)
			{	th->col=LineP(retval)->col;
				th->tcol=LineP(retval)->tcol;
				th->row=LineP(retval)->row;
				th->setfilename(LineP(retval)->filename());
			}
			return new Line(retval, someArgs);
		}
	}
	return 0;
}

const char Primitive::ok_if_no_arg = -2;
	// Special value for pmin, means that it's OK 
	// if primitive name is on a line by itself 
	//even though defltargs=1 (ED, CO)

LineP Procedure::gather_args(LinePP args, bool inparen, 
									  int oldPriority, ObjectPP ifnode)
// Gather the correct number of arguments to proc into a list. 
// Set args to immediately after the last arg.
{	int
		argmin = inparen ?   min  :  std,
		argmax = inparen ?   max  :  std;

	return gather_some_args(argmin, argmax, args, inparen, oldPriority, ifnode);
}
LineP Primitive::gather_args(LinePP args, bool inparen, 
									  int oldPriority, ObjectPP ifnode)
// Gather the correct number of arguments to proc into a list. 
// Set args to immediately after the last arg.
{
	int argmin = inparen ?   min  :  std;
	if (min == Primitive::ok_if_no_arg)	
	// Kludge to allow EDIT and CO without input without paren
		argmin=0;

	int argmax = inparen ?   max  :  std;

	return Procedure::gather_some_args(
		argmin, argmax, args, inparen, oldPriority, ifnode);
}

LineP Line::paren_expr(LinePP expr, bool inparen, int oldPriority) 
// Parenthesize an expression. 
// Set expr to the Node after the first full expression.
{
	ThreadP th=Thread::This();
	if throwing return *expr;
	th->autoThisLine=*expr;
	if (!*expr)
	{	if (inparen) error(PAREN_MISMATCH, *expr);
		return *expr;
	}
	NodeP nd = (*expr)->first();
	wxASSERT_MSG((*expr)->is.line,
		"somewhere I forgot to change a new Pair to a new Line.");

	if (LineP(*expr)->filename()->type==STRING)
	{	th->col=LineP(*expr)->col;
		th->tcol=LineP(*expr)->tcol;
		th->row=LineP(*expr)->row;
		th->setfilename(LineP(*expr)->filename());
//c_out << "PE(" << th->row << ')' << LineP(*expr) << "pe" << endL;
	}
	LineP expr1=*expr;
	*expr=LineP((*expr)->tail());
	LineP expr2=*expr;

	if (type_of(nd)==OBJECT  &&  !nd->numberP()) 
	{	ObjectP obj=ObjectP(nd);

		if (obj==_Left_Paren) 
		{	NodeP tree=0;
			tree = paren_expr(expr, true, -1);
			tree = tree->paren_infix(expr, -1, true, true);

			if (!*expr)
			{	error(PAREN_MISMATCH, expr1);
			}
			else if (ObjectP((*expr)->first()) != _Right_Paren)
			{	if (!throwing) 
					error(TOO_MUCH, 0);	// throw the rest away

				for (int parens=0;  *expr;  *expr=LineP((*expr)->tail()))
				{	if (ObjectP((*expr)->first())==_Left_Paren)
						parens++;
					else if (ObjectP((*expr)->first())==_Right_Paren)
						if (parens--==0) break;
				}
			}
			else 
				*expr=LineP((*expr)->tail());
			return LineP(tree);
		}
		else if (obj==_Right_Paren) 
		{	error(UNEXPECTED_PAREN, 0);
			if (inparen) 
			{	if (expr1)
				{	th->col=expr1->col;
					th->tcol=expr1->tcol;
					th->row=expr1->row;
					th->setfilename(expr1->filename());
				}
				*expr=new Line(obj, *expr);
			}
			return 0;
		}
		else if (obj==_Minus_Sign) 
		{	if (expr1)
			{	th->col=expr1->col;
				th->tcol=expr1->tcol;
				th->row=expr1->row;
				th->setfilename(expr1->filename());
			}
			LineP line=new Line(_Negate, *expr);
			if throwing return LineP(Unbound);
			NodeP retval=0;
//			retval = _Negate->proc_or_prim()->gather_args(
//					&line, inparen, priority_negate, 0);
//			retval = paren_expr(&line, false, priority_negate);
//			if (!throwing  &&  retval  &&  retval!=Unbound) 
				retval=LineP(line->paren_infix(&line, priority_negate, inparen, false));
			if (*expr)
				*expr=LineP((*expr)->tail());
			return LineP(retval);
		}else 
		{
			if (!obj->bit.permanent && obj->proc()==Unbound && th->thisobj)
			{	TableP t=th->thisobj->table();
				ObjectP o=t->find(ObjectP(obj)->name());
				if (o && o->bit.proc)
					obj=o;
			}
			if (!obj->bit.proc)	// only fix it if no proc or prim exists
			{	bool isvar=th->varStack->isvar(obj);
				if (isvar  &&  !throwing  &&  obj != _Null_Word )
				{
					if (expr1)
					{	th->col=expr1->col;
						th->tcol=expr1->tcol;
						th->row=expr1->row;
						th->setfilename(expr1->filename());
					}
					if (th->logoLanguage==_ucblogo)
					{	error(DK_HOW, obj); return 0;
					}else
					{	return new Colon(obj);	// fix colonless variables
					}
				}
			}
			// it must be a procedure
			NodeP proc_or_prim = Unbound;

			if ( obj->proc_or_prim()==Unbound  &&  obj != _Null_Word )
			{	obj->silent_load(environment.logolib.c_str()); // try <logolib>/<obj>.lg

				ObjectP procobj=th->namespce->find(obj->name());
				proc_or_prim = Unbound;
				if (procobj)
				{	proc_or_prim=procobj->proc_or_prim();
					if (proc_or_prim != Unbound)
						obj=procobj;
				}
			}
			if ( obj->proc_or_prim()==Unbound  &&  obj != _Null_Word)
			{	obj->silent_load(0);    // try ./<obj>.lg

				ObjectP procobj=th->namespce->find(obj->name());
				proc_or_prim = Unbound;
				if (procobj)
				{	proc_or_prim=procobj->proc_or_prim();
					if (proc_or_prim != Unbound)
						obj=procobj;
				}
			}
			if (proc_or_prim==Unbound || proc_or_prim==0)
				proc_or_prim=obj->proc_or_prim();

			if ((proc_or_prim==Unbound || !proc_or_prim) && !throwing) 
			{	Tree::dont_know_how_to=true;
//				if (theApp->autoloadEdit)
//				{	error(DK_HOW, obj); return 0;
//				}
				if (expr1)
				{	th->col=expr1->col;
					th->tcol=expr1->tcol;
					th->row=expr1->row;
					th->setfilename(expr1->filename());
				}
				return new Line(obj, 0);
			}
			else if ((	type_of(proc_or_prim)==INFIX 
						|| type_of(proc_or_prim)==INFIX)  &&  !throwing) 
			{	error(NOT_ENOUGH, obj);
				if (expr1)
				{	th->col=expr1->col;
					th->tcol=expr1->tcol;
					th->row=expr1->row;
					th->setfilename(expr1->filename());
				}
				return new Line(obj, 0);
			}else 
			{	// Kludge follows to turn IF to IFELSE sometimes.
				ObjectPP ifnode=0;
				if (obj==_If) 
					ifnode = &obj;

				LineP ret=LineP(Unbound);
				int col=th->col, tcol=th->tcol, row=th->row;
				StringP fn=th->filename();
				if (expr1)
				{	th->col=expr1->col;
					th->tcol=expr1->tcol;
					th->row=expr1->row;
					th->setfilename(expr1->filename());
				}
				int pri=-1;
//				if (!inparen)
					pri=proc_or_prim->priority();
				ret = proc_or_prim->gather_args(
					expr, inparen, pri, ifnode);

				if (ret != Unbound) 
				{	th->col=col;
					th->tcol=tcol;
					th->row=row;
					th->setfilename(fn);
					LineP l=new Line(obj, ret);
//c_out << "PE(" << l->row << ")" << l << "pe" << endL;
					return l;
				}
				else
					return LineP(Unbound);
			}
		}
	}
	else if (nd->islist())	// quoted list
	{	if (expr1)
		{	th->col=expr1->col;
			th->tcol=expr1->tcol;
			th->row=expr1->row;
			th->setfilename(expr1->filename());
		}
		return new Quote(nd);
	}else 
	{	if (expr1)
		{	th->col=expr1->col;
			th->tcol=expr1->tcol;
			th->row=expr1->row;
			th->setfilename(expr1->filename());
		}
		return LineP(nd);
	}
}

Priority Object::infixPriority()	// Calculate the infixPriority of a procedure.
{
//	if (ObjectP(this)==_Negate) 
//		return priority_negate;

	NodeP p=proc_or_prim();

	if (!p || p==Unbound  ||  (p->type != INFIX  &&  p->type != INFIXPROC))
		return priority_STOP;

	return PrimitiveP(p)->_priority;
}

// Parenthesize an infix expression.
// <this> is the expression on the left (already parenthesized), 
// and <rest> is a pointer to the list starting with the infix procedure, if it's there.  
// Set rest to after the right end of the infix expression.

NodeP Node::paren_infix(LinePP rest, int oldPriority, 
	bool inparen, bool firstInParen)
{
//c_out << "PIr|" << *rest << "|" << endL;
	if (!*rest) return this;	// this==left_arg

	NodeP infix_proc = NodeP((*rest)->first());

	if (!infix_proc) return this;

	ThreadP th=Thread::This();

	int priority = infix_proc->infixPriority();
	if (priority==priority_STOP  ||  priority <= oldPriority) 
		return this;

	*rest=LineP((*rest)->tail());
	NodeP retval=0;
	retval = Line::paren_expr(rest, inparen, oldPriority);
//c_out << "PIr2|" << *rest << "|" << endL;
//c_out << "PI|" << retval << "|" << endL;
	SET_th_LINE(retval);
	retval = retval->paren_infix(rest, priority, inparen, false);
//c_out << "PIr3|" << *rest << "|" << endL;
//c_out << "PI2|" << retval << "|" << endL;

	if (infix_proc==_Dot && *rest && firstInParen
	&& ObjectP((*rest)->first())!=_Right_Paren
	&& (!ObjectP((*rest)->first())->bit.proc
		|| (  ObjectP((*rest)->first())->bit.proc
			&& ObjectP((*rest)->first())->bit.permanent
			&& ObjectP((*rest)->first())->prim()->_priority==priority_prefix)))
	{	NodeP retval2=0;
		LineP tmpline=0, last=0;
		while (retval2!=Unbound
		&& *rest 
		&& ObjectP((*rest)->first())!=_Right_Paren 
		&& (*rest)->tail() 
		&& !throwing)
		{	retval2 = Line::paren_expr(rest, inparen, oldPriority);
			retval2 = retval2->paren_infix(rest, oldPriority, inparen, false);
//c_out << "PIr4|" << *rest << "|" << endL;
//c_out << "PI4|" << retval2 << "|" << endL;
			SET_th_LINE(retval);
			LineP tnode=new Line(retval2, 0);
			if (last)
			{	last->settail(tnode);
			}else
				tmpline=tnode;
			last=tnode;
		}
		if throwing return Unbound;
		SET_th_LINE(retval);
		retval = new Line(infix_proc, 
						new Line(this, 
							new Line(retval, tmpline)));
	}
	else
	{	if throwing return Unbound;
		SET_th_LINE(retval);
		retval = new Line(infix_proc, 
						new Line(this, 
							new Line(retval, 0)));
/*		SET_th_LINE(retval);
		LineP rv=new Line(retval, 0);
		SET_th_LINE(this);
		LineP rvp=new Line(this, rv);
		SET_th_LINE(infix_proc);
		retval = new Line(infix_proc, rvp);
*/	}
//c_out << "PI(" << LineP(retval)->linenr << ")" 
//<< LineP(retval)->filename() << "|" << retval << "|" << endL;
	if throwing return Unbound;
	retval=retval->paren_infix(rest, oldPriority, inparen, false);
//c_out << "PIr4|" << *rest << "|" << endL;
//c_out << "PI3|" << retval << "|" << endL;
	return retval;
}

LineP Line::paren_line() 
// Fully parenthesize a complete line, i.e. transform it from a flat list to a tree.
{
	if (!this) return this;
	ThreadP th=Thread::This();
	LineP line=this;
	NodeP retval=0;
	retval = paren_expr(&line, false, -1);
//c_out << "PL(" << LineP(retval)->col << ' ' << LineP(retval)->row << ')' << '|' << retval << "|" << endL;
//c_out << "PL|" << retval << "|" << endL;

	if (!throwing  &&  retval  &&  retval!=Unbound) 
	{	retval = retval->paren_infix(&line, -1, false, false);
		LineP plresult = line->paren_line();
//c_out << "PL2|" << plresult << "|" << endL;
		SET_th_LINE(retval);
		retval = new Line( retval, plresult );
//c_out << "PL2(" << LineP(retval)->linenr << ")" << filename() << "|" << retval << "|" << endL;
	}
	return LineP(retval);
}

void Pair::make_runparse(bool deep)
{
//	wxMutexError err=parenMutex.Lock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);

	ThreadP th=Thread::This();
	if (this && islist())
	{	if (!is.line)
		{	settype_LINE();
			LineP l=LineP((Line*)this);
			l->col=th->col;
			l->tcol=th->tcol;
			l->row=th->row;
			l->setfilename(th->filename());
		}else
		{
			LineP l=LineP((Line*)this);
			th->col =l->col;
			th->tcol=l->tcol;
			th->row =l->row;
			th->setfilename(l->filename());
		}
	}
	else
	{	error(BAD_DATA_UNREC, this);
		return;
	}
	LineP runparsed=runparse(deep);
	settype_RUNPARSED();
	((Runparsed*)this)->setparsed(runparsed);

//	err=parenMutex.Unlock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);
}

void Pair::makeTree(bool deep) // Treeify a list of tokens (runparsed or not).
{
	if (!this
	|| (type==TREE  &&  LineP(this)->generation==Line::the_generation))
		return;

//	wxMutexError err=parenMutex.Lock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);

	if (type==TREE)
		settype_RUNPARSED();
	else if (type!=RUNPARSED)
		make_runparse(deep);
	if throwing return;

	RunparsedP runp = (Runparsed*)this;
	Tree::dont_know_how_to=false;
	PairP tree=0;
wxASSERT(runp->parsed()->is.line || runp->parsed()->type==QUOTE);
	tree = runp->parsed()->paren_line();
//c_out << "MT" << tree << "E"<< endL; //lpause(0);

	if (tree  &&  tree != Unbound) 
	{
		wxASSERT(runp->type!=OBJECT);
		runp->settype_TREE();
		TreeP(runp)->settree(tree);

// Kludge! (temporarily commented out for making colonless local vars work)
		if (/*Tree::dont_know_how_to ||*/ throwing)
			LineP(runp)->generation=0;
		else
			LineP(runp)->generation=Line::the_generation;
	}
//	err=parenMutex.Unlock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);
}

NodeP lmakeTree(PairPP arg)
{
	PairP expr=arg[1]->a_runnable(1);
	expr->makeTree();
	return TreeP(expr)->tree()->first();
}

void Pair::makeTreeFromBody(bool deep) // Treeify a procedure`s body by appending the trees of the lines.
{
	Tree *body = (Tree*) this;

	if (!body
	||  (type==TREE  &&  body->generation==Line::the_generation) )
		return;

//	wxMutexError err=parenMutex.Lock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);

//	c_out << "B" << body << endL;
	ThreadP th=Thread::This();
	if (type==PAIR)
	{	settype_LINE();
		LineP l=LineP((Line*)this);
		l->col=th->col;
		l->tcol=th->tcol;
		l->row=th->row;
		l->setfilename(th->filename());
	}
	if (!(type==LINE || type==RUNPARSED || type==TREE))
	{	error(BAD_DATA, this); return;
	}
	settype_TREE();
	body->settree(NIL); 
	body->generation=0;

	PairP end_ptr=0;

	for (PairP body_ptr=body;  body_ptr;  next(body_ptr)) 
	{
		PairP line = PairP(body_ptr->first());

		if (!line) continue;  // skip blank line

		th->this_line=LineP(line);
//		c_out << "maketfb " << line << endL;
		line->makeTree(deep);
		if (line->type==TREE) 
		{	TreeP tree = TreeP(TreeP(line)->tree());
			tree->setline(body_ptr->first());
//			c_out << "E" << end_ptr << "T" << tree << endL;


			if (end_ptr)
			{	end_ptr->settail( tree );
/*
if (tree->circular())	// there`s a CIRCULAR-LIST-BUG somewhere in the parser !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
{	c_out << "CIRC";
	TreeP t=tree;
	for (int i=20;  i && t;  i--)
	{	if (t->first())	
			c_out << t->first() << endL;
		next(t);
	}
	console->getch();
	end_ptr->settail(0);	// really bad slow ugly KLUDGE!!!
}
*/
			}else
			{	wxASSERT(body->type==TREE);
//body->settype_TREE();
				body->settree(tree);
				body->generation=Line::the_generation;
			}
			if ( LineP(body_ptr->first())->generation==0)
				body->generation=0;

			while (tree->tail())
				next(tree);

			end_ptr=tree;
		}else				// error while treeifying
			body->untreeify();
	}
//c_out << this; lpause(0);
//	err=parenMutex.Unlock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);
}

NodeP lmakeTreeFromBody(PairPP arg)
{
	PairP expr=arg[1]->a_runnable(1);
	if throwing return Unbound;
	expr=new Line(expr, 0);
	if throwing return Unbound;
	expr->makeTreeFromBody(true);
	if throwing return Unbound;
	return TreeP(expr)->tree();
}
