#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif
	
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <wx/string.h>
#include <wx/thread.h>
#include <wx/textctrl.h>

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Tree.h"
#include "Quote.h"
#include "Procedure.h"
#include "Object.h"
#include "Array.h"

#include "Error.h"
#include "Init.h"
#include "Color.h"
#include "Security.h"
#include "Thread.h"
#include "Console.h"
#include "Consolew.h"
#include "Stack.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//static wxMutex parseMutex(wxMUTEX_RECURSIVE);

bool isNewline   (char c) {return c=='\n' || c=='\r';}
bool isWhite(char c) 
{	return c==' ' || c=='\t' || isNewline(c) || c=='\0';
}

bool wasSpecialForm(charP source, int pos, ObjectPP specialForm)
{
	ThreadP th=Thread::This();
	bool sawSpecialForm=false;
	*specialForm=_Null_Word;
	int cnt=th->specialForms->length();
	for (int i=0; i<cnt; i++)
	{	ObjectP obj=ObjectP(th->specialForms->item(i));
		StringP form=obj->name();
		int len=form->length();
		if (pos >= len)
		{	charP f=form->ptr(), 
				l=&source[pos-len];
			if (String::comp(f, l, len)==0)
			{	if (pos==len 
				|| isWhite(source[pos-len-1]))
				{	sawSpecialForm=true;
					*specialForm=obj;
//c_out << obj->upper();
				}
			}
		}
	}
	return sawSpecialForm;
}

bool isSpecialForm(charP source, charP limit, ObjectPP specialForm)
{
	ThreadP th=Thread::This();
	bool sawSpecialForm=false;
	*specialForm=_Null_Word;
	int pos=limit-source;
	int cnt=th->specialForms->length();
	for (int i=0; i<cnt; i++)
	{	ObjectP obj=ObjectP(th->specialForms->item(i));
		StringP form=obj->name();
		int len=form->length();
		if (pos >= len)
		{	charP f=form->ptr(), 
				l=source;
			if (String::comp(f, l, len)==0)
			{	if (pos==len 
				|| isWhite(source[len]))
				{	sawSpecialForm=true;
					*specialForm=obj;
//c_out << obj->upper();
				}
			}
		}
	}
	return sawSpecialForm;
}

StringP File::read(charP prompt) 
{
	ThreadP th=Thread::This();
	if (file()==conin && !th->security->keyboard)
	{	error(SECURITY_ERROR, new String(_("to use keyboard"))); 
		return StringP(Unbound);
	}
	int specialFormsCount=th->specialForms->length();
	ObjectP specialForm=_Null_Word;
	bool sawSpecialForm=false;
	int dribbling, 
		vbar=0, paren=0, bracket=0, brace=0, to=0,
		pos=0, contin=1, insemi=0, raw=0;

	if (!strcmp(prompt, "RW"))	// called by readWord
	{	prompt="";
		contin=0;
	}
	if (!strcmp(prompt, "RAW")) // called by readRawLine
	{	prompt="";
		contin=0;
		raw=1;
	}
	dribbling=0;

	if (!th->line) 
	{	th->line = new String(File::lineMax);
		if throwing return StringP(_Null_Word);
		th->lineEnd = File::lineMax-1;
	}
	bool oldCharMode;
	if (console)
		oldCharMode=console->charMode;
	else if (consolew)
		oldCharMode=consolew->charMode;
	charModeOff();

	if (file()==conin && *prompt) 
		if (th->interactive)
			printPrompt(prompt);

	if (file()==conin) 
	{	th->inputBlocking++;
		th->erract_errtype=FATAL;
		th->setfilename(Line::findFile(_Unnamed->name()));
	}

	bool sawWhite=true, lastSawWhite=true, backslash=false;
	int c=0, cl=0;
	NodeType thisType=STRING;
	int isp=*th->intStack;

	if (!setjmp(th->inputBlockingBuffer)) 
	{	c=read_getc();
		cl=wxTolower(c);

		while (c!=EOF 
		&& (vbar || paren || bracket || brace || to > 0 || c !='\n' || th->toPending)
		&& !throwing) 
		{
			if (dribbling) dribbler.putc_(c);

			if (!raw  &&  c=='\\')//  &&  (c=read_getc())!=EOF) 
			{	if (dribbling) 
					dribbler.putc_(c);

//				c = backslash.encode(c);
//				thisType=BACKSLASHED;
				
//				if (c==backslash.encode('\n') && file()==conin) 
//					if (th->interactive) 
//						printPrompt("\\ ");
				if (c!=EOF) 
				{	if (pos >= th->lineEnd)
					{	th->lineLen += File::lineMax;
						th->line = new String(th->line->ptr(), 0, th->lineLen, STRING, String::copy);
						if throwing 
						{	th->intStack->pop_till(isp);
							return StringP(_Null_Word);
						}
						th->lineEnd = th->lineLen-1;
					}
					th->line->ptr()[pos]=char(c);
					pos++;
				}
				c=read_getc();
				if (c!=EOF) 
				{	if (pos >= th->lineEnd)
					{	th->lineLen += File::lineMax;
						th->line = new String(th->line->ptr(), 0, th->lineLen, STRING, String::copy);
						if throwing 
						{	th->intStack->pop_till(isp);
							return StringP(_Null_Word);
						}
						th->lineEnd = th->lineLen-1;
					}
					th->line->ptr()[pos]=char(c);
					pos++;
					if (c!=EOF 
					&& (vbar || paren || bracket || brace || to > 0 || c !='\n' || th->toPending)
					&& !throwing) 
					{	c=read_getc();
						continue;
					}else
						break;
				}
			}
			if (c!=EOF) 
			{	if (pos >= th->lineEnd)
				{	th->lineLen += File::lineMax;
					th->line = new String(th->line->ptr(), 0, th->lineLen, STRING, String::copy);
					if throwing 
					{	th->intStack->pop_till(isp);
						return StringP(_Null_Word);
					}
					th->lineEnd = th->lineLen-1;
				}
				th->line->ptr()[pos]=char(c);
				pos++;
			}
			if (raw)
			{	c=read_getc();
				continue;
			}
			if (c=='\n') 
				insemi=0;	
			
			if (isWhite(c))
			{	sawWhite=true;
//c_out << ".";
			}
			cl=wxTolower(c);

			if (!paren && !bracket && !brace && !insemi && !vbar && contin)
			{
				if (sawSpecialForm && sawWhite)
				{	if (specialForm->prim()==_End->prim())
					{	Pop1(th->intStack, c, to); 
						if (c=='t')
							to--;
						else
						{	th->line->ptr()[pos]='\0';
							th->row=th->readnr;
							error(UNEXPECTED_END, 
								new Line(String::make(th->line->ptr(), 0, strlen(th->line->ptr()), 
									thisType, String::copy), 0));
						}
					}else
					{	to++;	Push(th->intStack, 't', to); 
					}
				}
				sawSpecialForm=wasSpecialForm(th->line->ptr(), pos, &specialForm);
			}
			if (!insemi && c=='|') 
			{	vbar=!vbar;	//thisType=VBAR_STRING;
			}
			else if (contin && !vbar && !insemi) 
			{	switch (c)
				{	case '(': 
						paren++;	
						if (!bracket && !brace) 
							Push(th->intStack, c, paren); 
						break;
					case ')': 
						if (!bracket && !brace)
						{	Pop1(th->intStack, c, paren); 
							if (c=='(')
								paren--;
							else
							{	th->line->ptr()[pos]='\0';
								th->row=th->readnr;
								error(UNEXPECTED_PAREN, 
									new Line(String::make(th->line->ptr(), 0, strlen(th->line->ptr()), 
										thisType, String::copy), 0));
							}
						}else if (paren)
							paren--;
						break;
					case '[': 
						bracket++; Push(th->intStack, c, bracket); 
						break;
					case ']': 
						Pop1(th->intStack, c, bracket); 
						if (c=='[')
							bracket--;
						else
						{	th->line->ptr()[pos]='\0';
							th->row=th->readnr;
							error(UNEXPECTED_BRACKET, 
								new Line(String::make(th->line->ptr(), 0, strlen(th->line->ptr()), 
									thisType, String::copy), 0));
						}
						break;
					case '{': 
						brace++; Push(th->intStack, c, brace); 
						break;
					case '}': 
						Pop1(th->intStack, c, brace); 
						if (c=='{')
							brace--;
						else
						{	th->line->ptr()[pos]='\0';
							th->row=th->readnr;
							error(UNEXPECTED_BRACE, 
								new Line(String::make(th->line->ptr(), 0, strlen(th->line->ptr()), 
									thisType, String::copy), 0));
						}
						break;
					case ';': insemi++;	break;
				}
			}
			if throwing break;
//			if (thisType==STRING && strchr(backslash.chars, c))
//				thisType=VBAR_STRING;

			while (!vbar && c=='~' && (c=read_getc())!=EOF) 
			{	while (c==' ' || c=='\t')
					c=read_getc();
				
				if (dribbling) dribbler.putc_(c);
				if (pos >= th->lineEnd)
				{	th->lineLen += File::lineMax;
					th->line = new String(th->line->ptr(), 0, th->lineLen, STRING, String::copy);
					if throwing 
					{	th->intStack->pop_till(isp);
						return StringP(_Null_Word);
					}
					th->lineEnd = th->lineLen-1;
				}
				th->line->ptr()[pos]=char(c);
				pos++;
				
				if (c=='\n' && file()==conin) 
				{	insemi=0;	//if (interactive) zrd_print_prompt("~ ");
				}
			}
			lastSawWhite=sawWhite;
			sawWhite=false;
//c_out << char(c); console->flush(); th->dispatchMessages();
			if (c!=EOF) c=read_getc();
			cl=wxTolower(c);
		}
	}
	if (isNewline(c))
	{	if (pos >= th->lineEnd)
		{	th->lineLen += File::lineMax;
			th->line = new String(th->line->ptr(), 0, th->lineLen, STRING, String::copy);
			if throwing 
			{	th->intStack->pop_till(isp);
				return StringP(_Null_Word);
			}
			th->lineEnd = th->lineLen-1;
		}
		th->line->ptr()[pos]=char(c);
		pos++;
	}

	th->line->ptr()[pos]='\0';
//c_out << th->line->ptr() << "</read>" << endL;
	th->inputBlocking=0;

	if (dribbling) dribbler.putc_('\n');
/*	if (c==EOF && file()==conin) 
	{
		if (th->interactive) 
			clearerr(conin);
		printPrompt("\n");
	}
*/
	if (to && sawSpecialForm)
	{	if (specialForm->prim()==_End->prim())
		{	Pop1(th->intStack, c, to); 
			if (c=='t')
				to--;
			else
			{	th->line->ptr()[pos]='\0';
				th->row=th->readnr;
				error(UNEXPECTED_END, 
					new Line(String::make(th->line->ptr(), 0, strlen(th->line->ptr()), 
						thisType, String::copy), 0));
			}
		}
	}
	File::charModeOn();
	th->intStack->pop_till(isp);
	if (vbar || paren || bracket || brace || to > 0)
	{	StringP result=String::make(th->line->ptr(), 0, strlen(th->line->ptr()), 
			thisType, String::copy);
		if (vbar)
			error(VBAR_MISMATCH, result);
		else if (paren)
			error(PAREN_MISMATCH, result);
		else if (bracket)
			error(BRACKET_MISMATCH, result);
		else if (brace)
			error(BRACE_MISMATCH, result);
		else if (to)
			error(END_MISMATCH, result);
		return StringP(_Null_Word);
	}
	if (pos==0) 
		return StringP(_Null_Word);		// so emptyp works
	
	StringP result=String::make(th->line->ptr(), 0, strlen(th->line->ptr()), 
		thisType, String::copy);
	return result;
}

PairP parser_iterate(charPP inln, charP inlimit, String::HeadP inhead, 
							bool semi, int endchar, StringP lexufun, bool inparen, 
							int &col, int &tcol, charP &lineStart, LineP &lines) 
{
	ThreadP th=Thread::This();
	int windex=0, vbar=0;
	bool broken=false, inString=false, lastInString=false, sawSpecialForm=false,
		backslash=false, wasBackslash=false;

	char ch='\0';
	charP wptr=0;
	static char terminate='\0';   // KLUDGE

	NodeType this_type=STRING;
	PairP tnode=0, lastnode=0, outline=0;
	bool run=true;
	ObjectP specialForm=_Null_Word;

	ch = **inln;
	if (*inln==inhead->str && ch=='#' && *inln+1 <= inlimit && (*inln)[1]=='!')
		while (ch && !isNewline(ch) && *inln < inlimit)
		{
//c_out << ch; console->flush(); console->dispatchMessages();
			(*inln)++;
			ch = **inln;
		}

	while (ch && run)
	{	ch = **inln;			// get the current character and increase pointer
//c_out << ch; console->flush(); th->dispatchMessages();
		//if (!vbar && windex==0)
		if (windex==0)
			wptr = *inln;

		(*inln)++;
		col++;
		tcol++;
		if (*inln >= inlimit)
			*inln = &terminate;

		if (!backslash && !vbar)
		{							 // skip through comments and line continuations
			while ((semi && ch==';') || (ch=='~' && isNewline(**inln))) 
			{	while (ch=='~' && isNewline(**inln)) 
				{	th->row++;
					col=0;
					tcol=0;
					(*inln)++;
					if (*inln >= inlimit) 
						*inln = &terminate;
					ch = **inln;
					if (windex==0) 
						wptr = *inln;
					else 
					{	if (**inln==']' || **inln=='[' || **inln=='{' || **inln=='}') 
						{	ch=' ';
							break;
						}else 
							broken=true;
					}
					(*inln)++;
					col++;
					tcol++;
					if (*inln >= inlimit) 
						*inln = &terminate;
				}
				if (semi && ch==';') 
				{
					if (!isNewline(ch))
					{	do
						{	ch = **inln;
							if (windex==0) 
								wptr=*inln;
							else 
								broken=true;

							(*inln)++;
							col++;
							tcol++;
							if (*inln >= inlimit)
								*inln = &terminate;
						} while (ch !='\0' && ch !='~' && !isNewline(ch));
					}
					if (isNewline(ch))
					{	col=0;
						tcol=0;
	//					th->row++;
					}
					if (ch !='\0') 
					{	ch='\n';
	//					(*inln)++;
						col++;
						tcol++;
					}
					if (ch=='~')
						(*inln)++;
				}
			}
		}
//		if (backslash.is(ch))// && this_type!=VBAR_STRING)	
//			this_type=BACKSLASHED;
			
		if (ch=='|') 
		{	vbar = !vbar;
//				this_type=VBAR_STRING;
			broken=true;	// so we'll copy the chars
			windex++;
		}else 
			if ( vbar || backslash
			|| (!isWhite(ch) && ch !=']' &&	ch !='{' && ch !='}' && ch !='[') )
				windex++;
	
		if (vbar) 
			goto vbar_continue;

		sawSpecialForm=isSpecialForm(*inln-1, inlimit, &specialForm);
		if (!inparen && !inString && endchar=='E' && *inln+2 <= inlimit
		&&  specialForm->prim()==_End->prim())
		{	int len=specialForm->name()->length();
			if (isWhite((*inln)[len-1]))
			{	*inln+=len-1;
				run=false;
			}
		}
/*		if (!inparen && !inString && endchar=='E' && *inln+2 <= inlimit)
		{	if (wxToupper(ch)=='E' 
			&&  wxToupper((*inln)[0])=='N' 
			&&  wxToupper((*inln)[1])=='D' && isWhite((*inln)[2]))
			{	if (*inln+3 <= inlimit)
				{	*inln+=2;
				}else
					*inln+=2;	// END at the end of text
				run=false;
//				th->row++;
			}
		}
*/		else if (endchar=='\n' && isNewline(ch))
			run=false;
		else if (ch==endchar && endchar!='E' && !backslash) 
		{	break;
		}
		lastInString=inString;
		if (isalnum((unsigned char)ch) 
		|| ch=='"' || ch=='_' || ch==':' || ch=='\\')
		{	inString=true;
			if (!lastInString)
			{	th->col=col;
				th->tcol=tcol;
			}
		}
		if (isWhite(ch) 
		|| ch=='[' || ch==']' 
		|| ch=='(' || ch==')'
		|| ch=='{' || ch=='}')
			inString=false;

		if (run && !lastInString && !inparen && sawSpecialForm)
		{	// there are just "TO", "BE" and "_MACRO" 
			// as special forms at the moment. 
			// Translations of them are now possible.
			int tmpcol=col, tmptcol=tcol, tmprow=th->row;
			int len=specialForm->name()->length();
			*inln+=len; col+=len; tcol+=len;

			bool dynFlag=true;
			if (specialForm->prim() == _Be->prim()
			||  specialForm->prim() == __BeMacro->prim())
				dynFlag=false;
			bool macroFlag=false;
			if (specialForm->prim() == __Macro_->prim()
			||  specialForm->prim() == __BeMacro->prim())
				macroFlag=true;

			charP cp=*inln;
			while (*cp && (*cp==' ' || *cp=='\t' || *cp=='\r' || *cp=='\n'))
			{	cp++;
			}
			*inln=cp;
			bool infix=false;
			StringP left=0;
			if (*cp && *cp==':')
			{	infix=true;
				while (*cp && *cp!=' ' && *cp!='\t' && *cp!='\r' && *cp!='\n')
				{	cp++;
				}
				left=String::make(*inln, 0, cp-(*inln), STRING, String::copy);
				*inln=cp;
				while (*cp && (*cp==' ' || *cp=='\t' || *cp=='\r' || *cp=='\n'))
				{	cp++;
				}
				*inln=cp;
			}
			while (*cp && *cp!=' ' && *cp!='\t' && *cp!='\r' && *cp!='\n')
			{	cp++;
			}
			StringP name=String::make(*inln, 0, cp-(*inln), STRING, String::copy);
			StringP name2=String::make(*inln-1, 0, cp-(*inln-1), STRING, String::copy);
			if throwing goto vbar_continue;
			if (lexufun)
				name=new String(wxString(lexufun->ptr())+"::"+name->ptr()); 
			if throwing goto vbar_continue;
			*inln=cp;
			
			StringP fn=th->filename();
			tmpcol=col;
			tmptcol=tcol;
			tmprow=th->row;
			LineP tmplines=lines;
			lines=0;
			lineStart=*inln;
			NodeP formals=parser_iterate(inln, inlimit, inhead, semi, '\n', 
				name, inparen, col, tcol, lineStart, lines);
			if throwing goto vbar_continue;
			if (left)
				formals=new Line(left, formals);
			if throwing goto vbar_continue;

			ObjectP symbol=specialForm; //_To;
//			if (!dynFlag)
//				symbol=_Be;
//			if (macroFlag)
//				symbol=__Macro_;
			lines=new Line(symbol->name()->lPut(name2)->lPut(lines->reverse_destruct()), 0);
			if throwing goto vbar_continue;
			lines=LineP(lines->reverse_destruct());
			if throwing goto vbar_continue;
			fn=th->filename();
			tmpcol=col;
			tmptcol=tcol;
			tmprow=th->row;

			NodeP body=parser_iterate(inln, inlimit, inhead, semi, 'E', 
				name, inparen, col, tcol, lineStart, lines);
			if throwing goto vbar_continue;

			StringP fn2=th->filename();
			int tmpcol2=col, tmptcol2=tcol, tmprow2=th->row;

			col=tmpcol;
			tcol=tmptcol;
			th->row=tmprow;
			th->setfilename(fn);

			if (body)
				body=new Line(body, 0);
			
//c_out << name << "</name>" << endL;
//c_out << formals << "</formals>" << endL;
//c_out << body << "</body>" << endL;
			LineP proc= new Line(formals, body);
			if throwing goto vbar_continue;
			ProcedureP procedure=proc->define_helper(macroFlag, dynFlag, PairP(name), left);
			if throwing goto vbar_continue;
			lines=new Line(_End, lines);
			if throwing goto vbar_continue;
			procedure->setbodywords(PairP(lines->reverse_destruct()));
			if throwing goto vbar_continue;
			lines=tmplines;
			lineStart=*inln;
			col=tmpcol2;
			tcol=tmptcol2;
			th->row=tmprow2;
			th->setfilename(fn2);
			inString=false;
			windex=0;
//			ch=**inln;
//			(*inln)++;
//			(*col)++;
//			(*tcol)++;
			if (**inln=='\0') ch='\0';
		}

		if (backslash)
			goto backslash_continue;
		
		switch (ch)
		{
		case ']':	
			th->col=col;
			error(UNEXPECTED_BRACKET, new Line(0,0));	break;
		case '}':	
			th->col=col;
			error(UNEXPECTED_BRACE  , new Line(0,0));	break;

		case '[':	// parse a new list
			{
				StringP fn=th->filename();
				th->col=col;
				th->tcol=tcol;
				int tmpcol=th->col, tmptcol=th->tcol, tmprow=th->row;

				NodeP nd=parser_iterate(inln, inlimit, inhead, semi, ']', 
					lexufun, true, col, tcol, lineStart, lines);
				
				StringP fn2=th->filename();
				int tmpcol2=th->col, tmptcol2=th->tcol, tmprow2=th->row;

				th->col=tmpcol;
				th->tcol=tmptcol;
				th->row=tmprow;
				th->setfilename(fn);
				tnode = new Line( nd, 0 );
				th->col=tmpcol2;
				th->tcol=tmptcol2;
				th->row=tmprow2;
				th->setfilename(fn2);

				if (**inln=='\0') ch='\0';
				break;
			}
		case '{': 
			{	PairP p=parser_iterate(inln, inlimit, inhead, semi, '}', 
					lexufun, true, col, tcol, lineStart, lines);
//				if (!p)
//				{	th->col=col;
//					error(UNEXPECTED_BRACE, new Line(0,0)); //   break;
//				}
				if throwing break;
				int origin=1;
				if (**inln=='@') 
				{	int i=0, sign=1;
					(*inln)++;
					col++;
					tcol++;
					if (**inln=='-') 
					{	sign=-1;	(*inln)++; col++; tcol++;
					}
					while ((ch=**inln) >='0' && ch <='9')	// Ooops! No expressions as array origins allowed yet!
					{	i=(i*10) + ch - '0';
						(*inln)++; col++; tcol++;
					}
					origin=sign*i;
					ch='}';
				}
				if (p)
					tnode = new Line(NodeP(p->array(origin)), 0);
				else
				{	ArrayP a=new Array(0);
					tnode= new Line(a, 0);
				}
				if (**inln=='\0') ch='\0';
				break;
			}
		default:	// if this character or the next one will terminate string, make the word
backslash_continue:
			if ((		(!backslash && ch!='\\') 
					|| (ch!='\\' && **inln==endchar))
			&& (isWhite(ch) || **inln==']' || **inln=='[' || **inln=='{' || **inln=='}'))
				{	if (windex > 0) 
					{	/*if (this_type==VBAR_STRING)
							tnode = new Line
							(	String::make
								(	wptr, 0, windex, this_type, String::copy_vbar
								), 
								0
							);
						else
						*/
						//	if (broken==false)
								StringP s=String::make(wptr, inhead, windex, this_type, String::copy);
								if (!s) return LineP(Unbound);
								s=s->noPepperIfNotInString();
//c_out << 'A' << s << 'E' << ch;
								tnode = new Line(s, 0);
						//	else 
						//	{	tnode = new Line
						//		(	String::make
						//			(	wptr, 0, windex, this_type, //String::copy
						//				semi ?   String::copy_mend  :  String::copy_mend_nosemi 
						//			), 
						//			0
						//		);
						//		broken=false;
						//	}
						this_type=STRING;
						windex=0;
					}
				}
		}
vbar_continue:
		if (isNewline(ch) || ch=='\n')//backslash.encode('\n'))
		{	int len=*inln-lineStart;
			if (len > 0)
				lines=new Line
				(	String::make(lineStart, inhead, *inln-lineStart, STRING, String::copy), 
					lines
				);
			lineStart=*inln;
//c_out << "PI(" << th->nr << ")" << th->filename() << "|" << outline << "|" << endL;
			col=0;
			tcol=0;
			th->row++;
		}
		if (ch=='\t')
		{	tcol+=String::tabSize-1-(col-1) % String::tabSize;
		}
		if (tnode)	// put the word onto the end of the return list
		{	if (outline)
			{	while (lastnode->tail())
					next(lastnode);
				lastnode->settail( tnode );
			}else
				outline=tnode;
			
			lastnode=tnode;
			tnode=0;
		}
		wasBackslash=backslash;
		if (ch=='\\' && !backslash)
			backslash=true;
		else
			backslash=false;
	}
//c_out << outline << endL; //lpause(0);
	return outline;
}

PairP String::parser(bool semi, bool inparen) 
{
	if (!this)
		return new Line(0,0);
	StringP str=0;	str = string();
	if (str==Unbound)
		return new Line(0,0);
	int slen = str->length();
	charP lnsav = str->ptr();

	if (!lnsav)
	{	c_out << this << endL; 
		return new Line(0, 0);
	}
//	wxMutexError err=parseMutex.Lock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);

	charP lineStart=lnsav;
	PairP ret=0;
	LineP lines=0;
	int col=0, tcol=0;
	ThreadP th=Thread::This();
	StringP parent=0;
	if (th->lexufun)
		parent=th->lexufun->name();
	ret = parser_iterate(&lnsav, lnsav+slen, str->head(), semi, -1, parent,
		inparen, col, tcol, lineStart, lines);

//	err=parseMutex.Unlock();
//	wxASSERT(err==wxMUTEX_NO_ERROR);
	return ret;
}

NodeP lparse(PairPP arg) 
{
	StringP s = arg[1]->a_string(1);
	if throwing return Unbound;
	return s->parser(false);
}
