#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "File.h"
#include "IntToPtr_.h"
#include "PtrToInt_.h"
#include "StringToPtr.h"
#include "PtrToString.h"
#include "String.h"
#include "Int.h"
#include "Primitive.h"
#include "Object.h"
#include "Table.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#else
	static char THIS_FILE[]= __FILE__;
#endif

Primitive::Primitive(PrimType &prim):
	_function(prim.function), _priority(prim.priority),
	min(prim.min), std(prim.std), max(prim.max)
{
	if (prim.priority==priority_MACRO)
		type=MACRO;
	else if (prim.priority <= priority_TAIL)
		type=TAILFORM;
	else if (prim.priority == prefix)
		type=PRIM;
	else if (prim.priority == priority_negate)
		type=PRIM;
	else
		type=INFIX;
}

DataInputStream &Primitive::read(DataInputStream &istr) 
{	wxString name;
	istr >> name
		>> (unsigned char&)_priority 
		>> (unsigned char&)min 
		>> (unsigned char&)std 
		>> (unsigned char&)max;
	_function = FunctionP(GC_::primitivesStringToPtr->find(name)); 
	if (!_function)
	{	error(FILE_ERROR, new String(
			(wxString(_("Function "))+name
			 +_(" not available in this release"))));
		_function=&lError;
	}
	return istr;
}
wxDataOutputStream &Primitive::write(wxDataOutputStream &ostr) 
{	ostr << type 
		<< GC_::primitivesPtrToString->find((void*)_function) 
		<< (unsigned char)_priority 
		<< (unsigned char)min 
		<< (unsigned char)std 
		<< (unsigned char)max;
	return ostr;
}

NodeP Primitive::copy(PtrToInt_ &ptrToInt)
{	PrimitiveP prim=new Primitive();
	if throwing return Unbound;
	prim->_function=_function;
	prim->_priority=_priority;
	prim->min=min;
	prim->std=std;
	prim->max=max;
	return prim;
}

void Primitive::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, _("Primitive ["));
	::print(strm, width, int(min));
	::print(strm, width, ' ');
	::print(strm, width, int(std));
	::print(strm, width, ' ');
	::print(strm, width, int(max));
	::print(strm, width, ' ');
	::print(strm, width, int(_priority));
	::print(strm, width, ']');
}

void Primitive::printBin(FILEP strm)
{
	int width=Int::max;
	print(strm, width, Int::max, -1, -1);
}

void Primitive::init(PrimType primitives[], bool caseIgnored)
{
	ThreadP th=Thread::This();
	th->caseIgnored=caseIgnored;

	for (int i=0;  primitives[i].name;  i++)
	{
		PrimType &prim=primitives[i];
		PrimitiveP proc=new Primitive(prim);
		if throwing return;

		StringP name;
		if (caseIgnored)
			name=new String(prim.name, 0, 
				strlen(prim.name), STRING, String::copy_low);
		else
			name=new String(prim.name, 0, 
				strlen(prim.name), STRING, String::copy);
		if throwing return;

		ObjectP obj=root->find(name);
		if (!obj)	// don't insert duplicate primitive names
			obj=root->insert(name, THIS_FILE, __LINE__);
//		else cout << name << " is double!" << endL;
		
		obj->setprim(proc);

		if (proc->type == MACRO)
			obj->bit.proc_macro=true;

		obj->bit.permanent=true;
	}
}
