#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>
#include <stdio.h>

#include <wx/datstrm.h>
#include <wx/thread.h>
#include <wx/list.h>

#include <m_apm.h>

#include <complex>

typedef std::complex<double> complexd;	
typedef complexd* complexdP;

#include "Main.h"
#include "Thread.h"
#include "File.h"
#include "String.h"
#include "Object.h"
#include "Pair.h"
#include "Line.h"
#include "Quote.h"
#include "Int.h"
#include "Error.h"
#include "NodePStack.h"
#include "Stack.h"
#include "Init.h"

//#define _DEBUG_FNS 1

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

void printIndents(FILEP strm, int &width)
{
	ThreadP th=Thread::This();
	if (width < th->indent)
		print(strm, width, "...");
	else
		for (int i=0;  i < th->indent;  i++)
			print(strm, width, '\t');
}

void print(FILEP strm, int &width, unsigned char i)
{
	ThreadP th=Thread::This();
	char ch, format[5], ca[15];
	charP cp;
	if (th->digits)
	{	sprintf(format, "%%%ii", th->digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%i", i);
	cp=ca;
	if (width < int(strlen(cp)))
		print(strm, width, "...");
	else
		while ((ch=*cp++) !='\0')
			print(strm, width, ch);
}

void print(FILEP strm, int &width, int16 i)
{
	ThreadP th=Thread::This();
	char ch, format[5], ca[15];
	charP cp;
	if (th->digits)
	{	sprintf(format, "%%%ii", th->digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%i", i);
	cp=ca;
	if (width < int(strlen(cp)))
		print(strm, width, "...");
	else
		while ((ch=*cp++) !='\0')
			print(strm, width, ch);
}

void print(FILEP strm, int &width, int i)
{
	ThreadP th=Thread::This();
	char ch, format[5], ca[15];
	charP cp;
	if (th->digits)
	{	sprintf(format, "%%%ii", th->digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%i", i);
	cp=ca;
	if (width < int(strlen(cp)))
		print(strm, width, "...");
	else
		while ((ch=*cp++) !='\0')
			print(strm, width, ch);
}

void printBin(FILEP strm, char i)
{
	ThreadP th=Thread::This();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen--;
		if (th->stringlen >= 0)     // printing to string
			*th->stringptr++ = i;
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print int8 to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	int width=Int::max;
		print(strm, width, i);
		return;
	}
	fwrite(&i, sizeof(char), 1, strm);
}

void printBin(FILEP strm, unsigned char i)
{
	ThreadP th=Thread::This();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen--;
		if (th->stringlen >= 0)     // printing to string
			*th->stringptr++ = char(i);
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print uint8 to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	int width=Int::max;
		print(strm, width, char(i));
		return;
	}
	fwrite(&i, sizeof(char), 1, strm);
}

void printBin(FILEP strm, int16 i)
{
	ThreadP th=Thread::This();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen-=sizeof(int16);
		if (th->stringlen >= 0)     // printing to string
		{	*int16P(th->stringptr) = i;
			th->stringptr+=sizeof(int16);
		}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print int16 to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(&i);
		int width=Int::max;
		for (int l=sizeof(int16); l; l--)
			print(strm, width, *cp++);
		return;
	}
	fwrite(&i, sizeof(int16), 1, strm);
}

void printBin(FILEP strm, int i)
{
	ThreadP th=Thread::This();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen-=sizeof(int);
		if (th->stringlen >= 0)     // printing to string
		{	*intP(th->stringptr) = i;
			th->stringptr+=sizeof(int);
		}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print int to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(&i);
		int width=Int::max;
		for (int l=sizeof(int); l; l--)
			print(strm, width, *cp++);
		return;
	}
	fwrite(&i, sizeof(int), 1, strm);
}

void printHex(FILEP strm, int &width, int i)
{
	ThreadP th=Thread::This();
	char ch, format[5], ca[15];
	charP cp;
	if (th->digits)
	{	sprintf(format, "%%%ix", th->digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%x", i);
	cp=ca;
	while ((ch=*cp++) !='\0')
		print(strm, width, ch);
}

void print(FILEP strm, int &width, double d)
{
	ThreadP th=Thread::This();
	char ch, format[10], ca[200];
	charP cp;
	if (th->digits)
	{	sprintf(format, "%%%i.%ig", th->digits, th->precision);
		sprintf(ca, format, d);
	}else
	{	sprintf(format, "%%-.%ig", th->precision);
		sprintf(ca, format, d);
	}
	cp=ca;
	if (width < int(strlen(cp)))
		print(strm, width, "...");
	else
		while ((ch=*cp++) !='\0')
			print(strm, width, ch);
}

void printBin(FILEP strm, double d)
{
	ThreadP th=Thread::This();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen-=sizeof(double);
		if (th->stringlen >= 0)     // printing to string
		{	*doubleP(th->stringptr) = d;
			th->stringptr+=sizeof(double);
		}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print float to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(&d);
		int width=Int::max;
		for (int l=sizeof(double); l; l--)
			print(strm, width, *cp++);
		return;
	}
	fwrite(&d, sizeof(double), 1, strm);
}

void print(FILEP strm, int &width, complexd c)
{
	ThreadP th=Thread::This();
	char ch, format[20], ca[400];
	charP cp;
	if (th->digits)
	{	if (c.imag() >= 0)
			sprintf(format, "%%%i.%ig+%%%i.%igi", 
				th->digits, th->precision, th->digits, th->precision);
		else
			sprintf(format, "%%%i.%ig%%%i.%igi", 
				th->digits, th->precision, th->digits, th->precision);
		sprintf(ca, format, c.real(), c.imag());
	}else
	{	if (c.imag() >= 0)
			sprintf(format, "%%-.%ig+%%-.%igi", th->precision, th->precision);
		else
			sprintf(format, "%%-.%ig%%-.%igi", th->precision, th->precision);
		sprintf(ca, format, c.real(), c.imag());
	}
	cp=ca;
	if (width < int(strlen(cp)))
		print(strm, width, "...");
	else
		while ((ch=*cp++) !='\0')
			print(strm, width, ch);
}

void printBin(FILEP strm, complexd c)
{
	ThreadP th=Thread::This();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen-=sizeof(complexd);
		if (th->stringlen >= 0)     // printing to string
		{	*complexdP(th->stringptr) = c;
			th->stringptr+=sizeof(complexd);
		}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print complex to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(&c);
		int width=Int::max;
		for (int l=sizeof(complexd); l; l--)
			print(strm, width, *cp++);
		return;
	}
	fwrite(&c, sizeof(complexd), 1, strm);
}

void File::print(FILEP strm, NodeP nd, int noBrackets) 
{
	ThreadP th=Thread::This();
	int width=th->printWidthLimit,
		 depth=th->printDepthLimit;
	if (nd)
		nd->print(strm, width, depth,  noBrackets);
	else if (noBrackets <= 0)
	{	::print(strm, width, '[');
		::print(strm, width, ']');
	}
}

void print(FILEP strm, PairPP arg) 
{
	ThreadP th=Thread::This();
	int width=th->printWidthLimit,
		 depth=th->printDepthLimit;

	if (NodePP(arg)!=th->argStack->index) 
		if (*++arg)
			(*arg)->print(strm, width, depth);
		else
		{	print(strm, width, '[');
			print(strm, width, ']');
		}

	while (NodePP(arg)!=th->argStack->index) 
	{	print(strm, width, ' ');
		if (*++arg)
			(*arg)->print(strm, width, depth);
		else
		{	print(strm, width, '[');
			print(strm, width, ']');
		}
	}
}

void File::printStack(PairPP arg, bool space) 
{
	ThreadP th=Thread::This();
	if (NodePP(arg)!=th->argStack->index) 
		if ((*++arg)->islist())
			print(writer.file(), *arg, 1);
		else
			print(writer.file(), *arg, 0);

	while (NodePP(arg)!=th->argStack->index) 
	{	if (space) 
			writer << ' ';

		if ((*++arg)->islist())
			print(writer.file(), *arg, 1);
		else
			print(writer.file(), *arg, 0);
	}
}

void File::printStackBin(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (NodePP(arg)!=th->argStack->index) 
		if (*++arg)
			(*arg)->printBin(writer.file());
	
	while (NodePP(arg)!=th->argStack->index) 
		if (*++arg)
			(*arg)->printBin(writer.file());
}

void __cdecl print(FILEP strm, int &width, const char* fmt, ...)	//VARARGS
{
	ThreadP th=Thread::This();
	va_list ap;
	va_start(ap, fmt);
	char ch;
	while ((ch=*fmt++) !='\0') 
	{	if (ch=='%') 
		{	th->digits=0;
next_format_ch:
			ch=*fmt++;
			switch(ch)
			{
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':th->digits=th->digits*10+ch-'0';
						goto next_format_ch;

			case 'p':{	NodeP nd=va_arg(ap, NodeP);	// print
							if (nd->islist())
								if (nd)
									nd->print(strm, width, -1, 1);
								else
								{	print(strm, width, '[');
									print(strm, width, ']');
								}
							else
								nd->print(strm, width,-1);
							break;
						}
			case 's':{	NodeP nd=va_arg(ap, NodeP);	// show
							File::print(strm, nd);
							break;
						}
			case 't':{	charP cp=va_arg(ap, charP);	// text
							while ((ch=*cp++) !='\0')
								print(strm, width, ch);
							break;
						}
			case 'i':print(strm, width, va_arg(ap, int));      break;  // int
			case 'f':print(strm, width, va_arg(ap, double));   break;  // double
#ifdef WIN32
			case 'c':print(strm, width, va_arg(ap, complexd)); break;  // complexd
#endif
			case 'x':printHex(strm, width, va_arg(ap, int));  break;  // int as hex
			default:	print(strm, width, '%'); print(strm, width, ch);
			}
		}else 
			print(strm, width, ch);
	}
	if (!strm) print(strm, width, '\0');
	va_end(ap);
}

File &File::operator<<(const   char c) 
{
	int width=Int::max;
	::print(file(), width, c); 
	return *this;
}
File &File::operator<<(const    int c) 
{
	int width=Int::max;
	::print(file(), width, c); 
	return *this;
}
File &File::operator<<(const double c) 
{
	int width=Int::max;
	::print(file(), width, c); 
	return *this;
}
File &File::operator<<(const  NodeP c) 
{
	int width=Int::max;
	if (c)
		c->print(file(), width, 0, 1); 
	return *this;
}
File &File::operator<<(const  LineP c) 
{
	int width=Int::max;
	if (c)
		c->print(file(), width, 0, 1); 
	return *this;
}
File &File::operator<<(const QuoteP c) 
{
	int width=Int::max;
	c->print(file(), width, 0); 
	return *this;
}
File &File::operator<<(const PairPP c) 
{
//	int width=Int::max;
	::print(file(), c); 
	return *this;
}
File &File::operator<<(const char *c) 
{
	int width=Int::max;
	while (width > 0 && *c)
		::print(file(), width, *c++); 
	return *this;
}
File &File::operator<<(const VarP var) 
{
	int width=Int::max;
	var->object->print(file(), width, 0); 
	var->val->print(file(), width, 0); 
	return *this;
}

NodeP lprint(PairPP arg) 
{
	int width=Int::max;
	File::printStack(arg, true);
	print(writer.file(), width, '\n');
	return Unbound;
}

NodeP ltype(PairPP arg) 
{
	File::printStack(arg, false);
	return Unbound;
}

NodeP ltypeBin(PairPP arg) 
{
	File::printStackBin(arg);
	return Unbound;
}

NodeP lshow(PairPP arg) 
{
	ThreadP th=Thread::This();
	int width=Int::max;
	print(th->Writer->file(), arg);
	print(th->Writer->file(), width, '\n');
	return Unbound;
}

NodeP lPrintDepthLimit(PairPP arg)
{
	ThreadP th=Thread::This();
	return new Int(th->printDepthLimit);
}

NodeP lsetPrintDepthLimit(PairPP arg)
{
	ThreadP th=Thread::This();
	int printDepthLimit=arg[1]->a_int(1);
	if throwing return Unbound;
	th->printDepthLimit=printDepthLimit;
	return Unbound;
}

NodeP lPrintWidthLimit(PairPP arg)
{
	ThreadP th=Thread::This();
	return new Int(th->printWidthLimit);
}

NodeP lsetPrintWidthLimit(PairPP arg)
{
	ThreadP th=Thread::This();
	int printWidthLimit=arg[1]->a_int(1);
	if throwing return Unbound;
	th->printWidthLimit=printWidthLimit;
	return Unbound;
}
