#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "File.h"
#include "IntToPtr_.h"
#include "PtrToInt_.h"
#include "StringToPtr.h"
#include "PtrToString.h"
#include "String.h"
#include "Pair.h"
#include "Procedure.h"
#include "IntNodeTable.h"
#include "Compiled.h"
#include "Object.h"
#include "Line.h"
#include "Colon.h"
#include "Tree.h"
#include "Error.h"
#include "Init.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

Procedure::Procedure(PairP formals_, PairP bodylist_, PairP bodywords_, 
	ObjectP object_, Priority priority_, char min_, char std_, char max_, NodeType type_):
	Node(type_), _object(object_), _compiled(0), min(min_), std(std_), max(max_) 
{
	setformals(formals_);  setbodylist(bodylist_);  setbodywords(bodywords_);
	setpriority(priority_);
}

void Procedure::setcompiled(CompiledP nd) 
{	
	_compiled=nd;  
	check_oldyoung(nd);
}

void Procedure::inc_gc() 
{	
	fifo << formals() << bodylist() << bodywords() << compiled();
}

int Procedure::youngs() 
{
	return gmark(formals())
			+gmark(bodylist())
			+gmark(bodywords())
			+gmark(compiled());
}

int Procedure::youngsPtr(PtrToInt_ &ptrToInt) 
{
	return gmarkPtr(formals(), ptrToInt)
			+gmarkPtr(bodylist(), ptrToInt)
			+gmarkPtr(bodywords(), ptrToInt)
			+gmarkPtr(compiled(), ptrToInt);
}

void Procedure::check_oldyoungs() 
{
	check_oldyoung(formals()); 
	check_oldyoung(bodylist()); 
	check_oldyoung(bodywords());
	check_oldyoung(compiled());
}

void Procedure::setPtrs(IntToPtr_ &intToPtr)
{	setformals  (PairP((Pair*)intToPtr.find(int(_formals  ))));
	setbodylist (PairP((Pair*)intToPtr.find(int(_bodylist ))));
	setbodywords(PairP((Pair*)intToPtr.find(int(_bodywords))));
	setcompiled (CompiledP((Compiled*)intToPtr.find(int(_compiled))));
	_object=(ObjectP((Object*)intToPtr.find(int(_object   ))));
}
DataInputStream &Procedure::read(DataInputStream &istr) 
{	return istr 
	>> (int&)_formals >> (int&)_bodylist >> (int&)_bodywords 
	>> (int&)_compiled >> (int&)_object
	>> (unsigned char&)min 
	>> (unsigned char&)std 
	>> (unsigned char&)max;
}
wxDataOutputStream &Procedure::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(formals()) 
		<< GC_::ptrToInt->find(bodylist())
		<< GC_::ptrToInt->find(bodywords())
		<< GC_::ptrToInt->find(compiled())
		<< GC_::ptrToInt->find(object()) 
		<< (unsigned char)min 
		<< (unsigned char)std 
		<< (unsigned char)max;
}

NodeP Procedure::copy(PtrToInt_ &ptrToInt)
{	ProcedureP proc=new Procedure();
	if throwing return Unbound;
	proc->_formals=PairP(ptrToInt.find(formals()));
	proc->_bodylist=PairP(ptrToInt.find(bodylist()));
	proc->_bodywords=PairP(ptrToInt.find(bodywords()));
	proc->_compiled=CompiledP(ptrToInt.find(compiled()));
	proc->_object=ObjectP(ptrToInt.find(object()));
	proc->min=min;
	proc->std=std;
	proc->max=max;
	return proc;
}

void Procedure::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	if (bodywords())
		bodywords()->print(strm, width, depth, noBrackets, noSpacers);
}

void Procedure::printBin(FILEP strm)
{
	if (bodywords())
		bodywords()->printBin(strm);
}

NodeP Procedure::contentsTree(	wxArrayString &namearr, 
											wxArrayString &valarr, 
											wxArrayPtrVoid &objarr, 
											ObjectType want, 
											ObjectType dont, 
											ObjectType must, 
											int &count, int level)
{
	PairP nd=bodywords();
	objarr[count]=nd;
	NodeP contentsList=0;
	namearr[count]=wxString(' ',level-1);
	if (nd)
	{	if (nd->expandable())
			if (nd->expanded())
				namearr[count]+='-';
			else
				namearr[count]+='+';
		if (nd->expanded())
		{	valarr[count]=nd->_typeof()->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
			contentsList=nd->contentsTree(
				namearr, valarr, objarr, want, dont, must, count, level+1);
		}else
		{	valarr[count]=_typeof()->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
		}
	}
	if (!contentsList || contentsList==Unbound)
		return 0;

	NodeP cmpld=Unbound;
	
	if (compiled())
//		if (bit.table_expand)
		{	cmpld=compiled()->contentsTree(
				namearr, valarr, objarr, want, dont, must, count, level+1);
		}
	return this;
}

NodeP Procedure::compare(NodeP nd)
{	return bodywords()->compare(nd);
}
int Procedure::count()
{	return bodywords()->count();
}
int Procedure::_sizeof()
{	return bodywords()->_sizeof();
}
bool Procedure::equalP(NodeP b)
{	return bodywords()->equalP(b);
}
NodeP Procedure::first_()
{	return bodywords()->first_();
}
NodeP Procedure::butFirst()
{	return bodywords()->butFirst();
}
NodeP Procedure::last()
{	return bodywords()->last();
}
NodeP Procedure::butLast()
{	return bodywords()->butLast();
}
PairP Procedure::list()
{	return bodywords();
}

/*
void Procedure::untreeify() 
{
	TreeP	body = TreeP(bodylist());

	for (TreeP body_ptr=body;  body_ptr;  body_ptr=TreeP(body_ptr->tail())) 
		TreeP(body_ptr->first())->untreeify();
	
	body->untreeify();
}
*/

PairP Procedure::get_bodywords(NodeP name)
{
	PairP
		val = bodywords(),
		head=0, tnode=0;
	
	if (val) return val;

	ObjectP obj = name->intern();

	ObjectP symbol=
		obj->bit.proc_dyn ?
			(obj->bit.proc_macro ? __Macro_ : _To)
			:
			_Be;

	head = 
		new Pair(symbol,
		new Pair(obj->name(), 0));
	if throwing return PairP(Unbound);
	tnode=head->tail();
	val=formals();

	while (val) 
	{
		if (val->first()->is.list)
			tnode->settail
			(	new Pair
				(	new Pair
					(	new Colon(val->first()->first()), 
						val->first()->tail()
					), 
					0
				)
			);
		else 
			if ( val->first()->type == INT_)
				tnode->settail( new Pair(val->first(), 0) );
			else
				tnode->settail
				(	new Pair
					(	new Colon(val->first()),
						0
					)
				);
		if throwing return PairP(Unbound);
		next(tnode);
		next(val);
	}
	tnode = head = new Pair(head, 0);
	
	for (val=bodylist();  val;  next(val))
	{	tnode->settail( new Pair( val->first()->runparse(), 0 ) );
		if throwing return PairP(Unbound);
		next(tnode);
	}
	tnode->settail
	(	new Pair
		(	new Pair(_End, 0), 
			0
		)
	);
	return head;
}
