//___________________________________________________________________
//
//                  PtrDebug Smart Pointer Classes
//___________________________________________________________________


#include <stdlib.h>

template <class T> class Ptr_
{	public:
		T *p;

		Ptr_() {p= (T*) 0xBAD00BAD;}
		Ptr_(T *p_);

		  operator T*() {return p; }
		T &operator*() {return *p; }
		T *operator->() {return p; }
		T &operator[](const int i);
		Ptr_ *operator&();

		Ptr_ &operator=(T *p_);
		Ptr_ &operator--();
		Ptr_ &operator++(int);
		Ptr_ &operator--(int);

		operator bool();
		operator!();
		bool operator==(const Ptr_ &np) {return p==np.p;}
		bool operator!=(const Ptr_ &np) {return p!=np.p;}

		bool operator&&(const Ptr_ &np) {return p? (np.p? true : false) : false;}
		bool operator||(const Ptr_ &np) {return p? true : (np.p? true : false);}
		bool operator&&(const bool &b) {return p && b;}
		bool operator||(const bool &b) {return p || b;}
		bool operator&&(const int &b) {return p && b;}
		bool operator||(const int &b) {return p || b;}
		bool operator&&(const unsigned &b) {return p && b;}
		bool operator||(const unsigned &b) {return p || b;}

		friend bool operator&&(const bool b, const Ptr_ &np) {return b && np.p;}
		friend bool operator||(const bool b, const Ptr_ &np) {return b || np.p;}
		friend bool operator&&(const int b, const Ptr_ &np) {return b && np.p;}
		friend bool operator||(const int b, const Ptr_ &np) {return b || np.p;}
		friend bool operator&&(const unsigned b, const Ptr_ &np) {return b && np.p;}
		friend bool operator||(const unsigned b, const Ptr_ &np) {return b || np.p;}
};
class Node;	typedef Ptr_<Node> NodeP;

template <class T> class NPtr:  public NodeP 
{	public:
		NPtr() {p= (Node*) 0xBAD01BAD;}
		NPtr(const NodeP p_);
		  operator T*() {return (T*)p; }
		T& operator*() {return *(T*)p; }
		T* operator->() {return (T*)p; }
		T &operator[](const int i);
		NPtr& operator=(T *p_);

		bool operator&&(const NodeP &np) {return p? (np.p? true : false) : false;}
		bool operator||(const NodeP &np) {return p? true : (np.p? true : false);}
		bool operator&&(const bool &b) {return p && b;}
		bool operator||(const bool &b) {return p || b;}
		bool operator&&(const int &b) {return p && b;}
		bool operator||(const int &b) {return p || b;}
		bool operator&&(const unsigned &b) {return p && b;}
		bool operator||(const unsigned &b) {return p || b;}
};
class Pair;	typedef NPtr<Pair> PairP;

template <class T> class PairPtr: public PairP 
{	public:
		PairPtr() {p= (Node*) 0xBAD02BAD;}
		PairPtr(const PairP p_);
		PairPtr(const T* p_);

		operator T*() {return (T*)p; }
		T& operator*() {return *(T*)p; }
		T* operator->() {return (T*)p; }
		PairPtr& operator=(T *p_);

		bool operator&&(const PairP &np) {return p? (np.p? true : false) : false;}
		bool operator||(const PairP &np) {return p? true : (np.p? true : false);}
		bool operator&&(const bool &b) {return p && b;}
		bool operator||(const bool &b) {return p || b;}
		bool operator&&(const int &b) {return p && b;}
		bool operator||(const int &b) {return p || b;}
		bool operator&&(const unsigned &b) {return p && b;}
		bool operator||(const unsigned &b) {return p || b;}
};
class Line;	typedef PairPtr<Line> LineP;	

template <class T> class LinePtr: public LineP 
{	public:
		LinePtr() {p= (Node*) 0xBAD03BAD;}
		LinePtr(const LineP p_);
		LinePtr(const T* p_);

		operator T*() {return (T*)p; }
		T& operator*() {return *(T*)p; }
		T* operator->() {return (T*)p; }
		LinePtr& operator=(T *p_);

		bool operator&&(const PairP &np) {return p? (np.p? true : false) : false;}
		bool operator||(const PairP &np) {return p? true : (np.p? true : false);}
		bool operator&&(const bool &b) {return p && b;}
		bool operator||(const bool &b) {return p || b;}
		bool operator&&(const int &b) {return p && b;}
		bool operator||(const int &b) {return p || b;}
		bool operator&&(const unsigned &b) {return p && b;}
		bool operator||(const unsigned &b) {return p || b;}
};

template <class T> class PtrPtr
{	public:
		T *p;

		PtrPtr() {p= (T*) 0xBAD03BAD;}
		PtrPtr(T *p_);

		  operator T*() {return p; }
		T &operator*() {return *p; }
		T *operator->() {return p; }
		T &operator[](const int i);
		T operator[](const int i) const;
		PtrPtr &operator=(T *p_);
		PtrPtr &operator++();
		PtrPtr &operator--();
		PtrPtr &operator++(int);
		PtrPtr &operator--(int);
		PtrPtr &operator+(const int &i);
		PtrPtr &operator-(const int &i);

		operator bool();
		operator!();
		bool operator==(const T *np) {return p==np;}
		bool operator==(const PtrPtr &np) {return p==np.p;}
		bool operator!=(const PtrPtr &np) {return p!=np.p;}
		bool operator!=(const NodeP *np) {return p!=np;}

		bool operator&&(const PtrPtr &np) {return p? (np.p? true : false) : false;}
		bool operator||(const PtrPtr &np) {return p? true : (np.p? true : false);}
		bool operator&&(const bool &b) {return p && b;}
		bool operator||(const bool &b) {return p || b;}
		bool operator&&(const int &b) {return p && b;}
		bool operator||(const int &b) {return p || b;}
		bool operator&&(const unsigned &b) {return p && b;}
		bool operator||(const unsigned &b) {return p || b;}

		friend bool operator&&(const bool b, const PtrPtr &np) {return b && np.p;}
		friend bool operator||(const bool b, const PtrPtr &np) {return b || np.p;}
		friend bool operator&&(const int b, const PtrPtr &np) {return b && np.p;}
		friend bool operator||(const int b, const PtrPtr &np) {return b || np.p;}
		friend bool operator&&(const unsigned b, const PtrPtr &np) {return b && np.p;}
		friend bool operator||(const unsigned b, const PtrPtr &np) {return b || np.p;}
};
typedef PtrPtr<NodeP> NodePP;

template <class T> class NPPtr:  public NodePP
{	public:
		NPPtr() {p= (NodeP*) 0xBAD04BAD;}
		NPPtr(NodeP *p_);
		  operator T*() {return (T*)p; }
		T &operator*() {return *(T*)p; }
		T *operator->() {return (T*)p; }
		T &operator[](const int i);
		T operator[](const int i) const;
		NPPtr &operator++();
		NPPtr &operator++(int);
		NPPtr &operator=(T *p_);
		NPPtr &operator+(const int &i);
		NPPtr &operator-(const int &i);

		bool operator&&(const NodePP &np) {return p? (np.p? true : false) : false;}
		bool operator||(const NodePP &np) {return p? true : (np.p? true : false);}
		bool operator&&(const bool &b) {return p && b;}
		bool operator||(const bool &b) {return p || b;}
		bool operator&&(const int &b) {return p && b;}
		bool operator||(const int &b) {return p || b;}
		bool operator&&(const unsigned &b) {return p && b;}
		bool operator||(const unsigned &b) {return p || b;}
};

	class Int;			typedef NPtr<Int> IntP;
	class Int16;		typedef NPtr<Int16> Int16P;
	class Int8;			typedef NPtr<Int8> Int8P;
	class UInt8;		typedef NPtr<UInt8> UInt8P;
	class Float;		typedef NPtr<Float> FloatP;
	class BigFloat;	typedef NPtr<BigFloat> BigFloatP;
	class Complexd;	typedef NPtr<Complexd> ComplexdP;
	class Primitive;	typedef NPtr<Primitive> PrimitiveP;
	class Array;		typedef NPtr<Array> ArrayP;
	class IntArray;	typedef NPtr<IntArray> IntArrayP;
	class Int16Array;	typedef NPtr<Int16Array> Int16ArrayP;
	class FloatArray;	typedef NPtr<FloatArray> FloatArrayP;
	class String;		typedef NPtr<String> StringP;
	class Bitmap;		typedef NPtr<Bitmap> BitmapP;
	class Texture;		typedef NPtr<Texture> TextureP;
	class TurtleNode;	typedef NPtr<TurtleNode> TurtleNodeP;
	class GraphicNode;typedef NPtr<GraphicNode> GraphicNodeP;

	typedef NPPtr<PairP> PairPP;
		class List;			typedef PairPtr<List> ListP;
		class WindowPair;	typedef PairPtr<WindowPair> WindowPairP;
		typedef NPPtr<LineP> LinePP;
			class Quote;		typedef LinePtr<Quote> QuoteP;
			class Colon;		typedef LinePtr<Colon> ColonP;
			class Runparsed;	typedef LinePtr<Runparsed> RunparsedP;
			class Tree;			typedef LinePtr<Tree> TreeP;
	class Object;		typedef NPtr<Object> ObjectP;		typedef NPPtr<ObjectP> ObjectPP;
	class LogoObject;	typedef NPtr<LogoObject> LogoObjectP;	typedef NPPtr<LogoObjectP> LogoObjectPP;
	class Procedure;	typedef NPtr<Procedure> ProcedureP;
	class File;			typedef NPtr<File> FileP;
	class ToDoNode;	typedef NPtr<ToDoNode> ToDoNodeP;
	class Table;		typedef NPtr<Table> TableP;
	class Struct;		typedef NPtr<Struct> StructP;
	class DynamicLibrary;typedef NPtr<DynamicLibrary> DynamicLibraryP;
	class SliderNode;			typedef NPtr<SliderNode> SliderNodeP;
	class GaugeNode;			typedef NPtr<GaugeNode> GaugeNodeP;
	class IntControlNode;	typedef NPtr<IntControlNode> IntControlNodeP;
	class FloatControlNode;	typedef NPtr<FloatControlNode> FloatControlNodeP;
	class FrameNode;			typedef NPtr<FrameNode> FrameNodeP;
	class GraphNode;			typedef NPtr<GraphNode> GraphNodeP;
	class ButtonNode;			typedef NPtr<ButtonNode> ButtonNodeP;
	class ToggleButtonNode;	typedef NPtr<ToggleButtonNode> ToggleButtonNodeP;
	class CheckBoxNode;		typedef NPtr<CheckBoxNode> CheckBoxNodeP;
	class RadioButtonNode;	typedef NPtr<RadioButtonNode> RadioButtonNodeP;
	class StaticTextNode;	typedef NPtr<StaticTextNode> StaticTextNodeP;
	class TextControlNode;	typedef NPtr<TextControlNode> TextControlNodeP;
	class ListBoxNode;		typedef NPtr<ListBoxNode> ListBoxNodeP;
	class ChoiceBoxNode;		typedef NPtr<ChoiceBoxNode> ChoiceBoxNodeP;
	class ComboBoxNode;		typedef NPtr<ComboBoxNode> ComboBoxNodeP;
	class ListControlNode;	typedef NPtr<ListControlNode> ListControlNodeP;
	class BoxSizerNode;		typedef NPtr<BoxSizerNode> BoxSizerNodeP;


extern bool is_nil_or_ok(volatile Node *nd);

template <class T> Ptr_<T>::Ptr_<T>(T *p_)
{	if (is_nil_or_ok(p_)) p = p_; 
	else
		exit(0xBAD0010);
}
template <class T> PairPtr<T>::PairPtr<T>(const T *p_)
{	if (is_nil_or_ok((volatile Node*)p_) && (!p_ || p_->is.list || p_->is.line)) 
		p = (Node*)p_;
	else
		exit(0xBAD0020);
}
template <class T> LinePtr<T>::LinePtr<T>(const T *p_)
{	if (is_nil_or_ok((volatile Node*)p_) && (!p_ || p_->is.line)) 
		p = (Node*)p_;
	else
		exit(0xBAD0030);
}
template <class T> PairPtr<T>::PairPtr<T>(const PairP p_)
{	if (is_nil_or_ok(p_.p)) p = p_.p; 
	else
		exit(0xBAD0040);
}
template <class T> LinePtr<T>::LinePtr<T>(const LineP p_)
{	if (is_nil_or_ok(p_.p)) p = p_.p; 
	else
		exit(0xBAD0048);
}
inline NodePP::PtrPtr<NodeP>(NodeP *p_)
{	if (!p_ || is_nil_or_ok(*p_)) p = p_; 
	else
		exit(0xBAD0050);
}
template <class T> NPPtr<T>::NPPtr<T>(NodeP *p_)
{	if (!p_ || is_nil_or_ok(*p_)) p = p_; 
	else
		exit(0xBAD0060);
}

template <class T> Ptr_<T> &Ptr_<T>::operator=(T *p_)
{	if (!p_ || is_nil_or_ok(p_)) p = p_; 
	else
		exit(0xBAD0070);
	return *this;
}
template <class T> NPtr<T> &NPtr<T>::operator=(T *p_)
{	if (!p_ || is_nil_or_ok(p_)) p = p_; 
	else
		exit(0xBAD0071);
	return *this;
}
template <class T> PairPtr<T> &PairPtr<T>::operator=(T *p_)
{	if (is_nil_or_ok((Node*)p_)) p = (Node*)p_; 
	else
		exit(0xBAD0072);
	return *this;
}
template <class T> LinePtr<T> &LinePtr<T>::operator=(T *p_)
{	if (is_nil_or_ok((Node*)p_)) p = (Node*)p_; 
	else
		exit(0xBAD0073);
	return *this;
}
inline NodePP &NodePP::operator=(NodeP *p_)
{	if (!p_ || is_nil_or_ok(*p_)) p = p_; 
	else
		exit(0xBAD0074);
	return *this;
}
inline PairPP &PairPP::operator=(PairP *p_)
{	if (!p_ || is_nil_or_ok(*p_)) p = p_; 
	else
		exit(0xBAD0075);
	return *this;
}
template <class T> Ptr_<T>::operator bool() 
{	if (p==0) return false;
	if (is_nil_or_ok(p)) return true;
	exit(0xBAD0080);	return false;
}
template <class T> PtrPtr<T>::operator bool() 
{	if (p==0) return false;
	if (is_nil_or_ok(*p)) return true;
	exit(0xBAD0081);	return false;
}
template <class T> Ptr_<T>::operator!() 
{	if (p==0) return true;
	if (is_nil_or_ok(p)) return false;
	exit(0xBAD0090);	return true;
}
template <class T> PtrPtr<T>::operator!() 
{	if (p==0) return true;
	if (is_nil_or_ok(*p)) return false;
	exit(0xBAD0091);	return true;
}

inline NodeP &NodePP::operator[](int i) 
{	NodeP &nd = p[i];
	if (is_nil_or_ok(nd)) return nd;
	exit(0xBAD00A0);	return NIL;
}
inline PairP &PairPP::operator[](int i) 
{	PairP &nd = (PairP&)p[i];
	if (is_nil_or_ok(nd)) return nd;
	exit(0xBAD00A1);	return NIL;
}
inline NodeP NodePP::operator[](int i) const
{	NodeP &nd = p[i];
	if (is_nil_or_ok(nd)) return nd;
	exit(0xBAD00A0);	return NIL;
}
inline PairP PairPP::operator[](int i) const
{	PairP &nd = (PairP&)p[i];
	if (is_nil_or_ok(nd)) return nd;
	exit(0xBAD00A1);	return NIL;
}
inline NodeP *NodeP::operator&() 
{	NodeP *np = this;
	if (is_nil_or_ok(*np)) return np;
	exit(0xBAD00B0);	return 0;
}
inline NodePP &NodePP::operator++() 
{	++p;
	if (!p || is_nil_or_ok(*p)) return *this;
	exit(0xBAD00C0);	return NodePP(0);
}
inline NodePP &NodePP::operator--() 
{	--p;
	if (!p || is_nil_or_ok(*p)) return *this;
	exit(0xBAD00C1);	return NodePP(0);
}
inline PairPP &PairPP::operator++() 
{	++p;
	if (!p || is_nil_or_ok(*p)) return *this;
	exit(0xBAD00D0);	return PairPP(0);
}
inline NodePP &NodePP::operator++(int) 
{	NodePP np=p;
	p++;
	if (!p || is_nil_or_ok(*p)) return np;
	exit(0xBAD00E0);	return NodePP(0);
}
inline PairPP &PairPP::operator++(int) 
{	PairPP np=PairPP(p);
	p++;
	if (!p || is_nil_or_ok(*p)) return np;
	exit(0xBAD00E1);	return PairPP(0);
}
inline ObjectPP &ObjectPP::operator++(int) 
{	ObjectPP np=ObjectPP(p);
	p++;
	if (!p || is_nil_or_ok(*p)) return np;
	exit(0xBAD00E2);	return ObjectPP(0);
}
inline NodePP &NodePP::operator--(int) 
{	NodePP np=p;
	p--;
	if (!p || is_nil_or_ok(*p)) return np;
	exit(0xBAD00F0);	return NodePP(0);
}
inline NodePP &NodePP::operator+(const int &i) 
{	NodePP np=NodePP(p+i);
	if (!np || is_nil_or_ok(*np)) return np;
	exit(0xBAD0100);	return NodePP(0);
}
inline PairPP &PairPP::operator+(const int &i) 
{	PairPP np=PairPP(p+i);
	if (!np || is_nil_or_ok(*np)) return np;
	exit(0xBAD0101);	return PairPP(0);
}
inline NodePP &NodePP::operator-(const int &i) 
{	NodePP np=NodePP(p-i);
	if (!np || is_nil_or_ok(*np)) return np;
	exit(0xBAD0100);	return NodePP(0);
}
inline PairPP &PairPP::operator-(const int &i) 
{	PairPP np=PairPP(p-i);
	if (!np || is_nil_or_ok(*np)) return np;
	exit(0xBAD0111);	return PairPP(0);
}
