#include "stdafx.h"
#include <malloc.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

#include "Main.h"
#include "File.h"
#include "Error.h"
#include "PtrIntTable.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

PtrIntTable::PtrIntTable(int size_):
	size(0)
{
	data = (PtrIntPairP*)(GC_::calloc(size_, sizeof(PtrIntPairP)));
	if (data)
		size=size_;
}

PtrIntTable::~PtrIntTable()
{
	for (int i=0; i < size; i++)
	{	if (data[i])
			delete data[i];
	}
	if (data)
	{	GC_::free(data); data=0;
	}
}

int PtrIntTable::hash(void *p)
{
	unsigned int index = (unsigned int)p;
	if (size)
		index %= size;
	else
		index=0;
	return index;
}

PtrIntTable::PtrIntPairP PtrIntTable::find(void *p)
{
	PtrIntPairP obj;
	for (obj=data[hash(p)];  obj;  obj=obj->tail)
		if (obj->p==p)
			return obj;
	return 0;
}

PtrIntTable::PtrIntPairP PtrIntTable::insert(void *p, int i)
{
	PtrIntPairP &obj=data[hash(p)];
	obj = new PtrIntPair(p, i, obj);
	if (!obj)
		error(OUT_OF_MEM);
	return obj;
}

int PtrIntTable::item(void *index)
{	PtrIntPairP obj=find(index);
	if (!obj)
		return 0;
	return obj->i;
}

void PtrIntTable::setitem(void *index, int newitem)
{	PtrIntPairP obj=find(index);
	if (!obj)
		obj=insert(index, newitem);
	obj->i=newitem;
}

void PtrIntTable::removeitem(void *index)
{	int i=hash(index);
	PtrIntPairP obj=data[i], last=0;
	while (obj)
	{	if (index==obj->p)
			break;
		last=obj;
		obj=obj->tail;
	}
	if (!obj)
	{//	error(BAD_DATA, index);
		return;
	}
	if (!last)
	{	data[i]=obj->tail;
	}
	else
		last->tail=obj->tail;
}

