//___________________________________________________________________
//
//                         Class RadioButton
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/radiobut.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "LogoObject.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Point.h"
#include "NodePStack.h"

#include "Event.h"
#include "RadioButton.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Error.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(RadioButton, wxRadioButton)
	EVT_RADIOBUTTON	(-1, RadioButton::OnClick	)
END_EVENT_TABLE()

RadioButton::RadioButton(RadioButtonNodeP radioButtonNode_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=radioButtonNode_;
	wxWindow *parent=(wxWindow*)console;
	if (radioButtonNode_->parent)
		parent=radioButtonNode_->parent->window;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			_("RadioButton"));
}

RadioButton::RadioButton(RadioButtonNodeP radioButtonNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxRadioButton(parent, id, label, 
			pos, size, style, 
			validator, name)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=radioButtonNode_;
	radioButtonNode_->window=WindowP(this);
	radioButtonNode_->setonClick(onClick_);
}

RadioButton::~RadioButton()
{
	if (*d(windowNode))
		RadioButtonNodeP(*d(windowNode))->window=0;
}

void RadioButton::OnClick(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventRadioButton(this);
	if throwing return;
	*th->eventFiFo << evt;
}

wxString RadioButton::wxstring()
{
	wxString str;
	str << _("RadioButton") << " [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << (*d(onClick))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " 
		<< GetValue();
	return str;
}

DataInputStream &RadioButton::read(DataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style, state;
	
	istr >> label >> name >> (int&)*d(onClick)
		>> pos.x >> pos.y >> size.x >> size.y >> style >> state;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetValue(state!=0);
	return istr;
}
wxDataOutputStream &RadioButton::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(*d(onClick))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle()) << int(GetValue());
	return ostr;
}

PairP RadioButtonNode::onClick()
{
	if (radioButton)
		return PairP(*radioButton->d(RadioButton::onClick));
	return 0;
}

void RadioButtonNode::setonClick(NodeP nd) 
{
	if (window)
	{	ThreadP th=Thread::This();
		*radioButton->d(RadioButton:: onClick) = nd;
		*radioButton->d(RadioButton::uonClick) = th->lexufun;
		*radioButton->d(RadioButton::tonClick) = th->thisobj;
		check_oldyoung(nd);
	}
}

RadioButtonNode::RadioButtonNode(): 
	WindowNode(RADIOBUTTON),
	radioButton((RadioButtonP&)window)
{}

RadioButtonNode::RadioButtonNode(WindowNodeP parent_): 
	WindowNode(RADIOBUTTON, parent_),
	radioButton((RadioButtonP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void RadioButtonNode::free() 
{	
	if (radioButton) 
	{	ThreadP th=Thread::This();
		*radioButton->d(RadioButton::windowNode)=0;
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		radioButton->Destroy(); 
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::tRadioButton:
				if (((EventRadioButton*)ev)->radioButton==radioButton)
					((EventRadioButton*)ev)->radioButton=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		radioButton=0;
	}
}
void RadioButtonNode::inc_gc() {fifo << onClick();}
int RadioButtonNode::youngs() {return gmark(onClick());}
int RadioButtonNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(onClick(), ptrToInt);
}
void RadioButtonNode::check_oldyoungs() {check_oldyoung(onClick());}

DataInputStream &RadioButtonNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr 
			>> (int&)parentNode;
		radioButton=new RadioButton(RadioButtonNodeP(this));
		if (!radioButton)
		{	error(OUT_OF_MEM); return istr;
		}
		radioButton->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &RadioButtonNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (radioButton)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent);
		radioButton->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void RadioButtonNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (radioButton)
	{	setonClick((Pair*)intToPtr.find(int(onClick())));
	}
}

void RadioButtonNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	RadioButtonP r=radioButton, 
		r2=new RadioButton(RadioButtonNodeP(this),
				p, -1, r->GetLabel(),
				PairP(*r->d(RadioButton::onClick)), 
				r->GetPosition(), r->GetSize(), r->GetWindowStyle(), 
				wxDefaultValidator, r->GetName());
	if (!r2)
	{	error(OUT_OF_MEM); return;
	}
	r2->SetValue(r->GetValue());
	radioButton=r2;
	setonClick(*r->d(RadioButton::onClick));
	delete r;
	radioButton=r2;
}

StringP RadioButtonNode::string()
{
	if (radioButton)
		return new String(radioButton->wxstring());
	return _Null_Word->name();
}

wxString RadioButtonNode::wxstring()
{
	if (radioButton)
		return radioButton->wxstring();
	return wxEmptyString;
}

void RadioButtonNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lRadioButton(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindow *parent=Thread::This()->graph;
	WindowNodeP parentNode=0;
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	PairP onClick=0;
	if (exist(3))
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxWindowID id=-1;
	if (exist(7))
	{	int i=arg[7]->a_int(7);
		if throwing return Unbound;
		id=i;	
	}
	wxString name=label;
	if (exist(8))
	{	StringP n=arg[8]->a_string(8);
		if throwing return Unbound;
		name=n->ptr();
	}
//	GUI_ENTER
	RadioButtonNodeP radioButtonNode=new RadioButtonNode(parentNode);
	if throwing return Unbound;
	RadioButton *rb=new RadioButton(radioButtonNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	if (!rb)
	{	error(OUT_OF_MEM); return Unbound;
	}
	radioButtonNode->radioButton=rb;
//	GUI_LEAVE
	return radioButtonNode;
}

NodeP lRadioButtonOnClick(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
//	GUI_ENTER
	radioButton->setonClick(function);
//	GUI_LEAVE
	return Unbound;
}

NodeP lRadioButtonDestroy(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=radioButton->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(radioButton));
			p->check_oldyoung(p->children);
		}
	}
	radioButton->radioButton->Destroy();
//	GUI_LEAVE
	return Unbound;
}

NodeP lRadioButtonSet(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	radioButton->radioButton->SetValue(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lRadioButtonValue(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
//	GUI_ENTER
	bool state=radioButton->radioButton->GetValue();
//	GUI_LEAVE
	return Node::boolnode(state);
}

NodeP lRadioButtonEnable(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	radioButton->radioButton->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

