//___________________________________________________________________
//
//                         Class Recorder
//
//                           Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/ptr_scpd.h>

#ifdef OPENGL
	#include <GL/glu.h>
	#include <gl2ps.h>
#endif

#include <stdio.h>
#include <math.h>

#include "Main.h"
#include "File.h"
#include "PtrToInt_.h"
#include "IntToPtr_.h"
#include "String.h"
#include "Pair.h"
#include "Int.h"
#include "Bitmap.h"
#include "Texture.h"

#include "Error.h"
#include "Init.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "Font.h"
#include "Turtle.h"

#include "Recorder.h"
#include "Graph.h"
#include "LogoApp.h"
#include "Shadow.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

#define DEBUG_RECORDER_SAVE
#define DEBUG_RECORDER_LOAD

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

wxDEFINE_SCOPED_PTR_TYPE(wxZipEntry);

//___________________________________________________________________

wxDataOutputStream &operator<<(wxDataOutputStream &ostr, bool &b)
{
	ostr.Write8((unsigned char)(b));
	return ostr;
}
//___________________________________________________________________

TextInputStream &operator>>(TextInputStream &istr, Point &p)
{
	return istr >> p.x >> p.y >> p.z;
}
DataInputStream &operator>>(DataInputStream &istr, Point &p)
{
	return istr >> p.x >> p.y >> p.z;
}

TextOutputStream &operator<<(TextOutputStream &ostr, Point &p)
{
	return ostr << p.x << ' ' << p.y << ' ' << p.z << ' ';
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Point &p)
{
	return ostr << p.x << p.y << p.z;
}
//___________________________________________________________________

TextInputStream &operator>>(TextInputStream &istr, PointMatrix &m)
{
	return istr >> m.x >> m.y >> m.z;
}
DataInputStream &operator>>(DataInputStream &istr, PointMatrix &m)
{
	return istr >> m.x >> m.y >> m.z;
}
TextOutputStream &operator<<(TextOutputStream &ostr, PointMatrix &m)
{
	return ostr << m.x << m.y << m.z;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, PointMatrix &m)
{
	return ostr << m.x << m.y << m.z;
}
//___________________________________________________________________

wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Color &c)
{
	return ostr << c.i;
}

//___________________________________________________________________

TextInputStream &operator>>(TextInputStream &istr, Font_ &f)
{
	return istr >> f.sizex >> f.sizey 
		>> f.alignx >> f.aligny >> f.weight >> f.name; 
}
DataInputStream &operator>>(DataInputStream &istr, Font_ &f)
{
	return istr >> f.sizex >> f.sizey 
		>> f.alignx >> f.aligny >> f.weight >> f.name; 
}

TextOutputStream &operator<<(TextOutputStream &ostr, Font_ &f)
{
	ostr 
		<< f.sizex << ' ' << f.sizey << ' ' 
		<< f.alignx << ' ' << f.aligny << ' ' 
		<< f.weight << ' ' << f.name << ' ';
	return ostr;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Font_ &f)
{
	ostr 
		<< f.sizex  << f.sizey
		<< f.alignx << f.aligny 
		<< f.weight << f.name;
	return ostr;
}
//___________________________________________________________________

TextInputStream &operator>>(TextInputStream &istr, Pen &p)
{
	istr >> p.down >> p.wide >> p.roundLineEnds
		>> (int&)p.internalMode >> (int&)p.mode
		>> p.size >> p.size_rotated
		>> p.color >> p.rawcolor >> p.floodcolor
		>> p.pattern; 
	return istr;
}
DataInputStream &operator>>(DataInputStream &istr, Pen &p)
{
	istr >> p.down >> p.wide >> p.roundLineEnds
		>> (int&)p.internalMode >> (int&)p.mode
		>> p.size >> p.size_rotated
		>> p.color >> p.rawcolor >> p.floodcolor
		>> p.pattern; 
	return istr;
}

TextOutputStream &operator<<(TextOutputStream &ostr, Pen &p)
{
	ostr << p.down << ' ' << p.wide << ' ' << p.roundLineEnds << ' '
		<< int(p.internalMode) << ' ' << int(p.mode) << ' ' 
		<< p.size << p.size_rotated
		<< p.color << p.rawcolor << p.floodcolor
		<< p.pattern << ' ';
	return ostr;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Pen &p)
{
	ostr << p.down << p.wide << p.roundLineEnds
		<< p.internalMode << p.mode
		<< p.size << p.size_rotated
		<< p.color << p.rawcolor << p.floodcolor
		<< p.pattern; 
	return ostr;
}
//___________________________________________________________________

#ifdef DEBUG_RECORDER_SAVE
	#define Dr(x) istr.ReadWord(); istr >> x;
#else
	#define Dr(x) istr >> x;
#endif

TextInputStream &Turtle::read(TextInputStream &istr)
{
	istr >> (Point&)*this;
	Dr(pen);	Dr(font);
	Dr(shown); Dr(outOfBounds); 
	Dr(inEraseMode); Dr(cylinderLines); 
	Dr(polyMode); Dr(polyShown); 
	Dr(tessMode); Dr(tessShown);
	Dr(surfaceMode); Dr(shadows);
	Dr((wxInt32&)mode); Dr(heading);
	Dr(polypt1); Dr(polypt2); Dr(normal);
	Dr(texpt1); Dr(texpt2); Dr(texpos); Dr(lightPos);
	Dr(polyc1); Dr(polyc2); Dr(shadowcolor);
	Dr(u); Dr(hbase);
	Dr(tessWindingRule);
	Dr(polyPointNr); Dr(tessPointNr);
	Dr(surfaceShown); Dr(surfaceColumnNr);
	Dr(size); Dr(half_bottom); 
	Dr(_sin); Dr(_cos); Dr(wrapDelta);
	Dr(shoulder); Dr(head); Dr(polyv1);
	return istr;
}

#undef Dr
#define Dr(x) istr >> x;

DataInputStream &Turtle::read(DataInputStream &istr)
{
	istr >> (Point&)*this;
	Dr(pen);	Dr(font);
	Dr(shown); Dr(outOfBounds);
	Dr(inEraseMode); Dr(cylinderLines); 
	Dr(polyMode); Dr(polyShown); 
	Dr(tessMode); Dr(tessShown);
	Dr(surfaceMode); Dr(shadows);
	Dr((wxInt32&)mode); Dr(heading);
	Dr(polypt1); Dr(polypt2); Dr(normal);
	Dr(texpt1); Dr(texpt2); Dr(texpos); Dr(lightPos);
	Dr(polyc1); Dr(polyc2); Dr(shadowcolor);
	Dr(u); Dr(hbase);
	Dr(tessWindingRule);
	Dr(polyPointNr); Dr(tessPointNr);
	Dr(surfaceShown); Dr(surfaceColumnNr);
	Dr(size); Dr(half_bottom); 
	Dr(_sin); Dr(_cos); Dr(wrapDelta);
	Dr(shoulder); Dr(head); Dr(polyv1);
	return istr;
}

#ifdef DEBUG_RECORDER_SAVE
	#define D(x) #x << ' ' << x
#else
	#define D(x) x
#endif

TextOutputStream &Turtle::write(TextOutputStream &ostr)
{
	ostr << (Point&)*this << D(pen) << D(font) 
		<< D(shown) << ' ' << D(outOfBounds) << ' ' 
		<< D(inEraseMode) << ' ' << D(cylinderLines) << ' ' 
		<< D(polyMode) << ' ' << D(polyShown) << ' ' 
		<< D(tessMode) << ' ' << D(tessShown) << ' '
		<< D(surfaceMode) << ' ' << D(shadows) << ' '
		<< D(mode) << ' ' << D(heading) << ' '
		<< D(polypt1) << D(polypt2) << D(normal) 
		<< D(texpt1) << D(texpt2) << D(texpos) << D(lightPos)
		<< D(polyc1) << D(polyc2) << D(shadowcolor)
		<< D(u) << D(hbase) 
		<< D(tessWindingRule) << ' '
		<< D(polyPointNr) << ' ' << D(tessPointNr) << ' '
		<< D(surfaceShown) << ' ' << D(surfaceColumnNr) << ' '
		<< D(size) << ' ' << D(half_bottom) << ' ' 
		<< D(_sin) << ' ' << D(_cos) << ' ' << D(wrapDelta) << ' '
		<< D(shoulder) << D(head) << D(polyv1);
	return ostr;
}

#undef D
#define D(x) x

wxDataOutputStream &Turtle::write(wxDataOutputStream &ostr)
{
	ostr << (Point&)*this << D(pen) << D(font) 
		<< D(shown) << D(outOfBounds) 
		<< D(inEraseMode) << D(cylinderLines)
		<< D(polyMode) << D(polyShown)
		<< D(tessMode) << D(tessShown)
		<< D(surfaceMode) << D(shadows)
		<< D(mode)  << D(heading)
		<< D(polypt1) << D(polypt2) << D(normal) 
		<< D(texpt1) << D(texpt2) << D(texpos) << D(lightPos)
		<< D(polyc1) << D(polyc2) << D(shadowcolor)
		<< D(u) << D(hbase) 
		<< D(tessWindingRule)
		<< D(polyPointNr) << D(tessPointNr)
		<< D(surfaceShown) << D(surfaceColumnNr)
		<< D(size) << D(half_bottom)
		<< D(_sin) << D(_cos) << D(wrapDelta)
		<< D(shoulder) << D(head) << D(polyv1);
	return ostr;
}
//___________________________________________________________________

TextInputStream &Turtle::readEssentials(TextInputStream &istr)
{
	istr >> (Point&)*this >> u >> heading >> shown;
	return istr;
}
DataInputStream &Turtle::readEssentials(DataInputStream &istr)
{
	istr >> (Point&)*this >> u >> heading >> shown;
	return istr;
}

TextOutputStream &Turtle::writeEssentials(TextOutputStream &ostr)
{
	ostr << (Point&)*this << u << heading << ' ' << shown << ' ';
	return ostr;
}
wxDataOutputStream &Turtle::writeEssentials(wxDataOutputStream &ostr)
{
	ostr << (Point&)*this << u << heading << shown;
	return ostr;
}
//___________________________________________________________________

#define Command_Classes(x)												\
	x(Command)																\
	x(Graphic)																\
	x(MoveTo) x(LineTo) x(LineTo3d)									\
	x(setMatrix) x(setMatrix4x4) x(pushMatrix) x(popMatrix)	\
	x(setPixel)																\
	x(PolyStart) x(PolyPoint) x(PolyEnd)							\
	x(glBegin) x(TessPoint)												\
	x(QuadStripStart) x(QuadStrip) x(QuadStripEnd)				\
	x(fill) x(fillRect) x(fillEllipse)								\
	x(Sphere) x(Ellipsoid) x(partialEllipsoid)					\
	x(fillPie) x(Cylinder) x(PartCylinder)							\
	x(Perspective) x(unPerspective) x(ortho)						\
	x(setEye) x(wrap) x(mode)											\
	x(setScreenColor) x(Material) x(Light)							\
	x(setPenColor) x(setFloodColor) x(setShadowColor)			\
	x(Shininess)															\
	x(setPenDown)															\
	x(setPolySmooth) x(setLineSmooth) x(setPointSmooth)		\
	x(setDepthTest) x(setLighting) x(setDither)					\
	x(setTextured) x(setTextureFont)									\
	x(setPenRLE) x(setCyLines) x(setShadows)						\
	x(setFogEnabled) x(setPenMode)									\
	x(setPenSize) x(setPenPattern) x(setDepthFunc)				\
	x(setPolygonMode) x(setColorLogicOp) x(setPointSize)		\
	x(Texture) x(BitPaste) x(TessWindingRule)						\
	x(Label)																	\
	x(setLightPos)															\

#define make_aCommand(x)	const R_##x R_##x::a##x;

Command_Classes(make_aCommand)

void R_Command::free()
{	
	this->~R_Command();
	GC_::free(this);
}

void R_CommandList::clear()
{
	while (!empty()) 
		takeOut(next)->free();
}
void R_CommandList::free()
{	
	R_Command *c=this->next, *mem;
	while (c != this)
	{	mem=c;
		c=c->next;
		GC_::free(mem); mem=0;
	}
//	clear();
	GC_::free(this);
}

const Recorder Recorder::aRecorder;

const wxString
	R_Command::penUpDownString[]={"Up", "Down"},

	R_Command::commandTypeString[]=  // for debugging
	{	"nop", 
		"Graphic", "EndGraphic", 
		"moveTo", "lineTo", "lineTo3d", 
		"setMatrix", "setMatrix4x4", "pushMatrix", "popMatrix",
		"setPixel",
		"PolyStart", "PolyPoint", "PolyEnd", 
		"glBegin", "TessPoint", 
		"QuadStripStart", "QuadStrip", "QuadStripEnd",
		"fill", "fillRect", "fillEllipse", 
		"Sphere", "Ellipsoid", "partialEllipsoid", "fillPie", "cylinder", "partCylinder",
		"Perspective", "unPerspective", "ortho", "setEye", "wrap", "mode",
		"setScreenColor", "Material", "Light", 
		"setPenColor", "setFloodColor", "setShadowColor", 
		"Shininess",
		"setPenDown",
		"setPolySmooth", "setLineSmooth", "setPointSmooth",
		"setDepthTest", "setLighting", "setDither", "setTextured", "setTextureFont", 
		"setPenRLE", "setCyLines", "setShadows", "setFogEnabled", "setPenMode",
		"setPenSize", "setPenPattern", "setDepthFunc", 
		"setPolygonMode", "setColorLogicOp", "setPointSize",
		"Texture", "BitPaste", "TessWindingRule", 
		"Label", 
		"SetLightPos", 
		wxEmptyString
	};

#ifdef _DEBUG
	#undef new
	#define new new

	#define newOp_decl	\
		::operator new(size_t nSize, int blocktype, const char* newFilename_, int newLine_)
#else
	#define newOp_decl	\
		::operator new(size_t s)
#endif

#define newOp(x)										  \
	newOp_decl												\
	{	union													 \
		{	int* i;											  \
			R_Command *c;										\
		};													       \
		c=(R_Command*)GC_::malloc(sizeof(R_##x::a##x));\
		if (c)														\
			*i= *intP(&R_##x::a##x);					       \
		return c;												     \
	}

#define make_newOps(x)	void *R_##x newOp(x)

Command_Classes(make_newOps)

void *Recorder newOp_decl
{	union
	{	int* i;
		R_Command *c;
	};
	c=(R_Command*)GC_::malloc(sizeof(Recorder::aRecorder));
	if (c)
		*i= *intP(&Recorder::aRecorder);
	return c;
}


#ifdef _DEBUG
	#undef new
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

TextInputStream &R_Command::read(TextInputStream &istr) 
{	return istr;
}
DataInputStream &R_Command::read(DataInputStream &istr) 
{	return istr;
}
TextOutputStream &R_Command::write(TextOutputStream &ostr) 
{	return ostr << type << endL;
}
wxDataOutputStream &R_Command::write(wxDataOutputStream &ostr) 
{	return ostr << type;
}
//___________________________________________________________________

TextInputStream &operator>>(TextInputStream &istr, CommandListIterator &i)
{	if (i) 
		return i->read(istr);
	else
		return istr;
}
DataInputStream &operator>>(DataInputStream &istr, CommandListIterator &i)
{	if (i) 
		return i->read(istr);
	else
		return istr;
}
TextOutputStream &operator<<(TextOutputStream &ostr, CommandListIterator &i)
{	if (i) 
		return i->write(ostr);
	else 
		return ostr;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, CommandListIterator &i)
{	if (i) 
		return i->write(ostr);
	else 
		return ostr;
}
//___________________________________________________________________

const int Recorder::tableSize=1021;

PtrToInt_ *Recorder::ptrToInt;
IntToPtr_ *Recorder::intToPtr;

Recorder::Recorder() {}

Recorder::Recorder(GraphP graph_) 
{
	refCount=1;
	items=0;
	recording=pendown=true; 
	inReplay=restartFlag=stop_play=stop_confirm=false; 
	action=0;
	lastMove=0;
	insave=inload=false;
	graph=graph_;
	if (graph)
	{	screen=graph->screen;
		turtle=graph->turtle;
	}else
	{	screen=0;
		turtle=0;
	}
//	turtle->refCount++;

	recturtle=new Turtle;
	recturtle->shown=false;
//	recturtle->direct=false;	//comment out for debug

	recturtle->create(this, graph->screen);
}

void Recorder::free() 
{	
	clear();
	delete recturtle;
	GC_::free(this);
}

R_CommandList &Recorder::operator<<(R_Command *command)
{
	if (recording)
	{	ThreadP th=Thread::This();
		if (th->security->recordsize < items)
		{	error(SECURITY_ERROR, new String(_("to record more"))); return *this;
		}
		if (command)
		{	push_back(command);
			items++;
		}else if (!throwing)
			error(OUT_OF_MEM);
	}
	return *this;
}

void Recorder::save(const wxString name)
{
	if (!graph)
		return;
	wxFFileOutputStream output(name);
	wxZipOutputStream zip(output);
	zip.PutNextEntry(_T("graphics.dat"));
	wxDataOutputStream ostr(zip);
   tabs=0;
	insave=true;
	ptrToInt=new PtrToInt_(tableSize);
	if (!ptrToInt)
	{	error(OUT_OF_MEM); return;
	}
	ostr << screen->eye << screen->center << screen->up;
	ostr << *this;
	graph->turtle->write(ostr);
	delete ptrToInt;
	insave=false;
}

void Recorder::load(const wxString name)
{
	if (!graph)
		return;
	clear();
	wxFFileInputStream input(name);
	if (!input.Ok())
		return;
	
	wxZipInputStream zip(input);
	wxZipEntryPtr entry;
	if (entry.reset(zip.GetNextEntry()), entry.get() != 0)
	{	// access meta-data
		wxString name = entry->GetName();
		// read 'zip' to access the entry's data
		DataInputStream istr(zip);
		inload=true;
		intToPtr=new IntToPtr_(tableSize);
		if (!intToPtr)
		{	error(OUT_OF_MEM); return;
		}
		istr >> screen->eye >> screen->center >> screen->up;
		istr >> *this;
		graph->turtle->read(istr);
		delete intToPtr;
		inload=false;
	}
	playAgain();
}

void Recorder::saveText(const wxString name)
{
	if (!graph)
		return;
	ptrToInt=new PtrToInt_(tableSize);
	if (!ptrToInt)
	{	error(OUT_OF_MEM); return;
	}
	wxFFileOutputStream output(name);
	TextOutputStream ostr(output);
//	ostr.SetDoubleFormat(wxT("%g"));
	tabs=0;
	for (int i=1; commandTypeString[i]; i++) 
		ostr << i << ' ' << commandTypeString[i] << endL;
	ostr << int(0) << endL;
	ostr << screen->eye << screen->center << screen->up << endL;
	ostr << *this;
	graph->turtle->write(ostr);
	delete ptrToInt;
}

void Recorder::loadText(const wxString name)
{
	if (!graph)
		return;
	intToPtr=new IntToPtr_(tableSize);
	if (!intToPtr)
	{	error(OUT_OF_MEM); return;
	}
	clear();
	wxFFileInputStream input(name);
	TextInputStream istr(input);
	int start=-1;
	while (!input.Eof() && start!=0)
	{	istr >> start;
		if (start)
			istr.ReadWord() << endL;
	}
	istr >> screen->eye >> screen->center >> screen->up;
	istr >> *this;
	graph->turtle->read(istr);
	delete intToPtr;
	playAgain();
}

TextOutputStream &operator <<(TextOutputStream &ostr, Recorder &recorder)
{	
	CommandListIterator i(&recorder);
	
	for (i=recorder.begin();  i;  i++)
	{
#ifdef DEBUG_RECORDER_SAVE
		for (int t=0; t<recorder.tabs; t++)
			ostr << '\t';
		ostr << R_Command::commandTypeString[i->type] << ' ';
#endif
		ostr << i;
/*		if (ostr.Fail()) 
		{	error(FILE_ERROR, new String(
				"File too big or something in Recorder::save"));
			return ostr;
		}
*/	}
#ifdef DEBUG_RECORDER_SAVE
	for (int t=0; t<recorder.tabs; t++)
		ostr << '\t';
	ostr << R_Command::commandTypeString[R_Command::ENDGRAPHIC] << ' ';
#endif
	ostr << R_Command::ENDGRAPHIC << endL;
	return ostr;
}

wxDataOutputStream &operator <<(wxDataOutputStream &ostr, Recorder &recorder)
{	
	CommandListIterator i(&recorder);
	
	for (i=recorder.begin();  i;  i++)
	{	ostr << i;
	}
	ostr << R_Command::ENDGRAPHIC;
	return ostr;
}

#define RECORDER_CASES \
case R_Command::GRAPHIC:			c=new R_Graphic;				break;\
case R_Command::ENDGRAPHIC:		break;									\
case R_Command::MOVETO:				c=new R_MoveTo;				break;\
case R_Command::LINETO:				c=new R_LineTo;				break;\
case R_Command::LINETO3D:			c=new R_LineTo3d;				break;\
case R_Command::SETMATRIX:			c=new R_setMatrix;			break;\
case R_Command::SETMATRIX4X4:		c=new R_setMatrix4x4;		break;\
case R_Command::PUSHMATRIX:		c=new R_pushMatrix;			break;\
case R_Command::POPMATRIX:			c=new R_popMatrix;			break;\
case R_Command::SETPIXEL:			c=new R_setPixel;				break;\
case R_Command::POLYSTART:			c=new R_PolyStart;			break;\
case R_Command::POLYPOINT:			c=new R_PolyPoint;			break;\
case R_Command::POLYEND:			c=new R_PolyEnd;				break;\
case R_Command::GLBEGIN:			c=new R_glBegin;				break;\
case R_Command::TESSPOINT:			c=new R_TessPoint;			break;\
case R_Command::QUADSTRIPSTART:	c=new R_QuadStripStart;		break;\
case R_Command::QUADSTRIP:			c=new R_QuadStrip;			break;\
case R_Command::QUADSTRIPEND:		c=new R_QuadStripEnd;		break;\
case R_Command::FILL:				c=new R_fill;					break;\
case R_Command::FILLRECT:			c=new R_fillRect;				break;\
case R_Command::FILLELLIPSE:		c=new R_fillEllipse;			break;\
case R_Command::SPHERE:				c=new R_Sphere;				break;\
case R_Command::ELLIPSOID:			c=new R_Ellipsoid;			break;\
case R_Command::PARTIALELLIPSOID:c=new R_partialEllipsoid;	break;\
case R_Command::FILLPIE:			c=new R_fillPie;				break;\
case R_Command::CYLINDER:			c=new R_Cylinder;				break;\
case R_Command::PARTCYLINDER:		c=new R_PartCylinder;				break;\
case R_Command::PERSPECTIVE:		c=new R_Perspective;			break;\
case R_Command::UNPERSPECTIVE:	c=new R_unPerspective;		break;\
case R_Command::ORTHO:				c=new R_ortho;					break;\
case R_Command::SETEYE:				c=new R_setEye;				break;\
case R_Command::WRAP:				c=new R_wrap;					break;\
case R_Command::MODE:				c=new R_mode;					break;\
case R_Command::SETSCREENCOLOR:	c=new R_setScreenColor;		break;\
case R_Command::MATERIAL:			c=new R_Material;				break;\
case R_Command::LIGHT:				c=new R_Light;					break;\
case R_Command::SETPENCOLOR:		c=new R_setPenColor;			break;\
case R_Command::SETFLOODCOLOR:	c=new R_setFloodColor;		break;\
case R_Command::SETSHADOWCOLOR:	c=new R_setShadowColor;		break;\
case R_Command::SHININESS:			c=new R_Shininess;			break;\
case R_Command::SETPENDOWN: 		c=new R_setPenDown;			break;\
case R_Command::SETPOLYSMOOTH:	c=new R_setPolySmooth;		break;\
case R_Command::SETLINESMOOTH:	c=new R_setLineSmooth;		break;\
case R_Command::SETPOINTSMOOTH:	c=new R_setPointSmooth;		break;\
case R_Command::SETDEPTHTEST:		c=new R_setDepthTest;		break;\
case R_Command::SETLIGHTING:		c=new R_setLighting;			break;\
case R_Command::SETDITHER:			c=new R_setDither;			break;\
case R_Command::SETTEXTURED:		c=new R_setTextured;			break;\
case R_Command::SETTEXTUREFONT:	c=new R_setTextureFont;		break;\
case R_Command::SETPENRLE: 		c=new R_setPenRLE;			break;\
case R_Command::SETCYLINES: 		c=new R_setCyLines;			break;\
case R_Command::SETSHADOWS: 		c=new R_setShadows;			break;\
case R_Command::SETFOGENABLED:	c=new R_setFogEnabled;		break;\
case R_Command::SETPENMODE: 		c=new R_setPenMode;			break;\
case R_Command::SETPENSIZE:		c=new R_setPenSize;			break;\
case R_Command::SETPENPATTERN:	c=new R_setPenPattern;		break;\
case R_Command::SETDEPTHFUNC:		c=new R_setDepthFunc;		break;\
case R_Command::SETPOLYGONMODE:	c=new R_setPolygonMode;		break;\
case R_Command::SETCOLORLOGICOP:	c=new R_setColorLogicOp;	break;\
case R_Command::SETPOINTSIZE:		c=new R_setPointSize;		break;\
case R_Command::TEXTURE:			c=new R_Texture;				break;\
case R_Command::BITPASTE:			c=new R_BitPaste;				break;\
case R_Command::TESSWINDINGRULE:	c=new R_TessWindingRule;	break;\
case R_Command::LABEL:				c=new R_Label;					break;\
case R_Command::SETLIGHTPOS:		c=new R_setLightPos;			break;\

TextInputStream &operator >>(TextInputStream &istr, Recorder &recorder)
{	
	while (!istr.Eof())
	{	int i;
#ifdef DEBUG_RECORDER_SAVE
		istr.ReadWord();
#endif
		istr >> i;
		if (istr.Eof()) 
		{	//error(FILE_ERROR, new String(
			//	"File bad format or some strange bug in Recorder::load"));
			recorder.playAgain();
			return istr;
		}
#ifdef DEBUG_RECORDER_LOAD
		for (int tab=0; tab<recorder.tabs; tab++)
			c_out << '\t';
		if (i >=0 && i < R_Command::LASTCOMMAND)
			c_out << R_Command::commandTypeString[i] << endL;
#endif
		R_Command::Type t=R_Command::Type(i);
		R_Command *c;
		switch(t)
		{
		RECORDER_CASES
		default: 
//			_pause(0);
					c=new R_Command;	// should never be reached if file format is correct!
					t=R_Command::NOP;
		}
		if (t==R_Command::ENDGRAPHIC)
			break;
		if throwing return istr;
		c->type=t;
		istr >> *c;
		recorder.push_back(c);
		recorder.items++;
	}
	return istr;
}

DataInputStream &operator >>(DataInputStream &istr, Recorder &recorder)
{	
	while (!istr.Eof())
	{	int i;
		istr >> i;
		if (istr.Eof()) 
		{	//error(FILE_ERROR, new String(
			//	"File bad format or some strange bug in Recorder::load"));
			recorder.playAgain();
			return istr;
		}
#ifdef DEBUG_RECORDER_LOAD
		for (int tab=0; tab<recorder.tabs; tab++)
			c_out << '\t';
		if (i >=0 && i < R_Command::LASTCOMMAND)
			c_out << R_Command::commandTypeString[i] << endL;
#endif
		R_Command::Type t=R_Command::Type(i);
		R_Command *c;
		switch(t)
		{
		RECORDER_CASES
		default: 
//			_pause(0);
					c=new R_Command;	// should never be reached if file format is correct!
					t=R_Command::NOP;
		}
		if throwing return istr;
		if (t==R_Command::ENDGRAPHIC)
			break;
		c->type=t;
		istr >> *c;
		recorder.push_back(c);
		recorder.items++;
	}
	return istr;
}

wxString Recorder::rgb(Color color)  // I hope to reimplement palettes when I find time
{
	wxString sstr;
	sstr << color.r/255. << ' ' << color.g/255. << ' ' << color.b/255. << ' ';
	return sstr;
}

void Recorder::writePS(const wxString name)
{
	if (!graph)
		return;
	wxFFileOutputStream output(name);
	TextOutputStream ostr(output);
//	ostr.SetDoubleFormat(wxT("% 8.3f"));
	ostr<< wxT("%!PS-Adobe-3.0 EPSF-3.0") << endL
		<< wxT("%Title: ") << name << endL
		<< wxT("%%Creator: ") << aUCBLogo << "-" << versionString << endL
		<< wxT("%%BoundingBox: ") 
			<< (int)graph->turtle->screen->Min.x << ' '
			<< (int)graph->turtle->screen->Min.y << ' '
			<< (int)graph->turtle->screen->Max.x << ' '
			<< (int)graph->turtle->screen->Max.y << endL
		<< "%%EndComments" << endL
		<< "%%Page: " << name << endL
		<< "1 setlinecap 1 setlinejoin" << endL
		<< "/Bookman 24 selectfont" << endL
		<< endL
		<< "gsave" << endL
		<< rgb(graph->turtle->screen->color) << "setrgbcolor" << endL
		<< (int)graph->turtle->screen->Min.x << ' ' << (int)graph->turtle->screen->Min.y << " moveto" << endL
		<< (int)graph->turtle->screen->Max.x << ' ' << (int)graph->turtle->screen->Min.y << " lineto" << endL
		<< (int)graph->turtle->screen->Max.x << ' ' << (int)graph->turtle->screen->Max.y << " lineto" << endL
		<< (int)graph->turtle->screen->Min.x << ' ' << (int)graph->turtle->screen->Max.y << " lineto" << endL
		<< "   closepath fill" << endL

		<< "grestore" << endL
		<< endL;

	if (lastMove)
		ostr<< lastMove->point.x << ' ' << lastMove->point.y << " moveto" << endL;

	CommandListIterator i(this);

	for (i=begin();  i;  i++)
	{	i->writePS(ostr, *this);
/*			if (ostr.Fail()) 
		{	error(FILE_ERROR, new String(
				"writing failed in Recorder::writePS"));
			return Unbound;
		}
*/		}
	ostr<< "stroke" << endL
		<< "showpage" << endL
		<< "%%EOF" << endL;
}

DWORD WINAPI Recorder::playCommands(LPVOID lpv)
{
	ThreadP th=Thread::This();
	Recorder &recorder=*(Recorder*) lpv;
	bool r=recorder.recording;
	recorder.recording=false;

	Turtle *turtle=recorder.recturtle;
	turtle->shown=false;
	turtle->colorTolerance=recorder.graph->colorTolerance;
	bool direct=recorder.graph->direct;
	recorder.graph->direct=false;
	recorder.graph->clearScreen();
	Point zero(0,0,0);
	turtle->moveTo(zero);
	
	Turtle *turtsav=recorder.turtle;
	recorder.turtle=turtle;

	CommandListIterator i(&recorder);
	i=recorder.begin();
	unsigned int item=0;
	while (i  &&  item < recorder.items  &&  th->stoppingFlag!=THROWING)
	{	i->play(recorder);
//mySwapBuffers();
		++i;
		item++;
		if (recorder.restartFlag)
		{	i=recorder.begin();
			recorder.restartFlag=false;
		}
		if (recorder.stop_play)
			break;
	}
	if (turtle->shadows)
	{
#ifdef OPENGL
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);
		ShadowObject *so=recorder.graph->shadowObject;
		so->castShadows(turtsav->lightPos, 1);
		glDisable(GL_CULL_FACE);
#endif
	}
	recorder.turtle=turtsav;
	turtle->setHeading(recorder.graph->turtle->heading);
	turtle->shown = recorder.graph->turtle->shown;
//	turtle->draw();
	recorder.graph->direct=direct;

	recorder.graph->swapBuffers();

	recorder.inReplay=false;
	turtle->shown=false;
//	recturtle->destroy();

	recorder.recording=r;
	recorder.stop_confirm=true;
	return 0;
}

// this is only a class for testing multithreading, it doesn't work.
class RecorderPlayThread: public wxThread
{
public:
	Recorder *recorder;

	RecorderPlayThread(Recorder *rec): recorder(rec) {}

	ExitCode Entry()
	{
		if (!recorder->graph)
			return 0;
		ThreadP th=Thread::This();
#ifdef OPENGL
#ifdef WIN32
		BOOL ok=wglMakeCurrent((HDC) recorder->graph->GetHDC(), //Context()->m_hDC, 
			recorder->graph->GetContext()->GetGLRC());//m_glContext);
#else
		recorder->graph->SetCurrent();
#endif
#endif

		bool r=recorder->recording;
		recorder->recording=false;

		Turtle turtle;
		turtle.shown=false;
//		turtle.direct=false;	//comment out for debug

		turtle.create(recorder, recorder->graph->screen);

		recorder->graph->clearScreen();
		Point zero(0,0,0);
		turtle.moveTo(zero);

		Turtle *turtsav=recorder->turtle;
		recorder->turtle=&turtle;

		CommandListIterator i(recorder);
		i=recorder->begin();
		while (i  &&  th->stoppingFlag!=THROWING)
		{	i->play(*recorder);
			++i;
			if (recorder->restartFlag)
			{	i=recorder->begin();
				recorder->restartFlag=false;
			}
			if (recorder->stop_play)
				break;
		}
		if (turtle.shadows)
		{
#ifdef OPENGL
			glCullFace(GL_BACK);
			glEnable(GL_CULL_FACE);
			ShadowObject *so=recorder->graph->shadowObject;
			so->castShadows(turtsav->lightPos, 1);
			glDisable(GL_CULL_FACE);
#endif
		}
		recorder->turtle=turtsav;
		turtle.setHeading(recorder->graph->turtle->heading);
		turtle.shown = recorder->graph->turtle->shown;
		turtle.draw();

		recorder->graph->swapBuffers();

		recorder->inReplay=false;
		bool direct=recorder->graph->direct;
		recorder->graph->direct=false;

//th->GuiEnter();
//		recorder->updateView(turtle);
//th->GuiLeave();

//		::SwapBuffers((HDC) recorder->graph->GetContext()->m_hDC);
#ifdef WIN32
		wglMakeCurrent(0,0);
#else
		glXMakeCurrent((Display*)wxGetDisplay(), 0,0);
#endif

		turtle.shown=false;
		turtle.destroy();

		recorder->recording=r;
		recorder->stop_confirm=true;
		return 0;
	}
};

void Recorder::stop()
{
	if (!inReplay) return;
	stop_confirm=false;
	stop_play=true;
}

void Recorder::playAgain()
{
	if (inReplay) 
	{	restartFlag=true; return;
	}
	inReplay=true; // so this proc should not be entered twice 
					  // by multiple threads - when some exist - if there are some bugs
				    // - and they always exist - nothing serious is going to happen
	stop_play=false;

	playCommands(this);	// plain & correct, but freezing
/*	RecorderPlayThread *rpt=new RecorderPlayThread(this);
	rpt->Create(0x100000);
	rpt->Run();
*/
}

void Recorder::clear() 
{
	lastMove=0; 
	R_CommandList::clear(); 
	items=0; 
	R_BitPaste::bmps=0;
}

void Recorder::updateView(Turtle &turtle)
{
	if (graph)
	{	wxPaintEvent pe;
		graph->OnPaint(pe);
	}
}
//___________________________________________________________________

R_Graphic::R_Graphic(Turtle &turtle, Recorder *recorder_, double scale_) 
{	type=GRAPHIC;
	recorder=recorder_;
	recorder->refCount++;
	recturtle=new Turtle;
	*recturtle=turtle;
	saveturtle=new Turtle(false);
	scale=scale_;
}
void R_Graphic::free()
{
	if (recorder && !--recorder->refCount)
	{	recorder->free();
		recorder=0;
	}
	delete recturtle;
	delete saveturtle;
	GC_::free(this);
}
void R_Graphic::play(Recorder &rec) 
{
	ThreadP th=Thread::This();
	bool rr=recorder->recording;
	recorder->recording=false;

	*saveturtle=*rec.turtle;
	*rec.turtle=*recturtle;
#ifdef OPENGL
	glPushMatrix();
#endif
	rec.turtle->setTurtleMatrix();
#ifdef OPENGL
	glScalef(scale, scale, scale);
	GLboolean normalized;
	normalized=glIsEnabled(GL_NORMALIZE);
	glEnable(GL_NORMALIZE);
#endif
	*rec.turtle=Point(0,0,0);
	rec.turtle->reset3d();
	rec.turtle->heading=0;

	recorder->screen=rec.screen;
	CommandListIterator i(recorder);
	i=recorder->begin();
	while (i  &&  th->stoppingFlag!=THROWING)
	{	i->play(rec);
		++i;
	}
#ifdef OPENGL
	if (!normalized)
		glDisable(GL_NORMALIZE);
	glPopMatrix();
#endif
	*rec.turtle=*saveturtle;
	recorder->recording=rr;
}
TextInputStream  &R_Graphic::read(TextInputStream &istr) 
{	int nr;
	recturtle=new Turtle;
	saveturtle=new Turtle(false);
	recturtle->readEssentials(istr);
	istr >> scale >> nr;
	recorder = (Recorder*)Recorder::intToPtr->find(nr);
	if (!recorder)
	{	recorder=new Recorder(Thread::This()->graph);
		if (!recorder)
		{	error(OUT_OF_MEM); return istr;
		}
		Recorder::intToPtr->insert(recorder);
#ifdef DEBUG_RECORDER_LOAD
		recorder->tabs++;
#endif
		istr >> *recorder;
#ifdef DEBUG_RECORDER_LOAD
		recorder->tabs--;
#endif
	}
	else
		recorder->refCount++;
	return istr;
}
DataInputStream  &R_Graphic::read(DataInputStream &istr) 
{	int nr;
	recturtle=new Turtle;
	saveturtle=new Turtle(false);
	recturtle->read(istr);//->readEssentials(istr);
	istr >> scale >> nr;
	recorder = (Recorder*)Recorder::intToPtr->find(nr);
	if (!recorder)
	{	recorder=new Recorder(Thread::This()->graph);
		if (!recorder)
		{	error(OUT_OF_MEM); return istr;
		}
		Recorder::intToPtr->insert(recorder);
#ifdef DEBUG_RECORDER_LOAD
		recorder->tabs++;
#endif
		istr >> *recorder;
#ifdef DEBUG_RECORDER_LOAD
		recorder->tabs--;
#endif
	}
	else
		recorder->refCount++;
	return istr;
}
void R_Graphic::setPtrs(IntToPtr_ &intToPtr)
{
	CommandListIterator i(recorder);
	i=recorder->begin();
	while (i)
	{	i->setPtrs(intToPtr);
		i++;
	}
}
TextOutputStream &R_Graphic::write(TextOutputStream &ostr) 
{
	ostr << type << ' ';
	recturtle->writeEssentials(ostr);
	ostr << scale << ' ';
	int nr;
	nr=Recorder::ptrToInt->find(recorder);
	if (!nr)
	{	Recorder::ptrToInt->insert(recorder);
#ifdef DEBUG_RECORDER_SAVE
		recorder->tabs++;
#endif
		ostr << int(0) << endL 
			<< *recorder << endL;
#ifdef DEBUG_RECORDER_SAVE

		recorder->tabs--;
#endif
	}
	else
		ostr << nr << endL;
	return ostr;
}
wxDataOutputStream &R_Graphic::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	recturtle->write(ostr);//->writeEssentials(ostr);
	ostr << scale;
	int nr;
	nr=Recorder::ptrToInt->find(recorder);
	if (!nr)
	{	Recorder::ptrToInt->insert(recorder);
#ifdef DEBUG_RECORDER_SAVE
		recorder->tabs++;
#endif
		ostr << int(0) 
			<< *recorder;
#ifdef DEBUG_RECORDER_SAVE
		recorder->tabs--;
#endif
	}
	else
		ostr << nr;
	return ostr;
}
//___________________________________________________________________

R_MoveTo::R_MoveTo(Turtle &turtle, Recorder &recorder) 
{	init(turtle, recorder, Point(turtle));
}
R_MoveTo::R_MoveTo(Turtle &turtle, Recorder &recorder, 
						 Point point_) 
{	init(turtle, recorder, point_);
}
R_MoveTo::R_MoveTo(Turtle &turtle, Recorder &recorder, 
						 double x, double y) 
{	init(turtle, recorder, Point(x,y,0));
}
void R_MoveTo::play(Recorder &rec)
{	rec.turtle->moveTo(point);
	*rec.turtle=point;
}
TextInputStream  &R_MoveTo::read(TextInputStream &istr) 
{	return istr >> point;
}
DataInputStream  &R_MoveTo::read(DataInputStream &istr) 
{	return istr >> point;
}
TextOutputStream &R_MoveTo::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << point << endL;
}
wxDataOutputStream &R_MoveTo::write(wxDataOutputStream &ostr) 
{	return ostr << type << point;
}
void R_MoveTo::writePS(TextOutputStream &ostr, Recorder &recorder)
{	ostr << point.x << ' ' << point.y << " moveto" << endL;
}
//___________________________________________________________________

R_LineTo::R_LineTo(Turtle &turtle, Recorder &recorder) 
{	init(turtle, recorder, Point(turtle));
}
R_LineTo::R_LineTo(Turtle &turtle, Recorder &recorder, 
						 Point point_) 
{	init(turtle, recorder, point_);
}
R_LineTo::R_LineTo(Turtle &turtle, Recorder &recorder, 
						 double x, double y) 
{	init(turtle, recorder, Point(x,y,0));
}
void R_LineTo::play(Recorder &rec) 
{//	rec.turtle->u=u;
	rec.turtle->lineTo(point);	*rec.turtle=point;
}
TextInputStream  &R_LineTo::read(TextInputStream &istr) 
{	return istr >> point;// >> u;
}
DataInputStream  &R_LineTo::read(DataInputStream &istr) 
{	return istr >> point;// >> u;
}
TextOutputStream &R_LineTo::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << point
//		<< u 
		<< endL;
}
wxDataOutputStream &R_LineTo::write(wxDataOutputStream &ostr) 
{	return ostr << type << point;// << u;
}
void R_LineTo::writePS(TextOutputStream &ostr, Recorder &recorder)
{	if (recorder.pendown)
	{	ostr << point.x << ' ' << point.y << " lineto" << endL;
		recorder.lastMove=this;
		recorder.action++;
	}
}
//___________________________________________________________________

R_LineTo3d::R_LineTo3d(Turtle &turtle, Recorder &recorder) 
{	init(turtle, recorder, Point(turtle));
}
R_LineTo3d::R_LineTo3d(Turtle &turtle, Recorder &recorder, 
							  Point point_) 
{	init(turtle, recorder, point_);
}
R_LineTo3d::R_LineTo3d(Turtle &turtle, Recorder &recorder, 
							  double x, double y) 
{	init(turtle, recorder, Point(x,y,0));
}
void R_LineTo3d::init(Turtle &turtle, Recorder &recorder, 
							 Point point_)
{	type=LINETO3D;
	point=point_;
//	u=turtle.u;
	recorder.lastMove=(R_MoveTo*)this;
}
void R_LineTo3d::play(Recorder &rec) 
{//	rec.turtle->u=u;
	rec.turtle->lineTo(point);	*rec.turtle=point;
}
TextInputStream  &R_LineTo3d::read(TextInputStream &istr) 
{	return istr >> point;// >> u;
}
DataInputStream  &R_LineTo3d::read(DataInputStream &istr) 
{	return istr >> point;// >> u;
}
TextOutputStream &R_LineTo3d::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << point
//		<< u 
		<< endL;
}
wxDataOutputStream &R_LineTo3d::write(wxDataOutputStream &ostr) 
{	return ostr << type << point;// << u;
}
void R_LineTo3d::writePS(TextOutputStream &ostr, Recorder &recorder)
{	if (recorder.pendown)
	{	ostr << point.x << ' ' << point.y << " lineto" << endL;
		recorder.lastMove=this;
		recorder.action++;
	}
}
//___________________________________________________________________

void R_setMatrix::play(Recorder &rec) 
{	rec.turtle->setMatrix(basex,basey,basez,trans);
}
TextInputStream  &R_setMatrix::read(TextInputStream &istr) 
{	return istr >> basex >> basey >> basez >> trans;
}
DataInputStream  &R_setMatrix::read(DataInputStream &istr) 
{	return istr >> basex >> basey >> basez >> trans;
}
TextOutputStream &R_setMatrix::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' 
	<< basex << basey << basez << trans << endL;
}
wxDataOutputStream &R_setMatrix::write(wxDataOutputStream &ostr) 
{	return ostr << type
	<< basex << basey << basez << trans;
}
void R_setMatrix::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_setMatrix4x4::play(Recorder &rec) 
{
#ifdef OPENGL
	glLoadMatrixd(m);
#endif
}
TextInputStream  &R_setMatrix4x4::read(TextInputStream &istr) 
{	for (int i=0; i<16; i++)
		istr >> m[i];
	return istr;
}
DataInputStream  &R_setMatrix4x4::read(DataInputStream &istr) 
{	for (int i=0; i<16; i++)
		istr >> m[i];
	return istr;
}
TextOutputStream &R_setMatrix4x4::write(TextOutputStream &ostr) 
{	ostr << type << ' ';
	for (int i=0; i<16; i++)
		ostr << m[i] << ' ';
	ostr << endL;
	return ostr;
}
wxDataOutputStream &R_setMatrix4x4::write(wxDataOutputStream &ostr) 
{	ostr << type;
	for (int i=0; i<16; i++)
		ostr << m[i];
	return ostr;
}
void R_setMatrix4x4::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_pushMatrix::play(Recorder &rec) 
{
#ifdef OPENGL
	glPushMatrix();
#endif
}
void R_pushMatrix::writePS(TextOutputStream &ostr, Recorder &recorder) 
{} // still to do
//___________________________________________________________________

void R_popMatrix::play(Recorder &rec) 
{
#ifdef OPENGL
	glPopMatrix();
#endif
}
void R_popMatrix::writePS(TextOutputStream &ostr, Recorder &recorder)
{} // still to do
//___________________________________________________________________

void R_setPixel::play(Recorder &rec) 
{	rec.turtle->setPixel(point,color);
}
TextInputStream  &R_setPixel::read(TextInputStream &istr) 
{	return istr >> point >> color;
}
DataInputStream  &R_setPixel::read(DataInputStream &istr) 
{	return istr >> point >> color;
}
TextOutputStream &R_setPixel::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << point << color << endL;
}
wxDataOutputStream &R_setPixel::write(wxDataOutputStream &ostr) 
{	return ostr << type << point << color;
}
void R_setPixel::writePS(TextOutputStream &ostr, Recorder &recorder) 
{} // still to do
//___________________________________________________________________

R_PolyStart::R_PolyStart(Turtle &turtle, Point &normal_, 
								 Point &p3_, Point &t3_, Color c3_)
{	type=POLYSTART; 
	normal=normal_;
	p1=turtle.polypt1;
	p2=turtle.polypt2;
	p3=p3_;
	t1=turtle.texpt1;
	t2=turtle.texpt2;
	t3=t3_;
	c1=turtle.polyc1;
	c2=turtle.polyc2;
	c3=c3_;
}
void R_PolyStart::play(Recorder &rec) 
{
#ifdef OPENGL
	glBegin(GL_POLYGON);
	gl2psNormal3fv(floatP(&normal));
	glColor4ubv((unsigned char*)&c1);
	glTexCoord2fv(floatP(&t1));
	glVertex3fv(floatP(&p1));
	glColor4ubv((unsigned char*)&c2);
	glTexCoord2fv(floatP(&t2));
	glVertex3fv(floatP(&p2));
	glColor4ubv((unsigned char*)&c3);
	glTexCoord2fv(floatP(&t3));
	glVertex3fv(floatP(&p3));
#endif
}
TextInputStream  &R_PolyStart::read(TextInputStream &istr) 
{	return istr >> normal >> p1 >> p2 >> p3 >> t1 >> t2 >> t3;
}
DataInputStream  &R_PolyStart::read(DataInputStream &istr) 
{	return istr >> normal >> p1 >> p2 >> p3 >> t1 >> t2 >> t3;
}
TextOutputStream &R_PolyStart::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << normal
		<< p1 << p2 << p3
		<< t1 << t2 << t3
		<< endL;
}
wxDataOutputStream &R_PolyStart::write(wxDataOutputStream &ostr) 
{	return ostr << type << normal
		<< p1 << p2 << p3
		<< t1 << t2 << t3;
}
void R_PolyStart::writePS(TextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________


void R_PolyPoint::play(Recorder &rec) 
{
#ifdef OPENGL
	glColor4ubv((unsigned char*)&c);
	glTexCoord2fv(floatP(&t));
	glVertex3fv(floatP(&p));
#endif
	*rec.turtle=p;
}
TextInputStream  &R_PolyPoint::read(TextInputStream &istr) 
{	return istr >> p >> t;
}
DataInputStream  &R_PolyPoint::read(DataInputStream &istr) 
{	return istr >> p >> t;
}
TextOutputStream &R_PolyPoint::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << p << t << endL;
}
wxDataOutputStream &R_PolyPoint::write(wxDataOutputStream &ostr) 
{	return ostr << type << p << t;
}
void R_PolyPoint::writePS(TextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_PolyEnd::play(Recorder &rec) 
{
#ifdef OPENGL
	glEnd();
#endif
}
void R_PolyEnd::writePS(TextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_glBegin::play(Recorder &rec) 
{
#ifdef OPENGL
	glBegin(prim);
	gl2psNormal3fv(floatP(&normal));
#endif
}
TextInputStream  &R_glBegin::read(TextInputStream &istr) 
{	return istr >> prim >> normal;
}
DataInputStream  &R_glBegin::read(DataInputStream &istr) 
{	return istr >> prim >> normal;
}
TextOutputStream &R_glBegin::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << prim << ' ' << normal << endL;
}
wxDataOutputStream &R_glBegin::write(wxDataOutputStream &ostr) 
{	return ostr << type << prim << normal;
}
void R_glBegin::writePS(TextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_TessPoint::play(Recorder &rec) 
{
#ifdef OPENGL
	glColor4ubv((GLubyte*)&c);
	glTexCoord2fv(floatP(&t));
	glVertex3fv(floatP(&p));
#endif
	*rec.turtle=p;
}
TextInputStream  &R_TessPoint::read(TextInputStream &istr) 
{	return istr >> p >> t >> c;
}
DataInputStream  &R_TessPoint::read(DataInputStream &istr) 
{	return istr >> p >> t >> c;
}
TextOutputStream &R_TessPoint::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << p << t << c << endL;
}
wxDataOutputStream &R_TessPoint::write(wxDataOutputStream &ostr) 
{	return ostr << type << p << t << c;
}
void R_TessPoint::writePS(TextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_QuadStripStart::play(Recorder &rec) 
{
#ifdef OPENGL
	glBegin(GL_QUAD_STRIP);
	gl2psNormal3fv(floatP(&normal));
	glColor4ubv((unsigned char*)&c1);
	glTexCoord2fv(floatP(&t1));
	glVertex3fv(floatP(&p1));

	gl2psNormal3fv(floatP(&onormal));
	glColor4ubv((unsigned char*)&c2);
	glTexCoord2fv(floatP(&t2));
	glVertex3fv(floatP(&p2));
	
	glColor4ubv((unsigned char*)&c3);
	glTexCoord2fv(floatP(&t3));
	glVertex3fv(floatP(&p3));
	
	glColor4ubv((unsigned char*)&c4);
	glTexCoord2fv(floatP(&t4));
	glVertex3fv(floatP(&p4));
#endif
}
TextInputStream  &R_QuadStripStart::read(TextInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c1 >> p1 >> t1
		>> c2 >> p2 >> t2
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
DataInputStream  &R_QuadStripStart::read(DataInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c1 >> p1 >> t1
		>> c2 >> p2 >> t2
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
TextOutputStream &R_QuadStripStart::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << normal << onormal
		<< c1 << p1 << t1
		<< c2 << p2 << t2
		<< c3 << p3 << t3
		<< c4 << p4 << t4
		<< endL;
}
wxDataOutputStream &R_QuadStripStart::write(wxDataOutputStream &ostr) 
{	return ostr << type << normal << onormal
		<< c1 << p1 << t1
		<< c2 << p2 << t2
		<< c3 << p3 << t3
		<< c4 << p4 << t4;
}
void R_QuadStripStart::writePS(TextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_QuadStrip::play(Recorder &rec) 
{
#ifdef OPENGL
	gl2psNormal3fv(floatP(&normal));
	glColor4ubv((unsigned char*)&c3);
	glTexCoord2fv(floatP(&t3));
	glVertex3fv(floatP(&p3));

	gl2psNormal3fv(floatP(&onormal));
	glColor4ubv((unsigned char*)&c4);
	glTexCoord2fv(floatP(&t4));
	glVertex3fv(floatP(&p4));
#endif
}
TextInputStream  &R_QuadStrip::read(TextInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
DataInputStream  &R_QuadStrip::read(DataInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
TextOutputStream &R_QuadStrip::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << normal << onormal
		<< c3 << p3 << t3 
		<< c4 << p4 << t4 << endL;
}
wxDataOutputStream &R_QuadStrip::write(wxDataOutputStream &ostr) 
{	return ostr << type << normal << onormal
		<< c3 << p3 << t3 
		<< c4 << p4 << t4;
}
void R_QuadStrip::writePS(TextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_QuadStripEnd::play(Recorder &rec) 
{
#ifdef OPENGL
	glEnd();
#endif
}
//___________________________________________________________________

TextInputStream  &R_bool::read(TextInputStream &istr) 
{	return istr >> on;
}
DataInputStream  &R_bool::read(DataInputStream &istr) 
{	return istr >> on;
}
TextOutputStream &R_bool::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << on << endL;
}
wxDataOutputStream &R_bool::write(wxDataOutputStream &ostr) 
{	return ostr << type << on;
}
//___________________________________________________________________

void R_fill::play(Recorder &rec) 
{	rec.turtle->fill(mode);
}
TextInputStream  &R_fill::read(TextInputStream &istr) 
{	return istr >> mode;
}
DataInputStream  &R_fill::read(DataInputStream &istr) 
{	return istr >> mode;
}
TextOutputStream &R_fill::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << mode << endL;
}
wxDataOutputStream &R_fill::write(wxDataOutputStream &ostr) 
{	return ostr << type << mode;
}
void R_fill::writePS(TextOutputStream &ostr, Recorder &recorder)
{	ostr
		<< "closepath fill" << endL
		<< endL
		<< "newpath" << endL
		<< recorder.lastMove->point.x 
		<< recorder.lastMove->point.y << " moveto" << endL;
	recorder.action=0;
}
//___________________________________________________________________

R_fillRect::R_fillRect(Turtle &turtle, Point p1_, Point p2_)
{	type=FILLRECT; 
	p1=p1_;	
	p2=p2_;
	ori=turtle.orientation();
}
void R_fillRect::play(Recorder &rec) 
{	if (rec.turtle->screen->perspective)
		rec.turtle->setOrientation(ori);
	else
		rec.turtle->setHeading(ori.z);
	rec.turtle->fillRect(p1,p2);
}
TextInputStream  &R_fillRect::read(TextInputStream &istr) 
{	return istr >> p1 >> p2 >> ori;
}
DataInputStream  &R_fillRect::read(DataInputStream &istr) 
{	return istr >> p1 >> p2 >> ori;
}
TextOutputStream &R_fillRect::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << p1 << p2 << ori << endL;
}
wxDataOutputStream &R_fillRect::write(wxDataOutputStream &ostr) 
{	return ostr << type << p1 << p2 << ori;
}
void R_fillRect::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

R_fillEllipse::R_fillEllipse(Turtle &turtle, Point radius_)
{	type=FILLELLIPSE; 
	radius=radius_;
	ori=turtle.orientation();
}
void R_fillEllipse::play(Recorder &rec) 
{	if (rec.screen->perspective)
		rec.turtle->setOrientation(ori);
	else
		rec.turtle->setHeading(ori.z);
	rec.turtle->fillEllipse(radius);
}
TextInputStream  &R_fillEllipse::read(TextInputStream &istr) 
{	return istr >> radius >> ori;
}
DataInputStream  &R_fillEllipse::read(DataInputStream &istr) 
{	return istr >> radius >> ori;
}
TextOutputStream &R_fillEllipse::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << radius << ori << endL;
}
wxDataOutputStream &R_fillEllipse::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius << ori;
}
void R_fillEllipse::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_Sphere::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->sphere(radius, slices, stacks);
}
TextInputStream  &R_Sphere::read(TextInputStream &istr) 
{	return istr >> radius >> slices >> stacks >> ori;
}
DataInputStream  &R_Sphere::read(DataInputStream &istr) 
{	return istr >> radius >> slices >> stacks  >> ori;
}
TextOutputStream &R_Sphere::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << radius 
		<< ' ' << slices << ' ' << stacks << ' ' << ori	<< endL;
}
wxDataOutputStream &R_Sphere::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius << slices << stacks << ori;
}
void R_Sphere::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_Ellipsoid::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->ellipsoid(radius, slices, stacks);
}
TextInputStream  &R_Ellipsoid::read(TextInputStream &istr) 
{	return istr >> radius >> slices >> stacks >> ori;
}
DataInputStream  &R_Ellipsoid::read(DataInputStream &istr) 
{	return istr >> radius >> slices >> stacks >> ori;
}
TextOutputStream &R_Ellipsoid::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << radius 
		<< slices << ' ' << stacks << ' ' << ori << endL;
}
wxDataOutputStream &R_Ellipsoid::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius << slices << stacks << ori;
}
void R_Ellipsoid::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_partialEllipsoid::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->partialEllipsoid(radius, 
		slicesstart, slicesend, slices, 
		stacksstart, stacksend, stacks);
}
TextInputStream  &R_partialEllipsoid::read(TextInputStream &istr) 
{	return istr >> radius 
		>> slicesstart >> slicesend >> slices 
		>> stacksstart >> stacksend >> stacks >> ori;
}
DataInputStream  &R_partialEllipsoid::read(DataInputStream &istr) 
{	return istr >> radius 
		>> slicesstart >> slicesend >> slices 
		>> stacksstart >> stacksend >> stacks >> ori;
}
TextOutputStream &R_partialEllipsoid::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << radius 
		<< slicesstart << ' ' << slicesend << ' ' << slices << ' ' 
		<< stacksstart << ' ' << stacksend << ' ' << stacks << ' ' << ori << endL;
}
wxDataOutputStream &R_partialEllipsoid::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius 
		<< slicesstart << slicesend << slices 
		<< stacksstart << stacksend << stacks << ori;
}
void R_partialEllipsoid::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

R_fillPie::R_fillPie(Turtle &turtle, double angle_, Point radius_)
{	type=FILLPIE; 
	heading=turtle.heading;
	angle=angle_;
	radius=radius_;
}
void R_fillPie::play(Recorder &rec) 
{	rec.turtle->setHeading(heading);
	rec.turtle->fillPie(angle, radius);
}
TextInputStream  &R_fillPie::read(TextInputStream &istr) 
{	return istr >> heading >> angle >> radius;
}
DataInputStream  &R_fillPie::read(DataInputStream &istr) 
{	return istr >> heading >> angle >> radius;
}
TextOutputStream &R_fillPie::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << heading << ' '
		<< angle << ' ' << radius << endL;
}
wxDataOutputStream &R_fillPie::write(wxDataOutputStream &ostr) 
{	return ostr << type << heading << angle << radius;
}
void R_fillPie::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

R_Cylinder::R_Cylinder(Turtle &turtle, 
	double length_, double baseradius_, double topradius_)
{	type=CYLINDER; 
	ori=Point(
		turtle.roll(), 
		turtle.pitch(), 
		turtle.getHeading());
	length=length_;
	baseradius=baseradius_;
	topradius=topradius_;
}
void R_Cylinder::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->cylinder(length, baseradius, topradius);
}
TextInputStream  &R_Cylinder::read(TextInputStream &istr) 
{	return istr >> ori >> baseradius >> topradius;
}
DataInputStream  &R_Cylinder::read(DataInputStream &istr) 
{	return istr >> ori >> baseradius >> topradius;
}
TextOutputStream &R_Cylinder::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << ori
		<< length << ' ' << baseradius << ' ' << topradius << endL;
}
wxDataOutputStream &R_Cylinder::write(wxDataOutputStream &ostr) 
{	return ostr << type << ori << length << baseradius << topradius;
}
void R_Cylinder::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

R_PartCylinder::R_PartCylinder(Turtle &turtle, 
	double length_, double baseradius_, double topradius_, 
	double startangle_, double endangle_, double scrunch_, int slices_)
{	type=CYLINDER; 
	ori=Point(
		turtle.roll(), 
		turtle.pitch(), 
		turtle.getHeading());
	length=length_;
	baseradius=baseradius_;
	topradius=topradius_;
	startangle=startangle_;
	endangle=endangle_;
	scrunch=scrunch_;
	slices=slices_;
}
void R_PartCylinder::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->partialCylinder(length, baseradius, topradius, startangle, endangle, scrunch, slices);
}
TextInputStream  &R_PartCylinder::read(TextInputStream &istr) 
{	return istr >> ori 
		>> length >> baseradius >> topradius 
		>> startangle >> endangle >> scrunch >> slices;
}
DataInputStream  &R_PartCylinder::read(DataInputStream &istr) 
{	return istr >> ori 
		>> length >> baseradius >> topradius 
		>> startangle >> endangle >> scrunch >> slices;
}
TextOutputStream &R_PartCylinder::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << ori
		<< length << ' ' << baseradius << ' ' << topradius << ' '
		<< startangle << ' ' << endangle << ' ' << scrunch << ' ' << slices << endL;
}
wxDataOutputStream &R_PartCylinder::write(wxDataOutputStream &ostr) 
{	return ostr << type << ori 
		<< length << baseradius << topradius 
		<< startangle << endangle << scrunch << slices;
}
void R_PartCylinder::writePS(TextOutputStream &ostr, Recorder &recorder)
{
}
//___________________________________________________________________

void R_Perspective::play(Recorder &rec)
{
	rec.screen->perspective=true;
	rec.screen->orthographic=false;
	if (!rec.graph->savingToObj)
	{	rec.graph->setFrustum(rec.graph->GetSize().x, rec.graph->GetSize().y);
		rec.screen->setEye();
		rec.screen->getMatrices();
	}
#ifdef OPENGL
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
#endif
	rec.turtle->reset3d();
	rec.turtle->mode=Turtle::windowmode;
}
//___________________________________________________________________

void R_unPerspective::play(Recorder &rec)
{
	rec.screen->perspective=false;
	rec.screen->orthographic=false;
	rec.graph->setFrustum(rec.graph->GetSize().x, rec.graph->GetSize().y);
#ifdef OPENGL
	glLoadIdentity();
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
#endif
	rec.screen->getMatrices();
	rec.turtle->reset3d();
}
//___________________________________________________________________

void R_ortho::play(Recorder &rec)
{
	rec.screen->perspective=true;
	rec.screen->orthographic=true;
	rec.graph->setFrustum(rec.graph->GetSize().x, rec.graph->GetSize().y);
#ifdef OPENGL
	glLoadIdentity();
#endif
	rec.screen->setEye();
	rec.screen->getMatrices();
	rec.turtle->reset3d();
}
//___________________________________________________________________

R_setEye::R_setEye(Point &eye_, Point &center_, Point &up_):
	eye(eye_), center(center_), up(up_)
{	type=SETEYE;
}
void R_setEye::play(Recorder &rec) 
{	rec.screen->eye=eye;
	rec.screen->center=center;
	rec.screen->up=up;
	rec.screen->setEye();
	rec.screen->getMatrices();
}
TextInputStream  &R_setEye::read(TextInputStream &istr) 
{	return istr >> eye >> center >> up;
}
DataInputStream  &R_setEye::read(DataInputStream &istr) 
{	return istr >> eye >> center >> up;
}
TextOutputStream &R_setEye::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << eye << center << up << endL;
}
wxDataOutputStream &R_setEye::write(wxDataOutputStream &ostr) 
{	return ostr << type << eye << center << up;
}
void R_setEye::writePS(TextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_wrap::play(Recorder &rec)
{	rec.turtle->wrap();
}
//___________________________________________________________________

R_mode::R_mode(Turtle &turtle)
{	type=MODE;
	mode=turtle.mode;
}

void R_mode::play(Recorder &rec)
{
//	rec.turtle->draw();
	rec.turtle->mode=mode;
//	rec.turtle->draw();
}
TextInputStream  &R_mode::read(TextInputStream &istr) 
{	int &m=(int&)mode;
	return istr >> m;
}
DataInputStream  &R_mode::read(DataInputStream &istr) 
{	int &m=(int&)mode;
	return istr >> m;
}
TextOutputStream &R_mode::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << mode << endL;
}
wxDataOutputStream &R_mode::write(wxDataOutputStream &ostr) 
{	return ostr << type << mode;
}
//___________________________________________________________________

TextInputStream  &R_Color::read(TextInputStream &istr) 
{	return istr >> color;
}
DataInputStream  &R_Color::read(DataInputStream &istr) 

{	return istr >> color;
}
TextOutputStream &R_Color::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << color << endL;
}
wxDataOutputStream &R_Color::write(wxDataOutputStream &ostr) 
{	return ostr << type << color;
}
//___________________________________________________________________

void R_setScreenColor::play(Recorder &rec)
{	rec.turtle->screen->color=color; 
	rec.turtle->setScreenColor(color);
}
//___________________________________________________________________

void R_Material::play(Recorder &rec)
{
	float cf[4]=
	{	float(color.r)/255,
		float(color.g)/255,
		float(color.b)/255,
		float(color.a)/255
	};
#ifdef OPENGL
	glMaterialfv(GL_FRONT_AND_BACK, gltype, cf);
#endif
}
TextInputStream  &R_Material::read(TextInputStream &istr) 
{	return istr >> gltype >> color;
}
DataInputStream  &R_Material::read(DataInputStream &istr) 
{	return istr >> gltype >> color;
}
TextOutputStream &R_Material::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << gltype << ' ' << color << endL;
}
wxDataOutputStream &R_Material::write(wxDataOutputStream &ostr) 
{	return ostr << type << gltype << color;
}
//___________________________________________________________________

void R_Light::play(Recorder &rec)
{
	float cf[4]=
	{	float(color.r)/255,
		float(color.g)/255,
		float(color.b)/255,
		float(color.a)/255
	};
#ifdef OPENGL
	glLightfv(GL_LIGHT0, gltype, cf);
#endif
}
TextInputStream  &R_Light::read(TextInputStream &istr) 
{	return istr >> gltype >> color;
}
DataInputStream  &R_Light::read(DataInputStream &istr) 
{	return istr >> gltype >> color;
}
TextOutputStream &R_Light::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << gltype << ' ' << color << endL;
}
wxDataOutputStream &R_Light::write(wxDataOutputStream &ostr) 
{	return ostr << type << gltype << color;
}
//___________________________________________________________________

R_setPenColor::R_setPenColor(Turtle &turtle) 
{	type=SETPENCOLOR; color=turtle.pen.color;
}
void R_setPenColor::play(Recorder &rec) 
{	rec.turtle->setColor(color);
}
void R_setPenColor::writePS(TextOutputStream &ostr, Recorder &recorder)
{	if (recorder.action) 
	{	ostr<< "stroke" << endL
			<< recorder.lastMove->point.x 
			<< recorder.lastMove->point.y << " moveto" << endL;
		recorder.action=0;
	}
	ostr << recorder.rgb(color) << "setrgbcolor" << endL;
}
//___________________________________________________________________

R_setFloodColor::R_setFloodColor(Turtle &turtle) 
{	type=SETFLOODCOLOR; color=turtle.pen.floodcolor;
}
void R_setFloodColor::play(Recorder &rec) 
{	rec.turtle->pen.floodcolor=color;
}
void R_setFloodColor::writePS(TextOutputStream &ostr, Recorder &recorder)
{	if (recorder.action) 
	{	ostr<< "stroke" << endL
			<< recorder.lastMove->point.x 
			<< recorder.lastMove->point.y << " moveto" << endL;
		recorder.action=0;
	}
	ostr << recorder.rgb(color) << "setfloodcolor" << endL;
}
//___________________________________________________________________

R_setShadowColor::R_setShadowColor(Turtle &turtle) 
{	type=SETSHADOWCOLOR; color=turtle.shadowcolor;
}
void R_setShadowColor::play(Recorder &rec) 
{	rec.turtle->shadowcolor=color;
}
void R_setShadowColor::writePS(TextOutputStream &ostr, Recorder &recorder)
{
}
//___________________________________________________________________

void R_Shininess::play(Recorder &rec)
{
#ifdef OPENGL
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, s);
#endif
}
TextInputStream  &R_Shininess::read(TextInputStream &istr) 
{	return istr >> s;
}
DataInputStream  &R_Shininess::read(DataInputStream &istr) 
{	return istr >> s;
}
TextOutputStream &R_Shininess::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << s << ' ' << endL;
}
wxDataOutputStream &R_Shininess::write(wxDataOutputStream &ostr) 
{	return ostr << type << s;
}
//___________________________________________________________________

R_setPenDown::R_setPenDown(Turtle &turtle) 
{	type=SETPENDOWN; on=turtle.pen.down;
}
void R_setPenDown::play(Recorder &rec) 
{	rec.turtle->pen.down=on;
}
void R_setPenDown::writePS(TextOutputStream &ostr, Recorder &recorder) 
{	recorder.pendown=on;
}
//___________________________________________________________________

void R_setPolySmooth::play(Recorder &rec)
{
#ifdef OPENGL
	if (on)
		glEnable(GL_POLYGON_SMOOTH);
	else
		glDisable(GL_POLYGON_SMOOTH);
#endif
}
void R_setPolySmooth::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_setLineSmooth::play(Recorder &rec)
{
#ifdef OPENGL
	if (on)
		glEnable(GL_LINE_SMOOTH);
	else
		glDisable(GL_LINE_SMOOTH);
#endif
}
void R_setLineSmooth::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_setPointSmooth::play(Recorder &rec)
{
#ifdef OPENGL
	if (on)
		glEnable(GL_POINT_SMOOTH);
	else
		glDisable(GL_POINT_SMOOTH);
#endif
}
void R_setPointSmooth::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_setDepthTest::play(Recorder &rec)
{
#ifdef OPENGL
	if (on)
		glEnable(GL_DEPTH_TEST);
	else
		glDisable(GL_DEPTH_TEST);
#endif
}
void R_setDepthTest::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_setLighting::play(Recorder &rec)
{
#ifdef OPENGL
	if (on)
		glEnable(GL_LIGHTING);
	else
		glDisable(GL_LIGHTING);
#endif
}
void R_setLighting::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_setDither::play(Recorder &rec)
{
#ifdef OPENGL
	if (on)
		glEnable(GL_DITHER);
	else
		glDisable(GL_DITHER);
#endif
}
void R_setDither::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_setTextured::play(Recorder &rec)
{
#ifdef OPENGL
	if (on)
	{	glEnable(GL_TEXTURE_2D);
	}else
	{	glDisable(GL_TEXTURE_2D);
	}
#endif
	rec.screen->textured=on; 
}
void R_setTextured::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_setTextureFont::play(Recorder &rec)
{	if (on)
	{	rec.screen->textureFont=true; 

	}else
	{	rec.screen->textureFont=false; 
	}
}
void R_setTextureFont::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_setPenRLE::R_setPenRLE(Turtle &turtle) 
{	type=SETPENRLE; on=turtle.pen.roundLineEnds;
}
void R_setPenRLE::play(Recorder &rec) 
{	rec.turtle->pen.roundLineEnds=on;
}
void R_setPenRLE::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_setCyLines::R_setCyLines(Turtle &turtle) 
{	type=SETCYLINES; on=turtle.cylinderLines;
}
void R_setCyLines::play(Recorder &rec) 
{	rec.turtle->cylinderLines=on;
}
void R_setCyLines::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_setShadows::R_setShadows(Turtle &turtle) 
{	type=SETSHADOWS; on=turtle.shadows;
}
void R_setShadows::play(Recorder &rec) 
{	rec.turtle->shadows=on;
}
void R_setShadows::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_setFogEnabled::R_setFogEnabled(bool fog) 
{	type=SETCYLINES; on=fog;
}
void R_setFogEnabled::play(Recorder &rec) 
{
#ifdef OPENGL
	if (on)
		glEnable(GL_FOG);
	else
		glDisable(GL_FOG);
#endif
}
void R_setFogEnabled::writePS(TextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_setPenMode::R_setPenMode(Turtle &turtle) 
{	type=SETPENMODE; 
	penMode=turtle.pen.mode; 
	internalMode=turtle.pen.internalMode;
}
void R_setPenMode::play(Recorder &rec) 
{	rec.turtle->setPenMode(penMode); 
	rec.turtle->pen.internalMode=internalMode;
}
TextInputStream  &R_setPenMode::read(TextInputStream &istr) 
{	int &pm=(int&)penMode, &im=(int&)internalMode;
	return istr >> pm >> im;
}
DataInputStream  &R_setPenMode::read(DataInputStream &istr) 
{	int &pm=(int&)penMode, &im=(int&)internalMode;
	return istr >> pm >> im;
}
TextOutputStream &R_setPenMode::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' 
		<< penMode << ' ' << internalMode << endL;
}
wxDataOutputStream &R_setPenMode::write(wxDataOutputStream &ostr) 
{	return ostr << type
		<< penMode << internalMode;
}
//___________________________________________________________________

R_setPenSize::R_setPenSize(Turtle &turtle) 
{	type=SETPENSIZE; penSize=turtle.pen.size;
}
void R_setPenSize::play(Recorder &rec) 
{	rec.turtle->setPenSize(penSize);
}
TextInputStream  &R_setPenSize::read(TextInputStream &istr) 
{	return istr >> penSize;
}
DataInputStream  &R_setPenSize::read(DataInputStream &istr) 
{	return istr >> penSize;
}
TextOutputStream &R_setPenSize::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << penSize << endL;
}
wxDataOutputStream &R_setPenSize::write(wxDataOutputStream &ostr) 
{	return ostr << type << penSize;
}
void R_setPenSize::writePS(TextOutputStream &ostr, Recorder &recorder)
{	if (recorder.action) 
	{	ostr<< "stroke" << endL
			<< recorder.lastMove->point.x 
			<< recorder.lastMove->point.y << " moveto" << endL;
		recorder.action=0;
	}
	ostr << penSize.x << " setlinewidth" << endL;
}
//___________________________________________________________________

R_setPenPattern::R_setPenPattern(Turtle &turtle) 
{	type=SETPENPATTERN; 
	penPattern=turtle.pen.getPattern();
}
void R_setPenPattern::play(Recorder &rec) 
{	rec.turtle->pen.setPattern(penPattern);
}
TextInputStream  &R_setPenPattern::read(TextInputStream &istr) 
{	istr >> penPattern;
	return istr;
}
DataInputStream  &R_setPenPattern::read(DataInputStream &istr) 
{	istr >> penPattern;
	return istr;
}
TextOutputStream &R_setPenPattern::write(TextOutputStream &ostr) 
{	ostr << type << ' ' << penPattern << endL;
	return ostr;
}
wxDataOutputStream &R_setPenPattern::write(wxDataOutputStream &ostr) 
{	ostr << type << penPattern;
	return ostr;
}
//___________________________________________________________________

R_setDepthFunc::R_setDepthFunc(int func_) 
{	type=SETDEPTHFUNC; 
	func=func_;
}
void R_setDepthFunc::play(Recorder &rec) 
{
#ifdef OPENGL
	glDepthFunc(GL_NEVER+func);
#endif
}
TextInputStream  &R_setDepthFunc::read(TextInputStream &istr) 
{	istr >> func;
	return istr;
}
DataInputStream  &R_setDepthFunc::read(DataInputStream &istr) 
{	istr >> func;
	return istr;
}
TextOutputStream &R_setDepthFunc::write(TextOutputStream &ostr) 
{	ostr << type << ' ' << func << endL;
	return ostr;
}
wxDataOutputStream &R_setDepthFunc::write(wxDataOutputStream &ostr) 
{	ostr << type << func;
	return ostr;
}
//___________________________________________________________________

R_setPolygonMode::R_setPolygonMode(int mode_) 
{	type=SETPOLYGONMODE; 
	mode=mode_;
}
void R_setPolygonMode::play(Recorder &rec) 
{
#ifdef OPENGL
	glPolygonMode(GL_FRONT_AND_BACK, GL_POINT+mode);
#endif
}
TextInputStream  &R_setPolygonMode::read(TextInputStream &istr) 
{	istr >> mode;
	return istr;
}
DataInputStream  &R_setPolygonMode::read(DataInputStream &istr) 
{	istr >> mode;
	return istr;
}
TextOutputStream &R_setPolygonMode::write(TextOutputStream &ostr) 
{	ostr << type << ' ' << mode << endL;
	return ostr;
}
wxDataOutputStream &R_setPolygonMode::write(wxDataOutputStream &ostr) 
{	ostr << type << mode;
	return ostr;
}
//___________________________________________________________________

R_setColorLogicOp::R_setColorLogicOp(int mode_) 
{	type=SETCOLORLOGICOP; 
	mode=mode_;
}
void R_setColorLogicOp::play(Recorder &rec) 
{
#ifdef OPENGL
	if (mode)
	{	glLogicOp(GL_CLEAR+mode);
		glEnable(GL_COLOR_LOGIC_OP);
	}else
		glDisable(GL_COLOR_LOGIC_OP);
#endif
}
TextInputStream  &R_setColorLogicOp::read(TextInputStream &istr) 
{	istr >> mode;
	return istr;
}
DataInputStream  &R_setColorLogicOp::read(DataInputStream &istr) 
{	istr >> mode;
	return istr;
}
TextOutputStream &R_setColorLogicOp::write(TextOutputStream &ostr) 
{	ostr << type << ' ' << mode << endL;
	return ostr;
}
wxDataOutputStream &R_setColorLogicOp::write(wxDataOutputStream &ostr) 
{	ostr << type << mode;
	return ostr;
}
//___________________________________________________________________

R_setPointSize::R_setPointSize(double size_) 
{	type=SETPOINTSIZE; 
	size=size_;
}
void R_setPointSize::play(Recorder &rec) 
{
#ifdef OPENGL
	glPointSize(size);
#endif
}
TextInputStream  &R_setPointSize::read(TextInputStream &istr) 
{	istr >> size;
	return istr;
}
DataInputStream  &R_setPointSize::read(DataInputStream &istr) 
{	istr >> size;
	return istr;
}
TextOutputStream &R_setPointSize::write(TextOutputStream &ostr) 
{	ostr << type << ' ' << size << endL;
	return ostr;
}
wxDataOutputStream &R_setPointSize::write(wxDataOutputStream &ostr) 
{	ostr << type << size;
	return ostr;
}
//___________________________________________________________________

int R_Texture::maxnr=0;
PairP R_Texture::textures=0;

R_Texture::R_Texture(TextureP tex_):
	x(tex_->_bitmap->winx), y(tex_->_bitmap->winy), 
	texi(tex_->_texnr), pixels(tex_->_bitmap->pixels),
	tex(tex_)
{	type=TEXTURE;
	set=true;
}
void R_Texture::play(Recorder &rec) 
{
#ifdef OPENGL
	glBindTexture(GL_TEXTURE_2D, tex->_texnr);
	maxnr=max(maxnr, tex->_texnr);
	glEnable(GL_TEXTURE_2D);
#endif
	rec.screen->textured=true;
}
void R_Texture::setPtrs(IntToPtr_ &intToPtr)
{
	if (!set)
	{	TextureP t = (Texture*)intToPtr.find(int(tex));
		tex=t;
		PairP tl=new Pair(tex, textures);
		if throwing return;
		textures=tl;
		set=true;
	}
}
TextInputStream  &R_Texture::read(TextInputStream &istr) 
{	ThreadP th=Thread::This();
	int nr;
	istr >> nr;
	if (RECORDER.inload)
	{	if (nr)
		{	tex = (Texture*)RECORDER.intToPtr->find(nr);
			set=true;
		}
		else
		{	istr >> x >> y;
			int size=sizeof(Color)*x*y;
			pixels=ColorP(GC_::malloc(size));
			if (!pixels)
			{	x=y=size=0;
				tex=0;
				return istr;
			}
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				istr >> *c++;
			BitmapP bm=new Bitmap(x,y, x,y, pixels);
			if throwing return istr;
			tex=new Texture(bm);
			if throwing return istr;
			PairP tl=new Pair(tex, textures);
			if throwing return istr;
			textures=tl;
			set=true;
			RECORDER.intToPtr->insert(&*tex);
		}
	}
	else
	{	if (nr)
		{	tex = (Texture*)nr;
			set=false;
		}
		else
		{	istr >> (int&)tex;
			set=false;
		}
	}
	return istr;
}
DataInputStream  &R_Texture::read(DataInputStream &istr) 
{	ThreadP th=Thread::This();
	int nr;
	istr >> nr;
	if (RECORDER.inload)
	{	if (nr)
		{	tex = (Texture*)RECORDER.intToPtr->find(nr);
			set=true;
		}
		else
		{	istr >> x >> y;
			int size=sizeof(Color)*x*y;
			pixels=ColorP(GC_::malloc(size));
			if (!pixels)
			{	x=y=size=0;
				tex=0;
				return istr;
			}
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				istr >> *c++;
			BitmapP bm=new Bitmap(x,y, x,y, pixels);
			if throwing return istr;
			tex=new Texture(bm);
			if throwing return istr;
			PairP tl=new Pair(tex, textures);
			if throwing return istr;
			textures=tl;
			set=true;
			RECORDER.intToPtr->insert(&*tex);
		}
	}
	else
	{	if (nr)
		{	tex = (Texture*)nr;
			set=false;
		}
		else
		{	istr >> (int&)tex;
			set=false;
		}
	}
	return istr;
}
TextOutputStream &R_Texture::write(TextOutputStream &ostr) 
{	ThreadP th=Thread::This();
	ostr << type << ' ';
	if (RECORDER.insave)
	{	int nr = RECORDER.ptrToInt->find(tex);
		if (nr)
			ostr << nr << ' ';
		else
		{	RECORDER.ptrToInt->insert(&*tex);
			nr=RECORDER.ptrToInt->Nr();
			x=tex->_bitmap->winx;
			y=tex->_bitmap->winy;
			texi=tex->_texnr;
			pixels=tex->_bitmap->pixels;
			ostr << int(0) << ' ' << x << ' ' << y << ' ';
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				ostr << *c++;
		}
	}
	else
	{	int nr = GC_::ptrToInt->find(tex);
		if (nr)
			ostr << nr;
		else
		{	ostr << int(0);
			GC_::ptrToInt->insert(&*tex);
			tex->write(ostr);
		}		
	}	
	return ostr;
	return ostr;
}
wxDataOutputStream &R_Texture::write(wxDataOutputStream &ostr) 
{	ThreadP th=Thread::This();
	ostr << type;
	if (RECORDER.insave)
	{	int nr = RECORDER.ptrToInt->find(tex);
		if (nr)
			ostr << nr;
		else
		{	RECORDER.ptrToInt->insert(&*tex);
			nr=RECORDER.ptrToInt->Nr();
			x=tex->_bitmap->winx;
			y=tex->_bitmap->winy;
			texi=tex->_texnr;
			pixels=tex->_bitmap->pixels;
			ostr << int(0) << x << y;
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				ostr << *c++;
		}
	}
	else
	{	int nr = GC_::ptrToInt->find(tex);
		if (nr)
			ostr << nr;
		else
		{	ostr << int(0);
			GC_::ptrToInt->insert(&*tex);
			tex->write(ostr);
		}		
	}	
	return ostr;
}
//___________________________________________________________________

PairP R_BitPaste::bmps=0;

R_BitPaste::R_BitPaste(BitmapP bmp_):
	bmp(bmp_)
{	type=BITPASTE;
	PairP bl=new Pair(bmp, bmps);
	if throwing return;
	bmps=bl;
	set=true;
}
void R_BitPaste::play(Recorder &rec) 
{
	rec.turtle->bitPaste(bmp);
}
void R_BitPaste::setPtrs(IntToPtr_ &intToPtr)
{
	if (!set)
	{	BitmapP b = (Bitmap*)intToPtr.find(int(bmp));
		bmp=b;
		PairP bl=new Pair(bmp, bmps);
		if throwing return;
		bmps=bl;
		set=true;
	}
}
TextInputStream  &R_BitPaste::read(TextInputStream &istr) 
{	ThreadP th=Thread::This();
	int nr;
	istr >> nr;
	if (RECORDER.inload)
	{	if (nr)
		{	bmp = (Bitmap*)RECORDER.intToPtr->find(nr);
			set=true;
		}
		else
		{	int x=bmp->winx, y=bmp->winy;
			istr >> x >> y;
			int size=sizeof(Color)*x*y;
			ColorP pixels=ColorP(GC_::malloc(size));
			if (!pixels)
			{	size=0;
				bmp=0;
				return istr;
			}
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				istr >> *c++;
			bmp=new Bitmap(x,y, x,y, pixels);
			if throwing return istr;
			PairP bl=new Pair(bmp, bmps);
			if throwing return istr;
			bmps=bl;
			set=true;
			RECORDER.intToPtr->insert(&*bmp);
		}
	}
	else
	{	if (nr)
		{	bmp = (Bitmap*)nr;
			set=false;
		}
		else
		{	istr >> (int&)bmp;
			set=false;
		}
	}
	return istr;
}
DataInputStream  &R_BitPaste::read(DataInputStream &istr) 
{	ThreadP th=Thread::This();
	int nr;
	istr >> nr;
	if (RECORDER.inload)
	{	if (nr)
		{	bmp = (Bitmap*)RECORDER.intToPtr->find(nr);
			set=true;
		}
		else
		{	int x=bmp->winx, y=bmp->winy;
			istr >> x >> y;
			int size=sizeof(Color)*x*y;
			ColorP pixels=ColorP(GC_::malloc(size));
			if (!pixels)
			{	x=y=size=0;
				bmp=0;
				return istr;
			}
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				istr >> *c++;
			bmp=new Bitmap(x,y, x,y, pixels);
			if throwing return istr;
			PairP bl=new Pair(bmp, bmps);
			if throwing return istr;
			bmps=bl;
			set=true;
			RECORDER.intToPtr->insert(&*bmp);
		}
	}
	else
	{	if (nr)
		{	bmp = (Bitmap*)nr;
			set=false;
		}
		else
		{	istr >> (int&)bmp;
			set=false;
		}
	}
	return istr;
}
TextOutputStream &R_BitPaste::write(TextOutputStream &ostr) 
{	ThreadP th=Thread::This();
	ostr << type << ' ';
	if (RECORDER.insave)
	{	int nr = RECORDER.ptrToInt->find(bmp);
		if (nr)
			ostr << nr << ' ';
		else
		{	RECORDER.ptrToInt->insert(&*bmp);
			nr=RECORDER.ptrToInt->Nr();
			int x=bmp->winx;
			int y=bmp->winy;
			ostr << int(0) << ' ' << x << ' ' << y << ' ';
			ColorP c=bmp->pixels;
			for (int i=x*y; i; i--)
				ostr << *c++;
		}
	}
	else
	{	int nr = GC_::ptrToInt->find(bmp);
		if (nr)
			ostr << nr;
		else
		{	ostr << int(0);
			GC_::ptrToInt->insert(&*bmp);
			bmp->write(ostr);
		}		
	}	
	return ostr;
}
wxDataOutputStream &R_BitPaste::write(wxDataOutputStream &ostr) 
{	ThreadP th=Thread::This();
	ostr << type;
	if (RECORDER.insave)
	{	int nr = RECORDER.ptrToInt->find(bmp);
		if (nr)
			ostr << nr;
		else
		{	RECORDER.ptrToInt->insert(&*bmp);
			nr=RECORDER.ptrToInt->Nr();
			int x=bmp->winx;
			int y=bmp->winy;
			ostr << int(0) << x << y;
			ColorP c=bmp->pixels;
			for (int i=x*y; i; i--)
				ostr << *c++;
		}
	}
	else
	{	int nr = GC_::ptrToInt->find(bmp);
		if (nr)
			ostr << nr;
		else
		{	ostr << int(0);
			GC_::ptrToInt->insert(&*bmp);
			bmp->write(ostr);
		}		
	}	
	return ostr;
}
//___________________________________________________________________

void R_TessWindingRule::play(Recorder &rec) 
{	rec.turtle->tessWindingRule=rule;
}
TextInputStream  &R_TessWindingRule::read(TextInputStream &istr) 
{	return istr >> rule;
}
DataInputStream  &R_TessWindingRule::read(DataInputStream &istr) 
{	return istr >> rule;
}
TextOutputStream &R_TessWindingRule::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << rule << endL;
}
wxDataOutputStream &R_TessWindingRule::write(wxDataOutputStream &ostr) 
{	return ostr << type << rule;
}
//___________________________________________________________________

R_Label::R_Label(Turtle &turtle, wxString str_) 
{	type=LABEL; str=str_; 
	ori=Point(
		turtle.roll(), 
		turtle.pitch(), 
		turtle.getHeading());
	font=turtle.font;
}
void R_Label::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->font=font;
	rec.turtle->Label(str); rec.turtle->moveTo();
}
TextInputStream  &R_Label::read(TextInputStream &istr) 
{	int len;
	istr >> len;
	char dummy=istr.GetChar();
	str=wxEmptyString;
	for (int i=1; i<=len && dummy!=EOF; i++)
	{	dummy=istr.GetChar();
		if (dummy!=EOF)
			str+=dummy;
	}
	return istr >> ori >> font;
}
DataInputStream  &R_Label::read(DataInputStream &istr) 
{	return istr >> str >> ori >> font;
}
TextOutputStream &R_Label::write(TextOutputStream &ostr) 
{	ostr << type << ' ' << str.Length() << ' ' << '"' << str << ' ' << ori << ' ' << font << endL;
	return ostr;
}
wxDataOutputStream &R_Label::write(wxDataOutputStream &ostr) 
{	return ostr << type << str << ori << font;
}
void R_Label::writePS(TextOutputStream &ostr, Recorder &recorder)
{	for (unsigned int i=0;  i < str.length();  i++)
	{	char ch=str[i];
		if (ch=='(' || ch==')' || ch=='\\')
			ostr << '\\';
		ostr << ch;

	}
	ostr<< ") show grestore" << endL
		<< recorder.lastMove->point.x 
		<< recorder.lastMove->point.y << " moveto" << endL;
}
//___________________________________________________________________

void R_setLightPos::play(Recorder &rec) 
{	float p[4]={point.x,point.y,point.z,0};
#ifdef OPENGL
	glLightfv(GL_LIGHT0, GL_POSITION, p);
#endif
}
TextInputStream  &R_setLightPos::read(TextInputStream &istr) 
{	return istr >> point;
}
DataInputStream  &R_setLightPos::read(DataInputStream &istr) 
{	return istr >> point;
}
TextOutputStream &R_setLightPos::write(TextOutputStream &ostr) 
{	return ostr << type << ' ' << point << endL;
}
wxDataOutputStream &R_setLightPos::write(wxDataOutputStream &ostr) 
{	return ostr << type << point;
}
void R_setLightPos::writePS(TextOutputStream &ostr, Recorder &recorder) 
{} // still to do
