//___________________________________________________________________
//
//                         Class Recorder
//___________________________________________________________________

class Graph;
class Screen_;
class Font_;
class Turtle;
class PtrToInt_;
class IntToPtr_;
class Recorder;
class R_MoveTo;
class TextInputStream;
class TextOutputStream;

class R_Command
{
	friend class R_CommandList;
	friend class CommandListIterator;
	R_Command *prev, *next;
	public:
		enum Type	   // types of graphics moves that can be recorded
		{	NOP,		// and some internal help commands to keep things weird
			GRAPHIC, ENDGRAPHIC, 
			MOVETO, LINETO, LINETO3D, 
			SETMATRIX, SETMATRIX4X4, PUSHMATRIX, POPMATRIX,
			SETPIXEL,
			POLYSTART, POLYPOINT, POLYEND, 
			GLBEGIN, TESSPOINT,
			QUADSTRIPSTART, QUADSTRIP, QUADSTRIPEND,
			FILL, FILLRECT, FILLELLIPSE, 
			SPHERE, ELLIPSOID, PARTIALELLIPSOID, FILLPIE, CYLINDER, PARTCYLINDER, 
			PERSPECTIVE, UNPERSPECTIVE, ORTHO, SETEYE, WRAP, MODE,
			SETSCREENCOLOR, MATERIAL, LIGHT, 
			SETPENCOLOR,SETFLOODCOLOR, SETSHADOWCOLOR, 
			SHININESS, 
			SETPENDOWN, 
			SETPOLYSMOOTH, SETLINESMOOTH, SETPOINTSMOOTH,
			SETDEPTHTEST, SETLIGHTING, SETDITHER, SETTEXTURED, SETTEXTUREFONT,
			SETPENRLE, SETCYLINES, SETSHADOWS, SETFOGENABLED, SETPENMODE, 
			SETPENSIZE, SETPENPATTERN, SETDEPTHFUNC, SETPOLYGONMODE, 
			SETCOLORLOGICOP, SETPOINTSIZE,
			TEXTURE, BITPASTE, TESSWINDINGRULE,
			LABEL, 
			SETLIGHTPOS,
			LASTCOMMAND
		};
		Type type;

		static const wxString 
			penUpDownString[], commandTypeString[];  // for debugging

		static const R_Command aCommand;
		operator_new;

		R_Command() {prev=next=this;}
		virtual ~R_Command() {}
		virtual void free();
		virtual void play(Recorder &rec) {}

		virtual void setPtrs(IntToPtr_ &intToPtr) {}
		virtual TextInputStream  &read(TextInputStream &istr);
		virtual TextOutputStream &write(TextOutputStream &ostr);
		virtual DataInputStream  &read(DataInputStream &istr);
		virtual wxDataOutputStream &write(wxDataOutputStream &ostr);
		virtual void writePS(TextOutputStream &ostr, Recorder &recorder) {}
//		virtual void writeVRML(TextOutputStream &ostr, Recorder &recorder);

	friend TextInputStream &operator>>(TextInputStream &istr, R_Command &command) {return command.read(istr);}
	friend TextOutputStream &operator<<(TextOutputStream &ostr, R_Command &command) {return command.write(ostr);}
	friend DataInputStream &operator>>(DataInputStream &istr, R_Command &command) {return command.read(istr);}
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, R_Command &command) {return command.write(ostr);}
};

class R_CommandList: public R_Command
{
	friend class CommandListIterator;
	public:
		static const R_CommandList aCommandList;
		operator_new;

		bool empty() {return next==this;}
		R_Command *valid(R_Command *command) {return command!=this ?   command  :  0;}
		R_Command *begin() {return valid(next);}
		R_Command *end() {return this;}

		R_Command *takeOut(R_Command *command)
		{	command=valid(command);
			if (command)
			{	command->next->prev = command->prev;
				command->prev->next = command->next;
			}
			return command;
		}
		void clear();
		void free();

		R_CommandList &push_back(R_Command *command)
		{	R_Command *here=prev;
			if (!here)
				here=next;
			command->prev=here;
			command->next=here->next;
			here->next->prev=command;
			here->next=command;
			return *this;
		}
};

class CommandListIterator // This might be templated in the future (but I don`t like templates so very much, so I probably won`t do it)
{
	R_Command *here;
	R_CommandList *list;
	public:
		static const CommandListIterator aCommandListIterator;
		operator_new;

		CommandListIterator(R_CommandList *list_): here(0), list(list_) {}
		R_Command *operator++() {here = list->valid( here ?   here->next  :  list->next ); return here;}
		R_Command *operator--() {here = list->valid( here ?   here->prev  :  list->prev ); return here;}
		R_Command *operator++(int) {here = list->valid( here ?   here->next  :  list->next ); return here;}
		R_Command *operator--(int) {here = list->valid( here ?   here->prev  :  list->prev ); return here;}
		R_Command *operator=(R_Command *command) {here=command; return here;}
		R_Command *operator->() {return here;}
		R_Command *operator*() {return here;}
		operator unsigned int() {return (unsigned int)(here);}
		bool operator!=(R_Command *command) {return command!=(R_Command *)this;}

	friend TextInputStream &operator>>(TextInputStream &istr, CommandListIterator &i);
	friend TextOutputStream &operator<<(TextOutputStream &ostr, CommandListIterator &i);
	friend DataInputStream &operator>>(DataInputStream &istr, CommandListIterator &i);
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, CommandListIterator &i);
};

class Recorder: public R_CommandList
{
	friend class R_MoveTo;
	public:
		unsigned int refCount, items;
		bool recording, pendown;
		bool /*volatile*/ inReplay, restartFlag, stop_play, stop_confirm;
		bool inload, insave;
		int action, tabs;
		R_MoveTo *lastMove;
		Graph *graph;
		Screen_ *screen;
		Turtle *turtle, *recturtle;

		static PtrToInt_ *ptrToInt;
		static IntToPtr_ *intToPtr;

		static const int tableSize;

		static const Recorder aRecorder;
		operator_new;

		Recorder();
		Recorder(Graph *graph_);
		void free();

		void off() {recording=false;}
		void  on() {recording=true;}

		void clear();
		void move();

		void playAgain(), stop(), updateView(Turtle &turtle);

		static DWORD WINAPI 
			playCommands(LPVOID lpv);

		wxString rgb(Color c);

		void
			save(const wxString name),
			load(const wxString name),
			saveText(const wxString name),
			loadText(const wxString name),
			writePS(const wxString name);

		R_CommandList &operator<<(R_Command *command);

	friend TextInputStream &operator >>(TextInputStream &istr, Recorder &recorder);
	friend TextOutputStream &operator <<(TextOutputStream &ostr, Recorder &recorder);
	friend DataInputStream &operator >>(DataInputStream &istr, Recorder &recorder);
	friend wxDataOutputStream &operator <<(wxDataOutputStream &ostr, Recorder &recorder);
};

class R_Graphic: public R_Command
{
	Turtle *recturtle, *saveturtle;
	double scale;
	Recorder *recorder;
	public:
		static const R_Graphic aGraphic;
		operator_new;

		R_Graphic() {type=GRAPHIC;}
		R_Graphic(Turtle &turtle, Recorder *recorder_, double scale);
		void free();
		void play(Recorder &rec);
		void setPtrs(IntToPtr_ &intToPtr);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_MoveTo: public R_Command
{
	public:
		Point point;

		static const R_MoveTo aMoveTo;
		operator_new;

		R_MoveTo() {type=MOVETO;}
		R_MoveTo(Turtle &turtle, Recorder &recorder);
		R_MoveTo(Turtle &turtle, Recorder &recorder, Point point_);
		R_MoveTo(Turtle &turtle, Recorder &recorder, double x, double y);

		virtual void init(Turtle &turtle, Recorder &recorder, 
			Point point_)
		{	type=MOVETO;
			point=point_;
			recorder.lastMove=this;
		}
		virtual void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_LineTo: public R_MoveTo
{
//	PointMatrix u;
	public:
		static const R_LineTo aLineTo;
		operator_new;

		R_LineTo() {type=LINETO;}
		R_LineTo(Turtle &turtle, Recorder &recorder);
		R_LineTo(Turtle &turtle, Recorder &recorder, Point point_);
		R_LineTo(Turtle &turtle, Recorder &recorder, double x, double y);

		void init(Turtle &turtle, Recorder &recorder, Point point_)
		{	type=LINETO;
			point=point_;
//			u=turtle.u;
			recorder.lastMove=(R_MoveTo*)this;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_LineTo3d: public R_MoveTo
{
//	PointMatrix u;
	public:
		static const R_LineTo3d aLineTo3d;
		operator_new;

		R_LineTo3d() {}
		R_LineTo3d(Turtle &turtle, Recorder &recorder);
		R_LineTo3d(Turtle &turtle, Recorder &recorder, Point point_);
		R_LineTo3d(Turtle &turtle, Recorder &recorder, double x, double y);

		void init(Turtle &turtle, Recorder &recorder, Point point_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setMatrix: public R_Command
{
	Point basex,basey,basez,trans;
	public:
		static const R_setMatrix asetMatrix;
		operator_new;

		R_setMatrix() {}
		R_setMatrix(Point basex_, Point basey_, Point basez_, Point trans_)
		{	type=SETMATRIX; 
			basex=basex_;	
			basey=basey_;
			basez=basez_;
			trans=trans_;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setMatrix4x4: public R_Command
{
	double m[16];
	public:
		static const R_setMatrix4x4 asetMatrix4x4;
		operator_new;

		R_setMatrix4x4() {}
		R_setMatrix4x4(double m_[16])
		{	type=SETMATRIX4X4; 
			for (int i=0; i<16; i++)
				m[i]=m_[i];
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_pushMatrix: public R_Command
{
	public:
		static const R_pushMatrix apushMatrix;
		operator_new;

		R_pushMatrix() 
		{	type=PUSHMATRIX;
		}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_popMatrix: public R_Command
{
	public:
		static const R_popMatrix apopMatrix;
		operator_new;

		R_popMatrix() 
		{	type=POPMATRIX;
		}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setPixel: public R_MoveTo
{
	Color color;
	public:
		static const R_setPixel asetPixel;
		operator_new;

		R_setPixel() {}
		R_setPixel(Point &p, Color c) 
		{	type=SETPIXEL; point=p; color=c;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_PolyStart: public R_Command
{
	Point normal, p1,p2,p3, t1,t2,t3;
	Color c1,c2,c3;
	public:
		static const R_PolyStart aPolyStart;
		operator_new;

		R_PolyStart() {}
		R_PolyStart(Turtle &turtle, Point &normal_, Point &p3_, Point &t3_, Color c3_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_PolyPoint: public R_Command
{
	Point p, t;
	Color c;
	public:
		static const R_PolyPoint aPolyPoint;
		operator_new;

		R_PolyPoint() {}
		R_PolyPoint(Point &p_, Point &t_, Color c_)
		{	type=POLYPOINT; 
			p=p_;
			t=t_;
			c=c_;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_PolyEnd: public R_Command
{
	public:
		static const R_PolyEnd aPolyEnd;
		operator_new;

		R_PolyEnd()
		{	type=POLYEND; 
		}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_glBegin: public R_Command
{
	unsigned int prim;
	Point normal;
	public:
		static const R_glBegin aglBegin;
		operator_new;

		R_glBegin() {}
		R_glBegin(unsigned int prim_, Point normal_):
			prim(prim_), normal(normal_)
		{	type=GLBEGIN;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_TessPoint: public R_Command
{
	Point p, t;
	Color c;
	public:
		static const R_TessPoint aTessPoint;
		operator_new;

		R_TessPoint() {}
		R_TessPoint(Point &p_, Point &t_, Color c_)
		{	type=TESSPOINT; 
			p=p_;
			t=t_;
			c=c_;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_QuadStripStart: public R_Command
{
	Point normal,onormal, 
		p1,t1, p2,t2, p3,t3, p4,t4;
	Color c1,c2,c3,c4;
	public:
		static const R_QuadStripStart aQuadStripStart;
		operator_new;

		R_QuadStripStart() {}
		R_QuadStripStart(Point &normal_,Point &onormal_, 
			Color &c1_, Point &p1_, Point &t1_, 
			Color &c2_, Point &p2_, Point &t2_,
			Color &c3_, Point &p3_, Point &t3_,
			Color &c4_, Point &p4_, Point &t4_)
		{	type=QUADSTRIPSTART; 
			normal=normal_;
			onormal=onormal_;
			c1=c1_; p1=p1_; t1=t1_;
			c2=c2_; p2=p2_; t2=t2_;
			c3=c3_; p3=p3_; t3=t3_;
			c4=c4_; p4=p4_; t4=t4_;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_QuadStrip: public R_Command
{
	Point normal,onormal, p3,t3, p4,t4;
	Color c3,c4;
	public:
		static const R_QuadStrip aQuadStrip;
		operator_new;

		R_QuadStrip() {}
		R_QuadStrip(Point &normal_, Point &onormal_, 
			Color &c3_, Point &p3_, Point &t3_,
			Color &c4_, Point &p4_, Point &t4_)
		{	type=QUADSTRIP; 
			normal=normal_;
			onormal=onormal_;
			c3=c3_; p3=p3_; t3=t3_;
			c4=c4_; p4=p4_; t4=t4_;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_QuadStripEnd: public R_Command
{
	public:
		static const R_QuadStripEnd aQuadStripEnd;
		operator_new;

		R_QuadStripEnd()
		{	type=QUADSTRIPEND; 
		}
		void play(Recorder &rec);
};

class R_bool: public R_Command
{
	public:
		bool on;

		static const R_bool abool;
		operator_new;

		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_fill: public R_Command
{
	public:
		int mode;
		static const R_fill afill;
		operator_new;

		R_fill() {}
		R_fill(int mode_)
		{	type=FILL; 
			mode=mode_;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_fillRect: public R_Command
{
	Point p1, p2, ori;
	public:
		static const R_fillRect afillRect;
		operator_new;

		R_fillRect() {}
		R_fillRect(Turtle &turtle, Point p1_, Point p2_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_fillEllipse: public R_Command
{
	Point radius, ori;
	public:
		static const R_fillEllipse afillEllipse;
		operator_new;

		R_fillEllipse() {}
		R_fillEllipse(Turtle &turtle, Point radius_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_Sphere: public R_Command
{
	double radius;
	int slices, stacks;
	Point ori;
	public:
		static const R_Sphere aSphere;
		operator_new;

		R_Sphere() {}

		R_Sphere(Turtle &turtle, double radius_, int slices_, int stacks_)
		{	type=SPHERE; 
			radius=radius_;
			slices=slices_;
			stacks=stacks_;
			ori=Point(
				turtle.roll(), 
				turtle.pitch(), 
				turtle.getHeading());
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_Ellipsoid: public R_Command
{
	Point radius, ori;
	int slices, stacks;
	public:
		static const R_Ellipsoid aEllipsoid;
		operator_new;

		R_Ellipsoid() {}
		R_Ellipsoid(Turtle &turtle, Point radius_, int slices_, int stacks_)
		{	type=ELLIPSOID; 
			radius=radius_;
			slices=slices_;
			stacks=stacks_;
			ori=Point(
				turtle.roll(), 
				turtle.pitch(), 
				turtle.getHeading());
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_partialEllipsoid: public R_Command
{
	Point radius, ori;
	double slicesstart, slicesend, stacksstart, stacksend;
	int slices, stacks;
	public:
		static const R_partialEllipsoid apartialEllipsoid;
		operator_new;

		R_partialEllipsoid() {}
		R_partialEllipsoid(Turtle &turtle, Point radius_, 
						double slicesstart_, double slicesend_, int slices_,
						double stacksstart_, double stacksend_, int stacks_)
		{	type=PARTIALELLIPSOID; 
			radius=radius_;
			slicesstart=slicesstart_;
			slicesend=slicesend_;
			slices=slices_;
			stacksstart=stacksstart_;
			stacksend=stacksend_;
			stacks=stacks_;
			ori=Point(
				turtle.roll(), 
				turtle.pitch(), 
				turtle.getHeading());
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_fillPie: public R_Command
{
	double heading, angle;
	Point radius;
	public:
		static const R_fillPie afillPie;
		operator_new;

		R_fillPie() {}
		R_fillPie(Turtle &turtle, double angle_, Point radius_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_Cylinder: public R_Command
{
	Point ori;
	double length, baseradius, topradius;
	public:
		static const R_Cylinder aCylinder;
		operator_new;

		R_Cylinder() {}
		R_Cylinder(Turtle &turtle, double length_, double baseradius_, double topradius_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_PartCylinder: public R_Command
{
	Point ori;
	double length, baseradius, topradius, startangle, endangle, scrunch;
	int slices;
	public:
		static const R_PartCylinder aPartCylinder;
		operator_new;

		R_PartCylinder() {}
		R_PartCylinder(Turtle &turtle, double length_, double baseradius_, double topradius_,
			double startangle_, double endangle_, double scrunch_, int slices_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_Perspective: public R_Command
{
	public:
		static const R_Perspective aPerspective;
		operator_new;

		R_Perspective() {type=PERSPECTIVE;}
		void play(Recorder &rec);
};

class R_unPerspective: public R_Command
{
	public:
		static const R_unPerspective aunPerspective;
		operator_new;

		R_unPerspective() {type=UNPERSPECTIVE;}
		void play(Recorder &rec);
};

class R_ortho: public R_Command
{
	public:
		static const R_ortho aortho;
		operator_new;

		R_ortho() {type=ORTHO;}
		void play(Recorder &rec);
};

class R_setEye: public R_Command
{
	Point eye, center, up;
	public:
		static const R_setEye asetEye;
		operator_new;

		R_setEye() {type=SETEYE;}
		R_setEye(Point &eye_, Point &center, Point &up);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_wrap: public R_Command
{
	public:
		static const R_wrap awrap;
		operator_new;

		R_wrap() {type=WRAP;}
		void play(Recorder &rec);
};

class R_mode: public R_Command
{
	public:
		Turtle::Mode mode;
		static const R_mode amode;
		operator_new;

		R_mode() {type=MODE;}
		R_mode(Turtle &turtle);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Color: public R_Command
{
	public:
		Color color;
		static const R_Color aColor;
		operator_new;

		R_Color() {}
		void play(Recorder &rec) {}
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setScreenColor: public R_Color
{
	public:
		static const R_setScreenColor asetScreenColor;
		operator_new;

		R_setScreenColor() {type=SETSCREENCOLOR;}
		R_setScreenColor(Color _color) {type=SETSCREENCOLOR; color=_color;}
		void play(Recorder &rec);
};

class R_Material: public R_Color
{
		int gltype;
	public:
		static const R_Material aMaterial;
		operator_new;

		R_Material() {type=MATERIAL;}
		R_Material(int _gltype, Color _color) 
		{	type=MATERIAL; gltype=_gltype; color=_color;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Light: public R_Color
{
		int gltype;
	public:
		static const R_Light aLight;
		operator_new;

		R_Light() {type=LIGHT;}
		R_Light(int _gltype, Color _color) 
		{	type=LIGHT; gltype=_gltype; color=_color;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setPenColor: public R_Color
{
	public:
		static const R_setPenColor asetPenColor;
		operator_new;

		R_setPenColor() {type=SETPENCOLOR;}
		R_setPenColor(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setFloodColor: public R_Color
{
	public:
		static const R_setFloodColor asetFloodColor;
		operator_new;

		R_setFloodColor() {type=SETFLOODCOLOR;}
		R_setFloodColor(Turtle &turtle);
		void play(Recorder &rec); 
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setShadowColor: public R_Color
{
	public:
		static const R_setShadowColor asetShadowColor;
		operator_new;

		R_setShadowColor() {type=SETSHADOWCOLOR;}
		R_setShadowColor(Turtle &turtle);
		void play(Recorder &rec); 
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_Shininess: public R_Command
{
	float s;
	public:
		static const R_Shininess aShininess;
		operator_new;

		R_Shininess() {type=SHININESS;}
		R_Shininess(float s_) 
		{	type=SHININESS; s=s_;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setPenDown: public R_bool
{
	public:
		static const R_setPenDown asetPenDown;
		operator_new;

		R_setPenDown() {type=SETPENDOWN;}
		R_setPenDown(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setPolySmooth: public R_bool
{
	public:
		static const R_setPolySmooth asetPolySmooth;
		operator_new;

		R_setPolySmooth() {type=SETPOLYSMOOTH;}
		R_setPolySmooth(bool smooth) {type=SETPOLYSMOOTH; on=smooth;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setLineSmooth: public R_bool
{
	public:
		static const R_setLineSmooth asetLineSmooth;
		operator_new;

		R_setLineSmooth() {type=SETLINESMOOTH;}
		R_setLineSmooth(bool smooth) {type=SETLINESMOOTH; on=smooth;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setPointSmooth: public R_bool
{
	public:
		static const R_setPointSmooth asetPointSmooth;
		operator_new;

		R_setPointSmooth() {type=SETPOINTSMOOTH;}
		R_setPointSmooth(bool on_) {type=SETPOINTSMOOTH; on=on_;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setDepthTest: public R_bool
{
	public:
		static const R_setDepthTest asetDepthTest;
		operator_new;

		R_setDepthTest() {type=SETDEPTHTEST;}
		R_setDepthTest(bool on_) {type=SETDEPTHTEST; on=on_;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setLighting: public R_bool
{
	public:
		static const R_setLighting asetLighting;
		operator_new;

		R_setLighting() {type=SETLIGHTING;}
		R_setLighting(bool on_) {type=SETLIGHTING; on=on_;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setDither: public R_bool
{
	public:
		static const R_setDither asetDither;
		operator_new;

		R_setDither() {type=SETDITHER;}
		R_setDither(bool on_) {type=SETDITHER; on=on_;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setTextured: public R_bool
{
	public:
		static const R_setTextured asetTextured;
		operator_new;

		R_setTextured() {type=SETTEXTURED;}
		R_setTextured(bool on_) {type=SETTEXTURED; on=on_;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setTextureFont: public R_bool
{
	public:
		static const R_setTextureFont asetTextureFont;
		operator_new;

		R_setTextureFont() {type=SETTEXTUREFONT;}
		R_setTextureFont(bool on_) {type=SETTEXTUREFONT; on=on_;}
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setPenRLE: public R_bool	// turtle->pen.roundLineEnds
{
	public:
		static const R_setPenRLE asetPenRLE;
		operator_new;

		R_setPenRLE() {type=SETPENRLE;}
		R_setPenRLE(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setCyLines: public R_bool
{
	public:
		static const R_setCyLines asetCyLines;
		operator_new;

		R_setCyLines() {type=SETCYLINES;}
		R_setCyLines(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setShadows: public R_bool
{
	public:
		static const R_setShadows asetShadows;
		operator_new;

		R_setShadows() {type=SETSHADOWS;}
		R_setShadows(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setFogEnabled: public R_bool
{
	public:
		static const R_setFogEnabled asetFogEnabled;
		operator_new;

		R_setFogEnabled() {type=SETFOGENABLED;}
		R_setFogEnabled(bool fog);
		void play(Recorder &rec);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setPenMode: public R_Command
{
	Pen::WinMode penMode;
	Pen::Mode internalMode;
	public:
		static const R_setPenMode asetPenMode;
		operator_new;

		R_setPenMode() {type=SETPENMODE;}
		R_setPenMode(Turtle &turtle);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setPenSize: public R_Command
{
	Point penSize;
	public:
		static const R_setPenSize asetPenSize;
		operator_new;

		R_setPenSize() {type=SETPENSIZE;}
		R_setPenSize(Turtle &turtle);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setPenPattern: public R_Command
{
	int penPattern;
	public:
		static const R_setPenPattern asetPenPattern;
		operator_new;

		R_setPenPattern() {type=SETPENPATTERN;}
		R_setPenPattern(Turtle &turtle);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setDepthFunc: public R_Command
{
	int func;
	public:
		static const R_setDepthFunc asetDepthFunc;
		operator_new;

		R_setDepthFunc() {type=SETDEPTHFUNC;}
		R_setDepthFunc(int func_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setPolygonMode: public R_Command
{
	int mode;
	public:
		static const R_setPolygonMode asetPolygonMode;
		operator_new;

		R_setPolygonMode() {type=SETPOLYGONMODE;}
		R_setPolygonMode(int mode_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setColorLogicOp: public R_Command
{
	int mode;
	public:
		static const R_setColorLogicOp asetColorLogicOp;
		operator_new;

		R_setColorLogicOp() {type=SETCOLORLOGICOP;}
		R_setColorLogicOp(int mode_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_setPointSize: public R_Command
{
	double size;
	public:
		static const R_setPointSize asetPointSize;
		operator_new;

		R_setPointSize() {type=SETPOINTSIZE;}
		R_setPointSize(double size_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Texture: public R_Command
{
	public:
		bool set;
		int x,y, texi;
		ColorP pixels;
		TextureP tex;

		static int maxnr;
		static PairP textures;

		static const R_Texture aTexture;
		operator_new;

		R_Texture() {type=TEXTURE; pixels=0; set=false;}
		R_Texture(TextureP tex_);
		void play(Recorder &rec);
		void setPtrs(IntToPtr_ &intToPtr);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_BitPaste: public R_Command
{
	public:
		bool set;
		BitmapP bmp;

		static PairP bmps;

		static const R_BitPaste aBitPaste;
		operator_new;

		R_BitPaste() {type=BITPASTE; set=false;}
		R_BitPaste(BitmapP bmp_);
		void play(Recorder &rec);
		void setPtrs(IntToPtr_ &intToPtr);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_TessWindingRule: public R_Command
{
	int rule;
	public:
		static const R_TessWindingRule aTessWindingRule;
		operator_new;

		R_TessWindingRule() {type=TESSWINDINGRULE;}
		R_TessWindingRule(int rule_) {type=TESSWINDINGRULE; rule=rule_;}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Label: public R_Command
{
	wxString str;
	Point ori;
	Font_ font;
	public:
		static const R_Label aLabel;
		operator_new;

		R_Label() {type=LABEL;}
		R_Label(Turtle &turtle, wxString str_);
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};

class R_setLightPos: public R_MoveTo
{
	public:
		static const R_setLightPos asetLightPos;
		operator_new;

		R_setLightPos() {}
		R_setLightPos(Point &p) 
		{	type=SETLIGHTPOS; point=p;
		}
		void play(Recorder &rec);
		TextInputStream  &read(TextInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		DataInputStream  &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(TextOutputStream &ostr, Recorder &recorder);
};
