//___________________________________________________________________
//
//                           Class  Slider
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/slider.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "File.h"
#include "String.h"
#include "Object.h"
#include "Table.h"
#include "LogoObject.h"
#include "Int.h"
#include "Pair.h"
#include "WindowPair.h"
#include "NodePStack.h"
#include "Point.h"

#include "Event.h"
#include "Slider.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Error.h"
#include "Graph.h"
#include "Color.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MySlider : public wxSlider
{
public:
	MySlider(wxWindow* parent, wxWindowID id, 
		int value, int minValue, int maxValue, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = _("Slider"));

	// MSW sends extra kill focus event
	virtual ~MySlider()
	{
		if (parent) 
			parent->slider = 0;
		parent=0;
	}

	Slider *parent;

	virtual void 
		OnChar	(wxKeyEvent &event),         // pass chars to Slider
		OnKeyDown(wxKeyEvent &event),
		OnKeyUp	(wxKeyEvent &event),
		OnLeftDown		(wxMouseEvent &event),
		OnRightDown		(wxMouseEvent &event),
		OnMiddleDown	(wxMouseEvent &event),
		OnLeftUp  		(wxMouseEvent &event),
		OnRightUp  		(wxMouseEvent &event),
		OnMiddleUp  	(wxMouseEvent &event),
		OnLeftDClick	(wxMouseEvent &event),
		OnRightDClick	(wxMouseEvent &event),
		OnMiddleDClick	(wxMouseEvent &event),
		OnMotion			(wxMouseEvent &event),
		OnEnterWindow	(wxMouseEvent &event),
		OnLeaveWindow	(wxMouseEvent &event),
		OnMouseWheel	(wxMouseEvent &event),
		OnMouse			(wxMouseEvent &event),
		OnScroll       (wxScrollEvent &event);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MySlider,wxSlider)
	EVT_CHAR					(MySlider::OnChar		)
	EVT_KEY_DOWN			(MySlider::OnKeyDown	)
	EVT_KEY_UP				(MySlider::OnKeyUp		)
	EVT_SCROLL				(MySlider::OnScroll)
	EVT_LEFT_DOWN			(MySlider::OnLeftDown		)
	EVT_RIGHT_DOWN			(MySlider::OnRightDown		)
	EVT_MIDDLE_DOWN		(MySlider::OnMiddleDown	)
	EVT_LEFT_UP				(MySlider::OnLeftUp			)
	EVT_RIGHT_UP			(MySlider::OnRightUp		)
	EVT_MIDDLE_UP			(MySlider::OnMiddleUp		)
	EVT_LEFT_DCLICK		(MySlider::OnLeftDClick	)	
	EVT_RIGHT_DCLICK		(MySlider::OnRightDClick	)
	EVT_MIDDLE_DCLICK		(MySlider::OnMiddleDClick)
	EVT_MOTION				(MySlider::OnMotion			)
	EVT_ENTER_WINDOW		(MySlider::OnEnterWindow	)
	EVT_LEAVE_WINDOW		(MySlider::OnLeaveWindow	)
	EVT_MOUSEWHEEL			(MySlider::OnMouseWheel	)
	EVT_MOUSE_EVENTS		(MySlider::OnMouse			)
END_EVENT_TABLE()

MySlider::MySlider(wxWindow *parent_, wxWindowID id,
						int value_, int minValue_, int maxValue_,
						const wxPoint &pos, const wxSize &size,
						long style,
						const wxValidator& validator,
						const wxString &name):
wxSlider(parent_, id, value_, minValue_, maxValue_, pos, size, style, validator, name)
{
	parent = (Slider*)parent_;
}

void MySlider::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MySlider::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MySlider::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}

void MySlider::OnScroll(wxScrollEvent &event)
{
	if (parent) 
		parent->OnScroll(event);
	event.Skip();
}

#define MySliderOnMouse(NAME)								\
void MySlider::On##NAME(wxMouseEvent &event)			\
{																	\
	if (parent)													\
		parent->On##NAME(event);							\
}

MySliderOnMouse(LeftDown)
MySliderOnMouse(RightDown)
MySliderOnMouse(MiddleDown)

MySliderOnMouse(LeftUp)
MySliderOnMouse(RightUp)
MySliderOnMouse(MiddleUp)

MySliderOnMouse(LeftDClick)
MySliderOnMouse(RightDClick)
MySliderOnMouse(MiddleDClick)
MySliderOnMouse(Motion)
MySliderOnMouse(EnterWindow)
MySliderOnMouse(LeaveWindow)
MySliderOnMouse(MouseWheel)
MySliderOnMouse(Mouse)

BEGIN_EVENT_TABLE(Slider, wxControl)
	EVT_SCROLL	(Slider::OnScroll)
	EVT_LEFT_DOWN			(Slider::OnLeftDown		)
	EVT_RIGHT_DOWN			(Slider::OnRightDown		)
	EVT_MIDDLE_DOWN		(Slider::OnMiddleDown	)
	EVT_LEFT_UP				(Slider::OnLeftUp			)
	EVT_RIGHT_UP			(Slider::OnRightUp		)
	EVT_MIDDLE_UP			(Slider::OnMiddleUp		)
	EVT_LEFT_DCLICK		(Slider::OnLeftDClick	)	
	EVT_RIGHT_DCLICK		(Slider::OnRightDClick	)
	EVT_MIDDLE_DCLICK		(Slider::OnMiddleDClick)
	EVT_MOTION				(Slider::OnMotion			)
	EVT_ENTER_WINDOW		(Slider::OnEnterWindow	)
	EVT_LEAVE_WINDOW		(Slider::OnLeaveWindow	)
	EVT_MOUSEWHEEL			(Slider::OnMouseWheel	)
	EVT_MOUSE_EVENTS		(Slider::OnMouse			)
END_EVENT_TABLE()

int Slider::defaultStyle=wxSL_VERTICAL | wxSL_AUTOTICKS | wxSL_LABELS;

Slider::Slider(SliderNodeP sliderNode_, int style):
label(0), slider(0), bestSize(-999, -999),
value(0), minValue(0), maxValue(0)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=sliderNode_;
	wxWindow *parent=(wxWindow*)console;
	if (WindowNodeP(*d(windowNode))->parent)
		parent=WindowNodeP(*d(windowNode))->parent->window;
	create(parent, -1, 
			0,0,0, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			wxDefaultValidator, 
			_("Slider"));
}

Slider::Slider(SliderNodeP sliderNode_, 
			wxWindow* parent, wxWindowID id, 
			int value_, int minValue_, int maxValue_, 
			PairP onScroll_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), slider(0), bestSize(-999, -999),
value(value_), minValue(minValue_), maxValue(maxValue_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=sliderNode_;
	*(PairP*)(d(onScroll))=onScroll_;
	ThreadP th=Thread::This();
	*d(uonScroll) = th->lexufun;
	*d(tonScroll) = th->thisobj;
	create(parent, -1, 
			value, minValue, maxValue,
			pos, 
			size, 
			style, 
			wxDefaultValidator, 
			name);
}

bool Slider::create(wxWindow* parent, wxWindowID id, 
			int value_, int minValue_, int maxValue_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
	{	
		label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxSize(width, -1), wxALIGN_CENTER, _("staticText"));
		if (!label)
		{	error(OUT_OF_MEM); return false;
		}
	}else
		label=0;

	if (label)
		slider = new MySlider(this, id, value, minValue, maxValue,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		slider = new MySlider(this, id, value, minValue, maxValue,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);
	if (!slider)
	{	error(OUT_OF_MEM); return false;
	}
	DoSetSize(pos.x, pos.y, width, height);
	SetInitialSize(wxSize(width, height));
	if (label)
		label->SetSize(width, label->GetSize().GetHeight());
	return true;
}

void Slider::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
	{	label->GetSize(&labelwidth, &labelheight);
		label->SetSize(width, labelheight);
	}
	if (slider)   
		slider->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize Slider::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxSlider *t=new wxSlider((wxWindow*)this, wxID_ANY, 
			value,minValue,maxValue, wxPoint(0,0), wxDefaultSize, GetWindowStyle());
		if (!t)
		{	error(OUT_OF_MEM); return wxDefaultSize;
		}
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

Slider::~Slider()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
}

void Slider::OnScroll(wxScrollEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventSlider(this);
	if throwing return;
	*th->eventFiFo << evt;
	event.Skip();
}

#define SliderOnKey(NAME)										\
void Slider::On##NAME(wxKeyEvent &event)					\
{																		\
	if (*d(on##NAME))												\
	{	ThreadP th=Thread::This();								\
		EventP evt=new EventWindow##NAME((Window_*)this, event);	\
		if throwing return;										\
		*th->eventFiFo << evt;									\
	}else																\
		event.Skip();												\
}

SliderOnKey(Char)
SliderOnKey(KeyDown)
SliderOnKey(KeyUp)

#define EVENT_ON_MOUSE(NAME)								\
void Slider::On##NAME(wxMouseEvent &event)			\
{																	\
	if (*d(on##NAME))											\
	{	ThreadP th=Thread::This();							\
		EventP evt=new EventWindow##NAME((Window_*)this, event);	\
		if throwing return;									\
		*th->eventFiFo << evt;								\
	}else															\
		event.Skip();											\
}

EVENT_ON_MOUSE(LeftDown)
EVENT_ON_MOUSE(RightDown)
EVENT_ON_MOUSE(MiddleDown)

EVENT_ON_MOUSE(LeftUp)
EVENT_ON_MOUSE(RightUp)
EVENT_ON_MOUSE(MiddleUp)

EVENT_ON_MOUSE(LeftDClick)
EVENT_ON_MOUSE(RightDClick)
EVENT_ON_MOUSE(MiddleDClick)
EVENT_ON_MOUSE(Motion)
EVENT_ON_MOUSE(EnterWindow)
EVENT_ON_MOUSE(LeaveWindow)
EVENT_ON_MOUSE(MouseWheel)
EVENT_ON_MOUSE(Mouse)

int Slider::GetValue() const
{
	return slider->GetValue();
}

wxString Slider::wxstring()
{
	wxString str;
	str << _("Slider") << " " 
		<< slider->GetMin() << ' ' << GetValue() << ' ' << slider->GetMax() << ' ' 
		<< '[' << GetName() << "] ";
	if (onScroll)
		str << (*d(onScroll))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &Slider::read(DataInputStream &istr) 
{	
	int value, min, max;
	wxPoint pos;
	wxSize size;
	wxString name;
	
	istr >> min >> value >> max >> name 
		>> (int&)*d(onScroll) >> (int&)*d(uonScroll) >> (int&)*d(tonScroll)
		>> pos.x >> pos.y >> size.x >> size.y;

	slider->SetRange(min, max);
	slider->SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	label=new wxStaticText(this, -1, name, 
		wxDefaultPosition, wxDefaultSize, 
		0, _("staticText"));
	if (!label)
	{	error(OUT_OF_MEM);
	}
	return istr;
}
wxDataOutputStream &Slider::write(wxDataOutputStream &ostr) 
{
	ostr
		<< slider->GetMin() << GetValue() << slider->GetMax() << GetName() 
		<< GC_::ptrToInt->find(*d(onScroll))
		<< GC_::ptrToInt->find(*d(uonScroll))
		<< GC_::ptrToInt->find(*d(tonScroll))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;
	return ostr;
}

PairP SliderNode::onScroll()
{
	if (slider)
		return PairP(*slider->d(Slider::onScroll));
	return 0;
}

ObjectP SliderNode::uonScroll()
{
	if (slider)
		return ObjectP(*slider->d(Slider::uonScroll));
	return 0;
}

LogoObjectP SliderNode::tonScroll()
{
	if (slider)
		return LogoObjectP(*slider->d(Slider::tonScroll));
	return 0;
}

void SliderNode::setonScroll(NodeP nd) 
{
	ThreadP th=Thread::This();
	if (slider)
	{	*slider->d(Slider::onScroll) = PairP(nd);  check_oldyoung(nd);
		*slider->d(Slider::uonScroll) = th->lexufun;
		*slider->d(Slider::tonScroll) = th->thisobj;
	}
}

SliderNode::SliderNode(): 
	WindowNode(SLIDER), 
	slider((SliderP&)window)
{}

SliderNode::SliderNode(WindowNodeP parent_): 
	WindowNode(SLIDER, parent_), 
	slider((SliderP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void SliderNode::free() 
{	
	if (slider) 
	{	ThreadP th=Thread::This();
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::tSlider:
				if (((EventSlider*)ev)->slider==slider)
					((EventSlider*)ev)->slider=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		*slider->d(Slider::windowNode)=0;
//		GUI_ENTER
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		slider->Destroy(); 
		slider=0;
//		GUI_LEAVE
	}
}
void SliderNode::inc_gc() 
{	fifo << onScroll() << uonScroll() << tonScroll();
}
int SliderNode::youngs() 
{	return gmark(onScroll())+gmark(uonScroll())+gmark(tonScroll());
}
int SliderNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return 
		 gmarkPtr( onScroll(), ptrToInt)
		+gmarkPtr(uonScroll(), ptrToInt)
		+gmarkPtr(tonScroll(), ptrToInt);
}
void SliderNode::check_oldyoungs() 
{	check_oldyoung(onScroll());
	check_oldyoung(uonScroll());
	check_oldyoung(tonScroll());
}

DataInputStream &SliderNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode >> style;
		slider=new Slider(SliderNodeP(this), style);
		if (!slider)
		{	error(OUT_OF_MEM); return istr;
		}
		slider->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &SliderNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (slider)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent) 
			<< style;
		slider->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void SliderNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (slider)
	{	setonScroll((Pair*)intToPtr.find(int(slider->onScroll)));
		*slider->d(Slider::uonScroll)=(    Object*)intToPtr.find(int(*slider->d(Slider::uonScroll)));
		*slider->d(Slider::tonScroll)=(LogoObject*)intToPtr.find(int(*slider->d(Slider::tonScroll)));
	}
}

void SliderNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	SliderP s=slider, 
		s2=new Slider(SliderNodeP(this),
			p, -1, s->GetValue(), s->slider->GetMin(), s->slider->GetMax(),
			PairP(*s->d(Slider::onScroll)), s->GetPosition(), s->GetSize(), style, 
			wxDefaultValidator, s->GetName());
	if (!s2)
	{	error(OUT_OF_MEM); return;
	}
	slider=s2;
	setonScroll(*s->d(Slider::onScroll));
	*slider->d(Slider::uonScroll)=*s->d(Slider::uonScroll);
	*slider->d(Slider::tonScroll)=*s->d(Slider::tonScroll);
	delete s;
	slider=s2;
}

StringP SliderNode::string()
{
	if (slider)
		return new String(slider->wxstring());
	return _Null_Word->name();
}

wxString SliderNode::wxstring()
{
	if (slider)
		return slider->wxstring();
	return wxEmptyString;
}

void SliderNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lSlider(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString name=_T(_("Slider"));
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	int minValue;
	{	int v=arg[3]->a_int(3);
		if throwing return Unbound;
		minValue=v;	
	}
	int value;
	{	int v=arg[4]->a_int(4);
		if throwing return Unbound;
		value=v;	
	}
	int maxValue;
	{	int v=arg[5]->a_int(5);
		if throwing return Unbound;
		maxValue=v;	
	}
	PairP onScroll=0;
	if (exist(6))
	{	PairP os=arg[6]->a_runnable(6);
		if throwing return Unbound;
		onScroll=os;
	}
	long style = Slider::defaultStyle;
	if (exist(7))
	{	int s=arg[7]->a_int(7);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(8))
	{	PairP p=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(9))
	{	PairP s=arg[9]->a_int_vector(9);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
//	GUI_ENTER
	SliderNodeP sliderNode=new SliderNode(parentNode);
	if throwing return Unbound;
	SliderP slider=new Slider(sliderNode,
			parent, id, value, minValue, maxValue, 
			onScroll, pos, size, style, wxDefaultValidator, name);
	if (!slider)
	{	error(OUT_OF_MEM); return Unbound;
	}
	sliderNode->slider=slider;
	sliderNode->style=style;
//	GUI_LEAVE
	return sliderNode;
}

NodeP lSliderDestroy(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=slider->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(slider));
			p->check_oldyoung(p->children);
		}
	}
	slider->slider->Destroy();
	slider->slider=0;
//	GUI_LEAVE
	return Unbound;
}

NodeP lSliderEnable(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	slider->slider->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lSliderOnScroll(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
//	GUI_ENTER
	slider->setonScroll(function);
//	GUI_LEAVE
	return Unbound;
}

NodeP lSliderSetLineSize(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
	slider->slider->slider->SetLineSize(value);
	return Unbound;
}

NodeP lSliderSetPageSize(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
	slider->slider->slider->SetPageSize(value);
	return Unbound;
}

NodeP lSliderSetRange(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int min=arg[2]->a_int(2);
	if throwing return Unbound;
	int max=arg[3]->a_int(3);
	if throwing return Unbound;
//	GUI_ENTER
	slider->slider->slider->SetRange(min, max);
//	GUI_LEAVE
	return Unbound;
}

NodeP lSliderSetValue(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	slider->slider->slider->SetValue(value);
//	GUI_LEAVE
	return Unbound;
}

NodeP lSliderValue(PairPP arg)
{
	ThreadP th=Thread::This();
	if (exist(1))
	{	SliderNodeP slider=arg[1]->a_slider(1);
		if throwing return Unbound;
//		GUI_ENTER
		IntP val=new Int(slider->slider->slider->GetValue());
//		GUI_LEAVE
		return val;
	}
	else
		return new Int(th->sliderValue);
}

