//___________________________________________________________________
//
//                           Class Stack
//___________________________________________________________________

#ifdef _DEBUG
	#define SDEBUG
#endif

#ifdef SDEBUG
	#define Push(stack, a, marker)	\
		stack->push((a), ESM_##marker, #a, __LINE__)
	#define Push2(stack, a, b, marker)	\
	{	stack->push((a), ESM_##marker, #a, __LINE__);	\
		stack->push((b), ESM_##marker, #b, __LINE__);	\
	}
	#define Push3(stack, a, b, c, marker)	\
	{	stack->push((a), ESM_##marker, #a, __LINE__);	\
		stack->push((b), ESM_##marker, #b, __LINE__);	\
		stack->push((c), ESM_##marker, #c, __LINE__);	\
	}
	#define Push7(stack, a, b, c, d, e, f, g, marker)	\
	{	stack->push((a), ESM_##marker, #a, __LINE__);	\
		stack->push((b), ESM_##marker, #b, __LINE__);	\
		stack->push((c), ESM_##marker, #c, __LINE__);	\
		stack->push((d), ESM_##marker, #d, __LINE__);	\
		stack->push((e), ESM_##marker, #e, __LINE__);	\
		stack->push((f), ESM_##marker, #f, __LINE__);	\
		stack->push((g), ESM_##marker, #g, __LINE__);	\
	}
	#define Push8(stack, a, b, c, d, e, f, g, h, marker)	\
	{	stack->push((a), ESM_##marker, #a, __LINE__);	\
		stack->push((b), ESM_##marker, #b, __LINE__);	\
		stack->push((c), ESM_##marker, #c, __LINE__);	\
		stack->push((d), ESM_##marker, #d, __LINE__);	\
		stack->push((e), ESM_##marker, #e, __LINE__);	\
		stack->push((f), ESM_##marker, #f, __LINE__);	\
		stack->push((g), ESM_##marker, #g, __LINE__);	\
		stack->push((h), ESM_##marker, #h, __LINE__);	\
	}
	#define PushVar(stack, obj,val, marker)	\
	stack->push((obj),(val), ESM_##marker, #obj, __LINE__)
	#define PushVarA(stack, obj,val,ali, marker)	\
		stack->push((obj),(val),(ali), ESM_##marker, #obj, __LINE__)
	#define PushList(stack, a, marker)	\
		stack->pushList((a), ESM_##marker, #a, __LINE__)
	#define PushError(stack, fun, ufun, this_line, marker)	\
		stack->push((fun), (ufun), (this_line), ESM_##marker, #fun, __LINE__)

	#define Pop(stack, marker)	\
		stack->pop(ESM_##marker);
	#define Pop1(stack, a, marker)	\
		stack->pop((a), ESM_##marker);
	#define Pop2(stack, a, b, marker)	\
	{	stack->pop((b), ESM_##marker);	\
		stack->pop((a), ESM_##marker);	\
	}
	#define Pop3(stack, a, b, c, marker)	\
	{	stack->pop((c), ESM_##marker);	\
		stack->pop((b), ESM_##marker);	\
		stack->pop((a), ESM_##marker);	\
	}
	#define Pop7(stack, a, b, c, d, e, f, g, marker)	\
	{	stack->pop((g), ESM_##marker);	\
		stack->pop((f), ESM_##marker);	\
		stack->pop((e), ESM_##marker);	\
		stack->pop((d), ESM_##marker);	\
		stack->pop((c), ESM_##marker);	\
		stack->pop((b), ESM_##marker);	\
		stack->pop((a), ESM_##marker);	\
	}
#else
	#define Push(stack, a, marker)	\
		stack->push(a)
	#define Push2(stack, a, b, marker)	\
		stack->push((a), (b))
	#define Push3(stack, a, b, c, marker)	\
		stack->push((a), (b), (c))
	#define Push7(stack, a,b,c,d,e,f,g, marker)	\
		stack->push((a),(b),(c),(d),(e),(f),(g))
	#define Push8(stack, a,b,c,d,e,f,g,h, marker)	\
		stack->push((a),(b),(c),(d),(e),(f),(g),(h))
	#define PushVar(stack, obj,val, marker)	\
		stack->push((obj),(val))
	#define PushVarA(stack, obj,val,ali, marker)	\
		stack->push((obj),(val),(ali))
	#define PushList(stack, a, marker)	\
		stack->pushList((a))
	#define PushError(stack, fun, ufun, this_line, marker)	\
		stack->push((fun), (ufun), (this_line))

	#define Pop(stack, marker)	\
		stack->pop();
	#define Pop1(stack, a, marker)	\
		stack->pop((a));
	#define Pop2(stack, a, b, marker)	\
		stack->pop((a), (b));
	#define Pop3(stack, a, b, c, marker)	\
		stack->pop((a), (b), (c));
	#define Pop7(stack, a, b, c, d, e, f, g, marker)	\
		stack->pop((a), (b), (c), (d), (e), (f), (g));
#endif

class IntStack
{
	public:
		struct Item
		{	int i;
#ifdef SDEBUG
			EvalStackMarker marker;
			charP name;
			int line;
#endif
		};
		typedef Item *ItemP;

		int item, maxitem;
		int size;
		union
		{	ItemP base;
			ItemP (*base100)[100];
		};
		ItemP index, max;

		static bool noisy;

		IntStack();
		IntStack(int size_);
		~IntStack();

		void free();
		operator int() {return index-base;}
		void copyTo(IntStack &out);
		bool grow();
		bool shrink();
#ifdef SDEBUG
		void 
			push(int val), 
			push(int val, EvalStackMarker marker, charP name, int line), 
			pop(int &val, EvalStackMarker marker),
			pop(bool &val, EvalStackMarker marker),
			pop(TailCall &val, EvalStackMarker marker),
			pop(ValStatus &val, EvalStackMarker marker);
/*
		IntStack 
			&operator<<(const int val),
			&operator>>(int &val),
			&operator>>(bool &val),
			&operator>>(TailCall &val),
			&operator>>(ValStatus &val);
*/
		void 
//			push(int  i1, int  i2),
//			push(int  i1, int  i2, int  i3),
			 pop(EvalStackMarker marker),
			 pop(int &i1, int &i2, EvalStackMarker marker),
			 pop(int &i1, int &i2, int &i3, EvalStackMarker marker),
			 pop(ValStatus &i1, TailCall &i2, EvalStackMarker marker),
			 pop(ValStatus &i1, TailCall &i2, int &i3, EvalStackMarker marker);
		void pop_till(int sp);
#else
	void push(const int val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		(++index)->i=val;
		return;
	}
	IntStack &operator<<(const int val) 
	{	if (index+1 > max)
			if (!grow())
				return *this;
		(++index)->i=val;
		return *this;
	}
	void pop(      int &val) {val = (index--)->i;}
	void pop(     bool &val) {(int&)val =     (index--)->i;}
	void pop( TailCall &val) {val =  TailCall((index--)->i);}
	void pop(ValStatus &val) {val = ValStatus((index--)->i);}

	void push(int i1, int i2)
	{	if (index+2 > max)
			if (!grow())
				return;
		index[1].i=i1;
		index[2].i=i2;
		index+=2;
	}
	void push(int i1, int i2, int i3)
	{	
/*		try
		{	i[3]=i3;
		}
		catch (unsigned)
		{	if (grow())
			{	i=index;
				i[3]=i3;
			}else
				error(STACK_OVERFLOW);
		}
		i[2]=i2;
		i[1]=i1;
		index=i+3;
*/		if (index+3 > max)
			if (!grow())
				return;
		index[1].i=i1;
		index[2].i=i2;
		index[3].i=i3;
		index+=3;
	}
	void pop()
	{	index--;
	}
	void pop(int &i1, int &i2)
	{	register ItemP i=index;
		i2=(i--)->i; i1=(i--)->i;
		index=i;
	}
	void pop(int &i1, int &i2, int &i3)
	{	register ItemP i=index;
		i3=(i--)->i; i2=(i--)->i; i1=(i--)->i;
		index=i;
	}
	void pop(int &i1, TailCall &i2) 
	{	register ItemP i=index;
		i2=TailCall((i--)->i); i1=int((i--)->i);
		index=i;
	}
	void pop(bool &i1, ValStatus &i2) 
	{	register ItemP i=index;
		i2=ValStatus((i--)->i); i1=bool((i--)->i !=0);
		index=i;
	}
	void pop(int &i1, ValStatus &i2) 
	{	register ItemP i=index;
		i2=ValStatus((i--)->i); i1=int((i--)->i);
		index=i;
	}
	void pop(ValStatus &i1, TailCall &i2) 
	{	register ItemP i=index;
		i2=TailCall((i--)->i); i1=ValStatus((i--)->i);
		index=i;
	}
	void pop(ValStatus &i1, TailCall &i2, int &i3)
	{	register ItemP i=index;
		i3=(i--)->i; i2=TailCall((i--)->i); i1=ValStatus((i--)->i);
		index=i;
	}
	void pop_till(ItemP oldSP) {index=oldSP;}
	void pop_till(int i) 
	{	pop_till(&base[i]);
	}
#endif
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class FloatStack
{
	public:
		struct Item
		{	double i;
#ifdef SDEBUG
			EvalStackMarker marker;
			charP name;
			int line;
#endif
		};
		typedef Item *ItemP;

		int item, maxitem;
		int size;
		union
		{	ItemP base;
			ItemP (*base100)[100];
		};
		ItemP index, max;

		FloatStack();
		~FloatStack();

		bool grow();
		bool shrink();
#ifdef SDEBUG
		void push(double val), 
			push(double val, EvalStackMarker marker, charP name, int line), 
			pop(double &val, EvalStackMarker marker);
/*		FloatStack 
			&operator<<(const double val),
			&operator>>(double &val);
*/
		void 
			push(double d1, double d2, double d3, EvalStackMarker marker),
			 pop(double &d1, double &d2, double &d3, EvalStackMarker marker);
#else
	void push(const double val) 
	{	if (index+2 > max)
			if (!grow())
				return;
		(++index)->i=val;
	}
	FloatStack &operator>>(      double &val) {val = (index--)->i; return *this;}

	void push(double d1, double d2, double d3)
	{	if (index+4 > max)
			if (!grow())
				return;
		index[1].i=d1;
		index[2].i=d2;
		index[3].i=d3;
		index+=3;
	}
	void pop(double &d1, double &d2, double &d3)
	{	register ItemP i=index;
		d3=(i--)->i; d2=(i--)->i; d1=(i--)->i;
		index=i;
	}
#endif
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class PairPPStack
{
	public:
		struct Item
		{	PairPP i;
#ifdef SDEBUG
			EvalStackMarker marker;
			charP name;
			int line;
#endif
		};
		typedef Item *ItemP;

		int item, maxitem;
		int size;

		union
		{	ItemP base;
			ItemP (*base100)[100];
		};
		ItemP index, max;

		PairPPStack();
		~PairPPStack();

		bool grow();
		bool shrink();
#ifdef SDEBUG
		void push(PairPP val), 
			push(PairPP val, EvalStackMarker marker, charP name, int line), 
			pop(PairPP &val, EvalStackMarker marker);
/*		PairPPStack 
			&operator<<(const PairPP val),
			&operator>>(PairPP &val);
*/
#else
	void push(const PairPP val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		(++index)->i=val;
	}
	PairPPStack &operator>>(PairPP &val) {val = (index--)->i; return *this;}
#endif
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);
};

Class Var
{	public:
		ObjectP object; NodeP val; int alias; 
		union {ObjectType flag; ObjectBits bit; int tint;};

		static Var aVar;

		void mark();
		void markPtr(PtrToInt_ &ptrToInt);
};

class VarStack
{
	public:
		struct Item
		{	Var i;
#ifdef SDEBUG
			EvalStackMarker marker;
			charP name;
			int line;
#endif
		};
		typedef Item *ItemP;

		int item, maxitem;
		int size;

		union
		{	ItemP base;
			ItemP (*base100)[100];
		};
		ItemP index, max;

		VarStack();
		VarStack(int size_);
		~VarStack();

		void free();
		void copyTo(VarStack &out);
		bool grow();
		bool shrink();

		void copy(VarStack &out, IntToPtr_ &intToPtr, PtrToInt_ &ptrToInt);

		bool operator!=(const ItemP b) {return index != b;}
		operator ItemP() {return index;}
		operator int() {return index-base;}
		void 
			push(ObjectP object, NodeP val, int alias), 
			push(ObjectP object, NodeP val),
			push(ObjectP object, NodeP val, int alias, EvalStackMarker marker, charP name, int line), 
			push(ObjectP object, NodeP val, EvalStackMarker marker, charP name, int line);
#ifdef SDEBUG
		void 
			pop(EvalStackMarker marker),
			pop_till(ItemP oldSP);	// pop saved vars till old stack pointer
#else
		void pop() {index--;}
		void pop_till(ItemP oldSP) {index=oldSP;}
#endif
		void pop_till(int i) 
		{	pop_till(&base[i]);
		}
		void push_alias(ObjectP object, ObjectP alias);
		NodeP find(ObjectP obj);
		bool isvar(ObjectP obj);
		int whereIsVar(ObjectP obj);
		void setVar(ObjectP obj, NodeP val);
		PairP contents(ObjectType want, ObjectType dont);
		bool setflag(ObjectP obj, ObjectType set);
		bool clearflag(ObjectP obj, ObjectType clear);

		void mark(), markPtr(PtrToInt_ &ptrToInt);

		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);
};

class VarPStack
{
	public:
		struct Item
		{	VarP i;
#ifdef SDEBUG
			EvalStackMarker marker;
			charP name;
			int line;
#endif
		};
		typedef Item *ItemP;

		int item, maxitem;
		int size;

		union
		{	ItemP base;
			ItemP (*base100)[100];
		};
		ItemP index, max;

		VarPStack();
		~VarPStack();

		bool grow();
		bool shrink();
#ifdef SDEBUG
		void push(VarP val), 
			push(VarP val, EvalStackMarker marker, charP name, int line), 
			pop(VarP &val, EvalStackMarker marker);
/*		VarPStack 
			&operator<<(const VarP val),
			&operator>>(VarP &val);
*/
#else
	void push(const VarP val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		(++index)->i=val;
	}
	VarPStack &operator>>(VarP &val) {val = (index--)->i; return *this;}
#endif
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);
};

class ErrorStack
{
	public:
		struct Item
		{	ErrorStruct i;
#ifdef SDEBUG
			EvalStackMarker marker;
			charP name;
			int line;
#endif
		};
		typedef Item *ItemP;

		int item, maxitem;
		int size;

		union
		{	ItemP base;
			ItemP (*base100)[100];
		};
		ItemP index, max;

		ErrorStack();
		~ErrorStack();

		bool grow();
		bool shrink();
#ifdef SDEBUG
		ErrorStructP top() {return &(index->i);}
		void push(ObjectP fun_, ObjectP ufun_, LineP this_line_), 
			push(ObjectP fun_, ObjectP ufun_, LineP this_line_, EvalStackMarker marker, charP name, int line), 
			pop(EvalStackMarker marker);
#else
	ErrorStructP top() {return &(index->i);}
	void push(ObjectP fun_, ObjectP ufun_, LineP this_line_) 
	{	if (index+2 > max)
			if (!grow())
				return;
		++index;
		index->i.fun = fun_;
		index->i.ufun = ufun_;	
		index->i.this_line = this_line_;
	}
	void pop() {index--;}
#endif
		void mark(), markPtr(PtrToInt_ &ptrToInt);

		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);
};
//#undef SDEBUG

extern Function 
	lshrinkStacks, lsetStackNoisy;
