//___________________________________________________________________
//
//                         Class StaticText
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "NodePStack.h"
#include "Color.h"
#include "Point.h"

#include "Event.h"
#include "StaticText.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Error.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

StaticText::StaticText(StaticTextNodeP staticTextNode_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=staticTextNode_;
	wxWindow *parent=(wxWindow*)console;
	if (staticTextNode_->parent)
		parent=staticTextNode_->parent->window;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			_("StaticText"));
}

StaticText::StaticText(StaticTextNodeP staticTextNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxString& name):
wxStaticText(parent, id, label, pos, size, style, name)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=staticTextNode_;
	staticTextNode_->staticText=this;
	SetLabel(label);
}

StaticText::~StaticText()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
}

wxString StaticText::wxstring()
{
	wxString str;
	str << _("StaticText") << " " << GetLabel() << ' ' 
		<< '[' << GetName() << "] "
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &StaticText::read(DataInputStream &istr) 
{	
	wxString label, name, fontName;
	wxPoint pos;
	wxSize size;
	int style, fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> label >> name
		>> pos.x >> pos.y >> size.x >> size.y >> style 
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &StaticText::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle());
	{	wxColour f=GetForegroundColour(), b=GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=GetFont();
	ostr << f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();
	return ostr;
}

StaticTextNode::StaticTextNode(): 
	WindowNode(STATICTEXT), 
	staticText((StaticTextP&)window)
{}

StaticTextNode::StaticTextNode(WindowNodeP parent_): 
	WindowNode(STATICTEXT, parent_), 
	staticText((StaticTextP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void StaticTextNode::free() 
{	
	if (staticText) 
	{	*staticText->d(StaticText::windowNode)=0;
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		staticText->Destroy(); 
		staticText=0;
	}
}

DataInputStream &StaticTextNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode;
		staticText=new StaticText(StaticTextNodeP(this));
		if (!staticText)
		{	error(OUT_OF_MEM); return istr;
		}
		staticText->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &StaticTextNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (staticText)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent);
		staticText->write(ostr);
	}

	else
		ostr << 0;
	return ostr;
}

void StaticTextNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
}

void StaticTextNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	StaticTextP s=staticText, s2=0;
	s2=new StaticText(StaticTextNodeP(this),
			p, -1, s->GetLabel(), 
			s->GetPosition(), s->GetSize(), 
			s->GetWindowStyle(), s->GetName());
	if (!s2)
	{	error(OUT_OF_MEM); return;
	}
	staticText=s2;
	s2->SetForegroundColour(s->GetForegroundColour());
	s2->SetBackgroundColour(s->GetBackgroundColour());
	wxFont f=s->GetFont();
	s2->SetFont(f);
	delete s;
	staticText=s2;
}

StringP StaticTextNode::string()
{
	if (staticText)
		return new String(staticText->wxstring());
	return _Null_Word->name();
}

wxString StaticTextNode::wxstring()
{
	if (staticText)
		return staticText->wxstring();
	return wxEmptyString;
}

void StaticTextNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lStaticText(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=Thread::This()->graph;
	WindowNodeP parentNode=0;
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	long style = 0;
	if (exist(3))
	{	int s=arg[3]->a_int(3);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(4))
	{	PairP p=arg[4]->a_int_vector(4);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(5))
	{	PairP s=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=_("StaticText");
	if (exist(6))
	{	StringP n=arg[6]->a_string(6);
		if throwing return Unbound;
		name=n->ptr();
	}
//	GUI_ENTER
	StaticTextNodeP staticTextNode=new StaticTextNode(parentNode);
	if throwing return Unbound;
	StaticText *st=new StaticText(staticTextNode,
			parent, id, label, pos, size, style, name);
	if (!st)
	{	error(OUT_OF_MEM); return Unbound;
	}
//	GUI_LEAVE
	return staticTextNode;
}

NodeP lStaticTextLabel(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
//	GUI_ENTER
	StringP label=new String(staticText->staticText->GetLabel());
//	GUI_LEAVE
	return label;
}

NodeP lStaticTextSetLabel(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	StringP label=arg[2]->a_string(2);
	if throwing return Unbound;
	wxString wxlabel=label->ptr();
//	GUI_ENTER
	staticText->staticText->SetLabel(wxlabel);
//	GUI_LEAVE
	return Unbound;
}

NodeP lStaticTextSetBackgroundColor(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	staticText->staticText->SetBackgroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lStaticTextSetColor(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	staticText->staticText->SetForegroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lStaticTextSetFontSize(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=staticText->staticText->GetFont();
	f.SetPointSize(s);
	staticText->staticText->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lStaticTextSetFontName(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	s=new String(s);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=staticText->staticText->GetFont();
	f.SetFaceName(s->ptr());
	staticText->staticText->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lStaticTextSetFontStyle(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=staticText->staticText->GetFont();
	f.SetStyle(s);
	staticText->staticText->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lStaticTextSetFontWeight(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=staticText->staticText->GetFont();
	f.SetWeight(s);
	staticText->staticText->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lStaticTextDestroy(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=staticText->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(staticText));
			p->check_oldyoung(p->children);
		}
	}
	staticText->staticText->Destroy();
//	GUI_LEAVE
	return Unbound;
}
