#include "stdafx.h"
#include <malloc.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

#include "Main.h"
#include "File.h"
#include "PtrToInt_.h"
#include "IntToPtr_.h"
#include "String.h"
#include "Object.h"
#include "Pair.h"
#include "Line.h"
#include "Procedure.h"
#include "Array.h"
#include "Table.h"
#include "Init.h"
#include "Error.h"
#include "Thread.h"
#include "Security.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#else
	static char THIS_FILE[]= __FILE__;
#endif

#ifdef GC_DEBUG
	#define print_gc(text, var)			  \
		File::print(cout.file(), text, var);\
		if (dribbler)								 \
			fprintf(dribbler.file(), text, var); \
		examined=0;										\

	#define print_gc_object							 \
		File::print(conout,							  \
			"%s p=%s v=%s mark=%2i gcNo.=%2i \n",	\
			obj, obj->proc, obj->val,					 \
			obj->mark_gc, gc_run_numero);				  \

	#define inc_examined	examined++
	#define freed_sofar_init	int freed_sofar=0
	#define freed_sofar_is(n)	freed_sofar=(n)
#else
	#define print_gc(text, var)
	#define print_gc_object
	#define inc_examined
	#define freed_sofar_init
	#define freed_sofar_is(n)
#endif

int Table::maxStringLength=65536;

Table::Table(int size_): 
	Node(TABLE), allocated(false), size(0), base(0)
{	flag=ObjectType(TABLE_OBJECT);
	self_resize(size_);
}

Table::~Table() 
{
	if (allocated) 
	{	GC_::free((NodeP*)base); base=0;
		allocated=false;
	}
}

void Table::free() 
{
	if (allocated) 
	{	GC_::free((NodeP*)base); base=0;
		allocated=false;
	}
}

void Table::inc_gc()
{
	if (!allocated) return;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			fifo << obj;
}

int Table::youngs() 
{
	if (!allocated) return 0;

	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmark(obj);
	return youngs;
}

int Table::youngsPtr(PtrToInt_ &ptrToInt) 
{
	if (!allocated) return 0;

	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmarkPtr(obj, ptrToInt);
	return youngs;
}

void Table::check_oldyoungs()
{
	if (!allocated) return;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			check_oldyoung(obj);
}

void Table::setPtrs(IntToPtr_ &intToPtr)
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
	{	*itemptr=(Object*)intToPtr.find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
DataInputStream &Table::read(DataInputStream &istr) 
{	
	int siz;
	istr >> allocated >> siz;
	size=0;
	base = (ObjectP*) GC_::calloc(sizeof(ObjectP), siz+1);
	if (!base)
		return istr;
	size=siz;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Table::write(wxDataOutputStream &ostr) 
{
	ostr << type << allocated << size;

	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;
}

NodeP Table::copy(PtrToInt_ &ptrToInt)
{
	TableP tab=new Table(size);
	if throwing return Unbound;
	ObjectPP itemptr=base, itemptr2=tab->base;
	for (int i=size;  i;  i--)
		*itemptr2++ = ObjectP(ptrToInt.find(*itemptr++));
	return tab;
}

StringP Table::string()
{
	char textbuf[65536];
	ThreadP th=Thread::This();
	th->stringptr=textbuf;
	th->stringlen=maxStringLength-1;
	stringbuffer << NodeP(this);
	*th->stringptr='\0';
	return new String(textbuf);
}

void Table::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	PairP list=contentsNameVar(VALS);
	if (list)
		list->print(strm, width, depth-1, noBrackets-1, noSpacers-1);
}

void Table::printBin(FILEP strm)
{
	PairP list=contentsNameVar(VALS);
	if (list)
		list->printBin(strm);
}

int Table::hash(StringP str)
{
	if (!str)
		str=_Null_Word->name();
	unsigned int *i= (unsigned int*) str->ptr(),   index=0;

	for (char len=char(str->length());  len>0;  len-=sizeof(int))
	{	switch (len)
		{	case 0:	break;
			case 1:	index^= *i & 0x000000ff;	break;
			case 2:	index^= *i & 0x0000ffff;	break;

			case 3:	index^= *i & 0x00ffffff;	break;
			case 4:	index^= *i;	break;
			default:	index^= *i++ << 5;
		}
	};
	index %= size;
	return index;
}

ObjectP Table::find(StringP str) 
{
	ObjectP obj;
	for (obj=base[hash(str)];  obj;  next(obj))
		if (str->equal(obj->name()))
			break;

	return obj;
}

ObjectP Table::operator[](StringP str) {return find(str);}

ProcedureP Table::findProc(StringP filename, int row)
{
	int maxrow=-1;
	ObjectP obj=0;
	ObjectPP itemptr=base;
	LineP line=0;
	ProcedureP result=0;

	for (int i=size;  i;  i--)
		for (obj = *itemptr++;  obj;  next(obj))
			if ((obj->flag & PROCS) && !obj->bit.permanent)
			{	line=LineP(obj->proc()->bodylist());
				if (line && line->filename()==filename)
					if (line->row <= row && line->row > maxrow)
					{	maxrow=line->row;
						if (obj->table()!=Unbound)
						{	result=obj->table()->findProc(filename, row);
							if (!result)
								result=obj->proc();
						}else
							result=obj->proc();
					}
			}

	return result;
}

ObjectP Table::insert(StringP str, charP file, int line)
{
	ObjectPP obj=&base[hash(str)];
	*obj = new Object(str, *obj);
	check_oldyoung(*obj);
	(*obj)->creationFile=file;
	(*obj)->creationLine=line;
	return *obj;
}

PairP Table::objectlist()
{
	PairP li=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			push(li, obj);

	li = PairP(li->combSort());
	return li;
}

void Table::self_resize(int newsize)
{
	ObjectPP newbase = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(newsize+1));

	if (newbase==ObjectPP(0))
		c_out << _("; Warning: Not enough memory to resize the table.") << endL;
	else
	{	if (allocated) 
		{	PairP li=0;	li=objectlist();

			GC_::free((NodeP*)base);	base=newbase;	size=newsize;
			while (li)
			{	int i=hash(ObjectP(li->first())->name());
				base[i] = ObjectP(li->first());
				next(li);
			}
		}else
		{	allocated=true;	base=newbase;	size=newsize;
		}
	}
}

void Table::mark()
{
	_marked = run_numero();
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))

			obj->mark();

	print_gc("%2i ", examined);
}

void Table::mark_all_but_objects()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->mark_all_but_objects();

	print_gc("%2i ", examined);
}

void Table::clean()
{
	_marked = run_numero();
	mark_all_but_objects();

	print_gc("Table::clean: ", 0);
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
	{	ObjectP lastobj=0;

		for (ObjectP obj = *itemptr;  obj;  next(obj)) 
		{//	print_gc_object;

			if ( obj->marked() != GC_::run_numero()
			&&  obj->proc_or_prim()==Unbound
			&&  obj->var()==Unbound
			&&  obj->table()==Unbound
			&& !(obj->bit.permanent | obj->bit.permanentobj) )
			{
				inc_examined;
				if (!lastobj)
					next(*itemptr);
				else
					lastobj->settail( obj->tail() );
			}else		// has a value, don't gctwa
				lastobj=obj;
		}
		itemptr++;
	}
	print_gc("%2i collected\n", examined);
}

int Table::count()
{
	int c=0;
	for (int i=0; i<size; i++)
		for (ObjectP obj=base[i]; obj; next(obj))
		{	c++;
		}
	return c;
}

int  Table::_sizeof()
{
	int size=0;
	for (int i=0; i<size; i++)
		for (ObjectP obj=base[i]; obj; next(obj))
			size+=obj->_sizeof();
	return size;
}

void Table::setflag(ObjectType want, ObjectType newflag)
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->setflag(want, newflag);
}

void Table::clearflag(ObjectType want, ObjectType newflag)
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->clearflag(want, newflag);
}

PairP Table::contents(ObjectType want, ObjectType dont)
{
	PairP contentsList=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			NodeP sub = obj->contents(want, dont);
			if (sub)
				if (sub->is.list)
					if (contentsList)
					{	PairP lst=PairP(sub);
						while (lst && lst->tail())
							next(lst);
						lst->settail(contentsList);
						contentsList=PairP(sub);
					}else
						contentsList=PairP(sub);
				else
					push(contentsList, sub);
		}
	if (contentsList)
		contentsList = PairP(contentsList->combSort());
	return contentsList;
}

PairP Table::contentsNameVar(ObjectType want) 
{
	PairP contentsList=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			StringP name = obj->name();
			if (name)
				if (obj->bit.proc && (want & PROCS))
					push(contentsList, 
						new Pair(name, 0));
				else if (obj->bit.val && (want & VALS))
					push(contentsList, 
						new Pair(name, 
						new Pair(obj->var(), 0)));
			if throwing return PairP(Unbound);
		}
	if (contentsList)
		contentsList = PairP(contentsList->combSort());
	return contentsList;
}

NodeP Table::contentsTree(	wxArrayString &namearr, 
									wxArrayString &valarr, 
									wxArrayPtrVoid &objarr, 
									ObjectType want, 
									ObjectType dont,
									ObjectType must, 
									int &count, int level)
{
	objarr[count]=this;
	if (!bit.table_expand)
	{	StringP str=string();
		if (str  &&  str->is.string)
		{	str=str->noFormat();
			if (str  &&  str->is.string)
			{	valarr[count]=str->ptr();
				incContentsArrays(namearr, valarr, objarr, count);
			}
		}
		return this;
	}
	PairP contentsList=0;
	int oldcount=count;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			if ((obj->flag & want && !(obj->flag & dont)) || obj->flag & must)
			{	contentsList=new Pair(obj, contentsList);
				if throwing return Unbound;
			}
		
	if (!contentsList)
		return 0;
	contentsList=PairP(contentsList->combSort());
	while (contentsList)
	{	contentsList->first()->contentsTree(
			namearr, valarr, objarr, want, dont, must, count, level);
		next(contentsList);
	}
	return this;
}

PairP Table::contents(ObjectType obj_type)
{
	PairP ret=0;
	push(ret, contents(ObjectType(TABLES & obj_type), PERMANENT));
	push(ret, contents(ObjectType(  VALS & obj_type), PERMANENT));
	push(ret, contents(ObjectType( PROCS & obj_type), PERMANENT));
	return ret;
}

NodeP Table::memberP(NodeP subobj, bool notp, bool substr)
{
	if (substr || !subobj) return _False;

	ObjectP obj=find(subobj->string());
	if (!obj)
		return _False;
	return _True;
}

NodeP Table::setitem(NodeP index, NodeP newitem, bool safe)
{	StringP str=index->a_string(1);
	ObjectP obj=find(str);
	if (!obj)
		obj=insert(str, THIS_FILE, __LINE__);
	obj->setvar(newitem);
	return Unbound;
}

void Table::removeitem(NodeP index)
{	StringP name=index->a_string(1);
	if throwing return;
	int i=hash(name);
	ObjectP obj=base[i], last=0;
	while (obj)
	{	if (name->equal(obj->name()))
			break;
		last=obj;
		next(obj);
	}
	if (!obj)
	{	error(BAD_DATA, index);
		return;
	}
	if (!last)
	{	base[i]=obj->tail();
	}
	else
		last->settail(obj->tail());
}

NodeP Table::reverse()
{
	int len=count();
	TableP result=new Table(len);
	if throwing return Unbound;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			result->setitem(obj->var(), obj, false);

	return result;
}

NodeP Table::upper()
{
	int len=count();
	TableP result=new Table(len);
	if throwing return Unbound;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			result->setitem(obj->upper(), obj->var()->upper(), false);

	return result;
}

NodeP Table::lower()
{
	int len=count();
	TableP result=new Table(len);
	if throwing return Unbound;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			result->setitem(obj->lower(), obj->var()->lower(), false);

	return result;
}

PairP Table::list() {return contentsNameVar(ObjectType(VALS | PROCS));}

TableP Table::merge(TableP b)
{
	int len=count()+b->count();
	TableP result=new Table(len);
	if throwing return TableP(Unbound);

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			result->setitem(obj, obj->var(), false);

	itemptr=b->base;

	for (int i=b->size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			result->setitem(obj, obj->var(), false);

	return result;
}

NodeP Table::erase_all(bool procs, bool vals, bool tables) 
{
	ThreadP th=Thread::This();
	if (!th->security->erase)
	{	error(SECURITY_ERROR, new String(_("to erase definitions"))); return Unbound;
	}
	bool redef = th->redefinable;
	
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)

		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			if (procs 
			&&  obj->bit.proc
			&& !obj->bit.proc_buried 
			&& !obj->proc_or_prim()->is.primitive)
			{	obj->setproc( ProcedureP(Unbound) );
				obj->tint &= ~PROCS;
			}
			if (vals  
			&&  obj->bit.val
			&& !obj->bit.val_buried)
			{	obj->setvar(Unbound);
				obj->tint &= ~VALS;
			}
			if (tables && !obj->bit.proc_buried)
			{	obj->settable(TableP(Unbound));
				obj->tint &= ~TABLES;
			}
		}
	th->setfilename(_Unnamed->name());
	PairP fns=new Pair(th->filename(), 0);
	if throwing return Unbound;
	th->filenames=fns;
	return Unbound;
}

void Table::makeTree(bool deep)
{
	ThreadP th=Thread::This();
	PairP contentsList=0;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			if (obj->flag & (PROCS | TABLES))
			{	contentsList=new Pair(obj, contentsList);
				if throwing return;
			}
	if (!contentsList)
		return;
	ArrayP contentsArray=contentsList->array();
	contentsArray->combSortInPlace();
	if throwing return;
	NodePP dp=contentsArray->data();
	for (int i=count();  i;  i--)
	{	ObjectP obj=ObjectP(*dp++), oufun=th->ufun;
		th->ufun=obj;
		if (obj)
		{	if (obj->proc()->bodylist())
				obj->proc()->bodylist()->makeTreeFromBody(deep);
			if throwing return;
			if (obj->bit.table)
				obj->table()->makeTree(deep);
			if throwing return;
		}
		th->ufun=oufun;
	}
/*	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{	if (obj->flag & PROCS)
				if (obj->proc()->bodylist())
				{	ObjectP oufun=th->ufun;
					th->ufun=obj;
					obj->proc()->bodylist()->makeTreeFromBody(deep);
					th->ufun=oufun;
				}
			if (obj->bit.table)
				obj->table()->makeTree(deep);
		}
*/
}
