//___________________________________________________________________
//
//                          Class Turtle
//
//                           Definitions
//___________________________________________________________________

#include "stdafx.h"

#include <wx/app.h>
#include <wx/textctrl.h>
#include <wx/listctrl.h>
#include <wx/frame.h>
#include <wx/splitter.h>
#include <wx/treectrl.h>
#include <wx/html/helpctrl.h>
#include <wx/dir.h>

#include <math.h>

#ifdef OPENGL
	#include <GL/glu.h>
	#include <gl2ps.h>
#endif

#if defined(OPENGL) && defined(FTGL)
#ifndef WIN32
	#include <setjmp.h>
	#define __FT2_BUILD_UNIX_H__
	#define FREETYPE /usr/include/freetype2/freetype
	#define FT_FREETYPE_H <FREETYPE/freetype.h>
	#define FT_ERRORS_H <FREETYPE/fterrors.h>
	#define FT_CONFIG_CONFIG_H <FREETYPE/config/ftconfig.h>
	#define FT_CONFIG_OPTIONS_H <FREETYPE/config/ftoption.h>
	#define FT_CONFIG_STANDARD_LIBRARY_H <FREETYPE/config/ftstdlib.h>
	#define FT_MODULE_ERRORS_H <FREETYPE/ftmoderr.h>
	#define FT_BEGIN_HEADER
	#define FT_END_HEADER
	#define FT_ERROR_DEFINITIONS_H <FREETYPE/fterrdef.h>
	#define FT_TYPES_H <FREETYPE/fttypes.h>
	#define FT_SYSTEM_H <FREETYPE/ftsystem.h>
	#define FT_IMAGE_H <FREETYPE/ftimage.h>
	#define FT_GLYPH_H <FREETYPE/ftglyph.h>
#endif
#include <FTGL/FTGLPolygonFont.h>
#include <FTGL/FTGLTextureFont.h>
#endif

#include "Main.h"

#include "String.h"

#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Object.h"
#include "Line.h"
#include "List.h"
#include "Array.h"

#include "Error.h"
#include "Init.h"
#include "File.h"
#include "TurtleNode.h"

#include "Environment.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "FTGLFontManager.h"
#include "StringToPtr.h"
#include "Font.h"
#include "Turtle.h"
#include "Mouse.h"

#include "Bitmap.h"
#include "Recorder.h"

#include "Graph.h"
#include "Vars.h"
#include "Splitter.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Shadow.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

//___________________________________________________________________
//
//                          Portable Part
//___________________________________________________________________

// This part should be portable and reentrant, 
//	because it is the core of the turtle graphics.
// It only uses standard functions, which appear in the Machine-specific Part.
// If one machine does not have that function, write a dummy. 
// You could also leave the machine-specific code in a separate file, 
// but I think it is better if it is here altogether, because it is ONE class.
//
// You may ask, why it is necessary to have such a big class. 
// First, the task drawing wrapping Turtle Graphics simply is not so easy. 
// Second, I didn`t have so much time as to split it up once more, 
//	but simplifying is still wanted.
// Third, I didn`t want to leave it global, 
//	because I really really wanted it to be reentrant.
// And in global functions reentrancy also certainly is possible 
//	but not as easy as in a Class.
//___________________________________________________________________

PointFiFo::~PointFiFo() {if (allocated) GC_::free(base); base=0;}

PointFiFo &PointFiFo::operator<<(const Point val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

PointFiFo &PointFiFo::operator>>(Point &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}

PointFiFo &PointFiFo::operator=(PointFiFo &b)
{
	PointP bout=b.out;
	in=out=base;
	Point p;
	while (b)
	{	b >> p;
		*this << p;
	}
	b.out=bout;
	return *this;
}

void PointFiFo::expand()	// transfer is untested!!!!
{
	PointP newbase = (Point*) GC_::calloc(1, sizeof(Point)*(size+expand_size));

	if (newbase==PointP(0))	// no room to increase FiFo
	{	c_out << _("; Warning: Not enough memory to expand PointFiFo!") << endL;
		_overflow=true;
	}
	else
	{	PointP i=newbase;

		if (out)			// transfer old FiFo to new FiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;
		}				
		if (allocated) 
			GC_::free(base);

		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}
//___________________________________________________________________

DPointFiFo::~DPointFiFo() {if (allocated) GC_::free(base); base=0;}

DPointFiFo &DPointFiFo::operator<<(const Point val) 
{
	if (overflow()) return *this;
	*in=DPoint(val.x,val.y,val.z);
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

DPointFiFo &DPointFiFo::operator>>(Point &val) 
{
	val = Point(out->x, out->y, out->z);
	if (out==max) out=base; else out++;
	return *this;
}

DPointFiFo &DPointFiFo::operator<<(const DPoint val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

DPointFiFo &DPointFiFo::operator>>(DPoint &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}

DPointFiFo &DPointFiFo::operator=(DPointFiFo &b)
{
	DPointP bout=b.out;
	in=out=base;
	DPoint p;
	while (b)
	{	b >> p;
		*this << p;
	}
	b.out=bout;
	return *this;
}

void DPointFiFo::expand()	// transfer is untested!!!!
{
	DPointP newbase = (DPoint*) GC_::calloc(1, sizeof(DPoint)*(size+expand_size));

	if (newbase==DPointP(0))	// no room to increase FiFo
	{	c_out << _("; Warning: Not enough memory to expand DPointFiFo!") << endL;
		_overflow=true;
	}
	else
	{	DPointP i=newbase;

		if (out)			// transfer old FiFo to new FiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;
		}				
		if (allocated) 
			GC_::free(base);

		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}
//___________________________________________________________________

FTGLFontManager::FTGLFontManager()
{
	fonts=new StringToPtr(31);
	if (!fonts)
		error(OUT_OF_MEM);
}
FTGLFontManager::~FTGLFontManager()
{
	clear();
	delete fonts;
}

void FTGLFontManager::clear()
{
#if defined(OPENGL) && defined(FTGL)
	StringToPtr::PtrStringPairP obj;
	for (int i=0; i < fonts->size; i++)
	{	for (obj=fonts->data[i];  obj;  obj=obj->tail)
			delete FTFontP(obj->p);
	}
	delete fonts;
	fonts=new StringToPtr(31);
	if (!fonts)
		error(OUT_OF_MEM);
#endif
}

#if defined(OPENGL) && defined(FTGL)
FTFontP FTGLFontManager::getFont(const wxString filename, int size)
{
	ThreadP th=Thread::This();
	wxString uniqueName=filename;
	uniqueName << size << ' '<< int(SCREEN->textureFont);
	FTFontP result = FTFontP(fonts->find(uniqueName));

	if (result)
		return result;

#ifdef WIN32
	wxString fullname = environment.fontsdir+filename+".ttf";
#else
	wxString name=wxFileName::GetPathSeparator()+filename+".ttf";
	wxDir dir(environment.fontsdir);
	wxString fn;
	dir.GetFirst(&fn, wxEmptyString, wxDIR_DIRS);
	if (!File::exists(environment.fontsdir+fn+name))
		while (dir.GetNext(&fn))
			if (File::exists(environment.fontsdir+fn+name))
				break;
	wxString fullname=environment.fontsdir+fn+name;
#endif
	if (!File::exists(fullname))
	{	error(FILE_ERROR, new String(
			wxString(_("Font "))+fullname+_(" doesn't exist")));
		return 0;
	}

	FTFontP font;
	if (SCREEN->textureFont)
		font = new FTGLTextureFont(fullname.c_str());
	else
		font = new FTGLPolygonFont(fullname.c_str());

	if (!font)
	{	error(FILE_ERROR, new String(
			wxString(_("Font "))+fullname+_(" failed to open")));

		delete font;
		return 0;
	}
	if (!font->FaceSize( size))
	{	error(FILE_ERROR, new Pair(
			new String(
				wxString(_("Font "))+filename+_(" failed to set size ")), 
			new Int(size)));
		delete font;
		return 0;
	}
	fonts->insert(uniqueName, font);
	return font;
}
#endif
//___________________________________________________________________

Font_::Font_()
{	
	sizex=20; sizey=20; weight=400; alignx=aligny=0; 
#ifdef WIN32
	name="Times";
#else
	name="FreeSerif";
#endif
	ftfont=0;
}

Font_::~Font_()
{
}

Font_ &Font_::operator=(const Font_ &b) 
{
	sizex=b.sizex; 
	sizey=b.sizey; 
	weight=b.weight;
	alignx=b.alignx;
	aligny=b.aligny;
	name=b.name;
	ftfont=b.ftfont;
	return *this;
}

void Font_::load(GraphP graph)
{
#if defined(OPENGL) && defined(FTGL)
	if (graph)
		ftfont=graph->ftglFontManager->getFont(name.c_str(), int(sizex));
#endif
}
//___________________________________________________________________

ShadowObject::ShadowObject(): 
	nPoints(0), nPlanes(0), 
	ptsize(0), plsize(0), 
	dptsize(256), dplsize(256), 
	points(0), colors(0), texs(0), planes(0) 
{
	nPoints=1;
}

ShadowObject::~ShadowObject()
{
	if (planes) {GC_::free(planes); planes=0;}
	if (points) {GC_::free(points); points=0;}
	if (colors) {GC_::free(colors); colors=0;}
	if (texs) {GC_::free(texs); texs=0;}
}

bool ShadowObject::checkPoints(int i)
{
	if (ptsize <= i)
	{	int size=ptsize+dptsize;
		Point *p=(Point*)GC_::realloc(points, size*sizeof(Point));
		if (!p)
			return false;
		points=p;
		Color *c=(Color*)GC_::realloc(colors, size*sizeof(Color));
		if (!c)
			return false;
		colors=c;
		Point *t=(Point*)GC_::realloc(texs, size*sizeof(Point));
		if (!t)
			return false;
		texs=t;
		ptsize=size;
		dptsize*=2;
	}
	return true;
}

bool ShadowObject::checkPlanes(int i)
{
	if (plsize <= i)
	{	int oplsize=plsize, size=plsize+dplsize;
		sPlane *p=(sPlane*)GC_::realloc(planes, size*sizeof(sPlane));
		if (!p)
			return false;
		planes=p;
		plsize=size;
		for (int i=oplsize; i<plsize;  i++)
		{	planes[i].neigh[0]=planes[i].neigh[1]=planes[i].neigh[2]=0;
		}
		dplsize*=2;
	}
	return true;
}

void ShadowObject::clear()
{
	if (planes) {GC_::free(planes); planes=0;}
	if (points) {GC_::free(points); points=0;}
	if (colors) {GC_::free(colors); colors=0;}
	if (texs) {GC_::free(texs); texs=0;}
	ptsize=plsize=nPoints=nPlanes=0;
	dptsize=dplsize=256;
}

int ShadowObject::read(char *st)
{
	FILE *file;
	int i,j;

	file = fopen(st, "r");
	if (!file) return FALSE;
	//points
	fscanf(file, "%d", &(nPoints));
	for (i=1;  i<=nPoints;  i++)
	{	if (!checkPoints(i))
		{	fclose(file);
			return false;
		}
		fscanf(file, "%f", &(points[i].x));
		fscanf(file, "%f", &(points[i].y));
		fscanf(file, "%f", &(points[i].z));
	}
	//planes
	fscanf(file, "%d", &(nPlanes));
	for (i=0;  i<nPlanes;  i++)
	{	if (!checkPlanes(i))
		{	fclose(file);
			return false;
		}
		fscanf(file, "%d", &(planes[i].p[0]));
		fscanf(file, "%d", &(planes[i].p[1]));
		fscanf(file, "%d", &(planes[i].p[2]));

		for (j=0;  j<3;  j++)
		{	fscanf(file, "%f", &(planes[i].normals[j].x));
			fscanf(file, "%f", &(planes[i].normals[j].y));
			fscanf(file, "%f", &(planes[i].normals[j].z));
		}
	}
	fclose(file);
	return TRUE;
}

void ShadowObject::setConnectivity()
{
	// connectivity procedure - based on Gamasutra's article
	if (!planes)
		return;

	int p1i, p2i, p1j, p2j;
	int P1i, P2i, P1j, P2j;
	int i,j,ki,kj;

	for (i=0;  i<nPlanes-1;  i++)
		for (j=i+1;  j<nPlanes;  j++)
			for (ki=0;  ki<3;  ki++)
				if (!planes[i].neigh[ki])
				{	for (kj=0;  kj<3;  kj++)
					{	p1i=ki;
						p1j=kj;
						p2i=(ki+1)%3;
						p2j=(kj+1)%3;

						p1i=planes[i].p[p1i];
						p2i=planes[i].p[p2i];
						p1j=planes[j].p[p1j];
						p2j=planes[j].p[p2j];

						P1i=((p1i+p2i)-abs(p1i-p2i))/2;
						P2i=((p1i+p2i)+abs(p1i-p2i))/2;
						P1j=((p1j+p2j)-abs(p1j-p2j))/2;
						P2j=((p1j+p2j)+abs(p1j-p2j))/2;

						if ((P1i==P1j) && (P2i==P2j))	//they are neighbours
						{	planes[i].neigh[ki] = j+1;	  
							planes[j].neigh[kj] = i+1;	  
						}
					}
				}
}

void ShadowObject::calcPlane(sPlane *plane)
{
	// function for computing a plane equation given 3 points
	Point v[4];
	int i;

	for (i=0;  i<3;  i++)
	{	v[i+1].x = points[plane->p[i]].x;
		v[i+1].y = points[plane->p[i]].y;
		v[i+1].z = points[plane->p[i]].z;
	}
	plane->PlaneEq.a = 
		v[1].y*(v[2].z-v[3].z) + 
		v[2].y*(v[3].z-v[1].z) + 
		v[3].y*(v[1].z-v[2].z);
	plane->PlaneEq.b = 
		v[1].z*(v[2].x-v[3].x) + 
		v[2].z*(v[3].x-v[1].x) + 
		v[3].z*(v[1].x-v[2].x);
	plane->PlaneEq.c = 
		v[1].x*(v[2].y-v[3].y) + 
		v[2].x*(v[3].y-v[1].y) + 
		v[3].x*(v[1].y-v[2].y);
	plane->PlaneEq.d =-( 
		v[1].x*(v[2].y*v[3].z - v[3].y*v[2].z) +
		v[2].x*(v[3].y*v[1].z - v[1].y*v[3].z) +
		v[3].x*(v[1].y*v[2].z - v[2].y*v[1].z) );
}

void ShadowObject::draw()
{
	if (!planes)
		return;
#ifdef OPENGL
	int i, j;
	glBegin(GL_TRIANGLES);
	for (i=0;  i<nPlanes;  i++)
	{	for (j=0;  j<3;  j++)
		{
			glColor4ubv((unsigned char*)&colors[planes[i].p[j]]);
			gl2psNormal3fv((float*)&planes[i].normals[j]);
			glVertex3fv((float*)&points[planes[i].p[j]]);
		}
	}
	glEnd();
#endif
}

void ShadowObject::castShadows(Point lp, double lp4)
{
#ifdef OPENGL
	ThreadP th=Thread::This();
	if (!planes || !points)
		return;
	const double f=100;
	int i, j, k, jj;
	int p1, p2;
	Point v1, v2;
	double	side;

	//set visual parameter
	for (i=0;  i<nPlanes;  i++)
	{	// check to see if light is in front or behind the plane (face plane)
		side= planes[i].PlaneEq.a*lp.x+
				planes[i].PlaneEq.b*lp.y+
				planes[i].PlaneEq.c*lp.z+
				planes[i].PlaneEq.d*lp4;
		if (side >0) 
			planes[i].visible = TRUE;
		else 
			planes[i].visible = FALSE;
	}
	glDisable(GL_LIGHTING);
	glDepthMask(GL_FALSE);
	glDepthFunc(GL_LEQUAL);

	glEnable(GL_STENCIL_TEST);
	glColorMask(0, 0, 0, 0);
	glStencilFunc(GL_ALWAYS, 1, 0xffffffff);

	for (int l=1;  l<=2;  l++)	// loop to avoid strange artefacts
	{
		// first pass, stencil operation decreases stencil value
		glFrontFace(GL_CCW);
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
		for (i=0;  i<nPlanes;  i++)
		{	if (planes[i].visible)
				for (j=0;  j<3;  j++)
				{	k = planes[i].neigh[j];
					if ((!k) || (!planes[k-1].visible))
					{	// here we have an edge, we must draw a polygon
						p1 = planes[i].p[j];
						jj = (j+1)%3;
						p2 = planes[i].p[jj];

						//calculate the length of the vector
						v1.x = (points[p1].x - lp.x)*f;
						v1.y = (points[p1].y - lp.y)*f;
						v1.z = (points[p1].z - lp.z)*f;

						v2.x = (points[p2].x - lp.x)*f;
						v2.y = (points[p2].y - lp.y)*f;
						v2.z = (points[p2].z - lp.z)*f;
						
						//draw the polygon
						glBegin(GL_TRIANGLE_STRIP);
							glVertex3d(	points[p1].x,
											points[p1].y,
											points[p1].z);
							glVertex3d(	points[p1].x + v1.x,
											points[p1].y + v1.y,
											points[p1].z + v1.z);

							glVertex3d(	points[p2].x,
											points[p2].y,
											points[p2].z);
							glVertex3d(	points[p2].x + v2.x,
											points[p2].y + v2.y,
											points[p2].z + v2.z);
						glEnd();
					}
				}
		}

		// second pass, stencil operation increases stencil value
		glFrontFace(GL_CW);
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
		for (i=0;  i<nPlanes;  i++)
		{	if (planes[i].visible)
				for (j=0;  j<3;  j++)
				{	k = planes[i].neigh[j];
					if ((!k) || (!planes[k-1].visible))
					{	// here we have an edge, we must draw a polygon
						p1 = planes[i].p[j];
						jj = (j+1)%3;
						p2 = planes[i].p[jj];

						//calculate the length of the vector
						v1.x = (points[p1].x - lp.x)*f;
						v1.y = (points[p1].y - lp.y)*f;
						v1.z = (points[p1].z - lp.z)*f;

						v2.x = (points[p2].x - lp.x)*f;
						v2.y = (points[p2].y - lp.y)*f;
						v2.z = (points[p2].z - lp.z)*f;
						
						//draw the polygon
						glBegin(GL_TRIANGLE_STRIP);
							glVertex3d(	points[p1].x,
											points[p1].y,
											points[p1].z);
							glVertex3d(	points[p1].x + v1.x,
											points[p1].y + v1.y,
											points[p1].z + v1.z);

							glVertex3d(	points[p2].x,
											points[p2].y,
											points[p2].z);
							glVertex3d(	points[p2].x + v2.x,
											points[p2].y + v2.y,
											points[p2].z + v2.z);
						glEnd();
					}
				}
		}
		lp*=1.05f;
	}

	glFrontFace(GL_CCW);
	glColorMask(1, 1, 1, 1);

	//draw a shadowing rectangle covering the entire screen
	glColor4ubv((unsigned char*)&TURTLE->shadowcolor);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glStencilFunc(GL_NOTEQUAL, 0, 0xffffffff);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	glPushMatrix();
	glLoadIdentity();
	glBegin(GL_TRIANGLE_STRIP);
		glVertex3f(-10000, 10000,1);
		glVertex3f(-10000,-10000,1);
		glVertex3f( 10000, 10000,1);
		glVertex3f( 10000,-10000,1);
	glEnd();
	glPopMatrix();
	glDisable(GL_BLEND);

	glDepthFunc(GL_LEQUAL);
	glDepthMask(GL_TRUE);
	glEnable(GL_LIGHTING);
	glDisable(GL_STENCIL_TEST);
	glShadeModel(GL_SMOOTH);
#endif
}
//___________________________________________________________________

Turtle::Turtle(bool makeOld)
{
	ThreadP th=Thread::This();
	size=30; half_bottom=size/3; 
	shown=true;
	cylinderLines=outOfBounds=inEraseMode=false;
	mode=wrapmode;
	heading=wrapDelta=0;

	polyMode=polyShown=tessMode=surfaceMode=false;
	polyPointNr=tessPointNr=0;
	texpos=0;
	tessWindingRule=0;
	surfaceShown=tessShown=false;
	surfaceColumnNr=0;
	shadows=false;
	shadowcolor=Color(0,0,0,0.6);
	sinCos();
	reset3d();
	old=0;
	turtleNode=0;
	shapeNr=0;
	if (theApp && theApp->graph)
		colorTolerance=theApp->graph->colorTolerance;
	if (makeOld)
		old=new Turtle(false);
}

Turtle::Turtle(const Turtle &b)
{
	if (b.old)
		old=new Turtle(false);
	*this=b;
}

Turtle::~Turtle()
{
	if (old)
	{	delete old; old=0;
	}
}

Turtle &Turtle::operator=(const Turtle &b) 
{
	x=b.x; y=b.y; z=b.z;
	pen=b.pen; font=b.font;
	shown=b.shown; outOfBounds=outOfBounds;
	polyMode=b.polyMode;
	polyShown=b.polyShown;
	tessMode=b.tessMode;
	surfaceMode=b.surfaceMode;
	cylinderLines=b.cylinderLines;
	mode=b.mode; heading=b.heading; 
	shoulder=b.shoulder; head=b.head;
	size=b.size;	half_bottom=b.half_bottom; 
	_sin=b._sin; _cos=b._cos; 
	u=b.u;
	hbase=b.hbase;
	surfaceShown=b.surfaceShown;
	tessShown=b.tessShown;
	surfaceColumnNr=b.surfaceColumnNr;
	turtleNode=b.turtleNode;
	shapeNr=b.shapeNr;
	colorTolerance=b.colorTolerance;
	return *this;
}

Turtle &Turtle::operator=(const Point &b)
{
	x=b.x; y=b.y; z=b.z; 
	return *this;
}

void Turtle::create(Recorder *recorder_, Screen_ *screen_)
{
	recorder=recorder_;
	screen=screen_;
	pen.size_rotated=pen.size;//*screen->scale;
	pen.create(pen.color);
}

void Turtle::destroy()
{
	pen.destroy();
}

void Turtle::reset3d()
{
	u=PointMatrix(Point(1,0,0),Point(0,1,0),Point(0,0,1));
	hbase=u;
}


void Turtle::setScreenColor(Color c)
{
	recorder->graph->setColor(c);
}

void Turtle::setColor(Color c)
{
//	draw();
	if (!inEraseMode) 
	{	pen.setColor(c);
	}
//	draw();
}

void Turtle::plainXOrPen()
{
	static Color white(0xffffffffu);
#ifdef OPENGL
	glColor4ubv((unsigned char*)&white);
	glEnable(GL_COLOR_LOGIC_OP);
	glLogicOp(GL_XOR);
#else
	pen.setColor(white);
	recorder->graph->dc->SetLogicalFunction(wxXOR);
#endif
}

void Turtle::setPenSize(double sizeX, double sizeY) 
{
	pen.size.x=sizeX;
	pen.size.y=sizeY;
	pen.size.z=0;
	pen.size_rotated=pen.size;//*screen->scale;

//	double w[2];
//	glGetDoublev(GL_LINE_WIDTH_RANGE, w);
//	pen.wide=(pen.size.x > w[1]);
	pen.wide=(pen.size.x > 0);
		
	pen.setColor(inEraseMode ?   screen->color  :  pen.color);
}

void Turtle::setPenMode(Pen::WinMode newmode) 
{
	int rop2_mode, newpc;
	pen.mode=newmode;
	inEraseMode=newmode==Pen::WIN_ERASE;
	
	if (newmode==Pen::WIN_REVERSE)
		rop2_mode=Pen::XOR_PEN;
	else
		rop2_mode=Pen::COPY_PEN;

	if (rop2_mode==Pen::XOR_PEN)
	{
#ifdef OPENGL
		glEnable(GL_COLOR_LOGIC_OP);
		glLogicOp(GL_INVERT);
#else
	recorder->graph->dc->SetLogicalFunction(wxXOR);
#endif
	}else
	{
#ifdef OPENGL
		glLogicOp(GL_COPY);
		glDisable(GL_COLOR_LOGIC_OP);
		glDisable(GL_LOGIC_OP);
#else
	recorder->graph->dc->SetLogicalFunction(wxCOPY);
#endif
	}
	if (newmode==Pen::WIN_ERASE)
		newpc=screen->color;
	else
		newpc=pen.color;
	
	pen.setColor(newpc);
}

void Turtle::penErase()   {setPenMode(Pen::WIN_ERASE  );}
void Turtle::penDown()    {setPenMode(Pen::WIN_DOWN   );}
void Turtle::penReverse() {setPenMode(Pen::WIN_REVERSE);}

void Turtle::restorePen() 

{
	if (in(screen->Min, screen->Max))
		moveTo();
	setPenMode(pen.mode);
	pen.setColor(inEraseMode ?   screen->color  :  pen.color);
}

NodeP Turtle::getNodePenMode()
{
	if (inEraseMode)
		return new String(_("erase"));
#ifdef OPENGL
	GLboolean logic=glIsEnabled(GL_LOGIC_OP);
	if (logic==GL_TRUE)
	{	GLint mode;
		glGetIntegerv(GL_LOGIC_OP_MODE, &mode);
		if (mode==GL_XOR)
			return new String(_("reverse"));
	}
#endif
	return new String(_("paint"));
}

void Turtle::moveTo() {moveTo(*this);}

void Turtle::moveTo(Point target) 
{
	*this=target;
	if (recorder->recording)
		*recorder << new R_MoveTo(*this,*recorder,target);
}
void Turtle::lineTo(Point target)
{
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
	if (pen.down) 
	{
#ifdef OPENGL
		if (polyMode)
		{	if (polyPointNr==0)
			{	polypt1=target;
				polyc1=pen.color;
				polyPointNr++;
				if (screen->textured) 
					texpt1=texpos;
			}else if (polyPointNr==1)
			{	polypt2=target;
				polyv1=polypt2-polypt1;
				polyc2=pen.color;
				polyPointNr++;
				if (screen->textured) 
					texpt2=texpos;
			}else if (polyPointNr==2)
			{	
				Point 
					polypt3=target,
					&v1=polyv1,
					v2=polypt3-polypt2;
				normal=v1.X(v2);
				double norm=normal.Norm();

				if (fabs(norm) > Float::epsilon)
				{	normal=normal/norm;				
				}
				glBegin(GL_POLYGON);
				if (fabs(norm) > Float::epsilon)
				{	gl2psNormal3fv((float*)&normal);
				}else
				{/*	error(BAD_DATA, 
						new String("polygon normal nullvector"));
					polymode=false;
					return;
				*/}
				if (screen->textured) glTexCoord2fv(floatP(&texpt1));
				glColor4ubv((unsigned char*)&polyc1);
				gl2psNormal3fv((float*)&normal);
				glVertex3fv(floatP(&polypt1));
				if (screen->textured) glTexCoord2fv(floatP(&texpt2));
				glColor4ubv((unsigned char*)&polyc2);
				gl2psNormal3fv((float*)&normal);
				glVertex3fv(floatP(&polypt2));
				if (screen->textured) glTexCoord2fv(floatP(&texpos));
				glColor4ubv((unsigned char*)&pen.color);
				gl2psNormal3fv((float*)&normal);
				glVertex3fv(floatP(&polypt3));
				
				if (recorder->recording)
					*recorder << new R_PolyStart(*this, normal, polypt3, texpos, pen.color);

				if (shadows)
				{	ShadowObject *so=th->graph->shadowObject;
					int &i=so->nPoints, &j=so->nPlanes;
					if (!so->checkPoints(i))
						return;
					so->points[i]=polypt1;
					i++;
					if (!so->checkPoints(i))
						return;
					so->points[i]=polypt2;
					i++;
					if (!so->checkPoints(i))
						return;
					so->points[i]=polypt3;
					i++;
					if (!so->checkPlanes(j))
						return;
					so->planes[j].p[0]=i-3;
					so->planes[j].p[1]=i-2;
					so->planes[j].p[2]=i-1;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=normal;
					j++;
					i++;
					polypt2=polypt3;
				}
				polyPointNr++;
			}else
			{	if (screen->textured) 
					glTexCoord2fv(floatP(&texpos));
				gl2psNormal3fv((float*)&normal);
				glVertex3fv(floatP(&target));
				if (shadows)
				{	ShadowObject *so=th->graph->shadowObject;
					int &i=so->nPoints, &j=so->nPlanes;
					if (!so->checkPoints(i))
						return;
					so->points[i]=polypt1;
					i++;
					if (!so->checkPoints(i))
						return;
					so->points[i]=polypt2;
					i++;
					if (!so->checkPoints(i))
						return;
					so->points[i]=target;
					i++;
					if (!so->checkPlanes(j))
						return;
					so->planes[j].p[0]=i-3;
					so->planes[j].p[1]=i-2;
					so->planes[j].p[2]=i-1;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=normal;
					j++;
					i++;
					polypt1=polypt2;
					polypt2=target;
				}
				if (recorder->recording)
					*recorder << new R_PolyPoint(target, texpos, pen.color);
				*this=target;
			}
		}else if (tessMode)
		{	(*tessFiFo) << target;
			(*tessFiFo) << texpos;
			(*tessFiFo) << DPoint(*(doubleP)&pen.color,0,0);
			tessPointNr++;
		}else if (surfaceMode)
		{	(*surfaceFiFo) << target;
			(*surfaceFiFo) << texpos;
			(*colorFiFo) << pen.color;
		}else
		{	if (screen->perspective)
			{	if (cylinderLines)
				{	glPushMatrix();
					glTranslated(x,y,z);

					Point v=target-(*this);
				//*
					float xz, theta, phi;

					xz=::sqrt(Float::sqr(v.x)+Float::sqr(v.z));
					if (fabs(xz) < Float::epsilon
					 && fabs(v.y) < Float::epsilon)
						theta=0;
					else

						theta=atan2(xz, v.y)/Float::deg2rad;

					if (fabs(v.x) < Float::epsilon
					 && fabs(v.z) < Float::epsilon)
						phi=0;
					else
						phi=atan2(v.x, v.z)/Float::deg2rad;

					glRotated(phi,0,1,0);
					glRotated(theta-90,1,0,0);
				/*/
					glRotated(roll(),0,1,0);
					glRotated(pitch()-90,1,0,0);
					glRotated(getHeading(),0,1,0);
				//*/
					double height=v.Norm();
						GLUquadricObj *q=gluNewQuadric();
						gluQuadricTexture(q, GL_TRUE);
						gluCylinder(q,
							pen.size.x/2,pen.size.x/2, 
							height, int(pen.size.x/4+10),1);
						gluDeleteQuadric(q);
					glPopMatrix();
				}else
				{	if (pen.wide)
					{	glPushMatrix();
						glTranslated(x,y,z);

						Point v=target-(*this),
							ortho=u.x*pen.size.x/2;
						Point
							p1=-ortho,
							p2=v-ortho,
							p3=v+ortho,
							p4=+ortho;

						glBegin(GL_QUADS);
						Point n=-u.z;
						gl2psNormal3fv((float*)&n);
						glVertex3fv((float*)&p1);
						glVertex3fv((float*)&p2);
						glVertex3fv((float*)&p3);
						glVertex3fv((float*)&p4);
						glEnd();

//						glLineWidth(1);
//						glBegin(GL_LINES);
//						glVertex3fv((float*)&p1);
//						glVertex3fv((float*)&p2);
//						glVertex3fv((float*)&p3);
//						glVertex3fv((float*)&p4);
//						glEnd();
						glPopMatrix();
					}
					else
					{	glBegin(GL_LINES);
						glVertex3fv((float*)this);
						glVertex3fv((float*)&target);
						glEnd();
					}
				}
				if (pen.size.x > 0 && pen.roundLineEnds)
				{	glPushMatrix();
					glTranslated(x,y,z);
					if (cylinderLines)
					{	GLUquadricObj *q=gluNewQuadric();
						gluSphere(q, pen.size.x/2, 
							int(pen.size.x/8+8), int(pen.size.x/8+8));
//						gluDeleteQuadric(q);
						glPopMatrix();

						glPushMatrix();
						glTranslated(target.x,target.y,target.z);
	
//						q=gluNewQuadric();
						gluSphere(q, pen.size.x/2, 
							int(pen.size.x/8+8), int(pen.size.x/8+8));
						gluDeleteQuadric(q);
					}
					else
					{	double r=roll(), p=pitch(), h=-getHeading();

						glRotated(r,0,1,0);
						glRotated(p,1,0,0);
						glRotated(h,0,0,1);
					
						GLUquadricObj *q=gluNewQuadric();
						if (screen->textured)
							gluQuadricTexture(q, GL_TRUE);
						gluPartialDisk(q, 0, pen.size.x/2, 
							int(3*pen.size.x), 1, 90, 180);
						glPopMatrix();

						glPushMatrix();
						glTranslated(target.x,target.y,target.z);
						glRotated(r,0,1,0);
						glRotated(p,1,0,0);
						glRotated(h,0,0,1);
					
						gluPartialDisk(q, 0, pen.size.x/2, 
							int(3*pen.size.x), 1, -90, 180);

						gluDeleteQuadric(q);
					}
					glPopMatrix();
				}
				if (cylinderLines)
				{	if (recorder->recording)
					{	*recorder << new R_LineTo3d(*this, *recorder, target);
			//			mySwapBuffers();
					}
				}else
					if (recorder->recording)
					{	*recorder << new R_LineTo(*this, *recorder, target);
			//			mySwapBuffers();
					}
			}
			else // !screen->perspective
			{	if (pen.wide)
				{	Point ortho=Point(target.y-this->y, this->x-target.x, 0);
					double on=ortho.Norm();
					if (on)
						ortho=(ortho/on/2)*pen.size.x;
					else
						ortho=(ortho/2)*pen.size.x;
					Point
						p1=Point(*this)+ortho,
						p2=target+ortho,
						p3=target-ortho,
						p4=Point(*this)-ortho;

					glBegin(GL_QUADS);
					glVertex3fv((float*)&p1);
					glVertex3fv((float*)&p2);
					glVertex3fv((float*)&p3);
					glVertex3fv((float*)&p4);
					glEnd();
					glLineWidth(1);
					glBegin(GL_LINES);
					glVertex3fv((float*)&p1);
					glVertex3fv((float*)&p2);
					glVertex3fv((float*)&p3);
					glVertex3fv((float*)&p4);
					glEnd();
				}else
				{	glBegin(GL_LINES);
					glVertex3fv((float*)this);
					glVertex3fv((float*)&target);
					glEnd();
				}
				if (pen.size.x > 0 && pen.roundLineEnds)
				{	glPushMatrix();
					glTranslated(x,y,z);
					double startangle=
						-atan2(target.y-y, target.x-x)/Float::deg2rad;

						GLUquadricObj *q=gluNewQuadric();
						gluPartialDisk(q, 0, 
							pen.size.x/2, int(3*pen.size.x/2), 1, startangle+180, 180);
						gluDeleteQuadric(q);
					glPopMatrix();

					glPushMatrix();
					glTranslated(target.x,target.y,target.z);
						q=gluNewQuadric();
						gluPartialDisk(q, 0, 
							pen.size.x/2, int(3*pen.size.x/2), 1, startangle, 180);
						gluDeleteQuadric(q);
					glPopMatrix();
				}
				if (recorder->recording)
				{	*recorder << new R_LineTo(*this, *recorder, target);
				}
			}
		}
#else
	th->graph->dc->DrawLine(x,y, target.x,target.y);
	if (recorder->recording)
		*recorder << new R_LineTo(*this, *recorder, target);
#endif
	}
	else
	{	if (recorder->recording)
			*recorder << new R_MoveTo(*this, *recorder, target);
	}
//	mySwapBuffers();
	Float::reset_FPU();
}

void Turtle::moveTo(double x, double y) {moveTo(Point(x,y,z));}
void Turtle::lineTo(double x, double y) {lineTo(Point(x,y,z));}

void Turtle::polyStart()
{
//	draw();
	polyShown=shown;
	shown=false;
	polyMode=true;
	polyPointNr=0;
}

void Turtle::polyEnd()
{
	polyMode=false;
#ifdef OPENGL
	glEnd();
#endif
	shown=polyShown;
//	draw();
	if (recorder->recording)
		*recorder << new R_PolyEnd();
}

#ifdef OPENGL
void CALLBACK Turtle::tessglBegin(GLenum prim)
{	ThreadP th=Thread::This();
	glBegin(prim);
	gl2psNormal3dv(doubleP(th->tessNormal));
	if (TURTLE->shadows && th->graph)
	{	th->tess_nPoints=th->graph->shadowObject->nPoints;
		th->tess_prim=prim;
	}
	if (RECORDER.recording)
		RECORDER << new R_glBegin(prim, Point(*th->tessNormal));
}
void CALLBACK Turtle::tessglVertex(void *data)
{	DPointP p=DPointP(data), t=p+1;
	ColorP c=(ColorP)(p+2);
	glColor4ubv((GLubyte*)c);
	glTexCoord2dv(doubleP(t));
	glVertex3dv(doubleP(p));
/*	if (c->r==0 && c->b==0)
	{	c_out 
		<< p->x << ' '
		<< p->y << ' '
		<< p->z << ' '
		<< (int)c->r << ' ' 
		<< (int)c->g << ' ' 
		<< (int)c->b << ' ' 
		<< (int)c->a << endL;
	}
*/
	ThreadP th=Thread::This();
	if (TURTLE->shadows && th->graph)
	{	ShadowObject *so=th->graph->shadowObject;
		int &i=so->nPoints, &j=so->nPlanes;
		if (!so->checkPoints(i))
			return;
		so->points[i]=*p;
		i++;
	}
	if (RECORDER.recording)
	{	Point pp(*p), tt(*t);
		RECORDER << new R_TessPoint(pp, tt, *c);
	}
}
void CALLBACK Turtle::tessglEnd()
{	ThreadP th=Thread::This();
	glEnd();
	if (TURTLE->shadows && th->graph)
	{	ShadowObject *so=th->graph->shadowObject;
		int i=th->tess_nPoints, &j=so->nPlanes;
		switch (th->tess_prim)
		{
		case GL_TRIANGLE_STRIP:
			while (i < so->nPoints-2)
			{	int k;
				if (!so->checkPlanes(j))
					return;
				so->planes[j].p[0]=i;
				so->planes[j].p[1]=i+1;
				so->planes[j].p[2]=i+2;
				for (k=0;  k<3;  k++)
					so->planes[j].normals[k]=*th->tessNormal;
				j++;
				i++;
				if (i >= so->nPoints-2)
					break;
				if (!so->checkPlanes(j))
					return;
				so->planes[j].p[0]=i+1;
				so->planes[j].p[1]=i;
				so->planes[j].p[2]=i+2;
				for (k=0;  k<3;  k++)
					so->planes[j].normals[k]=*th->tessNormal;
				j++;
				i++;
			}
			break;
		case GL_TRIANGLE_FAN:
			{	int i0=th->tess_nPoints;
				while (i < so->nPoints-2)
				{	if (!so->checkPlanes(j))
						return;
					so->planes[j].p[0]=i0;
					so->planes[j].p[1]=i+1;
					so->planes[j].p[2]=i+2;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=*th->tessNormal;
					j++;
					i++;
				}
			}
			break;
		case GL_TRIANGLES:
			{	while (i < so->nPoints-2)
				{	if (!so->checkPlanes(j))
						return;
					so->planes[j].p[0]=i;
					so->planes[j].p[1]=i+1;
					so->planes[j].p[2]=i+2;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=*th->tessNormal;
					j++;
					i+=3;
				}
			}
			break;
		default:
			i=i;
		}
	}
	if (RECORDER.recording)
		RECORDER << new R_PolyEnd();
}

void CALLBACK Turtle::tessglCombine(DPointP p, 
												DPointDPointColorP d[4], 
												GLfloat w[4], void **out)
{
	ThreadP th=Thread::This();
//c_out << " x0=" << d[0]->p.x << "  y0=" << d[0]->p.y << "  z0=" << d[0]->p.z << " w0=" << w[0] << endL;
//c_out << " x1=" << d[1]->p.x << "  y1=" << d[1]->p.y << "  z1=" << d[1]->p.z << " w1=" << w[1] << endL;
//c_out << " x2=" << d[2]->p.x << "  y2=" << d[2]->p.y << "  z2=" << d[2]->p.z << " w2=" << w[2] << endL;
//c_out << " x3=" << d[3]->p.x << "  y3=" << d[3]->p.y << "  z3=" << d[3]->p.z << " w3=" << w[3] << endL;
//c_out << endL;

//c_out << "tx0=" << d[0]->t.x << " ty0=" << d[0]->t.y << " w0=" << w[0] << endL;
//c_out << "tx1=" << d[1]->t.x << " ty1=" << d[1]->t.y << " w1=" << w[1] << endL;
//c_out << "tx2=" << d[2]->t.x << " ty2=" << d[2]->t.y << " w2=" << w[2] << endL;
//c_out << "tx3=" << d[3]->t.x << " ty3=" << d[3]->t.y << " w3=" << w[3] << endL;
//c_out << endL;

//c_out << " r=" << int(d[0]->c.r) << "  g=" << int(d[0]->c.g) << "  b=" << int(d[0]->c.b) << " a=" << int(d[0]->c.a) << endL;

	(*TURTLE->tessFiFo) << (*p);
	DPoint tex;
	Color c;
	if (d[3])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r + w[1]*d[1]->c.r + w[2]*d[2]->c.r + w[3]*d[3]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g + w[1]*d[1]->c.g + w[2]*d[2]->c.g + w[3]*d[3]->c.g);
		c.b=(unsigned char)(w[0]*d[0]->c.b + w[1]*d[1]->c.b + w[2]*d[2]->c.b + w[3]*d[3]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a + w[1]*d[1]->c.a + w[2]*d[2]->c.a + w[3]*d[3]->c.a);
	}else if (d[2])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r + w[1]*d[1]->c.r + w[2]*d[2]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g + w[1]*d[1]->c.g + w[2]*d[2]->c.g);
		c.b=(unsigned char)(w[0]*d[0]->c.b + w[1]*d[1]->c.b + w[2]*d[2]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a + w[1]*d[1]->c.a + w[2]*d[2]->c.a);
	}else if (d[1])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r + w[1]*d[1]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g + w[1]*d[1]->c.g);

		c.b=(unsigned char)(w[0]*d[0]->c.b + w[1]*d[1]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a + w[1]*d[1]->c.a);
	}else if (d[0])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g);
		c.b=(unsigned char)(w[0]*d[0]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a);
	}
	if (d[3])
	{	tex.x=(w[0]*d[0]->t.x + w[1]*d[1]->t.x + w[2]*d[2]->t.x + w[3]*d[3]->t.x);
		tex.y=(w[0]*d[0]->t.y + w[1]*d[1]->t.y + w[2]*d[2]->t.y + w[3]*d[3]->t.y);
	}else if (d[2])
	{	tex.x=(w[0]*d[0]->t.x + w[1]*d[1]->t.x + w[2]*d[2]->t.x);
		tex.y=(w[0]*d[0]->t.y + w[1]*d[1]->t.y + w[2]*d[2]->t.y);
	}else if (d[1])
	{	tex.x=(w[0]*d[0]->t.x + w[1]*d[1]->t.x);
		tex.y=(w[0]*d[0]->t.y + w[1]*d[1]->t.y);
	}else if (d[0])
	{	tex.x=(w[0]*d[0]->t.x);
		tex.y=(w[0]*d[0]->t.y);
	}
	tex.z=0;
//c_out << "tx=" << tex.x << " ty=" << tex.y << endL;
//c_out << endL;
	(*TURTLE->tessFiFo) << tex;
	(*TURTLE->tessFiFo) << DPoint(*(doubleP)&c,0,0);
	if (TURTLE->tessFiFo->in-3 >= TURTLE->tessFiFo->base)
		*out = TURTLE->tessFiFo->in-3; 
	else
		*out = TURTLE->tessFiFo->max-3; 
}

void CALLBACK Turtle::tessglError(GLenum errorCode)
{
   const char *estring;

   estring = (char*)gluErrorString(errorCode);
   c_out << _("Tessellation Error: ") << estring << endL;
}
#endif

void Turtle::tessStart()
{
//	draw();
	tessShown=shown;
	shown=false;
	tessMode=true;
	tessFiFo=new DPointFiFo(0x1000);
	if (!tessFiFo)
	{	error(OUT_OF_MEM); return;
	}
/*	(*tessFiFo) << *this;
	(*tessFiFo) << DPoint(*(double*)&pen.color,0,0);
*/	tessPointNr=0;

#ifdef OPENGL
	tess=gluNewTess();
	gluTessCallback(tess, GLU_TESS_BEGIN, (void (CALLBACK *) ())tessglBegin);  
	gluTessCallback(tess, GLU_TESS_COMBINE, (void (CALLBACK *) ())tessglCombine);
	gluTessCallback(tess, GLU_TESS_VERTEX, (void (CALLBACK *) ())tessglVertex);  
	gluTessCallback(tess, GLU_TESS_END, tessglEnd);  
	gluTessCallback(tess, GLU_TESS_ERROR, (void (CALLBACK *) ())tessglError);

	gluTessProperty(tess, GLU_TESS_WINDING_RULE, tessWindingRule);
	gluTessBeginPolygon(tess, 0);
#endif
}

void Turtle::tessDraw()
{
	if (tessPointNr < 3)
	{	error(BAD_DATA, new String(
			_("Need at least three points for tesselation!")));
		return;
	}
	ThreadP th=Thread::This();

	DPointP p1,p2,p3;
	p1=tessFiFo->out++;
	tessFiFo->out++;
	tessFiFo->out++;
	p2=tessFiFo->out++;
	tessFiFo->out++;
	tessFiFo->out++;
	p3=tessFiFo->out++;
	tessFiFo->out++;
	tessFiFo->out++;

	DPoint v1=*p2-*p1, v2=*p3-*p2;
	*th->tessNormal=v1.X(v2);
	double norm=th->tessNormal->Norm();

	if (norm > Float::epsilon)
		*th->tessNormal=(*th->tessNormal)/norm;				
	
#ifdef OPENGL
	gluTessBeginContour(tess);
	if (norm > Float::epsilon)
		gluTessNormal(tess, th->tessNormal->x,th->tessNormal->y,th->tessNormal->z);
	
	gluTessVertex(tess, doubleP(p1), doubleP(p1));
	gluTessVertex(tess, doubleP(p2), doubleP(p2));
	gluTessVertex(tess, doubleP(p3), doubleP(p3));

	while (tessFiFo->out < tessFiFo->in)
	{	DPointP p=tessFiFo->out++;	tessFiFo->out++;	tessFiFo->out++;
		gluTessVertex(tess, doubleP(p), doubleP(p));
	}
#endif
}

void Turtle::tessContour()
{
	if (!tessMode)
		return;

	tessDraw();
#ifdef OPENGL
	gluTessEndContour(tess); 
#endif
}

void Turtle::tessEnd()
{
	if (!tessMode)
		return;

	tessDraw();
	tessMode=false;
#ifdef OPENGL
	gluTessEndContour(tess); 
	gluTessEndPolygon(tess); 
	gluDeleteTess(tess);
#endif
	delete tessFiFo;
	shown=tessShown;
//	draw();
}

void Turtle::surfaceStart()
{
//	draw();
	surfaceShown=shown;
	shown=false;
	surfaceMode=true;
	surfaceFiFo=new PointFiFo(0x1000);
	if (!surfaceFiFo)
	{	error(OUT_OF_MEM); return;
	}
	surfaceFiFo2=new PointFiFo(0x1000);
	if (!surfaceFiFo2)
	{	delete surfaceFiFo;
		error(OUT_OF_MEM); return;
	}
	surfaceFiFo3=new PointFiFo(0x1000);
	if (!surfaceFiFo3)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		error(OUT_OF_MEM); return;
	}
	surfaceFiFo00=new PointFiFo(0x1000);
	if (!surfaceFiFo00)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		delete surfaceFiFo3;
		error(OUT_OF_MEM); return;
	}
	surfaceFiFo01=new PointFiFo(0x1000);
	if (!surfaceFiFo01)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		delete surfaceFiFo3;
		delete surfaceFiFo00;
		error(OUT_OF_MEM); return;
	}
	surfaceFiFo02=new PointFiFo(0x1000);
	if (!surfaceFiFo02)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		delete surfaceFiFo3;
		delete surfaceFiFo00;
		delete surfaceFiFo01;
		error(OUT_OF_MEM); return;
	}
	colorFiFo=new ColorFiFo(0x1000);
	if (!colorFiFo)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		delete surfaceFiFo3;
		delete surfaceFiFo00;
		delete surfaceFiFo01;
		delete surfaceFiFo02;
		error(OUT_OF_MEM); return;
	}
	colorFiFo2=new ColorFiFo(0x1000);
	if (!colorFiFo2)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		delete surfaceFiFo3;
		delete surfaceFiFo00;
		delete surfaceFiFo01;
		delete surfaceFiFo02;
		delete colorFiFo;
		error(OUT_OF_MEM); return;
	}
	colorFiFo00=new ColorFiFo(0x1000);
	if (!colorFiFo00)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		delete surfaceFiFo3;
		delete surfaceFiFo00;
		delete surfaceFiFo01;
		delete surfaceFiFo02;
		delete colorFiFo;
		delete colorFiFo2;
		error(OUT_OF_MEM); return;
	}
	colorFiFo01=new ColorFiFo(0x1000);
	if (!colorFiFo01)
	{	delete surfaceFiFo;
		delete surfaceFiFo2;
		delete surfaceFiFo3;
		delete surfaceFiFo00;
		delete surfaceFiFo01;
		delete surfaceFiFo02;
		delete colorFiFo;
		delete colorFiFo2;
		delete colorFiFo00;
		error(OUT_OF_MEM); return;
	}
	surfaceColumnNr=0;
}

void Turtle::surfaceColumn()
{
	if (!surfaceMode)
		return;

	if (shadows)
	{	surfaceColumnSh();
		return;
	}

	PointFiFo *h=surfaceFiFo;
	surfaceFiFo=surfaceFiFo2;
	surfaceFiFo2=surfaceFiFo3;
	surfaceFiFo3=h;

	ColorFiFo *hc=colorFiFo;
	colorFiFo=colorFiFo2;
	colorFiFo2=hc;

	if (surfaceColumnNr==0)
	{	*surfaceFiFo00=*surfaceFiFo3;
		*colorFiFo00=*colorFiFo2;
	}else if (surfaceColumnNr==1)
	{	*surfaceFiFo01=*surfaceFiFo3;
		*colorFiFo01=*colorFiFo2;
	}else if (surfaceColumnNr==2)
	{	*surfaceFiFo02=*surfaceFiFo3;
	}
	if (surfaceColumnNr >= 2)
	{
#ifdef OPENGL
		glBegin(GL_QUAD_STRIP);
#endif
		Point p1,t1, p2,t2, p3,t3, p4,t4, o1,o3, dummy;
		(*surfaceFiFo3) >> p1 >> t1 >> p3 >> t3;
		(*surfaceFiFo2) >> p2 >> t2 >> p4 >> t4;
		(*surfaceFiFo ) >> o1 >> dummy >> o3 >> dummy;
		Color c1,c2,c3,c4;
		(*colorFiFo2) >> c1 >> c3;
		(*colorFiFo ) >> c2 >> c4;

#ifdef OPENGL
		Point normal=(p1-p2).X(p2-p4);
		double norm=normal.Norm();
		if (norm > Float::epsilon)
		{	normal/=norm;
			gl2psNormal3fv((float*)&normal);
		}
		glColor4ubv((unsigned char*)&c1);
		glTexCoord2fv(floatP(&t1));
		glVertex3fv(floatP(&p1));

		Point onormal=(p2-o1).X(p2-p4);
		double onorm=onormal.Norm();
		if (onorm > Float::epsilon)
		{	onormal/=onorm;
			gl2psNormal3fv((float*)&onormal);
		}
		glColor4ubv((unsigned char*)&c2);
		glTexCoord2fv(floatP(&t2));
		glVertex3fv(floatP(&p2));

		normal=(p3-p1).X(p3-p4);
		norm=normal.Norm();
		if (norm > Float::epsilon)
		{	normal/=norm;
			gl2psNormal3fv((float*)&normal);
		}
		glColor4ubv((unsigned char*)&c3);
		glTexCoord2fv(floatP(&t3));
		glVertex3fv(floatP(&p3));

		onormal=(p4-p2).X(p4-o3);
		onorm=onormal.Norm();
		if (onorm > Float::epsilon)
		{	onormal/=onorm;
			gl2psNormal3fv((float*)&onormal);
		}
		glColor4ubv((unsigned char*)&c4);
		glTexCoord2fv(floatP(&t4));
		glVertex3fv(floatP(&p4));
		if (recorder->recording)
			*recorder << new R_QuadStripStart(normal,onormal, 
				c1,p1,t1, c2,p2,t2, c3,p3,t3, c4,p4,t4);

		c1=c3;	p1=p3;	t1=t3;
		c2=c4;	p2=p4;	t2=t4;
		o1=o3;

		while (*surfaceFiFo && *surfaceFiFo2)
		{	(*surfaceFiFo3) >> p3 >> t3;
			(*surfaceFiFo2) >> p4 >> t4;
			(*surfaceFiFo ) >> o3 >> dummy;
			(*colorFiFo2) >> c3;
			(*colorFiFo ) >> c4;

			normal=(p3-p1).X(p3-p4);
			norm=normal.Norm();
			if (norm > Float::epsilon)
			{	normal/=norm;
				gl2psNormal3fv((float*)&normal);
			}
			glColor4ubv((unsigned char*)&c3);
			glTexCoord2fv(floatP(&t3));
			glVertex3fv(floatP(&p3));

			Point onormal;
			onormal=(p4-p2).X(p4-o3);
			double onorm=onormal.Norm();
			if (onorm > Float::epsilon)
			{	onormal/=onorm;
				gl2psNormal3fv((float*)&onormal);
			}
			glColor4ubv((unsigned char*)&c4);
			glTexCoord2fv(floatP(&t4));
			glVertex3fv(floatP(&p4));

			if (recorder->recording)
				*recorder << new R_QuadStrip(
					normal,onormal, c3,p3,t3, c4,p4,t4);
			c1=c3;	p1=p3;	t1=t3;
			c2=c4;	p2=p4;	t2=t4;
			o1=o3;
		}
		glEnd();
		if (recorder->recording)
			*recorder << new R_QuadStripEnd();
#endif
		surfaceFiFo3->reset_out();
		surfaceFiFo2->reset_out();
		surfaceFiFo->reset();
		colorFiFo2->reset_out();
		colorFiFo->reset();
//mySwapBuffers();
	}
	surfaceColumnNr++;
}

void Turtle::surfaceColumnSh()
{
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
	if (!surfaceMode)
		return;

	PointFiFo *h=surfaceFiFo;
	surfaceFiFo=surfaceFiFo2;
	surfaceFiFo2=surfaceFiFo3;
	surfaceFiFo3=h;

	ColorFiFo *hc=colorFiFo;
	colorFiFo=colorFiFo2;
	colorFiFo2=hc;

	if (surfaceColumnNr==0)
	{	*surfaceFiFo00=*surfaceFiFo3;
		*colorFiFo00=*colorFiFo2;
	}else if (surfaceColumnNr==1)
	{	*surfaceFiFo01=*surfaceFiFo3;
		*colorFiFo01=*colorFiFo2;
	}else if (surfaceColumnNr==2)
	{	*surfaceFiFo02=*surfaceFiFo3;
	}
	if (surfaceColumnNr >= 2)
	{
#ifdef OPENGL
		glBegin(GL_QUAD_STRIP);

		Point p1,t1, p2,t2, p3,t3, p4,t4, o1,o3, dummy;
		(*surfaceFiFo3) >> p1 >> t1 >> p3 >> t3;
		(*surfaceFiFo2) >> p2 >> t2 >> p4 >> t4;
		(*surfaceFiFo ) >> o1 >> dummy >> o3 >> dummy;
		Color c1,c2,c3,c4;
		(*colorFiFo2) >> c1 >> c3;
		(*colorFiFo ) >> c2 >> c4;

		Point normal2=(p1-p2).X(p2-p4);
		double norm=normal2.Norm();
		if (norm > Float::epsilon)
		{	normal2/=norm;
			gl2psNormal3fv((float*)&normal2);
		}
		glColor4ubv((unsigned char*)&c1);
		glTexCoord2fv(floatP(&t1));
		glVertex3fv(floatP(&p1));
		ShadowObject *so=th->graph->shadowObject;
		int i=so->nPoints,
			j=so->nPlanes;
		if (!so->checkPoints(i))
			return;
		so->points[i]=p1;
		so->colors[i]=c1;
		so->texs[i]=t1;
		i++;

		Point onormal2=(p2-o1).X(p2-p4);
		double onorm=onormal2.Norm();
		if (onorm > Float::epsilon)
		{	onormal2/=onorm;
			gl2psNormal3fv((float*)&onormal2);
		}
		glColor4ubv((unsigned char*)&c2);
		glTexCoord2fv(floatP(&t2));
		glVertex3fv(floatP(&p2));
		if (!so->checkPoints(i))
			return;
		so->points[i]=p2;
		so->colors[i]=c2;
		so->texs[i]=t2;
		i++;

		Point normal=(p3-p1).X(p3-p4);
		norm=normal.Norm();
		if (norm > Float::epsilon)
		{	normal/=norm;
			gl2psNormal3fv((float*)&normal);
		}
		glColor4ubv((unsigned char*)&c3);
		glTexCoord2fv(floatP(&t3));
		glVertex3fv(floatP(&p3));
		if (!so->checkPoints(i))
			return;
		so->points[i]=p3;
		so->colors[i]=c3;
		so->texs[i]=t3;
		i++;

		Point onormal=(p4-p2).X(p4-o3);
		onorm=onormal.Norm();
		if (onorm > Float::epsilon)
		{	onormal/=onorm;
			gl2psNormal3fv((float*)&onormal);
		}
		glColor4ubv((unsigned char*)&c4);
		glTexCoord2fv(floatP(&t4));
		glVertex3fv(floatP(&p4));
		if (!so->checkPoints(i))
			return;
		so->points[i]=p4;
		so->colors[i]=c4;
		so->texs[i]=t4;
		i++;

		if (!so->checkPlanes(j))
			return;
		so->planes[j].p[2]=i-1;
		so->planes[j].p[1]=i-2;
		so->planes[j].p[0]=i-3;
		so->planes[j].normals[2]=onormal;
		so->planes[j].normals[1]=normal;
		so->planes[j].normals[0]=onormal2;
		so->planes[j].neigh[0]=j+2;
		so->planes[j].neigh[1]=j+4;
		j++;

		if (!so->checkPlanes(j))
			return;
		so->planes[j].p[2]=i-4;
		so->planes[j].p[1]=i-3;
		so->planes[j].p[0]=i-2;
		so->planes[j].normals[2]=normal2;
		so->planes[j].normals[1]=onormal2;
		so->planes[j].normals[0]=normal;
		so->planes[j].neigh[0]=j;
		j++;

		if (recorder->recording)
			*recorder << new R_QuadStripStart(normal,onormal, 
				c1,p1,t1, c2,p2,t2, c3,p3,t3, c4,p4,t4);

		c1=c3;	p1=p3;	t1=t3;
		c2=c4;	p2=p4;	t2=t4;
		o1=o3;
		normal2=normal;
		onormal2=onormal;

		while (*surfaceFiFo && *surfaceFiFo2)
		{	(*surfaceFiFo3) >> p3 >> t3;
			(*surfaceFiFo2) >> p4 >> t4;
			(*surfaceFiFo ) >> o3 >> dummy;
			(*colorFiFo2) >> c3;
			(*colorFiFo ) >> c4;

			normal=(p3-p1).X(p3-p4);
			norm=normal.Norm();
			if (norm > Float::epsilon)
			{	normal/=norm;
				gl2psNormal3fv((float*)&normal);
			}
			glColor4ubv((unsigned char*)&c3);
			glTexCoord2fv(floatP(&t3));
			glVertex3fv(floatP(&p3));
			if (!so->checkPoints(i))
				return;
			so->points[i]=p3;
			so->colors[i]=c3;
			so->texs[i]=t3;
			i++;

			Point onormal;
			onormal=(p4-p2).X(p4-o3);
			double onorm=onormal.Norm();
			if (onorm > Float::epsilon)
			{	onormal/=onorm;
				gl2psNormal3fv((float*)&onormal);
			}
			glColor4ubv((unsigned char*)&c4);
			glTexCoord2fv(floatP(&t4));
			glVertex3fv(floatP(&p4));
			if (!so->checkPoints(i))
				return;
			so->points[i]=p4;
			so->colors[i]=c4;
			so->texs[i]=t4;
			i++;

			if (!so->checkPlanes(j))
				return;
			so->planes[j].p[2]=i-1;
			so->planes[j].p[1]=i-2;
			so->planes[j].p[0]=i-3;
			so->planes[j].normals[2]=onormal;
			so->planes[j].normals[1]=normal;
			so->planes[j].normals[0]=onormal2;
			so->planes[j].neigh[0]=j+2;
			if (*surfaceFiFo)
				so->planes[j].neigh[1]=j+4;
			j++;

			if (!so->checkPlanes(j))
				return;
			so->planes[j].p[2]=i-4;
			so->planes[j].p[1]=i-3;
			so->planes[j].p[0]=i-2;
			so->planes[j].normals[2]=normal2;
			so->planes[j].normals[1]=onormal2;
			so->planes[j].normals[0]=normal;
			so->planes[j].neigh[0]=j;
			so->planes[j].neigh[1]=j-2;
			j++;

			if (recorder->recording)
				*recorder << new R_QuadStrip(
					normal,onormal, c3,p3,t3, c4,p4,t4);
			c1=c3;	p1=p3;	t1=t3;
			c2=c4;	p2=p4;	t2=t4;
			o1=o3;
			normal2=normal;
			onormal2=onormal;
		}
		glEnd();
		if (recorder->recording)
			*recorder << new R_QuadStripEnd();
		so->nPoints=i;
		so->nPlanes=j;
#endif
		surfaceFiFo3->reset_out();
		surfaceFiFo2->reset_out();
		surfaceFiFo->reset();
		colorFiFo2->reset_out();
		colorFiFo->reset();
//mySwapBuffers();
	}
	surfaceColumnNr++;
}

void Turtle::surfaceEnd()
{
	if (!surfaceMode)
		return;

	delete surfaceFiFo;
	delete surfaceFiFo2;
	delete surfaceFiFo3;
	delete colorFiFo;
	delete colorFiFo2;

	surfaceFiFo2=surfaceFiFo02;
	surfaceFiFo3=surfaceFiFo01;
	surfaceFiFo=surfaceFiFo00;

	colorFiFo2=colorFiFo01;
	colorFiFo=colorFiFo00;

	if (shadows)
		surfaceColumnSh();
	else
		surfaceColumn();

	surfaceMode=false;
	delete surfaceFiFo00;
	delete surfaceFiFo01;
	delete surfaceFiFo02;
	delete colorFiFo00;
	delete colorFiFo01;
	shown=surfaceShown;
//	draw();
}

void Turtle::setPenSize(Point size) {setPenSize(size.x, size.y);}

inline void Turtle::sinCos()
{
	double real_heading=(-heading + 90)*Float::deg2rad;
	_cos= Float::cutError(cos(real_heading));
	_sin= Float::cutError(sin(real_heading));
	shoulder=Point(_sin,-_cos, 0)*half_bottom;
		 head=Point(_cos, _sin, 0)*size;
}

void Turtle::drawRaw() 
{
	*old = *this;
//	pen.color=0xff000000;
	pen.color=0xffffffff;
	plainXOrPen();
	pen.down=true;
#ifdef OPENGL
	setPenSize(0,0);
	glLineWidth(1);
	GLboolean ls;
	glGetBooleanv(GL_LINE_SMOOTH, &ls);
	glDisable(GL_LINE_SMOOTH);
	bool isdt=(glIsEnabled(GL_DEPTH_TEST)==GL_TRUE);
	if (isdt)
		glDisable(GL_DEPTH_TEST);
#else
	setPenSize(1,1);
#endif
   Point
		 left = *this-shoulder,
		right = *this+shoulder,
		  top = *this+head;

	bool r=recorder->recording;
	recorder->recording=false;

	moveTo(right);	*this=right;
	lineTo(left);	*this=left;
	lineTo(top);	*this=top;
	lineTo(right);	*this=right;
	moveTo(*old);	*this=*old;

#ifdef OPENGL
	if (ls)
		glEnable(GL_LINE_SMOOTH);
	if (isdt)
		glEnable(GL_DEPTH_TEST);
#endif
	restorePen();
	setPenSize(old->pen.size.x, old->pen.size.y);

	recorder->recording=r;
}

void Turtle::drawp() 
{
	if (shown)
	{	Turtle saved(false);
		saved = *this;
		pen.color=0xffffffff;
		pen.down=true;
		setPenSize(0,0);
		plainXOrPen();

#ifdef OPENGL
		glLineWidth(1);
		GLboolean isls, isdt, isli;
		glGetBooleanv(GL_LINE_SMOOTH, &isls);
		glGetBooleanv(GL_LIGHTING, &isli);
		glGetBooleanv(GL_DEPTH_TEST, &isdt);
		glDisable(GL_LINE_SMOOTH);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
#endif
		Point
			 left = *this-u.x*half_bottom,
			right = *this+u.x*half_bottom,
			  top = *this+u.y*size;

		bool r=recorder->recording;
		recorder->recording=false;

		moveTo(right);	*this=right;
		lineTo(left);	*this=left;
		lineTo(top);	*this=top;
		lineTo(right);	*this=right;
		moveTo(saved);	*this = saved;
		
#ifdef OPENGL
		if (isls)
			glEnable(GL_LINE_SMOOTH);
		if (isli)
			glEnable(GL_LIGHTING);
		if (isdt)
			glEnable(GL_DEPTH_TEST);
#endif
		restorePen();

		recorder->recording=r;
	}
}

void Turtle::checkMinX() {if (x-size < screen->Min.x) {x+=screen->size.x; drawRaw(); x-=screen->size.x;}}
void Turtle::checkMaxX() {if (x+size >=screen->Max.x) {x-=screen->size.x; drawRaw(); x+=screen->size.x;}}
void Turtle::checkMinY() {if (y-size < screen->Min.y) {y+=screen->size.y; drawRaw(); checkMaxX(); checkMinX(); y-=screen->size.y;}}
void Turtle::checkMaxY() {if (y+size >=screen->Max.y) {y-=screen->size.y; drawRaw(); checkMaxX(); checkMinX(); y+=screen->size.y;}}

void Turtle::getInWindow() 
{	if (screen->size.x!=0.0 && screen->size.y!=0.0 && screen->size.z!=0.0)
		*this=(*this-screen->Min) % screen->size + screen->Min;	// get *this in window
}

void Turtle::draw() 
{
	if (turtleNode && turtleNode->shapes)
	{	ArrayP shapes=turtleNode->shapes;
		if (shapeNr >= shapes->count())
			shapeNr=0;
		if (shapes->data()[shapeNr])
			shapes->data()[shapeNr]->draw(this);
	}else
	{	if (screen->perspective)
			drawp();
		else if (shown)
		{	if (mode==wrapmode)
			{	if (!in(screen->Min, screen->Max))
				{	Point saved=*this;
					getInWindow();
					checkMinX(); checkMaxX(); checkMinY(); checkMaxY(); drawRaw();
					*this=saved;
				}else
				{	checkMinX(); checkMaxX(); checkMinY(); checkMaxY(); drawRaw();
				}
			}else 
				drawRaw();
		}
	}
}

void Turtle::setHeading(double angle) 
{
	if (screen->perspective)
	{
/*		double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
		Point
		 uy=hbase.y*c+hbase.x*s;
		drawp();

		u.x=hbase.x*c-hbase.y*s;
		u.y=uy;
		u.z=hbase.z;
		drawp();
		heading = Float::mod(angle, 360);
*/		right(angle-getHeading());
	}
	else
	{	//if (angle != heading)
		{//draw();
			heading = Float::mod(angle, 360);
			sinCos();
		//	draw();
		}
	}
}

void Turtle::right(double angle) 
{	
	if (screen->perspective)
	{	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
		Point 
		 uy=u.y*c+u.x*s;
//		drawp();
		u.x=u.x*c-u.y*s;
		u.y=uy;
//		drawp();
		heading+=Float::mod(angle, 360);

	}
	else
		setHeading(heading+angle);
}

void Turtle::upPitch(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point 
	 uz=u.z*c-u.y*s;
//	drawp();
	u.y=u.y*c+u.z*s;
	u.z=uz;
//	drawp();
	hbase=u;
	heading=0;
}

void Turtle::rightRoll(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point 
	 uz=u.z*c+u.x*s;
//	drawp();
	u.x=u.x*c-u.z*s;
	u.z=uz;
//	drawp();
	hbase=u;
	heading=0;
}

void Turtle::spinZ(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point ux, uy, uz;
//	drawp();
	ux.x = u.x.x*c-u.x.y*s;
	ux.y = u.x.y*c+u.x.x*s;
	ux.z = u.x.z;
	u.x=ux;

	uy.x = u.y.x*c-u.y.y*s;
	uy.y = u.y.y*c+u.y.x*s;
	uy.z = u.y.z;
	u.y=uy;

	uz.x = u.z.x*c-u.z.y*s;
	uz.y = u.z.y*c+u.z.x*s;
	uz.z = u.z.z;
	u.z=uz;

//	drawp();
	hbase=u;
	heading=0;
}


void Turtle::spinY(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point ux, uy, uz;
//	drawp();
	ux.z = u.x.z*c-u.x.x*s;
	ux.x = u.x.x*c+u.x.z*s;
	ux.y = u.x.y;
	u.x=ux;

	uy.z = u.y.z*c-u.y.x*s;
	uy.x = u.y.x*c+u.y.z*s;
	uy.y = u.y.y;
	u.y=uy;

	uz.z = u.z.z*c-u.z.x*s;
	uz.x = u.z.x*c+u.z.z*s;
	uz.y = u.z.y;
	u.z=uz;

//	drawp();
	hbase=u;
	heading=0;
}

void Turtle::spinX(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point ux, uy, uz;
//	drawp();
	ux.y = u.x.y*c-u.x.z*s;
	ux.z = u.x.z*c+u.x.y*s;
	ux.x = u.x.x;
	u.x=ux;

	uy.y = u.y.y*c-u.y.z*s;
	uy.z = u.y.z*c+u.y.y*s;
	uy.x = u.y.x;
	u.y=uy;

	uz.y = u.z.y*c-u.z.z*s;
	uz.z = u.z.z*c+u.z.y*s;
	uz.x = u.z.x;
	u.z=uz;

//	drawp();
	hbase=u;
	heading=0;
}

double Turtle::getHeading()
{
	if (screen->perspective)
	{	double result;


		if ((1.0-fabs(u.z.y)) < Float::epsilon)
		{	result=atan2(u.y.x, u.x.x)/Float::deg2rad;
		}else
		{	result=atan2(-u.x.y, u.y.y)/Float::deg2rad;
		}

		if (result < 0.0) return result + 360.0; else return result;
	}else
		return heading;
}

double Turtle::pitch()
{
//*
	float ry=atan2(-u.z.x, u.z.z), a;

	if (fabs(fabs(ry)-90.0*Float::deg2rad) < 1)
		a=u.z.x/sin(-ry);
	else
		a=u.z.z/cos(ry);

	return atan2(-u.z.y, a)/Float::deg2rad;
/*/
	double xx=::sqrt(Float::sqr(u.z.x)+Float::sqr(u.z.z));
	if (u.z.z < 0)
		xx=-xx;
	if (fabs( xx  ) < Float::epsilon 
	 && fabs(u.z.y) < Float::epsilon)
		return 0;
	return atan2(u.z.y, // why is this not correct?
		xx)/Float::deg2rad;
//*/
}

double Turtle::roll() 
{
	if (fabs(u.z.z) < Float::epsilon
	 && fabs(u.z.x) < Float::epsilon)
		return 0;
	else
		return atan2(u.z.x, u.z.z)/Float::deg2rad;
}

void Turtle::setRoll(double angle)
{
	angle = Float::mod(angle, 360);
	rightRoll(angle-roll());
}

void Turtle::setPitch(double angle)
{
	angle = Float::mod(angle, 360);
	upPitch(angle-pitch());
}

void Turtle::setOrientation(Point o)
{
//	draw();
	reset3d();
/*	right    (-getHeading());
	upPitch  (-pitch());
	rightRoll(-roll());
*/
	setRoll   (o.x);
	setPitch  (o.y);
	setHeading(o.z);
//	draw();
}

Point Turtle::orientation()
{
	Point ori;
	ori.x=roll();
	ori.y=pitch();
	ori.z=getHeading();
	return ori;
}

void Turtle::show()
{
	if (!shown) 
	{	shown=true;
//		draw();
	}
}

void Turtle::hide()
{
	if (shown) 
	{
//		draw();
		shown=false;
	}
}

bool Turtle::wrapLeft(double d, Point end)
{
	double yi, newd;
	if (end.x < screen->Min.x) 
	{
		yi = y-(x-screen->Min.x)*(end.y-y)/(end.x-x);

		if (screen->Min.y <= yi  &&  yi <= screen->Max.y) 
		{	lineTo(screen->Min.x, yi);
			if (mode==wrapmode) 
			{
				newd = d*(end.x-screen->Min.x)/(end.x-x);
				*this = Point(screen->Max.x, yi, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}
			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);	    
		}
	}
	return false;
}

bool Turtle::wrapRight(double d, Point end)
{
	double yi, newd;
	if (end.x > screen->Max.x) 
	{
		yi = y+(screen->Max.x-x)*(end.y-y)/(end.x-x);

		if (screen->Min.y <= yi  &&  yi <= screen->Max.y) 
		{	lineTo(screen->Max.x, yi);
			if (mode==wrapmode) 
			{
				newd = d*(end.x-screen->Max.x)/(end.x-x);
				*this = Point(screen->Min.x, yi, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}

			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);
		}
	}
	return false;
}

bool Turtle::wrapUp(double d, Point end)
{
	double xi, newd;
	if (end.y > screen->Max.y) 
	{
		xi = x-(y-screen->Max.y)*(end.x-x)/(end.y-y);

		if (screen->Min.x <= xi  &&  xi < screen->Max.x) 
		{	lineTo(xi, screen->Max.y);
			if (mode==wrapmode) 
			{
				newd = d*(end.y-screen->Max.y)/(end.y-y);
				*this = Point(xi, screen->Min.y, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}
			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);
		}else 
			if (xi <= screen->Min.x  &&  screen->Max.x<=xi)
			{
				xi= screen->Max.x<=xi ?   screen->Max.x  :  screen->Min.x;

				lineTo(xi, screen->Max.y);
				if (mode==wrapmode) 
				{
					newd = d*(end.y-screen->Max.y)/(end.y-y);
					*this = Point(xi, screen->Min.y, 0);
					if (newd*d > 0)
					{	wrapDelta=newd;
						return true;
					}
				}else 
					error(TURTLE_OUT_OF_BOUNDS, 0);
			}
	}
	return false;
}

bool Turtle::wrapDown(double d, Point end)
{
	double xi, newd;
	if (end.y < screen->Min.y) 

	{
		xi = x+(screen->Min.y-y)*(end.x-x)/(end.y-y);

		if (screen->Min.x <= xi  &&  xi <= screen->Max.x) 
		{	lineTo(xi, screen->Min.y);
			if (mode==wrapmode) 
			{
				newd = d*(end.y-screen->Min.y)/(end.y-y);
				*this = Point(xi, screen->Max.y, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}
			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);
		}else 
			if (xi <= screen->Min.x  &&  screen->Max.x <= xi) 
			{
				xi= screen->Max.x <= xi ?   screen->Max.x  :  screen->Min.x;

				lineTo(xi, screen->Min.y);
				if (mode==wrapmode) 
				{
					newd = d*(end.y-screen->Min.y)/(end.y-y);
					*this = Point(xi, screen->Max.y, 0);
					if (newd*d > 0) 
					{	wrapDelta=newd;
						return true;
					}
				}else 
					error(TURTLE_OUT_OF_BOUNDS, 0);
			}
	}
	return false;
}

void Turtle::forward(double d) 
{
	TurtleNodeP tn=turtleNode; tn=tn;	// pushing turtleNode on stack
	Point delta, target, endPoint;
	bool wraping=false;

	wrapDelta=0;
//	draw();
	if (screen->perspective)
	{	target = *this+u.y*d;
		lineTo(target);
		if (checkThrowing()) return;
		*this=target;
	}
	else if (mode==windowmode)
	{	target = *this+Point(_cos, _sin, 0)*d;
		lineTo(target);
		if (checkThrowing()) return;
		*this=target;
	}else
	{	do
		{	if (wrapDelta!=0) 
				d=wrapDelta;
			delta = Point(_cos, _sin, 0)*d;

			if (wrapDelta==0.) 
			{	endPoint = *this+delta;
				if (mode==wrapmode && !in(screen->Min, screen->Max))
				{	getInWindow();
					moveTo(*this);
				}
			}
			target = *this+delta;
			if (checkThrowing()) return;
//			if (pen.internalMode==Pen::Reverse && pen.down && d > 0) 
//				lineTo(target);

			if (target.in(screen->Min, screen->Max)) 
			{	lineTo(target);
				break;
			}
			wraping = wrapUp(d, target) || wrapDown(d, target) || wrapLeft(d, target) || wrapRight(d, target); 
			if (wraping)
				moveTo(*this);
		}
		while (wraping && !throwing);

//		if (pen.internalMode==Pen::Reverse && pen.down && d < 0) 
//			lineTo(*this);

		*this=endPoint;
	}
//	draw();
	outOfBounds=false;
}


double Turtle::towards_helper(Point end) 
{
	if (end != *this) 
	{	double a;
		a=-(atan2(end.y-y, end.x-x) / Float::deg2rad - 90);
		return a < 0 ?   360+a  :  a;
	}
	return 0;
}

void Turtle::setPos(Point target) 
{
	if (!throwing) 
	{	bool targetInScreen=target.in(screen->Min, screen->Max),
			wrapping= 
			( !targetInScreen
			|| !this->in(screen->Min, screen->Max));

		if (mode==fencemode && !targetInScreen)
			error(TURTLE_OUT_OF_BOUNDS, 0);
		else 
			if (mode==wrapmode && (wrapping || outOfBounds)) 
			{
//				draw();
				bool savedShown=shown;
				shown=false;
				double save_heading=heading;
				setHeading(towards_helper(target));

				forward((target - *this).Norm());

				setHeading(save_heading);
				shown=savedShown;
//				draw();
				outOfBounds=wrapping;
			}else 
			{	outOfBounds=false;
//				draw();
				lineTo(target); *this=target;
//				draw();
			}
	}
}

double Turtle::distance(Point target)
{
	return (target-*this).Norm();
}

void Turtle::setPos(double x, double y, double z) {setPos(Point(x,y,z));}

void Turtle::setPos(NodeP nodeX, NodeP nodeY, NodeP nodeZ) 
{
	Point target;
  	if (!throwing) 
	{	if (nodeX) target.x=nodeX->getnumber(); else target.x=x;
		if (nodeY) target.y=nodeY->getnumber(); else target.y=y;
		if (nodeZ) target.z=nodeZ->getnumber(); else target.z=z;
		setPos(target);
	}
}

void Turtle::_setPos(Point target) 
{
	if (!throwing) 
	{
//		draw();
		lineTo(target);	*this=target;
//		draw();
	}
}

void Turtle::_setPos(double x, double y, double z) {_setPos(Point(x,y,z));}


void Turtle::_setPos(NodeP nodeX, NodeP nodeY, NodeP nodeZ) 
{
	Point target;
  	if (!throwing) 
	{	if (nodeX) target.x=nodeX->getnumber(); else target.x=x;
		if (nodeY) target.y=nodeY->getnumber(); else target.y=y;
		if (nodeZ) target.z=nodeZ->getnumber(); else target.z=z;
		_setPos(target);
	}
}

void Turtle::setSpherePos(double r, double phi, double theta)
{
	setPos(
		r*cos(phi*Float::deg2rad)*sin(theta*Float::deg2rad), 
		r*cos(theta*Float::deg2rad), 
		r*sin(phi*Float::deg2rad)*sin(theta*Float::deg2rad));
}

void Turtle::setCylinderPos(double r, double phi, double y)
{
	setPos(r*cos(phi*Float::deg2rad), y, r*sin(phi*Float::deg2rad));
}

void Turtle::home()
{
	outOfBounds=false;
	setPos(0., 0., 0.);
	setHeading(0);
	reset3d();
}

NodeP Turtle::matrix()
{
/*	Point saved=*this;
	if (mode==wrapmode && !in(screen->Min, screen->Max))
		getInWindow();

	Point basex_,basey_,basez_, trans_,
		&centr_ = screen->center - screen->center0;

//	&centr_ = screen->center - screen->center0,

	if (screen->perspective)
	{	basex_ = screen->rotateNotScaled(u.x);
		basey_ = screen->rotateNotScaled(u.y);
		basez_ = screen->rotateNotScaled(u.z);
	}else
	{	basex_ = screen->rotateNotScaled(Point(_sin,-_cos, 0));
		basey_ = screen->rotateNotScaled(Point(_cos, _sin, 0));
		basez_ = screen->rotateNotScaled(Point(   0,    0, 1));
	}
	trans_ = screen->rotateNotScaled(Point(x,y,z));

	*this=saved;
	return new List
	(	new List

		(	new Float(basex_.x),
			new Float(basex_.y), 
			new Float(basex_.z), _ListEnd),
		new List
		(	new Float(basey_.x),
			new Float(basey_.y), 
			new Float(basey_.z), _ListEnd), 
		new List
		(	new Float(basez_.x),
			new Float(basez_.y), 
			new Float(basez_.z), _ListEnd), 
		new List
		(	new Float(trans_.x+centr_.x),
			new Float(trans_.y-centr_.y),	
			new Float(trans_.z+centr_.z),	_ListEnd), _ListEnd);
*/
	double m[16]={0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};
#ifdef OPENGL
	glPushMatrix();
		glTranslated(x,y,z);
		if (screen->perspective)
		{	glRotated(roll(),0,1,0);
			glRotated(pitch(),1,0,0);
			glRotated(-getHeading(),0,0,1);
		}else
			glRotated(-heading,0,0,1);
		glGetDoublev(GL_MODELVIEW_MATRIX, m);
	glPopMatrix();
#endif
	return new List
	(	new List
		(	new Float(m[0]),
			new Float(m[1]), 
			new Float(m[2]), _ListEnd),
		new List
		(	new Float(m[4]),
			new Float(m[5]), 
			new Float(m[6]), _ListEnd), 
		new List
		(	new Float(m[8]),
			new Float(m[9]), 
			new Float(m[10]), _ListEnd), 
		new List
		(	new Float(m[12]),
			new Float(m[13]),	
			new Float(m[14]),	_ListEnd), _ListEnd);
}

void Turtle::setMatrix(Point basex,Point basey,Point basez,Point trans)
{
//	draw();
/*	screen->base.x = basex;
	screen->base.y = basey;
	screen->base.z = basez;

	screen->base_scaled.x=screen->base.x*screen->scale.x;
	screen->base_scaled.y=screen->base.y*screen->scale.y;
	screen->base_scaled.z=screen->base.z*screen->scale.z;

	screen->center = screen->center0+trans*screen->scale;
*/
	double m[16]=
	{	basex.x,basex.y,basex.z,0,
		basey.x,basey.y,basey.z,0,
		basez.x,basez.y,basez.z,0,
		trans.x,trans.y,trans.z,1};

	for (int i=0; i<16; i++)
		screen->modelview_matrix[i]=m[i];

#ifdef OPENGL
	glLoadMatrixd(screen->modelview_matrix);
#endif
	if (recorder->recording)
		*recorder << new R_setMatrix(basex,basey,basez,trans);
//	draw();
}

void Turtle::setMatrix(PairP basex,PairP basey,PairP basez,PairP trans)
{
	Point 
		basex_ = Point(
			basex-> first()->getnumber(),
			basex->second()->getnumber(), 
			basex-> third()->getnumber()),
		basey_ = Point(
			basey-> first()->getnumber(),
			basey->second()->getnumber(),
			basey-> third()->getnumber()),
		basez_ = Point(
			basez-> first()->getnumber(),

			basez->second()->getnumber(),
			basez-> third()->getnumber()),
		trans_ = Point(
			trans-> first()->getnumber(),
			trans->second()->getnumber(),
			trans-> third()->getnumber());

	setMatrix(basex_,basey_,basez_,trans_);
}

void Turtle::setTurtleMatrix()
{
/*	Point saved=*this;
	if (mode==wrapmode && !in(screen->Min, screen->Max))
		getInWindow();

	Point basex,basey,basez, trans;
	
	if (screen->perspective)
	{	basex=screen->rotateNotScaled(u.x);
		basey=screen->rotateNotScaled(u.y);
		basez=screen->rotateNotScaled(u.z);
	}else
	{	basex = screen->rotateNotScaled(Point(_sin,-_cos, 0)),
		basey = screen->rotateNotScaled(Point(_cos, _sin, 0)),
		basez = screen->rotateNotScaled(Point(  0 ,   0 , 1));
	}
	trans = screen->rotateNotScaled(Point(x,-y,z));

	*this=saved;
	screen->base.x=basex;
	screen->base.y=basey;
	screen->base.z=basez;

	screen->base_scaled.x=basex*screen->scale.x;
	screen->base_scaled.y=basey*screen->scale.y;
	screen->base_scaled.z=basez*screen->scale.z;


	screen->center+= trans*screen->scale;

	if (recorder->recording)
		*recorder << new R_SetMatrix(basex,basey,basez,trans);
*/
#ifdef OPENGL
	glTranslated(x,y,z);
	if (screen->perspective || screen->orthographic)
	{	glRotated(roll(),0,1,0);
		glRotated(pitch(),1,0,0);
		glRotated(-getHeading(),0,0,1);
	}else
		glRotated(-heading,0,0,1);
	glGetDoublev(GL_MODELVIEW_MATRIX, screen->modelview_matrix);
#endif
	if (recorder->recording)
		*recorder << new R_setMatrix4x4(screen->modelview_matrix);
}

void Turtle::wrap()
{
//	draw();
	mode=wrapmode;
	getInWindow();
	moveTo(*this);
//	draw();
}

void Turtle::arc(double angle, double radius)
{
	if (!pen.down) 
		return;
	Turtle saved(false);
	saved=*this;

//	draw();
	shown=false;

	if (screen->perspective)
	{	Point point = u.y*radius;

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		if (fabs(angle) > 360)
			angle=fmod(angle, 360);
		int count = int(4*ceil(fabs(angle*radius/360))); // calculate resolution parameters
		if (count < 5)
			count=5;
		else if (count > 100)
			count=100;
		if (count)
		{	double delta=Float::deg2rad*angle/count,
				radAngle=0;

			for (int i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radAngle+=delta;
				point=(u.y*cos(radAngle)+u.x*sin(radAngle))*radius;
				setPos(saved+point);
			}
		}
	}
	else
	{	double headingAngle = 90-heading, 
			radAngle = headingAngle*Float::deg2rad;

		Point point = Point(cos(radAngle), sin(radAngle), 0)*radius;

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		if (fabs(angle) > 360)
			angle=fmod(angle, 360);
		int count = int(4*ceil(fabs(angle*radius/360))); // calculate resolution parameters
		if (count < 5)
			count=5;
		else if (count > 100)
			count=100;
		if (count)
		{	double delta=-Float::deg2rad*angle/count; // heed the minus! It`s because Logo angles go strangely clockwise.
			int i;
			for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radAngle+=delta;
				point=Point(cos(radAngle), sin(radAngle), 0)*radius;
				setPos(saved+point);
			}
			if (pen.size!=0)
			{	--count;	// This section is neccessary because of an OpenGL bug (gaps!)
				radAngle-=delta/2;
				for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
				{	radAngle-=delta;
					point=Point(cos(radAngle), sin(radAngle), 0)*radius;
					setPos(saved+point);
				}
			}
		}
	}
	*this=saved;
	moveTo();
//	draw();
}

void Turtle::EllipseArc(double angle, double xRadius, 
								double yRadius, double startAngle)
{
	if (!pen.down) 
		return;
	Turtle saved(false);
	saved=*this;

//	draw();
	shown=false;

	if (screen->perspective)
	{	double radStart = startAngle*Float::deg2rad;

		Point 
			point =-u.x*xRadius*sin(radStart)
					 -u.y*yRadius*cos(radStart);

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		// calculate resolution parameters

		if (fabs(angle) > 360)
			angle=fmod(angle, 360);
		int count = int(4*ceil(fabs(angle*(xRadius+yRadius)/360))); // calculate resolution parameters
		if (count < 5)
			count=5;
		else if (count > 100)
			count=100;
		if (count)
		{	double delta=Float::deg2rad*angle/count;
		
			for (int i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radStart+=delta;
				point =-u.x*xRadius*sin(radStart)
						 -u.y*yRadius*cos(radStart);
				setPos(saved+point);
			}
		}
	}
	else
	{	double headingAngle = 90-heading, 
			radHead  = headingAngle*Float::deg2rad,
			radStart = (270-startAngle)*Float::deg2rad;

		Point 
			ex= Point( cos(radHead),sin(radHead),0),
			ey= Point(-sin(radHead),cos(radHead),0),

			point = ex*yRadius*sin(radStart)
					 -ey*xRadius*cos(radStart);

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		// calculate resolution parameters
		if (fabs(angle) > 360)
			angle=fmod(angle, 360);
		int count = int(4*ceil(fabs(angle*(xRadius+yRadius)/360))); // calculate resolution parameters
		if (count < 5)
			count=5;
		else if (count > 100)
			count=100;
		if (count)
		{	double delta=-Float::deg2rad*angle/count; // heed the minus! It`s because Logo angles go strangely clockwise.
			int i;

			if (mode==windowmode)
			{	for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
				{	radStart+=delta;
					point = ex*yRadius*sin(radStart)
							 -ey*xRadius*cos(radStart);
					Point target=saved+point;
					lineTo(target);
					*this=target;
				}
			}
			else
			{	for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
				{	radStart+=delta;
					point = ex*yRadius*sin(radStart)
							 -ey*xRadius*cos(radStart);
					setPos(saved+point);
				}
				if (pen.size!=0)
				{	--count;	// This section is neccessary because of an OpenGL bug (gaps!)
					radStart-=delta/2;
					for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
					{	radStart-=delta;
						point = ex*yRadius*sin(radStart)
								 -ey*xRadius*cos(radStart);
						setPos(saved+point);
					}	
				}
			}
		}

	}
	*this=saved;
	moveTo();
//	draw();
}

void Turtle::fillRect(Point a, Point b)
{
//	draw();
	Point oldpt=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

#ifdef OPENGL
	glPushMatrix();
		glColor4ubv((unsigned char*)&pen.floodcolor);
		glTranslated(x,y,z);
		if (screen->perspective)
		{	glRotated(roll(),0,1,0);
			glRotated(pitch(),1,0,0);
			glRotated(-getHeading(),0,0,1);
		}else
			glRotated(-heading,0,0,1);
		glBegin(GL_QUADS); 
			if (screen->perspective)
			{	Point 
					v1=Point(a.x,a.y,0)-Point(a.x,b.y,0),
					v2=Point(b.x,b.y,0)-Point(a.x,b.y,0),
					normal=v1.X(v2);
				float n=normal.Norm();

				if (n!=0)
				{	normal/=n;
					gl2psNormal3fv((float*)&normal);
				}
			}
			if (screen->textured)
			{	glTexCoord2f(0,0);
				glVertex2f(a.x, a.y); 
				glTexCoord2f(1,0);
				glVertex2f(b.x, a.y); 
				glTexCoord2f(1,1);
				glVertex2f(b.x, b.y); 
				glTexCoord2f(0,1);
				glVertex2f(a.x, b.y); 
			}else
			{	glVertex2f(a.x, a.y); 
				glVertex2f(b.x, a.y); 
				glVertex2f(b.x, b.y); 
				glVertex2f(a.x, b.y); 
			}
		glEnd(); 
		glColor4ubv((unsigned char*)&pen.color);
	glPopMatrix();
#else
	wxDC &dc=*recorder->graph->dc;
	recorder->graph->brush.SetColour(wxColour(pen.floodcolor));
	dc.SetBrush(recorder->graph->brush);
	*old=*this;
	pen.down=false;
	wxPoint p[4];
	forward(a.y);
	right(90);
	forward(a.x);
	p[0]=wxPoint(x,y);
	forward(b.x-a.x);
	p[1]=wxPoint(x,y);
	right(-90);
	forward(b.y-a.y);
	p[2]=wxPoint(x,y);
	right(-90);
	forward(b.x-a.x);
	p[3]=wxPoint(x,y);
	*this=*old;	
	dc.DrawPolygon(4, p);
#endif
	*this=oldpt;
//	draw();
}

void Turtle::fillEllipse(Point radius)
{
	if (radius.x > 0)
	{
//		draw();
		Point old=*this;

		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

#ifdef OPENGL
		glPushMatrix();
			glColor4ubv((GLubyte*)&pen.floodcolor);
			glTranslated(x,y,z);
			if (screen->perspective)
			{	glRotated(roll(),0,1,0);
				glRotated(pitch(),1,0,0);
				glRotated(-getHeading(),0,0,1);
			}else
				glRotated(-heading,0,0,1);

			glScaled(1,radius.y/radius.x,1);
				GLUquadricObj *q=gluNewQuadric();
				if (screen->textured)
					gluQuadricTexture(q, GL_TRUE);
				gluDisk(q, 0, radius.x, int(3*radius.x), 1);
				gluDeleteQuadric(q);
			glColor4ubv((GLubyte*)&pen.color);
		glPopMatrix();
#endif
		*this=old;
//		draw();
	}
}

void Turtle::cylinder(double length, double baseradius, double topradius)
{
#ifdef OPENGL
	glPushMatrix();
	glTranslated(x,y,z);

	Point v=u.y;
//*
	float xz, theta, phi;

	xz=::sqrt(Float::sqr(v.x)+Float::sqr(v.z));
	if (fabs(xz) < Float::epsilon
	 && fabs(v.y) < Float::epsilon)
		theta=0;
	else

		theta=atan2(xz, v.y)/Float::deg2rad;

	if (fabs(v.x) < Float::epsilon
	 && fabs(v.z) < Float::epsilon)
		phi=0;
	else
		phi=atan2(v.x, v.z)/Float::deg2rad;

	glRotated(phi,0,1,0);
	glRotated(theta-90,1,0,0);
/*/
	glRotated(roll(),0,1,0);
	glRotated(pitch()-90,1,0,0);
	glRotated(getHeading(),0,1,0);
//*/
	GLUquadricObj *q=gluNewQuadric();
	gluQuadricTexture(q, GL_TRUE);
	gluCylinder(q,
		baseradius,topradius, 
		length, 30,1);
	gluDeleteQuadric(q);
	glPopMatrix();
#endif
}

void Turtle::partialCylinder(double length, double baseradius, double topradius, 
									  double startangle, double endangle, double scrunch, int slices)
{
#ifdef OPENGL
	glPushMatrix();
	glTranslated(x,y,z);

	glRotated(roll(),0,1,0);
	glRotated(pitch()-90,1,0,0);
	glRotated(getHeading(),0,1,0);
	
		int i;
		double x,y,z,s=0, angle=startangle*Float::deg2rad, 
			dangle=(endangle*Float::deg2rad-startangle)/slices;
		glBegin(GL_QUAD_STRIP);
		for (i = 0; i <= slices; i++) 
		{	angle+=dangle;
			x = sin(angle);
			y = cos(angle);
			z = 0;
			gl2psNormal3f(x,y,0);
			if (screen->textured) glTexCoord2f(s,0);
			glVertex3f(x*baseradius, y*baseradius*scrunch, z);
			x = sin(angle);
			y = cos(angle);
			z = length;
			gl2psNormal3f(x,y,0);
			if (screen->textured) glTexCoord2f(s,0);
			glVertex3f(x*topradius, y*topradius*scrunch, z);
			s += 1/slices;
		}
		glEnd();
	
	glPopMatrix();
#endif
}

void Turtle::sphere(double radius, int slices, int stacks)
{
	if (radius > 0)
	{
//		draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

#ifdef OPENGL
		glPushMatrix();
			glTranslated(x,y,z);
			glRotated(roll(),0,1,0);
			glRotated(pitch()-90,1,0,0);
			glRotated(getHeading(),0,1,0);
			GLUquadricObj *q=gluNewQuadric();
			if (screen->textured)
				gluQuadricTexture(q, GL_TRUE);
			if (slices==0)
				slices=int(radius/2+10);
			if (stacks==0)
				stacks=int(radius/2+10);
			gluSphere(q, radius, slices, stacks);
			gluDeleteQuadric(q);
		glPopMatrix();
#endif
		*this=old;
//		draw();
	}
}

void Turtle::ellipsoid(Point radius, int slices, int stacks)
{
	double r=radius.Norm();
	if (r > 0)
	{
//		draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

#ifdef OPENGL
		glPushMatrix();
			glTranslated(x,y,z);
			glRotated(roll(),0,1,0);
			glRotated(pitch()-90,1,0,0);
			glRotated(getHeading(),0,1,0);
			glScaled(radius.x/100,radius.z/100,radius.y/100);
			glEnable(GL_AUTO_NORMAL);
			GLUquadricObj *q=gluNewQuadric();
			if (screen->textured)
				gluQuadricTexture(q, GL_TRUE);
			if (slices==0)
				slices=int(r/4+10);
			if (stacks==0)
				stacks=int(r/4+10);
			gluSphere(q, 100, slices, stacks);
			gluDeleteQuadric(q);
			glDisable(GL_AUTO_NORMAL);
		glPopMatrix();
#endif
		*this=old;
//		draw();
	}
}

void Turtle::partialEllipsoid(Point radius, 
										double slicesstart, double slicesend, int slices,
										double stacksstart, double stacksend, int stacks)
{
	double r=radius.Norm();
	if (r > 0)
	{
//		draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

#ifdef OPENGL
		glPushMatrix();
		glTranslated(x,y,z);
		glRotated(roll(),0,1,0);
		glRotated(pitch()-90,1,0,0);
		glRotated(getHeading(),0,1,0);
	
		GLfloat rho, rho0, drho, theta, theta0, dtheta;
		GLfloat x, y, z;
		GLfloat s, t, ds, dt;
		GLint i, j, imin, imax;
		GLboolean normals;
		GLfloat nsign;
		
		normals = GL_TRUE;
		nsign = 1.0;
		
		rho0 = Float::deg2rad*stacksstart;
		theta0 = Float::deg2rad*slicesstart;
		drho = Float::deg2rad*(stacksend-stacksstart) / (GLfloat) stacks;
		dtheta = Float::deg2rad*(slicesend-slicesstart) / (GLfloat) slices;
		
		// texturing: s goes from 0.0/0.25/0.5/0.75/1.0 at +y/+x/-y/-x/+y axis 
		// t goes from -1.0/+1.0 at z = -radius/+radius (linear along longitudes)
		// cannot use triangle fan on texturing (s coord. at top/bottom tip varies)
		
		if (!screen->textured) 
		{	
			if (stacksstart==0)
			{	// draw +Z end as a triangle fan 
				glBegin(GL_TRIANGLE_FAN);
				gl2psNormal3f(0.0, 0.0, 1.0);
				glVertex3f(0.0, 0.0, nsign*radius.z);
				for (j = 0; j <= slices; j++) 
				{	theta = j*dtheta+theta0;
					x = -sin(theta) * sin(drho);
					y = cos(theta) * sin(drho);
					z = nsign * cos(drho);
					if (normals)
						gl2psNormal3f(x * nsign, y * nsign, z * nsign);
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}
		ds = 1.0 / slices;
		dt = 1.0 / stacks;
		t = 1.0; 		// because loop now runs from 0 
		if (screen->textured) 
		{	imin = 0;
			imax = stacks;
		}
		else 
		{	imin = 1;
			imax = stacks - 1;
		}
		// draw intermediate stacks as quad strips
		for (i = imin; i < imax; i++) 
		{	rho = i*drho+rho0;
			glBegin(GL_QUAD_STRIP);
			s = 0.0;
			for (j = 0; j <= slices; j++) 
			{	theta = j*dtheta+theta0;
				x = -sin(theta) * sin(rho);
				y = cos(theta) * sin(rho);
				z = nsign * cos(rho);
				if (normals)
					gl2psNormal3f(x * nsign, y * nsign, z * nsign);
				if (screen->textured) glTexCoord2f(s,t);
				glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				x = -sin(theta) * sin(rho + drho);
				y = cos(theta) * sin(rho + drho);
				z = nsign * cos(rho + drho);
				if (normals)
					gl2psNormal3f(x * nsign, y * nsign, z * nsign);
				if (screen->textured) glTexCoord2f(s,t-dt);
				s += ds;
				glVertex3f(x*radius.x, y*radius.y, z*radius.z);
			}
			glEnd();
			t -= dt;
		}
		if (!screen->textured) 
		{	if (stacksend==180)
			{
				// draw -Z end as a triangle fan 
				glBegin(GL_TRIANGLE_FAN);
				gl2psNormal3f(0.0, 0.0, -1);
				glVertex3f(0.0, 0.0, -nsign*radius.z);
				rho = Float::pi - drho;
				s = 1.0;
				t = dt;
				for (j = slices; j >= 0; j--) 
				{	theta = j * dtheta+theta0;
					x = -sin(theta) * sin(rho);
					y = cos(theta) * sin(rho);
					z = nsign * cos(rho);
					if (normals)
						gl2psNormal3f(x * nsign, y * nsign, z * nsign);
					s -= ds;
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}

		glPopMatrix();
#endif
		*this=old;
//		draw();
	}
}

void Turtle::partialEllipsoidSh(Point radius, 
										  double slicesstart, double slicesend, int slices,
										  double stacksstart, double stacksend, int stacks)
{
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
	double r=radius.Norm();
	if (r > 0)
	{
//		draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

#ifdef OPENGL
		GLfloat rho, rho0, drho, theta, theta0, dtheta;
		GLfloat s, t, ds, dt;
		GLint i, j, imin, imax;
		GLboolean normals;
		GLfloat nsign;
		
		glPushMatrix();
		glLoadIdentity();
		glRotated(roll(),0,1,0);
		glRotated(pitch()-90,1,0,0);
		glRotated(getHeading(),0,1,0);
		double mm[16];
		glGetDoublev(GL_MODELVIEW_MATRIX, mm);
		glPopMatrix();
		PointMatrix m;
		for (i=0;  i<3; i++)
		{	m.x.a[i]=mm[i];
			m.y.a[i]=mm[i+4];
			m.z.a[i]=mm[i+8];
		}
		glPushMatrix();
		glTranslated(x,y,z);
		glRotated(roll(),0,1,0);
		glRotated(pitch()-90,1,0,0);
		glRotated(getHeading(),0,1,0);
	
		normals = GL_TRUE;
		nsign = 1.0;
		
		rho0 = Float::deg2rad*stacksstart;
		theta0 = Float::deg2rad*slicesstart;
		drho = Float::deg2rad*(stacksend-stacksstart) / (GLfloat) stacks;
		dtheta = Float::deg2rad*(slicesend-slicesstart) / (GLfloat) slices;

		ShadowObject *so=th->graph->shadowObject;
		int si=so->nPoints,
			 sj=so->nPlanes;

		Point p, normal;

		// texturing: s goes from 0.0/0.25/0.5/0.75/1.0 at +y/+x/-y/-x/+y axis 
		// t goes from -1.0/+1.0 at z = -radius/+radius (linear along longitudes)
		// cannot use triangle fan on texturing (s coord. at top/bottom tip varies)
		
/*		if (!screen->textured) 
		{	
			if (stacksstart==0)
			{	// draw +Z end as a triangle fan
				glBegin(GL_TRIANGLE_FAN);
				gl2psNormal3f(0.0, 0.0, 1.0);
				glVertex3f(0.0, 0.0, nsign*radius.z);
				for (j = 0; j <= slices; j++) 
				{	theta = j*dtheta+theta0;
					x = -sin(theta) * sin(drho);
					y = cos(theta) * sin(drho);
					z = nsign * cos(drho);
					if (normals)
						gl2psNormal3f(x * nsign, y * nsign, z * nsign);
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}
*/		ds = 1.0 / slices;
		dt = 1.0 / stacks;
		t = 1.0; 		// because loop now runs from 0
//		if (screen->textured) 
		{	imin = 0;
			imax = stacks;
		}
/*		else 
		{	imin = 1;
			imax = stacks - 1;
		}
*/		// draw intermediate stacks as quad strips
		for (i = imin; i < imax; i++) 
		{	rho = i*drho+rho0;
			glBegin(GL_QUAD_STRIP);
			s = 0.0;
			
			theta = theta0;
			p=Point( -sin(theta) * sin(rho),
						 cos(theta) * sin(rho),
								nsign * cos(rho));
			normal=p*nsign;
			if (normals)
				gl2psNormal3fv((float*)&normal);
			if (screen->textured) glTexCoord2f(s,t);
			p=p*radius;
			glVertex3fv((float*)&p);
			if (!so->checkPoints(si))
				return;
			so->points[si]=m*p+(*this);
			si++;
			
			p=Point(-sin(theta) * sin(rho + drho),
						cos(theta) * sin(rho + drho),
							  nsign * cos(rho + drho));
			normal=p*nsign;
			if (normals)
				gl2psNormal3fv((float*)&normal);
			if (screen->textured) glTexCoord2f(s,t-dt);
			s += ds;
			p=p*radius;
			glVertex3fv((float*)&p);
			if (!so->checkPoints(si))
				return;
			so->points[si]=m*p+(*this);
			si++;

			for (j = 1; j <= slices; j++) 
			{	theta = j*dtheta+theta0;
				p=Point(	-sin(theta) * sin(rho),
							 cos(theta) * sin(rho),
									nsign * cos(rho));
				normal=p*nsign;
				if (normals)
					gl2psNormal3fv((float*)&normal);
				if (screen->textured) glTexCoord2f(s,t);
				p=p*radius;
				glVertex3fv((float*)&p);
				if (!so->checkPoints(si))
					return;
				so->points[si]=m*p+(*this);
				si++;
		
				p=Point(	-sin(theta) * sin(rho + drho),
							 cos(theta) * sin(rho + drho),
							      nsign * cos(rho + drho));
				normal=p*nsign;
				if (normals)
					gl2psNormal3fv((float*)&normal);
				if (screen->textured) glTexCoord2f(s,t-dt);
				s += ds;
				p=p*radius;
				glVertex3fv((float*)&p);
				if (!so->checkPoints(si))
					return;
				so->points[si]=m*p+(*this);
				si++;

				if (!so->checkPlanes(sj))
					return;
				so->planes[sj].p[2]=si-2;
				so->planes[sj].p[1]=si-3;
				so->planes[sj].p[0]=si-4;
//				so->planes[sj].neigh[0]=sj+2;
//				so->planes[sj].neigh[1]=sj+4;
				sj++;

				if (!so->checkPlanes(sj))
					return;
				so->planes[sj].p[2]=si-3;
				so->planes[sj].p[1]=si-2;
				so->planes[sj].p[0]=si-1;
//				so->planes[sj].neigh[0]=sj;
				sj++;
			}
			glEnd();
			t -= dt;
		}
/*		if (!screen->textured) 
		{	if (stacksend==180)
			{
				// draw -Z end as a triangle fan
				glBegin(GL_TRIANGLE_FAN);
				gl2psNormal3f(0.0, 0.0, -1.0);
				glVertex3f(0.0, 0.0, -nsign*radius.z);
				rho = Float::pi - drho;
				s = 1.0;
				t = dt;
				for (j = slices; j >= 0; j--) 
				{	theta = j * dtheta+theta0;
					x = -sin(theta) * sin(rho);
					y = cos(theta) * sin(rho);
					z = nsign * cos(rho);
					if (normals)
						gl2psNormal3f(x * nsign, y * nsign, z * nsign);
					s -= ds;
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}
*/		so->nPoints=si;
		so->nPlanes=sj;

		glPopMatrix();
#endif
		*this=old;
//		draw();
	}
}

void Turtle::fillPie(double angle, Point radius)
{
	if (radius.x > 0)
	{
//		draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

#ifdef OPENGL
		glPushMatrix();
			glColor4ubv((GLubyte*)&pen.floodcolor);
			glTranslated(x,y,z);
			if (screen->perspective)
			{	glRotated(roll(),0,1,0);
				glRotated(pitch(),1,0,0);
				glRotated(-getHeading(),0,0,1);
			}else
				glRotated(-heading,0,0,1);

				GLUquadricObj *q=gluNewQuadric();
				if (screen->textured)
					gluQuadricTexture(q, GL_TRUE);
				gluPartialDisk(q, 0, radius.x, 
					int(3*radius.x), 1, 0, angle);
				gluDeleteQuadric(q);
			glColor4ubv((GLubyte*)&pen.color);
		glPopMatrix();
#endif
		*this=old;
//		draw();
	}
}

Color *pixels;
int pixelsSize=0;

void Turtle::pix(int w,int h, Color color)
{
	pixels[w+h*screen->winsizex]=color; 

/*	static int i=0;
	if (!(i & 0xf))
	{	glDrawPixels(screen->winsizex, screen->winsizey, 
			GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
		mySwapBuffers();
	}
	i++;
*/
}

void Turtle::set2dRasterPos()
{
#ifdef OPENGL
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	double w=screen->winsizex, h=screen->winsizey;
	double sy=screen->Min.y, sx=screen->Min.x,
			 Y=screen->Max.y, X=screen->Max.x,
			 xm=(X+sx)/2, ym=(Y+sy)/2;
	if (h && (Y-sy) && double(w)/h >= (X-sx)/(Y-sy))
	{
//		glOrtho(((sy-ym)*w)/h+xm, ((Y-ym)*w)/h+xm, sy+ym, Y+ym, -1, 1);
		glOrtho(((sy-ym)*w)/h*screen->scale.x+xm, ((Y-ym)*w)/h*screen->scale.x+xm, 
			sy*screen->scale.y+ym, Y*screen->scale.y+ym, -1000, 1000);
	}else
	{
//		glOrtho(sx+xm, X+xm, ((sx-xm)*h)/w+ym, ((X-xm)*h)/w+ym, -1, 1);
		glOrtho(sx*screen->scale.x+xm, X*screen->scale.x+xm, 
			((sx-xm)*h)/w*screen->scale.y+ym, ((X-xm)*h)/w*screen->scale.y+ym, -1000, 1000);
	}
	Point p=Point(sx,sy,0);
	glRasterPos3fv((float*)&p);
#endif
}

// floodFillX is taken from UCBLogo's xgraphics.c 
// & improved & translated to OpenGL

void Turtle::floodFill(int x, int y, Color fill, Color rim)
{
#ifdef OPENGL
	pixelsSize=screen->winsizex*screen->winsizey*sizeof(Color);
	pixels=(Color*)GC_::malloc(pixelsSize);
	if (!pixels)
		return;

	glReadPixels(0,0, screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);

	if (x >=0 && x<=screen->winsizex && y >=0 && y<=screen->winsizey)
	{	floodFillRecursive(x, y, fill, rim);
//		pixels[x+y*screen->winsizex]=0xffff0000;
	}
	glDrawPixels(screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
/*
glBegin(GL_LINE_STRIP);
glVertex3d(0,0,0);
glVertex3d(200,0,0);
glVertex3d(200,100,0);
glVertex3d(0,100,0);
glVertex3d(0,0,0);
glEnd();
*/
	GC_::free(pixels); pixels=0;
#else
	wxDC &dc=*recorder->graph->dc;
	recorder->graph->brush.SetColour(wxColour(fill));
	dc.SetBrush(recorder->graph->brush);
	dc.FloodFill(x,y, wxColour(rim), wxFLOOD_BORDER);
#endif
}

void Turtle::floodFillRecursive(int x, int y, Color fill, Color rim)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*screen->winsizex];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (c==rim || c==fill) 
			break;
//		wxASSERT(p >= pixels);
		*p-- = fill;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
//	if (leftLimit > 0) 
//		leftLimit--;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*screen->winsizex];
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=*p;
		if (c==rim || c==fill)
			break;
//		wxASSERT(p < &pixels[pixelsSize]);
		*p++ = fill;
		rightLimit=i;
	}
//	if (rightLimit < int(screen->winsizex-1))
//		rightLimit++;

	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y-1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fill) 
				{	floodFillRecursive(i, y-1, fill, rim);
					lastBorder=false;
				}
			}
			else if (c==rim) 
				lastBorder=true;
			p++;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y+1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fill) 
				{	floodFillRecursive(i, y+1, fill, rim);
					lastBorder=false;
				}
			} 
			else if (c==rim) 
				lastBorder=true;
			p++;
		}
	}
}

void Turtle::floodFillSurface(int x, int y, Color fill, Color cin)
{
#ifdef OPENGL
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
	pixels=(Color*)GC_::malloc(screen->winsizex*screen->winsizey*sizeof(Color));
	if (!pixels)
		return;

	glReadPixels(0,0, screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);

	if (x >=0 && x<=screen->winsizex && y >=0 && y<=screen->winsizey)
	{	floodFillSurfaceRecursive(x,y, fill,cin);
	}
	glDrawPixels(screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
	GC_::free(pixels); pixels=0;
#else
	wxDC &dc=*recorder->graph->dc;
	recorder->graph->brush.SetColour(wxColour(fill));
	dc.SetBrush(recorder->graph->brush);
	dc.FloodFill(x,y, wxColour(cin), wxFLOOD_SURFACE);
#endif
}

void Turtle::floodFillSurfaceRecursive(int x, int y, Color fill, Color cin)
{
	int i, lastBorder, leftLimit, rightLimit;
	if (cin == fill)	// Nothing to be done
		return;
	
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*screen->winsizex];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (c!=cin || c==fill) 
			break;
		*p-- = fill;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*screen->winsizex];
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=*p;
		if (c!=cin || c==fill) 
			break;
		*p++ = fill;
		rightLimit=i;
	}
	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y-1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c==cin && c!=fill) 
				{	floodFillSurfaceRecursive(i, y-1, fill, cin);
					lastBorder=0;
				}
			}
			else if (c != cin) 
				lastBorder=1;
			p++;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y+1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c==cin && c!=fill) 
				{	floodFillSurfaceRecursive(i, y+1, fill, cin);
					lastBorder=0;
				}
			} 
			else if (c != cin) 
				lastBorder=1;
			p++;
		}
	}
}

BitmapP fillbmp=0;

void Turtle::floodFillTolerant(int xx, int yy, Color fill, Color rim)
{
#ifdef OPENGL
	pixelsSize=screen->winsizex*screen->winsizey*sizeof(Color);
	pixels=(Color*)GC_::malloc(pixelsSize);
	if (!pixels)
		return;

	glReadPixels(0,0, screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);

	if (xx >=0 && xx<=screen->winsizex && yy >=0 && yy<=screen->winsizey)
	{	floodFillTolerantRecursive(xx, yy, fill, rim);
//		pixels[xx+yy*screen->winsizex]=0xffff0000;
	}
	glDrawPixels(screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
/*
glBegin(GL_LINE_STRIP);
glVertex3d(0,0,0);
glVertex3d(200,0,0);
glVertex3d(200,100,0);
glVertex3d(0,100,0);
glVertex3d(0,0,0);
glEnd();
*/
	GC_::free(pixels); pixels=0;
#else
	*old=*this;
	*this=screen->Min;
	wxDC &dc=*recorder->graph->dc;
	xx=dc.LogicalToDeviceX(xx);
	yy=dc.LogicalToDeviceY(yy);
	wxSize wsize=recorder->graph->dc->GetSize();
	wxMemoryDC mdc;
	wxBitmap bmp(wsize.x, wsize.y);
	mdc.SelectObject(bmp);
	mdc.SetUserScale(1,1);
	mdc.SetDeviceOrigin(0, 0);
	double sx=1,sy=1;
	dc.GetUserScale(&sx, &sy);
	dc.SetUserScale(1,1);
	int ox=0,oy=0;
	dc.GetDeviceOrigin(&ox, &oy);
	int ok=mdc.Blit(0,0, wsize.x,wsize.y, &dc, 0,0);
	wxImage im=bmp.ConvertToImage();
	unsigned char *imp=im.GetData();
	pixels=(Color*)GC_::malloc(wsize.x*wsize.y*sizeof(Color));
	if (!pixels) return;
	for (int y=0; y<wsize.y; y++)
	{	unsigned char *s=&imp[wsize.x*y*3];
		ColorP d=&pixels[wsize.x*y];
		for (int x=0; x<wsize.x; x++)
		{	d->b=*s++;
			d->g=*s++;
			d->r=*s++;
			d->a=0xff;
			d++;
		}
	}
//fillbmp=new Bitmap(int(wsize.x),int(wsize.y), wsize.x,wsize.y, pixels);

	if throwing return;
	if (xx >=0 && xx<=screen->winsizex && yy >=0 && yy<=screen->winsizey)
		floodFillTolerantRecursive(xx, yy, fill, rim);

	int wx=wsize.x, wy=wsize.y;
	unsigned char *p=(unsigned char*)GC_::malloc(wx*wy*3);
	for (int yi=0; yi<wy; yi++)
	{	ColorP s=&pixels[(wy-1-yi)*wx];
		unsigned char *d=&p[yi*wx*3];
		for (int xi=wx-1; xi; xi--)
		{	*d++=s->r;
			*d++=s->g;
			*d++=s->b;
			s++;
		}
	}
	wxImage image(wsize.x, wsize.y, (unsigned char*)p, true);
	wxBitmap bitmap(image);
	mdc.SelectObject(bitmap);
	ok=dc.Blit(-wsize.x/2, -wsize.y/2, wsize.x,wsize.y, &mdc, 0,0);
	dc.SetUserScale(sx,sy);
	GC_::free(pixels);
	GC_::free(p);
	*this=*old;
#endif
}

void Turtle::floodFillTolerantRecursive(int x, int y, Color fill, Color rim)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*screen->winsizex];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (c.isLike(rim, colorTolerance) || c==fill) 
			break;
//		wxASSERT(p >= pixels);
		*p-- = fill;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
//	if (leftLimit > 0) 
//		leftLimit--;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*screen->winsizex];
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=*p;
		if (c.isLike(rim, colorTolerance) || c==fill)
			break;
//		wxASSERT(p < &pixels[pixelsSize]);
		*p++ = fill;
		rightLimit=i;
	}
//	if (rightLimit < int(screen->winsizex-1))
//		rightLimit++;

//bitPaste(fillbmp); mySwapBuffers();
	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y-1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c = *p;
			if (lastBorder) 
			{	if (!c.isLike(rim, colorTolerance) && c!=fill) 
				{	floodFillTolerantRecursive(i, y-1, fill, rim);
					lastBorder=false;
				}
			}
			else if (c.isLike(rim, colorTolerance)) 
				lastBorder=true;
			p++;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=true;
		p=&pixels[leftLimit+(y+1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c = *p;
			if (lastBorder) 
			{	if (!c.isLike(rim, colorTolerance) && c!=fill) 
				{	floodFillTolerantRecursive(i, y+1, fill, rim);
					lastBorder=false;
				}
			} 
			else if (c.isLike(rim, colorTolerance)) 
				lastBorder=true;
			p++;
		}
	}
}

void Turtle::floodFillSurfaceTolerant(int xx, int yy, Color fill, Color cin)
{
#ifdef OPENGL
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
	pixels=(Color*)GC_::malloc(screen->winsizex*screen->winsizey*sizeof(Color));
	if (!pixels)
		return;

	glReadPixels(0,0, screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);

	if (xx >=0 && xx<=screen->winsizex && yy >=0 && yy<=screen->winsizey)
	{	floodFillSurfaceRecursive(xx,yy, fill,cin);
	}
	glDrawPixels(screen->winsizex, screen->winsizey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
	GC_::free(pixels); pixels=0;
#else
	*old=*this;
	*this=screen->Min;
	wxDC &dc=*recorder->graph->dc;
	xx=dc.LogicalToDeviceX(xx);
	yy=dc.LogicalToDeviceY(yy);
	wxSize wsize=recorder->graph->dc->GetSize();
	wxMemoryDC mdc;
	wxBitmap bmp(wsize.x, wsize.y);
	mdc.SelectObject(bmp);
	mdc.SetUserScale(1,1);
	mdc.SetDeviceOrigin(0, 0);
	double sx=1,sy=1;
	dc.GetUserScale(&sx, &sy);
	dc.SetUserScale(1,1);
	int ox=0,oy=0;
	dc.GetDeviceOrigin(&ox, &oy);
	int ok=mdc.Blit(0,0, wsize.x,wsize.y, &dc, 0,0);
	wxImage im=bmp.ConvertToImage();
	unsigned char *imp=im.GetData();
	pixels=(Color*)GC_::malloc(wsize.x*wsize.y*sizeof(Color));
	if (!pixels) return;
	for (int y=0; y<wsize.y; y++)
	{	unsigned char *s=&imp[wsize.x*y*3];
		ColorP d=&pixels[wsize.x*y];
		for (int x=0; x<wsize.x; x++)
		{	d->b=*s++;
			d->g=*s++;
			d->r=*s++;
			d->a=0xff;
			d++;
		}
	}
//fillbmp=new Bitmap(int(wsize.x),int(wsize.y), wsize.x,wsize.y, pixels);

	if throwing return;
	if (xx >=0 && xx<=screen->winsizex && yy >=0 && yy<=screen->winsizey)
		floodFillSurfaceTolerantRecursive(xx, yy, fill, cin);

	int wx=wsize.x, wy=wsize.y;
	unsigned char *p=(unsigned char*)GC_::malloc(wx*wy*3);
	for (int yi=0; yi<wy; yi++)
	{	ColorP s=&pixels[(wy-1-yi)*wx];
		unsigned char *d=&p[yi*wx*3];
		for (int xi=wx-1; xi; xi--)
		{	*d++=s->r;
			*d++=s->g;
			*d++=s->b;
			s++;
		}
	}
	wxImage image(wsize.x, wsize.y, (unsigned char*)p, true);
	wxBitmap bitmap(image);
	mdc.SelectObject(bitmap);
	ok=dc.Blit(-wsize.x/2, -wsize.y/2, wsize.x,wsize.y, &mdc, 0,0);
	dc.SetUserScale(sx,sy);
	GC_::free(pixels);
	GC_::free(p);
	*this=*old;
#endif
}

void Turtle::floodFillSurfaceTolerantRecursive(int x, int y, Color fill, Color cin)
{
	int i, lastBorder, leftLimit, rightLimit;
	if (cin == fill)	// Nothing to be done
		return;
	
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*screen->winsizex];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (!c.isLike(cin, colorTolerance) || c==fill) 
			break;
		*p-- = fill;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*screen->winsizex];
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=*p;
		if (!c.isLike(cin, colorTolerance) || c==fill) 
			break;
		*p++ = fill;
		rightLimit=i;
	}
	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y-1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c = *p;
			if (lastBorder) 
			{	if (c.isLike(cin, colorTolerance) && c!=fill) 
				{	floodFillSurfaceTolerantRecursive(i, y-1, fill, cin);
					lastBorder=0;
				}
			}
			else if (!c.isLike(cin, colorTolerance)) 
				lastBorder=1;
			p++;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y+1)*screen->winsizex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c = *p;
			if (lastBorder) 
			{	if (c.isLike(cin, colorTolerance) && c!=fill) 
				{	floodFillSurfaceTolerantRecursive(i, y+1, fill, cin);
					lastBorder=0;
				}
			} 
			else if (!c.isLike(cin, colorTolerance)) 
				lastBorder=1;
			p++;
		}
	}
}

void Turtle::setPixel(int x, int y, Color c)
{
//c_out << 'S' << x << ' ' << y << endL;
#ifdef OPENGL
	Point p=Point(float(x)/screen->winsizex*2-1, float(y)/screen->winsizey*2-1, -1);
	glRasterPos3fv((float*)&p);
//c_out << p.x << ' ' << p.y << endL;
	glDrawPixels(1,1, GL_RGBA, GL_UNSIGNED_BYTE, &c);
	p=Point(-1,-1,-1);
	glRasterPos3fv((float*)&p);
/*	glBegin(GL_POINTS);
		glColor4ubv((GLubyte*)&c);
		glVertex2fv((GLfloat*)&p);
	glEnd();
*/
//mySwapBuffers();
#else
	pen.setColor(c);
	recorder->graph->dc->DrawPoint(x, y);
mySwapBuffers();
#endif
}

Color Turtle::getPixel(int x, int y)
{
	ThreadP th=Thread::This();
	Color c;
#ifdef OPENGL
	glReadPixels(x,y, 1,1, GL_RGBA, GL_UNSIGNED_BYTE, &c);
#else
	wxColour wxc;
	recorder->graph->dc->GetPixel(x,y, &wxc);
	c=wxc.GetPixel();
	c.a=0xff;
//c_out << 'G' << x << ' ' << y << ' ' << c << endL;
#endif
//	c=GetPixel(HDC(th->graph->GetHDC()), x,y);
	return c;
}

void Turtle::floodFillSmall(int x, int y, Color fill, Color rim, int safe)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c;

	if (safe > 5000)
		return;

	// Seek left
	leftLimit=-1;
	for (i=x; i >= 0; i--) 
	{	c=getPixel(i,y);
		if (c==rim || c==fill) 
			break;
//		wxASSERT(p >= pixels);
		setPixel(i,y,fill);
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
//	if (leftLimit > 0) 
//		leftLimit--;
	
	// Seek right
	rightLimit=x;
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=getPixel(i,y);
		if (c==rim || c==fill)
			break;
//		wxASSERT(p < &pixels[pixelsSize]);
		setPixel(i,y,fill);
		rightLimit=i;
	}
//	if (rightLimit < int(screen->winsizex-1))
//		rightLimit++;

	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c=getPixel(i,y-1);
			if (lastBorder) 
			{	if (c!=rim && c!=fill) 
				{	floodFillSmall(i, y-1, fill, rim, safe+1);
					lastBorder=false;
				}
			}
			else if (c==rim) 
				lastBorder=true;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c=getPixel(i,y+1);
			if (lastBorder) 
			{	if (c!=rim && c!=fill) 
				{	floodFillSmall(i, y+1, fill, rim, safe+1);
					lastBorder=false;
				}
			} 
			else if (c==rim) 
				lastBorder=true;
		}
	}
}

void Turtle::floodFillSmallSurface(int x, int y, Color fill, Color cin, int safe)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c;
	
	if (safe > 5000)
		return;

	// Seek left
	leftLimit=-1;
	for (i=x; i >= 0; i--) 
	{	c=getPixel(i,y);
		if (c!=cin || c==fill) 
			break;
//		wxASSERT(p >= pixels);
		setPixel(i,y,fill);
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
//	if (leftLimit > 0) 
//		leftLimit--;
	
	// Seek right
	rightLimit=x;
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=getPixel(i,y);
		if (c!=cin || c==fill)
			break;
//		wxASSERT(p < &pixels[pixelsSize]);
		setPixel(i,y,fill);
		rightLimit=i;
	}
//	if (rightLimit < int(screen->winsizex-1))
//		rightLimit++;

	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c=getPixel(i,y-1);
			if (lastBorder) 
			{	if (c==cin && c!=fill) 
				{	floodFillSmallSurface(i, y-1, fill, cin, safe+1);
					lastBorder=false;
				}
			}
			else if (c!=cin) 
				lastBorder=true;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c=getPixel(i,y+1);
			if (lastBorder) 
			{	if (c==cin && c!=fill) 
				{	floodFillSmallSurface(i, y+1, fill, cin, safe+1);
					lastBorder=false;
				}
			} 
			else if (c!=cin) 
				lastBorder=true;
		}
	}
}

void Turtle::floodFillSmallTolerant(int x, int y, Color fill, Color rim, int safe)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c;
	
	if (safe > 5000)
		return;

	// Seek left
	leftLimit=-1;
	for (i=x; i >= 0; i--) 
	{	c=getPixel(i,y);
		if (c.isLike(rim, colorTolerance) || c==fill) 
			break;
//		wxASSERT(p >= pixels);
		setPixel(i,y,fill);
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
//	if (leftLimit > 0) 
//		leftLimit--;
	
	// Seek right
	rightLimit=x;
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=getPixel(i,y);
		if (c.isLike(rim, colorTolerance) || c==fill)
			break;
//		wxASSERT(p < &pixels[pixelsSize]);
		setPixel(i,y,fill);
		rightLimit=i;
	}
//	if (rightLimit < int(screen->winsizex-1))
//		rightLimit++;

	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c=getPixel(i,y-1);
			if (lastBorder) 
			{	if (!c.isLike(rim, colorTolerance) && c!=fill) 
				{	floodFillSmallTolerant(i, y-1, fill, rim, safe+1);
					lastBorder=false;
				}
			}
			else if (c.isLike(rim, colorTolerance)) 
				lastBorder=true;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c=getPixel(i,y+1);
			if (lastBorder) 
			{	if (!c.isLike(rim, colorTolerance) && c!=fill) 
				{	floodFillSmallTolerant(i, y+1, fill, rim, safe+1);
					lastBorder=false;
				}
			} 
			else if (c.isLike(rim, colorTolerance)) 
				lastBorder=true;
		}
	}
}

void Turtle::floodFillSmallSurfaceTolerant(int x, int y, Color fill, Color cin, int safe)
{
	bool lastBorder;
	int i, leftLimit, rightLimit;
	Color c;

	if (safe > 5000)
		return;
	
	// Seek left
	leftLimit=-1;
	for (i=x; i >= 0; i--) 
	{	c=getPixel(i,y);
		if (!c.isLike(cin, colorTolerance) || c==fill) 
			break;
//		wxASSERT(p >= pixels);
		setPixel(i,y,fill);
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
//	if (leftLimit > 0) 
//		leftLimit--;
	
	// Seek right
	rightLimit=x;
	for (i=x+1; i < screen->winsizex; i++) 
	{	c=getPixel(i,y);
		if (!c.isLike(cin, colorTolerance) || c==fill)
			break;
//		wxASSERT(p < &pixels[pixelsSize]);
		setPixel(i,y,fill);
		rightLimit=i;
	}
//	if (rightLimit < int(screen->winsizex-1))
//		rightLimit++;

	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c=getPixel(i,y-1);
			if (lastBorder) 
			{	if (c.isLike(cin, colorTolerance) && c!=fill) 
				{	floodFillSmallSurfaceTolerant(i, y-1, fill, cin, safe+1);
					lastBorder=false;
				}
			}
			else if (!c.isLike(cin, colorTolerance)) 
				lastBorder=true;
		}
	}
	// Below
	if (y < screen->winsizey-1) 
	{	lastBorder=true;
		for (i=leftLimit; i <= rightLimit; i++) 
		{	c=getPixel(i,y+1);
			if (lastBorder) 
			{	if (c.isLike(cin, colorTolerance) && c!=fill) 
				{	floodFillSmallSurfaceTolerant(i, y+1, fill, cin, safe+1);
					lastBorder=false;
				}
			} 
			else if (!c.isLike(cin, colorTolerance)) 
				lastBorder=true;
		}
	}
}

void Turtle::fill(int version) 
{
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
//	if (direct)
//		console->dispatchMessages();
	if (checkThrowing()) return;
//	draw();
	Point old=*this;
	getInWindow();

	Point wp=getWindowPos();
	if (wp.x >=0 && wp.x<=screen->winsizex && wp.y >=0 && wp.y<=screen->winsizey)
	{
#ifdef OPENGL
		glColor4ubv((GLubyte*)&pen.floodcolor);

		GLboolean isat, isblend, isclop, iscm, isdi, isdt, isfog, 
			islop, isps, islight, istex;

		glGetBooleanv(GL_ALPHA_TEST, &isat);
		glGetBooleanv(GL_BLEND, &isblend);
		glGetBooleanv(GL_COLOR_LOGIC_OP, &isclop);
		glGetBooleanv(GL_COLOR_MATERIAL, &iscm);
		glGetBooleanv(GL_DEPTH_TEST, &isdt);
		glGetBooleanv(GL_DITHER, &isdi);
		glGetBooleanv(GL_FOG, &isfog);
		glGetBooleanv(GL_LIGHTING, &islight);
		glGetBooleanv(GL_LOGIC_OP, &islop);
		glGetBooleanv(GL_POINT_SMOOTH, &isps);
		glGetBooleanv(GL_TEXTURE_2D, &istex);

		glDisable(GL_ALPHA_TEST);
		glDisable(GL_BLEND);
		glDisable(GL_COLOR_LOGIC_OP);
		glDisable(GL_COLOR_MATERIAL);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_DITHER);
		glDisable(GL_FOG);
		glDisable(GL_LIGHTING);
		glDisable(GL_LOGIC_OP);
		glDisable(GL_POINT_SMOOTH);
		glDisable(GL_TEXTURE_2D);

		int mm;
		glGetIntegerv(GL_MATRIX_MODE, &mm);
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		Point p=Point(-1,-1,-1);
		glRasterPos3fv((float*)&p);
#endif
//		int result;
		switch (version)
		{
		case FILL_SURFACE:
			floodFillSurface(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
			break;
		case FILL_TOLERANT:
			floodFillTolerant(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
			break;
		case FILL_SURFACE_TOLERANT:
			floodFillSurfaceTolerant(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
			break;
		case FILL_SMALL:
#ifdef OPENGL
			floodFillSmall(int(wp.x), int(wp.y), pen.floodcolor, pen.color, 0);
#else
			floodFill(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
#endif
			break;
		case FILL_SMALL_SURFACE:
#ifdef OPENGL
			floodFillSmallSurface(int(wp.x), int(wp.y), pen.floodcolor, pen.color, 0);
#else
			floodFillSurface(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
#endif
			break;
		case FILL_SMALL_TOLERANT:
#ifdef OPENGL
			floodFillSmallTolerant(int(wp.x), int(wp.y), pen.floodcolor, pen.color, 0);
#else
			floodFillTolerant(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
#endif
			break;
		case FILL_SMALL_SURFACE_TOLERANT:
#ifdef OPENGL
			floodFillSmallSurfaceTolerant(int(wp.x), int(wp.y), pen.floodcolor, pen.color, 0);
#else
			floodFillSurfaceTolerant(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
#endif
			break;
		default:
			floodFill(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
/*			Color c=pen.floodcolor;
			Color rim=pen.color;
			HBRUSH brush=CreateSolidBrush(RGB(c.r,c.g,c.b));
			HDC hdc=HDC(th->graph->GetHDC());
			SelectObject(hdc, brush);
			result=ExtFloodFill(hdc, wp.x, screen->winsizey-wp.y, 
				RGB(rim.r,rim.g,rim.b), FLOODFILLBORDER);
			SelectObject(hdc, NULL);
			DeleteObject(brush);
*/		}
#ifdef OPENGL
		glClearDepth(1);
		glClear(GL_DEPTH_BUFFER_BIT);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(mm);

		if (isat)
			glEnable(GL_ALPHA_TEST);
		if (isblend)
			glEnable(GL_BLEND);
		if (isclop)
			glEnable(GL_COLOR_LOGIC_OP);
		if (iscm)
			glEnable(GL_COLOR_MATERIAL);
		if (isdt)
			glEnable(GL_DEPTH_TEST);
		if (isdi)
			glEnable(GL_DITHER);
		if (isfog)
			glEnable(GL_FOG);
		if (islight)
			glEnable(GL_LIGHTING);
		if (islop)
			glEnable(GL_LOGIC_OP);
		if (isps)
			glEnable(GL_POINT_SMOOTH);
		if (istex)
			glEnable(GL_TEXTURE_2D);

		glColor4ubv((GLubyte*)&pen.color);
		screen->getMatrices();
#endif
	}	
	*this=old;
//	draw();
	checkThrowing();
//	if (th->graph->direct)
//		mySwapBuffers();
	th->dispatchMessages();
}

// link with: freetype219MT.lib ftgl_static_MT.lib 

void Turtle::Label(wxString str)
{
#if defined(OPENGL)
#if defined(FTGL)
	font.load(recorder->graph);

	if (font.ftfont)
	{
//		draw();
		glPushMatrix();
			glTranslatef(x,y,z);
			if (screen->perspective)
				glScaled(1.5,1.5,1.5);

			double ad,as;
			ad=font.ftfont->Advance(str.c_str());
			as=font.ftfont->Ascender();
			if (screen->perspective)
			{	glRotated(roll(),0,1,0);
				glRotated(pitch(),1,0,0);
				glRotated(90-getHeading(),0,0,1);
			}else
				glRotated(90-heading,0,0,1);

			switch (font.alignx)
			{	case 1:	break;
				case 0:	glTranslatef(-ad/2,0,0); break;
				case -1:	glTranslatef(-ad,0,0); break;
			}
			switch (font.aligny)
			{	case 1: break;
				case 0: glTranslatef(0,-as/3,0); break;
				case -1: glTranslatef(0,-as,0); break;
			}
			if (screen->textureFont)
				glEnable(GL_TEXTURE_2D);			
			font.ftfont->Render(str.c_str());
			if (!screen->textured)
				glDisable(GL_TEXTURE_2D);
		glPopMatrix();
//		draw();
	}
#endif
#else
	bool rr=recorder->recording;
	recorder->recording=false;
	wxDC &dc=*recorder->graph->dc;
	dc.SetFont(wxFont(wxSize(font.sizex, font.sizey), 
		wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, 
		false, font.name));
	int w,h,lh;
	dc.GetTextExtent(str, &w,&lh);
	dc.GetMultiLineTextExtent(str, &w,&h);
	*old=*this;
	pen.down=false;
	right(-180);
	switch (font.alignx)
	{	case 1:	break;
		case 0:	forward(w/2); break;
		case -1:	forward(w); break;
	}
	right(90);
	switch (font.aligny)
	{	case 1: break;
		case 0: forward(h/2); break;
		case -1: forward(h); break;
	}
	right(90);
	while (str!=wxEmptyString)
	{	wxString s=str.BeforeFirst('\n');
		dc.DrawRotatedText(s, x,y, getHeading()-90);
		right(90); forward(lh); right(-90);
		str=str.AfterFirst('\n');
	}
	*this=*old;
	recorder->recording=rr;
#endif
}

Point Turtle::LabelSize(wxString str)
{
#if defined(OPENGL) && defined(FTGL)
	font.load(recorder->graph);

	if (font.ftfont)
	{	float llx,lly,llz,urx,ury,urz;
		if (screen->perspective)
			glScaled(1.5,1.5,1.5);
		font.ftfont->BBox(str.c_str(),llx,lly,llz,urx,ury,urz);
		return Point(urx-llx,ury-lly,urz-llz);
	}
#else
	wxDC &dc=*recorder->graph->dc;
	dc.SetFont(wxFont(wxSize(font.sizex, font.sizey), 
		wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, 
		false, font.name));
	int w,h;
	dc.GetMultiLineTextExtent(str, &w,&h);
	return Point(w,h,0);
#endif
	return Point();
}

void Turtle::setLabelSize(double x, double y)
{	font.sizex=x;
	font.sizey=y;
}
wxString Turtle::getLabelFont()
{	return font.name;
}
void Turtle::setLabelFont(wxString face)
{	font.name=face;
}
void Turtle::setLabelWeight(int weight)
{	font.weight=weight;
}
void Turtle::setLabelAlign(int alignx, int aligny)
{	font.alignx=alignx;
	font.aligny=aligny;
}

void Turtle::setPixel(Point p, Color c)
{
#ifdef OPENGL
	glBegin(GL_POINTS);
		glColor4ubv((GLubyte*)(&c));
		glVertex3fv((float*)&p);
	glEnd();
#else
	Color oldc=pen.color;
	pen.setColor(c);
	recorder->graph->dc->DrawPoint(p.x, p.y);
	pen.setColor(oldc);
#endif
	if (recorder->recording)
		*recorder << new R_setPixel(p, c);
}

Point Turtle::getWindowPos()
{
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

	double d[3]={0,0,0};
#ifdef OPENGL
	gluProject(x,y,z, 
		screen->modelview_matrix, 
		screen->projection_matrix, 
		screen->viewport, 
		&d[0],&d[1],&d[2]);
#else
	d[0]=x;
	d[1]=y;
	d[2]=z;
#endif
	*this=old;
	return Point(d[0],d[1],d[2]);
}

int Turtle::Pixel()	// Can this be made faster & easier? I hope so!
{
	Point wp=getWindowPos();
	Color c;
#ifdef OPENGL
	glReadPixels(int(wp.x),int(wp.y), 1,1, GL_RGBA, GL_UNSIGNED_BYTE, &c);
#else
	wxColour wxc;
	recorder->graph->dc->GetPixel(wp.x,wp.y, &wxc);
	c=wxc.GetPixel();
	c.a=0xff;
#endif
	return c;
}


BitmapP Turtle::bitCopy(double w, double h)
{
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

#ifdef OPENGL
	Point p1=getWindowPos().round();
	*this+=Point(w,h,0);
	Point win=(getWindowPos()-p1).abs().round();
	int winx=int(win.x), winy=int(win.y);
	*this=old;

	Color *pixels=(Color*)GC_::malloc(winx*winy*sizeof(Color));
	if (!pixels)
		return BitmapP(Unbound);
	glReadPixels(int(p1.x),int(p1.y), winx,winy, 
		GL_RGBA, GL_UNSIGNED_BYTE, pixels);
#else
	wxDC &dc=*recorder->graph->dc;
	wxCoord winx=dc.LogicalToDeviceXRel(w);
	wxCoord winy=dc.LogicalToDeviceYRel(h);
	wxCoord xx=dc.LogicalToDeviceX(x);
	wxCoord yy=dc.LogicalToDeviceY(y);
	wxMemoryDC mdc(&dc);
	wxBitmap bmp(winx, winy, mdc);
	mdc.SelectObject(bmp);
	mdc.SetUserScale(1,1);
	double sx=1,sy=1;
	dc.GetUserScale(&sx, &sy);
	dc.SetUserScale(1,1);
	int ok=mdc.Blit(0,0, winx,winy, &dc, xx,yy);
	dc.SetUserScale(sx,sy);
	wxImage im=bmp.ConvertToImage();
	unsigned char *imp=im.GetData();
	Color *pixels=(Color*)GC_::malloc(winx*winy*sizeof(Color));
	if (!pixels)
		return BitmapP(Unbound);
	for (int y=0; y<winy; y++)
	{	unsigned char *s=&imp[winx*y*3];
		ColorP d=&pixels[winx*y];
		for (int x=0; x<winx; x++)
		{	d->b=*s++;
			d->g=*s++;
			d->r=*s++;
			d->a=0xff;
			d++;
		}
	}
#endif
//	draw();
	BitmapP b=new Bitmap(int(w),int(h), winx,winy, pixels);
	return b;
}

void Turtle::bitPaste(BitmapP b)
{
//	draw();
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

//	double p[4];
//	glGetDoublev(GL_CURRENT_RASTER_POSITION, p);

#ifdef OPENGL
	glRasterPos3fv((float*)this);

	GLboolean isclop, iscm, isdi, isdt, isfog, 
		islop, isps, islight, istex;

	glGetBooleanv(GL_COLOR_LOGIC_OP, &isclop);
	glGetBooleanv(GL_COLOR_MATERIAL, &iscm);
	glGetBooleanv(GL_DEPTH_TEST, &isdt);
	glGetBooleanv(GL_DITHER, &isdi);
	glGetBooleanv(GL_FOG, &isfog);
	glGetBooleanv(GL_LIGHTING, &islight);
	glGetBooleanv(GL_LOGIC_OP, &islop);
	glGetBooleanv(GL_POINT_SMOOTH, &isps);
	glGetBooleanv(GL_TEXTURE_2D, &istex);

//	glDisable(GL_COLOR_LOGIC_OP);
	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_DITHER);
	glDisable(GL_FOG);
	glDisable(GL_LIGHTING);
	glDisable(GL_LOGIC_OP);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_TEXTURE_2D);

	glDrawPixels(b->winx,b->winy,
		GL_RGBA,GL_UNSIGNED_BYTE, b->pixels);

	if (isclop)
		glEnable(GL_COLOR_LOGIC_OP);
	if (iscm)
		glEnable(GL_COLOR_MATERIAL);
	if (isdt)
		glEnable(GL_DEPTH_TEST);
	if (isdi)
		glEnable(GL_DITHER);
	if (isfog)
		glEnable(GL_FOG);
	if (islight)
		glEnable(GL_LIGHTING);
	if (islop)
		glEnable(GL_LOGIC_OP);
	if (isps)
		glEnable(GL_POINT_SMOOTH);
	if (istex)
		glEnable(GL_TEXTURE_2D);

//	glRasterPos4fv(p);
//	Point p=Point(-screen->winsizex/2,-screen->winsizey/2,0);
	Point p=Point(screen->Min.x,screen->Min.y,0);
	glRasterPos3fv((float*)&p);
	
//	Float::reset_FPU();// I don't know why this is neccessary, but it HELPS!
#else
	int wx=b->winx, wy=b->winy;
	unsigned char *p=(unsigned char*)GC_::malloc(wx*wy*3);
	for (int yi=wy-1; yi; yi--)
	{	ColorP s=&b->pixels[yi*wx];
		unsigned char *d=&p[yi*wx*3];
		for (int xi=wx-1; xi; xi--)
		{	*d++=s->b;
			*d++=s->g;
			*d++=s->r;
			s++;
		}
	}
	wxImage image(b->winx, b->winy, (unsigned char*)p, true);
	wxBitmap bitmap(image);
	wxDC &dc=*recorder->graph->dc;
	wxCoord xx=dc.LogicalToDeviceXRel(x);
	wxCoord yy=dc.LogicalToDeviceYRel(y);
	double sx=1,sy=1;
	dc.GetUserScale(&sx, &sy);
	dc.SetUserScale(1,1);
	dc.DrawBitmap(bitmap, xx, yy);
	dc.SetUserScale(sx,sy);
	GC_::free(p);
#endif
	*this=old;
//	draw();
}

void Turtle::bitPasteFast(BitmapP b)
{
#ifdef OPENGL
//	draw();
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

	double sx,sy, dummyz;
	gluProject(x, y, 0, 
		screen->modelview_matrix, 
		screen->projection_matrix, 
		screen->viewport, 
		&sx, &sy, &dummyz);

	int wx=b->winx, wy=b->winy;
	ColorP p=ColorP(malloc(wx*wy*sizeof(Color)));
	for (int yi=wy; yi > 0; yi--)
	{	ColorP 
			s=&b->pixels[(yi-1)*wx],
			d=&p[(wy-yi)*wx];
		for (int xi=wx; xi > 0; xi--)
		{	*d++=Color(s->b, s->g, s->r);
			s++;
		}
	}
#ifdef WIN32
	HBITMAP bmp=CreateBitmap(b->winx, b->winy, 1, 32, p);
	HDC mdc=CreateCompatibleDC((HDC)recorder->graph->GetHDC());
	SelectObject(mdc, bmp);
	BitBlt((HDC)recorder->graph->GetHDC(), sx,sy, b->winx, b->winy, mdc, 0,0, SRCCOPY);
	DeleteObject(bmp);
	DeleteDC(mdc);
#else
//	TODO
#endif
	free(p);

	*this=old;
//	draw();
#else
	bitPaste(b);
#endif
}

void Turtle::scroll(int width, int height, double sx, double sy)
{
#ifdef OPENGL
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();
	glRasterPos3fv((float*)&th->graph->winorigin);
	*this=old;

	width=screen->winsizex;
	height=screen->winsizey;

	Color *pixels=(Color*)GC_::malloc(width*height*4);
	if (!pixels)
		return;
//	draw();

	GLboolean isclop, iscm, isdi, isdt, isfog, 
		islop, isps, islight, istex;

	glGetBooleanv(GL_COLOR_LOGIC_OP, &isclop);
	glGetBooleanv(GL_COLOR_MATERIAL, &iscm);
	glGetBooleanv(GL_DEPTH_TEST, &isdt);
	glGetBooleanv(GL_DITHER, &isdi);
	glGetBooleanv(GL_FOG, &isfog);
	glGetBooleanv(GL_LIGHTING, &islight);
	glGetBooleanv(GL_LOGIC_OP, &islop);
	glGetBooleanv(GL_POINT_SMOOTH, &isps);
	glGetBooleanv(GL_TEXTURE_2D, &istex);

	glDisable(GL_COLOR_LOGIC_OP);
	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_DITHER);
	glDisable(GL_FOG);
	glDisable(GL_LIGHTING);
	glDisable(GL_LOGIC_OP);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_TEXTURE_2D);

	glReadPixels(int(sx),int(sy), int(width-sx),int(height-sy), 
		GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	glDrawPixels(int(width-sx),int(height-sy),
		GL_RGBA,GL_UNSIGNED_BYTE, pixels);

	if (isclop)
		glEnable(GL_COLOR_LOGIC_OP);
	if (iscm)
		glEnable(GL_COLOR_MATERIAL);
	if (isdt)
		glEnable(GL_DEPTH_TEST);
	if (isdi)
		glEnable(GL_DITHER);
	if (isfog)
		glEnable(GL_FOG);
	if (islight)
		glEnable(GL_LIGHTING);
	if (islop)
		glEnable(GL_LOGIC_OP);
	if (isps)
		glEnable(GL_POINT_SMOOTH);
	if (istex)
		glEnable(GL_TEXTURE_2D);

	GC_::free(pixels); pixels=0;
//	glCopyPixels(sx,sy, width-sx,height-sy, GL_COLOR);
	if (th->graph->direct)
		mySwapBuffers();
//	draw();
#endif
}

Point Turtle::scrollCalibrate(double sx, double sy)
{
	return Point(
		(sx*screen->winsizex)/800,(sy*screen->winsizey)/600,0);
}
