#include "stdafx.h"
//___________________________________________________________________
//
//                           Class  Window_
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/window.h>
#include <wx/datstrm.h>
#include <wx/textctrl.h>
#include <wx/sizer.h>
#include <wx/caret.h>
#include <wx/aui/aui.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Table.h"
#include "LogoObject.h"
#include "Int.h"
#include "Bitmap.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "BoxSizer.h"
#include "Point.h"
#include "Graph.h"
#include "Window.h"
#include "Frame.h"
#include "UInt8.h"
#include "Color.h"
#include "Mouse.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"

#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"
#include "Mem.h"
#include "Aui.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Window_, wxWindow)
	EVT_CLOSE				(Window_::OnClose			)
	EVT_PAINT				(Window_::OnPaint			)
	EVT_ERASE_BACKGROUND	(Window_::OnErase			) 
	EVT_SET_FOCUS			(Window_::OnSetFocus		)
	EVT_KILL_FOCUS			(Window_::OnKillFocus	)
	EVT_MOVE					(Window_::OnMove			)
	EVT_SIZE					(Window_::OnSize			)
	EVT_CHAR					(Window_::OnChar			)
	EVT_KEY_DOWN			(Window_::OnKeyDown		)
	EVT_KEY_UP				(Window_::OnKeyUp			)
	EVT_LEFT_DOWN			(Window_::OnLeftDown		)
	EVT_RIGHT_DOWN			(Window_::OnRightDown	)
	EVT_MIDDLE_DOWN		(Window_::OnMiddleDown	)
	EVT_LEFT_UP				(Window_::OnLeftUp		)
	EVT_RIGHT_UP			(Window_::OnRightUp		)
	EVT_MIDDLE_UP			(Window_::OnMiddleUp		)
	EVT_LEFT_DCLICK		(Window_::OnLeftDClick	)	
	EVT_RIGHT_DCLICK		(Window_::OnRightDClick	)
	EVT_MIDDLE_DCLICK		(Window_::OnMiddleDClick)
	EVT_MOTION				(Window_::OnMotion		)
	EVT_ENTER_WINDOW		(Window_::OnEnterWindow	)
	EVT_LEAVE_WINDOW		(Window_::OnLeaveWindow	)
	EVT_MOUSEWHEEL			(Window_::OnMouseWheel	)
	EVT_MOUSE_EVENTS		(Window_::OnMouse			)
END_EVENT_TABLE()

Window_::Window_(WindowNodeP windowNode_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=windowNode_;
	thread=Thread::This();
	wxWindow *parent=0;
	if (WindowNodeP(*d(windowNode))->parent)
		parent=WindowNodeP(*d(windowNode))->parent->window;
	Create(parent, -1, 
			wxDefaultPosition, wxDefaultSize, 
			wxDEFAULT_FRAME_STYLE, _("Window_"));
}

Window_::Window_(WindowNodeP windowNode_,
			wxWindow* parent, wxWindowID id, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxString& name):
	wxWindow(parent, id, pos, size, style, name)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=windowNode_;
	thread=Thread::This();
}

Window_::~Window_()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
	if (!Thread::list.Find(thread))
		return;
	EventFiFo *fifo=thread->eventFiFo;
	EventPP out=fifo->out;
	EventP ev;
	while (*fifo)
	{	*fifo >> ev;
		switch (ev->type)
		{
		case Event::WindowChar:
		case Event::WindowKeyDown:
		case Event::WindowKeyUp:
		case Event::WindowClose:
			if (((EventWindow*)ev)->window==this)
				((EventWindow*)ev)->window=0;
			break;
		default:
			;
		}
	}
	fifo->out=out;
}

#define WINDOW_ON(NAME, EVENT)								\
void Window_::On##NAME(wx##EVENT &event)					\
{																		\
	if (*d(on##NAME))													\
	{	EventP evt=new EventWindow##NAME(this, event);	\
		if throwing return;										\
		*thread->eventFiFo << evt;								\
	}else																\
		event.Skip();												\
}

void Window_::OnPaint(wxPaintEvent &event)				
{																		
	if (*d(onPaint))													
	{	EventP evt=new EventWindowPaint(this, event);	
		if throwing return;										
		*thread->eventFiFo << evt;								
	}
	event.Skip();												
}

void Window_::OnErase(wxEraseEvent &event)				
{																		
	if (*d(onErase))													
	{	EventP evt=new EventWindowErase(this, event);	
		if throwing return;										
		*thread->eventFiFo << evt;								
	}
	event.Skip();												
}

void Window_::OnSize(wxSizeEvent &event)
{
	if (*d(onSize))
	{	EventP evt=new EventWindowSize(this, event);
		if throwing return;
		*thread->eventFiFo << evt;
	}else
		event.Skip();
}

WINDOW_ON(Char, KeyEvent)
//WINDOW_ON(Erase, EraseEvent)
WINDOW_ON(KeyDown, KeyEvent)
WINDOW_ON(KeyUp, KeyEvent)
WINDOW_ON(KillFocus, FocusEvent)
WINDOW_ON(Move, MoveEvent)
//WINDOW_ON(Paint, PaintEvent)
WINDOW_ON(SetFocus, FocusEvent)
//WINDOW_ON(Size, SizeEvent)


void Window_::OnClose(wxCloseEvent &event)
{
	if (onClose)
	{	EventP evt=new EventWindowClose(this, event);
		if throwing return;
		*thread->eventFiFo << evt;
		event.Skip(false);// wait for the onClose handler
	}else
	{	if (*d(windowNode))
			WindowNodeP(*d(windowNode))->window=0;
		delete this;
//		event.Skip();
	}
}

#define EVENT_ON_MOUSE(NAME)								\
void Window_::On##NAME(wxMouseEvent &event)			\
{																	\
	ThreadP th=Thread::This();								\
	EventP evt=new EventWindow##NAME(this, event);	\
	if throwing return;										\
	*th->eventFiFo << evt;									\
	event.Skip();												\
}

EVENT_ON_MOUSE(LeftDown)
EVENT_ON_MOUSE(RightDown)
EVENT_ON_MOUSE(MiddleDown)

EVENT_ON_MOUSE(LeftUp)
EVENT_ON_MOUSE(RightUp)
EVENT_ON_MOUSE(MiddleUp)

EVENT_ON_MOUSE(LeftDClick)
EVENT_ON_MOUSE(RightDClick)
EVENT_ON_MOUSE(MiddleDClick)
EVENT_ON_MOUSE(Motion)
EVENT_ON_MOUSE(EnterWindow)
EVENT_ON_MOUSE(LeaveWindow)
EVENT_ON_MOUSE(MouseWheel)
EVENT_ON_MOUSE(Mouse)

wxString Window_::wxstring()
{
	wxString str;
	str << _("Window") << wxT("[]")
		<< GetWindowStyle() << wxT(" [")
		<< GetPosition().x << ' ' << GetPosition().y << wxT("][")
		<< GetSize().x << ' ' << GetSize().y << wxT("] ")
		<< wxT(" [") << GetName() << ']';
	return str;
}

DataInputStream &Window_::read(DataInputStream &istr) 
{	
	wxString name, fontName;

	wxPoint pos;
	wxSize size;
	int style, fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> pos.x >> pos.y >> size.x >> size.y >> style >> name
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight
#define READ_ON(NAME)	>> (int&)*d(on##NAME) >> (int&)*d(uon##NAME) >> (int&)*d(ton##NAME)
		READ_ON(Char)
		READ_ON(KeyDown)
		READ_ON(KeyUp)
		READ_ON(Close)
		>> (int&)*d(boxSizer);

	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetName(name); 
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont font=GetFont();
	font.SetPointSize(fontSize);
	font.SetFaceName(fontName);
	font.SetStyle(fontStyle);
	font.SetWeight(fontWeight);
	SetFont(font);
	return istr;
}
wxDataOutputStream &Window_::write(wxDataOutputStream &ostr) 
{
	wxColour f=GetForegroundColour(), b=GetBackgroundColour();
	wxFont font=GetFont();
	ostr
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle()) << GetName()
		<< f.Red() << f.Green() << f.Blue() 
		<< b.Red() << b.Green() << b.Blue() 
		<< font.GetPointSize() << font.GetFaceName() 
		<< font.GetStyle() << font.GetWeight();
#define WRITE_ON(NAME)	\
	ostr << GC_::ptrToInt->find(*d( on##NAME));		\
	ostr << GC_::ptrToInt->find(*d(uon##NAME));	\
	ostr << GC_::ptrToInt->find(*d(ton##NAME));

	WRITE_ON(Char)
	WRITE_ON(KeyDown)
	WRITE_ON(KeyUp)
	WRITE_ON(Close)

	ostr << GC_::ptrToInt->find(*d(boxSizer));
	return ostr;
}

#define WINDOWNODE_ON(x)				\
PairP WindowNode::on##x()				\
{												\
	if (window)								\
		return PairP(*window->d(Window_::on##x));	\
	return 0;								\
}												\
												\
ObjectP WindowNode::uon##x()			\
{												\
	if (window)								\
		return ObjectP(*window->d(Window_::uon##x));	\
	return 0;								\
}												\
												\
LogoObjectP WindowNode::ton##x()		\
{												\
	if (window)								\
		return LogoObjectP(*window->d(Window_::ton##x));\
	return 0;								\
}												\
												\
void WindowNode::seton##x(NodeP nd)	\
{												\
	if (window)								\
	{	ThreadP th=Thread::This();		\
		*window->d(Window_:: on##x) = PairP(nd);		\
		*window->d(Window_::uon##x) = th->lexufun;	\
		*window->d(Window_::ton##x) = th->thisobj;	\
		check_oldyoung(nd);				\
	}											\
}

WINDOW_EVENTS(WINDOWNODE_ON)

BoxSizerNodeP WindowNode::boxSizer()
{
	if (window)
		return BoxSizerNodeP(*window->d(Window_::boxSizer));
	return 0;
}

void WindowNode::setSizer(BoxSizerNodeP boxSizer_)
{
	if ((window!=0) && boxSizer_)
	{	*window->d(Window_::boxSizer)=boxSizer_;
		check_oldyoung(boxSizer_);
		window->SetSizer(boxSizer_->boxSizer);
		window->Layout();
	}
}

void WindowNode::setSizerAndFit(BoxSizerNodeP boxSizer_)
{
	if ((window!=0) && boxSizer_)
	{	*window->d(Window_::boxSizer)=boxSizer_;
		check_oldyoung(boxSizer_);
		window->SetSizerAndFit(boxSizer_->boxSizer);
		window->Layout();
	}
}

void WindowNode::setParent(WindowNodeP window_)
{
	parent=window_;
}

WindowNode::WindowNode(): 
	Node(FRAME), window(0), parent(0), children(0) 
{}

WindowNode::WindowNode(NodeType type_): 
	Node(type_), window(0), parent(0), children(0)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

WindowNode::WindowNode(NodeType type_, WindowNodeP parent_): 
	Node(type_), window(0), parent(parent_), children(0)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void WindowNode::free() 
{	
	if (window) 
	{	
//		GUI_ENTER
		*window->d(Window_::windowNode)=0;
		WindowPairP c=children;
		while (c)
		{	if (c->first())
				c->first()->setParent(WindowNodeP(0));
			c=WindowPairP(c->tail());
		}				
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		window->Disable();
		if (window->thread)
		{	EventFiFo *fifo=window->thread->eventFiFo;
			EventPP out=fifo->out;
			EventP ev;
			while (*fifo)
			{	*fifo >> ev;
				switch (ev->type)
				{
				case Event::WindowChar:
				case Event::WindowKeyDown:
				case Event::WindowKeyUp:
				case Event::WindowClose:
					if (((EventWindow*)ev)->window==window)
						((EventWindow*)ev)->window=0;
					break;
				default:
					;
				}
			}
			fifo->out=out;
		}
		window->Destroy(); 
		window=0;
//		GUI_LEAVE
	}
}
void WindowNode::inc_gc() 
{	fifo
		<< children 
		WINDOW_EVENTS(WINDOW_MAKE_inc_gc)
		<< boxSizer();
}
int WindowNode::youngs() 
{	return 
		gmark(children)
		WINDOW_EVENTS(WINDOW_MAKE_youngs)
		+gmark(boxSizer());
}
int WindowNode::youngsPtr(PtrToInt_ &ptrToInt)
{	return 
		gmarkPtr(children, ptrToInt)
		WINDOW_EVENTS(WINDOW_MAKE_youngs)
		+gmarkPtr(boxSizer(), ptrToInt);
}
void WindowNode::check_oldyoungs() 
{	check_oldyoung(children);
	WINDOW_EVENTS(WINDOW_MAKE_check_oldyoung)
	check_oldyoung(boxSizer());
}

DataInputStream &WindowNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr 
			>> (int&)parentNode 
			>> (int&)children;
		window=new Window_(WindowNodeP(this));
		if (!window)
		{	error(OUT_OF_MEM); return istr;
		}
		window->read(istr);
		parent=parentNode;
	}
	return istr;
}
wxDataOutputStream &WindowNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (window)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent)
			<< GC_::ptrToInt->find(children);
		window->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void WindowNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	children = (WindowPair*)intToPtr.find(int(children));
	check_oldyoung(children);
	if (window)
	{
#define SETON(NAME)	\
	*window->d(Window_:: on##NAME)=((      Pair*)intToPtr.find(int(*window->d(Window_:: on##NAME))));	\
	*window->d(Window_::uon##NAME)=((    Object*)intToPtr.find(int(*window->d(Window_::uon##NAME))));	\
	*window->d(Window_::ton##NAME)=((LogoObject*)intToPtr.find(int(*window->d(Window_::ton##NAME))));

		SETON(Char)
		SETON(KeyDown)
		SETON(KeyUp)
		SETON(Close)

		*window->d(Window_::boxSizer) = (BoxSizerNode*)intToPtr.find(int(*window->d(Window_::boxSizer)));
		window->Show();
	}
}

void WindowNode::aftersetPtrs()
{
	if (parent)
		return;
	recreate();
}

void WindowNode::recreate()
{
	WindowP f=window, f2=0;
	if (!window)
		return;
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	if (!f) return;
	f2=new Window_(WindowNodeP(this),
			p, -1, f->GetPosition(), 
			f->GetSize(), f->GetWindowStyle(), f->GetName());
	if (!f2)
	{	error(OUT_OF_MEM); return;
	}
	f2->SetForegroundColour(f->GetForegroundColour());
	f2->SetBackgroundColour(f->GetBackgroundColour());
	wxFont font=f->GetFont();
	f2->SetFont(font);
	window=f2;
#define SET_ON(NAME) \
	*window->d(Window_:: on##NAME)=*(f->d(Window_:: on##NAME));	\
	*window->d(Window_::uon##NAME)=*(f->d(Window_::uon##NAME));	\
	*window->d(Window_::ton##NAME)=*(f->d(Window_::ton##NAME));	

		SET_ON(Char)
		SET_ON(KeyDown)
		SET_ON(KeyUp)
		SET_ON(Close)

		*window->d(Window_::boxSizer)=*f->d(Window_::boxSizer);
	delete f;

	window=f2;
	window->Show();

	PairP c=children;
	while (c)
	{	if (c->first())
			c->first()->recreate();
		next(c);
	}
}

void WindowNode::aftersetPtrs3()
{
	if (window)
		setSizer(BoxSizerNodeP(*window->d(Window_::boxSizer)));
}

StringP WindowNode::string()
{
	if (window)
		return new String(window->wxstring());
	return _Null_Word->name();
}

wxString WindowNode::wxstring()
{
	if (window)
		return window->wxstring();
	return wxEmptyString;
}

void WindowNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lWindow(PairPP arg)
{
	enum Paras
	{	_parent=1, _id, _pos, _size, _style, _name,
	};

	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[_parent])
	{	parentNode=arg[_parent]->a_window(_parent);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxWindowID id=wxID_ANY;
	id=arg[_id]->a_int(_id);
	wxPoint pos=wxDefaultPosition;
	if (exist(_pos) && arg[_pos])
	{	PairP p=arg[_pos]->a_int_vector(_pos);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(_size) && arg[_size])
	{	PairP s=arg[_size]->a_int_vector(_size);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	long style=wxDEFAULT_FRAME_STYLE;
	if (exist(_style) && arg[_style])
	{	style=arg[_style]->a_int(_style);
		if throwing return Unbound;
	}
	wxString name=_("Window");
	if (exist(_name))
	{	StringP n=arg[_name]->a_string(_name);
		if throwing return Unbound;
		name=n->ptr();
		name.Truncate(n->length());
	}
	WindowNodeP windowNode=new WindowNode(WINDOW, parentNode);
	if throwing return Unbound;
	WindowP window=new Window_(windowNode,
			parent, id, pos, size, style, name);
	if (!window)
	{	error(OUT_OF_MEM); return Unbound;
	}
	window->Show();
	windowNode->window=window;
	return windowNode;
}

NodeP lWindowCenter(PairPP arg)
{
	enum Paras {_window=1, _direction};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int direction=arg[_direction]->a_int(_direction);
	if throwing return Unbound;
	window->window->Center(direction);
	return Unbound;
}

NodeP lWindowCenterOnParent(PairPP arg)
{
	enum Paras {_window=1, _direction};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int direction=arg[_direction]->a_int(_direction);
	if throwing return Unbound;
	window->window->CenterOnParent(direction);
	return Unbound;
}

NodeP lWindowClearBackground(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->ClearBackground();
	return Unbound;
}

NodeP lWindowClientToScreen(PairPP arg)
{
	enum Paras {_window=1, _x, _y};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int x=arg[_x]->a_int(_x);
	if throwing return Unbound;
	int y=arg[_y]->a_int(_y);
	if throwing return Unbound;
	window->window->ClientToScreen(&x,&y);
	return new Pair(new Int(x), 
		new Pair(new Int(y), 0));
}

NodeP lWindowClose(PairPP arg)
{
	enum Paras {_window=1, _force};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int force=false;
	if (exist(_force))
	{	force=arg[_force]->a_bool(_force);
		if throwing return Unbound;
	}
	window->window->Close(force==1);
	return Unbound;
}

NodeP lWindowDestroy(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[1]->a_window(1);
	if throwing return Unbound;
	WindowPairP c=window->children;
	while (c)
	{	if (c->first())
			c->first()->setParent(WindowNodeP(0));
		c=WindowPairP(c->tail());
	}				
	{	WindowNodeP p=window->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(window));
			p->check_oldyoung(p->children);
		}
	}
	window->window->Destroy();
	window->window=0;
	return Unbound;
}

NodeP lWindowDestroyChildren(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->DestroyChildren();
	return Unbound;
}

NodeP lWindowDisable(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Disable();
	return Unbound;
}

NodeP lWindowEnable(PairPP arg)
{
	enum Paras {_window=1, _state};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int state=true;
	if (exist(_state))
	{	state=arg[_state]->a_bool(_state);
		if throwing return Unbound;
	}
	window->window->Enable(state!=0);
	return Unbound;
}

NodeP lWindowFindFocus(PairPP arg)
{
	wxWindow *window=WindowP(wxWindow::FindFocus());
	if (isLogoWindow(window))
	{	WindowNodeP windowNode=WindowNodeP(*window->d(Window_::windowNode));
		if (is_valid_pointer(windowNode))
			return windowNode;
	}
	return 0;
}

NodeP lWindowFindWindowById(PairPP arg)
{
	enum Paras {_id=1, _parent};
	int id=arg[_id]->a_int(_id);
	if throwing return Unbound;
	WindowP parent=0;
	if (exist(_parent))
	{	WindowNodeP parentNode=arg[_parent]->a_window(_parent);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	WindowP window=WindowP(wxWindow::FindWindowById(id, parent));
	if (isLogoWindow(window))
	{	WindowNodeP windowNode=WindowNodeP(*window->d(Window_::windowNode));
		if (is_valid_pointer(windowNode))
			return windowNode;
	}
	return 0;
}

NodeP lWindowFindWindowByLabel(PairPP arg)
{
	enum Paras {_label=1, _parent};
	StringP label=arg[_label]->a_string(_label);
	if throwing return Unbound;
	WindowP parent=0;
	if (exist(_parent))
	{	WindowNodeP parentNode=arg[_parent]->a_window(_parent);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	WindowP window=WindowP(wxWindow::FindWindowByLabel(label->ptr(), parent));
	if (isLogoWindow(window))
	{	WindowNodeP windowNode=WindowNodeP(*window->d(Window_::windowNode));
		if (is_valid_pointer(windowNode))
			return windowNode;
	}
	return 0;
}

NodeP lWindowFindWindowByName(PairPP arg)
{
	enum Paras {_name=1, _parent};
	StringP name=arg[_name]->a_string(_name);
	if throwing return Unbound;
	WindowP parent=0;
	if (exist(_parent))
	{	WindowNodeP parentNode=arg[_parent]->a_window(_parent);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	WindowP window=WindowP(wxWindow::FindWindowByLabel(name->ptr(), parent));
	if (isLogoWindow(window))
	{	WindowNodeP windowNode=WindowNodeP(*window->d(Window_::windowNode));
		if (is_valid_pointer(windowNode))
			return windowNode;
	}
	return 0;
}

NodeP lWindowFit(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Fit();
	return Unbound;
}

NodeP lWindowFitInside(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->FitInside();
	return Unbound;
}

NodeP lWindowFreeze(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Freeze();
	return Unbound;
}

NodeP lWindowGetBackgroundColor(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	Color color=window->window->GetBackgroundColour().GetPixel();
	return new Int(color);
}

NodeP lWindowGetBackgroundStyle(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int style=window->window->GetBackgroundStyle();
	return new Int(style);
}

NodeP lWindowGetBestSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxSize size=window->window->GetBestSize();
	return 
		new Pair(new Int(size.x),
		new Pair(new Int(size.y), 0));
}

NodeP lWindowGetCapture(PairPP arg)
{
	wxWindow *window=WindowP(wxWindow::GetCapture());
	if (isLogoWindow(window))
	{	WindowNodeP windowNode=WindowNodeP(*window->d(Window_::windowNode));
		if (is_valid_pointer(windowNode))
			return windowNode;
	}
	return 0;
}

NodeP lWindowGetCaret(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxCaret *caret=window->window->GetCaret();
	if (!caret)
		return 0;
	return 
		new Pair(new Int(caret->GetPosition().x),
		new Pair(new Int(caret->GetPosition().y), 0));
}

NodeP lWindowGetCharHeight(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int height=window->window->GetCharHeight();
	return new Int(height);
}

NodeP lWindowGetCharWidth(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int width=window->window->GetCharWidth();
	return new Int(width);
}

NodeP lWindowGetChildren(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxWindowList list=window->window->GetChildren();
	PairP result=0;
	for (wxWindowList::Node *node=list.GetFirst(); node; node=node->GetNext())
	{	wxWindow *window=node->GetData();
		if (isLogoWindow(window))
		{	WindowNodeP windowNode=WindowNodeP(*window->d(Window_::windowNode));
			if (is_valid_pointer(windowNode))
				push(result, windowNode);
		}
	}
	return result;
}

NodeP lWindowGetClientSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxSize size=window->window->GetClientSize();
	return 
		new Pair(new Int(size.x),
		new Pair(new Int(size.y), 0));
}

NodeP lWindowGetCursor(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxCursor cursor=window->window->GetCursor();
#ifdef WIN32
	int hcursor=int(cursor.GetHCURSOR());
#else
//	TODO
	int hcursor=0;
#endif
	return new Int(hcursor);
}

NodeP lWindowGetExtraStyle(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int style=window->window->GetExtraStyle();
	return new Int(style);
}

NodeP lWindowGetForegroundColor(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	Color c=window->window->GetForegroundColour().GetPixel();
	return new Int(c);
}

NodeP lWindowGetGrandParent(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxWindow *grandParent=window->window->GetGrandParent();
	if (isLogoWindow(grandParent))
	{	WindowNodeP windowNode=WindowNodeP(*grandParent->d(Window_::windowNode));
		if (is_valid_pointer(windowNode))
			return windowNode;
	}
	return 0;
}

NodeP lWindowGetHandle(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int handle=int(window->window->GetHandle());
	return new Int(handle);
}

NodeP lWindowGetHelpText(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxString helpText=window->window->GetHelpText();
	return new String(helpText);
}

NodeP lWindowGetId(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int id=window->window->GetId();
	return new Int(id);
}

NodeP lWindowGetLabel(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxString label=window->window->GetLabel();
	return new String(label);
}

NodeP lWindowGetMaxSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxSize size=window->window->GetMaxSize();
	return 
		new Pair(new Int(size.x),
		new Pair(new Int(size.y), 0));
}

NodeP lWindowGetMinSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxSize size=window->window->GetMinSize();
	return 
		new Pair(new Int(size.x),
		new Pair(new Int(size.y), 0));
}

NodeP lWindowGetName(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxString name=window->window->GetName();
	return new String(name);
}

NodeP lWindowGetParent(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxWindow *parent=window->window->GetParent();
	if (isLogoWindow(parent))
	{	WindowNodeP windowNode=WindowNodeP(*parent->d(Window_::windowNode));
		if (is_valid_pointer(windowNode))
			return windowNode;
	}
	return 0;
}

NodeP lWindowGetPosition(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxPoint position=window->window->GetPosition();
	return 
		new Pair(new Int(position.x),
		new Pair(new Int(position.y), 0));
}

NodeP lWindowGetRect(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxRect rect=window->window->GetRect();
	return
		new Pair(
			new Pair(new Int(rect.GetLeft()),
			new Pair(new Int(rect.GetBottom()), 0)),
		new Pair(
			new Pair(new Int(rect.GetRight()),
			new Pair(new Int(rect.GetTop()), 0)), 0));
}

NodeP lWindowGetScreenPosition(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxPoint screenPosition=window->window->GetScreenPosition();
	return 
		new Pair(new Int(screenPosition.x),
		new Pair(new Int(screenPosition.y), 0));
}

NodeP lWindowGetScreenRect(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxRect rect=window->window->GetScreenRect();
	return
		new Pair(
			new Pair(new Int(rect.GetLeft()),
			new Pair(new Int(rect.GetBottom()), 0)),
		new Pair(
			new Pair(new Int(rect.GetRight()),
			new Pair(new Int(rect.GetTop()), 0)), 0));
}

NodeP lWindowGetScrollPos(PairPP arg)
{
	enum Paras {_window=1, _orientation};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int orientation=arg[_orientation]->a_int(_orientation);
	if throwing return Unbound;
	int scrollPos=window->window->GetScrollPos(orientation);
	return new Int(scrollPos);
}

NodeP lWindowGetScrollRange(PairPP arg)
{
	enum Paras {_window=1, _orientation};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int orientation=arg[_orientation]->a_int(_orientation);
	if throwing return Unbound;
	int scrollRange=window->window->GetScrollRange(orientation);
	return new Int(scrollRange);
}

NodeP lWindowGetScrollThumb(PairPP arg)
{
	enum Paras {_window=1, _orientation};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int orientation=arg[_orientation]->a_int(_orientation);
	if throwing return Unbound;
	int scrollThumb=window->window->GetScrollThumb(orientation);
	return new Int(scrollThumb);
}

NodeP lWindowGetSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxSize size=window->window->GetSize();
	return 
		new Pair(new Int(size.x),
		new Pair(new Int(size.y), 0));
}

NodeP lWindowGetTextExtent(PairPP arg)
{
	enum Paras {_window=1, _string};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	StringP string=arg[_string]->a_string(_string);
	if throwing return Unbound;
	int x=0,y=0;
	window->window->GetTextExtent(string->ptr(), &x, &y);
	return 
		new Pair(new Int(x),
		new Pair(new Int(y), 0));
}

NodeP lWindowGetVirtualSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxSize size=window->window->GetVirtualSize();
	return 
		new Pair(new Int(size.x),
		new Pair(new Int(size.y), 0));
}

NodeP lWindowGetWindowBorderSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	wxSize size=window->window->GetWindowBorderSize();
	return 
		new Pair(new Int(size.x),
		new Pair(new Int(size.y), 0));
}

NodeP lWindowGetWindowStyle(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int style=window->window->GetWindowStyle();
	return new Int(style);
}

NodeP lWindowHasCapture(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool capture=window->window->HasCapture();
	if (capture)
		return _True;
	return _False;
}

NodeP lWindowHasFlag(PairPP arg)
{
	enum Paras {_window=1, _flag};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int flag=arg[_flag]->a_int(_flag);
	if throwing return Unbound;
	bool state=window->window->HasFlag(flag);
	if (state)
		return _True;
	return _False;
}

NodeP lWindowHasMultiplePages(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->HasMultiplePages();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowHasScrollbar(PairPP arg)
{
	enum Paras {_window=1, _orient};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int orient=arg[_orient]->a_int(_orient);
	if throwing return Unbound;
	bool state=window->window->HasScrollbar(orient);
	if (state)
		return _True;
	return _False;
}

NodeP lWindowHasTransparentBackground(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->HasTransparentBackground();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowHide(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Hide();
	return Unbound;
}

NodeP lWindowInheritAttributes(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->InheritAttributes();
	return Unbound;
}

NodeP lWindowInitDialog(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->InitDialog();
	return Unbound;
}

NodeP lWindowInvalidateBestSize(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->InvalidateBestSize();
	return Unbound;
}

NodeP lWindowIsDoubleBuffered(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->IsDoubleBuffered();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowIsEnabled(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->IsEnabled();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowIsExposed(PairPP arg)
{
	enum Paras {_window=1, _x, _y, _w, _h};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int x=arg[_x]->a_int(_x);
	if throwing return Unbound;
	int y=arg[_y]->a_int(_y);
	if throwing return Unbound;
	int w=0;
	if (exist(_w))
	{	w=arg[_w]->a_int(_w);
		if throwing return Unbound;
	}
	int h=0;
	if (exist(_h))
	{	h=arg[_h]->a_int(_h);
		if throwing return Unbound;
	}
	bool state=window->window->IsExposed(x,y,w,h);
	if (state)
		return _True;
	return _False;
}

NodeP lWindowIsFrozen(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->IsFrozen();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowIsRetained(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->IsRetained();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowIsShown(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->IsShown();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowIsShownOnScreen(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->IsShownOnScreen();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowIsTopLevel(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	bool state=window->window->IsTopLevel();
	if (state)
		return _True;
	return _False;
}

NodeP lWindowLayout(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=0;
	if (arg[_window])
	{	arg[_window]->a_window(_window);
		if throwing return Unbound;
		window->window->Layout();
	}else
		theApp->aui->Update();
	return Unbound;
}

NodeP lWindowLineDown(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->LineDown();
	return Unbound;
}

NodeP lWindowLineUp(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->LineUp();
	return Unbound;
}

NodeP lWindowLower(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Lower();
	return Unbound;
}

NodeP lWindowMakeModal(PairPP arg)
{
	enum Paras {_window=1, _flag};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int flag=arg[_flag]->a_bool(_flag);
	if throwing return Unbound;
	window->window->MakeModal(flag!=0);
	return Unbound;
}

NodeP lWindowMove(PairPP arg)
{
	enum Paras {_window=1, _x, _y};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int x=arg[_x]->a_int(_x);
	if throwing return Unbound;
	int y=arg[_y]->a_int(_y);
	if throwing return Unbound;
	window->window->Move(x,y);
	return Unbound;
}

NodeP lWindowMoveAfterInTabOrder(PairPP arg)
{
	enum Paras {_window=1, _win};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	WindowNodeP win=arg[_win]->a_window(_win);
	if throwing return Unbound;
	window->window->MoveAfterInTabOrder(win->window);
	return Unbound;
}

NodeP lWindowMoveBeforeInTabOrder(PairPP arg)
{
	enum Paras {_window=1, _win};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	WindowNodeP win=arg[_win]->a_window(_win);
	if throwing return Unbound;
	window->window->MoveBeforeInTabOrder(win->window);
	return Unbound;
}

NodeP lWindowNavigate(PairPP arg)
{
	enum Paras {_window=1, _flags};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int flags=arg[_flags]->a_int(_flags);
	if throwing return Unbound;
	window->window->Navigate(flags);
	return Unbound;
}

#define WINDOW_lWindowOn(NAME)							\
NodeP lWindowOn##NAME(PairPP arg)						\
{																	\
	enum Paras {_window=1, _commandlist};				\
	WindowNodeP window=										\
		arg[_window]->a_window(_window);					\
	if throwing return Unbound;							\
	PairP commandlist=										\
		arg[_commandlist]->a_runnable(_commandlist);	\
	if throwing return Unbound;							\
	window->seton##NAME(commandlist);					\
	return Unbound;											\
}

WINDOW_EVENTS(WINDOW_lWindowOn)

NodeP lWindowPageDown(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->PageDown();
	return Unbound;
}

NodeP lWindowPageUp(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->PageUp();
	return Unbound;
}

NodeP lWindowRaise(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Raise();
	return Unbound;
}

NodeP lWindowRefresh(PairPP arg)
{
	enum Paras {_window=1, _eraseBackground, _x, _y, _w, _h};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int eraseBackground=0;
	if (exist(_eraseBackground))
	{	eraseBackground=arg[_eraseBackground]->a_bool(_eraseBackground);
		if throwing return Unbound;
	}
	int x=0;
	if (exist(_x))
	{	x=arg[_x]->a_int(_x);
		if throwing return Unbound;
	}
	int y=0;
	if (exist(_y))
	{	y=arg[_y]->a_int(_y);
		if throwing return Unbound;
	}
	int w=-1;
	if (exist(_w))
	{	w=arg[_w]->a_int(_w);
		if throwing return Unbound;
	}
	int h=-1;
	if (exist(_h))
	{	h=arg[_h]->a_int(_h);
		if throwing return Unbound;
	}
	if (h>0)
	{	wxRect rect(x,y,w,h);
		window->window->Refresh(eraseBackground!=0, &rect);
	}else
		window->window->Refresh(eraseBackground!=0, 0);
	return Unbound;
}

NodeP lWindowRegisterHotKey(PairPP arg)
{
	enum Paras {_window=1, _hotkeyId, _modifiers, _virtualKeyCodex};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int hotkeyId=arg[_hotkeyId]->a_int(_hotkeyId);
	if throwing return Unbound;
	int modifiers=arg[_modifiers]->a_int(_modifiers);
	if throwing return Unbound;
	int virtualKeyCodex=arg[_virtualKeyCodex]->a_int(_virtualKeyCodex);
	if throwing return Unbound;
#ifdef WIN32
	bool state=window->window->RegisterHotKey(hotkeyId, modifiers, virtualKeyCodex);
#else
	bool state=false;
#endif
	if (state)
		return _True;
	return _False;
}

NodeP lWindowReleaseMouse(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->ReleaseMouse();
	return Unbound;
}

NodeP lWindowReparent(PairPP arg)
{
	enum Paras {_window=1, _win};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	WindowNodeP win=arg[_win]->a_window(_win);
	if throwing return Unbound;
	window->window->Reparent(win->window);
	return Unbound;
}

NodeP lWindowScreenToClient(PairPP arg)
{
	enum Paras {_window=1, _x, _y};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int x=arg[_x]->a_int(_x);
	if throwing return Unbound;
	int y=arg[_y]->a_int(_y);
	if throwing return Unbound;
	window->window->ScreenToClient(&x, &y);
	return 
		new Pair(new Int(x),
		new Pair(new Int(y), 0));
}

NodeP lWindowScrollLines(PairPP arg)
{
	enum Paras {_window=1, _lines};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int lines=arg[_lines]->a_int(_lines);
	if throwing return Unbound;
	window->window->ScrollLines(lines);
	return Unbound;
}

NodeP lWindowScrollPages(PairPP arg)
{
	enum Paras {_window=1, _pages};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int pages=arg[_pages]->a_int(_pages);
	if throwing return Unbound;
	window->window->ScrollPages(pages);
	return Unbound;
}

NodeP lWindowScrollWindow(PairPP arg)
{
	enum Paras {_window=1, _dx, _dy};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int dx=arg[_dx]->a_int(_dx);
	if throwing return Unbound;
	int dy=arg[_dy]->a_int(_dy);
	if throwing return Unbound;
	window->window->ScrollWindow(dx, dy);
	return Unbound;
}

NodeP lWindowSetAutoLayout(PairPP arg)
{
	enum Paras {_window=1, _autoLayout};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int autoLayout=arg[_autoLayout]->a_bool(_autoLayout);
	if throwing return Unbound;
	window->window->SetAutoLayout(autoLayout!=0);
	return Unbound;
}

NodeP lWindowSetBackgroundColor(PairPP arg)
{
	enum Paras {_window=1, _color};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	Color c=arg[_color]->a_color(_color);
	if throwing return Unbound;
	wxColour wc(c);
	window->window->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lWindowSetBackgroundStyle(PairPP arg)
{
	enum Paras {_window=1, _style};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int style=arg[_style]->a_int(_style);
	if throwing return Unbound;
	window->window->SetBackgroundStyle(wxBackgroundStyle(style));
	return Unbound;
}

NodeP lWindowSetCaret(PairPP arg)
{
	enum Paras {_window=1, _x, _y};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int x=arg[_x]->a_int(_x);
	if throwing return Unbound;
	int y=arg[_y]->a_int(_y);
	if throwing return Unbound;
	wxCaret *caret=window->window->GetCaret();
	if (caret)
		caret->Move(x,y);
	else
		window->window->SetCaret(
			new wxCaret(window->window, 10,15));
	return Unbound;
}

NodeP lWindowSetClientSize(PairPP arg)
{
	enum Paras {_window=1, _width, _height};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int width=arg[_width]->a_int(_width);
	if throwing return Unbound;
	int height=arg[_height]->a_int(_height);
	if throwing return Unbound;
	window->window->SetClientSize(width, height);
	return Unbound;
}

NodeP lWindowSetExtraStyle(PairPP arg)
{
	enum Paras {_window=1, _style};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int style=arg[_style]->a_int(_style);
	if throwing return Unbound;
	window->window->SetExtraStyle(style);
	return Unbound;
}

NodeP lWindowSetFocus(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->SetFocus();
	return Unbound;
}

NodeP lWindowSetFontName(PairPP arg)
{
	enum Paras {_window=1, _fontName};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	StringP fontName=arg[_fontName]->a_string(_fontName);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetFaceName(fontName->ptr());
	window->window->SetFont(f);
	return Unbound;
}

NodeP lWindowSetFontSize(PairPP arg)
{
	enum Paras {_window=1, _fontSize};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int fontSize=arg[_fontSize]->a_int(_fontSize);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetPointSize(fontSize);
	window->window->SetFont(f);
	return Unbound;
}

NodeP lWindowSetFontStyle(PairPP arg)
{
	enum Paras {_window=1, _fontStyle};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int fontStyle=arg[_fontStyle]->a_int(_fontStyle);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetStyle(fontStyle);
	window->window->SetFont(f);
	return Unbound;
}

NodeP lWindowSetFontWeight(PairPP arg)
{
	enum Paras {_window=1, _fontWeight};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int fontWeight=arg[_fontWeight]->a_int(_fontWeight);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetWeight(fontWeight);
	window->window->SetFont(f);
	return Unbound;
}

NodeP lWindowSetForegroundColor(PairPP arg)
{
	enum Paras {_window=1, _foregroundColor};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	Color foregroundColor=arg[_foregroundColor]->a_color(_foregroundColor);
	if throwing return Unbound;
	wxColour wc(foregroundColor);
	window->window->SetForegroundColour(wc);
	return Unbound;
}

NodeP lWindowSetHelpText(PairPP arg)
{
	enum Paras {_window=1, _helpText};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	StringP helpText=arg[_helpText]->a_string(_helpText);
	if throwing return Unbound;
	window->window->SetHelpText(helpText->ptr());
	return Unbound;
}

NodeP lWindowSetId(PairPP arg)
{
	enum Paras {_window=1, _id};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int id=arg[_id]->a_int(_id);
	if throwing return Unbound;
	window->window->SetId(id);
	return Unbound;
}

NodeP lWindowSetInitialSize(PairPP arg)
{
	enum Paras {_window=1, _width, _height};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int width=arg[_width]->a_int(_width);
	if throwing return Unbound;
	int height=arg[_height]->a_int(_height);
	if throwing return Unbound;
	window->window->SetInitialSize(wxSize(width, height));
	return Unbound;
}

NodeP lWindowSetLabel(PairPP arg)
{
	enum Paras {_window=1, _label};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	StringP label=arg[_label]->a_string(_label);
	if throwing return Unbound;
	window->window->SetLabel(label->ptr());
	return Unbound;
}

NodeP lWindowSetMaxSize(PairPP arg)
{
	enum Paras {_window=1, _width, _height};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int width=arg[_width]->a_int(_width);
	if throwing return Unbound;
	int height=arg[_height]->a_int(_height);
	if throwing return Unbound;
	window->window->SetMaxSize(wxSize(width, height));
	return Unbound;
}

NodeP lWindowSetMinSize(PairPP arg)
{
	enum Paras {_window=1, _width, _height};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int width=arg[_width]->a_int(_width);
	if throwing return Unbound;
	int height=arg[_height]->a_int(_height);
	if throwing return Unbound;
	window->window->SetMinSize(wxSize(width, height));
	return Unbound;
}

NodeP lWindowSetName(PairPP arg)
{
	enum Paras {_window=1, _name};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	StringP name=arg[_name]->a_string(_name);
	if throwing return Unbound;
	window->window->SetName(name->ptr());
	return Unbound;
}

NodeP lWindowSetOwnBackgroundColor(PairPP arg)
{
	enum Paras {_window=1, _color};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	Color c=arg[_color]->a_color(_color);
	if throwing return Unbound;
	wxColour wc(c);
	window->window->SetOwnBackgroundColour(wc);
	return Unbound;
}

NodeP lWindowSetOwnFontName(PairPP arg)
{
	enum Paras {_window=1, _fontName};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	StringP fontName=arg[_fontName]->a_string(_fontName);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetFaceName(fontName->ptr());
	window->window->SetOwnFont(f);
	return Unbound;
}

NodeP lWindowSetOwnFontSize(PairPP arg)
{
	enum Paras {_window=1, _fontSize};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int fontSize=arg[_fontSize]->a_int(_fontSize);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetPointSize(fontSize);
	window->window->SetOwnFont(f);
	return Unbound;
}

NodeP lWindowSetOwnFontStyle(PairPP arg)
{
	enum Paras {_window=1, _fontStyle};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int fontStyle=arg[_fontStyle]->a_int(_fontStyle);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetStyle(fontStyle);
	window->window->SetOwnFont(f);
	return Unbound;
}

NodeP lWindowSetOwnFontWeight(PairPP arg)
{
	enum Paras {_window=1, _fontWeight};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int fontWeight=arg[_fontWeight]->a_int(_fontWeight);
	if throwing return Unbound;
	wxFont f=window->window->GetFont();
	f.SetWeight(fontWeight);
	window->window->SetOwnFont(f);
	return Unbound;
}

NodeP lWindowSetOwnForegroundColor(PairPP arg)
{
	enum Paras {_window=1, _foregroundColor};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	Color foregroundColor=arg[_foregroundColor]->a_color(_foregroundColor);
	if throwing return Unbound;
	wxColour wc(foregroundColor);
	window->window->SetOwnForegroundColour(wc);
	return Unbound;
}

NodeP lWindowSetScrollbar(PairPP arg)
{
	enum Paras {_window=1, _orientation, _position, _thumbSize, _range, _refresh};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int orientation=arg[_orientation]->a_int(_orientation);
	if throwing return Unbound;
	int position=arg[_position]->a_int(_position);
	if throwing return Unbound;
	int thumbSize=arg[_thumbSize]->a_int(_thumbSize);
	if throwing return Unbound;
	int range=arg[_range]->a_int(_range);
	if throwing return Unbound;
	int refresh=arg[_refresh]->a_bool(_refresh);
	if throwing return Unbound;
	window->window->SetScrollbar(orientation, position, thumbSize, range, refresh!=0);
	return Unbound;
}

NodeP lWindowSetScrollPos(PairPP arg)
{
	enum Paras {_window=1, _orientation, _pos, _refresh};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int orientation=arg[_orientation]->a_int(_orientation);
	if throwing return Unbound;
	int pos=arg[_pos]->a_int(_pos);
	if throwing return Unbound;
	int refresh=arg[_refresh]->a_bool(_refresh);
	if throwing return Unbound;
	window->window->SetScrollPos(orientation, pos, refresh!=0);
	return Unbound;
}

NodeP lWindowSetSize(PairPP arg)
{
	enum Paras {_window=1, _width, _height};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int width=arg[_width]->a_int(_width);
	if throwing return Unbound;
	int height=arg[_height]->a_int(_height);
	if throwing return Unbound;
	if (!arg[_window])
	{	wxAuiPaneInfo &pane=theApp->aui->GetPane(window->window);
		wxSize mins=pane.min_size, maxs=pane.max_size, s(width, height);
		pane.BestSize(s).MinSize(s).MaxSize(s).FloatingSize(s);
		theApp->aui->Update();
		pane.MinSize(mins).MaxSize(maxs);
		theApp->aui->Update();
	}else
	{	window->window->SetSize(width, height);
	}
	return Unbound;
}

NodeP lWindowSetSizer(PairPP arg)
{
	enum Paras {_window=1, _boxSizer};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	BoxSizerNodeP boxSizer=arg[_boxSizer]->a_boxSizer(_boxSizer);
	if throwing return Unbound;
	window->setSizer(boxSizer);
	return Unbound;
}

NodeP lWindowSetSizerAndFit(PairPP arg)
{
	enum Paras {_window=1, _boxSizer};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	BoxSizerNodeP boxSizer=arg[_boxSizer]->a_boxSizer(_boxSizer);
	if throwing return Unbound;
	window->setSizerAndFit(boxSizer);
	return Unbound;
}

NodeP lWindowSetThemeEnabled(PairPP arg)
{
	enum Paras {_window=1, _enabled};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int enabled=arg[_enabled]->a_bool(_enabled);
	if throwing return Unbound;
	window->window->SetThemeEnabled(enabled!=0);
	return Unbound;
}

NodeP lWindowSetToolTip(PairPP arg)
{
	enum Paras {_window=1, _tip};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	StringP tip=arg[_tip]->a_string(_tip);
	if throwing return Unbound;
	window->window->SetToolTip(tip->ptr());
	return Unbound;
}

NodeP lWindowSetVirtualSize(PairPP arg)
{
	enum Paras {_window=1, _width, _height};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int width=arg[_width]->a_int(_width);
	if throwing return Unbound;
	int height=arg[_height]->a_int(_height);
	if throwing return Unbound;
	window->window->SetVirtualSize(width, height);
	return Unbound;
}

NodeP lWindowSetVirtualSizeHints(PairPP arg)
{
	enum Paras {_window=1, _minw, _minh, _maxw, _maxh};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int minw=arg[_minw]->a_int(_minw);
	if throwing return Unbound;
	int minh=arg[_minh]->a_int(_minh);
	if throwing return Unbound;
	int maxw=arg[_maxw]->a_int(_maxw);
	if throwing return Unbound;
	int maxh=arg[_maxh]->a_int(_maxh);
	if throwing return Unbound;
	window->window->SetVirtualSizeHints(minw, minh, maxw, maxh);
	return Unbound;
}

NodeP lWindowSetWindowStyle(PairPP arg)
{
	enum Paras {_window=1, _style};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int style=arg[_style]->a_int(_style);
	if throwing return Unbound;
	window->window->SetWindowStyle(style);
	return Unbound;
}

NodeP lWindowShow(PairPP arg)
{
	enum Paras {_window=1, _show};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int show=true;
	if (exist(_show))
	{	show=arg[_show]->a_bool(_show);
		if throwing return Unbound;
	}
	window->window->Show(show!=0);
	return Unbound;
}

NodeP lWindowThaw(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Thaw();
	return Unbound;
}

NodeP lWindowToggleWindowStyle(PairPP arg)
{
	enum Paras {_window=1, _flag};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int flag=arg[_flag]->a_int(_flag);
	if throwing return Unbound;
	window->window->ToggleWindowStyle(flag);
	return Unbound;
}

NodeP lWindowUnregisterHotKey(PairPP arg)
{
	enum Paras {_window=1, _hotkeyId};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	int hotkeyId=arg[_hotkeyId]->a_int(_hotkeyId);
	if throwing return Unbound;
#ifdef WIN32
	window->window->UnregisterHotKey(hotkeyId);
#endif
	return Unbound;
}

NodeP lWindowUpdate(PairPP arg)
{
	enum Paras {_window=1};
	WindowNodeP window=arg[_window]->a_window(_window);
	if throwing return Unbound;
	window->window->Update();
	return Unbound;
}

