/////////////////////////////////////////////////////////////////////////////
// Name:        spinctli.h
// Author:      John Labenski
// Created:     25/12/05
// Copyright:   John Labenski, 2002
// License:     wxWidgets
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#if defined(__GNUG__) && !defined(NO_GCC_PRAGMA)
    #pragma implementation "spinctli.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include <wx/valtext.h>     // for wxTextValidator
    #include <wx/textctrl.h>
    #include <wx/stattext.h>
#endif // WX_PRECOMP

#include "spinctli.h"
#include <math.h>

#if wxCHECK_VERSION(2,5,0)
    #include "wx/math.h"
#else
    #if defined(__VISUALC__) || defined(__BORLANDC__) || defined(__WATCOMC__)
        #include <float.h>
        #define wxFinite(x) _finite(x)
    #elif defined(__GNUG__)||defined(__GNUWIN32__)||defined(__DJGPP__)|| \
          defined(__SGI_CC__)||defined(__SUNCC__)||defined(__XLC__)|| \
          defined(__HPUX__)||defined(__MWERKS__)
        #define wxFinite(x) finite(x)
    #else
        #define wxFinite(x) ((x) == (x))
    #endif
#endif // wxCHECK_VERSION(2,5,0)

// NOTES : if the textctrl is focused and the program is ending, a killfocus
//         event is sent in MSW, this is why m_textCtrl is set to NULL in it's
//         destructor and there's so many checks for it not being NULL

//----------------------------------------------------------------------------
// wxSpinCtrlInt
//----------------------------------------------------------------------------

// the textctrl used for the wxSpinCtrlInt, needed for keypresses
class wxSpinCtrlIntTextCtrl : public wxTextCtrl
{
public:
    wxSpinCtrlIntTextCtrl( wxWindow *parent, wxWindowID id,
                           const wxString &value = wxEmptyString,
                           const wxPoint &pos = wxDefaultPosition,
                           const wxSize &size = wxDefaultSize,
                           long style = 0,
                           const wxValidator& validator = wxDefaultValidator,
                           const wxString &name = wxTextCtrlNameStr);

    // MSW sends extra kill focus event
    virtual ~wxSpinCtrlIntTextCtrl()
    {
        if (m_parent) m_parent->m_textCtrl = NULL;
        m_parent = NULL;
    }

    wxSpinCtrlInt *m_parent;

    void OnChar( wxKeyEvent &event );         // pass chars to wxSpinCtrlInt
    void OnKillFocus( wxFocusEvent &event );  // sync the spin to textctrl

private:
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(wxSpinCtrlIntTextCtrl,wxTextCtrl)
//  EVT_TEXT_ENTER( wxID_ANY, wxSpinCtrlIntTextCtrl::OnTextEnter ) // get them from spinctrldbl
//  EVT_TEXT( wxID_ANY, wxSpinCtrlIntTextCtrl::OnTextUpdate )      // get them from spinctrldbl
    EVT_CHAR( wxSpinCtrlIntTextCtrl::OnChar )
    EVT_KILL_FOCUS( wxSpinCtrlIntTextCtrl::OnKillFocus )
END_EVENT_TABLE()

wxSpinCtrlIntTextCtrl::wxSpinCtrlIntTextCtrl( wxWindow *parent, wxWindowID id,
                                              const wxString &value,
                                              const wxPoint &pos, const wxSize &size,
                                              long style,
                                              const wxValidator& validator,
                                              const wxString &name)
                       :wxTextCtrl( parent, id, value, pos, size, style,
                                    validator, name)
{
    m_parent = (wxSpinCtrlInt*)parent;
}

void wxSpinCtrlIntTextCtrl::OnChar( wxKeyEvent &event )
{
    if (m_parent) m_parent->OnChar( event );
}

void wxSpinCtrlIntTextCtrl::OnKillFocus( wxFocusEvent &event )
{
    if (m_parent) m_parent->SyncSpinToText(true);
    event.Skip();
}

//----------------------------------------------------------------------------
// wxSpinCtrlInt
//----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS( wxSpinCtrlInt, wxControl )

BEGIN_EVENT_TABLE(wxSpinCtrlInt,wxControl)
    EVT_SPIN_UP   ( wxID_ANY, wxSpinCtrlInt::OnSpinUp    )
    EVT_SPIN_DOWN ( wxID_ANY, wxSpinCtrlInt::OnSpinDown  )
    EVT_TEXT_ENTER( wxID_ANY, wxSpinCtrlInt::OnTextEnter )
    //EVT_TEXT      ( wxID_ANY, wxSpinCtrlInt::OnText      )
    EVT_SET_FOCUS ( wxSpinCtrlInt::OnFocus     )
    EVT_KILL_FOCUS( wxSpinCtrlInt::OnKillFocus )
END_EVENT_TABLE()

void wxSpinCtrlInt::Init()
{
    m_min = 0;
    m_max = 100;
    m_value = 0;
    m_default_value = 0;
    m_increment = 1;
    m_snap_ticks = false;
    m_spinButton = NULL;
    m_textCtrl = NULL;
    m_bestSize = wxSize(-999, -999);
}

bool wxSpinCtrlInt::Create( wxWindow *parent, wxWindowID id,
                            const wxString& value,
                            const wxPoint& pos, const wxSize& size,
                            long style,
                            int min, int max,
                            int initial,
                            int increment, 
                            const wxString& name)
{
    if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
                           wxDefaultValidator, name))
        return false;

    wxControl::SetLabel(name);
    wxControl::SetBackgroundColour(parent->GetBackgroundColour());
    wxControl::SetForegroundColour(parent->GetForegroundColour());

    int width = size.GetWidth(), height = size.GetHeight();

    wxSize best_size( DoGetBestSize() );
    if (width  == -1) width  = best_size.GetWidth();
    if (height == -1) height = best_size.GetHeight();

    // Create a validator for numbers, +-, and eE for exponential
    wxTextValidator validator(wxFILTER_INCLUDE_CHAR_LIST);

#if wxCHECK_VERSION(2, 5, 4)
    wxArrayString list;

    wxString valid_chars(wxT(" 0123456789+-eE"));
    size_t len = valid_chars.Length();
    for (size_t i=0; i<len; i++)
        list.Add(wxString(valid_chars.GetChar(i)));

    validator.SetIncludes(list);
#else
    wxStringList list;

    wxString valid_chars(wxT(" 0123456789+-eE"));
    size_t len = valid_chars.Length();
    for (size_t i=0; i<len; i++)
        list.Add(wxString(valid_chars.GetChar(i)));

    validator.SetIncludeList(list);
#endif // wxCHECK_VER(2, 5, 4)

	 if (name != wxEmptyString)
	    m_label = new wxStaticText(this, id, name, 
						wxPoint(0, 0), wxDefaultSize, 0, "staticText");
	 else
		 m_label=0;
    m_spinButton = new wxSpinButton( this, id, wxPoint(0,0), wxSize(-1, height),
                                     wxSP_ARROW_KEYS|wxSP_VERTICAL|wxSP_WRAP);
	 if (m_label)
	    m_textCtrl = new wxSpinCtrlIntTextCtrl( this, id, value,
                      wxPoint(0,0),
                      wxSize(width-m_spinButton->GetSize().GetWidth(), 
									  height-m_label->GetSize().GetHeight()),
                      wxTE_NOHIDESEL|wxTE_PROCESS_ENTER, validator);
    else
        m_textCtrl = new wxSpinCtrlIntTextCtrl( this, id, value,
                      wxPoint(0,0),
                      wxSize(width-m_spinButton->GetSize().GetWidth(), 
									  height),
                      wxTE_NOHIDESEL|wxTE_PROCESS_ENTER, validator);

    DoSetSize( pos.x, pos.y, width, height );
    SetInitialSize(wxSize(width, height));

    m_min = min;
    m_max = max;
    m_value = initial;
    m_default_value = initial;
    m_increment = increment;
	 m_textFormat = wxT("%i");

    // set the value here without generating an event
    if (!value.IsEmpty())
        m_textCtrl->SetValue(value);
    else
        m_textCtrl->SetValue(wxString::Format(m_textFormat.c_str(), initial));

    return true;
}

wxSpinCtrlInt::~wxSpinCtrlInt()
{
    if (m_textCtrl) // null this since MSW sends KILL_FOCUS on deletion
    {
        m_textCtrl->m_parent = NULL;

        wxSpinCtrlIntTextCtrl *text = m_textCtrl;
        m_textCtrl = NULL;
        delete text;
    }

    delete m_spinButton;
    m_spinButton = NULL;
}


#define wxSPINCTRLDBL_SPIN_WIDTH  15
#define wxSPINCTRLDBL_SPIN_HEIGHT 22

void wxSpinCtrlInt::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
    //wxPrintf(wxT("DoSetSize %d, %d %d %d %d %d\n"), GetId(), x, y, width, height, sizeFlags);

    wxSize bestSize( DoGetBestSize() );
    if (width < 0)  width  = bestSize.GetWidth();
    if (height < 0) height = bestSize.GetHeight();

    wxWindow::DoSetSize(x, y, width, height, sizeFlags);

    int spinwidth  = wxSPINCTRLDBL_SPIN_WIDTH;
    int spinheight = wxSPINCTRLDBL_SPIN_HEIGHT;
    if (m_spinButton)
        m_spinButton->GetSize( &spinwidth, &spinheight );

    int labelwidth  = 0;
    int labelheight = 0;
    if (m_label)
        m_label->GetSize(&labelwidth, &labelheight);


#ifdef __WIN95__   // humm... these used to be different
    if (m_textCtrl)   
		m_textCtrl->SetSize( 0, labelheight, 
			width - spinwidth, height-labelheight-2);
    if (m_spinButton) 
		m_spinButton->SetSize( width-spinwidth-2, labelheight, 
			-1, height-labelheight-2 );
    //m_textCtrl->SetSize( -3, -3, width - spinwidth, height );   // old wxWin < 2.3.2
    //m_spinButton->SetSize( width-spinwidth-4, -3, -1, height-1 );
#else
    if (m_textCtrl)   
		m_textCtrl->SetSize( 0, labelheight, 
			width - spinwidth, height-labelheight-2);
    if (m_spinButton) 
		m_spinButton->SetSize( width-spinwidth, labelheight, 
			-1, height-labelheight-2);
#endif
}

wxSize wxSpinCtrlInt::DoGetBestSize()
{
    //wxPrintf(wxT("GetBestSize %d\n"), GetId());
    if (m_bestSize.x == -999)
    {
        wxSpinCtrl spin((wxWindow*)this, wxID_ANY);
        m_bestSize = spin.GetBestSize();
		  if (GetLabel()!=wxEmptyString)
		  {   wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
            int y=label.GetBestSize().y;
            m_bestSize.y+=y; 
        }
        // oops something went wrong, set to reasonable value
        if (m_bestSize.GetWidth()  < 20)
            m_bestSize.SetWidth(95);
        if (m_bestSize.GetHeight() < 10)
            m_bestSize.SetHeight(wxSPINCTRLDBL_SPIN_HEIGHT);
    }
    return m_bestSize;
}

void wxSpinCtrlInt::DoSendEvent()
{
    wxCommandEvent event( wxEVT_COMMAND_SPINCTRL_UPDATED, GetId() );
    event.SetEventObject( this );
    event.SetInt( (int)(m_value+0.5) );
    if (m_textCtrl) event.SetString( m_textCtrl->GetValue() );
    GetEventHandler()->ProcessEvent( event );
}

void wxSpinCtrlInt::OnSpinUp( wxSpinEvent &WXUNUSED(event) )
{
    if (m_textCtrl && m_textCtrl->IsModified() )
        SyncSpinToText(false);

    if ( InRange(m_value + m_increment) )
    {
        m_value += m_increment;
        SetValue( m_value );
        DoSendEvent();
    }
}

void wxSpinCtrlInt::OnSpinDown( wxSpinEvent &WXUNUSED(event) )
{
    if (m_textCtrl && m_textCtrl->IsModified() )
        SyncSpinToText(false);

    if ( InRange(m_value - m_increment) )
    {
        m_value -= m_increment;
        SetValue( m_value );
        DoSendEvent();
    }
}

void wxSpinCtrlInt::OnTextEnter( wxCommandEvent &event )
{
    SyncSpinToText(true);
    event.Skip();
}

void wxSpinCtrlInt::OnText( wxCommandEvent &event )
{
    //wxPrintf(wxT("Text '%s'\n"), event.GetString()); fflush(stdout);
    event.Skip();
}

void wxSpinCtrlInt::OnChar( wxKeyEvent &event )
{
    int modifier = 1;
    if ( event.m_shiftDown   ) modifier  = 2;
    if ( event.m_controlDown ) modifier *= 10;
    if ( event.m_altDown     ) modifier *= 100;

    switch ( event.GetKeyCode() )
    {

        case WXK_UP :
        {
            if (m_textCtrl && m_textCtrl->IsModified()) SyncSpinToText(false);
            SetValue( m_value + m_increment*modifier );
            DoSendEvent();
            break;
        }
        case WXK_DOWN :
        {
            if (m_textCtrl && m_textCtrl->IsModified()) SyncSpinToText(false);
            SetValue( m_value - m_increment*modifier );
            DoSendEvent();
            break;
        }
        case WXK_PRIOR :  // pg-up
        {
            if (m_textCtrl && m_textCtrl->IsModified()) SyncSpinToText(false);
            SetValue( m_value + m_increment*10*modifier );
            DoSendEvent();
            break;
        }
        case WXK_NEXT :  // pg-down
        {
            if (m_textCtrl && m_textCtrl->IsModified()) SyncSpinToText(false);
            SetValue( m_value - m_increment*10*modifier );
            DoSendEvent();
            break;
        }
        case WXK_SPACE :
        {
            SetValue(m_value);

            event.Skip(false);
            break;
        }
        case WXK_ESCAPE :
        {
            SetDefaultValue();
            DoSendEvent();
            break;
        }
        case WXK_TAB :
        {
            wxNavigationKeyEvent new_event;
            new_event.SetEventObject( GetParent() );
            new_event.SetDirection( !event.ShiftDown() );
            // CTRL-TAB changes the (parent) window, i.e. switch notebook page
            new_event.SetWindowChange( event.ControlDown() );
            new_event.SetCurrentFocus( this );
            GetParent()->GetEventHandler()->ProcessEvent( new_event );
            break;
        }
        default : event.Skip(); break;
    }
}

void wxSpinCtrlInt::SetValue( int value )
{
    if (!m_textCtrl || !InRange(value))
        return;

    if ( m_snap_ticks && (m_increment != 0) )
    {
        int snap_value = (value - m_default_value) / m_increment;

        if (wxFinite(snap_value)) // FIXME what to do about a failure?
        {
            if (snap_value - floor(double(snap_value)) < ceil(double(snap_value)) - snap_value)
                value = int(m_default_value + floor(double(snap_value)) * m_increment);
            else
                value = int(m_default_value + ceil(double(snap_value)) * m_increment);
        }
    }

    wxString str(wxString::Format(m_textFormat.c_str(), value));

    if ((value != m_value) || (str != m_textCtrl->GetValue()))
    {
        m_textCtrl->SetValue( str );
        m_textCtrl->DiscardEdits();
        m_value = value;
		  long v=value;
        str.ToLong(&v);    // wysiwyg for textctrl
		  m_value=v;
    }
}

void wxSpinCtrlInt::SetValue( const wxString& text, bool force )
{
    if (!m_textCtrl) return;

    long value;
    if ( text.ToLong(&value) )
        SetValue( value );
    else if (force)
    {
        m_textCtrl->SetValue( text );
        m_textCtrl->DiscardEdits();
    }
}

void wxSpinCtrlInt::SetRange( int min_val, int max_val )
{
    //wxCHECK_RET(max_val > min_val, wxT("invalid spinctrl range"));
    m_min = min_val;
    m_max = max_val;

    if (HasRange())
    {
        if (m_value > m_max)
            SetValue(m_max);
        else if (m_value < m_min)
            SetValue(m_min);
    }
}

void wxSpinCtrlInt::SetIncrement( int increment )
{
    m_increment = increment;
    SetValue(m_value);
}

void wxSpinCtrlInt::SetFormat( const wxString& format )
{
    wxString wxstr;
    if ( wxstr.Printf(format.c_str(), 123456.123456) > 0 )
        m_textFormat = format;

    SetValue(m_value);
}

void wxSpinCtrlInt::SetDefaultValue( int default_value )
{
    if ( InRange(default_value) )
    {
        m_default_value = default_value;
        SetDefaultValue();
    }
}

void wxSpinCtrlInt::SetSnapToTicks(bool forceTicks)
{
    if (m_snap_ticks != forceTicks)
    {
        m_snap_ticks = forceTicks;
        SetValue( m_value );
    }
}

void wxSpinCtrlInt::OnFocus( wxFocusEvent &event )
{
    if (m_textCtrl)
        m_textCtrl->SetFocus(); // this is to pass TAB navigation

    event.Skip();
}

void wxSpinCtrlInt::OnKillFocus( wxFocusEvent &event )
{
    SyncSpinToText(true);
    event.Skip();
}

void wxSpinCtrlInt::SyncSpinToText(bool send_event, bool force_valid)
{
    if (!m_textCtrl)
        return;

    long txt_value;
    if ( m_textCtrl->GetValue().ToLong( &txt_value ) )
    {
        if ( force_valid || !HasRange() || InRange(txt_value) )
        {
            if (force_valid && HasRange())
            {
                if (txt_value > GetMax())
                    txt_value = GetMax();
                else if (txt_value < GetMin())
                    txt_value = GetMin();
            }

            if (m_value != txt_value)
            {
                SetValue( txt_value );
                if (send_event) DoSendEvent();
            }
        }
    }
    else if (force_valid)
    {
        // textctrl is out of sync, discard and reset
        SetValue(GetValue());
    }
}

bool wxSpinCtrlInt::SetFont( const wxFont &font )
{
    if (!m_textCtrl) return false;
    return m_textCtrl->SetFont( font );
}
wxFont wxSpinCtrlInt::GetFont() const
{
    if (!m_textCtrl) return GetFont();
    return m_textCtrl->GetFont();
}

bool wxSpinCtrlInt::SetBackgroundColour(const wxColour& colour)
{
    if (!m_textCtrl) return wxControl::SetBackgroundColour(colour);
    bool ret = false;
    ret = m_textCtrl->SetBackgroundColour(colour);
    m_textCtrl->Refresh(); // FIXME is this necessary in GTK/OSX
    return ret;
}
wxColour wxSpinCtrlInt::GetBackgroundColour() const
{
    if (!m_textCtrl) return wxControl::GetBackgroundColour();
    return m_textCtrl->GetBackgroundColour();
}

bool wxSpinCtrlInt::SetForegroundColour(const wxColour& colour)
{
    if (!m_textCtrl) return wxControl::SetForegroundColour(colour);
    bool ret = false;
    ret = m_textCtrl->SetForegroundColour(colour);
    m_textCtrl->Refresh();
    return ret;
}
wxColour wxSpinCtrlInt::GetForegroundColour() const

{
    if (!m_textCtrl) return wxControl::GetForegroundColour();
    return m_textCtrl->GetForegroundColour();
}
