////////////////////////////////////////////////////////////////////////////
// Name:		msw/spinctrl.h
// Purpose: 	wxSpinCtrl class declaration for Win32
// Author:		Vadim Zeitlin
// Modified by:
// Created: 	22.07.99
// RCS-ID:		$Id: spinctrl.h,v 1.24 2005/02/28 01:23:01 VZ Exp $
// Copyright:	(c) Vadim Zeitlin
// Licence: 	wxWindows licence
/////////////////////////////////////////////////////////////////////////////

#ifndef _WX_MSW_SPINREAL_H_
#define _WX_MSW_SPINREAL_H_

#if defined(__GNUG__) && !defined(NO_GCC_PRAGMA)
	#pragma interface "spinctrl.h"
#endif

#include <wx/spinbutt.h>	// the base class

#if wxUSE_SPINCTRL

#include <wx/dynarray.h>

class WXDLLEXPORT wxSpinReal;
WX_DEFINE_EXPORTED_ARRAY_PTR(wxSpinReal *, wxArraySpinReals);

// ----------------------------------------------------------------------------
// Under Win32, wxSpinCtrl is a wxSpinButton with a buddy (as MSDN docs call
// it) text window whose contents is automatically updated when the spin
// control is clicked.
// ----------------------------------------------------------------------------

class WXDLLEXPORT wxSpinReal: public wxSpinButton
{
public:
	wxSpinReal() { }

	wxSpinReal(wxWindow *parent,
			   wxWindowID id = wxID_ANY,
			   const wxString& value = wxEmptyString,
			   const wxPoint& pos = wxDefaultPosition,
			   const wxSize& size = wxDefaultSize,
			   long style = wxSP_ARROW_KEYS,
			   double min = 0, double max = 100, double initial = 0,
			   double increment = 0.1, int digits = -1,
			   const wxString& name = _T("wxSpinReal"))
	{
		Create(parent, id, value, pos, size, style, 
			min, max, initial, increment, digits, name);
	}

	bool Create(wxWindow *parent,
				wxWindowID id = wxID_ANY,
				const wxString& value = wxEmptyString,
				const wxPoint& pos = wxDefaultPosition,
				const wxSize& size = wxDefaultSize,
				long style = wxSP_ARROW_KEYS,
				double min = 0, double max = 100, double initial = 0,
				double increment = 0.1, int digits = -1,
				const wxString& name = _T("wxSpinReal"));

	// a wxTextCtrl-like method (but we can't have GetValue returning wxString
	// because the base class already has one returning int!)
	void SetValue(const wxString& text);

	// another wxTextCtrl-like method
	void SetSelection(long from, long to);

	// implementation only from now on
	// -------------------------------

	virtual ~wxSpinReal();

	virtual void SetValue(double val);// { wxSpinButton::SetValue(val); }
	virtual double GetValue();
	virtual void SetIncrement(double increment, int digits);
    virtual double GetIncrement() const { return m_increment; }
    virtual int GetDigits() const { return m_digits; }
    virtual void SetSnap(bool snap_to_increment);
    virtual bool GetSnap() const { return m_snap; }
	
	virtual bool SetFont(const wxFont &font);
	virtual void SetFocus();

	virtual bool Enable(bool enable = true);
	virtual bool Show(bool show = true);

	// wxSpinButton doesn't accept focus, but we do
	virtual bool AcceptsFocus() const { return wxWindow::AcceptsFocus(); }

	// for internal use only

	// get the subclassed window proc of the buddy text
	WXFARPROC GetBuddyWndProc() const { return m_wndProcBuddy; }

	// return the spinctrl object whose buddy is the given window or NULL
	static wxSpinReal *GetSpinForTextCtrl(WXHWND hwndBuddy);

	// process a WM_COMMAND generated by the buddy text control
	bool ProcessTextCommand(WXWORD cmd, WXWORD id);

protected:
	virtual void DoGetPosition(int *x, int *y) const;
	virtual void DoMoveWindow(int x, int y, int width, int height);
	virtual wxSize DoGetBestSize() const;
	virtual void DoGetSize(int *width, int *height) const;
#if wxUSE_TOOLTIPS
	virtual void DoSetToolTip( wxToolTip *tip );
#endif // wxUSE_TOOLTIPS

	// the handler for wxSpinButton events
	void OnSpinChange(wxSpinEvent& event);

	// Handle processing of special keys
	void OnChar(wxKeyEvent& event);
	void OnSetFocus(wxFocusEvent& event);

	// the data for the "buddy" text ctrl
	WXHWND	   m_hwndBuddy;
	WXFARPROC  m_wndProcBuddy;

	// all existing wxSpinCtrls - this allows to find the one corresponding to
	// the given buddy window in GetSpinForTextCtrl()
	static wxArraySpinReals ms_allSpins;

    double m_value;
    double m_max;
    double m_min;
    double m_increment;
    double m_modifier;      // for shift,ctrl,alt accelerators
    int m_digits;          // # of digits to show
    bool m_snap;           // snap to increment
    wxString m_textformat; // print format for textctrl, ie. %.3f

private:
	DECLARE_DYNAMIC_CLASS(wxSpinReal)
	DECLARE_EVENT_TABLE()
	DECLARE_NO_COPY_CLASS(wxSpinReal)
};

#endif // wxUSE_SPINREAL

#endif // _WX_MSW_SPINREAL_H_


