// Scintilla source code edit control
/** @file LexAUCBLOGO.cxx
 ** Lexer for aUCBLogo.
 **/

#include <wx/string.h>

#ifdef _DEBUG
	#include <crtdbg.h>
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>

#include "Platform.h"

#include "PropSet.h"
#include "Accessor.h"
#include "StyleContext.h"
#include "KeyWords.h"
#include "Scintilla.h"
#include "SciLexer.h"

#define KEYWORD_BOXHEADER 1
#define KEYWORD_FOLDCONTRACTED 2

static inline bool IsAWordChar(const int ch) 
{	return isalnum(ch) || ch == '_' || ch == '?';
}

static inline bool IsAWordStart(const int ch) 
{	return isalnum(ch) || ch == '_' || ch == '?';
}

static inline bool IsStateComment(const int state) 
{	return ((state == SCE_LG_COMMENT) ||
	        (state == SCE_LG_COMMENTLINE));
}

static inline bool IsStateString(const int state) 
{	return ((state == SCE_LG_STRING) || (state == SCE_LG_VERBATIM));
}

extern wxString aUCBLogo_dynkeystring;
wxString aUCBLogo_dynkeystring;

static void ColouriseaUCBLogoDoc(unsigned int startPos, int length, 
											int initStyle, WordList *keywordlists[],
											Accessor &styler, bool caseSensitive) 
{
	WordList &keywords  = *keywordlists[0];
	WordList &keywords2 = *keywordlists[1];
	WordList &keywords3 = *keywordlists[2];
	WordList &keywords4 = *keywordlists[3];
	WordList &keywords5 = *keywordlists[4];
	WordList dynkeylist;
	if (caseSensitive)
		dynkeylist.Set(aUCBLogo_dynkeystring);
	else
		dynkeylist.Set(aUCBLogo_dynkeystring.Lower());

	// Do not leak onto next line
	if (initStyle == SCE_LG_STRINGEOL)
		initStyle = SCE_LG_DEFAULT;

	int chPrevNonWhite = ' ';
	int visibleChars = 0;
	char s[100];

/*	StyleContext sc2(0, styler.Length(), initStyle, styler);
	
	if (sc2.state==SCE_LG_COMMENT)
		while (sc2.state==SCE_LG_COMMENT && sc2.Less())
		{	sc2.Back();
		}
	else if (sc2.state==SCE_LG_STRING)
		while (sc2.state==SCE_LG_STRING && sc2.Less())
		{	sc2.Back();
		}
*/
	StyleContext sc2(startPos, length, initStyle, styler);
	while (sc2.More())
	{	if (sc2.state==SCE_LG_COMMENT)
		{	length=styler.Length()-startPos;
			break;
		}
		sc2.Forward();
	}
	StyleContext sc(startPos, length, initStyle, styler);

	for (; sc.More(); sc.Forward()) 
	{
		if (sc.atLineStart && (sc.state == SCE_LG_STRING)) 
			// Prevent SCE_LG_STRINGEOL from leaking back to previous line
			sc.SetState(SCE_LG_STRING);

		// Handle line continuation generically.
		if (sc.ch == '~') 
		{	if (sc.chNext == '\n' || sc.chNext == '\r') 
			{	sc.Forward();
				if (sc.ch == '\r' && sc.chNext == '\n') 
					sc.Forward();
				continue;
			}
		}

		// Determine if the current state should terminate.
		switch (sc.state)
		{
		case SCE_LG_OPERATOR:
			sc.SetState(SCE_LG_DEFAULT); break;
		
		case SCE_LG_NUMBER:
			if (!IsAWordChar(sc.ch)) 
				sc.SetState(SCE_LG_DEFAULT);
			break;

		case SCE_LG_IDENTIFIER:
			if (!IsAWordChar(sc.ch) || (sc.ch == '.')) 
			{	if (caseSensitive) 
					sc.GetCurrent(s, sizeof(s));
				else 
					sc.GetCurrentLowered(s, sizeof(s));
				
				if (keywords.InList(s)) 
					sc.ChangeState(SCE_LG_WORD);
				else if (keywords2.InList(s)) 
					sc.ChangeState(SCE_LG_WORD2);
				else if (keywords3.InList(s)) 
					sc.ChangeState(SCE_LG_WORD3);
				else if (keywords4.InList(s)) 
					sc.ChangeState(SCE_LG_WORD4);
				else if (keywords5.InList(s)) 
					sc.ChangeState(SCE_LG_WORD5);
				else if (dynkeylist.InList(s)) 
					sc.ChangeState(SCE_LG_WORD6);
				sc.SetState(SCE_LG_DEFAULT);
			}
			break;

		case SCE_LG_TO:
			char s[100];
			if (caseSensitive) 
				sc.GetCurrent(s, sizeof(s));
			else 
				sc.GetCurrentLowered(s, sizeof(s));
			
			sc.ChangeState(SCE_LG_TO);
			sc.ForwardSetState(SCE_LG_TODEF);
			break;
		
		case SCE_LG_TODEF:
			if (IsASpace(sc.ch))
			{	char s[100];
				if (caseSensitive) 
					sc.GetCurrent(s, sizeof(s));
				else 
					sc.GetCurrentLowered(s, sizeof(s));
				
				if (!dynkeylist.InList(s))
				{	aUCBLogo_dynkeystring+=wxString(s)+" ";
					dynkeylist.Clear();
					dynkeylist.Set(aUCBLogo_dynkeystring);
					sc.ChangeState(SCE_LG_WORD6);
				}
				sc.ChangeState(SCE_LG_WORD6);
				sc.SetState(SCE_LG_DEFAULT);
			}
			break;

		case SCE_LG_END:
			if (caseSensitive) 
				sc.GetCurrent(s, sizeof(s));
			else 
				sc.GetCurrentLowered(s, sizeof(s));
			
			sc.ChangeState(SCE_LG_END);
			sc.SetState(SCE_LG_DEFAULT);
			break;
		
		case SCE_LG_COMMENTLINE:
			if (sc.atLineEnd) 
			{	sc.SetState(SCE_LG_DEFAULT);
				visibleChars = 0;
			}
			break;

		case SCE_LG_STRING:
			if (sc.ch == '|')
			{	sc.SetState(SCE_LG_VERBATIM);
				sc.Forward();
				while (sc.ch != '|' && sc.More())
				{	sc.SetState(SCE_LG_VERBATIM);
					sc.Forward();
				}
				sc.SetState(SCE_LG_STRING);
			}
			else if (sc.ch == '\\') 
			{	sc.Forward();
				sc.SetState(SCE_LG_STRING);
			}else if (sc.atLineEnd || sc.ch == ' ' || sc.ch == '\t' 
				  ||  sc.ch == '(' || sc.ch == ')' 
				  ||  sc.ch == '[' || sc.ch == ']' ||  sc.ch == ';') 
				sc.SetState(SCE_LG_DEFAULT);
			break;

		case SCE_LG_VERBATIM:
			while (sc.ch != '|' && sc.More())
			{	sc.SetState(SCE_LG_VERBATIM);
				sc.Forward();
			}
			sc.SetState(SCE_LG_STRING);
			break;

		case SCE_LG_UUID:
			if (sc.ch == '\r' || sc.ch == '\n' || sc.ch == ')') 
				sc.SetState(SCE_LG_DEFAULT);
			break;

		case SCE_LG_COMMENT:
			{	int brakets=0;
				while (sc.More())
				{	if (sc.ch == '[')
						brakets++;
					else if (sc.ch == ']') 
					{	brakets--;
						if (brakets<0)
							break;
					}
					sc.SetState(SCE_LG_COMMENT);
					sc.Forward();
				}
				sc.ForwardSetState(SCE_LG_DEFAULT);
			}
			break;

		default: ;
		}

		// Determine if a new state should be entered.
		if (sc.state == SCE_LG_DEFAULT) 
		{	if (sc.atLineStart 
			&& (	sc.ch==' ' 
				|| sc.ch=='\t' 
				|| toupper(sc.ch)=='T' 
				|| toupper(sc.ch)=='B' 
				|| toupper(sc.ch)=='E'))
			{	int opos=sc.currentPos;
				while ((sc.ch==' ' || sc.ch=='\t') && sc.More())
					sc.Forward();
				if (sc.Match("to ")) 
				{	sc.SetState(SCE_LG_TO);
					sc.Forward();
				}
				else if (sc.Match("be ")) 
				{	sc.SetState(SCE_LG_TO);
					sc.Forward();
				}
				else if (sc.Match("end")) 
				{	sc.SetState(SCE_LG_END);
					sc.Forward();
					sc.Forward();
				}
				else
				{	if (opos >= 1)
						sc.currentPos=opos-1;
					sc.Forward();
				}
			}
			else if (sc.Match("comment")) 
			{	sc.SetState(SCE_LG_COMMENT);
				int brakets=0;
				while (sc.More())
				{	if (sc.ch == '[')
						brakets++;
					else if (sc.ch == ']') 
					{	brakets--;
						if (brakets<=0)
							break;
					}
					sc.SetState(SCE_LG_COMMENT);
					sc.Forward();
				}
				sc.ForwardSetState(SCE_LG_DEFAULT);
			}
			else if (IsADigit(sc.ch) || (sc.ch == '.' && IsADigit(sc.chNext))) 
				sc.SetState(SCE_LG_NUMBER);
			
			else if (IsAWordStart(sc.ch) || (sc.ch == '@') || (sc.ch == ':')) 
				sc.SetState(SCE_LG_IDENTIFIER);
			
			else if (sc.ch == ';' && sc.chPrev!='\\') 
				sc.SetState(SCE_LG_COMMENTLINE);
			
			else if (sc.ch == '\"') 
				sc.SetState(SCE_LG_STRING);
			
			else if (isoperator(static_cast<char>(sc.ch))) 
				sc.SetState(SCE_LG_OPERATOR);
		}

		if (sc.atLineEnd) 
		{	// Reset states to begining of colourise so no surprises
			// if different sets of lines lexed.
			chPrevNonWhite = ' ';
			visibleChars = 0;
		}
		if (!IsASpace(sc.ch)) 
		{	chPrevNonWhite = sc.ch;
			visibleChars++;
		}
	}
	sc.Complete();
}

static bool IsStreamCommentStyle(int style) 
{
	return style == SCE_LG_COMMENT ||
	       style == SCE_LG_COMMENTDOC ||
	       style == SCE_LG_COMMENTDOCKEYWORD ||
	       style == SCE_LG_COMMENTDOCKEYWORDERROR;
}

// Store both the current line's fold level and the next lines in the
// level store to make it easy to pick up with each increment
// and to make it possible to fiddle the current level for "} else {".

static void FoldNoBoxaUCBLogoDoc(unsigned int startPos, int length, 
											int initStyle, Accessor &styler) 
{
	bool foldComment = styler.GetPropertyInt("fold.comment") != 0;
	bool foldPreprocessor = styler.GetPropertyInt("fold.preprocessor") != 0;
	bool foldCompact = styler.GetPropertyInt("fold.compact", 1) != 0;
	bool foldAtElse = styler.GetPropertyInt("fold.at.else", 0) != 0;
	unsigned int endPos = startPos + length;
	int visibleChars = 0;
	int lineCurrent = styler.GetLine(startPos);
	int levelCurrent = SC_FOLDLEVELBASE;
	if (lineCurrent > 0)
		levelCurrent = styler.LevelAt(lineCurrent-1) >> 16;
	int levelMinCurrent = levelCurrent;
	int levelNext = levelCurrent;
	char chNext = styler[startPos];
	int styleNext = styler.StyleAt(startPos);
	int style = initStyle;

	for (unsigned int i = startPos; i < endPos; i++) 
	{
		char ch = chNext;
		chNext = styler.SafeGetCharAt(i + 1);
		int stylePrev = style;
		style = styleNext;
		styleNext = styler.StyleAt(i + 1);

		bool atEOL = (ch == '\r' && chNext != '\n') || (ch == '\n');

		if (foldComment && IsStreamCommentStyle(style)) 
		{	if (!IsStreamCommentStyle(stylePrev)) 
				levelNext++;
			else if (!IsStreamCommentStyle(styleNext) && !atEOL) 
				// Comments don't end at end of line and the next character may be unstyled.
				levelNext--;
		}
		if (foldComment && (style == SCE_LG_COMMENTLINE)) 
		{
			if ((ch == '/') && (chNext == '/')) 
			{
				char chNext2 = styler.SafeGetCharAt(i + 2);
				if (chNext2 == '{') 
					levelNext++;
				else if (chNext2 == '}') 
					levelNext--;
			}
		}
		if (style == SCE_LG_OPERATOR) 
		{	if (ch == '[') 
			{	// Measure the minimum before a '{' to allow
				// folding on "} else {"
				if (levelMinCurrent > levelNext) 
					levelMinCurrent = levelNext;
				
				levelNext++;
			}else if (ch == ']') 
				levelNext--;
			else if (ch == '{') 
			{	// Measure the minimum before a '{' to allow
				// folding on "} else {"
				if (levelMinCurrent > levelNext) 
					levelMinCurrent = levelNext;
				
				levelNext++;
			} else if (ch == '}') 
				levelNext--;
		}
		if (style == SCE_LG_TO) 
			levelNext++;
		if (style == SCE_LG_END)
			levelNext--;
		
		if (atEOL) 
		{	int levelUse = levelCurrent;
			if (foldAtElse) 
				levelUse = levelMinCurrent;
			
			int lev = levelUse | levelNext << 16;
			if (visibleChars == 0 && foldCompact)
				lev |= SC_FOLDLEVELWHITEFLAG;
			if (levelUse < levelNext)
				lev |= SC_FOLDLEVELHEADERFLAG;
			if (lev != styler.LevelAt(lineCurrent)) 
				styler.SetLevel(lineCurrent, lev);
			
			lineCurrent++;
			levelCurrent = levelNext;
			levelMinCurrent = levelCurrent;
			visibleChars = 0;
		}
		if (!isspacechar(ch))
			visibleChars++;
	}
}

static void FoldaUCBLogoDoc(unsigned int startPos, int length, int initStyle, 
							WordList *[], Accessor &styler) 
{
	FoldNoBoxaUCBLogoDoc(startPos, length, initStyle, styler);
}

static const char * const aUCBLogoWordLists[] = 
	{	"Structural keywords",
		"Primitives",
		"Library procedures",
		"Abbreviations",
		"Special Variables",
		0,
	};

static void ColouriseaUCBLogoDocSensitive(unsigned int startPos, 
														int length, int initStyle, 
														WordList *keywordlists[],
														Accessor &styler) 
{
	ColouriseaUCBLogoDoc(startPos, length, initStyle, 
		keywordlists, styler, true);
}

static void ColouriseaUCBLogoDocInsensitive(unsigned int startPos, 
														  int length, int initStyle, 
														  WordList *keywordlists[],
														  Accessor &styler) 
{
	ColouriseaUCBLogoDoc(startPos, length, initStyle, 
		keywordlists, styler, false);
}

LexerModule lmaUCBLogo      (SCLEX_AUCBLOGO      , 
							 ColouriseaUCBLogoDocSensitive, 
							 "lg"      , FoldaUCBLogoDoc, aUCBLogoWordLists);

LexerModule lmaUCBLogoNoCase(SCLEX_AUCBLOGONOCASE, 
							 ColouriseaUCBLogoDocInsensitive, 
							 "lgnocase", FoldaUCBLogoDoc, aUCBLogoWordLists);
